#include "e_lib.h"
#include "patch.h"

static void EndNode(int n, GEN *data, int CurCon, INTERNAL_MESH *mesh,
                    int *node, double *x1, double *y1, double *z1)
{
  int patchID, linkID, xpos, ypos, nin;

  nin = data->Connections[CurCon].Nodes[n].Node;

  /* the node is a referenced surface */
  if(nin == SURFACE_NODE)
  {
    patchID = data->Connections[CurCon].Nodes[n].ElemID-1;
    xpos    = data->Connections[CurCon].Nodes[n].Position[0];
    ypos    = data->Connections[CurCon].Nodes[n].Position[1];
    *node   = data->Patches[patchID].RegNodes[ypos][xpos].Node;
    if((*node == 0) || (*node == -1))
      e_GeneralError("Registered position is not filled !");
  }
  /* the starting node is a referenced link element */
  else if(nin == LINK_NODE)
  {
    linkID = data->Connections[CurCon].Nodes[n].ElemID-1;
    xpos   = data->Connections[CurCon].Nodes[n].Position[0];
    *node  = data->Connections[linkID].RegNodes[xpos].Node;
    if((*node == 0) || (*node == -1))
      e_GeneralError("Registered position is not filled !");
  }
  /* the starting node should be "generated" */
  else
  {
    /* store the new coordinates */
    mesh->Coords[mesh->NPoints][0] = data->Coords[nin-1][0];
    mesh->Coords[mesh->NPoints][1] = data->Coords[nin-1][1];
    mesh->Coords[mesh->NPoints][2] = data->Coords[nin-1][2];

    /* increment the number of nodes */
    mesh->NPoints++;

    *node = mesh->NPoints;
  }

  *x1 = mesh->Coords[(*node)-1][0];
  *y1 = mesh->Coords[(*node)-1][1];
  *z1 = mesh->Coords[(*node)-1][2];

}


void p_GenerateConnections(GEN *data, int CurCon, INTERNAL_MESH *mesh)
{
  int     node, endnode, type;
  int     i;
  double  wc, wt, x1, y1, z1, x2, y2, z2;
  int    *mat_vector;
  int     mat_type;

  /* set internal values according to the element type to generate */
  if(data->Connections[CurCon].ElementType == E_ELEM_TYPE_LINK1)
  {
    type = 0;
    if(data->DefMat)
    {
      mat_vector = mesh->MatLinks[0];
      mat_type = MAT_LINK1;
    }
  }
  else if(data->Connections[CurCon].ElementType == E_ELEM_TYPE_LINK2)
  {
    type = 1;
    if(data->DefMat)
    {
      mat_vector = mesh->MatLinks[1];
      mat_type = MAT_LINK2;
    }
  }
  else if(data->Connections[CurCon].ElementType == E_ELEM_TYPE_LINK3)
  {
    type = 2;
    if(data->DefMat)
    {
      mat_vector = mesh->MatLinks[2];
      mat_type = MAT_LINK3;
    }
  }
  else if(data->Connections[CurCon].ElementType == E_ELEM_TYPE_LINK4)
  {
    type = 3;
    if(data->DefMat)
    {
      mat_vector = mesh->MatLinks[3];
      mat_type = MAT_LINK4;
    }
  }
  /* geodesic strings are special !!! */
  else if(data->Connections[CurCon].ElementType == E_ELEM_TYPE_LINK5)
  {
    type = 4;
    
    if(data->DefMat)
    {
      if(data->Connections[CurCon].Division == 1)
        e_GeneralError("Geodesic string without division, but with material generation!",
                       CurCon+1);
      if(data->Connections[CurCon].Material != 0)
        e_GeneralError("Geodesic string (%d) cannot have user defined material!",
                       CurCon+1);
      mat_vector = mesh->MatLinks[4];
      mat_type = MAT_MID_LINK5;
    }
  }

  /* first end node */
  EndNode(0, data, CurCon, mesh, &node, &x1, &y1, &z1);

  /* beginning of the very first link */
  mesh->Links[type][mesh->NLinks[type]][0] = node;
  data->Connections[CurCon].RegNodes[0].Node = node;

  /* second end node */
  EndNode(1, data, CurCon, mesh, &endnode, &x2, &y2, &z2);

  /* determine total weight of the divisions, so they can be normalised */
  wt = 0.0;
  if(data->Connections[CurCon].Division > 1)
  {
    for(i = 0; i < data->Connections[CurCon].Division; i++)
    {
      wt += data->Connections[CurCon].DivWeight[i];
    }
  }

  /* intermediate nodes and links */
  wc = 0.0;
  for(i = 0; i < data->Connections[CurCon].Division-1; i++)
  {
    wc += data->Connections[CurCon].DivWeight[i];
    mesh->Coords[mesh->NPoints][0] = x1+(x2-x1)*wc/wt;
    mesh->Coords[mesh->NPoints][1] = y1+(y2-y1)*wc/wt;
    mesh->Coords[mesh->NPoints][2] = z1+(z2-z1)*wc/wt;
    mesh->NPoints++;

    mesh->Links[type][mesh->NLinks[type]][1] = mesh->NPoints;
    data->Connections[CurCon].RegNodes[i+1].Node = mesh->NPoints;
    
    /* local (user-defined) material, overwrite the default */
    if(   data->DefMat
       && data->Connections[CurCon].Material)
    {
      mesh->MatLinks[type][mesh->NLinks[type]] =
        data->Connections[CurCon].Material + DEFAULT_N_MATERIAL-1;
    }
    
    mesh->NLinks[type]++;
    
    /* this node is the beginning of the next link */
    mesh->Links[type][mesh->NLinks[type]][0] = mesh->NPoints;
  }

  /* the end node of the last link */
  mesh->Links[type][mesh->NLinks[type]][1] = endnode;
  data->Connections[CurCon].RegNodes[data->Connections[CurCon].Division].Node = 
    endnode;

  /* local (user-defined) material for the last link */
  if(   (data->DefMat)
     && (data->Connections[CurCon].Material) )
  {
    mesh->MatLinks[type][mesh->NLinks[type]] =
      data->Connections[CurCon].Material + DEFAULT_N_MATERIAL-1;
  }

  mesh->NLinks[type]++;

}



