#include "e_lib.h"
#include "intersect.h"


#define ZERO 1e-6


static int p_CheckNodeSides(double px, double py, double nx, double ny,
                            double s1x, double s1y, double s2x, double s2y);


/* when it returns NO_CROSS then there is a problem with the two sections,
   if there is no error then it returns CROSS */

int e_2DSegmentIntersect(double px, double py, double nx, double ny,
                         double s1x, double s1y, double s2x, double s2y,
                         double *x_sect, double *y_sect)
{
  double a, b, c, s;
  int type;

  a = (s2y - s1y) * (nx - px)   - (s2x - s1x) * (ny - py);
  b = (s1x - px)  * (ny - py)   - (s1y - py)  * (nx - px);
  c = (s1x - px)  * (s2y - s1y) - (s1y - py)  * (s2x - s1x);

  if(E_ABS(a) < ZERO)
    return PARALLEL;

  if(a > 0)
  {
    if((b >= 0 && b <= a) &&
       (c >= 0 && c <= a))
      type = CROSS;
    else
      type = NO_CROSS;
  }
  else
  {
    if((b <= 0 && b >= a) &&
       (c <= 0 && c >= a))
      type = CROSS;
    else
      type = NO_CROSS;
  }

  /* probably intersection found */
  if(type == CROSS)
  {
    /* checking for the case when the nodes of the lines coincide */
    if(   E_ABS(b - a) < ZERO 
       || E_ABS(c - a) < ZERO
       || E_ABS(c)     < ZERO )
    {
      type = NO_CROSS;
      return type;
    }
    if(   (E_ABS(b) < ZERO)
       && (p_CheckNodeSides(px, py, nx, ny, s1x, s1y, s2x, s2y) >= 0) )
    {
      type = NO_CROSS;
      return type;
    }
    s = b / a;
    if(E_ABS(s1x - s2x) < ZERO)
      *x_sect = s1x;
    else
      *x_sect = s1x + (s2x - s1x) * s;

    if(E_ABS(s1y - s2y) < ZERO)
      *y_sect = s1y;
    else
      *y_sect = s1y + (s2y - s1y) * s;
  }
  return type;
}


static int p_CheckNodeSides(double px, double py, double nx, double ny,
                            double s1x, double s1y, double s2x, double s2y)
{
  double dx, dy, dx1, dy1, dx2, dy2, p1, p2;

  dx = nx - px;
  dy = ny - py;
  dx1 = s1x - px;
  dy1 = s1y - py;
  dx2 = s2x - nx;
  dy2 = s2y - ny;

  p1 = dx * dy1 - dy * dx1;
  p2 = dx * dy2 - dy * dx2;

  if((E_ABS(p1) < ZERO) || (E_ABS(p2) < ZERO))
    return 0;
  else
  {
    if((p1 > 0 && p2 < 0) || (p1 < 0 && p2 > 0))
      return -1;
    else
      return 1;
  }
}
