#include "e_lib.h"
#include "patch.h"
#include "intersect.h"

#define APPROX_ZERO  1e-4



void p_GenerateTriangPatch(GEN *data, int CurPatch, 
                           INTERNAL_MESH *mesh)
{
  double x1, y1, z1, x2, y2, z2, x3, y3, z3;
  double total_weight;
  int    idivx, ix, iy, nxdiv, jj;
  double x1norm, x2norm, x3norm;
  double ex1sp, ex2sp, et1sp, et2sp, ep1sp, ep2sp;
  int    patchID, linkID, xpos, ypos;
  double xn1, yn1, zn1, xn2, yn2, zn2, xn3, yn3, zn3;
  double xn4, yn4, zn4, xn5, yn5, zn5, xn6, yn6, zn6;
  double xc1, yc1, zc1, xc2, yc2, zc2, xc3, yc3, zc3;
  int  **Triangles, *NTriangles, *MatTriang;
  int    apoint, bpoint, cpoint;
  int    il, type_x, type_y, reg_y, diag;
  int    align1 = FALSE, align2 = FALSE, align3 = FALSE;

  /* what kind of element to generate */
  if(data->Patches[CurPatch].ElementType == E_ELEM_TYPE_TRIANG1)
  {
    Triangles = mesh->Triangles[0];
    NTriangles = mesh->NTriangles;
    if(data->DefMat)
      MatTriang = mesh->MatTriangles[0];
  }
  else if(data->Patches[CurPatch].ElementType == E_ELEM_TYPE_TRIANG3)
  {
    Triangles = mesh->Triangles[1];
    NTriangles = mesh->NTriangles+1;
    if(data->DefMat)
      MatTriang = mesh->MatTriangles[1];
  }
  else if(data->Patches[CurPatch].ElementType == E_ELEM_TYPE_TRIANG4)
  {
    Triangles = mesh->Triangles[2];
    NTriangles = mesh->NTriangles+2;
    if(data->DefMat)
      MatTriang = mesh->MatTriangles[2];
  }
  else if(data->Patches[CurPatch].ElementType == E_ELEM_TYPE_TRIANG5)
  {
    Triangles = mesh->Triangles[3];
    NTriangles = mesh->NTriangles+3;
    if(data->DefMat)
      MatTriang = mesh->MatTriangles[3];
  }
  else
  {
    e_GeneralError("invalid element type to generate in a triangle surface");
  }

  
  /* determine coordinates of the corner nodes, 
     maybe a reference */
  DetNodeCoord(data, mesh, data->Patches[CurPatch].Nodes+0,
               &x1, &y1, &z1);
  DetNodeCoord(data, mesh, data->Patches[CurPatch].Nodes+1,
               &x2, &y2, &z2);
  DetNodeCoord(data, mesh, data->Patches[CurPatch].Nodes+2,
               &x3, &y3, &z3);

  /* check that the first side of the triangle definition should be
     aligned with which side */
  for(il = 0; il < data->Patches[CurPatch].LinkSides[0]; il++)
  {
    if(data->Patches[CurPatch].LinkType[0][il] == 5)
    {
      align1 = TRUE;
      break;
    }
  }

  for(il = 0; il < data->Patches[CurPatch].LinkSides[1]; il++)
  {
    if(data->Patches[CurPatch].LinkType[1][il] == 5)
    {
      align2 = TRUE;
      break;
    }
  }

  for(il = 0; il < data->Patches[CurPatch].LinkSides[2]; il++)
  {
    if(data->Patches[CurPatch].LinkType[2][il] == 5)
    {
      align3 = TRUE;
      break;
    }
  }

  if(   (align1 == TRUE && align2 == TRUE)
     || (align2 == TRUE && align3 == TRUE)
     || (align3 == TRUE && align1 == TRUE) )
  {
    e_GeneralWarning("Surface %d has LINK 5 in more than one direction", CurPatch+1);
  }

  /* user specified weights are only relative, scale them to 1 */
  /* first side */
  total_weight = 0.0;
  for(idivx = 0; idivx < data->Patches[CurPatch].XDivision; idivx++)
  {
    if(data->Patches[CurPatch].XDivWeight[idivx] == 0.0)
      data->Patches[CurPatch].XDivWeight[idivx] = 1.0;
    total_weight += data->Patches[CurPatch].XDivWeight[idivx];
  }
  x1norm = 1.0 / total_weight;
  /* second side */
  total_weight = 0.0;
  for(idivx = 0; idivx < data->Patches[CurPatch].XDivision; idivx++)
  {
    if(data->Patches[CurPatch].YDivWeight[idivx] == 0.0)
      data->Patches[CurPatch].YDivWeight[idivx] = 1.0;
    total_weight += data->Patches[CurPatch].YDivWeight[idivx];
  }
  x2norm = 1.0 / total_weight;
  /* third side */
  total_weight = 0.0;
  for(idivx = 0; idivx < data->Patches[CurPatch].XDivision; idivx++)
  {
    if(data->Patches[CurPatch].TDivWeight[idivx] == 0.0)
      data->Patches[CurPatch].TDivWeight[idivx] = 1.0;
    total_weight += data->Patches[CurPatch].TDivWeight[idivx];
  }
  x3norm = 1.0 / total_weight;

  /* no of nodes to generate or use */
  nxdiv = data->Patches[CurPatch].XDivision+1;

  /* y direction */  
  et1sp = 0;
  ep1sp = 0;
  for(iy = 0; iy < nxdiv; iy++)
  {
    /* local coordinates */
    ex1sp = 0;
    et2sp = 0;

    ex2sp = 0;
    ep2sp = 0;
    if(iy < nxdiv-1)
    {
      for(jj = 0; jj < iy; jj++)
      {
        ex2sp += x1norm * data->Patches[CurPatch].XDivWeight[jj];
        ep2sp += x3norm * data->Patches[CurPatch].TDivWeight[nxdiv-2-jj];
      }
    }

    /* registered link, horizontal */
    reg_y = FALSE;
    for(il = 0; il < data->Patches[CurPatch].LinkSides[0]; il++)
    {
      if(iy == data->Patches[CurPatch].LinkDiv[0][il])
      {
        type_y = data->Patches[CurPatch].LinkType[0][il]-1;
        reg_y  = TRUE;
      }
    }

    /* x direction */
    for(ix = 0; ix < nxdiv-iy; ix++)
    {
      /* the node is either referenced or generated */
      if(data->Patches[CurPatch].RegNodes[iy][ix].Node == SURFACE_NODE)
      {
        patchID = data->Patches[CurPatch].RegNodes[iy][ix].ElemID-1;
        xpos    = data->Patches[CurPatch].RegNodes[iy][ix].Position[0];
        ypos    = data->Patches[CurPatch].RegNodes[iy][ix].Position[1];
        data->Patches[CurPatch].RegNodes[iy][ix].Node =
          data->Patches[patchID].RegNodes[ypos][xpos].Node;
      }
      else if(data->Patches[CurPatch].RegNodes[iy][ix].Node == LINK_NODE)
      {
        linkID = data->Patches[CurPatch].RegNodes[iy][ix].ElemID-1;
        xpos   = data->Patches[CurPatch].RegNodes[iy][ix].Position[0];
        data->Patches[CurPatch].RegNodes[iy][ix].Node =
          data->Connections[linkID].RegNodes[xpos].Node;
      }
      else
      {
        /* generate the three coordinates */
        /* special cases, nodes on triangle sides */
        if(iy == 0)
        {
          mesh->Coords[mesh->NPoints][0] = ex1sp * x2 + (1-ex1sp) * x1;
          mesh->Coords[mesh->NPoints][1] = ex1sp * y2 + (1-ex1sp) * y1;
          mesh->Coords[mesh->NPoints][2] = ex1sp * z2 + (1-ex1sp) * z1;
        }
        else if(ix == 0)
        {
          mesh->Coords[mesh->NPoints][0] = ep1sp * x3 + (1-ep1sp) * x1;
          mesh->Coords[mesh->NPoints][1] = ep1sp * y3 + (1-ep1sp) * y1;
          mesh->Coords[mesh->NPoints][2] = ep1sp * z3 + (1-ep1sp) * z1;
        }
        else if(ix == nxdiv-iy-1)
        {
          mesh->Coords[mesh->NPoints][0] = et1sp * x3 + (1-et1sp) * x2;
          mesh->Coords[mesh->NPoints][1] = et1sp * y3 + (1-et1sp) * y2;
          mesh->Coords[mesh->NPoints][2] = et1sp * z3 + (1-et1sp) * z2;
        }
        /* general case, internal nodes */
        else
        {
          /* coordinates on the sides of the triangle, 
             6 point will determine three lines, 
             the intersection of each lines are determined, 
             this forms a triangle and finally the centre of the triangle
             is the new node */
          xn1 = ex1sp * x2 + (1-ex1sp) * x1;
          yn1 = ex1sp * y2 + (1-ex1sp) * y1;
          zn1 = ex1sp * z2 + (1-ex1sp) * z1;

          xn2 = ex2sp * x2 + (1-ex2sp) * x1;
          yn2 = ex2sp * y2 + (1-ex2sp) * y1;
          zn2 = ex2sp * z2 + (1-ex2sp) * z1;

          xn3 = et1sp * x3 + (1-et1sp) * x2;
          yn3 = et1sp * y3 + (1-et1sp) * y2;
          zn3 = et1sp * z3 + (1-et1sp) * z2;

          xn4 = et2sp * x2 + (1-et2sp) * x3;
          yn4 = et2sp * y2 + (1-et2sp) * y3;
          zn4 = et2sp * z2 + (1-et2sp) * z3;

          xn5 = ep2sp * x3 + (1-ep2sp) * x1;
          yn5 = ep2sp * y3 + (1-ep2sp) * y1;
          zn5 = ep2sp * z3 + (1-ep2sp) * z1;

          xn6 = ep1sp * x3 + (1-ep1sp) * x1;
          yn6 = ep1sp * y3 + (1-ep1sp) * y1;
          zn6 = ep1sp * z3 + (1-ep1sp) * z1;
          
          /* intersection of two lines */
          if(e_2DSegmentIntersect(xn1, yn1, xn4, yn4,
                                  xn2, yn2, xn5, yn5, &xc1, &yc1) != CROSS)
          {
            /* in case of a triangle it is impossible, 
               but some problem can occur, someone never know */
            e_GeneralError("internal error: no intersection 1");
          }
          /* z coordinate is determined here */
          /* check for horizontal line, in that case these would be a
             division by zero */
          if(E_ABS(yn4-yn1) > APPROX_ZERO)
          {
            zc1 = zn1 + (yc1-yn1)*(zn4-zn1)/(yn4-yn1);
          }
          else if(E_ABS(xn4-xn1) > APPROX_ZERO)
          {
            zc1 = zn1 + (xc1-xn1)*(zn4-zn1)/(xn4-xn1);
          }
          else
          {
            e_GeneralError("internal error: zero length 1");
          }

          if(e_2DSegmentIntersect(xn1, yn1, xn4, yn4,
                                  xn3, yn3, xn6, yn6, &xc2, &yc2) != CROSS)
          {
            e_GeneralError("internal error: no intersection 2");
          }
          if(E_ABS(yn4-yn1) > APPROX_ZERO)
          {
            zc2 = zn1 + (yc2-yn1)*(zn4-zn1)/(yn4-yn1);
          }
          else if(E_ABS(xn4-xn1) > APPROX_ZERO)
          {
            zc2 = zn1 + (xc2-xn1)*(zn4-zn1)/(xn4-xn1);
          }
          else
          {
            e_GeneralError("internal error: zero length 2");
          }

          if(e_2DSegmentIntersect(xn2, yn2, xn5, yn5,
                                  xn3, yn3, xn6, yn6, &xc3, &yc3) != CROSS)
          {
            e_GeneralError("internal error: no intersection 3");
          }
          if(E_ABS(yn5-yn2) > APPROX_ZERO)
          {
            zc3 = zn2 + (yc3-yn2)*(zn5-zn2)/(yn5-yn2);
          }
          else if(E_ABS(xn5-xn2) > APPROX_ZERO)
          {
            zc3 = zn2 + (xc3-xn2)*(zn5-zn2)/(xn5-xn2);
          }
          else
          {
            e_GeneralError("internal error: zero length 3");
          }
          
          /* centre of the triangle */
          mesh->Coords[mesh->NPoints][0] = (xc1+xc2+xc3) / 3.0;
          mesh->Coords[mesh->NPoints][1] = (yc1+yc2+yc3) / 3.0;
          mesh->Coords[mesh->NPoints][2] = (zc1+zc2+zc3) / 3.0;
        }
        mesh->NPoints++;

        data->Patches[CurPatch].RegNodes[iy][ix].Node = mesh->NPoints;
      }

      /* registered links */
      /* parallel with first side, horizontal */
      if(ix && reg_y)
      {
        mesh->Links[type_y][mesh->NLinks[type_y]][0] =
          data->Patches[CurPatch].RegNodes[iy][ix-1].Node;
        mesh->Links[type_y][mesh->NLinks[type_y]][1] =
          data->Patches[CurPatch].RegNodes[iy][ix].Node;
        mesh->NLinks[type_y]++;
      }
      
      /* parallel with third side */
      if(iy)
      {
        for(il = 0; il < data->Patches[CurPatch].LinkSides[2]; il++)
        {
          if(ix == data->Patches[CurPatch].LinkDiv[2][il])
          {
            type_x = data->Patches[CurPatch].LinkType[2][il]-1;

            mesh->Links[type_x][mesh->NLinks[type_x]][0] =
              data->Patches[CurPatch].RegNodes[iy][ix].Node;
            mesh->Links[type_x][mesh->NLinks[type_x]][1] =
              data->Patches[CurPatch].RegNodes[iy-1][ix].Node;
            mesh->NLinks[type_x]++;
          }
        }
      }

      /* parallel with second side */
      if(iy)
      {
        for(il = 0; il < data->Patches[CurPatch].LinkSides[1]; il++)
        {
          diag = data->Patches[CurPatch].XDivision-iy+ix;
          if(ix+iy == data->Patches[CurPatch].XDivision-
                      data->Patches[CurPatch].LinkDiv[1][il])
          {
            type_x = data->Patches[CurPatch].LinkType[1][il]-1;

            mesh->Links[type_x][mesh->NLinks[type_x]][0] =
              data->Patches[CurPatch].RegNodes[iy][ix].Node;
            mesh->Links[type_x][mesh->NLinks[type_x]][1] =
              data->Patches[CurPatch].RegNodes[iy-1][ix+1].Node;
            mesh->NLinks[type_x]++;
          }
        }
      }

      /* generate the triangle elements */
      if(iy)
      {
        apoint = data->Patches[CurPatch].RegNodes[iy  ][ix  ].Node;
        bpoint = data->Patches[CurPatch].RegNodes[iy-1][ix  ].Node;
        cpoint = data->Patches[CurPatch].RegNodes[iy-1][ix+1].Node;
        
        if(align1 == TRUE)
        {
          Triangles[*NTriangles][0] = apoint;
          Triangles[*NTriangles][1] = bpoint;
          Triangles[*NTriangles][2] = cpoint;
        }
        else if(align2 == TRUE)
        {
          Triangles[*NTriangles][0] = bpoint;
          Triangles[*NTriangles][1] = cpoint;
          Triangles[*NTriangles][2] = apoint;
        }
        else
        {
          Triangles[*NTriangles][0] = cpoint;
          Triangles[*NTriangles][1] = apoint;
          Triangles[*NTriangles][2] = bpoint;
        }

        /* user defined material */
        if(   data->DefMat
           && data->Patches[CurPatch].Material)
        {
          MatTriang[*NTriangles] =
            data->Patches[CurPatch].Material +
            DEFAULT_N_MATERIAL-1;
        }
        (*NTriangles)++;
        
        if(ix)
        {
          apoint = data->Patches[CurPatch].RegNodes[iy  ][ix  ].Node;
          bpoint = data->Patches[CurPatch].RegNodes[iy  ][ix-1].Node;
          cpoint = data->Patches[CurPatch].RegNodes[iy-1][ix  ].Node;
          
          if(align1 == TRUE)
          {
            Triangles[*NTriangles][0] = cpoint;
            Triangles[*NTriangles][1] = apoint;
            Triangles[*NTriangles][2] = bpoint;
          }
          else if(align2 == TRUE)
          {
            Triangles[*NTriangles][0] = apoint;
            Triangles[*NTriangles][1] = bpoint;
            Triangles[*NTriangles][2] = cpoint;
          }
          else
          {
            Triangles[*NTriangles][0] = bpoint;
            Triangles[*NTriangles][1] = cpoint;
            Triangles[*NTriangles][2] = apoint;
          }

          /* user defined material */
          if(   data->DefMat
             && data->Patches[CurPatch].Material)
          {
            MatTriang[*NTriangles] =
              data->Patches[CurPatch].Material +
              DEFAULT_N_MATERIAL-1;
          }
          (*NTriangles)++;

        }
      }

      if(ix < nxdiv-1)
        ex1sp += x1norm * data->Patches[CurPatch].XDivWeight[ix];

      if(ix < nxdiv-2)
      {
        et2sp += x2norm * data->Patches[CurPatch].YDivWeight[nxdiv-1-ix-1];
      }

      if(jj < nxdiv-2)
      {
        ex2sp += x1norm * data->Patches[CurPatch].XDivWeight[jj];
        ep2sp += x3norm * data->Patches[CurPatch].TDivWeight[nxdiv-2-jj];
        jj++;
      }
    }

    if(iy < nxdiv-1)
    {
      et1sp += x2norm * data->Patches[CurPatch].YDivWeight[iy];
      ep1sp += x3norm * data->Patches[CurPatch].TDivWeight[nxdiv-1-iy-1];
    }

  }

}





