#include "e_lib.h"
#include "adjmemb.h"
#include "getvertices.h"

static void AllocElemData(int start, int n, int nsides, ADJDATA *adata);
static int FindSideInElem(MESH *mesh, int index, int n1, int n2);
static void Set2DElemMatrix(MESH *mesh, int *no_node_elem, int **node_elem,
                            int maxadjelem, int NSides, int NElements, 
                            int **Vertices, int StartIndex,
                            int *Sides, int **Degree, 
                            int ***Adjacency, int ***AdjacentSide);
static void Set1DElemMatrix(MESH *mesh, int *no_node_elem, int **node_elem,
                            int NElements, int **Vertices, int StartIndex,
                            int *Sides, int **Degree, 
                            int ***Adjacency, int ***AdjacentSide);

/* ------------------------------------------------------------------------------ */





void Set2DElemAdjacency(MESH *mesh, ADJDATA **data, 
                        int *no_node_elem, int **node_elem, 
                        int flag)
{
  int maxadj;

  if(mesh->NElemsTriang1 +
     mesh->NElemsTriang3 +
     mesh->NElemsTriang4 +
     mesh->NElemsTriang5 +
     mesh->NElemsQuad1 +
     mesh->NElemsLink1 +
     mesh->NElemsLink2 +
     mesh->NElemsLink3 +
     mesh->NElemsLink4 +
     mesh->NElemsLink5   != mesh->TotalNElems)
  {
    e_GeneralError("Set2DElemAdjacency: Unsupported element type !");
  }

  /* setup the data structures */
  *data = (ADJDATA *)e_calloc(1, sizeof(ADJDATA) );
  
  (*data)->Sides        = e_alloc_i_vect(mesh->TotalNElems);
  (*data)->Degree       = e_init_frag_i_matr(mesh->TotalNElems);
  (*data)->Adjacency    = e_init_frag_i_3D_matr(mesh->TotalNElems);
  (*data)->AdjacentSide = e_init_frag_i_3D_matr(mesh->TotalNElems);

  AllocElemData(mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1]-1, 
                mesh->NElemsTriang1, E_NEPE_TRIANG1, *data);

  AllocElemData(mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3]-1, 
                mesh->NElemsTriang3, E_NEPE_TRIANG3, *data);

  AllocElemData(mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4]-1, 
                mesh->NElemsTriang4, E_NEPE_TRIANG4, *data);

  AllocElemData(mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5]-1, 
                mesh->NElemsTriang5, E_NEPE_TRIANG5, *data);

  AllocElemData(mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1]-1, 
                mesh->NElemsQuad1, E_NEPE_QUAD1, *data);

  if(flag & ADJ_LINK)
  {
    AllocElemData(mesh->FirstNetIndex[E_ELEM_TYPE_LINK1]-1, 
                  mesh->NElemsLink1, 1, *data);
    AllocElemData(mesh->FirstNetIndex[E_ELEM_TYPE_LINK2]-1, 
                  mesh->NElemsLink2, 1, *data);
    AllocElemData(mesh->FirstNetIndex[E_ELEM_TYPE_LINK3]-1, 
                  mesh->NElemsLink3, 1, *data);
    AllocElemData(mesh->FirstNetIndex[E_ELEM_TYPE_LINK4]-1, 
                  mesh->NElemsLink4, 1, *data);
    AllocElemData(mesh->FirstNetIndex[E_ELEM_TYPE_LINK5]-1, 
                  mesh->NElemsLink5, 1, *data);
  }

  /* any restriction on the number of adjacent elements */
  maxadj = 0;
  if(flag & ADJ_FLAT)
  {
    maxadj = 1;
  }

  Set2DElemMatrix(mesh, no_node_elem, node_elem, maxadj,
                  E_NEPE_TRIANG1, mesh->NElemsTriang1, mesh->NodesTriang1,
                  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1],
                  (*data)->Sides, (*data)->Degree, 
                  (*data)->Adjacency, (*data)->AdjacentSide);

  Set2DElemMatrix(mesh, no_node_elem, node_elem, maxadj,
                  E_NEPE_TRIANG3, mesh->NElemsTriang3, mesh->NodesTriang3,
                  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3],
                  (*data)->Sides, (*data)->Degree, 
                  (*data)->Adjacency, (*data)->AdjacentSide);

  Set2DElemMatrix(mesh, no_node_elem, node_elem, maxadj,
                  E_NEPE_TRIANG4, mesh->NElemsTriang4, mesh->NodesTriang4,
                  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4],
                  (*data)->Sides, (*data)->Degree, 
                  (*data)->Adjacency, (*data)->AdjacentSide);

  Set2DElemMatrix(mesh, no_node_elem, node_elem, maxadj,
                  E_NEPE_TRIANG5, mesh->NElemsTriang5, mesh->NodesTriang5,
                  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5],
                  (*data)->Sides, (*data)->Degree, 
                  (*data)->Adjacency, (*data)->AdjacentSide);

  Set2DElemMatrix(mesh, no_node_elem, node_elem, maxadj,
                  E_NEPE_QUAD1, mesh->NElemsQuad1, mesh->NodesQuad1,
                  mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1],
                  (*data)->Sides, (*data)->Degree, 
                  (*data)->Adjacency, (*data)->AdjacentSide);


  if(flag & ADJ_LINK)
  {
    Set1DElemMatrix(mesh, no_node_elem, node_elem,
                    mesh->NElemsLink1, mesh->NodesLink1, 
                    mesh->FirstNetIndex[E_ELEM_TYPE_LINK1],
                    (*data)->Sides, (*data)->Degree, 
                    (*data)->Adjacency, (*data)->AdjacentSide);

    Set1DElemMatrix(mesh, no_node_elem, node_elem,
                    mesh->NElemsLink2, mesh->NodesLink2, 
                    mesh->FirstNetIndex[E_ELEM_TYPE_LINK2],
                    (*data)->Sides, (*data)->Degree, 
                    (*data)->Adjacency, (*data)->AdjacentSide);

    Set1DElemMatrix(mesh, no_node_elem, node_elem,
                    mesh->NElemsLink3, mesh->NodesLink3, 
                    mesh->FirstNetIndex[E_ELEM_TYPE_LINK3],
                    (*data)->Sides, (*data)->Degree, 
                    (*data)->Adjacency, (*data)->AdjacentSide);

    Set1DElemMatrix(mesh, no_node_elem, node_elem,
                    mesh->NElemsLink4, mesh->NodesLink4, 
                    mesh->FirstNetIndex[E_ELEM_TYPE_LINK4],
                    (*data)->Sides, (*data)->Degree, 
                    (*data)->Adjacency, (*data)->AdjacentSide);

    Set1DElemMatrix(mesh, no_node_elem, node_elem,
                    mesh->NElemsLink5, mesh->NodesLink5, 
                    mesh->FirstNetIndex[E_ELEM_TYPE_LINK5],
                    (*data)->Sides, (*data)->Degree, 
                    (*data)->Adjacency, (*data)->AdjacentSide);

  }

}


static void AllocElemData(int start, int n, int nsides, ADJDATA *adata)
{
  int i;
  
  for(i = 0; i < n; i++)
  {
    adata->Sides[start+i]        = nsides;
    adata->Degree[start+i]       = e_alloc_i_vect(nsides);
    adata->Adjacency[start+i]    = e_init_frag_i_matr(nsides);
    adata->AdjacentSide[start+i] = e_init_frag_i_matr(nsides);
  }
}


static int FindSideInElem(MESH *mesh, int index, int n1, int n2)
{
  int nver, *ver;
  int ii;
  int as, ax, a1, a2;

  /* find the side defined by 'n1' and 'n2' in the 'index' element */
  e_GetElementVertices(index, mesh, &nver, &ver);
  as = 0;
  for(ii = 0; ii < nver; ii++)
  {
    ax = nver - 1 - ii;
    a1 = ver[as]-1;
    a2 = ver[ax]-1;
    if(   (n1 == a1 && n2 == a2)
       || (n1 == a2 && n2 == a1) )
    {
      break;
    }
  
    as = ax;
  }
  /* sanity check */
  if(ii == nver)
  {
    e_GeneralError("FindSideInElem: internal error !");
  }

  return(ax);
}


static void Set2DElemMatrix(MESH *mesh, int *no_node_elem, int **node_elem,
                            int maxadjelem, int NSides, int NElements, 
                            int **Vertices, int StartIndex,
                            int *Sides, int **Degree, 
                            int ***Adjacency, int ***AdjacentSide)
{
  int i, i_node, j, k, is, index, ix;
  int ElementIndex, ns, nx, n1, n2, no_link;

  no_link = mesh->NElemsLink1 + mesh->NElemsLink2 + mesh->NElemsLink3 +
            mesh->NElemsLink4 + mesh->NElemsLink5;

  for(i = 0; i < NElements; i++)
  {
    ElementIndex = StartIndex - 1 + i;

    /* search for other triangle elements */
    ns = 0;
    for(i_node = 0; i_node < NSides; i_node++) /* there are n sides to check */
    {
      nx = NSides - 1 - i_node;
      n1 = Vertices[i][ns]-1;  /* vertices of the side */
      n2 = Vertices[i][nx]-1;

      /* check all possible combination of elements */
      for(j = 0; j < no_node_elem[n1]; j++)
      {
        for(k = 0; k < no_node_elem[n2]; k++)
        {
          /* the element contains the two nodes */
          if(   (node_elem[n1][j] > ElementIndex+1)       /* relation is processed only once */
             && (node_elem[n1][j] == node_elem[n2][k])    /* same element at the two node */
             && (node_elem[n1][j] > no_link)              /* it is not a link */
             && (node_elem[n1][j] != ElementIndex+1)   )  /* it is not itself */
          {
            /* adjacent element and its side */
            index = node_elem[n1][j]-1;
            ix = FindSideInElem(mesh, index, n1, n2);
            
            /* check whether we already stored it with 'ElementIndex' */
            for(is = 0; is < Degree[ElementIndex][nx]; is++)
            {
              if(Adjacency[ElementIndex][nx][is] == node_elem[n1][j])
                break;
            }
            /* store it */
            if(is == Degree[ElementIndex][nx])
            {
              /* store the element */
              Adjacency[ElementIndex][nx] = 
                e_incr_i_vect(Degree[ElementIndex][nx], 
                              Adjacency[ElementIndex][nx], 1);
              Adjacency[ElementIndex][nx][Degree[ElementIndex][nx]] = index;
              
              /* store the side */
              AdjacentSide[ElementIndex][nx] = 
                e_incr_i_vect(Degree[ElementIndex][nx], 
                              AdjacentSide[ElementIndex][nx], 1);

              AdjacentSide[ElementIndex][nx][Degree[ElementIndex][nx]] = ix;
              
              if(   maxadjelem > 0
                 && maxadjelem == Degree[ElementIndex][nx])
              {
                e_GeneralError("Set2DElemMatrix: maxadjelem exceeded !");
              }
              Degree[ElementIndex][nx]++;
            }

            /* check whether we already stored it with 'index' */
            for(is = 0; is < Degree[index][ix]; is++)
            {
              if(Adjacency[index][ix][is] == ElementIndex)
                break;
            }
            /* store it */
            if(is == Degree[index][ix])
            {
              /* store the element */
              Adjacency[index][ix] = 
                e_incr_i_vect(Degree[index][ix], Adjacency[index][ix], 1);
              Adjacency[index][ix][Degree[index][ix]] = ElementIndex;
              
              /* store the side */
              AdjacentSide[index][ix] = 
                e_incr_i_vect(Degree[index][ix], AdjacentSide[index][ix], 1);

              AdjacentSide[index][ix][Degree[index][ix]] = nx;

              if(   maxadjelem > 0
                 && maxadjelem == Degree[index][nx])
              {
                e_GeneralError("Set2DElemMatrix: maxadjelem exceeded !");
              }
              Degree[index][ix]++;

            }
          }
        }
      }
      
      ns = nx;
    }
  }
}


/* This function should be called after that the adjacency matrix has been 
   assembled for the 2D elements */
static void Set1DElemMatrix(MESH *mesh, int *no_node_elem, int **node_elem,
                            int NElements, int **Vertices, int StartIndex,
                            int *Sides, int **Degree, 
                            int ***Adjacency, int ***AdjacentSide)
{
  int   i, j, k, ElementIndex, index, ix;
  int   no_link, n1, n2;

  no_link = mesh->NElemsLink1 + mesh->NElemsLink2 + mesh->NElemsLink3 +
            mesh->NElemsLink4 + mesh->NElemsLink5;

  for(i = 0; i < NElements; i++)
  {
    n1 = Vertices[i][0] - 1;
    n2 = Vertices[i][1] - 1;

    ElementIndex = StartIndex - 1 + i;

    /* search for 2D elements sharing two nodes with the link */
    for(j = 0; j < no_node_elem[n1]; j++)
    {
      for(k = 0; k < no_node_elem[n2]; k++)
      {
        if((node_elem[n1][j] == node_elem[n2][k]) &&
           (node_elem[n1][j] > no_link)                )
        {
          /* this is the adjacent 2D element and its side */
          index = node_elem[n1][j]-1;
          ix = FindSideInElem(mesh, index, n1, n2);

#if 0
          /* if there are other adjacent 2D elements, 
             insert the 1D element in between */
          for(ii = 0; ii < Degree[index][ix]; ii++)
          {
            adjelem = Adjacency[index][ix][ii];
            adjside = AdjacentSide[index][ix][ii];
            
            for(jj = 0; jj < Degree[adjelem][adjside]; jj++)
            {
              if(Adjacency[adjelem][adjside][jj] == index)
              {
                Adjacency[ElementIndex][0] = 
                  e_incr_i_vect(Degree[ElementIndex][0], 
                                Adjacency[ElementIndex][0], 1);
                AdjacentSide[ElementIndex][0] = 
                  e_incr_i_vect(Degree[ElementIndex][0], 
                                AdjacentSide[ElementIndex][0], 1);
                Adjacency[ElementIndex][0][Degree[ElementIndex][0]] = adjelem;
                AdjacentSide[ElementIndex][0][Degree[ElementIndex][0]] = adjside;
                
                Adjacency[adjelem][adjside][jj] = index;
                AdjacentSide[adjelem][adjside][jj] = Degree[ElementIndex][0];
                Degree[ElementIndex][0]++;
                break;
              }
            }
            if(jj == Degree[adjelem][adjside])
            {
              e_GeneralError("Set1DElemMatrix: internal error 1 !");
            }
          }
#endif
          
          
          if(Degree[index][ix] == 0)
          {
            /* store the 1D element with the 2D element */
            Adjacency[index][ix]    = e_alloc_i_vect(1);
            AdjacentSide[index][ix] = e_alloc_i_vect(1);
          }
          Adjacency[index][ix][0] = ElementIndex;
          AdjacentSide[index][ix][0] = Degree[ElementIndex][0];
          Degree[index][ix] = 1;
         
          /* store the 2D element with the 1D element */
          Adjacency[ElementIndex][0] = 
            e_incr_i_vect(Degree[ElementIndex][0], 
                          Adjacency[ElementIndex][0], 1);
          AdjacentSide[ElementIndex][0] = 
            e_incr_i_vect(Degree[ElementIndex][0], 
                          AdjacentSide[ElementIndex][0], 1);
          Adjacency[ElementIndex][0][Degree[ElementIndex][0]]    = index;
          AdjacentSide[ElementIndex][0][Degree[ElementIndex][0]] = ix;
          Degree[ElementIndex][0]++;

        }
      }
    }

  }
}



void e_Free2DElemAdjacency(MESH *mesh, ADJDATA *adata)
{
  int i, j;

  for(i = 0; i < mesh->TotalNElems; i++)
  {
    for(j = 0; j < adata->Sides[i]; j++)
    {
      e_free(adata->Adjacency[i][j]);
      e_free(adata->AdjacentSide[i][j]);
    }
    e_free(adata->Degree[i]);
    e_free(adata->Adjacency[i]);
    e_free(adata->AdjacentSide[i]);
  }
  e_free(adata->Sides);
  e_free(adata->Degree);
  e_free(adata->Adjacency);
  e_free(adata->AdjacentSide);

}





