#ifndef __COMM_VECTOR_H__
#define __COMM_VECTOR_H__

#include "stdlib.h"

#include "config.h"


/* temporary solution */
#if !(__PARALLEL__) && !(__E_MPI__)
  #error !!!!! __PARALLEL__ or __E_MPI__ macro is not defined !!!!!
#endif

#include <mpi.h>

#ifndef MAX_VECTOR_COMM
#define MAX_VECTOR_COMM  6
#endif

/* data types in the meshes */
#define INTERNAL_ELEM  0
#define BOUNDARY_ELEM  1
#define HALO_ELEM      2

#define INTERNAL_NODE 0
#define BOUNDARY_NODE 1
#define HALO_NODE     2

typedef struct
{
  int             Par_ProcID;             /* process ID in MPI group */
  int             Par_NSubdomains;        /* how many processes in the group */
  
  /* members below are handled by "Init" functions */
  int            *Par_NElemSend;          /* no. elem info to send, per subdomain */
  int            *Par_NElemRecv;          /* no. elem info to receive, per subdomain */
  int           **Par_ElemSend;           /* local elems to send, per subdomain */
  int           **Par_ElemRecv;           /* local elems to receive, per subdomain */
  
  /* members below are handled by the "Final" function */
  int            *Par_SubdNoSend;         /* subdomain numbers where to send */
  int            *Par_SubdNoRecv;         /* subdomain numbers where from to receive */
  int             Par_NSubdNoSend;        /* to how many subdomain to send*/
  int             Par_NSubdNoRecv;        /* from how many subdomain to receive */

  double         *Values[MAX_VECTOR_COMM]; /* pointers to data to communicate */
  double        **Par_BufSend;            /* buffers to send the data simultanenously */
  double        **Par_BufRecv;            /* buffers to receive the data simultanenously */

  MPI_Request    *Par_SendCommReq;        /* communicators to send */
  MPI_Status     *Par_SendCommStat;       /* status of sending processes */

  MPI_Request    *Par_RecvCommReq;        /* communicators to receive */
}
COMM_VECTOR_DATA;


void SendCommDataInit(int NSubdomains, COMM_VECTOR_DATA *data, int i_subdom);
void RecvCommDataInit(int NSubdomains, COMM_VECTOR_DATA **data, int i_subdom);
void SetupCommDataFinal(COMM_VECTOR_DATA *data);


void LaunchDataSend(int n, COMM_VECTOR_DATA *data);
void WaitDataSend(COMM_VECTOR_DATA *data);
void LaunchDataRecv(int n, COMM_VECTOR_DATA *data);
void WaitDataRecv(int n, COMM_VECTOR_DATA *data);
void CommunicateData(int n, COMM_VECTOR_DATA *data);


#endif


