#ifndef __DR_H__
#define __DR_H__

#include "dr_len.h"

#ifdef __E_MPI__

#ifdef DD_TYPE_COMM
  #include "comm_vector.h"  /* COMM_VECTOR_DATA definition */
#endif

#ifdef AL_TYPE_COMM
  #include "dr_comm_al.h"  /* REDUCE_VECTOR_DATA definition */
#endif

#ifdef BU_TYPE_COMM
  #include "dr_comm_bu.h"  /* BUTTERFLY_DATA definition */
#endif

#endif /* __E_MPI__ */


#define TASK_FORM_FINDING        1
#define TASK_CUTTING_PATTERN     2
#define TASK_GEN_INIT_LENGTH     4
/* FIXME: these control parameters are not supposed to be here */
#define TASK_CHECK_STRESSES      8
#define TASK_WRITE_INIT_LENGTH   16
#define TASK_WRITE_ANALYSIS      32
#define TASK_WRITE_FORCE         64
#define TASK_WITHOUT_G_STRING    128

/*
 * Data structure for dynamic relaxation 
 */

typedef struct
{
  char      *Projectname;
  char      *Outputname;
  int        ProjectTask;
  int        OutputTask;
  int        Dimension;

  int        StopAtConv;
  double     ConvCrit;

  int        PressureOption;
  double     PressureValue;

  int        HydPressOption;
  double     HydPressDensity;
  
  int        AddMassOption;
  int        GlobalMassOption;
  double     GlobalMassMult;
  int        NAddMass;
  int       *AddMassNode;
  double   **AddMassMult;
  
  char      *CuttingMaterialFile;

  double   **InitCoords; /* Initial coordinates */
    
  INIT_LENGTH_DATA *InitLengthData; /* initial or slack length data per element */

  /* members below are not handled by functions in dr_data.c, 
     these members are setup through dr_io.c */

  int        NCompNodes;          /* no of nodes where residual should be */
                                  /* reversed due to compressed material*/
  int       *CompNodes;           /* list of the nodes */

  double     A, B;                /* constants for the DR iteration */
  double     CurrentEnergy;       /* current energy level */
  double   **Pressure;            /* pressure value per node per dimension */
  double   **Normal;              /* normal vector per node per dimension */
  double   **Residual;            /* residual force per node per dimension */
  double   **Velocity;            /* velocity, speed per node per dimension */
  double   **Mass;                /* mass per node per dimension */

  double  ***KeTriang3;           /* natural stiffness matrix (3x3) per element */
  double  ***KeTriang4;           /* natural stiffness matrix (3x3) per element */
  /* Triang 5 (form-finding element does not require stiffness matrix storage */

#ifdef __E_MPI__
  int        CommReq;             /* communication or interaction required */
                                  /* during parallel DR */
  int       *CommEnergy;          /* nodes participating in energy calculation */
  
#ifdef DD_TYPE_COMM
  COMM_VECTOR_DATA *CommData;
#endif

#ifdef AL_TYPE_COMM
  REDUCE_VECTOR_DATA *CommData;
#endif

#ifdef BU_TYPE_COMM
  BUTTERFLY_DATA *CommData;
#endif

#endif /* __E_MPI__ */
  
}
DR_DATA;


#endif


