#include "e_lib.h"
#include "ga.h"

#include "dr_cut.h"

static int       G_NSubdomains;
static G_CUTPAT *G_mdata;
static int      *G_subdom_elems;
static int       G_Printing;

double objective_function(long nvars, DESIGNPARAMS vars)
{
  int  i, j, i_subdom;
  int  elem_move, total_elem, result, ideal, subdom;

  /* clear the number of elements in the subdomains */
  for(i = 0; i < G_NSubdomains; i++)
  {
    G_subdom_elems[i] = 0;
  }

  /* determine the number of elements in the subdomain +
     how many element should be moved */
  elem_move = 0;
  for(i = 0; i < nvars; i++)
  {
    i_subdom = (int)vars[i].parameter;
    subdom = G_mdata->PartSubdomForStrip[i][i_subdom];

    for(j = 0; j < G_mdata->NPartForStrip[i]; j++)
    {
      G_subdom_elems[subdom] += G_mdata->NElemInPartForStrip[i][j];

      if(subdom != G_mdata->PartSubdomForStrip[i][j])
        elem_move += G_mdata->NElemInPartForStrip[i][j];
    }
  }

  total_elem = 0;

  /* 
   * add the elements in the whole strips to the subdomain
   * add up all the elements 
   */
  for(i = 0; i < G_NSubdomains; i++)
  {
    G_subdom_elems[i] += G_mdata->NElemInSubdom[i];
    total_elem += G_subdom_elems[i];
  }
  
  /* ideal number of elements */
  ideal = total_elem / G_NSubdomains;
  result = 0;

  /* check deviation from ideal number of elements */
  for(i = 0; i < G_NSubdomains; i++)
  {
    result += E_ABS(ideal - G_subdom_elems[i]);
    if(G_Printing == 2)
      printf("\n%d subdom diff from ideal = %d", i+1, E_ABS(ideal - G_subdom_elems[i]));
  }
  
  /* print number of elements to move */
  if(G_Printing == 2)
  {
    printf("\nNo elem to move = %d", elem_move);
    printf("\n----------------------------------------");
  }

  /* function value */
  return (double)(total_elem * 2 - (1.00 - G_mdata->Gamma) * result - G_mdata->Gamma * elem_move);

}


void DoBalancing(int NSubdomains, CUTTING_PATTERN *cdata, GA_DATA *gadata)
{
  int i, no_variable, pr, act_var;
  DESIGNPARAMS  design_var;
  int subdom, i_subdom;

  /* store it for the objective function */
  G_NSubdomains = NSubdomains;
  G_mdata       = cdata->MasterData;
  
/*  G_mdata->Gamma = 0.75; */

  G_subdom_elems = e_alloc_i_vect(NSubdomains);

  cdata->MasterData->NewSubdomForPartStrip = 
    e_alloc_i_vect(cdata->MasterData->NAssembledPartStrip);

  no_variable = cdata->MasterData->NAssembledPartStrip;

  design_var = (DESIGNPARAMS)e_calloc(no_variable, sizeof(PARAM));

  act_var = 0;
  for(i = 0; i < no_variable; i++)
  {
    pr = 1;
    while(cdata->MasterData->NPartForStrip[i] > pow(2, pr))
      pr++;

    design_var[act_var].minparam = 0.00;
    design_var[act_var].maxparam = cdata->MasterData->NPartForStrip[i]-1;
    design_var[act_var].lparam = pr;
    act_var++;
  }

  srand(gadata->GA_Seed);
  genetic_algorithm(gadata->GA_Popsize, gadata->GA_CrossProb, 
                    gadata->GA_MutationProb, no_variable, 
                    design_var, gadata->GA_MaxGen, gadata->GA_LinScale, 
                    gadata->GA_Printing);

  /* this is only to print the final result */
  G_Printing = 2;
  objective_function(no_variable, design_var);

  for(i = 0; i < cdata->MasterData->NAssembledPartStrip; i++)
  {
    i_subdom = (int)design_var[i].parameter;
    subdom = cdata->MasterData->PartSubdomForStrip[i][i_subdom];

    cdata->MasterData->NewSubdomForPartStrip[i] = subdom;
  }


}



