#include "dr_plt.h"

#include <stdio.h>
#include <math.h>

#include "e_lib.h"
#include "/home/peteri/work/GA/ga.h"

#include "dr_main.h"
#include "dr_io.h"
#include "dr_cut.h"
#include "dr_bal.h"


#include "mpi.h"
#include "dr_comm.h"



double objective_function(long nvars, DESIGNPARAMS vars)
{
  int  i, j, i_subdom;
  int  total_elem, result, ideal, subdom;

#ifdef __GA_PARETO__
  static  int very_first_time = 1, counter = 0;
  static  double n_this, n_new;
  double  min_prox, par_result, prox;
  short   pareto_smaller, pareto_equal, pareto_bigger;
#endif

  for(i = 0; i < GV_total_no_subdomains+1; i++)
    GA_obj_func[i] = 0;

  GA_obj_func[GV_total_no_subdomains] = 0;
  for(i = 0; i < nvars; i++)
  {
    i_subdom = (int)vars[i].parameter;
    subdom = GV_m_half_strips_subdoms[i][i_subdom];

    for(j = 0; j < GV_m_no_half_strips[i]; j++)
    {
      GA_obj_func[subdom] += GV_m_half_strips_elem_number[i][j];

/* elements to move */
      if(subdom != GV_m_half_strips_subdoms[i][j])
        GA_obj_func[GV_total_no_subdomains] +=
          GV_m_half_strips_elem_number[i][j];
    }
  }

  total_elem = 0;

/* add the elements in the whole strips to the subdomain
   add up all the elements 
*/
  for(i = 0; i < GV_total_no_subdomains; i++)
  {
    GA_obj_func[i] += GV_m_elems_in_subdoms[i];
    total_elem += GA_obj_func[i];

  }

  ideal = total_elem / GV_total_no_subdomains;

  for(i = 0; i < GV_total_no_subdomains; i++)
    GA_obj_func[i] = E_ABS(ideal - GA_obj_func[i]);

#ifdef __GA_PARETO__
  if(counter == GA_popsize)
  {
    n_this = n_new;
    counter = 0;
  }

  if(very_first_time)
  {
/* only for the very first time 
   contains initialization as well
*/

    n_this = total_elem;
    n_new = n_this;

    min_prox = 0.0;

    GA_no_Pareto = 0;
    i = GA_no_Pareto;
    GA_all_time_Pareto_length = GA_no_Pareto;
    GA_Pareto = NULL;

    very_first_time = 0;
  }
  else
  {

/* calculate proximity values */
    min_prox = 1e50;

    for(i = 0; i < GA_no_Pareto; i++)
    {
      prox = 0.00;
      for(j = 0; j < GV_total_no_subdomains+1; j++)
        prox += E_SQUARE( ((double)GA_Pareto[i][j] - (double)GA_obj_func[j]) / 
                           (double)GA_Pareto[i][j]                             );

      prox = sqrt(prox);
      if(prox < min_prox)
        min_prox = prox;
    }

/* check whether it is Pareto optimum and dominance */
    for(i = 0; i < GA_no_Pareto; i++)
    {
      pareto_smaller = 0;
      pareto_equal   = 0;
      pareto_bigger  = 0;
      for(j = 0; j < GV_total_no_subdomains+1; j++)
      {
        if(GA_Pareto[i][j] < GA_obj_func[j])
          pareto_smaller = 1;
        if(GA_Pareto[i][j] == GA_obj_func[j])
          pareto_equal = 1;
        if(GA_Pareto[i][j] > GA_obj_func[j])
          pareto_bigger = 1;
      }
      if((pareto_smaller  && !pareto_equal && !pareto_bigger) ||
         (pareto_smaller  &&  pareto_equal && !pareto_bigger) ||
         (!pareto_smaller &&  pareto_equal && !pareto_bigger)    )
        break;

      if((!pareto_smaller && !pareto_equal && pareto_bigger) ||
         (!pareto_smaller &&  pareto_equal && pareto_bigger)    )
        GA_Pareto[i][GV_total_no_subdomains+1] = 1;
    }

  } /* else very_first_time */


/* new Pareto optimum */
  if(i == GA_no_Pareto) 
  {

/* remove dominated Pareto optimums */
    for(i = 0; i < GA_no_Pareto; i++)
    {
      if(GA_Pareto[i][GV_total_no_subdomains+1])
      {
        e_free(GA_Pareto[i]);
        for(j = i+1; j < GA_no_Pareto; j++)
          GA_Pareto[j-1] = GA_Pareto[j];
        GA_no_Pareto--;
        i--;
      }
    }

/* add new Pareto optimum */
    if(GA_all_time_Pareto_length == GA_no_Pareto)
    {
      GA_Pareto = e_incr_frag_i_matr(GA_all_time_Pareto_length, GA_Pareto, 1);
      GA_all_time_Pareto_length++;
    }

    GA_Pareto[GA_no_Pareto] = e_alloc_i_vect(GV_total_no_subdomains+2);
    for(j = 0; j < GV_total_no_subdomains+1; j++)
      GA_Pareto[GA_no_Pareto][j] = GA_obj_func[j];
    GA_no_Pareto++;


    par_result = n_this + min_prox;
    if(par_result > n_new)
      n_new = par_result;
  }
  else
  {
    par_result = n_this - min_prox;
  }

  counter++;

  return par_result;

#else

  result = 0;

  for(i = 0; i < GV_total_no_subdomains; i++)
  {
    result += GA_obj_func[i];
    if(GA_printing == 2)
      printf("\n%d subdom diff from ideal = %d", i+1, GA_obj_func[i] );
  }

  if(GA_printing == 2)
  {
    printf("\nNo elem to move = %d", GA_obj_func[GV_total_no_subdomains] );
    printf("\n----------------------------------------");
  }

  return (double)(total_elem * 2 - (1.00 - GA_gamma) * result - 
                  GA_gamma * GA_obj_func[GV_total_no_subdomains] );

#endif

}



void do_balancing()
{
  int i, no_variable, pr, act_var;
  DESIGNPARAMS  design_var;
  double final;
  int subdom, i_subdom;

  GA_obj_func = e_alloc_i_vect(GV_total_no_subdomains+1);

  GV_m_where_subdom = e_alloc_i_vect(GV_m_no_assembled_strips);

  no_variable = GV_m_no_assembled_strips;

  design_var = (DESIGNPARAMS) e_calloc(no_variable, sizeof(PARAM));

  act_var = 0;
  for(i = 0; i < no_variable; i++)
  {
    pr = 1;
    while(GV_m_no_half_strips[i] > pow(2, pr))
      pr++;

    design_var[act_var].minparam = 0.00;
    design_var[act_var].maxparam = GV_m_no_half_strips[i]-1;
    design_var[act_var].lparam = pr;
    act_var++;
  }

  srand(GA_seed);
  genetic_algorithm(GA_popsize, GA_pcross, GA_pmutation, no_variable, 
                    design_var, GA_maxgen, GA_linear, GA_printing);


#ifdef __GA_PARETO__
  printf("\nNo of Pareto optimums: %d", GA_no_Pareto);

  for(i = 0; i < GA_no_Pareto; i++)
  {
    int j;

    printf("\n %d. ", i);
    for(j = 0; j < GV_total_no_subdomains+1; j++)
      printf("  %d", GA_Pareto[i][j]);
  }
  printf("\n");

#endif


/* this is only to print the final result */
  GA_printing = 2;
  final = objective_function(no_variable, design_var);

  for(i = 0; i < GV_m_no_assembled_strips; i++)
  {
    i_subdom = (int)design_var[i].parameter;
    subdom = GV_m_half_strips_subdoms[i][i_subdom];

    GV_m_where_subdom[i] = subdom;
  }

/* for debugging */
/*
  for(i = 0; i < GV_m_no_assembled_strips; i++)
  {
    GV_m_where_subdom[i] = GV_m_half_strips_subdoms[i][0];
  }
*/

}


