#ifdef AL_TYPE_COMM

void SetupSymCommDataInit(int dimension, int NSubdomains, int i_subdom,
                          int NInterSubDomNodes, int *InterSubDomNodes,
                          int *NNodeSubDoms, int **NodeSubDoms,
                          REDUCE_VECTOR_DATA **data)
{
  int i, j, NoSD_shared;

  (*data) = e_calloc(1, sizeof(REDUCE_VECTOR_DATA) );

  (*data)->Par_TotalNElem = NInterSubDomNodes * dimension;
  (*data)->Par_NElem      = 0;
  (*data)->Par_Elems      = NULL;
  (*data)->Par_ElemPos    = NULL;

  for(i = 0; i < NInterSubDomNodes; i++)
  {
    for(j = 0; j < NNodeSubDoms[i]; j++)
    {
      if(i_subdom == NodeSubDoms[i][j])
      {
        (*data)->Par_Elems = 
          e_incr_i_vect((*data)->Par_NElem, (*data)->Par_Elems, ALLOC_SIZE * dimension);
        (*data)->Par_ElemPos = 
          e_incr_i_vect((*data)->Par_NElem, (*data)->Par_ElemPos, ALLOC_SIZE * dimension);
        
        (*data)->Par_Elems[(*data)->Par_NElem] = InterSubDomNodes[i];
        (*data)->Par_ElemPos[(*data)->Par_NElem] = i * dimension;
        
        (*data)->Par_NElem += dimension;
        break;
      }
    }
  }
}


void FreeSymCommDataInit(int NSubdomains, REDUCE_VECTOR_DATA *data)
{
  e_free(data->Par_Elems);
  e_free(data->Par_ElemPos);
  data->Par_Elems    = NULL;
  data->Par_ElemPos  = NULL;

  e_free(data);
}


void ConvertSymCommDataInit(MESH *mesh, REDUCE_VECTOR_DATA *data, 
                            int dimension)
{
  int i, k, node, pos, index;

  for(i = 0; i < data->Par_NElem; i+=dimension)
  {
    node = data->Par_Elems[i];
    for(index = 0; index < mesh->NMeshPoints; index++)
    {
      if(mesh->GNIs[index] == node)
        break;
    }
    /* check that we have the node locally in the mesh */
    if(index == mesh->NMeshPoints)
    {
      e_GeneralError("%d proc: communication node %d is not found in mesh!", 
                     data->Par_ProcID, node);
    }

    /* setup access to the matrices as a vector */
    pos = data->Par_ElemPos[i];
    for(k = 0; k < dimension; k++)
    {
      data->Par_Elems[i+k]    = index * dimension + k + 1;
      data->Par_ElemPos[i+k]  = pos + k + 1;
    }
  }

}



/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/* !!!!! These functions belong to a "reduce_vector.c" file !!!!! */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */

#define AL_DATA   2

void SendCommDataInit(int NSubdomains, REDUCE_VECTOR_DATA *data, int i_subdom)
{
  int senddata[AL_DATA];

  senddata[0] = data->Par_NElem;
  senddata[1] = data->Par_TotalNElem;

  MPI_Send(senddata, AL_DATA, MPI_INT, 
           i_subdom, 444, MPI_COMM_WORLD);

  MPI_Send(data->Par_Elems, data->Par_NElem, MPI_INT, 
           i_subdom, 445, MPI_COMM_WORLD);

  MPI_Send(data->Par_ElemPos, data->Par_NElem, MPI_INT, 
           i_subdom, 446, MPI_COMM_WORLD);
}


void RecvCommDataInit(int NSubdomains, REDUCE_VECTOR_DATA **data, int i_subdom)
{
  MPI_Status status;
  int        recvdata[AL_DATA];

  (*data) = calloc(1, sizeof(REDUCE_VECTOR_DATA) );

  MPI_Recv(recvdata, AL_DATA, MPI_INT, 
           i_subdom, 444, MPI_COMM_WORLD, &status);

  (*data)->Par_NElem = recvdata[0];
  (*data)->Par_TotalNElem = recvdata[1];

  (*data)->Par_Elems = (int *)calloc((*data)->Par_NElem, sizeof(int) );
  MPI_Recv((*data)->Par_Elems, (*data)->Par_NElem, MPI_INT, 
               i_subdom, 445, MPI_COMM_WORLD, &status);

  (*data)->Par_ElemPos = (int *)calloc((*data)->Par_NElem, sizeof(int) );
  MPI_Recv((*data)->Par_ElemPos, (*data)->Par_NElem, MPI_INT, 
               i_subdom, 446, MPI_COMM_WORLD, &status);
}

void SetupCommDataFinal(REDUCE_VECTOR_DATA *data)
{
  data->Par_BufSend = 
    (double *)calloc(data->Par_TotalNElem * MAX_VECTOR_COMM, sizeof(double));
  data->Par_BufRecv = 
    (double *)calloc(data->Par_TotalNElem * MAX_VECTOR_COMM, sizeof(double));
}


void CommunicateData(int n, REDUCE_VECTOR_DATA *data)
{
  int k, j;

  for(k = 0; k < n; k++)
  {
    for(j = 0; j < data->Par_NElem; j++)
    {
      data->Par_BufSend[data->Par_ElemPos[j]-1+k*data->Par_NElem] = 
        data->Values[k][data->Par_Elems[j]-1];
    }
  }

  MPI_Allreduce(data->Par_BufSend, data->Par_BufRecv, data->Par_TotalNElem,
                MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);

  for(k = 0; k < n; k++)
  {
    for(j = 0; j < data->Par_NElem; j++)
    {
      data->Values[k][data->Par_Elems[j]-1] =
      data->Par_BufRecv[data->Par_ElemPos[j]-1+k*data->Par_NElem];
    }
  }

}

#endif



