#ifdef BU_TYPE_COMM


void SetupSymCommDataInit(int dimension, int NSubdomains, int i_subdom,
                          int NInterSubDomNodes, int *InterSubDomNodes,
                          int *NNodeSubDoms, int **NodeSubDoms,
                          BUTTERFLY_DATA **data)
{
  int i, j, nsubdom;

  (*data) = e_calloc(1, sizeof(BUTTERFLY_DATA) );

  (*data)->Par_NElem = NInterSubDomNodes;
  (*data)->Par_Elems = e_alloc_i_vect((*data)->Par_NElem);
  (*data)->Par_NSharedSubdom = e_alloc_i_vect((*data)->Par_NElem);
  (*data)->Par_SharedSubdoms = e_init_frag_i_matr((*data)->Par_NElem);

  for(i = 0; i < (*data)->Par_NElem; i++)
  {
    nsubdom = NNodeSubDoms[i];
    (*data)->Par_Elems[i] = InterSubDomNodes[i];
    (*data)->Par_NSharedSubdom[i] = nsubdom;
    (*data)->Par_SharedSubdoms[i] = e_alloc_i_vect(nsubdom);
    for(j = 0; j < nsubdom; j++)
    {
      (*data)->Par_SharedSubdoms[i][j] = NodeSubDoms[i][j];
    }
  }
}



void FreeSymCommDataInit(int NSubdomains, BUTTERFLY_DATA *data)
{
  e_free(data->Par_Elems);
  e_free(data->Par_NSharedSubdom);
  
  e_free_frag_i_matr(data->Par_NElem, data->Par_SharedSubdoms);

  e_free(data);
}


/* collects those subdomains from which information is received */
static void CollectRecvNeighbours(int MyProcID, int nwave, int iwave, int **destination,
                                  int *NNeighbour, int **Neighbours)
{
  int procID, i, j;

  for(i = nwave; i >= iwave+1; i--)
  {
    procID = MyProcID;
    *Neighbours = e_ext_i_vect(*NNeighbour, *Neighbours, 1);
    (*Neighbours)[*NNeighbour] = destination[procID][i]+1;
    (*NNeighbour)++;
    
    procID = destination[procID][i];
    
    for(j = i+1; j < nwave+1; j++)
    {
      *Neighbours = e_ext_i_vect(*NNeighbour, *Neighbours, 1);
      (*Neighbours)[*NNeighbour] = destination[procID][j]+1;
      (*NNeighbour)++;
    
      procID = destination[procID][j];
    }
  
  }
}

/* collect those to which we communicate, but maybe through other subdomains */
static void CollectSendNeighbours(int MyProcID, int iwave, int **destination,
                                  int *NNeighbour, int **Neighbours)
{
  int procID, i, j, ptr;

  *Neighbours = e_alloc_i_vect(1);
  (*Neighbours)[0] = destination[MyProcID][iwave]+1;
  (*NNeighbour)++;
  
  ptr = 0;
  while(ptr != (*NNeighbour))
  {
    procID = (*Neighbours)[ptr]-1;
    for(i = iwave-1; i >= 0; i--)
    {
      for(j = 0; j < *NNeighbour; j++)
      {
        if((*Neighbours)[j] == destination[procID][i]+1)
          break;
      }
      if(j == *NNeighbour)
      {
        *Neighbours = e_ext_i_vect(*NNeighbour, *Neighbours, 1);
        (*Neighbours)[*NNeighbour] = destination[procID][i]+1;
        (*NNeighbour)++;
      }
      
      procID = destination[procID][i];
    }
    ptr++;

  }

}






static void CollectElems(int dimension, int ProcID, int Wave, int NWave,
                         int **Destination, int NElem, int *Elems,
                         int *NSharedSubdom, int **SharedSubdoms,
                         int *NCommElem, int **CommElems)
{
  int  i, j, k, found;
  int  NRecvNeighbour;
  int *RecvNeighbours;
  int  NSendNeighbour;
  int *SendNeighbours;

  RecvNeighbours = NULL;
  NRecvNeighbour = 0;
  CollectRecvNeighbours(ProcID, NWave, Wave, Destination, 
                        &NRecvNeighbour, &RecvNeighbours);

  SendNeighbours = NULL;
  NSendNeighbour = 0;
  CollectSendNeighbours(ProcID, Wave, Destination, 
                        &NSendNeighbour, &SendNeighbours);

  (*NCommElem) = 0;
  (*CommElems) = NULL;

  /* all "elements" */
  for(i = 0; i < NElem; i++)
  {

#if 0
    /* either my subdomain or a received subdom */
    found = FALSE;
    for(j = 0; (j < NSharedSubdom[i]) && (found == FALSE); j++)
    {
      /* this subdomain has the info directly */
      if(SharedSubdoms[i][j] == ProcID+1)
      {
        found = TRUE;
        break;
      }
      /* it has been received by this subdomain */
      for(k = 0; k < NRecvNeighbour; k++)
      {
        if(SharedSubdoms[i][j] == RecvNeighbours[k])
        {
          found = TRUE;
          break;
        }
      }
    }
    if(found == FALSE)
      continue;
#endif

    found = FALSE;
    /* "elements" need info from these subdomains */
    for(j = 0; j < NSharedSubdom[i] && !found; j++)
    {
      for(k = 0; k < NSendNeighbour; k++)
      {
        if(SharedSubdoms[i][j] == SendNeighbours[k])
        {
          found = TRUE;
          break;
        }
      }
    }

    if(found)
    {
      (*CommElems) = e_incr_i_vect(*NCommElem, *CommElems, ALLOC_SIZE*dimension);
      for(j = 0; j < dimension; j++)
      {
        /* where to insert in the global vector */
        (*CommElems)[(*NCommElem)+j] = i*dimension+j;
      }
      (*NCommElem) += dimension;
    }
  }

  e_free(RecvNeighbours);
  e_free(SendNeighbours);
}


void ConvertSymCommDataInit(MESH *mesh, BUTTERFLY_DATA *data, int dimension)
{
  int i, k, node, index;
  int NWave, iwave, FromWhom;
  int i_sub, procID, mult, half;

  /* relation between per wave buffers and Par_AllElems buffer */
  data->Par_NLocalElemPos = 0;
  data->Par_LocalCommPos  = NULL;
  data->Par_LocalMeshPos  = NULL;

  for(i = 0; i < data->Par_NElem; i++)
  {
    node = data->Par_Elems[i];
    for(index = 0; index < mesh->NMeshPoints; index++)
    {
      if(mesh->GNIs[index] == node)
        break;
    }

    /* we may not have the node locally */
    if(index != mesh->NMeshPoints)
    {
      data->Par_LocalCommPos = 
        e_incr_i_vect(data->Par_NLocalElemPos, data->Par_LocalCommPos, ALLOC_SIZE*dimension);
      data->Par_LocalMeshPos = 
        e_incr_i_vect(data->Par_NLocalElemPos, data->Par_LocalMeshPos, ALLOC_SIZE*dimension);
    
      /* setup access to the matrices as a vector */
      for(k = 0; k < dimension; k++)
      {
        data->Par_LocalCommPos[data->Par_NLocalElemPos+k] = i*dimension+k;
        data->Par_LocalMeshPos[data->Par_NLocalElemPos+k] = index*dimension+k;
      }
      data->Par_NLocalElemPos+=dimension;
    }
  }

  /* number of communication waves */
  NWave = (int)(log(data->Par_NSubdomains) / log(2.0));

  /* setup total butterfly communication table */
  data->Par_Destination = 
    e_alloc_cont_i_matr(data->Par_NSubdomains, NWave);

  for(i_sub = 0; i_sub < data->Par_NSubdomains; i_sub++)
  {
    procID = i_sub;
    for(iwave = NWave-1; iwave >= 0; iwave--)
    {
      mult = 1;
      half = pow(2, iwave);
      if(procID >= half)
        mult = -1;
      data->Par_Destination[i_sub][iwave] = i_sub + half * mult;
      if(procID >= half)
        procID -= half;
    }
  }

  /* all data to be communicated, send/receive */
  data->Par_NAllElem = data->Par_NElem * dimension;
  data->Par_AllElems = 
    e_alloc_d_vect(data->Par_NAllElem * MAX_VECTOR_COMM);

  /* setup the communication buffers */
  data->Par_NElemWaveSend = e_alloc_i_vect(NWave);
  data->Par_IndexWaveSend = e_init_frag_i_matr(NWave);
  data->Par_ElemsWaveSend = e_init_frag_d_matr(NWave);
  data->Par_NElemWaveRecv = e_alloc_i_vect(NWave);
  data->Par_IndexWaveRecv = e_init_frag_i_matr(NWave);
  data->Par_ElemsWaveRecv = e_init_frag_d_matr(NWave);

  for(iwave = NWave-1; iwave >= 0; iwave--)
  {
    /* send */
    CollectElems(dimension, data->Par_ProcID-1, iwave, NWave-1, data->Par_Destination,
                 data->Par_NElem, data->Par_Elems,
                 data->Par_NSharedSubdom, data->Par_SharedSubdoms,
                 &(data->Par_NElemWaveSend[iwave]), &(data->Par_IndexWaveSend[iwave]) );

    /* receive */
    FromWhom = data->Par_Destination[data->Par_ProcID-1][iwave];

    CollectElems(dimension, FromWhom, iwave, NWave-1, data->Par_Destination,
                 data->Par_NElem, data->Par_Elems,
                 data->Par_NSharedSubdom, data->Par_SharedSubdoms,
                 &(data->Par_NElemWaveRecv[iwave]), &(data->Par_IndexWaveRecv[iwave]) );
    
    /* allocate the buffer itself */
    data->Par_ElemsWaveSend[iwave] = 
      e_alloc_d_vect(data->Par_NElemWaveSend[iwave] * MAX_VECTOR_COMM);
    data->Par_ElemsWaveRecv[iwave] = 
      e_alloc_d_vect(data->Par_NElemWaveRecv[iwave] * MAX_VECTOR_COMM);
  }


}


void SetupCommDataFinal(BUTTERFLY_DATA *data)
{
  data->Par_SendReq  = (MPI_Request *)e_calloc(1, sizeof(MPI_Request) );
  data->Par_RecvReq  = (MPI_Request *)e_calloc(1, sizeof(MPI_Request) );
  data->Par_SendStat = (MPI_Status *)e_calloc(1, sizeof(MPI_Status) );
  data->Par_RecvStat = (MPI_Status *)e_calloc(1, sizeof(MPI_Status) );

}


void SendCommDataInit(int NSubdomains, BUTTERFLY_DATA *data, int i_subdom)
{
  int i;

  MPI_Send(&(data->Par_NElem), 1, MPI_INT, 
           i_subdom, 444, MPI_COMM_WORLD);

  MPI_Send(data->Par_Elems, data->Par_NElem, MPI_INT, 
           i_subdom, 445, MPI_COMM_WORLD);

  MPI_Send(data->Par_NSharedSubdom, data->Par_NElem, MPI_INT, 
           i_subdom, 446, MPI_COMM_WORLD);

  for(i = 0; i < data->Par_NElem; i++)
  {
    MPI_Send(data->Par_SharedSubdoms[i], data->Par_NSharedSubdom[i], MPI_INT, 
             i_subdom, 447+i, MPI_COMM_WORLD);
  }
}


void RecvCommDataInit(int NSubdomains, BUTTERFLY_DATA **data, int i_subdom)
{
  int         i;
  MPI_Status  status;

  (*data) = e_calloc(1, sizeof(BUTTERFLY_DATA) );

  MPI_Recv(&((*data)->Par_NElem), 1, MPI_INT, 
           i_subdom, 444, MPI_COMM_WORLD, &status);

  (*data)->Par_Elems = e_alloc_i_vect((*data)->Par_NElem);
  MPI_Recv((*data)->Par_Elems, (*data)->Par_NElem, MPI_INT, 
           i_subdom, 445, MPI_COMM_WORLD, &status);

  (*data)->Par_NSharedSubdom = e_alloc_i_vect((*data)->Par_NElem);
  MPI_Recv((*data)->Par_NSharedSubdom, (*data)->Par_NElem, MPI_INT, 
           i_subdom, 446, MPI_COMM_WORLD, &status);

  (*data)->Par_SharedSubdoms = e_init_frag_i_matr((*data)->Par_NElem);
  for(i = 0; i < (*data)->Par_NElem; i++)
  {
    (*data)->Par_SharedSubdoms[i] = e_alloc_i_vect((*data)->Par_NSharedSubdom[i]);
    MPI_Recv((*data)->Par_SharedSubdoms[i], (*data)->Par_NSharedSubdom[i], MPI_INT, 
             i_subdom, 447+i, MPI_COMM_WORLD, &status);
  }
}


/* ------------------------------------------------------------------------------- */


void LaunchDataSend(int n, int iwave, BUTTERFLY_DATA *data)
{
  int j, k, nvec, dest;

  /* assemble buffers and start send */
  for(k = 0; k < n; k++)
  {
    nvec = k * data->Par_NElemWaveSend[iwave];
    for(j = 0; j < data->Par_NElemWaveSend[iwave]; j++)
    {
      data->Par_ElemsWaveSend[iwave][j+nvec] = 
        data->Par_AllElems[data->Par_IndexWaveSend[iwave][j]+nvec];
    }
  }

  /* initiate the sending procedure, non-blocking */
  dest = data->Par_Destination[data->Par_ProcID-1][iwave];
  MPI_Isend(data->Par_ElemsWaveSend[iwave], data->Par_NElemWaveSend[iwave]*n, MPI_DOUBLE,
            dest, 900+dest, MPI_COMM_WORLD, data->Par_SendReq);

}

void WaitDataSend(BUTTERFLY_DATA *data)
{
  MPI_Wait(data->Par_SendReq, data->Par_SendStat);
}

void LaunchDataRecv(int n, int iwave, BUTTERFLY_DATA *data)
{

  MPI_Irecv(data->Par_ElemsWaveRecv[iwave], data->Par_NElemWaveRecv[iwave]*n, MPI_DOUBLE,
            data->Par_Destination[data->Par_ProcID-1][iwave], 900+data->Par_ProcID-1, 
            MPI_COMM_WORLD, data->Par_RecvReq);
}

void WaitDataRecv(int n, int iwave, BUTTERFLY_DATA *data)
{
  int k, j, nvec;
  
  MPI_Wait(data->Par_RecvReq, data->Par_RecvStat);

  for(k = 0; k < n; k++)
  {
    nvec = k * data->Par_NElemWaveRecv[iwave];
    for(j = 0; j < data->Par_NElemWaveRecv[iwave]; j++)
    {
      data->Par_AllElems[data->Par_IndexWaveRecv[iwave][j]+nvec] +=
        data->Par_ElemsWaveRecv[iwave][j+nvec];
    }
  }

}


void CommunicateData(int n, BUTTERFLY_DATA *data)
{
  int i, k, nvec;
  int NWave, iwave;

  /* clear it */
  memset((void *)(data->Par_AllElems), 0,
         sizeof(double) * data->Par_NAllElem);

  /* copy from vector/matrix to our vector */
  for(k = 0; k < n; k++)
  {
    nvec = k*data->Par_NLocalElemPos;
    for(i = 0; i < data->Par_NLocalElemPos; i++)
    {
      data->Par_AllElems[data->Par_LocalCommPos[i+nvec]] =
        data->Values[k][data->Par_LocalMeshPos[i+nvec]];
    }
  }


  NWave = (int)(log(data->Par_NSubdomains) / log(2.0));

  for(iwave = NWave-1; iwave >= 0; iwave--)
  {
    if(data->Par_NElemWaveSend[iwave])
    {
      LaunchDataSend(n, iwave, data);
    }
    if(data->Par_NElemWaveRecv[iwave])
    {
      LaunchDataRecv(n, iwave, data);
    }
    if(data->Par_NElemWaveSend[iwave])
    {
      WaitDataSend(data);
    }
    if(data->Par_NElemWaveRecv[iwave])
    {
      WaitDataRecv(n, iwave, data);
    }
  }

  /* copy from our vector to vector/matrix*/
  for(k = 0; k < n; k++)
  {
    nvec = k*data->Par_NLocalElemPos;
    for(i = 0; i < data->Par_NLocalElemPos; i++)
    {
      data->Values[k][data->Par_LocalMeshPos[i+nvec]] = 
        data->Par_AllElems[data->Par_LocalCommPos[i+nvec]];
    }
  }
}



#endif


