#ifndef __DR_COMM_BU_H__
#define __DR_COMM_BU_H__

#include "stdlib.h"
#include <mpi.h>

#ifndef __E_MPI__
#error "__E_MPI__ is not defined for dr_comm_bu.h"
#endif

/* notify user about the communication type */
#define COMM_NAME "BU"

typedef struct
{
  int       Par_ProcID;
  int       Par_NSubdomains;

  /* members below are handled by "Init" functions */
  int       Par_NElem;
  int      *Par_Elems;
  int      *Par_NSharedSubdom;
  int     **Par_SharedSubdoms;

  /* members below are handled by "Convert" function */
  int       Par_NLocalElemPos;  /* no of nodes locally available */
  int      *Par_LocalCommPos;   /* pos. in Par_AllElems */
  int      *Par_LocalMeshPos;   /* position in the vector/matrix */

  int     **Par_Destination;    /* no of subdomains * no of communication waves */

  int       Par_NAllElem;       /* number of all comm elements */
  double   *Par_AllElems;       /* all data which will ever be communicated (send/receive) */
  
  int      *Par_NElemWaveSend;  /* no of elements sent in each wave */
  int     **Par_IndexWaveSend;  /* index of the sent elements in Par_AllElems */
  double  **Par_ElemsWaveSend;  /* sending buffer */

  int      *Par_NElemWaveRecv;  /* no of elements received in each wave */
  int     **Par_IndexWaveRecv;  /* index of the received elements in Par_AllElems */
  double  **Par_ElemsWaveRecv;  /* receiving buffer */

  MPI_Request *Par_SendReq;
  MPI_Request *Par_RecvReq;
  MPI_Status  *Par_SendStat;
  MPI_Status  *Par_RecvStat;

  double   *Values[MAX_VECTOR_COMM];
}
BUTTERFLY_DATA;


void SetupSymCommDataInit(int dimension, int NSubdomains, int i_subdom,
                          int NInterSubDomNodes, int *InterSubDomNodes,
                          int *NNodeSubDoms, int **NodeSubDoms,
                          BUTTERFLY_DATA **data);
void FreeSymCommDataInit(int NSubdomains, BUTTERFLY_DATA *data);
void ConvertSymCommDataInit(MESH *mesh, BUTTERFLY_DATA *data, 
                            int dimension);

void SendCommDataInit(int NSubdomains, BUTTERFLY_DATA *data, int i_subdom);
void RecvCommDataInit(int NSubdomains, BUTTERFLY_DATA **data, int i_subdom);
void SetupCommDataFinal(BUTTERFLY_DATA *data);

void LaunchDataSend(int n, int iwave, BUTTERFLY_DATA *data);
void WaitDataSend(BUTTERFLY_DATA *data);
void LaunchDataRecv(int n, int iwave, BUTTERFLY_DATA *data);
void WaitDataRecv(int n, int iwave, BUTTERFLY_DATA *data);
void CommunicateData(int n, BUTTERFLY_DATA *data);

#endif


