#ifndef DR_CUT_H
#define DR_CUT_H

#include "config.h"
#include "e_lib.h"
#include "dr.h"

#ifdef __E_MPI__
#include "ga.h"

/* this structure is valid only on the master process */
typedef struct
{
  int  **StripInfo;
  int  **NElemPerStrip;
  int  **NCommNodePerStrip;
  int  **CommNodesPerStrip;

  int    NAssembledPartStrip;
  int   *NPartForStrip;
  int  **PartSubdomForStrip;
  int  **PartNoForStrip;
  int  **NElemInPartForStrip;

  double Gamma;                 /* objective function parameter */

  int   *NElemInSubdom;
  int   *NewSubdomForPartStrip;
}
G_CUTPAT;
#endif


/* A strip mesh */
typedef struct
{
  int   ResultOK;
  MESH *StripMesh;

#ifdef __E_MPI__
#endif
}
STRIP;


typedef struct
{
#ifdef __E_MPI__
  G_CUTPAT  *MasterData;            /* data for strip balancing */

  int       *NElemInStrip;          /* number of elements in each strip */
  int        TotalNCommNodeInStrip; /* total number of comm nodes */
  int       *NCommNodeInStrip;      /* number of comm nodes per strip */
  int       *CommNodesInStrip;      /* comm nodes for all strips */
  
  int        NPartToSend;           /* number of half strips to send */
  int       *PartCommInfo;          /* Communication info for sending */
#endif

  int        NStrips;               /* number of strips to flatten */
  STRIP    **Strips;                /* strip information */

  MESH      *CableMesh;             /* cable mesh to unstress */

}
CUTTING_PATTERN;


void dr_cutting_pattern(char *ProjectName, DR_DATA *data, MESH *mesh, 
                        CUTTING_PATTERN **cdata);
#ifdef __E_MPI__
void DoBalancing(int NSubdomains, CUTTING_PATTERN *cdata, GA_DATA *gadata);
#endif

#endif /* DR_CUT_H */



