#include "ga.h"
#include "e_lib.h"

/* 
 * Returns 1 at success and 0 at failure. 
 * At failure the data pointer is NULL.
 */
int GetGAParam(char *name, GA_DATA **data)
{
  FILE *fp;
  char  line[E_LINE_MAX+1];
  int   linecount;
  char *GA_file;
  int   result;

  GA_file = (char *)e_calloc( strlen(name) + strlen(E_GA) + 2, sizeof(char) );
  sprintf(GA_file, "%s.%s", name, E_GA);

  if((fp = fopen(GA_file, "r")) == NULL)
  {
    fprintf(stderr, "Can not open GA init data file: %s!", GA_file);
    e_free(GA_file);
    return 0;
  }

  (*data) = (GA_DATA *)calloc(1, sizeof(GA_DATA));

  result = 1;
  linecount = 0;
  if(!e_GetLine(line, E_LINE_MAX, &linecount, fp))
  {
    fprintf(stderr, "Could not read line in file \"%s\" !", GA_file);
    result = 0;
  }
  else if(sscanf(line, "%ld", &((*data)->GA_Popsize)) != 1)
  {
    fprintf(stderr, "Could not read population size at line %d in file \"%s\" !", 
           linecount, GA_file);
    result = 0;
  }
  else if(!e_GetLine(line, E_LINE_MAX, &linecount, fp))
  {
    fprintf(stderr, "Could not read line in file \"%s\" !", GA_file);
    result = 0;
  }
  else if(sscanf(line, "%ld", &((*data)->GA_MaxGen)) != 1)
  {
    fprintf(stderr, "Could not read max. generation number at line %d in file \"%s\" !", 
           linecount, GA_file);
    result = 0;
  }
  else if(!e_GetLine(line, E_LINE_MAX, &linecount, fp))
  {
    fprintf(stderr, "Could not read line in file \"%s\" !", GA_file);
    result = 0;
  }
  else if(sscanf(line, "%lf", &((*data)->GA_CrossProb)) != 1)
  {
    fprintf(stderr, "Could not read crossover probability at line %d in file \"%s\" !", 
           linecount, GA_file);
    result = 0;
  }
  else if(!e_GetLine(line, E_LINE_MAX, &linecount, fp))
  {
    fprintf(stderr, "Could not read line in file \"%s\" !", GA_file);
    result = 0;
  }
  else if(sscanf(line, "%lf", &((*data)->GA_MutationProb)) != 1)
  {
    fprintf(stderr, "Could not read mutation probability at line %d in file \"%s\" !", 
            linecount, GA_file);
    result = 0;
  }
  else if(!e_GetLine(line, E_LINE_MAX, &linecount, fp))
  {
    fprintf(stderr, "Could not read line in file \"%s\" !", GA_file);
    result = 0;
  }
  else if(sscanf(line, "%d", &((*data)->GA_Seed)) != 1)
  {
    fprintf(stderr, "Could not read seeding value at line %d in file \"%s\" !", 
            linecount, GA_file);
    result = 0;
  }
  else if(!e_GetLine(line, E_LINE_MAX, &linecount, fp))
  {
    fprintf(stderr, "Could not read line in file \"%s\" !", GA_file);
    result = 0;
  }
  else if(sscanf(line, "%d", &((*data)->GA_LinScale)) != 1)
  {
    fprintf(stderr, "Could not read linear scaling option at line %d in file \"%s\" !", 
            linecount, GA_file);
    result =0;
  }
  else if(((*data)->GA_LinScale != 0) && ((*data)->GA_LinScale != 1))
  {
    fprintf(stderr, "Linear scaling option can have only zero or one value!");
    result = 0;
  }
  else if(!e_GetLine(line, E_LINE_MAX, &linecount, fp))
  {
    fprintf(stderr, "Could not read line in file \"%s\" !", GA_file);
    result = 0;
  }
  else if(sscanf(line, "%d", &((*data)->GA_Printing)) != 1)
  {
    fprintf(stderr, "Could not read value for printing level at line %d in file \"%s\" !", 
            linecount, GA_file);
    result = 0;
  }

  fclose(fp);
  e_free(GA_file);
  
  if(result == 0)
  {
    e_free(*data);
    *data = NULL;
  }
  else
  {
    printf("\nGenetic Algorithm parameters:");
    printf("\nPopulation size: %ld",                  (*data)->GA_Popsize);
    printf("\nMax. number of generations: %ld",       (*data)->GA_MaxGen);
    printf("\nCrossover probability: %lf",            (*data)->GA_CrossProb);
    printf("\nMutation probability: %lf",             (*data)->GA_MutationProb);
    printf("\nRandom number generation seeding: %ud", (*data)->GA_Seed);
    printf("\nLinear scaling: %d",                    (*data)->GA_LinScale);
    printf("\nPrinting level: %d\n\n",                (*data)->GA_Printing);
  }

  return result;

}




