#include "myarg.h"

static int FindArg(int argc, char **argv, char *opt);

/* ------------------------------------------------------------------ */

/* find the specified option string, like "-i" */
static int FindArg(int argc, char **argv, char *opt)
{
  int i;
  
  for(i = 0; i < argc; i++)
  {
    if(strcmp(argv[i], opt) == 0)
      return i;
  }

  return -1;
}

/* check whether an option has been specified */
int GetSimpleArg(int argc, char **argv, char *opt)
{
  int i;
  
  i = FindArg(argc, argv, opt);
  
  if(i < 0)
    return 0;
  else
    return 1;
}

/* read a string after an option */
int GetStringArg(int argc, char **argv, char *opt, char **value)
{
  int i;
  
  i = FindArg(argc, argv, opt);
  
  if(i < 0)
    return 0;
  
  if(i+1 >= argc)
  {
    fprintf(stderr, "\nERROR: Missing string value for option ``%s''!\n", opt);
    return 0;
  }
  
  *value = (char *)calloc( strlen(argv[i+1])+1, sizeof(char) );
  strcpy(*value, argv[i+1]);
  
  return 1;
}

/* read an integer after an option */
int GetIntegerArg(int argc, char **argv, char *opt, int *value)
{
  int  i, ret;
  char c;
  
  i = FindArg(argc, argv, opt);
  
  if(i < 0)
    return 0;
  
  if(i+1 >= argc)
  {
    fprintf(stderr, "\nERROR: Missing integer value for option ``%s''!\n", opt);
    return 0;
  }
  
  ret = sscanf(argv[i+1], "%d%c", value, &c);

  if(ret == 2)
  {
    if(c != ' ' && c != '\t' && c != '\n')
    {
      fprintf(stderr, "\nERROR: Option ``%s'' is followed by a non-integer value!\n", opt);
      return 0;
    }
  }
  else if(ret == 0)
  {
    fprintf(stderr, "\nERROR: Cannot read integer value for option ``%s''!\n", opt);
    return 0;
  }
  else if(ret != 1)
  {
    fprintf(stderr, "\nERROR: Option ``%s'' is followed by a non-integer value!\n", opt);
    return 0;
  }
  
  return 1;
}

/* read a double after an option */
int GetDoubleArg(int argc, char **argv, char *opt, double *value)
{
  int  i, ret;
  char c;
  
  i = FindArg(argc, argv, opt);
  
  if(i < 0)
    return 0;
  
  if(i+1 >= argc)
  {
    fprintf(stderr, "\nERROR: Missing value for option ``%s''\n", opt);
    return 0;
  }

  ret = sscanf(argv[i+1], "%lf%c", value, &c);

  if(ret == 2)
  {
    if(c != ' ' && c != '\t' && c != '\n')
    {
      fprintf(stderr, "\nERROR: Option ``%s'' is followed by a non-double value!\n", opt);
      return 0;
    }
  }
  else if(ret == 0)
  {
    fprintf(stderr, "\nERROR: Cannot read double value for option ``%s''!\n", opt);
    return 0;
  }
  else if(ret != 1)
  {
    fprintf(stderr, "\nERROR: Option ``%s'' is followed by a non-double value!\n", opt);
    return 0;
  }

  return 1;
}



