dnl @synopsis E_MPI
dnl Provide configuration options to compile auxiliary support 
dnl to use E-Lib with MPI
dnl
dnl @author Roman Putanowicz <putanowr@twins.pk.edu.pl>
dnl @version $Id: e_mpi.m4,v 1.1.1.1 2002/10/24 20:10:31 peteri Exp $ 

# _E_MPI_PATH_DIRECT
# --------------------
# Internal subroutine of _E_MPI_PATH
# Set e_mpi_includes and e_mpi_lib and e_mpi_bin by directly checking specified
# directories
m4_define([_E_MPI_PATH_DIRECT],[
e_mpi_headers_dirs='
/usr/include
/usr/local/include'

e_mpi_lib_dirs='
/usr/lib
/usr/local/lib'

e_mpi_bin_dirs='
/usr/bin/
/usr/local/bin'

for ac_dir in $e_mpi_include $e_mpi_headers_dirs; do
  if test -r "$ac_dir/mpi.h"; then
     e_mpi_include=$ac_dir
     break
  fi  
done

for ac_dir in $e_mpi_lib $e_mpi_lib_dirs
do
  for ac_extension in a so sl; do
    if test -r $ac_dir/libmpi.$ac_extension; then
        e_mpi_libe=$ac_dir
         break
    fi
    if test -r $ac_dir/libmpich.$ac_extension; then
        e_mpi_lib=$ac_dir
        break
    fi
  done
  if test x"$e_mpi_lib" != x; then 
    break
  fi
done

for ac_dir in $e_mpi_bin $e_mpi_bin_dirs
do
  if test -r "$ac_dir/mpicc"; then
     e_mpi_bin=$ac_dir
     break
  fi
done
]) # _E_MPI_PATH_DIRECT


# _E_MPI_PATH
# -----------
AC_DEFUN([_E_MPI_PATH],[_E_MPI_PATH_DIRECT])
  
# E_MPI
# ----------
# If MPI was found set the vars e_mpi_cc and e_mpi_include to
# the paths otherwise set no_e_mpi=yes
#
AC_DEFUN([E_MPI],
[
m4_divert_once([HELP_WITH], [
Specify if to add MIP support to E-Lib and where to find MPI
AC_HELP_STRING([--with-mpi],
[try to guess where mpi was installed])
AC_HELP_STRING([--with-mpi-prefix=DIR],
[MPI installed in DIR/bin, DIR/inclued and DIR/lib])
AC_HELP_STRING([--with-mpi-bin=DIR],
[MPI binaries insatlled in DIR])
AC_HELP_STRING([--with-mpi-include=DIR],
[MPI header files installed in DIR])
AC_HELP_STRING([--with-mpi-lib=DIR],
[MPI library files installed in DIR])
])
dnl
AC_ARG_WITH([mpi],[],e_have_mpi=yes,[])
dnl
AC_ARG_WITH([mpi-prefix],[],e_mpi_prefix=$withval
e_have_mpi=yes,[])
dnl
AC_ARG_WITH([mpi-bin],[],e_mpi_bin=$withval
e_have_mpi=yes,[])
dnl
AC_ARG_WITH([mpi-include],[],e_mpi_include=$withval
e_have_mpi=yes,[])
dnl
AC_ARG_WITH([mpi-lib],[],e_mpi_lib=$withval
e_have_mpi=yes,[])
dnl
AC_MSG_CHECKING([for MPI])
if test x"$e_have_mpi" != x; then 
  # try to guess the MPI path
  _E_MPI_PATH

  # override the default gues

  if test x"$e_mpi_prefix" != x; then
     e_mpi_include=$e_mpi_prefix/include
     e_mpi_lib=$e_mpi_prefix/lib
     e_mpi_bin=$e_mpi_prefix/bin
  fi

  if test ! -r "$e_mpi_include/mpi.h"; then
        e_with_mpi=no 
        unset e_mpi_include
  fi  
  if test x"$e_with_mpi" != xno; then
    for ac_extension in a so sl 
    do
      if test -r $e_mpi_lib/libmpi.$ac_extension; then
         e_mpi_lib_name=mpi
         break
      fi
      if test -r $e_mpi_lib/libmpich.$ac_extension; then
        e_mpi_lib_name=mpich
        break
      fi
    done
    if test x"$e_mpi_lib_name" = x; then
      e_with_mpi=no
      unset e_mpi_lib
    else
      save_LDFLAGS="$LDFLAGS"
      LDFLAGS=-L$e_mpi_lib
      save_LIBS="$LIBS"
      AC_CHECK_LIB($e_mpi_lib_name, MPI_Init,[], 
       AC_MSG_ERROR(
       [Could not find MPI_Init in -L$e_mpi_lib -l$e_mpi_lib_name]),[])
       LDFLAGS="$save_LDFLAGS"
       LIBS="$save_LIBS"
    fi
  fi
  if test x"$e_with_mpi" != xno -a ! -r "$e_mpi_bin/mpicc"; then
        e_with_mpi=no 
        unset e_mpi_bin
  fi  
else
  e_have_mpi="skipping" 
  AC_MSG_RESULT([$e_have_mpi])  
fi
if test x"$e_with_mpi" = xno; then
  if test x"$e_mpi_include" = x; then
     AC_MSG_ERROR([Could not find MPI header files])
  fi
  if test x"$e_mpi_lib" = x; then
     AC_MSG_ERROR([Could not find MPI library file])
  fi
  if test x"$e_mpi_bin" = x; then
     AC_MSG_ERROR([Could not find MPI binaries])
  fi
fi

E_MPI_INCLUDE="-I$e_mpi_include"
E_MPI_LIB="-L$e_mpi_lib -l$e_mpi_lib_name"
E_MPI_BIN="$e_mpi_bin"
AC_SUBST(E_MPI_INCLUDE)
AC_SUBST(E_MPI_LIB)
AC_SUBST(E_MPI_BIN)
])# E_MPI 
