/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                    e_GetDecompData(), e_PutDecompData()                    */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include "e_pltfrm.h"
#include <stdio.h>
#include <string.h>
#include <e_lib.h>
#include "comsrc.h"
#include "decomp.h"


/* ////////////////////////////////////////////////////////////////////////// */
/* e_GetDecompData: To read in mesh decomposition data. */
/* The return value is 1 if no input errors occur, else 0. */


int e_GetDecompData(MESH *mesh, char *InputFile)
{
   FullPathName = OpenInputFile(InputFile,E_DOM,&fp);

   if (FullPathName == NULL || !Initialise(mesh))
   {
      e_free(FullPathName);
      return 0;
   }

   if (GetData(mesh) && AllCompulsoryKeywordsSet())
   {
      SetMeshData(mesh);
      e_free(FullPathName);
      fclose(fp);
      return 1;
   }
   else
   {
      DeallocateAllMemory();
      fclose(fp);
      return 0;
   }
}


int e_SetupDecompData(MESH *mesh)
{
   int sum = 0;
   int i;

   mesh->SubStructSet[E_SS_NUM_DECOMP] = TRUE;
   mesh->Decomp.SubDoms = e_alloc_i_vect(mesh->TotalNElems);
   mesh->Decomp.NSubDoms = 1;
   for(i = 0; i < mesh->TotalNElems; i++)
   {
     mesh->Decomp.SubDoms[i] = 1;
   }

   if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
   {
      mesh->Decomp.SubDomsLink1 = mesh->Decomp.SubDoms + sum;
      sum += mesh->NElemsLink1;
   }

   if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2]) 
   {
      mesh->Decomp.SubDomsLink2 = mesh->Decomp.SubDoms + sum;
      sum += mesh->NElemsLink2;
   }

   if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3]) 
   {
      mesh->Decomp.SubDomsLink3 = mesh->Decomp.SubDoms + sum;
      sum += mesh->NElemsLink3;
   }

   if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4]) 
   {
      mesh->Decomp.SubDomsLink4 = mesh->Decomp.SubDoms + sum;
      sum += mesh->NElemsLink4;
   }

   if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5]) 
   {
      mesh->Decomp.SubDomsLink5 = mesh->Decomp.SubDoms + sum;
      sum += mesh->NElemsLink5;
   }

   if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
   {
      mesh->Decomp.SubDomsTriang1 = mesh->Decomp.SubDoms + sum;
      sum += mesh->NElemsTriang1;
   }

   if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2]) 
   {
      mesh->Decomp.SubDomsTriang2 = mesh->Decomp.SubDoms + sum;
      sum += mesh->NElemsTriang2;
   }

   if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3]) 
   {
      mesh->Decomp.SubDomsTriang3 = mesh->Decomp.SubDoms + sum;
      sum += mesh->NElemsTriang3;
   }

   if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4]) 
   {
      mesh->Decomp.SubDomsTriang4 = mesh->Decomp.SubDoms + sum;
      sum += mesh->NElemsTriang4;
   }

   if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5]) 
   {
      mesh->Decomp.SubDomsTriang5 = mesh->Decomp.SubDoms + sum;
      sum += mesh->NElemsTriang5;
   }

   if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1]) 
   {
      mesh->Decomp.SubDomsQuad1 = mesh->Decomp.SubDoms + sum;
      sum += mesh->NElemsQuad1;
   }

   if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3]) 
   {
      mesh->Decomp.SubDomsQuad3 = mesh->Decomp.SubDoms + sum;
      sum += mesh->NElemsQuad3;
   }

   if(mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1]) 
   {
      mesh->Decomp.SubDomsTetrah1 = mesh->Decomp.SubDoms + sum;
      sum += mesh->NElemsTetrah1;
   }

   if(mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1]) 
   {
      mesh->Decomp.SubDomsBlock1 = mesh->Decomp.SubDoms + sum;
      sum += mesh->NElemsBlock1;
   }

   return 1;
}



/* ////////////////////////////////////////////////////////////////////////// */
/* initialise: To initialise variables. */


static int Initialise(MESH *mesh)
{
   int sum = 0;


   NElemsLink1 = mesh->NElemsLink1;
   NElemsLink2 = mesh->NElemsLink2;
   NElemsLink3 = mesh->NElemsLink3;
   NElemsLink4 = mesh->NElemsLink4;
   NElemsLink5 = mesh->NElemsLink5;
   NElemsTriang1 = mesh->NElemsTriang1;
   NElemsTriang2 = mesh->NElemsTriang2;
   NElemsTriang3 = mesh->NElemsTriang3;
   NElemsTriang4 = mesh->NElemsTriang4;
   NElemsTriang5 = mesh->NElemsTriang5;
   NElemsQuad1 = mesh->NElemsQuad1;
   NElemsQuad3 = mesh->NElemsQuad3;
   NElemsTetrah1 = mesh->NElemsTetrah1;
   NElemsBlock1 = mesh->NElemsBlock1;
   TotalNElems = mesh->TotalNElems;
   LineNumber = 0;
   AllCompKeywordsSet = FALSE;

   SubDoms = e_alloc_i_vect(TotalNElems);

   if (SubDoms == NULL)
      return 0;

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
   {
      SubDomsLink1 = SubDoms + sum;
      sum += mesh->NElemsLink1;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2]) 
   {
      SubDomsLink2 = SubDoms + sum;
      sum += mesh->NElemsLink2;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3]) 
   {
      SubDomsLink3 = SubDoms + sum;
      sum += mesh->NElemsLink3;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4]) 
   {
      SubDomsLink4 = SubDoms + sum;
      sum += mesh->NElemsLink4;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5]) 
   {
      SubDomsLink5 = SubDoms + sum;
      sum += mesh->NElemsLink5;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
   {
      SubDomsTriang1 = SubDoms + sum;
      sum += mesh->NElemsTriang1;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2]) 
   {
      SubDomsTriang2 = SubDoms + sum;
      sum += mesh->NElemsTriang2;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3]) 
   {
      SubDomsTriang3 = SubDoms + sum;
      sum += mesh->NElemsTriang3;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4]) 
   {
      SubDomsTriang4 = SubDoms + sum;
      sum += mesh->NElemsTriang4;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5]) 
   {
      SubDomsTriang5 = SubDoms + sum;
      sum += mesh->NElemsTriang5;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1]) 
   {
      SubDomsQuad1 = SubDoms + sum;
      sum += mesh->NElemsQuad1;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3]) 
   {
      SubDomsQuad3 = SubDoms + sum;
      sum += mesh->NElemsQuad3;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1]) 
   {
      SubDomsTetrah1 = SubDoms + sum;
      sum += mesh->NElemsTetrah1;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1]) 
   {
      SubDomsBlock1 = SubDoms + sum;
      sum += mesh->NElemsBlock1;
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetData: To read in the input data. */

/* The return value is 1 if the data are read successfully or the end of file
   is encountered, else 0.                                                    */


static int GetData(MESH *mesh)
{
   char word[E_LINE_MAX + 1];      /* potential keyword */


   if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
      return 1;
   else if (   !WordMatchesKeyword(E_KW_NSUBDOMS,word)
            || !GetNSubDoms()                         )
      return 0;

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_SUBDOMS_LINK1,word)
               || !GetSubDoms(NElemsLink1,SubDomsLink1)       )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_SUBDOMS_LINK2,word)
               || !GetSubDoms(NElemsLink2,SubDomsLink2)       )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_SUBDOMS_LINK3,word)
               || !GetSubDoms(NElemsLink3,SubDomsLink3)       )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_SUBDOMS_LINK4,word)
               || !GetSubDoms(NElemsLink4,SubDomsLink4)       )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_SUBDOMS_LINK5,word)
               || !GetSubDoms(NElemsLink5,SubDomsLink5)       )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_SUBDOMS_TRIANG1,word)
               || !GetSubDoms(NElemsTriang1,SubDomsTriang1)     )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_SUBDOMS_TRIANG2,word)
               || !GetSubDoms(NElemsTriang2,SubDomsTriang2)     )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_SUBDOMS_TRIANG3,word)
               || !GetSubDoms(NElemsTriang3,SubDomsTriang3)     )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_SUBDOMS_TRIANG4,word)
               || !GetSubDoms(NElemsTriang4,SubDomsTriang4)     )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_SUBDOMS_TRIANG5,word)
               || !GetSubDoms(NElemsTriang5,SubDomsTriang5)     )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_SUBDOMS_QUAD1,word)
               || !GetSubDoms(NElemsQuad1,SubDomsQuad1)       )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_SUBDOMS_QUAD3,word)
               || !GetSubDoms(NElemsQuad3,SubDomsQuad3)       )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_SUBDOMS_TETRAH1,word)
               || !GetSubDoms(NElemsTetrah1,SubDomsTetrah1)     )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_SUBDOMS_BLOCK1,word)
               || !GetSubDoms(NElemsBlock1,SubDomsBlock1)      )
         return 0;
   }

   if (GetKeyword(line,word,&LineNumber,FullPathName,fp) != EOF)
   {
      e_FileLineError(LineNumber,FullPathName,"undefined keyword");
      return 0;
   }

   AllCompKeywordsSet = TRUE;
   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* WordMatchesKeyword: To return whether a word is a specified keyword. */

/* The return value is 1 if the word matches the keyword, else 0. */


static int WordMatchesKeyword(char *Keyword, char *word)
{
   if (strcmp(Keyword,word) != 0)
   {
      if (IsAKeyword(word))
         e_FileLineError(LineNumber,FullPathName,"keyword out of sequence");
      else
         e_FileLineError(LineNumber,FullPathName,"undefined keyword");

      return 0;
   }
   else
      return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* Get... input functions: To read in data corresponding to a keyword. */

/* The return value is 1 if the data are read successfully, else 0. */


static int GetNSubDoms()
{
   if (sscanf(line,"%*s %d",&NSubDoms) == 1)
   {
      if (NSubDoms < 1)
      {
         e_FileLineError(LineNumber,FullPathName,
                         "ill-defined number of subdomains");
         return 0;
      }
   }
   else
   {
      e_FileLineError(LineNumber,FullPathName,
                      "cannot read number of subdomains");
      return 0;
   }

   return 1;
}


/* ////////////////////////////////// */


static int GetSubDoms(int NElements, int *SubDoms)
{
   int i;
   int SubdomainIndex;


   for (i = 0; i < NElements; i++)
   {
      if (e_GetNextLine(line,&LineNumber,FullPathName,fp) == 0)
         return 0;

      if (sscanf(line,"%d %d",&SubdomainIndex,&SubDoms[i]) == 2)
      {
         if (SubdomainIndex != i + 1)
         {
            e_FileLineError(LineNumber,FullPathName,
                            "element index out of sequence");
            return 0;
         }

         if (SubDoms[i] < 1 || SubDoms[i] > NSubDoms)
         {
            e_FileLineError(LineNumber,FullPathName,
                            "ill-defined subdomain index");
            return 0;
         }
      }
      else
      {
         e_FileLineError(LineNumber,FullPathName,
                         "cannot read subdomain index");
         return 0;
      }
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* IsAKeyword: To return whether a word is a keyword. */

/* The return value is 1 if "word" is a keyword, else 0. */


static int IsAKeyword(char *word)
{
   if (          strcmp(E_KW_NSUBDOMS,word) == 0
       ||   strcmp(E_KW_SUBDOMS_LINK1,word) == 0
       ||   strcmp(E_KW_SUBDOMS_LINK2,word) == 0
       ||   strcmp(E_KW_SUBDOMS_LINK3,word) == 0
       ||   strcmp(E_KW_SUBDOMS_LINK4,word) == 0
       ||   strcmp(E_KW_SUBDOMS_LINK5,word) == 0
       || strcmp(E_KW_SUBDOMS_TRIANG1,word) == 0
       || strcmp(E_KW_SUBDOMS_TRIANG2,word) == 0
       || strcmp(E_KW_SUBDOMS_TRIANG3,word) == 0
       || strcmp(E_KW_SUBDOMS_TRIANG4,word) == 0
       || strcmp(E_KW_SUBDOMS_TRIANG5,word) == 0
       ||   strcmp(E_KW_SUBDOMS_QUAD1,word) == 0
       ||   strcmp(E_KW_SUBDOMS_QUAD3,word) == 0
       || strcmp(E_KW_SUBDOMS_TETRAH1,word) == 0
       ||  strcmp(E_KW_SUBDOMS_BLOCK1,word) == 0)
   {
      return TRUE;
   }
   else
   {
      return FALSE;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* AllCompulsorySet: To return whether all compulsory keywords have been set. */


static int AllCompulsoryKeywordsSet()
{
   if (!AllCompKeywordsSet)
   {
      e_FileError(FullPathName,"unexpected end of file");
      return 0;
   }
   else
      return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetMeshData: To set the MESH structure data. */


static void SetMeshData(MESH *mesh)
{
   mesh->Decomp.NSubDoms = NSubDoms;
   mesh->Decomp.SubDoms = SubDoms;
   mesh->Decomp.SubDomsLink1 = SubDomsLink1;
   mesh->Decomp.SubDomsLink2 = SubDomsLink2;
   mesh->Decomp.SubDomsLink3 = SubDomsLink3;
   mesh->Decomp.SubDomsLink4 = SubDomsLink4;
   mesh->Decomp.SubDomsLink5 = SubDomsLink5;
   mesh->Decomp.SubDomsTriang1 = SubDomsTriang1;
   mesh->Decomp.SubDomsTriang2 = SubDomsTriang2;
   mesh->Decomp.SubDomsTriang3 = SubDomsTriang3;
   mesh->Decomp.SubDomsTriang4 = SubDomsTriang4;
   mesh->Decomp.SubDomsTriang5 = SubDomsTriang5;
   mesh->Decomp.SubDomsQuad1 = SubDomsQuad1;
   mesh->Decomp.SubDomsQuad3 = SubDomsQuad3;
   mesh->Decomp.SubDomsTetrah1 = SubDomsTetrah1;
   mesh->Decomp.SubDomsBlock1 = SubDomsBlock1;

   mesh->SubStructSet[E_SS_NUM_DECOMP] = TRUE;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* DeallocateAllMemory: To deallocate all memory. */


static void DeallocateAllMemory()
{
   e_free(SubDoms);
   e_free(FullPathName);
}


/* ////////////////////////////////////////////////////////////////////////// */










/* ////////////////////////////////////////////////////////////////////////// */
/* e_PutDecompData: To output mesh decomposition data. */
/* The return value is 1 if the data are written successfully, else 0. */


int e_PutDecompData(MESH *mesh, char *OutputFile)
{
   int i;

   FILE *fp;            /* output file pointer */


   if (!mesh->SubStructSet[E_SS_NUM_DECOMP])
   {
      e_GeneralError("e_PutDecompData: decomposition data not set");
      return 0;
   }
   else if (OpenOutputFile(OutputFile,E_DOM,&fp) == NULL)
   {
      return 0;
   }

   fprintf(fp,"%s %d\n",E_KW_NSUBDOMS,mesh->Decomp.NSubDoms);

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
   {
      fprintf(fp,"\n\n# LINK1 element subdomains\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#       subdomain index\n");

      fprintf(fp,"\n%s\n",E_KW_SUBDOMS_LINK1);

      for (i = 0; i < mesh->NElemsLink1; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%5d\n",mesh->Decomp.SubDomsLink1[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
   {
      fprintf(fp,"\n\n# LINK2 element subdomains\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#       subdomain index\n");

      fprintf(fp,"\n%s\n",E_KW_SUBDOMS_LINK2);

      for (i = 0; i < mesh->NElemsLink2; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%5d\n",mesh->Decomp.SubDomsLink2[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
   {
      fprintf(fp,"\n\n# LINK3 element subdomains\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#       subdomain index\n");

      fprintf(fp,"\n%s\n",E_KW_SUBDOMS_LINK3);

      for (i = 0; i < mesh->NElemsLink3; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%5d\n",mesh->Decomp.SubDomsLink3[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
   {
      fprintf(fp,"\n\n# LINK4 element subdomains\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#       subdomain index\n");

      fprintf(fp,"\n%s\n",E_KW_SUBDOMS_LINK4);

      for (i = 0; i < mesh->NElemsLink4; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%5d\n",mesh->Decomp.SubDomsLink4[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
   {
      fprintf(fp,"\n\n# LINK5 element subdomains\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#       subdomain index\n");

      fprintf(fp,"\n%s\n",E_KW_SUBDOMS_LINK5);

      for (i = 0; i < mesh->NElemsLink5; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%5d\n",mesh->Decomp.SubDomsLink5[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
   {
      fprintf(fp,"\n\n# TRIANG1 element subdomains\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#       subdomain index\n");

      fprintf(fp,"\n%s\n",E_KW_SUBDOMS_TRIANG1);

      for (i = 0; i < mesh->NElemsTriang1; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%5d\n",mesh->Decomp.SubDomsTriang1[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
   {
      fprintf(fp,"\n\n# TRIANG2 element subdomains\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#       subdomain index\n");

      fprintf(fp,"\n%s\n",E_KW_SUBDOMS_TRIANG2);

      for (i = 0; i < mesh->NElemsTriang2; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%5d\n",mesh->Decomp.SubDomsTriang2[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
   {
      fprintf(fp,"\n\n# TRIANG3 element subdomains\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#       subdomain index\n");

      fprintf(fp,"\n%s\n",E_KW_SUBDOMS_TRIANG3);

      for (i = 0; i < mesh->NElemsTriang3; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%5d\n",mesh->Decomp.SubDomsTriang3[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
   {
      fprintf(fp,"\n\n# TRIANG4 element subdomains\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#       subdomain index\n");

      fprintf(fp,"\n%s\n",E_KW_SUBDOMS_TRIANG4);

      for (i = 0; i < mesh->NElemsTriang4; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%5d\n",mesh->Decomp.SubDomsTriang4[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
   {
      fprintf(fp,"\n\n# TRIANG5 element subdomains\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#       subdomain index\n");

      fprintf(fp,"\n%s\n",E_KW_SUBDOMS_TRIANG5);

      for (i = 0; i < mesh->NElemsTriang5; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%5d\n",mesh->Decomp.SubDomsTriang5[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
   {
      fprintf(fp,"\n\n# QUAD1 element subdomains\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#       subdomain index\n");

      fprintf(fp,"\n%s\n",E_KW_SUBDOMS_QUAD1);

      for (i = 0; i < mesh->NElemsQuad1; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%5d\n",mesh->Decomp.SubDomsQuad1[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
   {
      fprintf(fp,"\n\n# QUAD3 element subdomains\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#       subdomain index\n");

      fprintf(fp,"\n%s\n",E_KW_SUBDOMS_QUAD3);

      for (i = 0; i < mesh->NElemsQuad3; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%5d\n",mesh->Decomp.SubDomsQuad3[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
   {
      fprintf(fp,"\n\n# TETRAH1 element subdomains\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#       subdomain index\n");

      fprintf(fp,"\n%s\n",E_KW_SUBDOMS_TETRAH1);

      for (i = 0; i < mesh->NElemsTetrah1; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%5d\n",mesh->Decomp.SubDomsTetrah1[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
   {
      fprintf(fp,"\n\n# BLOCK1 element subdomains\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#       subdomain index\n");

      fprintf(fp,"\n%s\n",E_KW_SUBDOMS_BLOCK1);

      for (i = 0; i < mesh->NElemsBlock1; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%5d\n",mesh->Decomp.SubDomsBlock1[i]);
      }
   }

   fclose(fp);
   return 1;
}
