/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                            The E-Lib Functions                             */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include "e_pltfrm.h"
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#ifdef __E_PVM__
#include <pvm3.h>
#endif

#ifdef __E_MSWIN__
#include <windows.h>
#endif

#ifdef __E_T800RTL__
#include <par.h>
#include <sema.h>
#endif

#ifdef __E_T800SAL__
#include <par.h>
#include <sema.h>
#endif

#include <e_lib.h>


#ifndef __E_T800SAL__

/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                          Input-Output Functions                            */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#ifdef __E_MSWIN__
#  if _MSC_VER
/* Microsoft Visual Studio defines it */
#    define vsnprintf _vsnprintf
#  else
/* In other case we may not have the vsnprintf implementation */
#    define vsnprintf(buf, size, format, arg) vsprintf(buf, format, arg)
#  endif
#endif


/* ////////////////////////////////////////////////////////////////////////// */
/* e_GetNextLine: To read in the next data line from a file. */

/* The return value is 1 if successful, else 0. */


int e_GetNextLine(char *line, int *LineNumber, char *FullPathName, FILE *fp)
{
   int ReturnValue;             /* return value of 'e_GetLine' */


   if ((ReturnValue = e_GetLine(line,E_LINE_MAX,LineNumber,fp)) == EOF)
   {
      e_FileError(FullPathName,"unexpected end of file");
      return 0;
   }
   else if (ReturnValue == 0)
   {
      e_FileError(FullPathName,"system read error");
      return 0;
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_GetLine: To read a line of text of up to "MaxLineLength" characters from
   an input stream into an array. Lines whose first non-white-space character
   is the macro character COMMENT_CHARACTER are ignored, the file being read
   until a non-comment line or end of file is encountered. The line count
   "LineNumber" is incremented appropriately. The return value is 1 if the
   array is filled successfully, EOF if the end of file is encountered before
   a non-comment line, else 0.                                                */


#define COMMENT_CHARACTER  '#'


int e_GetLine(char *line, int MaxLineLength, int *LineNumber, FILE *fp)
{
   char FirstNonWhiteCharacter;
   char InsideComment;
   char *pLine;


   InsideComment = FALSE;
   FirstNonWhiteCharacter = COMMENT_CHARACTER;

   while (   InsideComment
          || FirstNonWhiteCharacter == COMMENT_CHARACTER
          || FirstNonWhiteCharacter == '\0'             )
   {
      if ((pLine = fgets(line,MaxLineLength + 1,fp)) != NULL)
      {
         if (strncmp(line,"/*",2) == 0)
         {
            InsideComment = TRUE;
         }
         else if (InsideComment && strncmp(line,"*/",2) == 0)
         {
            (*LineNumber)++;
            return e_GetLine(line,MaxLineLength,LineNumber,fp);
         }

         if (strchr(line,'\n') == NULL)
            while (getc(fp) != '\n' && !feof(fp));

         while (isspace((int)(*pLine)))
            pLine++;

         FirstNonWhiteCharacter = *pLine;
         (*LineNumber)++;
      }
      else
      {
         if (feof(fp))
            return EOF;
         else
            return 0;
      }
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_beep: To output a beep. */


void e_beep()
{

#ifndef __E_XWIN__
#ifndef __E_MSWIN__
   printf("\a");
#endif
#endif

#ifdef __E_XWIN__
   /* To Do - Put X-Windows code here */
#endif

#ifdef __E_MSWIN__
   MessageBeep(MB_OK);
#endif

}


#endif  /* ifndef __E_T800SAL__ */


/* ////////////////////////////////////////////////////////////////////////// */
/* e_PlatformMessage: To display a message for the appropriate platform.      */


void e_PlatformMessage(char *message)
{

#ifdef __E_CLANSIC__
   printf(message);
   fflush(stdout);
#endif

#ifdef __E_XWIN__
   /* To Do - Put X-Windows code here */
   printf(message);
   fflush(stdout);
#endif

#ifdef __E_MSWIN__
   MessageBox(NULL,message,"E-Lib Message",
              MB_OK | MB_ICONEXCLAMATION | MB_TASKMODAL);
#endif

#ifdef __E_T800RTL__
   par_fprintf(stdout,message);
   sema_wait(&par_sema);
   fflush(stdout);
   sema_signal(&par_sema);
#endif

#ifdef __E_T800SAL__
   return;
#endif

}


/* ////////////////////////////////////////////////////////////////////////// */
/* ErrorWarningMessage: To display an error/warning message for the
   appropriate platform.                                                      */


static void ErrorWarningMessage(char *message)
{

#ifdef __E_CLANSIC__
   fprintf(stderr,message);
#endif

#ifdef __E_XWIN__
   /* To Do - Put X-Windows code here */
   fprintf(stderr,message);
#endif

#ifdef __E_MSWIN__
   MessageBox(NULL,message,"E-Lib Message",
              MB_OK | MB_ICONEXCLAMATION | MB_TASKMODAL);
#endif

#ifdef __E_T800RTL__
   par_fprintf(stderr,message);
   sema_wait(&par_sema);
   fflush(stderr);
   sema_signal(&par_sema);
#endif

#ifdef __E_T800SAL__
   return;
#endif

}


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                         Warning Message Functions                          */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_GeneralWarning: General warning. */


void e_GeneralWarning(char *Warning, ...)
{
   char message[E_LINE_MAX + 1];
   char prefix[] = "WARNING: ";


#ifdef __E_CLANSIC__
   va_list ap;
   strcpy(message, "\n");
   strcat(message, prefix);
   va_start(ap, Warning);
#ifdef __SYSTEM_IRIX__
   vsprintf (message+sizeof(prefix), Warning, ap);
#else
   vsnprintf (message+sizeof(prefix), E_LINE_MAX-1, Warning, ap);
#endif
   va_end(ap);
   strcat(message, "\n");
#endif

#ifdef __E_XWIN__
   va_list ap;
   strcpy(message, "\n");
   strcat(message, prefix);
   va_start(ap, Warning);
#ifdef __SYSTEM_IRIX__
   vsprintf (message+sizeof(prefix), Warning, ap);
#else
   vsnprintf (message+sizeof(prefix), E_LINE_MAX-1, Warning, ap);
#endif
   va_end(ap);
   strcat(message, "\n");
#endif

#ifdef __E_MSWIN__
   va_list ap;
   strcat(message, prefix);
   va_start(ap, Warning);
   vsnprintf (message+sizeof(prefix)-1, E_LINE_MAX, Warning, ap);
   va_end(ap);
#endif

#ifdef __E_T800RTL__
   sema_wait(&par_sema);
   sprintf(message, "\nWARNING: %s\n", Warning);
   sema_signal(&par_sema);
#endif

   ErrorWarningMessage(message);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_FileWarning: File warning. */


void e_FileWarning(char *FullPathName, char *Warning)
{
   char message[E_LINE_MAX + 1];


#ifdef __E_CLANSIC__
   sprintf(message, "\nWARNING: %s: '%s'\n", Warning,FullPathName);
#endif

#ifdef __E_XWIN__
   sprintf(message, "\nWARNING: %s: '%s'\n", Warning,FullPathName);
#endif

#ifdef __E_MSWIN__
   sprintf(message, "WARNING: %s: '%s'", Warning,FullPathName);
#endif

#ifdef __E_T800RTL__
   sema_wait(&par_sema);
   sprintf(message, "\nWARNING: %s: '%s'\n", Warning,FullPathName);
   sema_signal(&par_sema);
#endif

   ErrorWarningMessage(message);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_LineWarning: Line warning. */


void e_LineWarning(int LineNumber, char *Warning)
{
   char message[E_LINE_MAX + 1];


#ifdef __E_CLANSIC__
   sprintf(message, "\nWARNING: %s: line %d\n",Warning,LineNumber);
#endif

#ifdef __E_XWIN__
   sprintf(message, "\nWARNING: %s: line %d\n",Warning,LineNumber);
#endif

#ifdef __E_MSWIN__
   sprintf(message, "WARNING: %s: line %d",Warning,LineNumber);
#endif

#ifdef __E_T800RTL__
   sema_wait(&par_sema);
   sprintf(message, "\nWARNING: %s: line %d\n",Warning,LineNumber);
   sema_signal(&par_sema);
#endif

   ErrorWarningMessage(message);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_FileLineWarning: File line warning. */


void e_FileLineWarning(int LineNumber, char *FullPathName, char *Warning)
{
   char message[E_LINE_MAX + 1];


#ifdef __E_CLANSIC__
   sprintf(message, "\nWARNING: %s: line %d: '%s'\n",
                    Warning,LineNumber,FullPathName);
#endif

#ifdef __E_XWIN__
   sprintf(message, "\nWARNING: %s: line %d: '%s'\n",
                    Warning,LineNumber,FullPathName);
#endif

#ifdef __E_MSWIN__
   sprintf(message, "WARNING: %s: line %d: '%s'",
                    Warning,LineNumber,FullPathName);
#endif

#ifdef __E_T800RTL__
   sema_wait(&par_sema);
   sprintf(message, "\nWARNING: %s: line %d: '%s'\n",
                    Warning,LineNumber,FullPathName);
   sema_signal(&par_sema);
#endif

   ErrorWarningMessage(message);
}


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                          Error Message Functions                           */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_GeneralError: General error. */


void e_GeneralError(char *ErrorMessage, ...)
{
   char message[E_LINE_MAX + 1];
   char prefix[] = "ERROR: ";

#ifdef __E_CLANSIC__
   va_list ap;
   strcpy(message, "\n");
   strcat(message, prefix);
   va_start(ap, ErrorMessage);
#ifdef __SYSTEM_IRIX__
   vsprintf (message+sizeof(prefix), ErrorMessage, ap);
#else
   vsnprintf (message+sizeof(prefix), E_LINE_MAX-1, ErrorMessage, ap);
#endif
   va_end(ap);
   strcat(message, "\n");
#endif

#ifdef __E_XWIN__
   va_list ap;
   strcpy(message, "\n");
   strcat(message, prefix);
   va_start(ap, ErrorMessage);
#ifdef __SYSTEM_IRIX__
   vsprintf (message+sizeof(prefix), ErrorMessage, ap);
#else
   vsnprintf (message+sizeof(prefix), E_LINE_MAX-1, ErrorMessage, ap);
#endif
   va_end(ap);
   strcat(message, "\n");
#endif

#ifdef __E_MSWIN__
   va_list ap;
   strcat(message, prefix);
   va_start(ap, ErrorMessage);
   vsnprintf (message+sizeof(prefix)-1, E_LINE_MAX, ErrorMessage, ap);
   va_end(ap);
#endif

#ifdef __E_T800RTL__
   sema_wait(&par_sema);
   sprintf(message, "\nERROR: %s\n", ErrorMessage);
   sema_signal(&par_sema);
#endif

   ErrorWarningMessage(message);

   e_exit(1);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_FileError: File error. */


void e_FileError(char *FullPathName, char *ErrorMessage)
{
   char message[E_LINE_MAX + 1];


#ifdef __E_CLANSIC__
   sprintf(message, "\nERROR: %s: '%s'\n", ErrorMessage,FullPathName);
#endif

#ifdef __E_XWIN__
   sprintf(message, "\nERROR: %s: '%s'\n", ErrorMessage,FullPathName);
#endif

#ifdef __E_MSWIN__
   sprintf(message, "ERROR: %s: '%s'", ErrorMessage,FullPathName);
#endif

#ifdef __E_T800RTL__
   sema_wait(&par_sema);
   sprintf(message, "\nERROR: %s: '%s'\n", ErrorMessage,FullPathName);
   sema_signal(&par_sema);
#endif

   ErrorWarningMessage(message);

   e_exit(1);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_LineError: Line error. */


void e_LineError(int LineNumber, char *ErrorMessage)
{
   char message[E_LINE_MAX + 1];


#ifdef __E_CLANSIC__
   sprintf(message, "\nERROR: %s: line %d\n",ErrorMessage,LineNumber);
#endif

#ifdef __E_XWIN__
   sprintf(message, "\nERROR: %s: line %d\n",ErrorMessage,LineNumber);
#endif

#ifdef __E_MSWIN__
   sprintf(message, "ERROR: %s: line %d",ErrorMessage,LineNumber);
#endif

#ifdef __E_T800RTL__
   sema_wait(&par_sema);
   sprintf(message, "\nERROR: %s: line %d\n",ErrorMessage,LineNumber);
   sema_signal(&par_sema);
#endif

   ErrorWarningMessage(message);

   e_exit(1);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_FileLineError: File line error. */


void e_FileLineError(int LineNumber, char *FullPathName, char *ErrorMessage)
{
   char message[E_LINE_MAX + 1];


#ifdef __E_CLANSIC__
   sprintf(message, "\nERROR: %s: line %d: '%s'\n",
             ErrorMessage,LineNumber,FullPathName);
#endif

#ifdef __E_XWIN__
   sprintf(message, "\nERROR: %s: line %d: '%s'\n",
           ErrorMessage,LineNumber,FullPathName);
#endif

#ifdef __E_MSWIN__
   sprintf(message, "ERROR: %s: line %d: '%s'",
           ErrorMessage,LineNumber,FullPathName);
#endif

#ifdef __E_T800RTL__
   sema_wait(&par_sema);
   sprintf(message, "\nERROR: %s: line %d: '%s'\n",
             ErrorMessage,LineNumber,FullPathName);
   sema_signal(&par_sema);
#endif

   ErrorWarningMessage(message);

   e_exit(1);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_SystemError: System error. */


void e_SystemError()
{

#ifdef __E_CLANSIC__
   ErrorWarningMessage("\nSystem error\n");
#endif

#ifdef __E_XWIN__
   ErrorWarningMessage("\nSystem error\n");
#endif

#ifdef __E_MSWIN__
   ErrorWarningMessage("System error");
#endif

#ifdef __E_T800RTL__
   ErrorWarningMessage("\nSystem error\n");
#endif

   e_exit(1);
}


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                       Memory Allocation Functions                          */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* The following functions allocate memory for a vector of length "length" and
   initialise it to zero. The return value is a pointer to the allocated
   memory if the allocation is successful, else NULL.                         */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_calloc: Generalised "calloc" function. */


#ifndef __E_T800__

void *e_calloc(unsigned int length, unsigned int NBytes)
{
   void *vector;


   if (length == 0 || NBytes == 0)
   {
      length = 1;
      NBytes = 1;
   }

   if ((vector = (void *)calloc(length,NBytes)) == NULL)
   {
      e_GeneralError("insufficient memory");
      return NULL;
   }
   else
     return vector;
}

#endif


#ifdef __E_T800__

void *e_calloc(unsigned int length, unsigned int NBytes)
{
   void *vector;


   if (length == 0 || NBytes == 0)
   {
      length = 1;
      NBytes = 1;
   }

   if ((vector = (void *)par_malloc(length*NBytes)) == NULL)
   {
      e_GeneralError("insufficient memory");
      return NULL;
   }
   else
   {
      memset(vector,0,length*NBytes);
      return vector;
   }
}

#endif


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_i_vect: "int" integer vector. */


int *e_alloc_i_vect(unsigned int length)
{
   int *vector;


   if ((vector = (int *)e_calloc(length, sizeof(int))) == NULL)
      return NULL;
   else
      return vector;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_f_vect: Float vector. */


float *e_alloc_f_vect(unsigned int length)
{
   float *vector;


   if ((vector = (float *)e_calloc(length, sizeof(float))) == NULL)
      return NULL;
   else
      return vector;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_d_vect: Double precision vector. */


double *e_alloc_d_vect(unsigned int length)
{
   double *vector;


   if ((vector = (double *)e_calloc(length, sizeof(double))) == NULL)
      return NULL;
   else
      return vector;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_c_vect: Character vector. */


char *e_alloc_c_vect(unsigned int length)
{
   char *vector;


   if ((vector = (char *)e_calloc(length, sizeof(char))) == NULL)
      return NULL;
   else
      return vector;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_vp_vect: Void pointer vector. */


void **e_alloc_vp_vect(unsigned int length)
{
   register unsigned int i;
   void **vector;


   if ((vector = (void **)e_calloc(length, sizeof(void *))) == NULL)
      return NULL;

   for (i = 0; i < length; i++)
      vector[i] = NULL;

   return vector;
}


/* ////////////////////////////////////////////////////////////////////////// */

/* The following functions allocate memory for a vectors of pointers. The
   return value is a pointer to the allocated memory if the allocation is
   successful, else NULL.                                                     */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_init_frag_i_matr: "int" integer matrix. */


int **e_init_frag_i_matr(unsigned int NRows)
{
   register unsigned int i;
   int **matrix;


   if ((matrix = (int **)e_calloc(NRows, sizeof(int *))) == NULL)
      return NULL;

   for (i = 0; i < NRows; i++)
      matrix[i] = NULL;

   return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_init_frag_f_matr: Float matrix. */


float **e_init_frag_f_matr(unsigned int NRows)
{
   register unsigned int i;
   float **matrix;


   if ((matrix = (float **)e_calloc(NRows, sizeof(float *))) == NULL)
      return NULL;

   for (i = 0; i < NRows; i++)
      matrix[i] = NULL;

   return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_init_frag_d_matr: Double precision matrix. */


double **e_init_frag_d_matr(unsigned int NRows)
{
   register unsigned int i;
   double **matrix;


   if ((matrix = (double **)e_calloc(NRows, sizeof(double *))) == NULL)
      return NULL;

   for (i = 0; i < NRows; i++)
      matrix[i] = NULL;

   return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_init_frag_c_matr: Character matrix. */


char **e_init_frag_c_matr(unsigned int NRows)
{
   register unsigned int i;
   char **matrix;


   if ((matrix = (char **)e_calloc(NRows, sizeof(char *))) == NULL)
      return NULL;

   for (i = 0; i < NRows; i++)
      matrix[i] = NULL;

   return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_init_frag_vp_matr: Void pointer matrix. */


void ***e_init_frag_vp_matr(unsigned int NRows)
{
   register unsigned int i;
   void ***matrix;


   if ((matrix = (void ***)e_calloc(NRows, sizeof(void **))) == NULL)
      return NULL;

   for (i = 0; i < NRows; i++)
      matrix[i] = NULL;

   return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_init_frag_i_3D_matr: 3D "int" integer matrix. */


int ***e_init_frag_i_3D_matr(unsigned int NDepth)
{
   register unsigned int i;
   int ***matrix3D;


   if ((matrix3D = (int ***)e_calloc(NDepth, sizeof(int **))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
      matrix3D[i] = NULL;

   return matrix3D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_init_frag_f_3D_matr: 3D float matrix. */


float ***e_init_frag_f_3D_matr(unsigned int NDepth)
{
   register unsigned int i;

   float ***matrix3D;


   if ((matrix3D = (float ***)e_calloc(NDepth, sizeof(float **))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
      matrix3D[i] = NULL;

   return matrix3D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_init_frag_d_3D_matr: 3D double precision "matrix. */


double ***e_init_frag_d_3D_matr(unsigned int NDepth)
{
   register unsigned int i;

   double ***matrix3D;


   if ((matrix3D = (double ***)e_calloc(NDepth, sizeof(double **))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
      matrix3D[i] = NULL;

   return matrix3D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_init_frag_c_3D_matr: 3D character matrix. */


char ***e_init_frag_c_3D_matr(unsigned int NDepth)
{
   register unsigned int i;

   char ***matrix3D;


   if ((matrix3D = (char ***)e_calloc(NDepth, sizeof(char **))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
      matrix3D[i] = NULL;

   return matrix3D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_init_frag_vp_3D_matr: 3D void pointer matrix. */


void ****e_init_frag_vp_3D_matr(unsigned int NDepth)
{
   register unsigned int i;

   void ****matrix3D;


   if ((matrix3D = (void ****)e_calloc(NDepth, sizeof(void ***))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
      matrix3D[i] = NULL;

   return matrix3D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_init_frag_i_4D_matr: 4D "int" integer matrix. */


int ****e_init_frag_i_4D_matr(unsigned int NDepth)
{
   register unsigned int i;
   int ****matrix4D;


   if ((matrix4D = (int ****)e_calloc(NDepth, sizeof(int ***))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
      matrix4D[i] = NULL;

   return matrix4D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_init_frag_f_4D_matr: 4D float matrix. */


float ****e_init_frag_f_4D_matr(unsigned int NDepth)
{
   register unsigned int i;

   float ****matrix4D;


   if ((matrix4D = (float ****)e_calloc(NDepth, sizeof(float ***))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
      matrix4D[i] = NULL;

   return matrix4D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_init_frag_d_4D_matr: 4D double precision matrix. */


double ****e_init_frag_d_4D_matr(unsigned int NDepth)
{
   register unsigned int i;

   double ****matrix4D;


   if ((matrix4D = (double ****)e_calloc(NDepth, sizeof(double ***))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
      matrix4D[i] = NULL;

   return matrix4D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_init_frag_c_4D_matr: 4D character matrix. */


char ****e_init_frag_c_4D_matr(unsigned int NDepth)
{
   register unsigned int i;

   char ****matrix4D;


   if ((matrix4D = (char ****)e_calloc(NDepth, sizeof(char ***))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
      matrix4D[i] = NULL;

   return matrix4D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_init_frag_vp_4D_matr: 4D void pointer matrix. */


void *****e_init_frag_vp_4D_matr(unsigned int NDepth)
{
   register unsigned int i;

   void *****matrix4D;


   if ((matrix4D = (void *****)e_calloc(NDepth, sizeof(void ****))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
      matrix4D[i] = NULL;

   return matrix4D;
}


/* ////////////////////////////////////////////////////////////////////////// */

/* The following functions allocate memory for an "NRows" x "NCols" matrix
   where each row is allocated separately. The matrix is initialised to zero.
   The return value is a pointer to the allocated memory if the allocation is
   successful, else NULL.                                                     */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_frag_i_matr: "int" integer matrix. */


int **e_alloc_frag_i_matr(unsigned int NRows, unsigned int NCols)
{
   register unsigned int i;
   int **matrix;


   if ((matrix = (int **)e_calloc(NRows, sizeof(int *))) == NULL)
      return NULL;

   for (i = 0; i < NRows; i++)
   {
      if ((matrix[i] = (int *)e_calloc(NCols, sizeof(int))) == NULL)
      {
         e_free_frag_i_matr(i,matrix);
         return NULL;
      }
   }

   return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_frag_f_matr: Float matrix. */


float **e_alloc_frag_f_matr(unsigned int NRows, unsigned int NCols)
{
   register unsigned int i;
   float **matrix;


   if ((matrix = (float **)e_calloc(NRows, sizeof(float *))) == NULL)
      return NULL;

   for (i = 0; i < NRows; i++)
   {
      if ((matrix[i] = (float *)e_calloc(NCols, sizeof(float))) == NULL)
      {
         e_free_frag_f_matr(i,matrix);
         return NULL;
      }
   }

   return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_frag_d_matr: Double precision matrix. */


double **e_alloc_frag_d_matr(unsigned int NRows, unsigned int NCols)
{
   register unsigned int i;
   double **matrix;


   if ((matrix = (double **)e_calloc(NRows, sizeof(double *))) == NULL)
      return NULL;

   for (i = 0; i < NRows; i++)
   {
      if ((matrix[i] = (double *)e_calloc(NCols, sizeof(double))) == NULL)
      {
         e_free_frag_d_matr(i,matrix);
         return NULL;
      }
   }

   return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_frag_c_matr: Character matrix. */


char **e_alloc_frag_c_matr(unsigned int NRows, unsigned int NCols)
{
   register unsigned int i;
   char **matrix;


   if ((matrix = (char **)e_calloc(NRows, sizeof(char *))) == NULL)
      return NULL;

   for (i = 0; i < NRows; i++)
   {
      if ((matrix[i] = (char *)e_calloc(NCols, sizeof(char))) == NULL)
      {
         e_free_frag_c_matr(i,matrix);
         return NULL;
      }
   }

   return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_frag_vp_matr: Void pointer matrix. */


void ***e_alloc_frag_vp_matr(unsigned int NRows, unsigned int NCols)
{
   register unsigned int i;
   void ***matrix;


   if ((matrix = (void ***)e_calloc(NRows, sizeof(void **))) == NULL)
      return NULL;

   for (i = 0; i < NRows; i++)
   {
      if ((matrix[i] = e_alloc_vp_vect(NCols)) == NULL)
      {
         e_free_frag_vp_matr(i,matrix);
         return NULL;
      }
   }

   return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */

/* The following functions allocate memory for an "NRows" x "NCols" matrix
   where the entire matrix resides in contiguous memory. The matrix is
   initialised to zero. The return value is a pointer to the allocated memory
   if the allocation is successful, else NULL.                                */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_cont_i_matr: "int" integer matrix. */


int **e_alloc_cont_i_matr(unsigned int NRows, unsigned int NCols)
{
   register unsigned int i;
   int **matrix;


   if ((matrix = (int **)e_calloc(NRows, sizeof(int *))) == NULL)
      return NULL;
   else if (NRows == 0)
      return matrix;

   if ((matrix[0] = (int *)e_calloc(NRows*NCols, sizeof(int))) == NULL)
   {
      e_free(matrix);
      return NULL;
   }

   for (i = 1; i < NRows; i++)
      matrix[i] = matrix[i - 1] + NCols;

   return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_cont_f_matr: Float matrix. */


float **e_alloc_cont_f_matr(unsigned int NRows, unsigned int NCols)
{
   register unsigned int i;
   float **matrix;


   if ((matrix = (float **)e_calloc(NRows, sizeof(float *))) == NULL)
      return NULL;
   else if (NRows == 0)
      return matrix;

   if ((matrix[0] = (float *)e_calloc(NRows*NCols, sizeof(float))) == NULL)
   {
      e_free(matrix);
      return NULL;
   }

   for (i = 1; i < NRows; i++)
      matrix[i] = matrix[i - 1] + NCols;

   return matrix;
} 


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_cont_d_matr: Double precision matrix. */


double **e_alloc_cont_d_matr(unsigned int NRows, unsigned int NCols)
{
   register unsigned int i;
   double **matrix;


   if ((matrix = (double **)e_calloc(NRows, sizeof(double *))) == NULL)
      return NULL;
   else if (NRows == 0)
      return matrix;

   if ((matrix[0] = (double *)e_calloc(NRows*NCols, sizeof(double))) == NULL)
   {
      e_free(matrix);
      return NULL;
   }

   for (i = 1; i < NRows; i++)
      matrix[i] = matrix[i - 1] + NCols;

   return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_cont_c_matr: Character matrix. */


char **e_alloc_cont_c_matr(unsigned int NRows, unsigned int NCols)
{
   register unsigned int i;
   char **matrix;


   if ((matrix = (char **)e_calloc(NRows, sizeof(char *))) == NULL)
      return NULL;
   else if (NRows == 0)
      return matrix;

   if ((matrix[0] = (char *)e_calloc(NRows*NCols, sizeof(char))) == NULL)
   {
      e_free(matrix);
      return NULL;
   }

   for (i = 1; i < NRows; i++)
      matrix[i] = matrix[i - 1] + NCols;

   return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_cont_vp_matr: Void pointer matrix. */


void ***e_alloc_cont_vp_matr(unsigned int NRows, unsigned int NCols)
{
   register unsigned int i;
   void ***matrix;


   if ((matrix = (void ***)e_calloc(NRows, sizeof(void **))) == NULL)
      return NULL;
   else if (NRows == 0)
      return matrix;

   if ((matrix[0] = e_alloc_vp_vect(NRows*NCols)) == NULL)
   {
      e_free(matrix);
      return NULL;
   }

   for (i = 1; i < NRows; i++)
      matrix[i] = matrix[i - 1] + NCols;

   return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */

/* The following functions allocate memory for a vector of "NDepth"
   fragmented "NRows" x "NCols" matrices. The resulting 3D matrix is
   initialised to zero. The return value is a pointer to the allocated memory
   if the allocation is successful, else NULL.                                */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_frag_i_3D_matr: 3D "int" integer matrix. */


int ***e_alloc_frag_i_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                   unsigned int NCols )
{
   register unsigned int i;
   int ***matrix3D;


   if ((matrix3D = (int ***)e_calloc(NDepth, sizeof(int **))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
   {
      if ((matrix3D[i] = e_alloc_frag_i_matr(NRows,NCols)) == NULL)
      {
         e_free_frag_i_3D_matr(i,NRows,matrix3D);
         return NULL;
      }
   }

   return matrix3D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_frag_f_3D_matr: 3D float matrix. */


float ***e_alloc_frag_f_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                     unsigned int NCols )
{
   register unsigned int i;

   float ***matrix3D;


   if ((matrix3D = (float ***)e_calloc(NDepth, sizeof(float **))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
   {
      if ((matrix3D[i] = e_alloc_frag_f_matr(NRows,NCols)) == NULL)
      {
         e_free_frag_f_3D_matr(i,NRows,matrix3D);
         return NULL;
      }
   }

   return matrix3D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_frag_d_3D_matr: 3D double precision "matrix. */


double ***e_alloc_frag_d_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                      unsigned int NCols )
{
   register unsigned int i;

   double ***matrix3D;


   if ((matrix3D = (double ***)e_calloc(NDepth, sizeof(double **))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
   {
      if ((matrix3D[i] = e_alloc_frag_d_matr(NRows,NCols)) == NULL)
      {
         e_free_frag_d_3D_matr(i,NRows,matrix3D);
         return NULL;
      }
   }

   return matrix3D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_frag_c_3D_matr: 3D character matrix. */


char ***e_alloc_frag_c_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                    unsigned int NCols )
{
   register unsigned int i;

   char ***matrix3D;


   if ((matrix3D = (char ***)e_calloc(NDepth, sizeof(char **))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
   {
      if ((matrix3D[i] = e_alloc_frag_c_matr(NRows,NCols)) == NULL)
      {
         e_free_frag_c_3D_matr(i,NRows,matrix3D);
         return NULL;
      }
   }

   return matrix3D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_frag_vp_3D_matr: 3D void pointer matrix. */


void ****e_alloc_frag_vp_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                      unsigned int NCols )
{
   register unsigned int i;

   void ****matrix3D;


   if ((matrix3D = (void ****)e_calloc(NDepth, sizeof(void ***))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
   {
      if ((matrix3D[i] = e_alloc_frag_vp_matr(NRows,NCols)) == NULL)
      {
         e_free_frag_vp_3D_matr(i,NRows,matrix3D);
         return NULL;
      }
   }

   return matrix3D;
}


/* ////////////////////////////////////////////////////////////////////////// */

/* The following functions allocate memory for an
   "NDepth" x "NRows" x "NCols" matrix where the entire matrix resides in
   contiguous memory. The matrix is initialised to zero. The return value is
   a pointer to the allocated memory if the allocation is successful, else
   NULL.                                                                      */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_cont_i_3D_matr: 3D "int" integer matrix. */


int ***e_alloc_cont_i_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                   unsigned int NCols )
{
   register unsigned int i,k;
   int ***matrix3D;
   int *vector;


   matrix3D = NULL;
   vector = NULL;

   if (   (matrix3D = (int ***)e_calloc(NDepth, sizeof(int **))) == NULL
       ||          (vector = e_alloc_i_vect(NDepth*NRows*NCols)) == NULL)
   {
      e_free(matrix3D);
      e_free(vector);
      return NULL;
   }
   else if (NDepth == 0 || NRows == 0)
   {
      e_free(vector);
      return matrix3D;
   }

   for (k = 0; k < NDepth; k++)
   {
      if ((matrix3D[k] = (int **)e_calloc(NRows, sizeof(int *))) == NULL)
      {
         for (k--; k > 0; k--)
            e_free(matrix3D[k]);

         e_free(matrix3D[0]);
         e_free(matrix3D);
         e_free(vector);
         return NULL;
      }

      for (i = 0; i < NRows; i++)
      {
         matrix3D[k][i] = vector;
         vector += NCols;
      }
   }

   return matrix3D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_cont_f_3D_matr: 3D float matrix. */


float ***e_alloc_cont_f_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                     unsigned int NCols )
{
   register unsigned int i,k;

   float ***matrix3D;
   float *vector;


   matrix3D = NULL;
   vector = NULL;

   if (   (matrix3D = (float ***)e_calloc(NDepth, sizeof(float **))) == NULL
       ||              (vector = e_alloc_f_vect(NDepth*NRows*NCols)) == NULL)
   {
      e_free(matrix3D);
      e_free(vector);
      return NULL;
   }
   else if (NDepth == 0 || NRows == 0)
   {
      e_free(vector);
      return matrix3D;
   }

   for (k = 0; k < NDepth; k++)
   {
      if ((matrix3D[k] = (float **)e_calloc(NRows, sizeof(float *))) == NULL)
      {
         for (k--; k > 0; k--)
            e_free(matrix3D[k]);

         e_free(matrix3D[0]);
         e_free(matrix3D);
         e_free(vector);
         return NULL;
      }

      for (i = 0; i < NRows; i++)
      {
         matrix3D[k][i] = vector;
         vector += NCols;
      }
   }

   return matrix3D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_cont_d_3D_matr: 3D double precision matrix. */


double ***e_alloc_cont_d_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                      unsigned int NCols )
{
   register unsigned int i,k;

   double ***matrix3D;
   double *vector;


   matrix3D = NULL;
   vector = NULL;

   if (   (matrix3D = (double ***)e_calloc(NDepth, sizeof(double **))) == NULL
       ||                (vector = e_alloc_d_vect(NDepth*NRows*NCols)) == NULL)
   {
      e_free(matrix3D);
      e_free(vector);
      return NULL;
   }
   else if (NDepth == 0 || NRows == 0)
   {
      e_free(vector);
      return matrix3D;
   }

   for (k = 0; k < NDepth; k++)
   {
      if ((matrix3D[k] = (double **)e_calloc(NRows, sizeof(double *))) == NULL)
      {
         for (k--; k > 0; k--)
            e_free(matrix3D[k]);

         e_free(matrix3D[0]);
         e_free(matrix3D);
         e_free(vector);
         return NULL;
      }

      for (i = 0; i < NRows; i++)
      {
         matrix3D[k][i] = vector;
         vector += NCols;
      }
   }

   return matrix3D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_cont_c_3D_matr: 3D character matrix. */


char ***e_alloc_cont_c_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                    unsigned int NCols )
{
   register unsigned int i,k;

   char ***matrix3D;
   char *vector;


   matrix3D = NULL;
   vector = NULL;

   if (   (matrix3D = (char ***)e_calloc(NDepth, sizeof(char **))) == NULL
       ||            (vector = e_alloc_c_vect(NDepth*NRows*NCols)) == NULL)
   {
      e_free(matrix3D);
      e_free(vector);
      return NULL;
   }
   else if (NDepth == 0 || NRows == 0)
   {
      e_free(vector);
      return matrix3D;
   }

   for (k = 0; k < NDepth; k++)
   {
      if ((matrix3D[k] = (char **)e_calloc(NRows, sizeof(char *))) == NULL)
      {
         for (k--; k > 0; k--)
            e_free(matrix3D[k]);

         e_free(matrix3D[0]);
         e_free(matrix3D);
         e_free(vector);
         return NULL;
      }

      for (i = 0; i < NRows; i++)
      {
         matrix3D[k][i] = vector;
         vector += NCols;
      }
   }

   return matrix3D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_cont_vp_3D_matr: 3D void pointer matrix. */


void ****e_alloc_cont_vp_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                      unsigned int NCols )
{
   register unsigned int i,k;

   void ****matrix3D;
   void **vector;


   matrix3D = NULL;
   vector = NULL;

   if (   (matrix3D = (void ****)e_calloc(NDepth, sizeof(void ***))) == NULL
       ||             (vector = e_alloc_vp_vect(NDepth*NRows*NCols)) == NULL)
   {
      e_free(matrix3D);
      e_free(vector);
      return NULL;
   }
   else if (NDepth == 0 || NRows == 0)
   {
      e_free(vector);
      return matrix3D;
   }

   for (k = 0; k < NDepth; k++)
   {
      if ((matrix3D[k] = (void ***)e_calloc(NRows, sizeof(void **))) == NULL)
      {
         for (k--; k > 0; k--)
            e_free(matrix3D[k]);

         e_free(matrix3D[0]);
         e_free(matrix3D);
         e_free(vector);
         return NULL;
      }

      for (i = 0; i < NRows; i++)
      {
         matrix3D[k][i] = vector;
         vector += NCols;
      }
   }

   return matrix3D;
}


/* ////////////////////////////////////////////////////////////////////////// */

/* The following functions allocate memory for a vector of "NDepth"
   fragmented "NWidth" x "NRows" x "NCols" 3D matrices. The resulting 4D
   matrix is initialised to zero. The return value is a pointer to the
   allocated memory if the allocation is successful, else NULL.               */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_frag_i_4D_matr: 4D "int" integer matrix. */


int ****e_alloc_frag_i_4D_matr(unsigned int NDepth, unsigned int NWidth,
                               unsigned int NRows,  unsigned int NCols  )
{
   register unsigned int i;
   int ****matrix4D;


   if ((matrix4D = (int ****)e_calloc(NDepth, sizeof(int ***))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
   {
      if ((matrix4D[i] = e_alloc_frag_i_3D_matr(NWidth,NRows,NCols)) == NULL)
      {
         e_free_frag_i_4D_matr(i,NWidth,NRows,matrix4D);
         return NULL;
      }
   }

   return matrix4D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_frag_f_4D_matr: 4D float matrix. */


float ****e_alloc_frag_f_4D_matr(unsigned int NDepth, unsigned int NWidth,
                                 unsigned int NRows,  unsigned int NCols  )
{
   register unsigned int i;

   float ****matrix4D;


   if ((matrix4D = (float ****)e_calloc(NDepth, sizeof(float ***))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
   {
      if ((matrix4D[i] = e_alloc_frag_f_3D_matr(NWidth,NRows,NCols)) == NULL)
      {
         e_free_frag_f_4D_matr(i,NWidth,NRows,matrix4D);
         return NULL;
      }
   }

   return matrix4D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_frag_d_4D_matr: 4D double precision matrix. */


double ****e_alloc_frag_d_4D_matr(unsigned int NDepth, unsigned int NWidth,
                                  unsigned int NRows,  unsigned int NCols  )
{
   register unsigned int i;

   double ****matrix4D;


   if ((matrix4D = (double ****)e_calloc(NDepth, sizeof(double ***))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
   {
      if ((matrix4D[i] = e_alloc_frag_d_3D_matr(NWidth,NRows,NCols)) == NULL)
      {
         e_free_frag_d_4D_matr(i,NWidth,NRows,matrix4D);
         return NULL;
      }
   }

   return matrix4D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_frag_c_4D_matr: 4D character matrix. */


char ****e_alloc_frag_c_4D_matr(unsigned int NDepth, unsigned int NWidth,
                                unsigned int NRows,  unsigned int NCols  )
{
   register unsigned int i;

   char ****matrix4D;


   if ((matrix4D = (char ****)e_calloc(NDepth, sizeof(char ***))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
   {
      if ((matrix4D[i] = e_alloc_frag_c_3D_matr(NWidth,NRows,NCols)) == NULL)
      {
         e_free_frag_c_4D_matr(i,NWidth,NRows,matrix4D);
         return NULL;
      }
   }

   return matrix4D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_frag_vp_4D_matr: 4D void pointer matrix. */


void *****e_alloc_frag_vp_4D_matr(unsigned int NDepth, unsigned int NWidth,
                                  unsigned int NRows,  unsigned int NCols  )
{
   register unsigned int i;

   void *****matrix4D;


   if ((matrix4D = (void *****)e_calloc(NDepth, sizeof(void ****))) == NULL)
      return NULL;

   for (i = 0; i < NDepth; i++)
   {
      if ((matrix4D[i] = e_alloc_frag_vp_3D_matr(NWidth,NRows,NCols)) == NULL)
      {
         e_free_frag_vp_4D_matr(i,NWidth,NRows,matrix4D);
         return NULL;
      }
   }

   return matrix4D;
}


/* ////////////////////////////////////////////////////////////////////////// */

/* The following functions allocate memory for an
   "NDepth" x "NWidth" x "NRows" x "NCols" matrix where the entire matrix
   resides in contiguous memory. The matrix is initialised to zero. The
   return value is a pointer to the allocated memory if the allocation is
   successful, else NULL.                                                     */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_cont_i_4D_matr: 4D "int" integer matrix. */


int ****e_alloc_cont_i_4D_matr(unsigned int NDepth, unsigned int NWidth,
                               unsigned int NRows,  unsigned int NCols  )
{
   register unsigned int i,k,l,r;
   int ****matrix4D;
   int *vector;


   matrix4D = NULL;
   vector = NULL;

   if (   (matrix4D = (int ****)e_calloc(NDepth, sizeof(int ***))) == NULL
       ||     (vector = e_alloc_i_vect(NDepth*NWidth*NRows*NCols)) == NULL)
   {
      e_free(matrix4D);
      e_free(vector);
      return NULL;
   }
   else if (NDepth == 0 || NWidth == 0 || NRows == 0)
   {
      e_free(vector);
      return matrix4D;
   }

   for (k = 0; k < NDepth; k++)
   {
      if ((matrix4D[k] = (int ***)e_calloc(NWidth, sizeof(int **))) == NULL)
      {
         r = k;

         for (k = 0; k < r; k++)
         {
            for (l = 0; l < NWidth; l++)
               e_free(matrix4D[k][l]);

            e_free(matrix4D[k]);
         }

         e_free(matrix4D);
         e_free(vector);
         return NULL;
      }

      for (l = 0; l < NWidth; l++)
      {
         if ((matrix4D[k][l] = (int **)e_calloc(NRows, sizeof(int *))) == NULL)
         {
            r = l;

            for (l = 0; l < r; l++)
               e_free(matrix4D[k][l]);

            e_free(matrix4D[k]);

            r = k;

            for (k = 0; k < r; k++)
            {
               for (l = 0; l < NWidth; l++)
                  e_free(matrix4D[k][l]);

               e_free(matrix4D[k]);
            }

            e_free(matrix4D);
            e_free(vector);
            return NULL;
         }

         for (i = 0; i < NRows; i++)
         {
            matrix4D[k][l][i] = vector;
            vector += NCols;
         }
      }
   }

   return matrix4D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_cont_f_4D_matr: 4D float matrix. */


float ****e_alloc_cont_f_4D_matr(unsigned int NDepth, unsigned int NWidth,
                                 unsigned int NRows,  unsigned int NCols  )
{
   register unsigned int i,k,l,r;

   float ****matrix4D;
   float *vector;


   matrix4D = NULL;
   vector = NULL;

   if (   (matrix4D = (float ****)e_calloc(NDepth, sizeof(float ***))) == NULL
       ||         (vector = e_alloc_f_vect(NDepth*NWidth*NRows*NCols)) == NULL)
   {
      e_free(matrix4D);
      e_free(vector);
      return NULL;
   }
   else if (NDepth == 0 || NWidth == 0 || NRows == 0)
   {
      e_free(vector);
      return matrix4D;
   }

   for (k = 0; k < NDepth; k++)
   {
      if ((matrix4D[k]
              = (float ***)e_calloc(NWidth, sizeof(float **))) == NULL)
      {
         r = k;

         for (k = 0; k < r; k++)
         {
            for (l = 0; l < NWidth; l++)
               e_free(matrix4D[k][l]);

            e_free(matrix4D[k]);
         }

         e_free(matrix4D);
         e_free(vector);
         return NULL;
      }

      for (l = 0; l < NWidth; l++)
      {
         if ((matrix4D[k][l]
                 = (float **)e_calloc(NRows, sizeof(float *))) == NULL)
         {
            r = l;

            for (l = 0; l < r; l++)
               e_free(matrix4D[k][l]);

            e_free(matrix4D[k]);

            r = k;

            for (k = 0; k < r; k++)
            {
               for (l = 0; l < NWidth; l++)
                  e_free(matrix4D[k][l]);

               e_free(matrix4D[k]);
            }

            e_free(matrix4D);
            e_free(vector);
            return NULL;
         }

         for (i = 0; i < NRows; i++)
         {
            matrix4D[k][l][i] = vector;
            vector += NCols;
         }
      }
   }

   return matrix4D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_cont_d_4D_matr: 4D double precision matrix. */


double ****e_alloc_cont_d_4D_matr(unsigned int NDepth, unsigned int NWidth,
                                  unsigned int NRows,  unsigned int NCols  )
{
   register unsigned int i,k,l,r;

   double ****matrix4D;
   double *vector;


   matrix4D = NULL;
   vector = NULL;

   if (   (matrix4D
              = (double ****)e_calloc(NDepth, sizeof(double ***))) == NULL
       ||     (vector = e_alloc_d_vect(NDepth*NWidth*NRows*NCols)) == NULL)
   {
      e_free(matrix4D);
      e_free(vector);
      return NULL;
   }
   else if (NDepth == 0 || NWidth == 0 || NRows == 0)
   {
      e_free(vector);
      return matrix4D;
   }

   for (k = 0; k < NDepth; k++)
   {
      if ((matrix4D[k]
              = (double ***)e_calloc(NWidth, sizeof(double **))) == NULL)
      {
         r = k;

         for (k = 0; k < r; k++)
         {
            for (l = 0; l < NWidth; l++)
               e_free(matrix4D[k][l]);

            e_free(matrix4D[k]);
         }

         e_free(matrix4D);
         e_free(vector);
         return NULL;
      }

      for (l = 0; l < NWidth; l++)
      {
         if ((matrix4D[k][l]
                 = (double **)e_calloc(NRows, sizeof(double *))) == NULL)
         {
            r = l;

            for (l = 0; l < r; l++)
               e_free(matrix4D[k][l]);

            e_free(matrix4D[k]);

            r = k;

            for (k = 0; k < r; k++)
            {
               for (l = 0; l < NWidth; l++)
                  e_free(matrix4D[k][l]);

               e_free(matrix4D[k]);
            }

            e_free(matrix4D);
            e_free(vector);
            return NULL;
         }

         for (i = 0; i < NRows; i++)
         {
            matrix4D[k][l][i] = vector;
            vector += NCols;
         }
      }
   }

   return matrix4D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_cont_c_4D_matr: 4D character matrix. */


char ****e_alloc_cont_c_4D_matr(unsigned int NDepth, unsigned int NWidth,
                                unsigned int NRows,  unsigned int NCols  )
{
   register unsigned int i,k,l,r;

   char ****matrix4D;
   char *vector;


   matrix4D = NULL;
   vector = NULL;

   if (   (matrix4D = (char ****)e_calloc(NDepth, sizeof(char ***))) == NULL
       ||       (vector = e_alloc_c_vect(NDepth*NWidth*NRows*NCols)) == NULL)
   {
      e_free(matrix4D);
      e_free(vector);
      return NULL;
   }
   else if (NDepth == 0 || NWidth == 0 || NRows == 0)
   {
      e_free(vector);
      return matrix4D;
   }

   for (k = 0; k < NDepth; k++)
   {
      if ((matrix4D[k] = (char ***)e_calloc(NWidth, sizeof(char **))) == NULL)
      {
         r = k;

         for (k = 0; k < r; k++)
         {
            for (l = 0; l < NWidth; l++)
               e_free(matrix4D[k][l]);

            e_free(matrix4D[k]);
         }

         e_free(matrix4D);
         e_free(vector);
         return NULL;
      }

      for (l = 0; l < NWidth; l++)
      {
         if ((matrix4D[k][l]
                 = (char **)e_calloc(NRows, sizeof(char *))) == NULL)
         {
            r = l;

            for (l = 0; l < r; l++)
               e_free(matrix4D[k][l]);

            e_free(matrix4D[k]);

            r = k;

            for (k = 0; k < r; k++)
            {
               for (l = 0; l < NWidth; l++)
                  e_free(matrix4D[k][l]);

               e_free(matrix4D[k]);
            }

            e_free(matrix4D);
            e_free(vector);
            return NULL;
         }

         for (i = 0; i < NRows; i++)
         {
            matrix4D[k][l][i] = vector;
            vector += NCols;
         }
      }
   }

   return matrix4D;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_alloc_cont_vp_4D_matr: 4D void pointer matrix. */


void *****e_alloc_cont_vp_4D_matr(unsigned int NDepth, unsigned int NWidth,
                                  unsigned int NRows,  unsigned int NCols  )
{
   register unsigned int i,k,l,r;

   void *****matrix4D;
   void **vector;


   matrix4D = NULL;
   vector = NULL;

   if (   (matrix4D = (void *****)e_calloc(NDepth, sizeof(void ****))) == NULL
       ||        (vector = e_alloc_vp_vect(NDepth*NWidth*NRows*NCols)) == NULL)
   {
      e_free(matrix4D);
      e_free(vector);
      return NULL;
   }
   else if (NDepth == 0 || NWidth == 0 || NRows == 0)
   {
      e_free(vector);
      return matrix4D;
   }

   for (k = 0; k < NDepth; k++)
   {
      if ((matrix4D[k] = (void ****)e_calloc(NWidth, sizeof(void ***))) == NULL)
      {
         r = k;

         for (k = 0; k < r; k++)
         {
            for (l = 0; l < NWidth; l++)
               e_free(matrix4D[k][l]);

            e_free(matrix4D[k]);
         }

         e_free(matrix4D);
         e_free(vector);
         return NULL;
      }

      for (l = 0; l < NWidth; l++)
      {
         if ((matrix4D[k][l]
                 = (void ***)e_calloc(NRows, sizeof(void **))) == NULL)
         {
            r = l;

            for (l = 0; l < r; l++)
               e_free(matrix4D[k][l]);

            e_free(matrix4D[k]);

            r = k;

            for (k = 0; k < r; k++)
            {
               for (l = 0; l < NWidth; l++)
                  e_free(matrix4D[k][l]);

               e_free(matrix4D[k]);
            }

            e_free(matrix4D);
            e_free(vector);
            return NULL;
         }

         for (i = 0; i < NRows; i++)
         {
            matrix4D[k][l][i] = vector;
            vector += NCols;
         }
      }
   }

   return matrix4D;
}


/* ////////////////////////////////////////////////////////////////////////// */

/* The following functions increase the length of a vector of length "length"
   by "VectMax0" if "length" is a multiple of "VectMax0". If "length" equals
   zero then "vector" should be set to NULL. If the vector is not incremented
   then the return value is a pointer to the unchanged "vector". If additional
   memory is allocated successfully, the return value is a pointer to the
   incremented vector, otherwise the return value is NULL. The additional
   components are set to zero.                                                */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_incr_i_vect: "int" integer vector. */


int *e_incr_i_vect(unsigned int length, int *vector, unsigned int VectMax0)
{
   int *TempVector;


   if (length % VectMax0 == 0)
   {
      if ((TempVector = e_alloc_i_vect(length + VectMax0)) == NULL)
      {
         e_free(vector);
         return NULL;
      }

      memcpy(TempVector,vector,length*sizeof(int));
      e_free(vector);
      return TempVector;
   }
   else
      return vector;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_incr_f_vect: Float vector. */


float *e_incr_f_vect(unsigned int length, float *vector, unsigned int VectMax0)
{
   float *TempVector;


   if (length % VectMax0 == 0)
   {
      if ((TempVector = e_alloc_f_vect(length + VectMax0)) == NULL)
      {
         e_free(vector);
         return NULL;
      }
   
      memcpy(TempVector,vector,length*sizeof(float));
      e_free(vector);
      return TempVector;
   }
   else
      return vector;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_incr_d_vect: Double precision vector. */


double *e_incr_d_vect(unsigned int length, double *vector,
                                     unsigned int VectMax0)
{
   double *TempVector;


   if (length % VectMax0 == 0)
   {
      if ((TempVector = e_alloc_d_vect(length + VectMax0)) == NULL)
      {
         e_free(vector);
         return NULL;
      }
   
      memcpy(TempVector,vector,length*sizeof(double));
      e_free(vector);
      return TempVector;
   }
   else
      return vector;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_incr_c_vect: Character vector. */


char *e_incr_c_vect(unsigned int length, char *vector, unsigned int VectMax0)
{
   char *TempVector;


   if (length % VectMax0 == 0)
   {
      if ((TempVector = e_alloc_c_vect(length + VectMax0)) == NULL)
      {
         e_free(vector);
         return NULL;
      }
   
      memcpy(TempVector,vector,length*sizeof(char));
      e_free(vector);
      return TempVector;
   }
   else
      return vector;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_incr_vp_vect: Void pointer vector. */


void **e_incr_vp_vect(unsigned int length, void **vector,
                                    unsigned int VectMax0)
{
   void **TempVector;


   if (length % VectMax0 == 0)
   {
      if ((TempVector = e_alloc_vp_vect(length + VectMax0)) == NULL)
      {
         e_free(vector);
         return NULL;
      }
   
      memcpy(TempVector,vector,length*sizeof(void *));
      e_free(vector);
      return TempVector;
   }
   else
      return vector;
}


/* ////////////////////////////////////////////////////////////////////////// */

/* The following functions increase the number of rows "NRows" of the
   fragmented matrix "matrix" by "NRowsMax0" if "NRows" is a multiple of
   "NRowsMax0". If "NRows" equals zero then "matrix" should be set to NULL. If
   the matrix is not incremented then the return value is a pointer to the
   unchanged "matrix". If additional memory is allocated successfully, the
   return value is a pointer to the incremented matrix, otherwise the return
   value is NULL. The additional row pointers are set to NULL.                */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_incr_frag_i_matr: "int" integer matrix. */


int **e_incr_frag_i_matr(unsigned int NRows, int **matrix,
                                    unsigned int NRowsMax0)
{
   int **TempMatrix;


   if (NRows % NRowsMax0 == 0)
   {
      if ((TempMatrix = e_init_frag_i_matr(NRows + NRowsMax0)) == NULL)
      {
         e_free_frag_i_matr(NRows,matrix);
         return NULL;
      }

      memcpy(TempMatrix,matrix,NRows*sizeof(int *));
      e_free(matrix);
      return TempMatrix;
   }
   else
      return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_incr_frag_f_matr: Float matrix. */


float **e_incr_frag_f_matr(unsigned int NRows, float **matrix,
                                        unsigned int NRowsMax0)
{
   float **TempMatrix;


   if (NRows % NRowsMax0 == 0)
   {
      if ((TempMatrix = e_init_frag_f_matr(NRows + NRowsMax0)) == NULL)
      {
         e_free_frag_f_matr(NRows,matrix);
         return NULL;
      }

      memcpy(TempMatrix,matrix,NRows*sizeof(float *));
      e_free(matrix);
      return TempMatrix;
   }
   else
      return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_incr_frag_d_matr: Double precision matrix. */


double **e_incr_frag_d_matr(unsigned int NRows, double **matrix,
                                          unsigned int NRowsMax0)
{
   double **TempMatrix;


   if (NRows % NRowsMax0 == 0)
   {
      if ((TempMatrix = e_init_frag_d_matr(NRows + NRowsMax0)) == NULL)
      {
         e_free_frag_d_matr(NRows,matrix);
         return NULL;
      }

      memcpy(TempMatrix,matrix,NRows*sizeof(double *));
      e_free(matrix);
      return TempMatrix;
   }
   else
      return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_incr_frag_c_matr: Character matrix. */


char **e_incr_frag_c_matr(unsigned int NRows, char **matrix,
                                      unsigned int NRowsMax0)
{
   char **TempMatrix;


   if (NRows % NRowsMax0 == 0)
   {
      if ((TempMatrix = e_init_frag_c_matr(NRows + NRowsMax0)) == NULL)
      {
         e_free_frag_c_matr(NRows,matrix);
         return NULL;
      }

      memcpy(TempMatrix,matrix,NRows*sizeof(char *));
      e_free(matrix);
      return TempMatrix;
   }
   else
      return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_incr_frag_vp_matr: Void pointer matrix. */


void ***e_incr_frag_vp_matr(unsigned int NRows, void ***matrix,
                                         unsigned int NRowsMax0)
{
   void ***TempMatrix;


   if (NRows % NRowsMax0 == 0)
   {
      if ((TempMatrix = e_init_frag_vp_matr(NRows + NRowsMax0)) == NULL)
      {
         e_free_frag_vp_matr(NRows,matrix);
         return NULL;
      }

      memcpy(TempMatrix,matrix,NRows*sizeof(void **));
      e_free(matrix);
      return TempMatrix;
   }
   else
      return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */

/* The following functions increase the number of rows "NRows" of the
   contiguous matrix "matrix" by "NRowsMax0" if "NRows" is a multiple of
   "NRowsMax0". If "NRows" equals zero then "matrix" should be set to NULL. If
   the matrix is not incremented then the return value is a pointer to the
   unchanged "matrix". If additional memory is allocated successfully, the
   return value is a pointer to the incremented matrix, otherwise the return
   value is NULL. The additional rows are initialised to zero.                */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_incr_cont_i_matr: "int" integer matrix. */


int **e_incr_cont_i_matr(unsigned int NRows, unsigned int NCols,
                            int **matrix, unsigned int NRowsMax0)
{
   int **TempMatrix;


   if (NRows % NRowsMax0 == 0)
   {
      if ((TempMatrix = e_alloc_cont_i_matr(NRows + NRowsMax0,NCols)) == NULL)
      {
         e_free_cont_i_matr(NRows,matrix);
         return NULL;
      }

      if (matrix != NULL)
         memcpy(TempMatrix[0],matrix[0],NRows*NCols*sizeof(int));
      e_free_cont_i_matr(NRows,matrix);
      return TempMatrix;
   }
   else
      return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_incr_cont_f_matr: Float matrix. */


float **e_incr_cont_f_matr(unsigned int NRows, unsigned int NCols,
                            float **matrix, unsigned int NRowsMax0)
{
   float **TempMatrix;


   if (NRows % NRowsMax0 == 0)
   {
      if ((TempMatrix = e_alloc_cont_f_matr(NRows + NRowsMax0,NCols)) == NULL)
      {
         e_free_cont_f_matr(NRows,matrix);
         return NULL;
      }

      if (matrix != NULL)
         memcpy(TempMatrix[0],matrix[0],NRows*NCols*sizeof(float));
      e_free_cont_f_matr(NRows,matrix);
      return TempMatrix;
   }
   else
      return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_incr_cont_d_matr: Double precision matrix. */


double **e_incr_cont_d_matr(unsigned int NRows, unsigned int NCols,
                            double **matrix, unsigned int NRowsMax0)
{
   double **TempMatrix;


   if (NRows % NRowsMax0 == 0)
   {
      if ((TempMatrix = e_alloc_cont_d_matr(NRows + NRowsMax0,NCols)) == NULL)
      {
         e_free_cont_d_matr(NRows,matrix);
         return NULL;
      }

      if (matrix != NULL)
         memcpy(TempMatrix[0],matrix[0],NRows*NCols*sizeof(double));
      e_free_cont_d_matr(NRows,matrix);
      return TempMatrix;
   }
   else
      return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_incr_cont_c_matr: Character matrix. */


char **e_incr_cont_c_matr(unsigned int NRows, unsigned int NCols,
                            char **matrix, unsigned int NRowsMax0)
{
   char **TempMatrix;


   if (NRows % NRowsMax0 == 0)
   {
      if ((TempMatrix = e_alloc_cont_c_matr(NRows + NRowsMax0,NCols)) == NULL)
      {
         e_free_cont_c_matr(NRows,matrix);
         return NULL;
      }

      if (matrix != NULL)
         memcpy(TempMatrix[0],matrix[0],NRows*NCols*sizeof(char));
      e_free_cont_c_matr(NRows,matrix);
      return TempMatrix;
   }
   else
      return matrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_incr_cont_vp_matr: Void pointer matrix. */


void ***e_incr_cont_vp_matr(unsigned int NRows, unsigned int NCols,
                             void ***matrix, unsigned int NRowsMax0)
{
   void ***TempMatrix;


   if (NRows % NRowsMax0 == 0)
   {
      if ((TempMatrix = e_alloc_cont_vp_matr(NRows + NRowsMax0,NCols)) == NULL)
      {
         e_free_cont_vp_matr(NRows,matrix);
         return NULL;
      }

      if (matrix != NULL)
         memcpy(TempMatrix[0],matrix[0],NRows*NCols*sizeof(void *));
      e_free_cont_vp_matr(NRows,matrix);
      return TempMatrix;
   }
   else
      return matrix;
}

/* ////////////////////////////////////////////////////////////////////////// */

/* The following functions increase the length of a vector of length "length"
   by "VectMax0". If additional memory is allocated successfully, the return
   value is a pointer to the incremented vector, otherwise the return value 
   is NULL. The additional components are set to zero.                        */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_ext_i_vect: "int" integer vector. */


int *e_ext_i_vect(unsigned int length, int *vector, unsigned int VectMax0)
{
   int *TempVector;

    if ((TempVector = e_alloc_i_vect(length + VectMax0)) == NULL)
    {
       e_free(vector);
       return NULL;
    }

    memcpy(TempVector,vector,length*sizeof(int));
    e_free(vector);
    return TempVector;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_ext_f_vect: Float vector. */


float *e_ext_f_vect(unsigned int length, float *vector, unsigned int VectMax0)
{
   float *TempVector;

   if ((TempVector = e_alloc_f_vect(length + VectMax0)) == NULL)
   {
     e_free(vector);
     return NULL;
   }
   
   memcpy(TempVector,vector,length*sizeof(float));
   e_free(vector);
   return TempVector;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_ext_d_vect: Double precision vector. */


double *e_ext_d_vect(unsigned int length, double *vector,
                                     unsigned int VectMax0)
{
   double *TempVector;

   if ((TempVector = e_alloc_d_vect(length + VectMax0)) == NULL)
   {
     e_free(vector);
     return NULL;
   }
   
   memcpy(TempVector,vector,length*sizeof(double));
   e_free(vector);
   return TempVector;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_ext_c_vect: Character vector. */


char *e_ext_c_vect(unsigned int length, char *vector, unsigned int VectMax0)
{
   char *TempVector;

   if ((TempVector = e_alloc_c_vect(length + VectMax0)) == NULL)
   {
     e_free(vector);
     return NULL;
   }
   
   memcpy(TempVector,vector,length*sizeof(char));
   e_free(vector);
   return TempVector;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_ext_vp_vect: Void pointer vector. */


void **e_ext_vp_vect(unsigned int length, void **vector,
                                    unsigned int VectMax0)
{
   void **TempVector;

   if ((TempVector = e_alloc_vp_vect(length + VectMax0)) == NULL)
   {
      e_free(vector);
      return NULL;
   }
   
   memcpy(TempVector,vector,length*sizeof(void *));
   e_free(vector);
   return TempVector;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_ext_frag_i_matr: "int" integer matrix. */


int **e_ext_frag_i_matr(unsigned int NRows, int **matrix,
                                    unsigned int NRowsMax0)
{
   int **TempMatrix;

   if ((TempMatrix = e_init_frag_i_matr(NRows + NRowsMax0)) == NULL)
   {
      e_free_frag_i_matr(NRows,matrix);
      return NULL;
   }

   memcpy(TempMatrix,matrix,NRows*sizeof(int *));
   e_free(matrix);
   return TempMatrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_ext_frag_f_matr: Float matrix. */


float **e_ext_frag_f_matr(unsigned int NRows, float **matrix,
                                        unsigned int NRowsMax0)
{
   float **TempMatrix;


   if ((TempMatrix = e_init_frag_f_matr(NRows + NRowsMax0)) == NULL)
   {
      e_free_frag_f_matr(NRows,matrix);
      return NULL;
   }

   memcpy(TempMatrix,matrix,NRows*sizeof(float *));
   e_free(matrix);
   return TempMatrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_ext_frag_d_matr: Double precision matrix. */


double **e_ext_frag_d_matr(unsigned int NRows, double **matrix,
                                          unsigned int NRowsMax0)
{
   double **TempMatrix;


   if ((TempMatrix = e_init_frag_d_matr(NRows + NRowsMax0)) == NULL)
   {
      e_free_frag_d_matr(NRows,matrix);
      return NULL;
   }

   memcpy(TempMatrix,matrix,NRows*sizeof(double *));
   e_free(matrix);
   return TempMatrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_ext_frag_c_matr: Character matrix. */


char **e_ext_frag_c_matr(unsigned int NRows, char **matrix,
                                      unsigned int NRowsMax0)
{
   char **TempMatrix;


   if ((TempMatrix = e_init_frag_c_matr(NRows + NRowsMax0)) == NULL)
   {
      e_free_frag_c_matr(NRows,matrix);
      return NULL;
   }

   memcpy(TempMatrix,matrix,NRows*sizeof(char *));
   e_free(matrix);
   return TempMatrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_ext_frag_vp_matr: Void pointer matrix. */


void ***e_ext_frag_vp_matr(unsigned int NRows, void ***matrix,
                                         unsigned int NRowsMax0)
{
   void ***TempMatrix;


   if ((TempMatrix = e_init_frag_vp_matr(NRows + NRowsMax0)) == NULL)
   {
      e_free_frag_vp_matr(NRows,matrix);
      return NULL;
   }

   memcpy(TempMatrix,matrix,NRows*sizeof(void **));
   e_free(matrix);
   return TempMatrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_ext_cont_i_matr: "int" integer matrix. */


int **e_ext_cont_i_matr(unsigned int NRows, unsigned int NCols,
                            int **matrix, unsigned int NRowsMax0)
{
   int **TempMatrix;


   if ((TempMatrix = e_alloc_cont_i_matr(NRows + NRowsMax0,NCols)) == NULL)
   {
      e_free_cont_i_matr(NRows,matrix);
      return NULL;
   }

   if (matrix != NULL)
      memcpy(TempMatrix[0],matrix[0],NRows*NCols*sizeof(int));
   e_free_cont_i_matr(NRows,matrix);
   return TempMatrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_ext_cont_f_matr: Float matrix. */


float **e_ext_cont_f_matr(unsigned int NRows, unsigned int NCols,
                            float **matrix, unsigned int NRowsMax0)
{
   float **TempMatrix;


   if ((TempMatrix = e_alloc_cont_f_matr(NRows + NRowsMax0,NCols)) == NULL)
   {
      e_free_cont_f_matr(NRows,matrix);
      return NULL;
   }

   if (matrix != NULL)
      memcpy(TempMatrix[0],matrix[0],NRows*NCols*sizeof(float));
   e_free_cont_f_matr(NRows,matrix);
   return TempMatrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_ext_cont_d_matr: Double precision matrix. */


double **e_ext_cont_d_matr(unsigned int NRows, unsigned int NCols,
                            double **matrix, unsigned int NRowsMax0)
{
   double **TempMatrix;


   if ((TempMatrix = e_alloc_cont_d_matr(NRows + NRowsMax0,NCols)) == NULL)
   {
      e_free_cont_d_matr(NRows,matrix);
      return NULL;
   }

   if (matrix != NULL)
      memcpy(TempMatrix[0],matrix[0],NRows*NCols*sizeof(double));
   e_free_cont_d_matr(NRows,matrix);
   return TempMatrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_ext_cont_c_matr: Character matrix. */


char **e_ext_cont_c_matr(unsigned int NRows, unsigned int NCols,
                            char **matrix, unsigned int NRowsMax0)
{
   char **TempMatrix;


   if ((TempMatrix = e_alloc_cont_c_matr(NRows + NRowsMax0,NCols)) == NULL)
   {
      e_free_cont_c_matr(NRows,matrix);
      return NULL;
   }

   if (matrix != NULL)
      memcpy(TempMatrix[0],matrix[0],NRows*NCols*sizeof(char));
   e_free_cont_c_matr(NRows,matrix);
   return TempMatrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_ext_cont_vp_matr: Void pointer matrix. */


void ***e_ext_cont_vp_matr(unsigned int NRows, unsigned int NCols,
                             void ***matrix, unsigned int NRowsMax0)
{
   void ***TempMatrix;

   if ((TempMatrix = e_alloc_cont_vp_matr(NRows + NRowsMax0,NCols)) == NULL)
   {
      e_free_cont_vp_matr(NRows,matrix);
      return NULL;
   }

   if (matrix != NULL)
      memcpy(TempMatrix[0],matrix[0],NRows*NCols*sizeof(void *));
   e_free_cont_vp_matr(NRows,matrix);
   return TempMatrix;
}


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                      Memory deallocation functions                         */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free: Generalised "e_free" function. */



void e_free(void *vector)
{
   if (vector == NULL)
      return;

#ifndef __E_T800__
   free(vector);
#endif

#ifdef __E_T800__
   par_free(vector);
#endif

}


/* ////////////////////////////////////////////////////////////////////////// */

/* The following deallocate the memory of an "NRows" matrix where each row
   has been allocated separately.                                             */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_frag_i_matr: "int" integer matrix. */


void e_free_frag_i_matr(unsigned int NRows, int **matrix)
{
   register unsigned int i;


   if (matrix == NULL) return;

   for (i = 0; i < NRows; i++)
      e_free(matrix[i]);

   e_free(matrix);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_frag_f_matr: Float matrix. */


void e_free_frag_f_matr(unsigned int NRows, float **matrix)
{
   register unsigned int i;


   if (matrix == NULL) return;

   for (i = 0; i < NRows; i++)
      e_free(matrix[i]);

   e_free(matrix);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_frag_d_matr: Double precision matrix. */


void e_free_frag_d_matr(unsigned int NRows, double **matrix)
{
   register unsigned int i;


   if (matrix == NULL) return;

   for (i = 0; i < NRows; i++)
      e_free(matrix[i]);

   e_free(matrix);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_frag_c_matr: Character matrix. */


void e_free_frag_c_matr(unsigned int NRows, char **matrix)
{
   register unsigned int i;


   if (matrix == NULL) return;

   for (i = 0; i < NRows; i++)
      e_free(matrix[i]);

   e_free(matrix);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_frag_vp_matr: Void pointer matrix. */


void e_free_frag_vp_matr(unsigned int NRows, void ***matrix)
{
   register unsigned int i;


   if (matrix == NULL) return;

   for (i = 0; i < NRows; i++)
      e_free(matrix[i]);

   e_free(matrix);
}


/* ////////////////////////////////////////////////////////////////////////// */

/* The following deallocate the memory of an "NRows" matrix where the matrix
   resides in contiguous memory.                                              */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_cont_i_matr: "int" integer matrix. */


void e_free_cont_i_matr(unsigned int NRows, int **matrix)
{
   if (matrix == NULL)
      return;
   else if (NRows > 0)
      e_free(matrix[0]);

   e_free(matrix);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_cont_f_matr: Float matrix. */


void e_free_cont_f_matr(unsigned int NRows, float **matrix)
{
   if (matrix == NULL)
      return;
   else if (NRows > 0)
      e_free(matrix[0]);

   e_free(matrix);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_cont_d_matr: Double precision matrix. */


void e_free_cont_d_matr(unsigned int NRows, double **matrix)
{
   if (matrix == NULL)
      return;
   else if (NRows > 0)
      e_free(matrix[0]);

   e_free(matrix);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_cont_c_matr: Character matrix. */


void e_free_cont_c_matr(unsigned int NRows, char **matrix)
{
   if (matrix == NULL)
      return;
   else if (NRows > 0)
      e_free(matrix[0]);

   e_free(matrix);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_cont_vp_matr: Void pointer matrix. */


void e_free_cont_vp_matr(unsigned int NRows, void ***matrix)
{
   if (matrix == NULL)
      return;
   else if (NRows > 0)
      e_free(matrix[0]);

   e_free(matrix);
}


/* ////////////////////////////////////////////////////////////////////////// */

/* The following deallocate the memory of an "NDepth" x "NRows" x "NCols"
   fragmented 3D matrix.                                                      */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_frag_i_3D_matr: 3D "int" integer matrix. */


void e_free_frag_i_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                    int ***matrix3D)
{
   register unsigned int i;


   if (matrix3D == NULL) return;

   for (i = 0; i < NDepth; i++)
      e_free_frag_i_matr(NRows,matrix3D[i]);

   e_free(matrix3D);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_frag_f_3D_matr: 3D float matrix. */


void e_free_frag_f_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                  float ***matrix3D)
{
   register unsigned int i;


   if (matrix3D == NULL) return;

   for (i = 0; i < NDepth; i++)
      e_free_frag_f_matr(NRows,matrix3D[i]);

   e_free(matrix3D);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_frag_d_3D_matr: 3D double precision "matrix. */


void e_free_frag_d_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                 double ***matrix3D)
{
   register unsigned int i;


   if (matrix3D == NULL) return;

   for (i = 0; i < NDepth; i++)
      e_free_frag_d_matr(NRows,matrix3D[i]);

   e_free(matrix3D);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_frag_c_3D_matr: 3D character matrix. */


void e_free_frag_c_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                   char ***matrix3D)
{
   register unsigned int i;


   if (matrix3D == NULL) return;

   for (i = 0; i < NDepth; i++)
      e_free_frag_c_matr(NRows,matrix3D[i]);

   e_free(matrix3D);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_frag_vp_3D_matr: 3D void pointer matrix. */


void e_free_frag_vp_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                   void ****matrix3D)
{
   register unsigned int i;


   if (matrix3D == NULL) return;

   for (i = 0; i < NDepth; i++)
      e_free_frag_vp_matr(NRows,matrix3D[i]);

   e_free(matrix3D);
}


/* ////////////////////////////////////////////////////////////////////////// */

/* The following deallocate the memory of an "NDepth" x "NRows" x "NCols"
   contiguous 3D matrix.                                                      */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_cont_i_3D_matr: 3D "int" integer matrix. */


void e_free_cont_i_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                    int ***matrix3D)
{
   register unsigned int k;


   if (matrix3D == NULL || NDepth == 0 || NRows == 0)
   {
      e_free(matrix3D);
      return;
   }

   e_free(matrix3D[0][0]);

   for (k = 0; k < NDepth; k++)
      e_free(matrix3D[k]);

   e_free(matrix3D);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_cont_f_3D_matr: 3D float matrix. */


void e_free_cont_f_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                  float ***matrix3D)
{
   register unsigned int k;


   if (matrix3D == NULL || NDepth == 0 || NRows == 0)
   {
      e_free(matrix3D);
      return;
   }

   e_free(matrix3D[0][0]);

   for (k = 0; k < NDepth; k++)
      e_free(matrix3D[k]);

   e_free(matrix3D);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_cont_d_3D_matr: 3D double precision matrix. */


void e_free_cont_d_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                 double ***matrix3D)
{
   register unsigned int k;


   if (matrix3D == NULL || NDepth == 0 || NRows == 0)
   {
      e_free(matrix3D);
      return;
   }

   e_free(matrix3D[0][0]);

   for (k = 0; k < NDepth; k++)
      e_free(matrix3D[k]);

   e_free(matrix3D);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_cont_c_3D_matr: 3D character matrix. */


void e_free_cont_c_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                   char ***matrix3D)
{
   register unsigned int k;


   if (matrix3D == NULL || NDepth == 0 || NRows == 0)
   {
      e_free(matrix3D);
      return;
   }

   e_free(matrix3D[0][0]);

   for (k = 0; k < NDepth; k++)
      e_free(matrix3D[k]);

   e_free(matrix3D);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_cont_vp_3D_matr: 3D void pointer matrix. */


void e_free_cont_vp_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                   void ****matrix3D)
{
   register unsigned int k;


   if (matrix3D == NULL || NDepth == 0 || NRows == 0)
   {
      e_free(matrix3D);
      return;
   }

   e_free(matrix3D[0][0]);

   for (k = 0; k < NDepth; k++)
      e_free(matrix3D[k]);

   e_free(matrix3D);
}


/* ////////////////////////////////////////////////////////////////////////// */

/* The following deallocate the memory of an 
   "NDepth" x "NWidth" x "NRows" x "NCols" fragmented 4D matrix.              */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_frag_i_4D_matr: 4D "int" integer matrix. */


void e_free_frag_i_4D_matr(unsigned int NDepth, unsigned int NWidth,
                           unsigned int NRows,  int ****matrix4D    )
{
   register unsigned int i;


   if (matrix4D == NULL) return;

   for (i = 0; i < NDepth; i++)
      e_free_frag_i_3D_matr(NWidth,NRows,matrix4D[i]);

   e_free(matrix4D);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_frag_f_4D_matr: 4D float matrix. */


void e_free_frag_f_4D_matr(unsigned int NDepth, unsigned int NWidth,
                           unsigned int NRows,  float ****matrix4D  )
{
   register unsigned int i;


   if (matrix4D == NULL) return;

   for (i = 0; i < NDepth; i++)
      e_free_frag_f_3D_matr(NWidth,NRows,matrix4D[i]);

   e_free(matrix4D);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_frag_d_4D_matr: 4D double precision matrix. */


void e_free_frag_d_4D_matr(unsigned int NDepth, unsigned int NWidth,
                           unsigned int NRows,  double ****matrix4D )
{
   register unsigned int i;


   if (matrix4D == NULL) return;

   for (i = 0; i < NDepth; i++)
      e_free_frag_d_3D_matr(NWidth,NRows,matrix4D[i]);

   e_free(matrix4D);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_frag_c_4D_matr: 4D character matrix. */


void e_free_frag_c_4D_matr(unsigned int NDepth, unsigned int NWidth,
                           unsigned int NRows,  char ****matrix4D   )
{
   register unsigned int i;


   if (matrix4D == NULL) return;

   for (i = 0; i < NDepth; i++)
      e_free_frag_c_3D_matr(NWidth,NRows,matrix4D[i]);

   e_free(matrix4D);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_frag_vp_4D_matr: 4D void pointer matrix. */


void e_free_frag_vp_4D_matr(unsigned int NDepth, unsigned int NWidth,
                            unsigned int NRows,  void *****matrix4D  )
{
   register unsigned int i;


   if (matrix4D == NULL) return;

   for (i = 0; i < NDepth; i++)
      e_free_frag_vp_3D_matr(NWidth,NRows,matrix4D[i]);

   e_free(matrix4D);
}


/* ////////////////////////////////////////////////////////////////////////// */

/* The following deallocate the memory of an
   "NDepth" x "NWidth" x "NRows" x "NCols" contiguous 4D matrix.              */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_cont_i_4D_matr: 4D "int" integer matrix. */


void e_free_cont_i_4D_matr(unsigned int NDepth, unsigned int NWidth,
                           unsigned int NRows,  int ****matrix4D    )
{
   register unsigned int k,l;


   if (matrix4D == NULL || NDepth == 0 || NWidth == 0 || NRows == 0)
   {
      e_free(matrix4D);
      return;
   }

   e_free(matrix4D[0][0][0]);

   for (k = 0; k < NDepth; k++)
   {
      for (l = 0; l < NWidth; l++)
         e_free(matrix4D[k][l]);

      e_free(matrix4D[k]);
   }

   e_free(matrix4D);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_cont_f_4D_matr: 4D float matrix. */


void e_free_cont_f_4D_matr(unsigned int NDepth, unsigned int NWidth,
                           unsigned int NRows,  float ****matrix4D  )
{
   register unsigned int k,l;


   if (matrix4D == NULL || NDepth == 0 || NWidth == 0 || NRows == 0)
   {
      e_free(matrix4D);
      return;
   }

   e_free(matrix4D[0][0][0]);

   for (k = 0; k < NDepth; k++)
   {
      for (l = 0; l < NWidth; l++)
         e_free(matrix4D[k][l]);

      e_free(matrix4D[k]);
   }

   e_free(matrix4D);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_cont_d_4D_matr: 4D double precision matrix. */


void e_free_cont_d_4D_matr(unsigned int NDepth, unsigned int NWidth,
                           unsigned int NRows,  double ****matrix4D )
{
   register unsigned int k,l;


   if (matrix4D == NULL || NDepth == 0 || NWidth == 0 || NRows == 0)
   {
      e_free(matrix4D);
      return;
   }

   e_free(matrix4D[0][0][0]);

   for (k = 0; k < NDepth; k++)
   {
      for (l = 0; l < NWidth; l++)
         e_free(matrix4D[k][l]);

      e_free(matrix4D[k]);
   }

   e_free(matrix4D);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_cont_c_4D_matr: 4D character matrix. */


void e_free_cont_c_4D_matr(unsigned int NDepth, unsigned int NWidth,
                           unsigned int NRows,  char ****matrix4D   )
{
   register unsigned int k,l;


   if (matrix4D == NULL || NDepth == 0 || NWidth == 0 || NRows == 0)
   {
      e_free(matrix4D);
      return;
   }

   e_free(matrix4D[0][0][0]);

   for (k = 0; k < NDepth; k++)
   {
      for (l = 0; l < NWidth; l++)
         e_free(matrix4D[k][l]);

      e_free(matrix4D[k]);
   }

   e_free(matrix4D);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_free_cont_vp_4D_matr: 4D void pointer matrix. */


void e_free_cont_vp_4D_matr(unsigned int NDepth, unsigned int NWidth,
                            unsigned int NRows,  void *****matrix4D  )
{
   register unsigned int k,l;


   if (matrix4D == NULL || NDepth == 0 || NWidth == 0 || NRows == 0)
   {
      e_free(matrix4D);
      return;
   }

   e_free(matrix4D[0][0][0]);

   for (k = 0; k < NDepth; k++)
   {
      for (l = 0; l < NWidth; l++)
         e_free(matrix4D[k][l]);

      e_free(matrix4D[k]);
   }

   e_free(matrix4D);
}

/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                        Mesh Initialisation Function                        */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_NullifyMesh: To set all mesh integers to zero, all mesh Booleans to
   FALSE and all mesh pointers to NULL.                                       */


void e_NullifyMesh(MESH *mesh)
{
   int i;


   mesh->NMeshPoints = 0;
   mesh->NNodes = 0;
   mesh->NElemsLink1 = 0;
   mesh->NElemsLink2 = 0;
   mesh->NElemsLink3 = 0;
   mesh->NElemsLink4 = 0;
   mesh->NElemsLink5 = 0;
   mesh->NElemsTriang1 = 0;
   mesh->NElemsTriang2 = 0;
   mesh->NElemsTriang3 = 0;
   mesh->NElemsTriang4 = 0;
   mesh->NElemsTriang5 = 0;
   mesh->NElemsQuad1 = 0;
   mesh->NElemsQuad3 = 0;
   mesh->NElemsTetrah1 = 0;
   mesh->NElemsBlock1 = 0;
   mesh->TotalNElems = 0;
   mesh->NBCNodes = 0;
   mesh->NLoadedNodes = 0;
   mesh->NMats = 0;
   mesh->NMdfMats = 0;
   mesh->NCompMatsType1 = 0;
   mesh->NTimeSteps = 0;
   mesh->NInternalTimeSteps = 0;
   mesh->NExternalTimeSteps = 0;
   mesh->TotalNodalDOF = 0;

   mesh->NStressPointsLink1   = 0;
   mesh->NStressPointsLink2   = 0;
   mesh->NStressPointsLink3   = 0;
   mesh->NStressPointsLink4   = 0;
   mesh->NStressPointsLink5   = 0;
   mesh->NStressPointsTriang1 = 0;
   mesh->NStressPointsTriang2 = 0;
   mesh->NStressPointsTriang3 = 0;
   mesh->NStressPointsTriang4 = 0;
   mesh->NStressPointsTriang5 = 0;
   mesh->NStressPointsQuad1   = 0;
   mesh->NStressPointsQuad3   = 0;
   mesh->NStressPointsTetrah1 = 0;
   mesh->NStressPointsBlock1  = 0;

   mesh->Title = NULL;
   mesh->MeshPointCoords = NULL;
   mesh->NodesLink1 = NULL;
   mesh->NodesLink2 = NULL;
   mesh->NodesLink3 = NULL;
   mesh->NodesLink4 = NULL;
   mesh->NodesLink5 = NULL;
   mesh->NodesTriang1 = NULL;
   mesh->NodesTriang2 = NULL;
   mesh->NodesTriang3 = NULL;
   mesh->NodesTriang4 = NULL;
   mesh->NodesTriang5 = NULL;
   mesh->NodesQuad1 = NULL;
   mesh->NodesQuad3 = NULL;
   mesh->NodesTetrah1 = NULL;
   mesh->NodesBlock1 = NULL;
   mesh->NodalDOF = NULL;
   mesh->BCNodes = NULL;
   mesh->BCTypes = NULL;
   mesh->BCDispls = NULL;
   mesh->BCSpringConsts = NULL;
   mesh->LoadedNodes = NULL;
   mesh->Loads = NULL;
   mesh->Mats = NULL;
   mesh->MatsLink1 = NULL;
   mesh->MatsLink2 = NULL;
   mesh->MatsLink3 = NULL;
   mesh->MatsLink4 = NULL;
   mesh->MatsLink5 = NULL;
   mesh->MatsTriang1 = NULL;
   mesh->CompMatsTriang2 = NULL;
   mesh->MatsTriang3 = NULL;
   mesh->MatsTriang4 = NULL;
   mesh->MatsTriang5 = NULL;
   mesh->MatsQuad1 = NULL;
   mesh->CompMatsQuad3 = NULL;
   mesh->MatsTetrah1 = NULL;
   mesh->MatsBlock1 = NULL;
   mesh->NodeTypes = NULL;
   mesh->NNurbsCurves = NULL;
   mesh->NurbsCurveParams = NULL;
   mesh->NurbsCurves = NULL;
   mesh->GNIs = NULL;
   mesh->GEIs = NULL;
   mesh->GEIsLink1 = NULL;
   mesh->GEIsLink2 = NULL;
   mesh->GEIsLink3 = NULL;
   mesh->GEIsLink4 = NULL;
   mesh->GEIsLink5 = NULL;
   mesh->GEIsTriang1 = NULL;
   mesh->GEIsTriang2 = NULL;
   mesh->GEIsTriang3 = NULL;
   mesh->GEIsTriang4 = NULL;
   mesh->GEIsTriang5 = NULL;
   mesh->GEIsQuad1 = NULL;
   mesh->GEIsQuad3 = NULL;
   mesh->GEIsTetrah1 = NULL;
   mesh->GEIsBlock1 = NULL;
   mesh->Props = NULL;
   mesh->CompPropsType1 = NULL;

   mesh->Geom.NEndPoints = 0;
   mesh->Geom.NNurbsCurves = 0;
   mesh->Geom.EndPointCoords = NULL;
   mesh->Geom.NurbsCurves = NULL;

   mesh->Decomp.NSubDoms = 0;
   mesh->Decomp.SubDoms = NULL;
   mesh->Decomp.SubDomsLink1 = NULL;
   mesh->Decomp.SubDomsLink2 = NULL;
   mesh->Decomp.SubDomsLink3 = NULL;
   mesh->Decomp.SubDomsLink4 = NULL;
   mesh->Decomp.SubDomsLink5 = NULL;
   mesh->Decomp.SubDomsTriang1 = NULL;
   mesh->Decomp.SubDomsTriang2 = NULL;
   mesh->Decomp.SubDomsTriang3 = NULL;
   mesh->Decomp.SubDomsTriang4 = NULL;
   mesh->Decomp.SubDomsTriang5 = NULL;
   mesh->Decomp.SubDomsQuad1 = NULL;
   mesh->Decomp.SubDomsQuad3 = NULL;
   mesh->Decomp.SubDomsTetrah1 = NULL;
   mesh->Decomp.SubDomsBlock1 = NULL;

   mesh->MeshParam.NodalMeshParamsSet = FALSE;
   mesh->MeshParam.ElemMeshParamsSet = FALSE;
   mesh->MeshParam.NodalMeshParams = NULL;
   mesh->MeshParam.ElemMeshParams = NULL;
   mesh->MeshParam.ElemMeshParamsLink1 = NULL;
   mesh->MeshParam.ElemMeshParamsLink2 = NULL;
   mesh->MeshParam.ElemMeshParamsLink3 = NULL;
   mesh->MeshParam.ElemMeshParamsLink4 = NULL;
   mesh->MeshParam.ElemMeshParamsLink5 = NULL;
   mesh->MeshParam.ElemMeshParamsTriang1 = NULL;
   mesh->MeshParam.ElemMeshParamsTriang2 = NULL;
   mesh->MeshParam.ElemMeshParamsTriang3 = NULL;
   mesh->MeshParam.ElemMeshParamsTriang4 = NULL;
   mesh->MeshParam.ElemMeshParamsTriang5 = NULL;
   mesh->MeshParam.ElemMeshParamsQuad1 = NULL;
   mesh->MeshParam.ElemMeshParamsQuad3 = NULL;
   mesh->MeshParam.ElemMeshParamsTetrah1 = NULL;
   mesh->MeshParam.ElemMeshParamsBlock1 = NULL;

   mesh->FEError.FEErrors = NULL;
   mesh->FEError.FEErrorsLink1 = NULL;
   mesh->FEError.FEErrorsLink2 = NULL;
   mesh->FEError.FEErrorsLink3 = NULL;
   mesh->FEError.FEErrorsLink4 = NULL;
   mesh->FEError.FEErrorsLink5 = NULL;
   mesh->FEError.FEErrorsTriang1 = NULL;
   mesh->FEError.FEErrorsTriang2 = NULL;
   mesh->FEError.FEErrorsTriang3 = NULL;
   mesh->FEError.FEErrorsTriang4 = NULL;
   mesh->FEError.FEErrorsTriang5 = NULL;
   mesh->FEError.FEErrorsQuad1 = NULL;
   mesh->FEError.FEErrorsQuad3 = NULL;
   mesh->FEError.FEErrorsTetrah1 = NULL;
   mesh->FEError.FEErrorsBlock1 = NULL;

   mesh->Stresses.StressesLink1   = NULL;
   mesh->Stresses.StressesLink2   = NULL;
   mesh->Stresses.StressesLink3   = NULL;
   mesh->Stresses.StressesLink4   = NULL;
   mesh->Stresses.StressesLink5   = NULL;
   mesh->Stresses.StressesTriang1 = NULL;
   mesh->Stresses.StressesTriang2 = NULL;
   mesh->Stresses.StressesTriang3 = NULL;
   mesh->Stresses.StressesTriang4 = NULL;
   mesh->Stresses.StressesTriang5 = NULL;
   mesh->Stresses.StressesQuad1   = NULL;
   mesh->Stresses.StressesQuad3   = NULL;
   mesh->Stresses.StressesTetrah1 = NULL;
   mesh->Stresses.StressesBlock1  = NULL;

   for (i = 0; i < E_NELEM_TYPES; i++)
      mesh->FirstNetIndex[i] = 0;

   for (i = 0; i < E_NKEYWORDS; i++)
      mesh->KeywordSet[i] = FALSE;

   for (i = 0; i < E_NSUBSTRUCTS; i++)
      mesh->SubStructSet[i] = FALSE;
}


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                            Mesh Type Functions                             */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_IsOfOneElementType: To return whether a mesh is composed solely of one
   element type.                                                              */


int e_IsOfOneElementType(MESH *mesh)
{
   int NElementTypes = 0;


   if (mesh->NElemsLink1   > 0) NElementTypes++;
   if (mesh->NElemsLink2   > 0) NElementTypes++;
   if (mesh->NElemsLink3   > 0) NElementTypes++;
   if (mesh->NElemsLink4   > 0) NElementTypes++;
   if (mesh->NElemsLink5   > 0) NElementTypes++;
   if (mesh->NElemsTriang1 > 0) NElementTypes++;
   if (mesh->NElemsTriang2 > 0) NElementTypes++;
   if (mesh->NElemsTriang3 > 0) NElementTypes++;
   if (mesh->NElemsTriang4 > 0) NElementTypes++;
   if (mesh->NElemsTriang5 > 0) NElementTypes++;
   if (mesh->NElemsQuad1   > 0) NElementTypes++;
   if (mesh->NElemsQuad3   > 0) NElementTypes++;
   if (mesh->NElemsTetrah1 > 0) NElementTypes++;
   if (mesh->NElemsBlock1  > 0) NElementTypes++;

   if (NElementTypes == 1)
      return 1;
   else
      return 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_LinkElementsOnly: To return whether a mesh contains only link (1-D)
   elements.                                                                  */


int e_LinkElementsOnly(MESH *mesh)
{
   int NLinkElems = 0;
   int NPlanarElems = 0;
   int N3DElems = 0;


   NLinkElems =   mesh->NElemsLink1 + mesh->NElemsLink2
                + mesh->NElemsLink3 + mesh->NElemsLink4
                + mesh->NElemsLink5;
   NPlanarElems =   mesh->NElemsTriang1 + mesh->NElemsTriang2
                  + mesh->NElemsTriang3 + mesh->NElemsTriang4
                  + mesh->NElemsTriang5 + mesh->NElemsQuad1
                  + mesh->NElemsQuad3;
   N3DElems = mesh->NElemsTetrah1 + mesh->NElemsBlock1;

   if (NLinkElems > 0 && NPlanarElems == 0 && N3DElems == 0)
      return 1;
   else
      return 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_PlanarElementsOnly: To return whether a mesh contains only planar
   elements.                                                                  */


int e_PlanarElementsOnly(MESH *mesh)
{
   int NLinkElems = 0;
   int NPlanarElems = 0;
   int N3DElems = 0;


   NLinkElems =   mesh->NElemsLink1 + mesh->NElemsLink2
                + mesh->NElemsLink3 + mesh->NElemsLink4
                + mesh->NElemsLink5;
   NPlanarElems =   mesh->NElemsTriang1 + mesh->NElemsTriang2
                  + mesh->NElemsTriang3 + mesh->NElemsTriang4
                  + mesh->NElemsTriang5 + mesh->NElemsQuad1
                  + mesh->NElemsQuad3;
   N3DElems = mesh->NElemsTetrah1 + mesh->NElemsBlock1;

   if (NPlanarElems > 0 && NLinkElems == 0 && N3DElems == 0)
      return 1;
   else
      return 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                               Exit Function                                */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */
/* e_exit: To terminate execution. */


void e_exit(int ExitStatus)
{

#ifdef __E_CLANSIC__
#ifdef __E_PVM__
   pvm_exit();
#endif
   exit(ExitStatus);
#endif

#ifdef __E_XWIN__
   /* To Do - Put X-Windows code here */
   exit(ExitStatus);
#endif

#ifdef __E_MSWIN__
   exit(ExitStatus);
#endif

#ifdef __E_T800RTL__
   exit(ExitStatus);
#endif

#ifdef __E_T800SAL__
   for(;;);
#endif

}
