/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                      e_GetFEErrors(), e_PutFEErrors()                      */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include "e_pltfrm.h"
#include <stdio.h>
#include <string.h>
#include <e_lib.h>
#include "comsrc.h"
#include "feerrors.h"


/* ////////////////////////////////////////////////////////////////////////// */
/* e_GetFEErrors: To read in finite element error data. */
/* The return value is 1 if no input errors occur, else 0. */


int e_GetFEErrors(MESH *mesh, char *InputFile)
{
   FullPathName = OpenInputFile(InputFile,E_FEE,&fp);

   if (FullPathName == NULL || !Initialise(mesh))
   {
      e_free(FullPathName);
      return 0;
   }

   if (GetData(mesh) && AllCompulsoryKeywordsSet())
   {
      SetMeshData(mesh);
      e_free(FullPathName);
      fclose(fp);
      return 1;
   }
   else
   {
      DeallocateAllMemory();
      fclose(fp);
      return 0;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* initialise: To initialise variables. */


static int Initialise(MESH *mesh)
{
   int sum = 0;

   NElemsLink1 = mesh->NElemsLink1;
   NElemsLink2 = mesh->NElemsLink2;
   NElemsLink3 = mesh->NElemsLink3;
   NElemsLink4 = mesh->NElemsLink4;
   NElemsLink5 = mesh->NElemsLink5;
   NElemsTriang1 = mesh->NElemsTriang1;
   NElemsTriang2 = mesh->NElemsTriang2;
   NElemsTriang3 = mesh->NElemsTriang3;
   NElemsTriang4 = mesh->NElemsTriang4;
   NElemsTriang5 = mesh->NElemsTriang5;
   NElemsQuad1 = mesh->NElemsQuad1;
   NElemsQuad3 = mesh->NElemsQuad3;
   NElemsTetrah1 = mesh->NElemsTetrah1;
   NElemsBlock1 = mesh->NElemsBlock1;
   TotalNElems = mesh->TotalNElems;
   LineNumber = 0;
   AllCompKeywordsSet = FALSE;

   FEErrors = e_alloc_d_vect(TotalNElems);

   if (FEErrors == NULL)
      return 0;

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
   {
      FEErrorsLink1 = FEErrors + sum;
      sum += mesh->NElemsLink1;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2]) 
   {
      FEErrorsLink2 = FEErrors + sum;
      sum += mesh->NElemsLink2;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3]) 
   {
      FEErrorsLink3 = FEErrors + sum;
      sum += mesh->NElemsLink3;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4]) 
   {
      FEErrorsLink4 = FEErrors + sum;
      sum += mesh->NElemsLink4;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5]) 
   {
      FEErrorsLink5 = FEErrors + sum;
      sum += mesh->NElemsLink5;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
   {
      FEErrorsTriang1 = FEErrors + sum;
      sum += mesh->NElemsTriang1;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2]) 
   {
      FEErrorsTriang2 = FEErrors + sum;
      sum += mesh->NElemsTriang2;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3]) 
   {
      FEErrorsTriang3 = FEErrors + sum;
      sum += mesh->NElemsTriang3;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4]) 
   {
      FEErrorsTriang4 = FEErrors + sum;
      sum += mesh->NElemsTriang4;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5]) 
   {
      FEErrorsTriang5 = FEErrors + sum;
      sum += mesh->NElemsTriang5;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1]) 
   {
      FEErrorsQuad1 = FEErrors + sum;
      sum += mesh->NElemsQuad1;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3]) 
   {
      FEErrorsQuad3 = FEErrors + sum;
      sum += mesh->NElemsQuad3;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1]) 
   {
      FEErrorsTetrah1 = FEErrors + sum;
      sum += mesh->NElemsTetrah1;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1]) 
   {
      FEErrorsBlock1 = FEErrors + sum;
      sum += mesh->NElemsBlock1;
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetData: To read in the input data. */

/* The return value is 1 if the data are read successfully or the end of file
   is encountered, else 0.                                                    */


static int GetData(MESH *mesh)
{
   char word[E_LINE_MAX + 1];      /* potential keyword */

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_FEERRORS_LINK1,word)
               || !GetFEErrors(NElemsLink1,FEErrorsLink1)       )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_FEERRORS_LINK2,word)
               || !GetFEErrors(NElemsLink2,FEErrorsLink2)       )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_FEERRORS_LINK3,word)
               || !GetFEErrors(NElemsLink3,FEErrorsLink3)       )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_FEERRORS_LINK4,word)
               || !GetFEErrors(NElemsLink4,FEErrorsLink4)       )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_FEERRORS_LINK5,word)
               || !GetFEErrors(NElemsLink5,FEErrorsLink5)       )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_FEERRORS_TRIANG1,word)
               || !GetFEErrors(NElemsTriang1,FEErrorsTriang1)     )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_FEERRORS_TRIANG2,word)
               || !GetFEErrors(NElemsTriang2,FEErrorsTriang2)     )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_FEERRORS_TRIANG3,word)
               || !GetFEErrors(NElemsTriang3,FEErrorsTriang3)     )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_FEERRORS_TRIANG4,word)
               || !GetFEErrors(NElemsTriang4,FEErrorsTriang4)     )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_FEERRORS_TRIANG5,word)
               || !GetFEErrors(NElemsTriang5,FEErrorsTriang5)     )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_FEERRORS_QUAD1,word)
               || !GetFEErrors(NElemsQuad1,FEErrorsQuad1)       )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_FEERRORS_QUAD3,word)
               || !GetFEErrors(NElemsQuad3,FEErrorsQuad3)       )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_FEERRORS_TETRAH1,word)
               || !GetFEErrors(NElemsTetrah1,FEErrorsTetrah1)     )
         return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
   {
      if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
         return 1;
      else if (   !WordMatchesKeyword(E_KW_FEERRORS_BLOCK1,word)
               || !GetFEErrors(NElemsBlock1,FEErrorsBlock1)      )
         return 0;
   }

   if (GetKeyword(line,word,&LineNumber,FullPathName,fp) != EOF)
   {
      e_FileLineError(LineNumber,FullPathName,"undefined keyword");
      return 0;
   }

   AllCompKeywordsSet = TRUE;
   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* WordMatchesKeyword: To return whether a word is a specified keyword. */

/* The return value is 1 if the word matches the keyword, else 0. */


static int WordMatchesKeyword(char *Keyword, char *word)
{
   if (strcmp(Keyword,word) != 0)
   {
      if (IsAKeyword(word))
         e_FileLineError(LineNumber,FullPathName,"keyword out of sequence");
      else
         e_FileLineError(LineNumber,FullPathName,"undefined keyword");

      return 0;
   }
   else
      return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* Get... input functions: To read in data corresponding to a keyword. */

/* The return value is 1 if the data are read successfully, else 0. */


static int GetFEErrors(int NElements, double *FEErrors)
{
   int i;
   int FEErrorIndex;


   for (i = 0; i < NElements; i++)
   {
      if (e_GetNextLine(line,&LineNumber,FullPathName,fp) == 0)
         return 0;

      if (sscanf(line,"%d %lf",&FEErrorIndex,&FEErrors[i]) == 2)
      {
         if (FEErrorIndex != i + 1)
         {
            e_FileLineError(LineNumber,FullPathName,
                            "error index out of sequence");
            return 0;
         }
      }
      else
      {
         e_FileLineError(LineNumber,FullPathName,"cannot read error");
         return 0;
      }
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* IsAKeyword: To return whether a word is a keyword. */

/* The return value is 1 if "word" is a keyword, else 0. */


static int IsAKeyword(char *word)
{
   if (     strcmp(E_KW_FEERRORS_LINK1,word) == 0
       ||   strcmp(E_KW_FEERRORS_LINK2,word) == 0
       ||   strcmp(E_KW_FEERRORS_LINK3,word) == 0
       ||   strcmp(E_KW_FEERRORS_LINK4,word) == 0
       ||   strcmp(E_KW_FEERRORS_LINK5,word) == 0
       || strcmp(E_KW_FEERRORS_TRIANG1,word) == 0
       || strcmp(E_KW_FEERRORS_TRIANG2,word) == 0
       || strcmp(E_KW_FEERRORS_TRIANG3,word) == 0
       || strcmp(E_KW_FEERRORS_TRIANG4,word) == 0
       || strcmp(E_KW_FEERRORS_TRIANG5,word) == 0
       ||   strcmp(E_KW_FEERRORS_QUAD1,word) == 0
       ||   strcmp(E_KW_FEERRORS_QUAD3,word) == 0
       || strcmp(E_KW_FEERRORS_TETRAH1,word) == 0
       ||  strcmp(E_KW_FEERRORS_BLOCK1,word) == 0)
   {
      return TRUE;
   }
   else
   {
      return FALSE;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* AllCompulsorySet: To return whether all compulsory keywords have been set. */


static int AllCompulsoryKeywordsSet()
{
   if (!AllCompKeywordsSet)
   {
      e_FileError(FullPathName,"unexpected end of file");
      return 0;
   }
   else
      return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetMeshData: To set the MESH structure data. */


static void SetMeshData(MESH *mesh)
{
   mesh->FEError.FEErrors = FEErrors;
   mesh->FEError.FEErrorsLink1 = FEErrorsLink1;
   mesh->FEError.FEErrorsLink2 = FEErrorsLink2;
   mesh->FEError.FEErrorsLink3 = FEErrorsLink3;
   mesh->FEError.FEErrorsLink4 = FEErrorsLink4;
   mesh->FEError.FEErrorsLink5 = FEErrorsLink5;
   mesh->FEError.FEErrorsTriang1 = FEErrorsTriang1;
   mesh->FEError.FEErrorsTriang2 = FEErrorsTriang2;
   mesh->FEError.FEErrorsTriang3 = FEErrorsTriang3;
   mesh->FEError.FEErrorsTriang4 = FEErrorsTriang4;
   mesh->FEError.FEErrorsTriang5 = FEErrorsTriang5;
   mesh->FEError.FEErrorsQuad1 = FEErrorsQuad1;
   mesh->FEError.FEErrorsQuad3 = FEErrorsQuad3;
   mesh->FEError.FEErrorsTetrah1 = FEErrorsTetrah1;
   mesh->FEError.FEErrorsBlock1 = FEErrorsBlock1;

   mesh->SubStructSet[E_SS_NUM_FEERROR] = TRUE;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* DeallocateAllMemory: To deallocate all memory. */


static void DeallocateAllMemory()
{
   e_free(FEErrors);
   e_free(FullPathName);
}


/* ////////////////////////////////////////////////////////////////////////// */










/* ////////////////////////////////////////////////////////////////////////// */
/* e_PutFEErrors: To output finite element errors. */
/* The return value is 1 if the data are written successfully, else 0. */


int e_PutFEErrors(MESH *mesh, char *OutputFile)
{
   int i;

   FILE *fp;                       /* output file pointer */


   if (!mesh->SubStructSet[E_SS_NUM_FEERROR])
   {
      e_GeneralError("e_PutFEErrors: finite element errors not set");
      return 0;
   }
   else if (OpenOutputFile(OutputFile,E_FEE,&fp) == NULL)
   {
      return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
   {
      fprintf(fp,"\n\n# LINK1 finite element errors\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#              error\n");

      fprintf(fp,"\n%s\n",E_KW_FEERRORS_LINK1);

      for (i = 0; i < mesh->NElemsLink1; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%20.8f\n",mesh->FEError.FEErrorsLink1[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
   {
      fprintf(fp,"\n\n# LINK2 finite element errors\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#              error\n");

      fprintf(fp,"\n%s\n",E_KW_FEERRORS_LINK2);

      for (i = 0; i < mesh->NElemsLink2; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%20.8f\n",mesh->FEError.FEErrorsLink2[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
   {
      fprintf(fp,"\n\n# LINK3 finite element errors\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#              error\n");

      fprintf(fp,"\n%s\n",E_KW_FEERRORS_LINK3);

      for (i = 0; i < mesh->NElemsLink3; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%20.8f\n",mesh->FEError.FEErrorsLink3[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
   {
      fprintf(fp,"\n\n# LINK4 finite element errors\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#              error\n");

      fprintf(fp,"\n%s\n",E_KW_FEERRORS_LINK4);

      for (i = 0; i < mesh->NElemsLink4; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%20.8f\n",mesh->FEError.FEErrorsLink4[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
   {
      fprintf(fp,"\n\n# LINK5 finite element errors\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#              error\n");

      fprintf(fp,"\n%s\n",E_KW_FEERRORS_LINK5);

      for (i = 0; i < mesh->NElemsLink5; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%20.8f\n",mesh->FEError.FEErrorsLink5[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
   {
      fprintf(fp,"\n\n# TRIANG1 finite element errors\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#              error\n");

      fprintf(fp,"\n%s\n",E_KW_FEERRORS_TRIANG1);

      for (i = 0; i < mesh->NElemsTriang1; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%20.8f\n",mesh->FEError.FEErrorsTriang1[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
   {
      fprintf(fp,"\n\n# TRIANG2 finite element errors\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#              error\n");

      fprintf(fp,"\n%s\n",E_KW_FEERRORS_TRIANG2);

      for (i = 0; i < mesh->NElemsTriang2; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%20.8f\n",mesh->FEError.FEErrorsTriang2[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
   {
      fprintf(fp,"\n\n# TRIANG3 finite element errors\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#              error\n");

      fprintf(fp,"\n%s\n",E_KW_FEERRORS_TRIANG3);

      for (i = 0; i < mesh->NElemsTriang3; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%20.8f\n",mesh->FEError.FEErrorsTriang3[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
   {
      fprintf(fp,"\n\n# TRIANG4 finite element errors\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#              error\n");

      fprintf(fp,"\n%s\n",E_KW_FEERRORS_TRIANG4);

      for (i = 0; i < mesh->NElemsTriang4; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%20.8f\n",mesh->FEError.FEErrorsTriang4[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
   {
      fprintf(fp,"\n\n# TRIANG5 finite element errors\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#              error\n");

      fprintf(fp,"\n%s\n", E_KW_FEERRORS_TRIANG5);

      for (i = 0; i < mesh->NElemsTriang5; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%20.8f\n",mesh->FEError.FEErrorsTriang5[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
   {
      fprintf(fp,"\n\n# QUAD1 finite element errors\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#              error\n");

      fprintf(fp,"\n%s\n",E_KW_FEERRORS_QUAD1);

      for (i = 0; i < mesh->NElemsQuad1; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%20.8f\n",mesh->FEError.FEErrorsQuad1[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
   {
      fprintf(fp,"\n\n# QUAD3 finite element errors\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#              error\n");

      fprintf(fp,"\n%s\n",E_KW_FEERRORS_QUAD3);

      for (i = 0; i < mesh->NElemsQuad3; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%20.8f\n",mesh->FEError.FEErrorsQuad3[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
   {
      fprintf(fp,"\n\n# TETRAH1 finite element errors\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#              error\n");

      fprintf(fp,"\n%s\n",E_KW_FEERRORS_TETRAH1);

      for (i = 0; i < mesh->NElemsTetrah1; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%20.8f\n",mesh->FEError.FEErrorsTetrah1[i]);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
   {
      fprintf(fp,"\n\n# BLOCK1 finite element errors\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#              error\n");

      fprintf(fp,"\n%s\n",E_KW_FEERRORS_BLOCK1);

      for (i = 0; i < mesh->NElemsBlock1; i++)
      {
         fprintf(fp,"%7d   ",i + 1);
         fprintf(fp,"%20.8f\n",mesh->FEError.FEErrorsBlock1[i]);
      }
   }

   fclose(fp);
   return 1;
}
