/*  File: E_MPI-AU.C   */


/*
    This is an auxiliary library for MPI program development
    to aid with the E_LIB library.
    Functions have been written to send and receive a mesh definiton
    to MPI processes using the new FE mesh data format.
    
    There is no limit on the size of the messages.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <mpi.h>

#ifndef __E_MPI__
#define __E_MPI__
#endif

#include <e_lib.h>
#include "mpi_aux.h"

#define    MESH_STRUC_IBUF_LENGTH    41
#define    MESH_STRUC_DBUF_LENGTH    4

#ifdef __E_T800__
void NotEnoughMemory()
{
  e_GeneralError("Not enough memory");
}
#endif


void e_MPI_Send_MeshData(MESH *mesh, int dest)
{
  register  i,j;
  int      *ibuf_MeshStruc;
  double   *dbuf_MeshStruc;
  int      *vect_ind;


  if( (ibuf_MeshStruc = e_alloc_i_vect(MESH_STRUC_IBUF_LENGTH)) == NULL)
#ifdef __E_T800__
    NotEnoughMemory();;
#else
    e_GeneralError("Not enough memory for ibuf_MeshStruc");
#endif
  if( (dbuf_MeshStruc = e_alloc_d_vect(MESH_STRUC_DBUF_LENGTH)) == NULL)
#ifdef __E_T800__
    NotEnoughMemory();
#else
    e_GeneralError("Not enough memory for dbuf_MeshStruc");
#endif

  /* assemble i&d_buffers to transfer the mesh struct.  */
  ibuf_MeshStruc[0]  = strlen(mesh->Title) + 1;
  ibuf_MeshStruc[1]  = mesh->NMeshPoints;
  ibuf_MeshStruc[2]  = mesh->NNodes;
  ibuf_MeshStruc[3]  = mesh->NElemsLink1;
  ibuf_MeshStruc[4]  = mesh->NElemsLink2;
  ibuf_MeshStruc[5]  = mesh->NElemsLink3;
  ibuf_MeshStruc[6]  = mesh->NElemsLink4;
  ibuf_MeshStruc[7]  = mesh->NElemsLink5;
  ibuf_MeshStruc[8]  = mesh->NElemsTriang1;
  ibuf_MeshStruc[9]  = mesh->NElemsTriang2;
  ibuf_MeshStruc[10] = mesh->NElemsTriang3;
  ibuf_MeshStruc[11] = mesh->NElemsTriang4;
  ibuf_MeshStruc[12] = mesh->NElemsTriang5;
  ibuf_MeshStruc[13] = mesh->NElemsQuad1;
  ibuf_MeshStruc[14] = mesh->NElemsQuad3;
  ibuf_MeshStruc[15] = mesh->NElemsTetrah1;
  ibuf_MeshStruc[16] = mesh->NElemsBlock1;
  ibuf_MeshStruc[17] = mesh->TotalNElems;
  ibuf_MeshStruc[18] = mesh->NBCNodes;
  ibuf_MeshStruc[19] = mesh->NLoadedNodes;
  ibuf_MeshStruc[20] = mesh->NMats;
  ibuf_MeshStruc[21] = mesh->NMdfMats;
  ibuf_MeshStruc[22] = mesh->NCompMatsType1;
  ibuf_MeshStruc[23] = mesh->NTimeSteps;
  ibuf_MeshStruc[24] = mesh->NInternalTimeSteps;
  ibuf_MeshStruc[25] = mesh->NExternalTimeSteps;
  ibuf_MeshStruc[26] = mesh->TotalNodalDOF;

  ibuf_MeshStruc[27] = mesh->NStressPointsLink1;
  ibuf_MeshStruc[28] = mesh->NStressPointsLink2;
  ibuf_MeshStruc[29] = mesh->NStressPointsLink3;
  ibuf_MeshStruc[30] = mesh->NStressPointsLink4;
  ibuf_MeshStruc[31] = mesh->NStressPointsLink5;
  ibuf_MeshStruc[32] = mesh->NStressPointsTriang1;
  ibuf_MeshStruc[33] = mesh->NStressPointsTriang2;
  ibuf_MeshStruc[34] = mesh->NStressPointsTriang3;
  ibuf_MeshStruc[35] = mesh->NStressPointsTriang4;
  ibuf_MeshStruc[36] = mesh->NStressPointsTriang5;
  ibuf_MeshStruc[37] = mesh->NStressPointsQuad1;
  ibuf_MeshStruc[38] = mesh->NStressPointsQuad3;
  ibuf_MeshStruc[39] = mesh->NStressPointsTetrah1;
  ibuf_MeshStruc[40] = mesh->NStressPointsBlock1;

  dbuf_MeshStruc[0]  = mesh->TimeStep;
  dbuf_MeshStruc[1]  = mesh->DampingFactor;
  dbuf_MeshStruc[2]  = mesh->Beta;
  dbuf_MeshStruc[3]  = mesh->Gamma;


  MPI_Send(ibuf_MeshStruc, MESH_STRUC_IBUF_LENGTH, MPI_INT,    
           dest, 101, MPI_COMM_WORLD);
  MPI_Send(dbuf_MeshStruc, MESH_STRUC_DBUF_LENGTH, MPI_DOUBLE, 
           dest, 102, MPI_COMM_WORLD);

  /* send the title  */
  MPI_Send(mesh->Title, ibuf_MeshStruc[0], MPI_CHAR, 
           dest, 103, MPI_COMM_WORLD);

  MPI_Send(mesh->FirstNetIndex, E_NELEM_TYPES, MPI_INT, 
           dest, 104, MPI_COMM_WORLD);
  MPI_Send(mesh->KeywordSet, E_NKEYWORDS, MPI_INT, 
           dest, 105, MPI_COMM_WORLD);
  MPI_Send(mesh->SubStructSet,  E_NSUBSTRUCTS, MPI_INT, 
           dest, 106, MPI_COMM_WORLD);

  /* mesh point co-ordinates   */
  MPI_Send(mesh->MeshPointCoords[0], mesh->NMeshPoints * 3, MPI_DOUBLE, 
           dest, 107, MPI_COMM_WORLD);


  /* element nodes  */
  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
    MPI_Send(mesh->NodesLink1[0], mesh->NElemsLink1 * E_NNPE_LINK1, MPI_INT,
             dest, 108, MPI_COMM_WORLD);

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
    MPI_Send(mesh->NodesLink2[0], mesh->NElemsLink2 * E_NNPE_LINK2, MPI_INT,
             dest, 109, MPI_COMM_WORLD);

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
    MPI_Send(mesh->NodesLink3[0], mesh->NElemsLink3 * E_NNPE_LINK3, MPI_INT, 
             dest, 110, MPI_COMM_WORLD);

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
    MPI_Send(mesh->NodesLink4[0], mesh->NElemsLink4 * E_NNPE_LINK4, MPI_INT, 
             dest, 111, MPI_COMM_WORLD);

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
    MPI_Send(mesh->NodesLink5[0], mesh->NElemsLink5 * E_NNPE_LINK5, MPI_INT, 
             dest, 112, MPI_COMM_WORLD);

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
    MPI_Send(mesh->NodesTriang1[0], mesh->NElemsTriang1 * E_NNPE_TRIANG1, 
             MPI_INT, dest, 113, MPI_COMM_WORLD);

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
    MPI_Send(mesh->NodesTriang2[0], mesh->NElemsTriang2 * E_NNPE_TRIANG2, 
             MPI_INT, dest, 114, MPI_COMM_WORLD);

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
    MPI_Send(mesh->NodesTriang3[0], mesh->NElemsTriang3 * E_NNPE_TRIANG3, 
             MPI_INT, dest, 115, MPI_COMM_WORLD);

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
    MPI_Send(mesh->NodesTriang4[0], mesh->NElemsTriang4 * E_NNPE_TRIANG4, 
             MPI_INT, dest, 116, MPI_COMM_WORLD);

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
    MPI_Send(mesh->NodesTriang5[0], mesh->NElemsTriang5 * E_NNPE_TRIANG5, 
             MPI_INT, dest, 117, MPI_COMM_WORLD);

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
    MPI_Send(mesh->NodesQuad1[0], mesh->NElemsQuad1 * E_NNPE_QUAD1, 
             MPI_INT, dest, 118, MPI_COMM_WORLD);

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
    MPI_Send(mesh->NodesQuad3[0], mesh->NElemsQuad3 * E_NNPE_QUAD3, 
             MPI_INT, dest, 119, MPI_COMM_WORLD);

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
    MPI_Send(mesh->NodesTetrah1[0], mesh->NElemsTetrah1 * E_NNPE_TETRAH1, 
             MPI_INT, dest, 120, MPI_COMM_WORLD);

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
    MPI_Send(mesh->NodesBlock1[0], mesh->NElemsBlock1 * E_NNPE_BLOCK1, 
             MPI_INT, dest, 121, MPI_COMM_WORLD);

  /*  NodalDOF  */
  MPI_Send(mesh->NodalDOF, mesh->NNodes, MPI_INT, dest, 122, MPI_COMM_WORLD);

  /*  Boundary Conditions  */
  if(mesh->KeywordSet[E_KW_NUM_BCS]       && 
     mesh->KeywordSet[E_KW_NUM_NBC_NODES]    )
  {
    MPI_Send(mesh->BCNodes, mesh->NBCNodes, MPI_INT, 
             dest, 130, MPI_COMM_WORLD);

    for(i = 0;i < mesh->NBCNodes;i++)
    {
      MPI_Send(mesh->BCTypes[i], E_DOF_MAX, MPI_INT, 
               dest, 131, MPI_COMM_WORLD);
      MPI_Send(mesh->BCDispls[i], E_DOF_MAX, MPI_DOUBLE, 
               dest, 132, MPI_COMM_WORLD);
      MPI_Send(mesh->BCSpringConsts[i], E_DOF_MAX, MPI_DOUBLE, 
               dest, 133, MPI_COMM_WORLD);
    }
  }

  /*  Loading Conditions  */
  if(mesh->KeywordSet[E_KW_NUM_LOADS]         && 
     mesh->KeywordSet[E_KW_NUM_NLOADED_NODES]    )
  {
    MPI_Send(mesh->LoadedNodes, mesh->NLoadedNodes, MPI_INT, 
             dest, 134, MPI_COMM_WORLD);
    for(i = 0;i < mesh->NLoadedNodes; i++)
      MPI_Send(mesh->Loads[i], E_DOF_MAX, MPI_DOUBLE, 
               dest, 135, MPI_COMM_WORLD);
  }

  /* Material information */
  if(mesh->SubStructSet[E_SS_NUM_MAT_PROPS])
  {
    e_MPI_Send_Mats(mesh,dest);

    if((vect_ind = e_alloc_i_vect(mesh->TotalNElems)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Vect_ind");
#endif

    if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
    {
      for(j = 0;j < mesh->NElemsLink1; j++)
        vect_ind[j] = (mesh->MatsLink1[j]) - (mesh->Props);

      MPI_Send(vect_ind, mesh->NElemsLink1, MPI_INT, 
               dest, 150, MPI_COMM_WORLD);
    }

    if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
    {
      for(j = 0;j < mesh->NElemsLink2; j++)
        vect_ind[j] = (mesh->MatsLink2[j]) - (mesh->Props);

      MPI_Send(vect_ind, mesh->NElemsLink2, MPI_INT, 
               dest, 151, MPI_COMM_WORLD);
    }

    if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
    {
      for(j = 0;j < mesh->NElemsLink3; j++)
        vect_ind[j] = (mesh->MatsLink3[j]) - (mesh->Props);

      MPI_Send(vect_ind, mesh->NElemsLink3, MPI_INT, 
               dest, 152, MPI_COMM_WORLD);
    }

    if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
    {
      for(j = 0;j < mesh->NElemsLink4; j++)
        vect_ind[j] = (mesh->MatsLink4[j]) - (mesh->Props);

      MPI_Send(vect_ind, mesh->NElemsLink4, MPI_INT, 
               dest, 153, MPI_COMM_WORLD);
    }

    if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
    {
      for(j = 0;j < mesh->NElemsLink5; j++)
        vect_ind[j] = (mesh->MatsLink5[j]) - (mesh->Props);

      MPI_Send(vect_ind, mesh->NElemsLink5, MPI_INT, 
               dest, 154, MPI_COMM_WORLD);
    }

    if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
    {
      for(j = 0;j < mesh->NElemsTriang1; j++)
        vect_ind[j] = (mesh->MatsTriang1[j]) - (mesh->Props);

      MPI_Send(vect_ind, mesh->NElemsTriang1, MPI_INT, 
               dest, 155, MPI_COMM_WORLD);
    }

    if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
    {
      for(j = 0;j < mesh->NElemsTriang2;j++)
        vect_ind[j] = (mesh->CompMatsTriang2[j]) - (mesh->CompPropsType1);

      MPI_Send(vect_ind, mesh->NElemsTriang2, MPI_INT, 
               dest, 156, MPI_COMM_WORLD);
    }

    if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
    {
      for(j = 0;j < mesh->NElemsTriang3; j++)
        vect_ind[j] = (mesh->MatsTriang3[j]) - (mesh->Props);

      MPI_Send(vect_ind, mesh->NElemsTriang3, MPI_INT, 
               dest, 157, MPI_COMM_WORLD);
    }

    if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
    {
      for(j = 0;j < mesh->NElemsTriang4; j++)
        vect_ind[j] = (mesh->MatsTriang4[j]) - (mesh->Props);

      MPI_Send(vect_ind, mesh->NElemsTriang4, MPI_INT, 
               dest, 158, MPI_COMM_WORLD);
    }

    if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
    {
      for(j = 0;j < mesh->NElemsTriang5; j++)
        vect_ind[j] = (mesh->MatsTriang5[j]) - (mesh->Props);

      MPI_Send(vect_ind, mesh->NElemsTriang5, MPI_INT, 
               dest, 159, MPI_COMM_WORLD);
    }

    if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
    {
      for(j = 0;j < mesh->NElemsQuad1; j++)
        vect_ind[j] = (mesh->MatsQuad1[j]) - (mesh->Props);

      MPI_Send(vect_ind, mesh->NElemsQuad1, MPI_INT, 
               dest, 160, MPI_COMM_WORLD);
    }

    if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
    {
      for(j = 0;j < mesh->NElemsQuad3;j++)
        vect_ind[j] = (mesh->CompMatsQuad3[j]) - (mesh->CompPropsType1);

      MPI_Send(vect_ind, mesh->NElemsQuad3, MPI_INT,
               dest, 161, MPI_COMM_WORLD);
    }

    if(mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
    {
      for(j = 0;j < mesh->NElemsTetrah1; j++)
        vect_ind[j] = (mesh->MatsTetrah1[j]) - (mesh->Props);

      MPI_Send(vect_ind, mesh->NElemsTetrah1, MPI_INT, 
               dest, 162, MPI_COMM_WORLD);
    }

    if(mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
    {
      for(j = 0;j < mesh->NElemsBlock1; j++)
        vect_ind[j] = (mesh->MatsBlock1[j]) - (mesh->Props);

      MPI_Send(vect_ind, mesh->NElemsBlock1, MPI_INT, 
              dest, 163, MPI_COMM_WORLD);
    }
    e_free(vect_ind);
  }

  /* GNIs  */
  MPI_Send(mesh->GNIs, mesh->NNodes, MPI_INT, dest, 122, MPI_COMM_WORLD);

  /* GlobalElementIndices  */
  MPI_Send(mesh->GEIs, mesh->TotalNElems, MPI_INT, dest, 123, MPI_COMM_WORLD);

  e_free(ibuf_MeshStruc);
  e_free(dbuf_MeshStruc);

} /* end of function e_MPI_Send_MeshData()  */



void e_MPI_Recv_MeshData(MESH **mesh, int source)
{
  register    i,j;
  int        *ibuf_MeshStruc;
  double     *dbuf_MeshStruc;
  int        *vect_ind;
  MPI_Status  status;

  if( (ibuf_MeshStruc = e_alloc_i_vect(MESH_STRUC_IBUF_LENGTH)) == NULL)
#ifdef __E_T800__
    NotEnoughMemory();
#else
    e_GeneralError("Not enough memory for ibuf_MeshStruc");
#endif
  if( (dbuf_MeshStruc = e_alloc_d_vect(MESH_STRUC_DBUF_LENGTH)) == NULL)
#ifdef __E_T800__
    NotEnoughMemory();
#else
    e_GeneralError("Not enough memory for dbuf_MeshStruc");
#endif

  MPI_Recv(ibuf_MeshStruc, MESH_STRUC_IBUF_LENGTH, MPI_INT,    
           source, 101, MPI_COMM_WORLD, &status);
  MPI_Recv(dbuf_MeshStruc, MESH_STRUC_DBUF_LENGTH, MPI_DOUBLE, 
           source, 102, MPI_COMM_WORLD, &status);

      /* mesh structure  */
  if((*mesh = calloc(1, sizeof(MESH))) == NULL)
#ifdef __E_T800__
    NotEnoughMemory();
#else
    e_GeneralError("Not enough memory for mesh");
#endif

  (*mesh)->NMeshPoints        = ibuf_MeshStruc[1];
  (*mesh)->NNodes             = ibuf_MeshStruc[2];
  (*mesh)->NElemsLink1        = ibuf_MeshStruc[3];
  (*mesh)->NElemsLink2        = ibuf_MeshStruc[4];
  (*mesh)->NElemsLink3        = ibuf_MeshStruc[5];
  (*mesh)->NElemsLink4        = ibuf_MeshStruc[6];
  (*mesh)->NElemsLink5        = ibuf_MeshStruc[7];
  (*mesh)->NElemsTriang1      = ibuf_MeshStruc[8];
  (*mesh)->NElemsTriang2      = ibuf_MeshStruc[9];
  (*mesh)->NElemsTriang3      = ibuf_MeshStruc[10];
  (*mesh)->NElemsTriang4      = ibuf_MeshStruc[11];
  (*mesh)->NElemsTriang5      = ibuf_MeshStruc[12];
  (*mesh)->NElemsQuad1        = ibuf_MeshStruc[13];
  (*mesh)->NElemsQuad3        = ibuf_MeshStruc[14];
  (*mesh)->NElemsTetrah1      = ibuf_MeshStruc[15];
  (*mesh)->NElemsBlock1       = ibuf_MeshStruc[16];
  (*mesh)->TotalNElems        = ibuf_MeshStruc[17];
  (*mesh)->NBCNodes           = ibuf_MeshStruc[18];
  (*mesh)->NLoadedNodes       = ibuf_MeshStruc[19];
  (*mesh)->NMats              = ibuf_MeshStruc[20];
  (*mesh)->NMdfMats           = ibuf_MeshStruc[21];
  (*mesh)->NCompMatsType1     = ibuf_MeshStruc[22];
  (*mesh)->NTimeSteps         = ibuf_MeshStruc[23];
  (*mesh)->NInternalTimeSteps = ibuf_MeshStruc[24];
  (*mesh)->NExternalTimeSteps = ibuf_MeshStruc[25];
  (*mesh)->TotalNodalDOF      = ibuf_MeshStruc[26];

  (*mesh)->NStressPointsLink1   = ibuf_MeshStruc[27];
  (*mesh)->NStressPointsLink2   = ibuf_MeshStruc[28];
  (*mesh)->NStressPointsLink3   = ibuf_MeshStruc[29];
  (*mesh)->NStressPointsLink4   = ibuf_MeshStruc[30];
  (*mesh)->NStressPointsLink5   = ibuf_MeshStruc[31];
  (*mesh)->NStressPointsTriang1 = ibuf_MeshStruc[32];
  (*mesh)->NStressPointsTriang2 = ibuf_MeshStruc[33];
  (*mesh)->NStressPointsTriang3 = ibuf_MeshStruc[34];
  (*mesh)->NStressPointsTriang4 = ibuf_MeshStruc[35];
  (*mesh)->NStressPointsTriang5 = ibuf_MeshStruc[36];
  (*mesh)->NStressPointsQuad1   = ibuf_MeshStruc[37];
  (*mesh)->NStressPointsQuad3   = ibuf_MeshStruc[38];
  (*mesh)->NStressPointsTetrah1 = ibuf_MeshStruc[39];
  (*mesh)->NStressPointsBlock1  = ibuf_MeshStruc[40];

  (*mesh)->TimeStep           = dbuf_MeshStruc[0];
  (*mesh)->DampingFactor      = dbuf_MeshStruc[1];
  (*mesh)->Beta               = dbuf_MeshStruc[2];
  (*mesh)->Gamma              = dbuf_MeshStruc[3];


  /* title string  */
  if(((*mesh)->Title = e_alloc_c_vect(ibuf_MeshStruc[0])) == NULL)
#ifdef __E_T800__
    NotEnoughMemory();
#else
    e_GeneralError("Not enough memory for mesh Title");
#endif
  MPI_Recv((*mesh)->Title, ibuf_MeshStruc[E_KW_NUM_TITLE], MPI_CHAR, 
           source, 103, MPI_COMM_WORLD, &status);

  MPI_Recv((*mesh)->FirstNetIndex, E_NELEM_TYPES, MPI_INT, 
           source, 104, MPI_COMM_WORLD, &status);
  MPI_Recv((*mesh)->KeywordSet,    E_NKEYWORDS,   MPI_INT, 
           source, 105, MPI_COMM_WORLD, &status);
  MPI_Recv((*mesh)->SubStructSet,  E_NSUBSTRUCTS, MPI_INT, 
           source, 106, MPI_COMM_WORLD, &status);

  /* mesh point co-ordinates   */
  if(((*mesh)->MeshPointCoords = 
        e_alloc_cont_d_matr((*mesh)->NMeshPoints, 3)) == NULL)
#ifdef __E_T800__
    NotEnoughMemory();
#else
    e_GeneralError("Not enough memory for mesh point co-ords");
#endif
  MPI_Recv((*mesh)->MeshPointCoords[0], (*mesh)->NMeshPoints * 3, 
           MPI_DOUBLE, source, 107, MPI_COMM_WORLD, &status);


  /* element nodes  */
  if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_LINK1]) 
  {
    if(((*mesh)->NodesLink1 = 
          e_alloc_cont_i_matr((*mesh)->NElemsLink1, E_NNPE_LINK1)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Link1 vertices");
#endif
    MPI_Recv((*mesh)->NodesLink1[0], (*mesh)->NElemsLink1 * E_NNPE_LINK1, 
             MPI_INT, source, 108, MPI_COMM_WORLD, &status);
  }

  if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_LINK2]) 
  {
    if (((*mesh)->NodesLink2 = 
           e_alloc_cont_i_matr((*mesh)->NElemsLink2, E_NNPE_LINK2)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Link2 vertices");
#endif
    MPI_Recv((*mesh)->NodesLink2[0], (*mesh)->NElemsLink2 * E_NNPE_LINK2, 
             MPI_INT, source, 109, MPI_COMM_WORLD, &status);
  }

  if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_LINK3]) 
  {
    if(((*mesh)->NodesLink3 = 
          e_alloc_cont_i_matr((*mesh)->NElemsLink3, E_NNPE_LINK3)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Link3 vertices");
#endif
    MPI_Recv((*mesh)->NodesLink3[0], (*mesh)->NElemsLink3 * E_NNPE_LINK3, 
             MPI_INT, source, 110, MPI_COMM_WORLD, &status);
  }

  if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_LINK4]) 
  {
    if(((*mesh)->NodesLink4 = 
          e_alloc_cont_i_matr((*mesh)->NElemsLink4, E_NNPE_LINK4)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Link4 vertices");
#endif
    MPI_Recv((*mesh)->NodesLink4[0], (*mesh)->NElemsLink4 * E_NNPE_LINK4,
             MPI_INT, source, 111, MPI_COMM_WORLD, &status);
  }

  if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_LINK5]) 
  {
    if(((*mesh)->NodesLink5 = 
          e_alloc_cont_i_matr((*mesh)->NElemsLink5, E_NNPE_LINK5)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Link5 vertices");
#endif
    MPI_Recv((*mesh)->NodesLink5[0], (*mesh)->NElemsLink5 * E_NNPE_LINK5, 
             MPI_INT, source, 112, MPI_COMM_WORLD, &status);
  }

  if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG1]) 
  {
    if(((*mesh)->NodesTriang1 = 
          e_alloc_cont_i_matr((*mesh)->NElemsTriang1, E_NNPE_TRIANG1)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Triang1 vertices");
#endif
    MPI_Recv((*mesh)->NodesTriang1[0], (*mesh)->NElemsTriang1 * E_NNPE_TRIANG1,
             MPI_INT, source, 113, MPI_COMM_WORLD, &status);
  }

  if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG2]) 
  {
    if(((*mesh)->NodesTriang2 = 
          e_alloc_cont_i_matr((*mesh)->NElemsTriang2, E_NNPE_TRIANG2)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Triang2 vertices");
#endif
    MPI_Recv((*mesh)->NodesTriang2[0], (*mesh)->NElemsTriang2 * E_NNPE_TRIANG2,
             MPI_INT, source, 114, MPI_COMM_WORLD, &status);
  }

  if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG3]) 
  {
    if(((*mesh)->NodesTriang3 = 
          e_alloc_cont_i_matr((*mesh)->NElemsTriang3, E_NNPE_TRIANG3)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Triang3 vertices");
#endif
    MPI_Recv((*mesh)->NodesTriang3[0], (*mesh)->NElemsTriang3 * E_NNPE_TRIANG3,
             MPI_INT, source, 115, MPI_COMM_WORLD, &status);
  }

  if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG4]) 
  {
    if(((*mesh)->NodesTriang4 = 
          e_alloc_cont_i_matr((*mesh)->NElemsTriang4, E_NNPE_TRIANG4)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Triang4 vertices");
#endif
    MPI_Recv((*mesh)->NodesTriang4[0], (*mesh)->NElemsTriang4 * E_NNPE_TRIANG4,
             MPI_INT, source, 116, MPI_COMM_WORLD, &status);
  }

  if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG5]) 
  {
    if(((*mesh)->NodesTriang5 = 
          e_alloc_cont_i_matr((*mesh)->NElemsTriang5, E_NNPE_TRIANG5)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Triang5 vertices");
#endif
    MPI_Recv((*mesh)->NodesTriang5[0], (*mesh)->NElemsTriang5 * E_NNPE_TRIANG5,
             MPI_INT, source, 117, MPI_COMM_WORLD, &status);
  }

  if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_QUAD1]) 
  {
    if(((*mesh)->NodesQuad1 = 
          e_alloc_cont_i_matr((*mesh)->NElemsQuad1, E_NNPE_QUAD1)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Quad1 vertices");
#endif
    MPI_Recv((*mesh)->NodesQuad1[0], (*mesh)->NElemsQuad1 * E_NNPE_QUAD1,
             MPI_INT, source, 118, MPI_COMM_WORLD, &status);
  }

  if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_QUAD3]) 
  {
    if(((*mesh)->NodesQuad3 = 
          e_alloc_cont_i_matr((*mesh)->NElemsQuad3, E_NNPE_QUAD3)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Quad3 vertices");
#endif
    MPI_Recv((*mesh)->NodesQuad3[0], (*mesh)->NElemsQuad3 * E_NNPE_QUAD3,
             MPI_INT, source, 119, MPI_COMM_WORLD, &status);
  }

  if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TETRAH1]) 
  {
    if(((*mesh)->NodesTetrah1 = 
          e_alloc_cont_i_matr((*mesh)->NElemsTetrah1, E_NNPE_TETRAH1)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Tetrah1 vertices");
#endif
    MPI_Recv((*mesh)->NodesTetrah1[0], (*mesh)->NElemsTetrah1 * E_NNPE_TETRAH1,
             MPI_INT, source, 120, MPI_COMM_WORLD, &status);
  }

  if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_BLOCK1]) 
  {
    if(((*mesh)->NodesBlock1 = 
          e_alloc_cont_i_matr((*mesh)->NElemsBlock1, E_NNPE_BLOCK1)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Block1 vertices");
#endif
    MPI_Recv((*mesh)->NodesBlock1[0], (*mesh)->NElemsBlock1 * E_NNPE_BLOCK1,
             MPI_INT, source, 121, MPI_COMM_WORLD, &status);
  }

  /*  NodalDOF  */
  if(((*mesh)->NodalDOF = e_alloc_i_vect((*mesh)->NNodes)) == NULL)
#ifdef __E_T800__
    NotEnoughMemory();
#else
    e_GeneralError("Not enough memory for NodalDOFs");
#endif
  MPI_Recv((*mesh)->NodalDOF, (*mesh)->NNodes, MPI_INT, 
           source, 122, MPI_COMM_WORLD, &status);

  /*  Boundary Conditions  */
  if((*mesh)->KeywordSet[E_KW_NUM_BCS]       && 
     (*mesh)->KeywordSet[E_KW_NUM_NBC_NODES]    )
  {
    if(((*mesh)->BCNodes = e_alloc_i_vect((*mesh)->NBCNodes)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for BCNodes");
#endif
    MPI_Recv((*mesh)->BCNodes, (*mesh)->NBCNodes, MPI_INT, 
             source, 130, MPI_COMM_WORLD, &status);

    if(((*mesh)->BCTypes = e_init_frag_i_matr((*mesh)->NBCNodes)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for BCTypes");
#endif
    if(((*mesh)->BCDispls = e_init_frag_d_matr((*mesh)->NBCNodes)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for BCDispls");
#endif
    if(((*mesh)->BCSpringConsts = e_init_frag_d_matr((*mesh)->NBCNodes)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for BCSpringConsts");
#endif

    for(i = 0;i < (*mesh)->NBCNodes;i++)
    {
      if(((*mesh)->BCTypes[i] = e_alloc_i_vect(E_DOF_MAX)) == NULL)
#ifdef __E_T800__
        NotEnoughMemory();
#else
        e_GeneralError("Not enough memory for BCTypes");
#endif
      MPI_Recv((*mesh)->BCTypes[i], E_DOF_MAX, MPI_INT, 
               source, 131, MPI_COMM_WORLD, &status);
      if(((*mesh)->BCDispls[i] = e_alloc_d_vect(E_DOF_MAX)) == NULL)
#ifdef __E_T800__
        NotEnoughMemory();
#else
        e_GeneralError("Not enough memory for BCDispls");
#endif
      MPI_Recv((*mesh)->BCDispls[i], E_DOF_MAX, MPI_DOUBLE, 
               source, 132, MPI_COMM_WORLD, &status);
      if(((*mesh)->BCSpringConsts[i] = e_alloc_d_vect(E_DOF_MAX)) == NULL)
#ifdef __E_T800__
        NotEnoughMemory();
#else
        e_GeneralError("Not enough memory for BCSpringConsts");
#endif
      MPI_Recv((*mesh)->BCSpringConsts[i], E_DOF_MAX, MPI_DOUBLE, 
               source, 133, MPI_COMM_WORLD, &status);
    }
  }


  /*  Loading Conditions  */
  if((*mesh)->KeywordSet[E_KW_NUM_LOADS]         && 
     (*mesh)->KeywordSet[E_KW_NUM_NLOADED_NODES]    )
  {
    if(((*mesh)->LoadedNodes = e_alloc_i_vect((*mesh)->NLoadedNodes)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for LoadedNodes");
#endif

    MPI_Recv((*mesh)->LoadedNodes, (*mesh)->NLoadedNodes, MPI_INT, 
             source, 134, MPI_COMM_WORLD, &status);

    if(((*mesh)->Loads = e_init_frag_d_matr((*mesh)->NLoadedNodes)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Loads");
#endif

    for(i = 0;i < (*mesh)->NLoadedNodes;i++)
    {
      if(((*mesh)->Loads[i] = e_alloc_d_vect(E_DOF_MAX)) == NULL)
#ifdef __E_T800__
        NotEnoughMemory();
#else
        e_GeneralError("Not enough memory for Loads");
#endif
      MPI_Recv((*mesh)->Loads[i], E_DOF_MAX, MPI_DOUBLE, 
               source, 135, MPI_COMM_WORLD, &status);
    }
  }

  /* Material information */
  if((*mesh)->SubStructSet[E_SS_NUM_MAT_PROPS])
  {
    e_MPI_Recv_Mats(*mesh,source);

    if((vect_ind = e_alloc_i_vect((*mesh)->TotalNElems)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Vect_ind");
#endif

    if(((*mesh)->Mats = e_alloc_vp_vect((*mesh)->TotalNElems)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Mats");
#endif

    if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_LINK1])
    {
      (*mesh)->MatsLink1 = (MAT **)((*mesh)->Mats +
                           (*mesh)->FirstNetIndex[E_ELEM_TYPE_LINK1] - 1);

      MPI_Recv(vect_ind, (*mesh)->NElemsLink1, MPI_INT, 
               source, 150, MPI_COMM_WORLD, &status);
      for(j = 0;j < (*mesh)->NElemsLink1;j++)
        (*mesh)->MatsLink1[j] = (*mesh)->Props + vect_ind[j];
    }
    else
      (*mesh)->MatsLink1 = NULL;

    if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_LINK2])
    {
      (*mesh)->MatsLink2 = (MAT **)((*mesh)->Mats + 
                           (*mesh)->FirstNetIndex[E_ELEM_TYPE_LINK2] - 1);

      MPI_Recv(vect_ind, (*mesh)->NElemsLink2, MPI_INT, 
               source, 151, MPI_COMM_WORLD, &status);
      for(j = 0;j < (*mesh)->NElemsLink2;j++)
        (*mesh)->MatsLink2[j] = (*mesh)->Props + vect_ind[j];
    }
    else
      (*mesh)->MatsLink2 = NULL;

    if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_LINK3])
    {
      (*mesh)->MatsLink3 = (MAT **)((*mesh)->Mats +
                           (*mesh)->FirstNetIndex[E_ELEM_TYPE_LINK3] - 1);

      MPI_Recv(vect_ind, (*mesh)->NElemsLink3, MPI_INT, 
               source, 152, MPI_COMM_WORLD, &status);
      for(j = 0;j < (*mesh)->NElemsLink3;j++)
        (*mesh)->MatsLink3[j] = (*mesh)->Props + vect_ind[j];
    }
    else
      (*mesh)->MatsLink3 = NULL;

    if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_LINK4])
    {
      (*mesh)->MatsLink4 = (MAT **)((*mesh)->Mats + 
                           (*mesh)->FirstNetIndex[E_ELEM_TYPE_LINK4] - 1);

      MPI_Recv(vect_ind, (*mesh)->NElemsLink4, MPI_INT, 
               source, 153, MPI_COMM_WORLD, &status);
      for(j = 0;j < (*mesh)->NElemsLink4;j++)
        (*mesh)->MatsLink4[j] = (*mesh)->Props + vect_ind[j];
    }
    else
      (*mesh)->MatsLink4 = NULL;

    if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_LINK5])
    {
      (*mesh)->MatsLink5 = (MAT **)((*mesh)->Mats + 
                           (*mesh)->FirstNetIndex[E_ELEM_TYPE_LINK5] - 1);

      MPI_Recv(vect_ind, (*mesh)->NElemsLink5, MPI_INT, 
               source, 154, MPI_COMM_WORLD, &status);
      for(j = 0;j < (*mesh)->NElemsLink5;j++)
        (*mesh)->MatsLink5[j] = (*mesh)->Props + vect_ind[j];
    }
    else
      (*mesh)->MatsLink5 = NULL;


    if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
    {
      (*mesh)->MatsTriang1 = (MAT **)((*mesh)->Mats +
                             (*mesh)->FirstNetIndex[E_ELEM_TYPE_TRIANG1] - 1);

      MPI_Recv(vect_ind, (*mesh)->NElemsTriang1, MPI_INT, 
               source, 155, MPI_COMM_WORLD, &status);
      for(j = 0;j < (*mesh)->NElemsTriang1;j++)
        (*mesh)->MatsTriang1[j] = (*mesh)->Props + vect_ind[j];
    }
    else
      (*mesh)->MatsTriang1 = NULL;

    if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
    {
      (*mesh)->CompMatsTriang2 = (COMPMAT1 **)((*mesh)->Mats + 
                                 (*mesh)->FirstNetIndex[E_ELEM_TYPE_TRIANG2] -
                                 1 );

      MPI_Recv(vect_ind, (*mesh)->NElemsTriang2, MPI_INT, 
               source, 156, MPI_COMM_WORLD, &status);
      for(j = 0;j < (*mesh)->NElemsTriang2;j++)
        (*mesh)->CompMatsTriang2[j] = (*mesh)->CompPropsType1 + vect_ind[j];
    }
    else
      (*mesh)->CompMatsTriang2 = NULL;

    if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
    {
      (*mesh)->MatsTriang3 = (MAT **)((*mesh)->Mats + 
                             (*mesh)->FirstNetIndex[E_ELEM_TYPE_TRIANG3] - 1);

      MPI_Recv(vect_ind, (*mesh)->NElemsTriang3, MPI_INT, 
               source, 157, MPI_COMM_WORLD, &status);
      for(j = 0;j < (*mesh)->NElemsTriang3;j++)
        (*mesh)->MatsTriang3[j] = (*mesh)->Props + vect_ind[j];
    }
    else
      (*mesh)->MatsTriang3 = NULL;

    if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
    {
      (*mesh)->MatsTriang4 = (MAT **)((*mesh)->Mats + 
                             (*mesh)->FirstNetIndex[E_ELEM_TYPE_TRIANG4] - 1);

      MPI_Recv(vect_ind, (*mesh)->NElemsTriang4, MPI_INT, 
               source, 158, MPI_COMM_WORLD, &status);
      for(j = 0;j < (*mesh)->NElemsTriang4;j++)
        (*mesh)->MatsTriang4[j] = (*mesh)->Props + vect_ind[j];
    }
    else
      (*mesh)->MatsTriang4 = NULL;

    if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
    {
      (*mesh)->MatsTriang5 = (MAT **)((*mesh)->Mats + 
                             (*mesh)->FirstNetIndex[E_ELEM_TYPE_TRIANG5] - 1);

      MPI_Recv(vect_ind, (*mesh)->NElemsTriang5, MPI_INT, 
               source, 159, MPI_COMM_WORLD, &status);
      for(j = 0;j < (*mesh)->NElemsTriang5;j++)
        (*mesh)->MatsTriang5[j] = (*mesh)->Props + vect_ind[j];
    }
    else
      (*mesh)->MatsTriang5 = NULL;

    if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
    {
      (*mesh)->MatsQuad1 = (MAT **)((*mesh)->Mats + 
                           (*mesh)->FirstNetIndex[E_ELEM_TYPE_QUAD1] - 1);

      MPI_Recv(vect_ind, (*mesh)->NElemsQuad1, MPI_INT, 
               source, 160, MPI_COMM_WORLD, &status);
      for(j = 0;j < (*mesh)->NElemsQuad1;j++)
        (*mesh)->MatsQuad1[j] = (*mesh)->Props + vect_ind[j];
    }
    else
      (*mesh)->MatsQuad1 = NULL;

    if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
    {
      (*mesh)->CompMatsQuad3 = (COMPMAT1 **)((*mesh)->Mats + 
                               (*mesh)->FirstNetIndex[E_ELEM_TYPE_QUAD3] - 1);

      MPI_Recv(vect_ind, (*mesh)->NElemsQuad3, MPI_INT, 
               source, 161, MPI_COMM_WORLD, &status);
      for(j = 0;j < (*mesh)->NElemsQuad3;j++)
        (*mesh)->CompMatsQuad3[j] = (*mesh)->CompPropsType1 + vect_ind[j];
    }
    else
      (*mesh)->CompMatsQuad3 = NULL;

    if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
    {
      (*mesh)->MatsTetrah1 = (MAT **)((*mesh)->Mats +
                             (*mesh)->FirstNetIndex[E_ELEM_TYPE_TETRAH1] - 1);

      MPI_Recv(vect_ind, (*mesh)->NElemsTetrah1, MPI_INT, 
               source, 162, MPI_COMM_WORLD, &status);
      for(j = 0;j < (*mesh)->NElemsTetrah1;j++)
        (*mesh)->MatsTetrah1[j] = (*mesh)->Props + vect_ind[j];
    }
    else
      (*mesh)->MatsTetrah1 = NULL;

    if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
    {
      (*mesh)->MatsBlock1 = (MAT **)((*mesh)->Mats +
                            (*mesh)->FirstNetIndex[E_ELEM_TYPE_BLOCK1] - 1);

      MPI_Recv(vect_ind, (*mesh)->NElemsBlock1, MPI_INT, 
               source, 163, MPI_COMM_WORLD, &status);
      for(j = 0;j < (*mesh)->NElemsBlock1;j++)
        (*mesh)->MatsBlock1[j] = (*mesh)->Props + vect_ind[j];
    }
    else
      (*mesh)->MatsBlock1 = NULL;

    e_free(vect_ind);
  }


  /* GNIs  */
  if (((*mesh)->GNIs = e_alloc_i_vect((*mesh)->NNodes)) == NULL)
#ifdef __E_T800__
    NotEnoughMemory();
#else
    e_GeneralError("Not enough memory for GNIs");
#endif
  MPI_Recv((*mesh)->GNIs, (*mesh)->NNodes, MPI_INT, 
           source, 122, MPI_COMM_WORLD, &status);


  /* GlobalElementIndices  */
  if (((*mesh)->GEIs = e_alloc_i_vect((*mesh)->TotalNElems)) == NULL)
#ifdef __E_T800__
    NotEnoughMemory();
#else
    e_GeneralError("Not enough memory for GlobalElementIndices");
#endif
  MPI_Recv((*mesh)->GEIs, (*mesh)->TotalNElems, MPI_INT, 
           source, 123, MPI_COMM_WORLD, &status);


  if ((*mesh)->KeywordSet[E_KW_NUM_NELEMS_LINK1])
    (*mesh)->GEIsLink1 = (*mesh)->GEIs + 
                         (*mesh)->FirstNetIndex[E_ELEM_TYPE_LINK1] - 1;
  else
    (*mesh)->GEIsLink1 = NULL;

  if ((*mesh)->KeywordSet[E_KW_NUM_NELEMS_LINK2])
    (*mesh)->GEIsLink2 = (*mesh)->GEIs + 
                         (*mesh)->FirstNetIndex[E_ELEM_TYPE_LINK2] - 1;
  else
    (*mesh)->GEIsLink2 = NULL;

  if ((*mesh)->KeywordSet[E_KW_NUM_NELEMS_LINK3])
    (*mesh)->GEIsLink3 = (*mesh)->GEIs + 
                         (*mesh)->FirstNetIndex[E_ELEM_TYPE_LINK3] - 1;
  else
    (*mesh)->GEIsLink3 = NULL;

  if ((*mesh)->KeywordSet[E_KW_NUM_NELEMS_LINK4])
    (*mesh)->GEIsLink4 = (*mesh)->GEIs + 
                         (*mesh)->FirstNetIndex[E_ELEM_TYPE_LINK4] - 1;
  else
    (*mesh)->GEIsLink4 = NULL;

  if ((*mesh)->KeywordSet[E_KW_NUM_NELEMS_LINK5])
    (*mesh)->GEIsLink5 = (*mesh)->GEIs + 
                         (*mesh)->FirstNetIndex[E_ELEM_TYPE_LINK5] - 1;
  else
    (*mesh)->GEIsLink5 = NULL;

  if ((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
    (*mesh)->GEIsTriang1 = (*mesh)->GEIs + 
                           (*mesh)->FirstNetIndex[E_ELEM_TYPE_TRIANG1] - 1;
  else
    (*mesh)->GEIsTriang1 = NULL;

  if ((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
    (*mesh)->GEIsTriang2 = (*mesh)->GEIs + 
                           (*mesh)->FirstNetIndex[E_ELEM_TYPE_TRIANG2] - 1;
  else
    (*mesh)->GEIsTriang2 = NULL;

  if ((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
    (*mesh)->GEIsTriang3 = (*mesh)->GEIs + 
                           (*mesh)->FirstNetIndex[E_ELEM_TYPE_TRIANG3] - 1;
  else
    (*mesh)->GEIsTriang3 = NULL;

  if ((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
    (*mesh)->GEIsTriang4 = (*mesh)->GEIs + 
                           (*mesh)->FirstNetIndex[E_ELEM_TYPE_TRIANG4] - 1;
  else
    (*mesh)->GEIsTriang4 = NULL;

  if ((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
    (*mesh)->GEIsTriang5 = (*mesh)->GEIs + 
                           (*mesh)->FirstNetIndex[E_ELEM_TYPE_TRIANG5] - 1;
  else
    (*mesh)->GEIsTriang5 = NULL;

  if ((*mesh)->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
    (*mesh)->GEIsQuad1 = (*mesh)->GEIs + 
                         (*mesh)->FirstNetIndex[E_ELEM_TYPE_QUAD1] - 1;
  else
    (*mesh)->GEIsQuad1 = NULL;

  if ((*mesh)->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
    (*mesh)->GEIsQuad3 = (*mesh)->GEIs + 
                         (*mesh)->FirstNetIndex[E_ELEM_TYPE_QUAD3] - 1;
  else
    (*mesh)->GEIsQuad3 = NULL;

  if ((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
    (*mesh)->GEIsTetrah1 = (*mesh)->GEIs + 
                           (*mesh)->FirstNetIndex[E_ELEM_TYPE_TETRAH1] - 1;
  else
    (*mesh)->GEIsTetrah1 = NULL;

  if ((*mesh)->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
    (*mesh)->GEIsBlock1 = (*mesh)->GEIs + 
                          (*mesh)->FirstNetIndex[E_ELEM_TYPE_BLOCK1] - 1;
  else
    (*mesh)->GEIsBlock1 = NULL;

  e_free(ibuf_MeshStruc);
  e_free(dbuf_MeshStruc);

}  /* end of function e_MPI_Recv_MeshData()  */


#define No_MESHPARAMSET 2

void  e_MPI_Send_MeshParams(MESH *mesh, int dest)
{
  int  ibuf[No_MESHPARAMSET];

  if(mesh->SubStructSet[E_SS_NUM_MESHPARAM])
  {
    ibuf[0] = mesh->MeshParam.NodalMeshParamsSet;
    ibuf[1] = mesh->MeshParam.ElemMeshParamsSet;
    MPI_Send(ibuf, No_MESHPARAMSET, MPI_INT, dest, 401, MPI_COMM_WORLD);

    if(mesh->MeshParam.ElemMeshParamsSet)
      MPI_Send(mesh->MeshParam.ElemMeshParams, mesh->TotalNElems, 
               MPI_DOUBLE, dest, 402, MPI_COMM_WORLD);
    if(mesh->MeshParam.NodalMeshParamsSet)
      MPI_Send(mesh->MeshParam.NodalMeshParams, mesh->NMeshPoints, 
               MPI_DOUBLE, dest, 403, MPI_COMM_WORLD);
  }
} /* end of function e_MPI_Send_MeshParams()  */


void  e_MPI_Recv_MeshParams(MESH *mesh, int source)
{
  MPI_Status      status;
  int             ibuf[No_MESHPARAMSET];

  MPI_Recv(ibuf, No_MESHPARAMSET, MPI_INT, source, 401, 
           MPI_COMM_WORLD, &status);
  mesh->MeshParam.NodalMeshParamsSet  = ibuf[0];
  mesh->MeshParam.ElemMeshParamsSet   = ibuf[1];

  if(mesh->MeshParam.ElemMeshParamsSet)
  {
    if((mesh->MeshParam.ElemMeshParams = 
          e_alloc_d_vect(mesh->TotalNElems)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for ElemMeshParams");
#endif
    MPI_Recv(mesh->MeshParam.ElemMeshParams, mesh->TotalNElems, 
             MPI_DOUBLE, source, 402, MPI_COMM_WORLD, &status);

    if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
      mesh->MeshParam.ElemMeshParamsLink1 = mesh->MeshParam.ElemMeshParams +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_LINK1] - 1;
    else
      mesh->MeshParam.ElemMeshParamsLink1 = NULL;

    if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
      mesh->MeshParam.ElemMeshParamsLink2 = mesh->MeshParam.ElemMeshParams +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_LINK2] - 1;
    else
      mesh->MeshParam.ElemMeshParamsLink2 = NULL;

    if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
      mesh->MeshParam.ElemMeshParamsLink3 = mesh->MeshParam.ElemMeshParams +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_LINK3] - 1;
    else
      mesh->MeshParam.ElemMeshParamsLink3 = NULL;

    if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
      mesh->MeshParam.ElemMeshParamsLink4 = mesh->MeshParam.ElemMeshParams +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_LINK4] - 1;
    else
      mesh->MeshParam.ElemMeshParamsLink4 = NULL;

    if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
      mesh->MeshParam.ElemMeshParamsLink5 = mesh->MeshParam.ElemMeshParams +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_LINK5] - 1;
    else
      mesh->MeshParam.ElemMeshParamsLink5 = NULL;

    if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
      mesh->MeshParam.ElemMeshParamsTriang1 = mesh->MeshParam.ElemMeshParams +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] - 1;
    else
      mesh->MeshParam.ElemMeshParamsTriang1 = NULL;

    if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
      mesh->MeshParam.ElemMeshParamsTriang2 = mesh->MeshParam.ElemMeshParams +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2] - 1;
    else
      mesh->MeshParam.ElemMeshParamsTriang2 = NULL;

    if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
      mesh->MeshParam.ElemMeshParamsTriang3 = mesh->MeshParam.ElemMeshParams +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] - 1;
    else
      mesh->MeshParam.ElemMeshParamsTriang3 = NULL;

    if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
      mesh->MeshParam.ElemMeshParamsTriang4 = mesh->MeshParam.ElemMeshParams +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] - 1;
    else
      mesh->MeshParam.ElemMeshParamsTriang4 = NULL;

    if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
      mesh->MeshParam.ElemMeshParamsTriang5 = mesh->MeshParam.ElemMeshParams +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] - 1;
    else
      mesh->MeshParam.ElemMeshParamsTriang5 = NULL;

    if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
      mesh->MeshParam.ElemMeshParamsQuad1 = mesh->MeshParam.ElemMeshParams +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1] - 1;
    else
      mesh->MeshParam.ElemMeshParamsQuad1 = NULL;

    if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
      mesh->MeshParam.ElemMeshParamsQuad3 = mesh->MeshParam.ElemMeshParams +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_QUAD3] - 1;
    else
      mesh->MeshParam.ElemMeshParamsQuad3 = NULL;

    if (mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
      mesh->MeshParam.ElemMeshParamsTetrah1 = mesh->MeshParam.ElemMeshParams +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1] - 1;
    else
      mesh->MeshParam.ElemMeshParamsTetrah1 = NULL;

    if (mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
      mesh->MeshParam.ElemMeshParamsBlock1 = mesh->MeshParam.ElemMeshParams +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1] - 1;
    else
      mesh->MeshParam.ElemMeshParamsBlock1 = NULL;
  }

  if(mesh->MeshParam.NodalMeshParamsSet)
  {
    if((mesh->MeshParam.NodalMeshParams = 
          e_alloc_d_vect(mesh->NMeshPoints)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for NodalMeshParams");
#endif

    MPI_Recv(mesh->MeshParam.NodalMeshParams, mesh->NMeshPoints, 
             MPI_DOUBLE, source, 403, MPI_COMM_WORLD, &status);
  }

  mesh->SubStructSet[E_SS_NUM_MESHPARAM] = TRUE;

} /* end of function e_MPI_Recv_MeshParams()  */


#define No_COMPMAT1_INTS 2

void e_MPI_Send_Mats(MESH *mesh, int dest)
{
  register  i,j;
  int      *vect_ind;
  int       i_offset;
  int      *ibuf_MatStruc;
  double   *dbuf_MatStruc;
  int      *ibuf_CompMat1Struc;


  if(mesh->SubStructSet[E_SS_NUM_MAT_PROPS])
  {
    /* from here I assume that: mesh->KeywordSet[E_KW_NUM_NMATS] == TRUE
       and the value of mesh->NMats is the same on the other side        */
    if((ibuf_MatStruc = 
          e_alloc_i_vect(E_NMAT_PROP_KEYWORDS * mesh->NMats)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for ibuf_MatStruc");
#endif
    if((dbuf_MatStruc = 
          e_alloc_d_vect(E_NMAT_PROP_KEYWORDS * mesh->NMats)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for dbuf_MatStruc");
#endif

    /* let's send all the KeywordSet vectors over first... */
    for(i = 0;i < mesh->NMats;i++)
      MPI_Send(mesh->Props[i].KeywordSet, E_NMAT_PROP_KEYWORDS, 
               MPI_INT, dest, 200, MPI_COMM_WORLD);

    for(i = 0;i < mesh->NMats;i++)
    {
      i_offset = i * E_NMAT_PROP_KEYWORDS;
      ibuf_MatStruc[i_offset + E_MKW_NUM_END] = strlen(mesh->Props[i].Name)+1;
      ibuf_MatStruc[i_offset + E_MKW_NUM_MAT_TYPE]  = mesh->Props[i].MatType;
      ibuf_MatStruc[i_offset + E_MKW_NUM_IPARAM1]   = mesh->Props[i].iParam1;
      ibuf_MatStruc[i_offset + E_MKW_NUM_IPARAM2]   = mesh->Props[i].iParam2;
      ibuf_MatStruc[i_offset + E_MKW_NUM_IPARAM3]   = mesh->Props[i].iParam3;
      ibuf_MatStruc[i_offset + E_MKW_NUM_IPARAM4]   = mesh->Props[i].iParam4;
      ibuf_MatStruc[i_offset + E_MKW_NUM_IPARAM5]   = mesh->Props[i].iParam5;
      ibuf_MatStruc[i_offset + E_MKW_NUM_IPARAM6]   = mesh->Props[i].iParam6;

      dbuf_MatStruc[i_offset + E_MKW_NUM_DENSITY]   = mesh->Props[i].Density;
      dbuf_MatStruc[i_offset + E_MKW_NUM_YMOD]      = mesh->Props[i].YMod;
      dbuf_MatStruc[i_offset + E_MKW_NUM_YMOD_X]    = mesh->Props[i].YMod_x;
      dbuf_MatStruc[i_offset + E_MKW_NUM_YMOD_Y]    = mesh->Props[i].YMod_y;
      dbuf_MatStruc[i_offset + E_MKW_NUM_YMOD_Z]    = mesh->Props[i].YMod_z;
      dbuf_MatStruc[i_offset + E_MKW_NUM_POISSONS_RATIO] = 
                                                mesh->Props[i].PoissonsRatio;
      dbuf_MatStruc[i_offset + E_MKW_NUM_THICKNESS] = mesh->Props[i].Thickness;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM1]   = mesh->Props[i].dParam1;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM2]   = mesh->Props[i].dParam2;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM3]   = mesh->Props[i].dParam3;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM4]   = mesh->Props[i].dParam4;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM5]   = mesh->Props[i].dParam5;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM6]   = mesh->Props[i].dParam6;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM7]   = mesh->Props[i].dParam7;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM8]   = mesh->Props[i].dParam8;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM9]   = mesh->Props[i].dParam9;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM10]  = mesh->Props[i].dParam10;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM11]  = mesh->Props[i].dParam11;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM12]  = mesh->Props[i].dParam12;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM13]  = mesh->Props[i].dParam13;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM14]  = mesh->Props[i].dParam14;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM15]  = mesh->Props[i].dParam15;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM16]  = mesh->Props[i].dParam16;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM17]  = mesh->Props[i].dParam17;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM18]  = mesh->Props[i].dParam18;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM19]  = mesh->Props[i].dParam19;
      dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM20]  = mesh->Props[i].dParam20;
    }  /* for(i) 0 - mesh->NMats  */

    MPI_Send(ibuf_MatStruc, mesh->NMats * E_NMAT_PROP_KEYWORDS, 
             MPI_INT, dest, 201, MPI_COMM_WORLD);
    MPI_Send(dbuf_MatStruc, mesh->NMats * E_NMAT_PROP_KEYWORDS, 
             MPI_DOUBLE, dest, 202, MPI_COMM_WORLD);

    for(i = 0;i < mesh->NMats;i++)
      MPI_Send(mesh->Props[i].Name,
               ibuf_MatStruc[i * E_NMAT_PROP_KEYWORDS + E_MKW_NUM_END], 
               MPI_CHAR, dest, 203, MPI_COMM_WORLD);

    if(mesh->KeywordSet[E_KW_NUM_NCOMP_MATS_TYPE1])
    {
      /* from here I assume that: 
         the value of mesh->NCompMatsType1 is the same on the other side */
      if((ibuf_CompMat1Struc = 
            e_alloc_i_vect(No_COMPMAT1_INTS * mesh->NCompMatsType1)) == NULL)
#ifdef __E_T800__
        NotEnoughMemory();
#else
        e_GeneralError("Not enough memory for ibuf_CompMat1Struc");
#endif

      for(i = 0;i < mesh->NCompMatsType1;i++)
      {
        i_offset = i * No_COMPMAT1_INTS;
        ibuf_CompMat1Struc[i_offset]   = strlen(mesh->CompPropsType1[i].Name)+1;
        ibuf_CompMat1Struc[i_offset+1] = mesh->CompPropsType1[i].NLayers;
      }  /* for(i) 0 - mesh->NCompMatsType1  */

      MPI_Send(ibuf_CompMat1Struc, mesh->NCompMatsType1 * No_COMPMAT1_INTS,
               MPI_INT, dest, 204, MPI_COMM_WORLD);

      for(i = 0;i < mesh->NCompMatsType1;i++)
      {
        i_offset = i * No_COMPMAT1_INTS;
        MPI_Send(mesh->CompPropsType1[i].Thicknesses,
                 mesh->CompPropsType1[i].NLayers, 
                 MPI_DOUBLE, dest, 205, MPI_COMM_WORLD);

        if((vect_ind = 
              e_alloc_i_vect(mesh->CompPropsType1[i].NLayers)) == NULL)
#ifdef __E_T800__
          NotEnoughMemory();
#else
          e_GeneralError("Not enough memory for Vect_ind");
#endif

        for(j = 0;j < mesh->CompPropsType1[i].NLayers;j++)
          vect_ind[j] = (mesh->CompPropsType1[i].LayerMats[j]) - (mesh->Props);
        MPI_Send(vect_ind, mesh->CompPropsType1[i].NLayers, 
                 MPI_INT, dest, 206, MPI_COMM_WORLD);
        e_free(vect_ind);

        MPI_Send(mesh->CompPropsType1[i].Name, ibuf_CompMat1Struc[i_offset],
                 MPI_CHAR, dest, 207, MPI_COMM_WORLD);
      }  /* for(i) 0 - mesh->NCompMatsType1  */
      e_free(ibuf_CompMat1Struc);
    }

    e_free(ibuf_MatStruc);
    e_free(dbuf_MatStruc);
  }      /*  if(mesh->SubStructSet[E_SS_NUM_MAT_PROPS])  */

}   /* end of function e_MPI_Send_Mats()  */


void e_MPI_Recv_Mats(MESH *mesh, int source)
{
  register    i,j;
  int        *vect_ind;
  int         i_offset;
  MPI_Status  status;
  int        *ibuf_MatStruc;
  double     *dbuf_MatStruc;
  int        *ibuf_CompMat1Struc;


  if(mesh->SubStructSet[E_SS_NUM_MAT_PROPS])
  {
    /* from here I assume that: mesh->KeywordSet[E_KW_NUM_NMATS] == TRUE
       and the value of mesh->NMats is the same on the other side        */
    if((mesh->Props = e_calloc(sizeof(MAT),mesh->NMats)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for Props");
#endif
    if((ibuf_MatStruc = 
          e_alloc_i_vect(E_NMAT_PROP_KEYWORDS * mesh->NMats)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for ibuf_MatStruc");
#endif
    if((dbuf_MatStruc = 
          e_alloc_d_vect(E_NMAT_PROP_KEYWORDS * mesh->NMats)) == NULL)
#ifdef __E_T800__
      NotEnoughMemory();
#else
      e_GeneralError("Not enough memory for dbuf_MatStruc");
#endif

           /* let's receive all the KeywordSet vectors first... */
    for(i = 0;i < mesh->NMats;i++)
      MPI_Recv(mesh->Props[i].KeywordSet, E_NMAT_PROP_KEYWORDS, 
               MPI_INT, source, 200, MPI_COMM_WORLD, &status);

    MPI_Recv(ibuf_MatStruc, mesh->NMats * E_NMAT_PROP_KEYWORDS, 
             MPI_INT, source, 201, MPI_COMM_WORLD, &status);
    MPI_Recv(dbuf_MatStruc, mesh->NMats * E_NMAT_PROP_KEYWORDS, 
             MPI_DOUBLE, source, 202, MPI_COMM_WORLD, &status);


    for(i = 0; i < mesh->NMats; i++)
    {
      i_offset = i * E_NMAT_PROP_KEYWORDS;

      if((mesh->Props[i].Name = 
            e_alloc_c_vect(ibuf_MatStruc[i_offset + E_MKW_NUM_END])) == NULL)
#ifdef __E_T800__
        NotEnoughMemory();
#else
        e_GeneralError("Not enough memory for mesh MatNames");
#endif
      MPI_Recv(mesh->Props[i].Name, ibuf_MatStruc[i_offset + E_MKW_NUM_END],
               MPI_CHAR, source, 203, MPI_COMM_WORLD, &status);

      mesh->Props[i].MatType = ibuf_MatStruc[i_offset + E_MKW_NUM_MAT_TYPE];
      mesh->Props[i].iParam1 = ibuf_MatStruc[i_offset + E_MKW_NUM_IPARAM1];
      mesh->Props[i].iParam2 = ibuf_MatStruc[i_offset + E_MKW_NUM_IPARAM2];
      mesh->Props[i].iParam3 = ibuf_MatStruc[i_offset + E_MKW_NUM_IPARAM3];
      mesh->Props[i].iParam4 = ibuf_MatStruc[i_offset + E_MKW_NUM_IPARAM4];
      mesh->Props[i].iParam5 = ibuf_MatStruc[i_offset + E_MKW_NUM_IPARAM5];
      mesh->Props[i].iParam6 = ibuf_MatStruc[i_offset + E_MKW_NUM_IPARAM6];

      mesh->Props[i].Density = dbuf_MatStruc[i_offset + E_MKW_NUM_DENSITY];
      mesh->Props[i].YMod    = dbuf_MatStruc[i_offset + E_MKW_NUM_YMOD];
      mesh->Props[i].YMod_x  = dbuf_MatStruc[i_offset + E_MKW_NUM_YMOD_X];
      mesh->Props[i].YMod_y  = dbuf_MatStruc[i_offset + E_MKW_NUM_YMOD_Y];
      mesh->Props[i].YMod_z  = dbuf_MatStruc[i_offset + E_MKW_NUM_YMOD_Z];
      mesh->Props[i].PoissonsRatio = dbuf_MatStruc[i_offset + 
                                                E_MKW_NUM_POISSONS_RATIO];
      mesh->Props[i].Thickness = dbuf_MatStruc[i_offset + E_MKW_NUM_THICKNESS];
      mesh->Props[i].dParam1   = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM1];
      mesh->Props[i].dParam2   = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM2];
      mesh->Props[i].dParam3   = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM3];
      mesh->Props[i].dParam4   = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM4];
      mesh->Props[i].dParam5   = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM5];
      mesh->Props[i].dParam6   = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM6];
      mesh->Props[i].dParam7   = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM7];
      mesh->Props[i].dParam8   = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM8];
      mesh->Props[i].dParam9   = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM9];
      mesh->Props[i].dParam10  = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM10];
      mesh->Props[i].dParam11  = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM11];
      mesh->Props[i].dParam12  = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM12];
      mesh->Props[i].dParam13  = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM13];
      mesh->Props[i].dParam14  = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM14];
      mesh->Props[i].dParam15  = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM15];
      mesh->Props[i].dParam16  = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM16];
      mesh->Props[i].dParam17  = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM17];
      mesh->Props[i].dParam18  = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM18];
      mesh->Props[i].dParam19  = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM19];
      mesh->Props[i].dParam20  = dbuf_MatStruc[i_offset + E_MKW_NUM_DPARAM20];
    }  /* for(i) 0 - mesh->NMats  */


    if(mesh->KeywordSet[E_KW_NUM_NCOMP_MATS_TYPE1])
    {
      /* from here I assume that: 
         the value of mesh->NCompMatsType1 is the same on the other side */
      if((mesh->CompPropsType1 = 
            e_calloc(sizeof(COMPMAT1),mesh->NCompMatsType1)) == NULL)
#ifdef __E_T800__
        NotEnoughMemory();
#else
        e_GeneralError("Not enough memory for CompPropsType1");
#endif
      if((ibuf_CompMat1Struc = 
            e_alloc_i_vect(No_COMPMAT1_INTS * mesh->NCompMatsType1)) == NULL)
#ifdef __E_T800__
        NotEnoughMemory();
#else
        e_GeneralError("Not enough memory for ibuf_CompMat1Struc");
#endif

      MPI_Recv(ibuf_CompMat1Struc, mesh->NCompMatsType1 * No_COMPMAT1_INTS, 
               MPI_INT, source, 204, MPI_COMM_WORLD, &status);

      for(i = 0; i < mesh->NCompMatsType1; i++)
      {
        i_offset = i * No_COMPMAT1_INTS;

        mesh->CompPropsType1[i].NLayers = ibuf_CompMat1Struc[i_offset+1];
        if((mesh->CompPropsType1[i].Thicknesses = 
              e_alloc_d_vect(mesh->CompPropsType1[i].NLayers)) == NULL)
#ifdef __E_T800__
          NotEnoughMemory();
#else
          e_GeneralError("Not enough memory for CompPropsType1.Thicknesses");
#endif
        MPI_Recv(mesh->CompPropsType1[i].Thicknesses,
                 mesh->CompPropsType1[i].NLayers, 
                 MPI_DOUBLE, source, 205, MPI_COMM_WORLD, &status);
        if((vect_ind = 
              e_alloc_i_vect(mesh->CompPropsType1[i].NLayers)) == NULL)
#ifdef __E_T800__
          NotEnoughMemory();
#else
          e_GeneralError("Not enough memory for Vect_ind");
#endif
        if((mesh->CompPropsType1[i].LayerMats = 
            e_calloc(sizeof(COMPMAT1),mesh->CompPropsType1[i].NLayers)) == NULL)
#ifdef __E_T800__
          NotEnoughMemory();
#else
          e_GeneralError("Not enough memory for CompPropsType1.LayerMats");
#endif

        MPI_Recv(vect_ind, mesh->CompPropsType1[i].NLayers, 
                 MPI_INT, source, 206, MPI_COMM_WORLD, &status);
        for(j = 0;j < mesh->CompPropsType1[i].NLayers;j++)
          mesh->CompPropsType1[i].LayerMats[j] = mesh->Props + vect_ind[j];
        e_free(vect_ind);

        if((mesh->CompPropsType1[i].Name = 
              e_alloc_c_vect(ibuf_CompMat1Struc[i_offset])) == NULL)
#ifdef __E_T800__
          NotEnoughMemory();
#else
          e_GeneralError("Not enough memory for mesh CompMatNamesType1");
#endif
        MPI_Recv(mesh->CompPropsType1[i].Name,ibuf_CompMat1Struc[i_offset],
                 MPI_CHAR, source, 207, MPI_COMM_WORLD, &status);
      }
      e_free(ibuf_CompMat1Struc);
    }
    e_free(ibuf_MatStruc);
    e_free(dbuf_MatStruc);
  }

}   /* end of function e_MPI_Recv_Mats()  */

void e_MPI_Send_DecompData(MESH * mesh, int dest)
{
  if(mesh->SubStructSet[E_SS_NUM_DECOMP])
  {
    MPI_Send(&(mesh->Decomp.NSubDoms), 1, MPI_INT, 
             dest, 100, MPI_COMM_WORLD);
    MPI_Send(mesh->Decomp.SubDoms, mesh->TotalNElems, MPI_INT, 
             dest, 101, MPI_COMM_WORLD);
  }
} /* end of function e_MPI_Send_DecompData()  */


void e_MPI_Recv_DecompData(MESH * mesh, int source)
{
  MPI_Status      status;

  MPI_Recv(&(mesh->Decomp.NSubDoms), 1, MPI_INT, 
           source, 100, MPI_COMM_WORLD, &status);

  if( (mesh->Decomp.SubDoms = e_alloc_i_vect(mesh->TotalNElems)) == NULL)
#ifdef __E_T800__
    NotEnoughMemory();
#else
    e_GeneralError("Not enough memory for Decomp.SubDoms");
#endif
  MPI_Recv(mesh->Decomp.SubDoms, mesh->TotalNElems, MPI_INT, 
           source, 101, MPI_COMM_WORLD, &status);

  mesh->SubStructSet[E_SS_NUM_DECOMP] = TRUE;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
    mesh->Decomp.SubDomsLink1 = mesh->Decomp.SubDoms + 
                                mesh->FirstNetIndex[E_ELEM_TYPE_LINK1] - 1;
  else
    mesh->Decomp.SubDomsLink1 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
    mesh->Decomp.SubDomsLink2 = mesh->Decomp.SubDoms + 
                                mesh->FirstNetIndex[E_ELEM_TYPE_LINK2] - 1;
  else
    mesh->Decomp.SubDomsLink2 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
    mesh->Decomp.SubDomsLink3 = mesh->Decomp.SubDoms + 
                                mesh->FirstNetIndex[E_ELEM_TYPE_LINK3] - 1;
  else
    mesh->Decomp.SubDomsLink3 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
    mesh->Decomp.SubDomsLink4 = mesh->Decomp.SubDoms + 
                                mesh->FirstNetIndex[E_ELEM_TYPE_LINK4] - 1;
  else
    mesh->Decomp.SubDomsLink4 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
    mesh->Decomp.SubDomsLink5 = mesh->Decomp.SubDoms + 
                                mesh->FirstNetIndex[E_ELEM_TYPE_LINK5] - 1;
  else
    mesh->Decomp.SubDomsLink5 = NULL;


  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
    mesh->Decomp.SubDomsTriang1 = mesh->Decomp.SubDoms + 
                                  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] - 1;
  else
    mesh->Decomp.SubDomsTriang1 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
    mesh->Decomp.SubDomsTriang2 = mesh->Decomp.SubDoms + 
                                  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2] - 1;
  else
    mesh->Decomp.SubDomsTriang2 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
    mesh->Decomp.SubDomsTriang3 = mesh->Decomp.SubDoms + 
                                  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] - 1;
  else
    mesh->Decomp.SubDomsTriang3 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
    mesh->Decomp.SubDomsTriang4 = mesh->Decomp.SubDoms + 
                                  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] - 1;
  else
    mesh->Decomp.SubDomsTriang4 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
    mesh->Decomp.SubDomsTriang5 = mesh->Decomp.SubDoms + 
                                  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] - 1;
  else
    mesh->Decomp.SubDomsTriang5 = NULL;


  if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
    mesh->Decomp.SubDomsQuad1 = mesh->Decomp.SubDoms + 
                                mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1] - 1;
  else
    mesh->Decomp.SubDomsQuad1 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
    mesh->Decomp.SubDomsQuad3 = mesh->Decomp.SubDoms + 
                                mesh->FirstNetIndex[E_ELEM_TYPE_QUAD3] - 1;
  else
    mesh->Decomp.SubDomsQuad3 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
    mesh->Decomp.SubDomsTetrah1 = mesh->Decomp.SubDoms + 
                                  mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1] - 1;
  else
    mesh->Decomp.SubDomsTetrah1 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
    mesh->Decomp.SubDomsBlock1 = mesh->Decomp.SubDoms + 
                                 mesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1] - 1;
  else
    mesh->Decomp.SubDomsBlock1 = NULL;

} /* end of function e_MPI_Recv_DecompData()  */


void e_MPI_Send_FEErrorData(MESH *mesh, int dest)
{
  if(mesh->SubStructSet[E_SS_NUM_FEERROR])
  {
    MPI_Send(mesh->FEError.FEErrors, mesh->TotalNElems, MPI_DOUBLE, 
             dest, 301, MPI_COMM_WORLD);
  }
} /* end of function e_MPI_Send_FEErrorData()  */


void e_MPI_Recv_FEErrorData(MESH *mesh, int source)
{
  MPI_Status      status;

  if( (mesh->FEError.FEErrors = e_alloc_d_vect(mesh->TotalNElems)) == NULL)
#ifdef __E_T800__
    NotEnoughMemory();
#else
    e_GeneralError("Not enough memory for FEError.FEErrors");
#endif
  MPI_Recv(mesh->FEError.FEErrors, mesh->TotalNElems, MPI_DOUBLE, 
           source, 301, MPI_COMM_WORLD, &status);

  mesh->SubStructSet[E_SS_NUM_FEERROR] = TRUE;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
    mesh->FEError.FEErrorsLink1 = mesh->FEError.FEErrors + 
                                  mesh->FirstNetIndex[E_ELEM_TYPE_LINK1] - 1;
  else
    mesh->FEError.FEErrorsLink1 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
    mesh->FEError.FEErrorsLink2 = mesh->FEError.FEErrors + 
                                  mesh->FirstNetIndex[E_ELEM_TYPE_LINK2] - 1;
  else
    mesh->FEError.FEErrorsLink2 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
    mesh->FEError.FEErrorsLink3 = mesh->FEError.FEErrors + 
                                  mesh->FirstNetIndex[E_ELEM_TYPE_LINK3] - 1;
  else
    mesh->FEError.FEErrorsLink3 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
    mesh->FEError.FEErrorsLink4 = mesh->FEError.FEErrors + 
                                  mesh->FirstNetIndex[E_ELEM_TYPE_LINK4] - 1;
  else
    mesh->FEError.FEErrorsLink4 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
    mesh->FEError.FEErrorsLink5 = mesh->FEError.FEErrors + 
                                  mesh->FirstNetIndex[E_ELEM_TYPE_LINK5] - 1;
  else
    mesh->FEError.FEErrorsLink5 = NULL;


  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
    mesh->FEError.FEErrorsTriang1 = mesh->FEError.FEErrors + 
                                 mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] - 1;
  else
    mesh->FEError.FEErrorsTriang1 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
    mesh->FEError.FEErrorsTriang2 = mesh->FEError.FEErrors + 
                                 mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2] - 1;
  else
    mesh->FEError.FEErrorsTriang2 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
    mesh->FEError.FEErrorsTriang3 = mesh->FEError.FEErrors + 
                                 mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] - 1;
  else
    mesh->FEError.FEErrorsTriang3 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
    mesh->FEError.FEErrorsTriang4 = mesh->FEError.FEErrors + 
                                 mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] - 1;
  else
    mesh->FEError.FEErrorsTriang4 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
    mesh->FEError.FEErrorsTriang5 = mesh->FEError.FEErrors + 
                                 mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] - 1;
  else
    mesh->FEError.FEErrorsTriang5 = NULL;


  if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
    mesh->FEError.FEErrorsQuad1 = mesh->FEError.FEErrors + 
                                  mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1] - 1;
  else
    mesh->FEError.FEErrorsQuad1 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
    mesh->FEError.FEErrorsQuad3 = mesh->FEError.FEErrors + 
                                  mesh->FirstNetIndex[E_ELEM_TYPE_QUAD3] - 1;
  else
    mesh->FEError.FEErrorsQuad3 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
    mesh->FEError.FEErrorsTetrah1 = mesh->FEError.FEErrors + 
                                mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1] - 1;
  else
    mesh->FEError.FEErrorsTetrah1 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
    mesh->FEError.FEErrorsBlock1 = mesh->FEError.FEErrors + 
                                   mesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1] - 1;
  else
    mesh->FEError.FEErrorsBlock1 = NULL;

} /* end of function e_MPI_Recv_FEErrorData()  */

void  e_MPI_Send_Stresses(MESH *mesh, int dest)
{
  int  total, i;
  int  maxNLayerTriang2, maxNLayerQuad3;

  if(mesh->SubStructSet[E_SS_NUM_STRESS])
  {
    maxNLayerTriang2 = 0;
    maxNLayerQuad3   = 0;
    if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
    {
      MPI_Send(&(mesh->Stresses.StressesLink1[0][0][0]),
               mesh->NElemsLink1 * 
               mesh->NStressPointsLink1 * 
               E_NUM_STRESS_COMPONENTS, 
               MPI_DOUBLE, dest, 501, MPI_COMM_WORLD);
    }
    if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
    {
      MPI_Send(&(mesh->Stresses.StressesLink2[0][0][0]),
               mesh->NElemsLink2 * 
               mesh->NStressPointsLink2 * 
               E_NUM_STRESS_COMPONENTS, 
               MPI_DOUBLE, dest, 502, MPI_COMM_WORLD);
    }
    if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
    {
      MPI_Send(&(mesh->Stresses.StressesLink3[0][0][0]),
               mesh->NElemsLink3 * 
               mesh->NStressPointsLink3 * 
               E_NUM_STRESS_COMPONENTS, 
               MPI_DOUBLE, dest, 503, MPI_COMM_WORLD);
    }
    if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
    {
      MPI_Send(&(mesh->Stresses.StressesLink4[0][0][0]),
               mesh->NElemsLink4 * 
               mesh->NStressPointsLink4 * 
               E_NUM_STRESS_COMPONENTS, 
               MPI_DOUBLE, dest, 504, MPI_COMM_WORLD);
    }
    if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
    {
      MPI_Send(&(mesh->Stresses.StressesLink5[0][0][0]),
               mesh->NElemsLink5 * 
               mesh->NStressPointsLink5 * 
               E_NUM_STRESS_COMPONENTS, 
               MPI_DOUBLE, dest, 505, MPI_COMM_WORLD);
    }
    if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
    {
      MPI_Send(&(mesh->Stresses.StressesTriang1[0][0][0]),
               mesh->NElemsTriang1 * 
               mesh->NStressPointsTriang1 * 
               E_NUM_STRESS_COMPONENTS, 
               MPI_DOUBLE, dest, 506, MPI_COMM_WORLD);
    }
    if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
    {
      for(i = 0; i < mesh->NElemsTriang2; i++)
        if(maxNLayerTriang2 < mesh->CompMatsTriang2[i]->NLayers)
          maxNLayerTriang2 = mesh->CompMatsTriang2[i]->NLayers;

      MPI_Send(&(mesh->Stresses.StressesTriang2[0][0][0][0]),
               mesh->NElemsTriang2 * maxNLayerTriang2 *
               mesh->NStressPointsTriang2 * E_NUM_STRESS_COMPONENTS, 
               MPI_DOUBLE, dest, 507, MPI_COMM_WORLD);
    }
    if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
    {
      MPI_Send(&(mesh->Stresses.StressesTriang3[0][0][0]),
               mesh->NElemsTriang3 * 
               mesh->NStressPointsTriang3 * 
               E_NUM_STRESS_COMPONENTS, 
               MPI_DOUBLE, dest, 508, MPI_COMM_WORLD);
    }
    if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
    {
      MPI_Send(&(mesh->Stresses.StressesTriang4[0][0][0]),
               mesh->NElemsTriang4 * 
               mesh->NStressPointsTriang4 * 
               E_NUM_STRESS_COMPONENTS, 
               MPI_DOUBLE, dest, 509, MPI_COMM_WORLD);
    }
    if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
    {
      MPI_Send(&(mesh->Stresses.StressesTriang5[0][0][0]),
               mesh->NElemsTriang5 * 
               mesh->NStressPointsTriang5 * 
               E_NUM_STRESS_COMPONENTS, 
               MPI_DOUBLE, dest, 510, MPI_COMM_WORLD);
    }
    if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
    {
      MPI_Send(&(mesh->Stresses.StressesQuad1[0][0][0]),
               mesh->NElemsQuad1 * 
               mesh->NStressPointsQuad1 * 
               E_NUM_STRESS_COMPONENTS, 
               MPI_DOUBLE, dest, 511, MPI_COMM_WORLD);
    }
    if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
    {
      for(i = 0; i < mesh->NElemsQuad3; i++)
        if(maxNLayerQuad3 < mesh->CompMatsQuad3[i]->NLayers)
          maxNLayerQuad3 = mesh->CompMatsQuad3[i]->NLayers;

      MPI_Send(&(mesh->Stresses.StressesQuad3[0][0][0][0]),
               mesh->NElemsQuad3 * maxNLayerQuad3 *
               mesh->NStressPointsQuad3 * E_NUM_STRESS_COMPONENTS, 
               MPI_DOUBLE, dest, 512, MPI_COMM_WORLD);
    }
    if(mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
    {
      MPI_Send(&(mesh->Stresses.StressesTetrah1[0][0][0]),
               mesh->NElemsTetrah1 * 
               mesh->NStressPointsTetrah1 * 
               E_NUM_STRESS_COMPONENTS, 
               MPI_DOUBLE, dest, 513, MPI_COMM_WORLD);
    }
    if(mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
    {
      MPI_Send(&(mesh->Stresses.StressesBlock1[0][0][0]),
               mesh->NElemsBlock1 * 
               mesh->NStressPointsBlock1 * 
               E_NUM_STRESS_COMPONENTS, 
               MPI_DOUBLE, dest, 514, MPI_COMM_WORLD);
    }
  }
  else
    e_GeneralError("e_MPI_Send_Stresses: no stresses declared");

} /* end of function e_MPI_Send_Stresses()  */


void e_MPI_Recv_Stresses(MESH *mesh, int source)

{
  MPI_Status      status;
  int             total, i;
  int  maxNLayerTriang2, maxNLayerQuad3;

  maxNLayerTriang2 = 0;
  maxNLayerQuad3   = 0;

  if(!e_SetupStresses(mesh))
    e_GeneralError("e_MPI_Recv_Stresses: stress setup failed !");

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
  {
    MPI_Recv(&(mesh->Stresses.StressesLink1[0][0][0]),
             mesh->NElemsLink1 * 
             mesh->NStressPointsLink1 * 
             E_NUM_STRESS_COMPONENTS, 
             MPI_DOUBLE, source, 501, MPI_COMM_WORLD, &status);
  }
  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
  {
    MPI_Recv(&(mesh->Stresses.StressesLink2[0][0][0]),
             mesh->NElemsLink2 * 
             mesh->NStressPointsLink2 * 
             E_NUM_STRESS_COMPONENTS, 
             MPI_DOUBLE, source, 502, MPI_COMM_WORLD, &status);
  }
  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
  {
    MPI_Recv(&(mesh->Stresses.StressesLink3[0][0][0]),
             mesh->NElemsLink3 * 
             mesh->NStressPointsLink3 * 
             E_NUM_STRESS_COMPONENTS, 
             MPI_DOUBLE, source, 503, MPI_COMM_WORLD, &status);
  }
  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
  {
    MPI_Recv(&(mesh->Stresses.StressesLink4[0][0][0]),
             mesh->NElemsLink4 * 
             mesh->NStressPointsLink4 * 
             E_NUM_STRESS_COMPONENTS, 
             MPI_DOUBLE, source, 504, MPI_COMM_WORLD, &status);
  }
  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
  {
    MPI_Recv(&(mesh->Stresses.StressesLink5[0][0][0]),
             mesh->NElemsLink5 * 
             mesh->NStressPointsLink5 * 
             E_NUM_STRESS_COMPONENTS, 
             MPI_DOUBLE, source, 505, MPI_COMM_WORLD, &status);
  }
  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
  {
    MPI_Recv(&(mesh->Stresses.StressesTriang1[0][0][0]),
             mesh->NElemsTriang1 * 
             mesh->NStressPointsTriang1 * 
             E_NUM_STRESS_COMPONENTS, 
             MPI_DOUBLE, source, 506, MPI_COMM_WORLD, &status);
  }
  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
  {
    for(i = 0; i < mesh->NElemsTriang2; i++)
      if(maxNLayerTriang2 < mesh->CompMatsTriang2[i]->NLayers)
        maxNLayerTriang2 = mesh->CompMatsTriang2[i]->NLayers;

    MPI_Recv(&(mesh->Stresses.StressesTriang2[0][0][0][0]),
             mesh->NElemsTriang2 * maxNLayerTriang2 *
             mesh->NStressPointsTriang2 * E_NUM_STRESS_COMPONENTS, 
             MPI_DOUBLE, source, 507, MPI_COMM_WORLD, &status);
  }
  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
  {
    MPI_Recv(&(mesh->Stresses.StressesTriang3[0][0][0]),
             mesh->NElemsTriang3 * 
             mesh->NStressPointsTriang3 * 
             E_NUM_STRESS_COMPONENTS, 
             MPI_DOUBLE, source, 508, MPI_COMM_WORLD, &status);
  }
  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
  {
    MPI_Recv(&(mesh->Stresses.StressesTriang4[0][0][0]),
             mesh->NElemsTriang4 * 
             mesh->NStressPointsTriang4 * 
             E_NUM_STRESS_COMPONENTS, 
             MPI_DOUBLE, source, 509, MPI_COMM_WORLD, &status);
  }
  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
  {
    MPI_Recv(&(mesh->Stresses.StressesTriang5[0][0][0]),
             mesh->NElemsTriang5 * 
             mesh->NStressPointsTriang5 * 
             E_NUM_STRESS_COMPONENTS, 
             MPI_DOUBLE, source, 510, MPI_COMM_WORLD, &status);
  }
  if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
  {
    MPI_Recv(&(mesh->Stresses.StressesQuad1[0][0][0]),
             mesh->NElemsQuad1 * 
             mesh->NStressPointsQuad1 * 
             E_NUM_STRESS_COMPONENTS, 
             MPI_DOUBLE, source, 511, MPI_COMM_WORLD, &status);
  }
  if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
  {
    for(i = 0; i < mesh->NElemsQuad3; i++)
      if(maxNLayerQuad3 < mesh->CompMatsQuad3[i]->NLayers)
        maxNLayerQuad3 = mesh->CompMatsQuad3[i]->NLayers;

    MPI_Recv(&(mesh->Stresses.StressesQuad3[0][0][0][0]),
             mesh->NElemsQuad3 * maxNLayerQuad3 *
             mesh->NStressPointsQuad3 * E_NUM_STRESS_COMPONENTS, 
             MPI_DOUBLE, source, 512, MPI_COMM_WORLD, &status);
  }
  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
  {
    MPI_Recv(&(mesh->Stresses.StressesTetrah1[0][0][0]),
             mesh->NElemsTetrah1 * 
             mesh->NStressPointsTetrah1 * 
             E_NUM_STRESS_COMPONENTS, 
             MPI_DOUBLE, source, 513, MPI_COMM_WORLD, &status);
  }
  if(mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
  {
    MPI_Recv(&(mesh->Stresses.StressesBlock1[0][0][0]),
             mesh->NElemsBlock1 * 
             mesh->NStressPointsBlock1 * 
             E_NUM_STRESS_COMPONENTS, 
             MPI_DOUBLE, source, 514, MPI_COMM_WORLD, &status);
  }
  mesh->SubStructSet[E_SS_NUM_STRESS] = TRUE;
}        /* end of function e_MPI_Recv_Stresses()  */


/*  End Of File: E_MPI-AU.C   */
