/*  File: TRAN_AUX.C   */


/*
    This is an auxiliary library for transputer program development
    to aid with the E_LIB library.
    Functions have been written to send and receive a mesh definiton
    via transputter channels using the new FE mesh data format.

    Two special channel in/out functions were developed to overcome
    the limited bufferspace of the FE router task.
    The maximum allowed buffer space is defined below.
    These functions are:
        void e_chan_out_Buffer(char *pointer, int hossz, CHAN *channel);
        int   e_chan_in_Buffer(char *pointer, CHAN *channel);
    These routines are using the standard
    chan_in/out_word() and chan_in/out_message() function pairs.

*/




#include<chan.h>
#include<stdlib.h>
#include<string.h>
#include<par.h>


#include"e_pltfrm.h"

#ifndef __E_T800SAL__
#include <stdio.h>
#endif

#include<e_lib.h>


/* temporarly: */
#include"e_tr-aux.h"





void e_chan_out_MeshData(MESH *mesh, CHAN *chan)

{
register  i,j;
int       title_size;
int      *vect_ind;

e_chan_out_Buffer(sizeof(MESH),mesh,chan);

title_size = strlen(mesh->Title);
e_chan_out_Buffer(sizeof(int),&title_size,chan);
e_chan_out_Buffer((title_size+1)*sizeof(char),mesh->Title,chan);

       /*  mesh point co-ordinates   */
e_chan_out_Buffer(mesh->NMeshPoints * 3 * sizeof(double),mesh->MeshPointCoords[0],chan);

       /* element nodes  */
if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
    e_chan_out_Buffer(mesh->NElemsTriang1 * E_NNPE_TRIANG1 * sizeof(int),mesh->NodesTriang1[0],chan);

if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
    e_chan_out_Buffer(mesh->NElemsTriang2 * E_NNPE_TRIANG2 * sizeof(int),mesh->NodesTriang2[0],chan);

if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
    e_chan_out_Buffer(mesh->NElemsQuad1 * E_NNPE_QUAD1 * sizeof(int),mesh->NodesQuad1[0],chan);

if(mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
    e_chan_out_Buffer(mesh->NElemsTetrah1 * E_NNPE_TETRAH1 * sizeof(int),mesh->NodesTetrah1[0],chan);

if(mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
    e_chan_out_Buffer(mesh->NElemsBlock1 * E_NNPE_BLOCK1 * sizeof(int),mesh->NodesBlock1[0],chan);


       /*  NodalDOF  */
e_chan_out_Buffer(mesh->NNodes * sizeof(int),mesh->NodalDOF,chan);


       /*  Boundary Conditions  */
if(mesh->KeywordSet[E_KW_NUM_BCS] && mesh->KeywordSet[E_KW_NUM_NBC_NODES])
    {
    e_chan_out_Buffer(mesh->NBCNodes * sizeof(int),mesh->BCNodes,chan);
    for(i = 0;i < mesh->NBCNodes;i++)
        {
        e_chan_out_Buffer(E_DOF_MAX * sizeof(int),   mesh->BCTypes[i],       chan);
        e_chan_out_Buffer(E_DOF_MAX * sizeof(double),mesh->BCDispls[i],      chan);
        e_chan_out_Buffer(E_DOF_MAX * sizeof(double),mesh->BCSpringConsts[i],chan);
        }
    }


       /*  Loading Conditions  */
if(mesh->KeywordSet[E_KW_NUM_LOADS] && mesh->KeywordSet[E_KW_NUM_NLOADED_NODES])
    {
    e_chan_out_Buffer(mesh->NLoadedNodes * sizeof(int),mesh->LoadedNodes,chan);
    for(i = 0;i < mesh->NLoadedNodes;i++)
        e_chan_out_Buffer(E_DOF_MAX * sizeof(double),mesh->Loads[i],chan);
    }



       /* Material information */

if(mesh->SubStructSet[E_SS_NUM_MAT_PROPS])
    {
    e_chan_out_Mats(mesh,chan);


    if( (vect_ind = e_alloc_i_vect(mesh->TotalNElems)) == NULL)
        e_GeneralError("Not enough memory for Vect_ind");

    if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
        {
        for(j = 0;j < mesh->NElemsTriang1;j++)
            vect_ind[j] = (mesh->MatsTriang1[j]) - (mesh->Props);

        e_chan_out_Buffer(mesh->NElemsTriang1 * sizeof(int),vect_ind,chan);
        }
    if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
        {
        for(j = 0;j < mesh->NElemsTriang2;j++)
            vect_ind[j] = (mesh->CompMatsTriang2[j]) - (mesh->CompPropsType1);

        e_chan_out_Buffer(mesh->NElemsTriang2 * sizeof(int),vect_ind,chan);
        }


    e_free(vect_ind);
    }


       /*  GNIs  */
e_chan_out_Buffer(mesh->NNodes * sizeof(int),mesh->GNIs,chan);

       /*  GlobalElementIndices  */
e_chan_out_Buffer(mesh->TotalNElems * sizeof(int),mesh->GEIs,chan);



}  /* end of function e_chan_out_MeshData()  */

















void e_chan_in_MeshData(MESH **mesh, CHAN *chan)

{
register  i,j;
int       msglen;
int       title_size;
int      *vect_ind;


              /* mesh structure  */
if( (*mesh = e_calloc(1,sizeof(MESH))) == NULL)
    e_GeneralError("Not enough memory for mesh");
msglen = e_chan_in_Buffer(*mesh,chan);
if(msglen != sizeof(MESH))
    e_GeneralError("wrong size received");


              /* title string  */
msglen = e_chan_in_Buffer(&title_size,chan);
if(msglen != sizeof(int))
    e_GeneralError("wrong size received");

if( ((*mesh)->Title = e_alloc_c_vect(title_size+1)) == NULL)
    e_GeneralError("Not enough memory for mesh Title");
msglen = e_chan_in_Buffer((*mesh)->Title,chan);
if(msglen != (title_size+1)*sizeof(char))
    e_GeneralError("wrong size received");


              /* mesh point co-ordinates   */
if( ((*mesh)->MeshPointCoords = e_alloc_cont_d_matr((*mesh)->NMeshPoints,3)) == NULL)
    e_GeneralError("Not enough memory for mesh point co-ords");
msglen = e_chan_in_Buffer((*mesh)->MeshPointCoords[0],chan);
if(msglen != (*mesh)->NMeshPoints * 3 * sizeof(double))
    e_GeneralError("wrong size received");


       /* element vertices  */
if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
    {
    if( ((*mesh)->NodesTriang1 = e_alloc_cont_i_matr((*mesh)->NElemsTriang1,E_NNPE_TRIANG1)) == NULL)
        e_GeneralError("Not enough memory for triang1 vertices");
    msglen = e_chan_in_Buffer((*mesh)->NodesTriang1[0],chan);
    if(msglen != (*mesh)->NElemsTriang1 * E_NNPE_TRIANG1 * sizeof(int))
        e_GeneralError("wrong size received");
    }  /* E_KW_NUM_NELEMS_TRIANG1  */
if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
    {
    if( ((*mesh)->NodesTriang2 = e_alloc_cont_i_matr((*mesh)->NElemsTriang2,E_NNPE_TRIANG2)) == NULL)
        e_GeneralError("Not enough memory for triang2 vertices");
    msglen = e_chan_in_Buffer((*mesh)->NodesTriang2[0],chan);
    if(msglen != (*mesh)->NElemsTriang2 * E_NNPE_TRIANG2 * sizeof(int))
        e_GeneralError("wrong size received");
    }  /* E_KW_NUM_NELEMS_TRIANG2 */
if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
    {
    if( ((*mesh)->NodesQuad1 = e_alloc_cont_i_matr((*mesh)->NElemsQuad1,E_NNPE_QUAD1)) == NULL)
        e_GeneralError("Not enough memory for Quad1 vertices");
    msglen = e_chan_in_Buffer((*mesh)->NodesQuad1[0],chan);
    if(msglen != (*mesh)->NElemsQuad1 * E_NNPE_QUAD1 * sizeof(int))
        e_GeneralError("wrong size received");
    }  /* E_KW_NUM_NELEMS_QUAD1 */
if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
    {
    if( ((*mesh)->NodesTetrah1 = e_alloc_cont_i_matr((*mesh)->NElemsTetrah1,E_NNPE_TETRAH1)) == NULL)
        e_GeneralError("Not enough memory for Tetrah1 vertices");
    msglen = e_chan_in_Buffer((*mesh)->NodesTetrah1[0],chan);
    if(msglen != (*mesh)->NElemsTetrah1 * E_NNPE_TETRAH1 * sizeof(int))
        e_GeneralError("wrong size received");
    }  /* E_KW_NUM_NELEMS_TETRAH1 */
if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
    {
    if( ((*mesh)->NodesBlock1 = e_alloc_cont_i_matr((*mesh)->NElemsBlock1,E_NNPE_BLOCK1)) == NULL)
        e_GeneralError("Not enough memory for Block1 vertices");
    msglen = e_chan_in_Buffer((*mesh)->NodesBlock1[0],chan);
    if(msglen != (*mesh)->NElemsBlock1 * E_NNPE_BLOCK1 * sizeof(int))
        e_GeneralError("wrong size received");
    }  /* E_KW_NUM_NELEMS_BLOCK1 */


       /*  NodalDOF  */
if( ((*mesh)->NodalDOF = e_alloc_i_vect((*mesh)->NNodes)) == NULL)
    e_GeneralError("Not enough memory for GNIs");
msglen = e_chan_in_Buffer((*mesh)->NodalDOF,chan);
if(msglen != (*mesh)->NNodes * sizeof(int))
    e_GeneralError("wrong size received");



       /*  Boundary Conditions  */
if((*mesh)->KeywordSet[E_KW_NUM_BCS] && (*mesh)->KeywordSet[E_KW_NUM_NBC_NODES])
    {
    if( ((*mesh)->BCNodes = e_alloc_i_vect((*mesh)->NBCNodes)) == NULL)
        e_GeneralError("Not enough memory for BCNodes");

    msglen = e_chan_in_Buffer((*mesh)->BCNodes,chan);
    if(msglen != (*mesh)->NBCNodes * sizeof(int))
        e_GeneralError("wrong size received");


    if( ((*mesh)->BCTypes = e_init_frag_i_matr((*mesh)->NBCNodes)) == NULL)
        e_GeneralError("Not enough memory for BCTypes");
    if( ((*mesh)->BCDispls = e_init_frag_d_matr((*mesh)->NBCNodes)) == NULL)
        e_GeneralError("Not enough memory for BCDispls");
    if( ((*mesh)->BCSpringConsts = e_init_frag_d_matr((*mesh)->NBCNodes)) == NULL)
        e_GeneralError("Not enough memory for BCSpringConsts");

    for(i = 0;i < (*mesh)->NBCNodes;i++)
        {
        if( ((*mesh)->BCTypes[i] = e_alloc_i_vect(E_DOF_MAX)) == NULL)
            e_GeneralError("Not enough memory for BCTypes");
        msglen = e_chan_in_Buffer((*mesh)->BCTypes[i],chan);
        if(msglen != E_DOF_MAX * sizeof(int))
            e_GeneralError("wrong size received");

        if( ((*mesh)->BCDispls[i] = e_alloc_d_vect(E_DOF_MAX)) == NULL)
            e_GeneralError("Not enough memory for BCDispls");
        msglen = e_chan_in_Buffer((*mesh)->BCDispls[i],chan);
        if(msglen != E_DOF_MAX * sizeof(double))
            e_GeneralError("wrong size received");

        if( ((*mesh)->BCSpringConsts[i] = e_alloc_d_vect(E_DOF_MAX)) == NULL)
            e_GeneralError("Not enough memory for BCSpringConsts");
        msglen = e_chan_in_Buffer((*mesh)->BCSpringConsts[i],chan);
        if(msglen != E_DOF_MAX * sizeof(double))
            e_GeneralError("wrong size received");
        }
    }




       /*  Loading Conditions  */
if((*mesh)->KeywordSet[E_KW_NUM_LOADS] && (*mesh)->KeywordSet[E_KW_NUM_NLOADED_NODES])
    {
    if( ((*mesh)->LoadedNodes = e_alloc_i_vect((*mesh)->NLoadedNodes)) == NULL)
        e_GeneralError("Not enough memory for LoadedNodes");

    msglen = e_chan_in_Buffer((*mesh)->LoadedNodes,chan);
    if(msglen != (*mesh)->NLoadedNodes * sizeof(int))
        e_GeneralError("wrong size received");


    if( ((*mesh)->Loads = e_init_frag_d_matr((*mesh)->NLoadedNodes)) == NULL)
        e_GeneralError("Not enough memory for Loads");

    for(i = 0;i < (*mesh)->NLoadedNodes;i++)
        {
        if( ((*mesh)->Loads[i] = e_alloc_d_vect(E_DOF_MAX)) == NULL)
            e_GeneralError("Not enough memory for Loads");
        msglen = e_chan_in_Buffer((*mesh)->Loads[i],chan);
        if(msglen != E_DOF_MAX * sizeof(double))
            e_GeneralError("wrong size received");
        }
    }




       /* Material information */

if((*mesh)->SubStructSet[E_SS_NUM_MAT_PROPS])
    {
    e_chan_in_Mats(*mesh,chan);

    if( (vect_ind = e_alloc_i_vect((*mesh)->TotalNElems)) == NULL)
        e_GeneralError("Not enough memory for Vect_ind");

    if( ((*mesh)->Mats = e_alloc_vp_vect((*mesh)->TotalNElems)) == NULL)
        e_GeneralError("Not enough memory for Mats");


    if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
        {
        (*mesh)->MatsTriang1 = (MAT **)((*mesh)->Mats + (*mesh)->FirstNetIndex[E_ELEM_TYPE_TRIANG1] - 1);

        msglen = e_chan_in_Buffer(vect_ind,chan);
        if(msglen != (*mesh)->NElemsTriang1 * sizeof(int))
            e_GeneralError("wrong size received");

        for(j = 0;j < (*mesh)->NElemsTriang1;j++)
            (*mesh)->MatsTriang1[j] = (*mesh)->Props + vect_ind[j];
        }  /* E_KW_NUM_NELEMS_TRIANG1  */
    else
        (*mesh)->MatsTriang1 = NULL;

    if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
        {
        (*mesh)->CompMatsTriang2 = (COMPMAT1 **)((*mesh)->Mats + (*mesh)->FirstNetIndex[E_ELEM_TYPE_TRIANG2] - 1);

        msglen = e_chan_in_Buffer(vect_ind,chan);
        if(msglen != (*mesh)->NElemsTriang2 * sizeof(int))
            e_GeneralError("wrong size received");

        for(j = 0;j < (*mesh)->NElemsTriang2;j++)
            (*mesh)->CompMatsTriang2[j] = (*mesh)->CompPropsType1 + vect_ind[j];
        }  /* E_KW_NUM_NELEMS_TRIANG1  */
    else
        (*mesh)->MatsTriang1 = NULL;

    e_free(vect_ind);
    }



       /*  GNIs  */
if( ((*mesh)->GNIs = e_alloc_i_vect((*mesh)->NNodes)) == NULL)
    e_GeneralError("Not enough memory for GNIs");
msglen = e_chan_in_Buffer((*mesh)->GNIs,chan);
if(msglen != (*mesh)->NNodes * sizeof(int))
    e_GeneralError("wrong size received");


       /*  GlobalElementIndices  */
if( ((*mesh)->GEIs = e_alloc_i_vect((*mesh)->TotalNElems)) == NULL)
    e_GeneralError("Not enough memory for GlobalElementIndices");
msglen = e_chan_in_Buffer((*mesh)->GEIs,chan);
if(msglen != (*mesh)->TotalNElems * sizeof(int))
    e_GeneralError("wrong size received");

if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
    (*mesh)->GEIsTriang1 = (*mesh)->GEIs + (*mesh)->FirstNetIndex[E_ELEM_TYPE_TRIANG1] - 1;
else
    (*mesh)->GEIsTriang1 = NULL;
if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
    (*mesh)->GEIsTriang2 = (*mesh)->GEIs + (*mesh)->FirstNetIndex[E_ELEM_TYPE_TRIANG2] - 1;
else
    (*mesh)->GEIsTriang2 = NULL;
if((*mesh)->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
    (*mesh)->GEIsQuad1 = (*mesh)->GEIs + (*mesh)->FirstNetIndex[E_ELEM_TYPE_QUAD1] - 1;
else
    (*mesh)->GEIsQuad1 = NULL;





}  /* end of function e_chan_in_MeshData()  */










void e_chan_out_DecompData(MESH  *mesh, CHAN *chan)

{
if(mesh->SubStructSet[E_SS_NUM_DECOMP])
    {
    e_chan_out_Buffer(sizeof(int),&(mesh->Decomp.NSubDoms),chan);

    e_chan_out_Buffer(mesh->TotalNElems * sizeof(int),mesh->Decomp.SubDoms,chan);
    }

}  /* end of function e_chan_out_DecompData()  */











void e_chan_in_DecompData( MESH  *mesh, CHAN *chan)

{
int msglen;

msglen = e_chan_in_Buffer(&(mesh->Decomp.NSubDoms),chan);
if(msglen != sizeof(int))
    e_GeneralError("wrong size received");

if( (mesh->Decomp.SubDoms = e_alloc_i_vect(mesh->TotalNElems)) == NULL)
    e_GeneralError("Not enough memory for Decomp.SubDoms");
msglen = e_chan_in_Buffer(mesh->Decomp.SubDoms,chan);
if(msglen != mesh->TotalNElems * sizeof(int))
    e_GeneralError("wrong size received");

mesh->SubStructSet[E_SS_NUM_DECOMP] = TRUE;

if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
    mesh->Decomp.SubDomsTriang1 = mesh->Decomp.SubDoms + mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] - 1;
else
    mesh->Decomp.SubDomsTriang1 = NULL;
if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
    mesh->Decomp.SubDomsTriang2 = mesh->Decomp.SubDoms + mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2] - 1;
else
    mesh->Decomp.SubDomsTriang2 = NULL;
if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
    mesh->Decomp.SubDomsQuad1 = mesh->Decomp.SubDoms + mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1] - 1;
else
    mesh->Decomp.SubDomsQuad1 = NULL;

}   /*  end of function e_chan_in_DecompData()  */














/*


void e_chan_out_MeshParams(MESH  *mesh, CHAN *chan)

{
if(mesh->SubStructSet[E_SS_NUM_MESHPARAM])
    {
    e_chan_out_Buffer(mesh->TotalNElems * sizeof(double),mesh->MeshParam.MeshParams,chan);
    }

}
*/
 /* end of function e_chan_out_MeshParams()  */








/*


void e_chan_in_MeshParams( MESH  *mesh, CHAN *chan)

{
int msglen;

if( (mesh->MeshParam.MeshParams = e_alloc_d_vect(mesh->TotalNElems)) == NULL)
    e_GeneralError("Not enough memory for MeshParameters");
msglen = e_chan_in_Buffer(mesh->MeshParam.MeshParams,chan);
if(msglen != mesh->TotalNElems * sizeof(double))
    e_GeneralError("wrong size received");

mesh->SubStructSet[E_SS_NUM_MESHPARAM] = TRUE;

if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
    mesh->MeshParam.MeshParamsTriang1 = mesh->MeshParam.MeshParams + mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] - 1;
else
    mesh->MeshParam.MeshParamsTriang1 = NULL;
if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
    mesh->MeshParam.MeshParamsTriang2 = mesh->MeshParam.MeshParams + mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2] - 1;
else
    mesh->MeshParam.MeshParamsTriang2 = NULL;
if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
    mesh->MeshParam.MeshParamsQuad1 = mesh->MeshParam.MeshParams + mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1] - 1;
else
    mesh->MeshParam.MeshParamsQuad1 = NULL;

}

*/
  /*  end of function e_chan_in_MeshParams()  */








void e_chan_out_Mats(MESH *mesh, CHAN *chan)

{
register  i,j;
int      *vect_ind;
int       str_size;

if(mesh->SubStructSet[E_SS_NUM_MAT_PROPS])
    {
    if(mesh->KeywordSet[E_KW_NUM_NMATS])
        {
        e_chan_out_Buffer(mesh->NMats * sizeof(MAT),mesh->Props,chan);

        for(i = 0;i < mesh->NMats;i++)
            {
            str_size = strlen(mesh->Props[i].Name);
            e_chan_out_Buffer(sizeof(int),&str_size,chan);
            e_chan_out_Buffer((str_size+1)*sizeof(char),mesh->Props[i].Name,chan);
            }
        }
    if(mesh->KeywordSet[E_KW_NUM_NCOMP_MATS_TYPE1])
        {
        e_chan_out_Buffer(mesh->NCompMatsType1 * sizeof(COMPMAT1),mesh->CompPropsType1,chan);

        for(i = 0;i < mesh->NCompMatsType1;i++)
            {
            e_chan_out_Buffer(mesh->CompPropsType1[i].NLayers * sizeof(double),mesh->CompPropsType1[i].Thicknesses,chan);

            if( (vect_ind = e_alloc_i_vect(mesh->CompPropsType1[i].NLayers)) == NULL)
                e_GeneralError("Not enough memory for Vect_ind");

            for(j = 0;j < mesh->CompPropsType1[i].NLayers;j++)
                vect_ind[j] = (mesh->CompPropsType1[i].LayerMats[j]) - (mesh->Props);
            e_chan_out_Buffer(mesh->CompPropsType1[i].NLayers * sizeof(int),vect_ind,chan);
            e_free(vect_ind);

            str_size = strlen(mesh->CompPropsType1[i].Name);
            e_chan_out_Buffer(sizeof(int),&str_size,chan);
            e_chan_out_Buffer((str_size+1)*sizeof(char),mesh->CompPropsType1[i].Name,chan);
            }
        }
    }
}   /* end of function e_chan_out_Mats()  */







void e_chan_in_Mats(MESH *mesh, CHAN *chan)

{
register  i,j;
int      *vect_ind;
int       msglen;
int       str_size;

if(mesh->SubStructSet[E_SS_NUM_MAT_PROPS])
    {
    if(mesh->KeywordSet[E_KW_NUM_NMATS])
        {
        if( (mesh->Props = e_calloc(sizeof(MAT),mesh->NMats)) == NULL)
            e_GeneralError("Not enough memory for Props");
        msglen = e_chan_in_Buffer(mesh->Props,chan);
        if(msglen != mesh->NMats * sizeof(MAT))
            e_GeneralError("wrong size received");

        for(i = 0;i < mesh->NMats;i++)
            {
            msglen = e_chan_in_Buffer(&str_size,chan);
            if(msglen != sizeof(int))
                e_GeneralError("wrong size received");

            if( (mesh->Props[i].Name = e_alloc_c_vect(str_size+1)) == NULL)
                e_GeneralError("Not enough memory for mesh MatNames");
            msglen = e_chan_in_Buffer(mesh->Props[i].Name,chan);
            if(msglen != (str_size+1)*sizeof(char))
                e_GeneralError("wrong size received");
            }
        }
    if(mesh->KeywordSet[E_KW_NUM_NCOMP_MATS_TYPE1])
        {
        if( (mesh->CompPropsType1 = e_calloc(sizeof(COMPMAT1),mesh->NCompMatsType1)) == NULL)
            e_GeneralError("Not enough memory for CompPropsType1");
        msglen = e_chan_in_Buffer(mesh->CompPropsType1,chan);
        if(msglen != mesh->NCompMatsType1 * sizeof(COMPMAT1))
            e_GeneralError("wrong size received");

        for(i = 0;i < mesh->NCompMatsType1;i++)
            {
            if( (mesh->CompPropsType1[i].Thicknesses = e_alloc_d_vect(mesh->CompPropsType1[i].NLayers)) == NULL)
                e_GeneralError("Not enough memory for CompPropsType1.Thicknesses");
            msglen = e_chan_in_Buffer(mesh->CompPropsType1[i].Thicknesses,chan);
            if(msglen != mesh->CompPropsType1[i].NLayers * sizeof(double))
                e_GeneralError("wrong size received");

            if( (vect_ind = e_alloc_i_vect(mesh->CompPropsType1[i].NLayers)) == NULL)
                e_GeneralError("Not enough memory for Vect_ind");
            if( (mesh->CompPropsType1[i].LayerMats = e_calloc(sizeof(COMPMAT1),mesh->CompPropsType1[i].NLayers)) == NULL)
                e_GeneralError("Not enough memory for CompPropsType1.LayerMats");

            msglen = e_chan_in_Buffer(vect_ind,chan);
            if(msglen != mesh->CompPropsType1[i].NLayers * sizeof(int))
                e_GeneralError("wrong size received");

            for(j = 0;j < mesh->CompPropsType1[i].NLayers;j++)
                mesh->CompPropsType1[i].LayerMats[j] = mesh->Props + vect_ind[j];
            e_free(vect_ind);

            msglen = e_chan_in_Buffer(&str_size,chan);
            if(msglen != sizeof(int))
                e_GeneralError("wrong size received");

            if( (mesh->CompPropsType1[i].Name = e_alloc_c_vect(str_size+1)) == NULL)
                e_GeneralError("Not enough memory for mesh CompMatNamesType1");
            msglen = e_chan_in_Buffer(mesh->CompPropsType1[i].Name,chan);
            if(msglen != (str_size+1)*sizeof(char))
                e_GeneralError("wrong size received");
            }
        }
    }

}   /* end of function e_chan_in_Mats()  */











            /*   This value is the maximum lenght of the buffer
                 which could be handled by the router tasks
                 written by Mr. Asad Khan from Pakistan
                 (but he was in reality a honest Afgan!!!)
            */
#define ROUTER_MAX_BYTE  1999
            /* on the top of it, when he sais 10kb
               it might just work with 2kb....   what can we say?
            */

static char S_msg_buf[ROUTER_MAX_BYTE];




void e_chan_out_Buffer(int length, void *pointer, CHAN *channel)

{
register i;
int      nof;                 /* Num. of full (MAX_BYTE) length messages to be sent  */
int      rem;                 /* the remainder in the string to be sent  */
int      cur_length;          /* the current message length  */
char    *ptr = pointer;

nof = length / (ROUTER_MAX_BYTE-1);
rem = length % (ROUTER_MAX_BYTE-1);

for(i = 0; i <= nof; i++)
    {
    cur_length = (i == nof ? rem : (ROUTER_MAX_BYTE-1) );
    memcpy(S_msg_buf,ptr,cur_length);
    *(S_msg_buf + cur_length) = (char)(i == nof ? FALSE : TRUE );
    chan_out_word(cur_length+1,channel);
    chan_out_message(cur_length+1,S_msg_buf,channel);
    ptr += cur_length;
    }
}



int e_chan_in_Buffer(void *pointer, CHAN *channel)

{
int           cur_length_pp;
int           counter = 0;
char    *ptr = pointer;

do
    {
    chan_in_word(&cur_length_pp,channel);
    chan_in_message(cur_length_pp,S_msg_buf,channel);
    counter += cur_length_pp-1;
    memcpy(ptr,S_msg_buf,cur_length_pp-1);
    ptr += cur_length_pp-1;
    }
while( *(S_msg_buf+cur_length_pp-1) == (char)TRUE );

return(counter);
}








/*
void e_chan_out_Buffer(int hossz, void *ptr, CHAN *channel)

{
int db, maradek, k_hossz, i;
char tmp;
char *pointer=ptr;

db = hossz / MAX_BYTE;
maradek = hossz % MAX_BYTE;
for(i = 0; i <= db; i++)
    {
    k_hossz = (i == db ? maradek : MAX_BYTE);
    tmp = *(pointer+k_hossz);
    *(pointer+k_hossz) = (i == db ? FALSE : TRUE );
    chan_out_word(k_hossz+1, channel);
    chan_out_message(k_hossz+1, pointer, channel);
    *(pointer+k_hossz) = tmp;
    pointer += k_hossz;
    }
}


int e_chan_in_Buffer(void *ptr, CHAN *channel)

{
int           hossz;
char          tmp;
int           szamlalo = 0;
char          *pointer = ptr;

do
    {
    chan_in_word(&hossz, channel);
    tmp = *(pointer+hossz-1);
    chan_in_message(hossz, pointer, channel);
    szamlalo += hossz-1;
    pointer += hossz-1;
    }
while(*pointer);
*pointer = tmp;

return(szamlalo);
}

*/



/*  End Of File: TRAN_AUX.C   */
