#include "e_lib.h"


#define dist(x, y) \
  sqrt( ((x)[0] - (y)[0]) * ((x)[0] - (y)[0]) + \
        ((x)[1] - (y)[1]) * ((x)[1] - (y)[1]) + \
        ((x)[2] - (y)[2]) * ((x)[2] - (y)[2]) )

int main(int argc, char **argv)
{
  double a[3] = {1, 0, 0};
  double b[3] = {0, 1, 0};
  double ma, mb;
  int    n, i, mode, type;
  double **coord, *mpr;
  double dd;

  if(argc < 11)
  {
    printf("\nUsage: segtest x1 y1 z1 m1 x2 y2 z2 m2 mode type \n");
    printf("\nwhere:\n");
    printf("\t\tx1, y1, z1 - coordinates of the starting point\n");
    printf("\t\tm1         - mesh parameters of the starting point\n");
    printf("\t\tx2, y2, z2 - coordinates of the end point\n");
    printf("\t\tm2         - mesh parameters of the end point\n");
    printf("\t\tmode       - 0 -> NONE\n");
    printf("\t\t             1 -> ODD\n");
    printf("\t\t             2 -> EVEN\n");
    printf("\t\ttype       - 0 -> line\n");
    printf("\t\t             1 -> NURBS\n");
    printf("\nOptionally:\n");
    printf("\n\n");

    return(1);
  }

  for(i = 0; i < 3; i++)
  {
    if(sscanf(argv[1+i], "%lf", &(a[i])) != 1)
    {
      e_GeneralError("Invalid coordinate for starting point");
    }
  }
  if(sscanf(argv[4], "%lf", &ma) != 1)
  {
    e_GeneralError("Invalid mesh parameter for starting point");
  }
  for(i = 0; i < 3; i++)
  {
    if(sscanf(argv[5+i], "%lf", &(b[i])) != 1)
    {
      e_GeneralError("Invalid coordinate for end point");
    }
  }
  if(sscanf(argv[8], "%lf", &mb) != 1)
  {
    e_GeneralError("Invalid mesh parameter for end point");
  }
  /* mode */
  if(sscanf(argv[9], "%d", &mode) != 1)
  {
    e_GeneralError("Invalid mode");
  }
  switch(mode)
  {
    case 0:
        mode = E_SEGREF_NONE;
        break;
    case 1:
        mode = E_SEGREF_ODD;
        break;
    case 2:
        mode = E_SEGREF_EVEN;
        break;
    default:
        e_GeneralError("Invalid mode");
  }

  /* type of segment */
  if(sscanf(argv[10], "%d", &type) != 1)
  {
    e_GeneralError("Invalid type");
  }
  switch(type)
  {
    case 0:
    {
      n = e_segment_refine(a, ma, b, mb, &coord, &mpr, 1, mode);
      break;
    }
    case 1:
    {
      e_GeneralError("Not implemented");
      break;
    }
  }
  
  printf("\n");
  for(i = 0; i < n; i++)
  {
    printf("%d %f %f %f -- mpr: %f",
           i+1, coord[i][0], coord[i][1], coord[i][2], mpr[i]);
    if(i > 0 && i < n-1)
    {
      dd = dist(coord[i-1], coord[i]);
      printf(" -- d1: %f", dd);
      dd = dist(coord[i], coord[i+1]);
      printf(" -- d2: %f", dd);
    }
    printf("\n");
  }
  printf("\n\n");
  
  return 0;
}
