dnl @synopsis PETI_PEDANTIC_GCC
dnl
dnl For development purposes, it is desirable to have autoconf
dnl automatically enable warnings when compiling C or C++ sources. In
dnl case the underlying compiler is a gcc, the appropriate flags are
dnl "-Wall -pedantic". This macro will add them to $CFLAGS and
dnl $CXXFLAGS if $CC is found to be a gcc.
dnl
dnl @author Peter Simons <simons@computer.org>
dnl @version $Id: eg_config.m4,v 1.1 2002/08/06 11:23:22 romek Exp $
AC_DEFUN([PETI_PEDANTIC_GCC],
    [
    if test "$GCC" = yes; then
        CFLAGS="$CFLAGS -Wall -pedantic"
        CXXFLAGS="$CXXFLAGS -Wall -pedantic"
        fi
    ])


dnl @synopsis EG_GCOV_GCC
dnl Enable use of gcov: a Test Coverage Program in case
dnl the underlying compile is gcc by setting the flags
dnl "-fprofile-arcs -ftest-coverage". This macro will add
dnl them to $CFLAGS and $CXXFLAGS if $CC is found to be a gcc.
dnl
dnl Caution: This macro removes and optimization flag form
dnl          $CFLAGS and $CXXFLAGS
dnl
dnl Modeled after PETI_PEDANTIC_GCC by Peter Simons.
dnl @author Roman Putanowicz <putanowr@twins.pk.edu.pl>
dnl @version $Id: eg_config.m4,v 1.1 2002/08/06 11:23:22 romek Exp $ 
AC_DEFUN([EG_GCOV_GCC],
    [[
    if test "$GCC" = yes; then
        eg_gcov_cflags="-g -fprofile-arcs -ftest-coverage"
        for i in $CFLAGS
        do
          case $i in
            -O | -O? | -g) ;;
             *) eg_gcov_cflags="$eg_gcov_cflags $i";;
          esac
        done 
        CFLAGS="$eg_gcov_cflags"
        eg_gcov_cflags="-g -fprofile-arcs -ftest-coverage"
        for i in $CXXFLAGS
        do
          case $i in
            -O | -O? | -g) ;;
             *) eg_gcov_cflags="$eg_gcov_cflags $i";;
          esac
        done 
        CXXFLAGS="$eg_gcov_cflags"
    fi
    ]])

dnl @synopsis EG_MAINTANENCE_OPTIONS
dnl Adds new section header to the default help message
dnl produced by `configure --help'
dnl
dnl @author Roman Putanowicz <putanowr@twins.pk.edu.pl>
AC_DEFUN([EG_MAINTANENCE_OPTIONS],
    [
    AC_ARG_ENABLE([],[
eg-Lib Maintanence Options:],[],[])
    ])



dnl @synopsis EG_FULLWARN_GCC
dnl Enable full set of gcc warnings.
dnl
dnl @author Roman Putanowicz <putanowr@twins.pk.edu.pl>
dnl @version $Id: eg_config.m4,v 1.1 2002/08/06 11:23:22 romek Exp $ 
AC_DEFUN([EG_FULLWARN_GCC],
    [[
    if test "$GCC" = yes; then
        eg_fullwarn_cflags="-ansi -pedantic -W -Wall -Wtraditional -Wconversion -Wshadow -Wpointer-arith -Wcast-qual -Wcast-align -Wwrite-strings -Wstrict-prototypes -fshort-enums -fno-common -Wmissing-prototypes -Wnested-externs -Dinline= -g"
        temp_flags=$eg_fullwarn_cflags
        for i in $CFLAGS
        do
          case $i in
            -g | -Wall | -ansi | -pedantic) ;;
             *) temp_flags="$temp_flags $i";;
          esac
        done 
        CFLAGS="$temp_flags"
        temp_flags=$eg_fullwarn_cflags
        for i in $CXXFLAGS
        do
          case $i in
            -g | -Wall | -ansi | -pedantic) ;;
             *) temp_flags="$temp_flag $i";;
          esac
        done 
        CXXFLAGS="$temp_flags"
    fi
    ]])


dnl @synopsis EG_ARG_ENABLE_GCOV_GCC
dnl Provide configure option for enabling gcov support
dnl in caset the compiler is gcc
dnl
dnl @author Roman Putanowicz <putanowr@twins.pk.edu.pl>
AC_DEFUN(EG_ARG_ENABLE_GCOV_GCC,
     [AC_ARG_ENABLE(gcov-gcc, 
                   AC_HELP_STRING([--enable-gcov-gcc],
                                  [enable gcov support for gcc (default=no)]),
                   EG_GCOV_GCC, [])
     ]) 

dnl @synopsis EG_ARG_ENABLE_FULLWARN_GCC
dnl Provide configure option for enabling full
dnl set of gcc warnings 
dnl
dnl @author Roman Putanowicz <putanowr@twins.pk.edu.pl>
AC_DEFUN(EG_ARG_ENABLE_FULLWARN_GCC,
     [AC_ARG_ENABLE(fullwarn-gcc, 
                   AC_HELP_STRING([--enable-fullwarn-gcc],
                                  [enable full set of warnings support for gcc (default=no)]),
                   EG_FULLWARN_GCC, [])
     ]) 


