/********************************************************************
 eg_error.c  --
     Error handling routines for Eg-Lib. Eg-Lib was using E-Lib 
     handling routines and the one bellow are derived form E-Lib
     ones.

 Jelle Muylle <jelle@sect.mce.hw.ac.uk>
 Roman Putanowicz <romek@sect.mce.hw.ac.uk>
 SECT Research Group, Heriot-Watt University, Edinburgh    
 ********************************************************************/

/* $Id: eg_error.c,v 1.1 2002/10/16 19:05:03 peteri Exp $ */
#include <stdio.h>
#include <stdlib.h>
#include "eg_lib.h"

static void ErrorWarningMessage(char *message);
static void eg_exit(int ExitStatus);

static void ErrorWarningMessage(char *message)
{
#if defined(__E_XWIN__)
   /* To Do - Put X-Windows code here */
   fprintf(stderr,message);
#elif defined(__E_MSWIN__)
   MessageBox(NULL,message,"E-Lib Message",
              MB_OK | MB_ICONEXCLAMATION | MB_TASKMODAL);
#else
   fprintf(stderr,message);
#endif
}

void eg_GeneralWarning(char *Warning, ...)
{
   char message[EG_LINE_MAX + 1];
   char prefix[] = "WARNING: ";

#ifdef __E_CLANSIC__
   va_list ap;
   strcpy(message, "\n");
   strcat(message, prefix);
   va_start(ap, Warning);
#ifdef __SYSTEM_IRIX__
   vsprintf (message+sizeof(prefix), Warning, ap);
#else
   vsnprintf (message+sizeof(prefix), EG_LINE_MAX-1, Warning, ap);
#endif
   va_end(ap);
   strcat(message, "\n");
#endif

#ifdef __E_XWIN__
   va_list ap;
   strcpy(message, "\n");
   strcat(message, prefix);
   va_start(ap, Warning);
#ifdef __SYSTEM_IRIX__
   vsprintf (message+sizeof(prefix), Warning, ap);
#else
   vsnprintf (message+sizeof(prefix), EG_LINE_MAX-1, Warning, ap);
#endif
   va_end(ap);
   strcat(message, "\n");
#endif

#ifdef __E_MSWIN__
   va_list ap;
   strcat(message, prefix);
   va_start(ap, Warning);
   vsnprintf (message+sizeof(prefix)-1, EG_LINE_MAX, Warning, ap);
   va_end(ap);
#endif
   ErrorWarningMessage(message);
}

void eg_GeneralError(char *ErrorMessage, ...)
{
   char message[EG_LINE_MAX + 1];
   char prefix[] = "ERROR: ";

#ifdef __E_CLANSIC__
   va_list ap;
   strcpy(message, "\n");
   strcat(message, prefix);
   va_start(ap, ErrorMessage);
#ifdef __SYSTEM_IRIX__
   vsprintf (message+sizeof(prefix), ErrorMessage, ap);
#else
   vsnprintf (message+sizeof(prefix), EG_LINE_MAX-1, ErrorMessage, ap);
#endif
   va_end(ap);
   strcat(message, "\n");
#endif

#ifdef __E_XWIN__
   va_list ap;
   strcpy(message, "\n");
   strcat(message, prefix);
   va_start(ap, ErrorMessage);
#ifdef __SYSTEM_IRIX__
   vsprintf (message+sizeof(prefix), ErrorMessage, ap);
#else
   vsnprintf (message+sizeof(prefix), EG_LINE_MAX-1, ErrorMessage, ap);
#endif
   va_end(ap);
   strcat(message, "\n");
#endif

#ifdef __E_MSWIN__
   va_list ap;
   strcat(message, prefix);
   va_start(ap, ErrorMessage);
   vsnprintf (message+sizeof(prefix)-1, EG_LINE_MAX, ErrorMessage, ap);
   va_end(ap);
#endif

   ErrorWarningMessage(message);

#ifndef __E_XWIN__
#ifndef __E_MSWIN__
   eg_exit(1);
#endif
#endif
}

static void eg_exit(int ExitStatus)
{

#if defined(__E_MSWIN__)
   char message[EG_LINE_MAX + 1];
#elif defined(__E_CLANSIC__)
#  ifdef __E_PVM__
      pvm_exit();
#  endif
   exit(ExitStatus);
#elif defined(__E_XWIN__)
   /* To Do - Put X-Windows code here */
#elif defined(__E_MSWIN__)
   sprintf(message,"The function 'exit(%d)' has been called.\nTerminate the program.",ExitStatus);
   ErrorWarningMessage(message);
#else
   exit(ExitStatus);
#endif
}

/* Modifications:
 * $Log: eg_error.c,v $
 * Revision 1.1  2002/10/16 19:05:03  peteri
 * - dependence on e_lib is removed
 *
 * Revision 1.1  2002/10/16 09:28:11  putanowr
 * Dependency on E-Lib removed. Missing functions added in eg_error.c and eg_math.c
 *
 */

