#include "eps.h"


/* ------------------------------------------------------------------------- */
/* Drawing functions and parameters */


/* colors */
#define ERROR_COLOR_R  1.0
#define ERROR_COLOR_G  0.0
#define ERROR_COLOR_B  0.0

             /* GREEN     */
#define STE_XX_POS_R_MAX 0.00
#define STE_XX_POS_G_MAX 1.00
#define STE_XX_POS_B_MAX 0.35
              /* RED       */
#define STE_XX_NEG_R_MAX 1.00
#define STE_XX_NEG_G_MAX 0.00
#define STE_XX_NEG_B_MAX 0.42

              /* PURPLE    */
#define STE_YY_POS_R_MAX 0.71
#define STE_YY_POS_G_MAX 0.00
#define STE_YY_POS_B_MAX 1.00

              /* YELLOW    */

#define STE_YY_NEG_R_MAX 1.00
#define STE_YY_NEG_G_MAX 1.00
#define STE_YY_NEG_B_MAX 0.00

              /* CYAN      */
#define STE_XY_POS_R_MAX 0.00
#define STE_XY_POS_G_MAX 0.83
#define STE_XY_POS_B_MAX 1.00

              /* ORANGE    */
#define STE_XY_NEG_R_MAX 1.00
#define STE_XY_NEG_G_MAX 0.39
#define STE_XY_NEG_B_MAX 0.00

/* color scheme */
float   colorscheme[NO_COLOR][3] = {{0.32, 1.0,  0.03},
                                    {1.0,  0.32, 0.03},
                                    {0.03, 0.32, 1.0},
                                    {0.03, 1.0,  1.0},
                                    {1.0,  0.68, 0.03},
                                    {1.0,  0.03, 0.68},
                                    {0.55, 0.48, 0.32},
                                    {0.35, 0.35, 0.35}  };
/* greyscale scheme */
/* this could be simple shading as well instead of RGB colors */
float   greyscheme[NO_COLOR][3] = {{1.0,  1.0,  1.0},
                                   {0.67, 0.67, 0.67},
                                   {0.87, 0.87, 0.87},
                                   {0.40, 0.40, 0.40},
                                   {0.94, 0.94, 0.94},
                                   {0.78, 0.78, 0.78},
                                   {0.54, 0.54, 0.54},
                                   {0.23, 0.23, 0.23}  };


static double scale_color(double value, double correction);

/* ############################################################################ */

static double scale_color(double value, double correction)
{
  value = fabs(value);
  return(value + correction * (sqrt(value) - value));
}


void draw_simple(POLYGON polygon, OPTIONS options)
{
  int i_p, i;
  double centre[2];

  /* tiling */
  if(options.tile)
  {
    centre[0] = 0.0;
    centre[1] = 0.0;
    for(i = 0; i < polygon.no_vertices; i++)
    {
      centre[0] += polygon.screen_coords[i][0];
      centre[1] += polygon.screen_coords[i][1];
    }
    centre[0] /= polygon.no_vertices;
    centre[1] /= polygon.no_vertices;

    fprintf(options.fp, "\n%.3f %.3f m", 
            (centre[0] + (polygon.screen_coords[0][0] - centre[0]) * options.percent / 100),
            options.picture_size_y - 
            (centre[1] + (polygon.screen_coords[0][1] - centre[1]) * options.percent / 100)  );

    for(i_p = 1; i_p < polygon.no_vertices; i_p++)
      fprintf(options.fp, " %.3f %.3f l", 
              (centre[0] + (polygon.screen_coords[i_p][0] - centre[0]) * options.percent / 100),
              options.picture_size_y - 
              (centre[1] + (polygon.screen_coords[i_p][1] - centre[1]) * options.percent / 100)  );
    if(polygon.no_vertices > 2)
    {
      fprintf(options.fp, " %.3f %.3f l", 
              (centre[0] + (polygon.screen_coords[0][0] - centre[0]) * options.percent / 100),
              options.picture_size_y - 
              (centre[1] + (polygon.screen_coords[0][1] - centre[1]) * options.percent / 100)  );
    }
  }
  /* no tiling */
  else
  {
    fprintf(options.fp, "\n%.3f %.3f m", 
            polygon.screen_coords[0][0],
            options.picture_size_y-polygon.screen_coords[0][1]  );
    for(i_p = 1; i_p < polygon.no_vertices; i_p++)
      fprintf(options.fp, " %.3f %.3f l",
              polygon.screen_coords[i_p][0],
              options.picture_size_y-polygon.screen_coords[i_p][1]  );
    if(polygon.no_vertices > 2)
    {
      fprintf(options.fp, " %.3f %.3f l", 
              polygon.screen_coords[0][0],
              options.picture_size_y-polygon.screen_coords[0][1]  );
    }

  }
}


int draw_polygon(MESH *mesh, POLYGON polygon, OPTIONS options, int error)
{
  double pos_r_max, pos_g_max, pos_b_max;
  double neg_r_max, neg_g_max, neg_b_max;
  double str, str_color;
  double feerror, errcolor;
  int    dom, index;
  static int prev_dom = -1;

  /* determine the global index of the element */
  index = mesh->FirstNetIndex[polygon.elem_type]-1+polygon.mesh_index;

  /* subdomain color setup */
  if(   (options.type == EPS_SUBDOMAIN)
     && (options.subdom_style & SUBDOM_COLOR) )
  {
    dom = (mesh->Decomp.SubDoms[index]-1) % NO_COLOR;
    if((dom != prev_dom) || options.tile || (options.hidden_line == 1) )
    {
      /* color */
      if(options.color)
        fprintf(options.fp, "\n%.3f %.3f %.3f c",
                colorscheme[dom][0],
                colorscheme[dom][1],
                colorscheme[dom][2]  );
      /* grey scale */
      else
        fprintf(options.fp, "\n%.3f %.3f %.3f c",
                greyscheme[dom][0],
                greyscheme[dom][1],
                greyscheme[dom][2]  );

      prev_dom = dom;
    }
  }
  /* setup what kind of color will be used for the error plot */
  else if(options.type == EPS_ERROR)
  {
    feerror = mesh->FEError.FEErrors[index];
    /* color error plot */
    if(options.color)
    {
      /* anything below the specified error will be colored proportionally */
      if(feerror < options.max_error)
      {
        errcolor = scale_color(feerror / options.max_error, 
                               options.color_correction);
        fprintf(options.fp, "\n%.3f %.3f %.3f c",
                ERROR_COLOR_R * errcolor,
                ERROR_COLOR_G * errcolor,
                ERROR_COLOR_B * errcolor  );
      }
      /* above the specified error there is only one color */
      else
      {
        fprintf(options.fp, "\n%.3f %.3f %.3f c",
                ERROR_COLOR_R,
                ERROR_COLOR_G,
                ERROR_COLOR_B  );
      }
    }
    /* greyscale error plot */
    else
    {
      /* scales from black to white */
      if(feerror < options.max_error)
      {
        errcolor = scale_color(feerror / options.max_error,
                               options.color_correction);
        fprintf(options.fp, "\n%.3f %.3f %.3f c",
                errcolor,
                errcolor,
                errcolor  );
      }
      else
      {
        /* white if above limit */
        fprintf(options.fp, "\n1.0 1.0 1.0 c");
      }
    }
  }
  /* stress color setup */
  else if(options.type == EPS_STRESS)
  {
    if(options.color)
    {
      /* determine the color to use */
      switch(options.stress_type)
      {
        case 1:
          pos_r_max = STE_XX_POS_R_MAX;
          pos_g_max = STE_XX_POS_G_MAX;
          pos_b_max = STE_XX_POS_B_MAX;
          neg_r_max = STE_XX_NEG_R_MAX;
          neg_g_max = STE_XX_NEG_G_MAX;
          neg_b_max = STE_XX_NEG_B_MAX;
          break;
        case 2:
          pos_r_max = STE_YY_POS_R_MAX;
          pos_g_max = STE_YY_POS_G_MAX;
          pos_b_max = STE_YY_POS_B_MAX;
          neg_r_max = STE_YY_NEG_R_MAX;
          neg_g_max = STE_YY_NEG_G_MAX;
          neg_b_max = STE_YY_NEG_B_MAX;
          break;
        case 3:
          pos_r_max = STE_XX_POS_R_MAX;
          pos_g_max = STE_XX_POS_G_MAX;
          pos_b_max = STE_XX_POS_B_MAX;
          neg_r_max = STE_XX_NEG_R_MAX;
          neg_g_max = STE_XX_NEG_G_MAX;
          neg_b_max = STE_XX_NEG_B_MAX;
          break;
        case 4:
          pos_r_max = STE_XY_POS_R_MAX;
          pos_g_max = STE_XY_POS_G_MAX;
          pos_b_max = STE_XY_POS_B_MAX;
          neg_r_max = STE_XY_NEG_R_MAX;
          neg_g_max = STE_XY_NEG_G_MAX;
          neg_b_max = STE_XY_NEG_B_MAX;
          break;
        case 5:
          pos_r_max = STE_XY_POS_R_MAX;
          pos_g_max = STE_XY_POS_G_MAX;
          pos_b_max = STE_XY_POS_B_MAX;
          neg_r_max = STE_XY_NEG_R_MAX;
          neg_g_max = STE_XY_NEG_G_MAX;
          neg_b_max = STE_XY_NEG_B_MAX;
          break;
        case 6:
          pos_r_max = STE_XY_POS_R_MAX;
          pos_g_max = STE_XY_POS_G_MAX;
          pos_b_max = STE_XY_POS_B_MAX;
          neg_r_max = STE_XY_NEG_R_MAX;
          neg_g_max = STE_XY_NEG_G_MAX;
          neg_b_max = STE_XY_NEG_B_MAX;
          break;
        default:
          if(error)
          {
            e_GeneralError("Invalid stress type !");
          }
          else
            return 0;
      }

      if(polygon.elem_type == E_ELEM_TYPE_LINK1)
      {
        str = mesh->Stresses.StressesLink1[index]
                                          [options.stress_point-1]
                                          [options.stress_type-1];
      }
      else if(polygon.elem_type == E_ELEM_TYPE_LINK2)
      {
        str = mesh->Stresses.StressesLink2[index]
                                          [options.stress_point-1]
                                          [options.stress_type-1];
      }
      else if(polygon.elem_type == E_ELEM_TYPE_LINK3)
      {
        str = mesh->Stresses.StressesLink3[index]
                                          [options.stress_point-1]
                                          [options.stress_type-1];
      }
      else if(polygon.elem_type == E_ELEM_TYPE_LINK4)
      {
        str = mesh->Stresses.StressesLink4[index]
                                          [options.stress_point-1]
                                          [options.stress_type-1];
      }
      else if(polygon.elem_type == E_ELEM_TYPE_LINK5)
      {
        str = mesh->Stresses.StressesLink5[index]
                                          [options.stress_point-1]
                                          [options.stress_type-1];
      }
      else if(polygon.elem_type == E_ELEM_TYPE_TRIANG1)
      {
        str = mesh->Stresses.StressesTriang1[index]
                                            [options.stress_point-1]
                                            [options.stress_type-1];
      }
      else if(polygon.elem_type == E_ELEM_TYPE_TRIANG2)
      {
        str = mesh->Stresses.StressesTriang2[index]
                                            [options.stress_layer-1]
                                            [options.stress_point-1]
                                            [options.stress_type-1];
      }
      else if(polygon.elem_type == E_ELEM_TYPE_TRIANG3)
      {
        str = mesh->Stresses.StressesTriang3[index]
                                            [options.stress_point-1]
                                            [options.stress_type-1];
      }
      else if(polygon.elem_type == E_ELEM_TYPE_TRIANG4)
      {
        str = mesh->Stresses.StressesTriang4[index]
                                            [options.stress_point-1]
                                            [options.stress_type-1];
      }
      else if(polygon.elem_type == E_ELEM_TYPE_TRIANG5)
      {
        str = mesh->Stresses.StressesTriang5[index]
                                            [options.stress_point-1]
                                            [options.stress_type-1];
      }
      else if(polygon.elem_type == E_ELEM_TYPE_QUAD1)
      {
        str = mesh->Stresses.StressesQuad1[index]
                                          [options.stress_point-1]
                                          [options.stress_type-1];
      }
      else if(polygon.elem_type == E_ELEM_TYPE_QUAD3)
      {
        str = mesh->Stresses.StressesQuad3[index]
                                          [options.stress_layer-1]
                                          [options.stress_point-1]
                                          [options.stress_type-1];
      }
      else if(polygon.elem_type == E_ELEM_TYPE_TETRAH1)
      {
        str = mesh->Stresses.StressesTetrah1[index]
                                            [options.stress_point-1]
                                            [options.stress_type-1];
      }
      else if(polygon.elem_type == E_ELEM_TYPE_BLOCK1)
      {
        str = mesh->Stresses.StressesBlock1[index]
                                           [options.stress_point-1]
                                           [options.stress_type-1];
      }

      str /= (options.max_total_stress * options.max_stress_intensity / 100.00);

      if(str >= 0.00)
      {
        if(str <= 1.0)
        {
          str_color = scale_color(str, options.color_correction);
          fprintf(options.fp, "\n%.3f %.3f %.3f c",
                  pos_r_max * str_color,
                  pos_g_max * str_color,
                  pos_b_max * str_color  );
        }
        else
        {
          fprintf(options.fp, "\n%.3f %.3f %.3f c",
                  pos_r_max,
                  pos_g_max,
                  pos_b_max  );
        }
      }
      else
      {
        if(str >= -1.0)
        {
          str_color = scale_color(fabs(str), options.color_correction);
          fprintf(options.fp, "\n%.3f %.3f %.3f c",
                  neg_r_max * str_color,
                  neg_g_max * str_color,
                  neg_b_max * str_color  );
        }
        else
        {
          fprintf(options.fp, "\n%.3f %.3f %.3f c",
                  neg_r_max,
                  neg_g_max,
                  neg_b_max  );
        }
      }
    }
    /* grey scale */
    else
    {
      if(error)
      {
        e_GeneralError("Grey scale stress plot is not supported !");
      }
      else
        return 0;
    }
  }

  /* draw out a color filled element */
  if(   (options.type == EPS_ERROR)
     || (options.type == EPS_STRESS)
     || (   (options.type == EPS_SUBDOMAIN)
         && (options.subdom_style & SUBDOM_COLOR) ) )
  {
    draw_simple(polygon, options);
    fprintf(options.fp, " p");

    /* reset to the boundary color, but only 
     * if the user asked for and hidden line is on, otherwise
     * it is done later 
     */
    if(options.border && options.hidden_line)
    {
      if(options.bcolor)
        fprintf(options.fp, "\nb"); /* black */
      else
        fprintf(options.fp, "\nw"); /* white */

      draw_simple(polygon, options);
      fprintf(options.fp, " s");

    }
  }
  /* hidden line 'normal' drawing */
  else if(   (options.type == EPS_PLAINMESH)
          && (options.plot_type == 3)
          && (options.hidden_line == 1) )
  {
    /* white filled element */
    fprintf(options.fp, "\nw");
    draw_simple(polygon, options);
    fprintf(options.fp, " p");
    /* black border */
    fprintf(options.fp, "\nb");
    draw_simple(polygon, options);
    fprintf(options.fp, " s");

  }

  return 1;
}




