#include <unistd.h>
#include <math.h>

#include <Xm/Xm.h>
#include <Xm/MainW.h>
#include <Xm/CascadeB.h>
#include <Xm/MessageB.h>
#include <Xm/PushB.h>
#include <Xm/FileSB.h>
#include <Xm/DrawingA.h>
#include <Xm/SelectioB.h>

#include <e_lib.h>
extern int GetKeyword(char *line, char *word, int *LineNumber,
                      char *FullPathName, FILE *fp);
#include "jgr.h"
#include "cursors.h"

#define VERSION "$Revision"

void FileOpen_call();
void FileClose_call();
void FileReload_call();
void FileExit_call();
void View_call(Widget cascade_button, int item_no);
void Display_call(Widget cascade_button, int item_no);
void Plot_call(Widget cascade_button, int item_no);
void EPS_call(Widget cascade_button, int item_no);
void quit_activate();
void zoom_prompt();
void tilesize_prompt();
void EPSprompt(char *question, double *answer);
void ScrubDial(Widget wid, int dial);
void File_call(Widget cascade_button, int item_no);
void Help_call(Widget cascade_button, char *text,
               XmPushButtonCallbackStruct * cbs);
void select_activate(Widget widget, XtPointer client_data,
                     XmFileSelectionBoxCallbackStruct * selection);
void prompt_activate(Widget widget, XtPointer client_data,
                     XmSelectionBoxCallbackStruct * selection);
void tilesizeprompt_activate(Widget widget, XtPointer client_data,
                             XmSelectionBoxCallbackStruct * selection);
void EPSprompt_activate(Widget widget, XtPointer client_data,
                        XmSelectionBoxCallbackStruct * selection);
void cancel(Widget widget, XtPointer client_data,
            XmFileSelectionBoxCallbackStruct * selection);
void expose(Widget draw, XtPointer client_data,
            XmDrawingAreaCallbackStruct * cbk);
void resize(Widget draw, XtPointer client_data,
            XmDrawingAreaCallbackStruct * cbk);
void draw_cbk(Widget w, XButtonEvent * event, String * args, int *num_args);
void draw_mesh();
void draw_simplefigure();
void draw_info();
void load_removetable();
void reset_removetable();
char *GetCurrentDir();


/* global variables */
MESH *mesh;
int *rmv_table;                 /* the remove table */
Widget menu_bar;
int mdfopen = FALSE;
GC gc;
XGCValues gcv;
Widget draw;
Display *display;
Screen *screen;
Colormap cmap;
Pixmap pix;
Dimension w_width, w_height;
long int fill_pixel = 1;
Widget wg_Hl, wg_AHl;
Widget wg_Az, wg_Dm, wg_So, wg_Tp, wg_Te;
Widget wg_EPSsd1, wg_EPSsd2, wg_EPSsd3;
Widget top_wid;
double tilesize = 80.0;
unsigned long colourpixels[8 + MAXCOLS + 9 + MAXALTCOLS];
char zoomstr[100];
POLYGON *polygons;
int Npoly;
int AHl_once = FALSE;
char currentdir[1000];
double *G_EPSanswer;

int main(int argc, char *argv[])
{
  XtAppContext app;
  Widget main_w;
  Widget whelp, pmenu, vmenu, dmenu, emenu;
  XmString strFile, strDisplay, strView, strOpen, strClose, strRld, strExit;
  XmString strAz, strDm, strSo, strTp, strTe, strNice, strPlan, strZoom;
  XmString strDen, strDmpn, strTel, strTl, strHl, strRm, strPlot, strTs;
  XmString strPlain, strSubdom, strBorders, strInfo, strAHl;
  XmString strHN, strLG, strSG, strEPS, strZC, strAltCols;
  XmString strEPSplotplan, strEPSsizex, strEPSsizey, strEPSmargin;
  XmString strEPSbcolor, strEPScolor, strEPSexport;
/*  XmString strEPSsdstyle3, strEPSsdstyle1, strEPSsdstyle2, 
  XmString strEPSsdbwidth, strEPSsdtsize; */
  Arg arg[1];
  XtActionsRec actions;
  int i;

  static String fallback_resources[] = {
    "background: Grey", "*eplx.height: 600",
    "*eplx.width: 800", "*.title: E_PlotX11", NULL
  };
  static String translations =
    "<Btn1Motion>: draw_cbk(motion) ManagerGadgetButtonMotion() \n\
     <Btn1Down>: draw_cbk(down) ManagerGadgetArm() \n\
     <Btn1Up>: draw_cbk(up) ManagerGadgetActivate() \n\
     <KeyPress>q: draw_cbk(key_q) ManagerGadgetKeyInput() \n\
     <KeyPress>f: draw_cbk(key_f) ManagerGadgetKeyInput() \n\
     <KeyPress>c: draw_cbk(key_c) ManagerGadgetKeyInput() \n\
     <KeyPress>o: draw_cbk(key_o) ManagerGadgetKeyInput() \n\
     <KeyPress>x: draw_cbk(key_x) ManagerGadgetKeyInput() \n\
     <KeyPress>p: draw_cbk(key_p) ManagerGadgetKeyInput() \n\
     <KeyPress>d: draw_cbk(key_d) ManagerGadgetKeyInput() \n\
     <KeyPress>r: draw_cbk(key_r) ManagerGadgetKeyInput() \n\
     <KeyPress>z: draw_cbk(key_z) ManagerGadgetKeyInput() \n\
     <KeyPress>a: draw_cbk(key_a) ManagerGadgetKeyInput() \n\
     <KeyPress>e: draw_cbk(key_e) ManagerGadgetKeyInput() \n\
     <KeyPress>s: draw_cbk(key_s) ManagerGadgetKeyInput() \n\
     <KeyPress>v: draw_cbk(key_v) ManagerGadgetKeyInput() \n\
     <KeyPress>t: draw_cbk(key_t) ManagerGadgetKeyInput() \n\
     <KeyPress>k: draw_cbk(key_k) ManagerGadgetKeyInput() \n\
     <KeyPress>h: draw_cbk(key_h) ManagerGadgetKeyInput() \n\
     <KeyPress>l: draw_cbk(key_l) ManagerGadgetKeyInput()";

  /* printf("Welcome to e_plotX11 (v%s)!\n", VERSION); */

  wPlotSelect = PS_PLAINMESH;
  wSelection = DS_AZIM_ELEV;
  wDisplElemNo = wDisplPointNo = wTileElems = wTileLinks = wHN = FALSE;
  wDisplScale = wHidden = wClipPlane = wNoBorders = wRemove = FALSE;
  wEPS_plotplan = wEPS_bcolor = wEPS_color = wZC = wAltCols = FALSE;
  wEPS_subdom_style = SUBDOM_COLOR;
  wEPS_margin = 0.0;
  wEPS_subdom_b_width = 4.0;
  wEPS_subdom_t_size = 10.0;

  wAdvHidden = FALSE;
  w_Element_Size_percent = 100.00;
  w_Link_Size_percent = 100.00;

  wPlotSelect = PS_PLAINMESH;
  wSelection = DS_AZIM_ELEV;

  sprintf(zoomstr, "NODE 1");
  strcpy(currentdir, GetCurrentDir());

  top_wid = XtVaAppInitialize(&app, "eplx", NULL, 0,
                              &argc, argv, fallback_resources, NULL);
  main_w = XtVaCreateManagedWidget("main_window",
                                   xmMainWindowWidgetClass, top_wid, NULL);

  strFile = XmStringCreateLocalized("File");
  strDisplay = XmStringCreateLocalized("Display");
  strView = XmStringCreateLocalized("3D-View");
  strPlot = XmStringCreateLocalized("Plot Options");
  strEPS = XmStringCreateLocalized("EPS Export");
  menu_bar = XmVaCreateSimpleMenuBar(main_w, "main_list",
                                     XmVaCASCADEBUTTON, strFile, 'F',
                                     XmVaCASCADEBUTTON, strDisplay, 'D',
                                     XmVaCASCADEBUTTON, strView, 'V',
                                     XmVaCASCADEBUTTON, strPlot, 'P',
                                     XmVaCASCADEBUTTON, strEPS, 'E', NULL);
  XmStringFree(strFile);
  XmStringFree(strDisplay);
  XmStringFree(strView);
  XmStringFree(strPlot);
  XmStringFree(strEPS);
  XtManageChild(menu_bar);

  strOpen = XmStringCreateLocalized("Open");
  strClose = XmStringCreateLocalized("Close");
  strRld = XmStringCreateLocalized("Reload");
  strExit = XmStringCreateLocalized("Exit");
  XmVaCreateSimplePulldownMenu(menu_bar, "File_menu", 0,
                               (XtCallbackProc) File_call, XmVaPUSHBUTTON,
                               strOpen, 'O', NULL, NULL, XmVaPUSHBUTTON,
                               strClose, 'C', NULL, NULL, XmVaPUSHBUTTON,
                               strRld, 'R', NULL, NULL, XmVaSEPARATOR,
                               XmVaPUSHBUTTON, strExit, 'x', NULL, NULL,
                               NULL);
  XmStringFree(strOpen);
  XmStringFree(strClose);
  XmStringFree(strRld);
  XmStringFree(strExit);

  strDen = XmStringCreateLocalized("Display Element Numbers");
  strDmpn = XmStringCreateLocalized("Display Mesh Point Numbers");
  strHN = XmStringCreateLocalized("Highlight Nodes");
  strTel = XmStringCreateLocalized("Tile Elements");
  strTl = XmStringCreateLocalized("Tile Links");
  strTs = XmStringCreateLocalized("Set Tiling Percentage");
  strHl = XmStringCreateLocalized("Hidden Lines");
  strAHl = XmStringCreateLocalized("Advanced Hidden Lines");
  strRm = XmStringCreateLocalized("Remove Selected Elements");
  strZC = XmStringCreateLocalized("Grayscale Z colouring");
  dmenu = XmVaCreateSimplePulldownMenu(menu_bar, "Display_menu", 1,
                                       (XtCallbackProc) Display_call,
                                       XmVaTOGGLEBUTTON, strDen, 'E', NULL,
                                       NULL, XmVaTOGGLEBUTTON, strDmpn, 'P',
                                       NULL, NULL, XmVaTOGGLEBUTTON, strHN,
                                       'G', NULL, NULL, XmVaSEPARATOR,
                                       XmVaTOGGLEBUTTON, strTel, 'T', NULL,
                                       NULL, XmVaTOGGLEBUTTON, strTl, 'L',
                                       NULL, NULL, XmVaPUSHBUTTON, strTs, 'S',
                                       NULL, NULL, XmVaSEPARATOR,
                                       XmVaTOGGLEBUTTON, strHl, 'H', NULL,
                                       NULL, XmVaTOGGLEBUTTON, strAHl, 'A',
                                       NULL, NULL, XmVaTOGGLEBUTTON, strRm,
                                       'R', NULL, NULL, XmVaTOGGLEBUTTON,
                                       strZC, 'Z', NULL, NULL, NULL);

  wg_Hl = XtNameToWidget(dmenu, "button_6");
  wg_AHl = XtNameToWidget(dmenu, "button_7");
  wg_TE = XtNameToWidget(dmenu, "button_3");
  wg_TL = XtNameToWidget(dmenu, "button_4");
  wg_HN = XtNameToWidget(dmenu, "button_2");
  XmStringFree(strDen);
  XmStringFree(strDmpn);
  XmStringFree(strHN);
  XmStringFree(strTel);
  XmStringFree(strTl);
  XmStringFree(strTs);
  XmStringFree(strHl);
  XmStringFree(strAHl);
  XmStringFree(strRm);
  XmStringFree(strZC);

  strAz = XmStringCreateLocalized("Azimuth, Elevation adjustment");
  strDm = XmStringCreateLocalized("Distance, Magnification adjustment");
  strSo = XmStringCreateLocalized("Screen Origin adjustment");
  strTp = XmStringCreateLocalized("Target Point (x-y) co-ord. adjustment");
  strTe =
    XmStringCreateLocalized("Target Point Elevation (z) co-ord. adjustment");
  strNice = XmStringCreateLocalized("Nice View");
  strPlan = XmStringCreateLocalized("Plan View");
  strZoom = XmStringCreateLocalized("Zoom to");
  strInfo = XmStringCreateLocalized("Print Graphics Info");
  strSG = XmStringCreateLocalized("Save Graphics Info");
  strLG = XmStringCreateLocalized("Load Graphics Info");
  vmenu = XmVaCreateSimplePulldownMenu(menu_bar, "View_menu", 2,
                                       (XtCallbackProc) View_call,
                                       XmVaRADIOBUTTON, strAz, 'A', NULL,
                                       NULL, XmVaRADIOBUTTON, strDm, 'D',
                                       NULL, NULL, XmVaRADIOBUTTON, strSo,
                                       'O', NULL, NULL, XmVaRADIOBUTTON,
                                       strTp, 'T', NULL, NULL,
                                       XmVaRADIOBUTTON, strTe, 'E', NULL,
                                       NULL, XmNradioBehavior, True,
                                       XmNradioAlwaysOne, True, XmVaSEPARATOR,
                                       XmVaPUSHBUTTON, strNice, 'N', NULL,
                                       NULL, XmVaPUSHBUTTON, strPlan, 'P',
                                       NULL, NULL, XmVaPUSHBUTTON, strZoom,
                                       'Z', NULL, NULL, XmVaSEPARATOR,
                                       XmVaPUSHBUTTON, strInfo, 'I', NULL,
                                       NULL, XmVaPUSHBUTTON, strSG, 'S', NULL,
                                       NULL, XmVaPUSHBUTTON, strLG, 'L', NULL,
                                       NULL, NULL);
  XmStringFree(strAz);
  XmStringFree(strDm);
  XmStringFree(strSo);
  XmStringFree(strTp);
  XmStringFree(strTe);
  XmStringFree(strNice);
  XmStringFree(strPlan);
  XmStringFree(strZoom);
  XmStringFree(strInfo);
  XmStringFree(strLG);
  XmStringFree(strSG);
  wg_Az = XtNameToWidget(vmenu, "button_0");
  wg_Dm = XtNameToWidget(vmenu, "button_1");
  wg_So = XtNameToWidget(vmenu, "button_2");
  wg_Tp = XtNameToWidget(vmenu, "button_3");
  wg_Te = XtNameToWidget(vmenu, "button_4");
  XtVaSetValues(wg_Az, XmNset, TRUE, NULL);
  

  strPlain = XmStringCreateLocalized("Plain Mesh Plot");
  strSubdom = XmStringCreateLocalized("Sub-Domain Plot");
  strBorders = XmStringCreateLocalized("Don't Show Borders");
  strAltCols = XmStringCreateLocalized("Alternative SD Colours");
  pmenu = XmVaCreateSimplePulldownMenu(menu_bar, "Plot_menu", 3,
                                       (XtCallbackProc) Plot_call,
                                       XmVaRADIOBUTTON, strPlain, 'M', NULL,
                                       NULL, XmVaRADIOBUTTON, strSubdom, 'D',
                                       NULL, NULL, XmNradioBehavior, True,
                                       XmNradioAlwaysOne, False,
                                       XmVaSEPARATOR, XmVaTOGGLEBUTTON,
                                       strBorders, 'B', NULL, NULL,
                                       XmVaTOGGLEBUTTON, strAltCols, 'A',
                                       NULL, NULL, NULL);
  XmStringFree(strPlain);
  XmStringFree(strSubdom);
  XmStringFree(strBorders);
  XmStringFree(strAltCols);
  if ((wg_Plain = XtNameToWidget(pmenu, "button_0")))
    XtVaSetValues(wg_Plain, XmNset, TRUE, NULL);
  wg_Subdom = XtNameToWidget(pmenu, "button_1");
  wg_NB = XtNameToWidget(pmenu, "button_2");

  strEPSplotplan = XmStringCreateLocalized("Plot Plan View Only");
  strEPSsizex = XmStringCreateLocalized("Set X Size");
  strEPSsizey = XmStringCreateLocalized("Set Y Size");
  strEPSmargin = XmStringCreateLocalized("Set Margin");
  strEPSbcolor = XmStringCreateLocalized("Use Black Border Color");
  strEPScolor = XmStringCreateLocalized("Color Picture");
/*    strEPSsdstyle1=XmStringCreateLocalized("SD Style 1: Color");
    strEPSsdstyle2=XmStringCreateLocalized("SD Style 2: Boundary");
    strEPSsdstyle3=XmStringCreateLocalized("SD Style 3: Number");
    strEPSsdbwidth=XmStringCreateLocalized("SD Style: Set Border Width");
    strEPSsdtsize=XmStringCreateLocalized("SD Style: Set Number Size"); */
  strEPSexport = XmStringCreateLocalized("*.eps Export");
  emenu = XmVaCreateSimplePulldownMenu(menu_bar, "EPS_menu", 4,
                                       (XtCallbackProc) EPS_call,
                                       XmVaTOGGLEBUTTON, strEPSplotplan, 'P',
                                       NULL, NULL, XmVaSEPARATOR,
                                       XmVaPUSHBUTTON, strEPSsizex, 'X', NULL,
                                       NULL, XmVaPUSHBUTTON, strEPSsizey, 'Y',
                                       NULL, NULL, XmVaPUSHBUTTON,
                                       strEPSmargin, 'M', NULL, NULL,
                                       XmVaSEPARATOR, XmVaTOGGLEBUTTON,
                                       strEPSbcolor, 'B', NULL, NULL,
                                       XmVaTOGGLEBUTTON, strEPScolor, 'C',
                                       NULL, NULL,
/*        XmVaSEPARATOR,
        XmVaTOGGLEBUTTON, strEPSsdstyle1, '1',NULL,NULL,
        XmVaTOGGLEBUTTON, strEPSsdstyle2, '2',NULL,NULL,
        XmVaTOGGLEBUTTON, strEPSsdstyle3, '3',NULL,NULL,
        XmVaPUSHBUTTON, strEPSsdbwidth, 'W',NULL,NULL,
        XmVaPUSHBUTTON, strEPSsdtsize, 'S',NULL,NULL, */
                                       XmVaSEPARATOR,
                                       XmVaPUSHBUTTON, strEPSexport, 'E',
                                       NULL, NULL, NULL);
/*    wg_EPSsd1=XtNameToWidget(dmenu, "button_6");
    wg_EPSsd2=XtNameToWidget(dmenu, "button_7");
    wg_EPSsd3=XtNameToWidget(dmenu, "button_8"); */
  XmStringFree(strEPSplotplan);
  XmStringFree(strEPSsizex);
  XmStringFree(strEPSsizey);
  XmStringFree(strEPSmargin);
  XmStringFree(strEPSbcolor);
  XmStringFree(strEPScolor);
/*    XmStringFree(strEPSsdstyle1);
    XmStringFree(strEPSsdstyle2);
    XmStringFree(strEPSsdstyle3);
    XmStringFree(strEPSsdbwidth);
    XmStringFree(strEPSsdtsize); */
  XmStringFree(strEPSexport);

  whelp = XtVaCreateManagedWidget("Help",
                                  xmCascadeButtonWidgetClass, menu_bar,
                                  XmNmnemonic, 'H', NULL);
  XtAddCallback(whelp, XmNactivateCallback, (XtCallbackProc) Help_call, NULL);

  XtSetArg(arg[0], XmNmenuHelpWidget, whelp);
  XtSetValues(menu_bar, arg, 1);

  actions.string = "draw_cbk";
  actions.proc = (XtActionProc) draw_cbk;
  XtAppAddActions(app, &actions, 1);

  XtVaGetValues(top_wid, XmNwidth, &w_width, XmNheight, &w_height, NULL);
  draw = XtVaCreateWidget("draw",
                          xmDrawingAreaWidgetClass, main_w,
                          XmNtranslations,
                          XtParseTranslationTable(translations),
                          XmNbackground, BlackPixelOfScreen(XtScreen(main_w)),
                          XmNwidth, w_width, XmNheight, w_height, NULL);

  wEPS_sizex = (double) w_width;
  wEPS_sizey = (double) w_height;

  cmap = DefaultColormapOfScreen(XtScreen(draw));
  display = XtDisplay(draw);
  screen = XtScreen(draw);

  j_init_colors(display, cmap);

  gcv.foreground = WhitePixelOfScreen(XtScreen(draw));
  gc = XCreateGC(XtDisplay(draw),
                 RootWindowOfScreen(XtScreen(draw)), GCForeground, &gcv);

  XtVaGetValues(draw, XmNwidth, &w_width, XmNheight, &w_height, NULL);

  pix = XCreatePixmap(display, RootWindowOfScreen(screen),
                      2 * w_width, 2 * w_height,
                      DefaultDepthOfScreen(screen));

  XtVaSetValues(main_w, XmNmenuBar, menu_bar, XmNworkWindow, draw, NULL);

  XtAddCallback(draw, XmNexposeCallback, (XtCallbackProc) expose, NULL);
  XtAddCallback(draw, XmNresizeCallback, (XtCallbackProc) resize, NULL);
  XtManageChild(draw);

  XtRealizeWidget(top_wid);

  if (argc > 1)
  {
    strcpy(mdfname, argv[1]);
    if (e_GetMeshData(&mesh, mdfname))
    {
      mdfopen = TRUE;
      wPlotSelect = PS_PLAINMESH;
      XtVaSetValues(wg_Plain, XmNset, TRUE, NULL);
      XtVaSetValues(wg_Subdom, XmNset, FALSE, NULL);
      rmv_table = e_alloc_i_vect(mesh->TotalNElems);
      for (i = 0; i < mesh->TotalNElems; i++)
        rmv_table[i] = 0;
      if (wRemove)
        load_removetable();
      setup_polygons(mesh, &polygons, &Npoly);
      order = e_alloc_i_vect(Npoly);
      if (wAdvHidden)
        setup_hidden_line1(Npoly, polygons);

      j_init_perspective();
      draw_mesh();
    }
    else
      draw_mesh();
  }
  else
    draw_mesh();

  XtAppMainLoop(app);

  return 1;  /* to keep compiler happy */
}

void View_call(Widget cascade_button, int item_no)
{
  switch (item_no)
  {
    case 0:
      wSelection = DS_AZIM_ELEV;
      break;
    case 1:
      wSelection = DS_TCDIST_SCALE;
      break;
    case 2:
      wSelection = DS_SCRORIGIN;
      break;
    case 3:
      wSelection = DS_TARGETCOORDS;
      break;
    case 4:
      wSelection = DS_TARGETELEV;
      break;
    case 5:                    /* Nice View */
      j_init_perspective();
      draw_mesh();
      break;
    case 6:                    /* Plan View */
      XtVaGetValues(draw, XmNwidth, &w_width, XmNheight, &w_height, NULL);
      G_Xscrorigin = (w_width / 2);
      G_Yscrorigin = (w_height / 2);
      G_Target[XDIR] = w_dGlobCoordL + (w_dGlobCoordR - w_dGlobCoordL) / 2.00;
      G_Target[YDIR] = w_dGlobCoordB + (w_dGlobCoordT - w_dGlobCoordB) / 2.00;
      G_Target[ZDIR] =
        w_dGlobCoordD + (w_dGlobCoordU - w_dGlobCoordD) / 2.00;;
      G_Twist = 0.00;
      G_Azim = 0.00;
      G_Elev = M_PI / 2.00;
      j_3Drecalc_rot_matrix();
      G_dTC = w_maxsize * 20000.00;
      G_scale = E_MIN((double) w_width, (double) w_height) * 0.90 / w_maxsize;
      draw_mesh();
      break;
    case 7:                    /* zoom to */
      zoom_prompt();
      break;
    case 8:                    /* graphics info */
      graphics_info();
      break;
    case 9:
      save_graphics_info();
      break;
    case 10:
      load_graphics_info();
      break;
  }
}

void Plot_call(Widget cascade_button, int item_no)
{
  switch (item_no)
  {
    case 0:
      if (wPlotSelect == PS_SUBDOM)
        e_FreeDecompData(mesh);
      wPlotSelect = PS_PLAINMESH;
      XtVaSetValues(wg_Plain, XmNset, TRUE, NULL);
      XtVaSetValues(wg_Subdom, XmNset, FALSE, NULL);
      break;
    case 1:
      if (!mdfopen)
      {
        XtVaSetValues(wg_Plain, XmNset, TRUE, NULL);
        XtVaSetValues(wg_Subdom, XmNset, FALSE, NULL);
      }
      else
      {
        if (e_GetDecompData(mesh, mdfname))
        {
          wPlotSelect = PS_SUBDOM;
          XtVaSetValues(wg_Plain, XmNset, FALSE, NULL);
          XtVaSetValues(wg_Subdom, XmNset, TRUE, NULL);
        }
        else
        {
          XtVaSetValues(wg_Plain, XmNset, TRUE, NULL);
          XtVaSetValues(wg_Subdom, XmNset, FALSE, NULL);
        }
      }
      break;
    case 2:
      wNoBorders = !wNoBorders;
      break;
    case 3:
      wAltCols = !wAltCols;
      break;
  }
  draw_mesh();
}

void EPS_call(Widget cascade_button, int item_no)
{
  switch (item_no)
  {
    case 0:
      wEPS_plotplan = !wEPS_plotplan;
      break;
    case 1:
      EPSprompt("Set X Size", &wEPS_sizex);
      break;
    case 2:
      EPSprompt("Set Y Size", &wEPS_sizey);
      break;
    case 3:
      EPSprompt("Set Margin", &wEPS_margin);
      break;
    case 4:
      wEPS_bcolor = !wEPS_bcolor;
      break;
    case 5:
      wEPS_color = !wEPS_color;
      break;
/*        case 6:
            wEPS_subdom_style=SUBDOM_COLOR;
            XtVaSetValues(wg_EPSsd1, XmNset, TRUE, NULL);
            XtVaSetValues(wg_EPSsd2, XmNset, FALSE, NULL);
            XtVaSetValues(wg_EPSsd3, XmNset, FALSE, NULL);
            break;
        case 7:
            wEPS_subdom_style=SUBDOM_BOUNDARY;
            XtVaSetValues(wg_EPSsd1, XmNset, FALSE, NULL);
            XtVaSetValues(wg_EPSsd2, XmNset, TRUE, NULL);
            XtVaSetValues(wg_EPSsd3, XmNset, FALSE, NULL);
            break;
        case 8:
            wEPS_subdom_style=SUBDOM_NUMBER;
            XtVaSetValues(wg_EPSsd1, XmNset, FALSE, NULL);
            XtVaSetValues(wg_EPSsd2, XmNset, FALSE, NULL);
            XtVaSetValues(wg_EPSsd3, XmNset, TRUE, NULL);
            break;
        case 9:
            EPSprompt("Set Border Width",&wEPS_subdom_b_width);
            break;
        case 10:
            EPSprompt("Set Number Size",&wEPS_subdom_t_size);
            break; */
    case 6:
      EPS_export(polygons, Npoly);
      break;
  }
}

void Display_call(Widget cascade_button, int item_no)
{
  switch (item_no)
  {
    case 0:
      wDisplElemNo = !wDisplElemNo;
      break;
    case 1:
      wDisplPointNo = !wDisplPointNo;
      break;
    case 2:
      wHN = !wHN;
      break;
    case 3:
      wTileElems = !wTileElems;
      if (wTileElems)
        w_Element_Size_percent = tilesize;
      else
        w_Element_Size_percent = 100.00;
      break;
    case 4:
      wTileLinks = !wTileLinks;
      if (wTileLinks)
        w_Link_Size_percent = tilesize;
      else
        w_Link_Size_percent = 100.00;
      break;
    case 5:
      tilesize_prompt();
      break;
    case 6:
      wHidden = !wHidden;
      if (wAdvHidden)
      {
        XtVaSetValues(wg_AHl, XmNset, FALSE, NULL);
        wAdvHidden = FALSE;
      }
      break;
    case 7:
      wAdvHidden = !wAdvHidden;
      if (wHidden)
      {
        XtVaSetValues(wg_Hl, XmNset, FALSE, NULL);
        wHidden = FALSE;
      }
      if ((mdfopen) && (!AHl_once))
      {
        set_watch_cursor(draw);
        setup_hidden_line1(Npoly, polygons);
        AHl_once = TRUE;
      }
      break;
    case 8:
      wRemove = !wRemove;
      if (wRemove)
        load_removetable();
      else
        reset_removetable();
      break;
    case 9:
      wZC = !wZC;
      break;
  }
  draw_mesh();
}

void load_removetable()
{
  FILE *fp;
  char filename[400];
  char word[100];
  int i;
  char line[E_LINE_MAX];
  int rmv;
  int ln = 0;
  int l;

  strcpy(filename, mdfname);
  strcat(filename, ".rmv");
  if (!(fp = fopen(filename, "r")))
    e_GeneralError("can't load remove info");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
  {
    if (!GetKeyword(line, word, &ln, filename, fp))
      e_LineError(ln, "can't read element removal keyword");
    if (strcmp(word, "REMOVE_LINK1") != 0)
      e_LineError(ln, "element removal keyword incorrect");
    for (i = 0; i < mesh->NElemsLink1; i++)
    {
      if (!e_GetNextLine(line, &ln, filename, fp))
        e_LineError(ln, "can't read removal info");
      if (sscanf(line, "%d %d", &l, &rmv) != 2)
        e_LineError(ln, "missing information in removal info");
      if (l != i + 1)
        e_LineError(ln, "missing information in removal info");
      rmv_table[mesh->FirstNetIndex[E_ELEM_TYPE_LINK1] + i - 1] = rmv;
    }
  }
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
  {
    if (!GetKeyword(line, word, &ln, filename, fp))
      e_LineError(ln, "can't read element removal keyword");
    if (strcmp(word, "REMOVE_LINK2") != 0)
      e_LineError(ln, "element removal keyword incorrect");
    for (i = 0; i < mesh->NElemsLink2; i++)
    {
      if (!e_GetNextLine(line, &ln, filename, fp))
        e_LineError(ln, "can't read removal info");
      if (sscanf(line, "%d %d", &l, &rmv) != 2)
        e_LineError(ln, "missing information in removal info");
      if (l != i + 1)
        e_LineError(ln, "missing information in removal info");
      rmv_table[mesh->FirstNetIndex[E_ELEM_TYPE_LINK2] + i - 1] = rmv;
    }
  }
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
  {
    if (!GetKeyword(line, word, &ln, filename, fp))
      e_LineError(ln, "can't read element removal keyword");
    if (strcmp(word, "REMOVE_LINK3") != 0)
      e_LineError(ln, "element removal keyword incorrect");
    for (i = 0; i < mesh->NElemsLink3; i++)
    {
      if (!e_GetNextLine(line, &ln, filename, fp))
        e_LineError(ln, "can't read removal info");
      if (sscanf(line, "%d %d", &l, &rmv) != 2)
        e_LineError(ln, "missing information in removal info");
      if (l != i + 1)
        e_LineError(ln, "missing information in removal info");
      rmv_table[mesh->FirstNetIndex[E_ELEM_TYPE_LINK3] + i - 1] = rmv;
    }
  }
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
  {
    if (!GetKeyword(line, word, &ln, filename, fp))
      e_LineError(ln, "can't read element removal keyword");
    if (strcmp(word, "REMOVE_LINK4") != 0)
      e_LineError(ln, "element removal keyword incorrect");
    for (i = 0; i < mesh->NElemsLink4; i++)
    {
      if (!e_GetNextLine(line, &ln, filename, fp))
        e_LineError(ln, "can't read removal info");
      if (sscanf(line, "%d %d", &l, &rmv) != 2)
        e_LineError(ln, "missing information in removal info");
      if (l != i + 1)
        e_LineError(ln, "missing information in removal info");
      rmv_table[mesh->FirstNetIndex[E_ELEM_TYPE_LINK4] + i - 1] = rmv;
    }
  }
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
  {
    if (!GetKeyword(line, word, &ln, filename, fp))
      e_LineError(ln, "can't read element removal keyword");
    if (strcmp(word, "REMOVE_LINK5") != 0)
      e_LineError(ln, "element removal keyword incorrect");
    for (i = 0; i < mesh->NElemsLink5; i++)
    {
      if (!e_GetNextLine(line, &ln, filename, fp))
        e_LineError(ln, "can't read removal info");
      if (sscanf(line, "%d %d", &l, &rmv) != 2)
        e_LineError(ln, "missing information in removal info");
      if (l != i + 1)
        e_LineError(ln, "missing information in removal info");
      rmv_table[mesh->FirstNetIndex[E_ELEM_TYPE_LINK5] + i - 1] = rmv;
    }
  }
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
  {
    if (!GetKeyword(line, word, &ln, filename, fp))
      e_LineError(ln, "can't read element removal keyword");
    if (strcmp(word, "REMOVE_TRIANG1") != 0)
      e_LineError(ln, "element removal keyword incorrect");
    for (i = 0; i < mesh->NElemsTriang1; i++)
    {
      if (!e_GetNextLine(line, &ln, filename, fp))
        e_LineError(ln, "can't read removal info");
      if (sscanf(line, "%d %d", &l, &rmv) != 2)
        e_LineError(ln, "missing information in removal info");
      if (l != i + 1)
        e_LineError(ln, "missing information in removal info");
      rmv_table[mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] + i - 1] = rmv;
    }
  }
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
  {
    if (!GetKeyword(line, word, &ln, filename, fp))
      e_LineError(ln, "can't read element removal keyword");
    if (strcmp(word, "REMOVE_TRIANG2") != 0)
      e_LineError(ln, "element removal keyword incorrect");
    for (i = 0; i < mesh->NElemsTriang2; i++)
    {
      if (!e_GetNextLine(line, &ln, filename, fp))
        e_LineError(ln, "can't read removal info");
      if (sscanf(line, "%d %d", &l, &rmv) != 2)
        e_LineError(ln, "missing information in removal info");
      if (l != i + 1)
        e_LineError(ln, "missing information in removal info");
      rmv_table[mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2] + i - 1] = rmv;
    }
  }
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
  {
    if (!GetKeyword(line, word, &ln, filename, fp))
      e_LineError(ln, "can't read element removal keyword");
    if (strcmp(word, "REMOVE_TRIANG3") != 0)
      e_LineError(ln, "element removal keyword incorrect");
    for (i = 0; i < mesh->NElemsTriang3; i++)
    {
      if (!e_GetNextLine(line, &ln, filename, fp))
        e_LineError(ln, "can't read removal info");
      if (sscanf(line, "%d %d", &l, &rmv) != 2)
        e_LineError(ln, "missing information in removal info");
      if (l != i + 1)
        e_LineError(ln, "missing information in removal info");
      rmv_table[mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] + i - 1] = rmv;
    }
  }
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
  {
    if (!GetKeyword(line, word, &ln, filename, fp))
      e_LineError(ln, "can't read element removal keyword");
    if (strcmp(word, "REMOVE_TRIANG4") != 0)
      e_LineError(ln, "element removal keyword incorrect");
    for (i = 0; i < mesh->NElemsTriang4; i++)
    {
      if (!e_GetNextLine(line, &ln, filename, fp))
        e_LineError(ln, "can't read removal info");
      if (sscanf(line, "%d %d", &l, &rmv) != 2)
        e_LineError(ln, "missing information in removal info");
      if (l != i + 1)
        e_LineError(ln, "missing information in removal info");
      rmv_table[mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] + i - 1] = rmv;
    }
  }
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
  {
    if (!GetKeyword(line, word, &ln, filename, fp))
      e_LineError(ln, "can't read element removal keyword");
    if (strcmp(word, "REMOVE_TRIANG5") != 0)
      e_LineError(ln, "element removal keyword incorrect");
    for (i = 0; i < mesh->NElemsTriang5; i++)
    {
      if (!e_GetNextLine(line, &ln, filename, fp))
        e_LineError(ln, "can't read removal info");
      if (sscanf(line, "%d %d", &l, &rmv) != 2)
        e_LineError(ln, "missing information in removal info");
      if (l != i + 1)
        e_LineError(ln, "missing information in removal info");
      rmv_table[mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] + i - 1] = rmv;
    }
  }
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
  {
    if (!GetKeyword(line, word, &ln, filename, fp))
      e_LineError(ln, "can't read element removal keyword");
    if (strcmp(word, "REMOVE_QUAD1") != 0)
      e_LineError(ln, "element removal keyword incorrect");
    for (i = 0; i < mesh->NElemsQuad1; i++)
    {
      if (!e_GetNextLine(line, &ln, filename, fp))
        e_LineError(ln, "can't read removal info");
      if (sscanf(line, "%d %d", &l, &rmv) != 2)
        e_LineError(ln, "missing information in removal info");
      if (l != i + 1)
        e_LineError(ln, "missing information in removal info");
      rmv_table[mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1] + i - 1] = rmv;
    }
  }
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
  {
    if (!GetKeyword(line, word, &ln, filename, fp))
      e_LineError(ln, "can't read element removal keyword");
    if (strcmp(word, "REMOVE_QUAD3") != 0)
      e_LineError(ln, "element removal keyword incorrect");
    for (i = 0; i < mesh->NElemsQuad3; i++)
    {
      if (!e_GetNextLine(line, &ln, filename, fp))
        e_LineError(ln, "can't read removal info");
      if (sscanf(line, "%d %d", &l, &rmv) != 2)
        e_LineError(ln, "missing information in removal info");
      if (l != i + 1)
        e_LineError(ln, "missing information in removal info");
      rmv_table[mesh->FirstNetIndex[E_ELEM_TYPE_QUAD3] + i - 1] = rmv;
    }
  }
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
  {
    if (!GetKeyword(line, word, &ln, filename, fp))
      e_LineError(ln, "can't read element removal keyword");
    if (strcmp(word, "REMOVE_TETRAH1") != 0)
      e_LineError(ln, "element removal keyword incorrect");
    for (i = 0; i < mesh->NElemsTetrah1; i++)
    {
      if (!e_GetNextLine(line, &ln, filename, fp))
        e_LineError(ln, "can't read removal info");
      if (sscanf(line, "%d %d", &l, &rmv) != 2)
        e_LineError(ln, "missing information in removal info");
      if (l != i + 1)
        e_LineError(ln, "missing information in removal info");
      rmv_table[mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1] + i - 1] = rmv;
    }
  }
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
  {
    if (!GetKeyword(line, word, &ln, filename, fp))
      e_LineError(ln, "can't read element removal keyword");
    if (strcmp(word, "REMOVE_BLOCK1") != 0)
      e_LineError(ln, "element removal keyword incorrect");
    for (i = 0; i < mesh->NElemsBlock1; i++)
    {
      if (!e_GetNextLine(line, &ln, filename, fp))
        e_LineError(ln, "can't read removal info");
      if (sscanf(line, "%d %d", &l, &rmv) != 2)
        e_LineError(ln, "missing information in removal info");
      if (l != i + 1)
        e_LineError(ln, "missing information in removal info");
      rmv_table[mesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1] + i - 1] = rmv;
    }
  }
}

void reset_removetable()
{
  int i;

  for (i = 0; i < mesh->TotalNElems; i++)
    rmv_table[i] = 0;
}

void File_call(Widget cascade_button, int item_no)
{
  switch (item_no)
  {
    case 3:
      FileExit_call();
      break;
    case 2:
      FileReload_call();
      break;
    case 1:
      FileClose_call();
      break;
    case 0:
      FileOpen_call();
      break;
  }
}

void Help_call(Widget cascade_button, char *text,
               XmPushButtonCallbackStruct * cbs)
{
  static Widget wg_dialog;
  Arg args[2];
  int n = 0;
  XmString msg =
    XmStringCreateLtoR("contact Jelle!", XmSTRING_DEFAULT_CHARSET);

  if (!wg_dialog)
  {
    XtSetArg(args[n], XmNmessageString, msg);
    n++;
    wg_dialog = XmCreateInformationDialog(top_wid, "help_dialog", args, n);
  }
  ScrubDial(wg_dialog, XmDIALOG_HELP_BUTTON);
  ScrubDial(wg_dialog, XmDIALOG_CANCEL_BUTTON);
  XtManageChild(wg_dialog);
  XtPopup(XtParent(wg_dialog), XtGrabNone);
}

void FileExit_call()
{
  Widget dialog;
  XmString xm_string;
  Arg args[2];

  xm_string = XmStringCreateLocalized("Are you sure you want to quit?");
  XtSetArg(args[0], XmNmessageString, xm_string);
  XtSetArg(args[1], XmNdefaultButtonType, XmDIALOG_OK_BUTTON);

  /* Create the WarningDialog */
  dialog = XmCreateWarningDialog(top_wid, "quit", args, 2);
  ScrubDial(dialog, XmDIALOG_HELP_BUTTON);

  XmStringFree(xm_string);

  XtAddCallback(dialog, XmNokCallback, (XtCallbackProc) quit_activate, NULL);

  XtManageChild(dialog);
  XtPopup(XtParent(dialog), XtGrabNone);
}

void zoom_prompt()
{
  Widget dialog, remove;
  XmString xm_string_label, xm_string_text;
  Arg args[3];

  if (!mdfopen)
    return;

  xm_string_label = XmStringCreateLocalized("Zoom to:");
  xm_string_text = XmStringCreateLocalized(zoomstr);
  XtSetArg(args[0], XmNselectionLabelString, xm_string_label);
  XtSetArg(args[1], XmNtextString, xm_string_text);
  XtSetArg(args[2], XmNdefaultButtonType, XmDIALOG_OK_BUTTON);

  dialog = XmCreatePromptDialog(top_wid, "prompt", args, 3);

  XmStringFree(xm_string_label);
  XmStringFree(xm_string_text);

  XtAddCallback(dialog, XmNokCallback, (XtCallbackProc) prompt_activate,
                NULL);
  XtAddCallback(dialog, XmNcancelCallback, (XtCallbackProc) cancel, NULL);

  remove = XmSelectionBoxGetChild(dialog, XmDIALOG_HELP_BUTTON);
  XtUnmanageChild(remove);

  XtManageChild(dialog);
  XtPopup(XtParent(dialog), XtGrabNone);
}

void tilesize_prompt()
{
  Widget dialog, remove;
  XmString xm_string_label, xm_string_text;
  Arg args[3];
  char str[10];

  if (!mdfopen)
    return;

  xm_string_label = XmStringCreateLocalized("tiling percentage:");
  sprintf(str, "%.1f", tilesize);
  xm_string_text = XmStringCreateLocalized(str);
  XtSetArg(args[0], XmNselectionLabelString, xm_string_label);
  XtSetArg(args[1], XmNtextString, xm_string_text);
  XtSetArg(args[2], XmNdefaultButtonType, XmDIALOG_OK_BUTTON);

  dialog = XmCreatePromptDialog(top_wid, "prompt", args, 3);

  XmStringFree(xm_string_label);
  XmStringFree(xm_string_text);

  XtAddCallback(dialog, XmNokCallback,
                (XtCallbackProc) tilesizeprompt_activate, NULL);
  XtAddCallback(dialog, XmNcancelCallback, (XtCallbackProc) cancel, NULL);

  remove = XmSelectionBoxGetChild(dialog, XmDIALOG_HELP_BUTTON);
  XtUnmanageChild(remove);

  XtManageChild(dialog);
  XtPopup(XtParent(dialog), XtGrabNone);
}

void EPSprompt(char *question, double *answer)
{
  Widget dialog, remove;
  XmString xm_string_label, xm_string_text;
  Arg args[3];
  char str[10];

  xm_string_label = XmStringCreateLocalized(question);
  sprintf(str, "%.1f", *answer);
  xm_string_text = XmStringCreateLocalized(str);
  XtSetArg(args[0], XmNselectionLabelString, xm_string_label);
  XtSetArg(args[1], XmNtextString, xm_string_text);
  XtSetArg(args[2], XmNdefaultButtonType, XmDIALOG_OK_BUTTON);

  dialog = XmCreatePromptDialog(top_wid, "prompt", args, 3);

  XmStringFree(xm_string_label);
  XmStringFree(xm_string_text);

  G_EPSanswer = answer;
  XtAddCallback(dialog, XmNokCallback, (XtCallbackProc) EPSprompt_activate,
                NULL);
  XtAddCallback(dialog, XmNcancelCallback, (XtCallbackProc) cancel, NULL);

  remove = XmSelectionBoxGetChild(dialog, XmDIALOG_HELP_BUTTON);
  XtUnmanageChild(remove);

  XtManageChild(dialog);
  XtPopup(XtParent(dialog), XtGrabNone);
}

void FileClose_call()
{
  if (!mdfopen)
    return;
  if (wPlotSelect == PS_SUBDOM)
    e_FreeDecompData(mesh);
  e_FreeMeshData(mesh);
  e_free(rmv_table);
  free_polygons(Npoly, polygons);
  e_free(order);
  mdfopen = FALSE;
  AHl_once = FALSE;

  XSetForeground(XtDisplay(draw), gc, BlackPixelOfScreen(XtScreen(draw)));
  XFillRectangle(display, pix, gc, 0, 0, w_width, w_height);
  draw_info();
}

void FileReload_call()
{
  int i;

  if (!mdfopen)
    return;
  if (wPlotSelect == PS_SUBDOM)
    e_FreeDecompData(mesh);
  e_FreeMeshData(mesh);
  e_free(rmv_table);
  free_polygons(Npoly, polygons);
  e_free(order);
  mdfopen = FALSE;
  AHl_once = FALSE;
  if (e_GetMeshData(&mesh, mdfname))
  {
    mdfopen = TRUE;
    wPlotSelect = PS_PLAINMESH;
    XtVaSetValues(wg_Plain, XmNset, TRUE, NULL);
    XtVaSetValues(wg_Subdom, XmNset, FALSE, NULL);
    rmv_table = e_alloc_i_vect(mesh->TotalNElems);
    for (i = 0; i < mesh->TotalNElems; i++)
      rmv_table[i] = 0;
    if (wRemove)
      load_removetable();
    setup_polygons(mesh, &polygons, &Npoly);
    order = e_alloc_i_vect(Npoly);
    if (wAdvHidden)
      setup_hidden_line1(Npoly, polygons);

    /* j_init_perspective(); */
    draw_mesh();
  }
  else
    draw_info();
}

void FileOpen_call()
{
  Widget dialog;
  Widget remove;
  XmString mask;
  char str[1024];
  Arg args[2];

  if (mdfopen)
  {
    if (wPlotSelect == PS_SUBDOM)
      e_FreeDecompData(mesh);
    e_FreeMeshData(mesh);
    e_free(rmv_table);
    free_polygons(Npoly, polygons);
    mdfopen = FALSE;
  }

  strcpy(str, currentdir);
  strcat(str, "/*.mdf");
  mask = XmStringCreateLocalized(str);
  XtSetArg(args[0], XmNdirMask, mask);
  XtSetArg(args[1], XmNdirMask, mask);

  dialog = XmCreateFileSelectionDialog(top_wid, "select", args, 1);
  XtAddCallback(dialog, XmNokCallback, (XtCallbackProc) select_activate,
                NULL);
  XtAddCallback(dialog, XmNcancelCallback, (XtCallbackProc) cancel, NULL);

  remove = (Widget) XmSelectionBoxGetChild(dialog, XmDIALOG_HELP_BUTTON);
  XtUnmanageChild(remove);

  XtManageChild(dialog);
  XtPopup(XtParent(dialog), XtGrabNone);
}

void ScrubDial(Widget wid, int dial)
{
  Widget remove;

  remove = XmMessageBoxGetChild(wid, dial);
  XtUnmanageChild(remove);
}

void quit_activate()
{

  if (mdfopen)
  {
    e_FreeMeshData(mesh);
    free_polygons(Npoly, polygons);
  }
  /* printf("Bye...\n"); */
  XFreeColors(display, cmap, colourpixels, MAXCOLS + 8, 0);
  exit(0);
}

void select_activate(Widget widget, XtPointer client_data,
                     XmFileSelectionBoxCallbackStruct * selection)
{
  int i;

  char *filename;
  char *last_slash;

  XmStringGetLtoR(selection->value, XmSTRING_DEFAULT_CHARSET, &filename);
  strncpy(mdfname, filename, strlen(filename) - 4);
  mdfname[strlen(filename) - 4] = '\0';

  last_slash = strrchr(filename, '/');
  if (last_slash != NULL)
  {
    strncpy(currentdir, filename, strlen(filename) - strlen(last_slash));
    currentdir[strlen(filename) - strlen(last_slash)] = '\0';
  }

  XtUnmanageChild(widget);
  if (e_GetMeshData(&mesh, mdfname))
  {
    mdfopen = TRUE;
    wPlotSelect = PS_PLAINMESH;
    XtVaSetValues(wg_Plain, XmNset, TRUE, NULL);
    XtVaSetValues(wg_Subdom, XmNset, FALSE, NULL);
    rmv_table = e_alloc_i_vect(mesh->TotalNElems);
    for (i = 0; i < mesh->TotalNElems; i++)
      rmv_table[i] = 0;
    if (wRemove)
      load_removetable();
    AHl_once = FALSE;
    j_init_perspective();
    setup_polygons(mesh, &polygons, &Npoly);
    order = e_alloc_i_vect(Npoly);
    if (wAdvHidden)
      setup_hidden_line1(Npoly, polygons);
    draw_mesh();
  }
}

void prompt_activate(Widget widget, XtPointer client_data,
                     XmSelectionBoxCallbackStruct * selection)
{
  char *zmstr;
  char keyword[E_LINE_MAX];
  int number, i, s;

  XmStringGetLtoR(selection->value, XmSTRING_DEFAULT_CHARSET, &zmstr);

  if (sscanf(zmstr, "%s %d", keyword, &number) != 2)
  {
    printf("can't understand your zoom to expression\n");
    return;
  }
  if ((strcmp(keyword, "NODE") != 0) && (strcmp(keyword, "TRIANG1") != 0))
  {
    printf("don't recognize your zoom to keyword\n");
    return;
  }
  if (strcmp(keyword, "NODE") == 0)
  {
    if ((number < 1) || (number > mesh->NNodes))
    {
      printf("invalid zoom to node. only %d nodes in mesh\n", mesh->NNodes);
      return;
    }
    for (i = 0; i < 3; i++)
      G_Target[i] = mesh->MeshPointCoords[number - 1][i];
  }
  else if (strcmp(keyword, "TRIANG1") == 0)
  {
    if ((number < 1) || (number > mesh->NElemsTriang1))
    {
      printf("invalid zoom to element. only %d triang1's in mesh\n",
             mesh->NElemsTriang1);
      return;
    }
    for (i = 0; i < 3; i++)
    {
      G_Target[i] = 0.0;
      for (s = 0; s < 3; s++)
        G_Target[i] +=
          mesh->MeshPointCoords[mesh->NodesTriang1[number - 1][s] - 1][i];
      G_Target[i] /= 3.0;
    }
  }

  G_scale *= 3;
  strcpy(zoomstr, zmstr);
  draw_mesh();
}

void tilesizeprompt_activate(Widget widget, XtPointer client_data,
                             XmSelectionBoxCallbackStruct * selection)
{
  char *str;

  XmStringGetLtoR(selection->value, XmSTRING_DEFAULT_CHARSET, &str);

  if (sscanf(str, "%lf", &tilesize) != 1)
  {
    printf("can't understand your tiling percentage\n");
    return;
  }
  if (wTileElems)
    w_Element_Size_percent = tilesize;
  if (wTileLinks)
    w_Link_Size_percent = tilesize;
  if (wTileElems || wTileLinks)
    draw_mesh();
}

void EPSprompt_activate(Widget widget, XtPointer client_data,
                        XmSelectionBoxCallbackStruct * selection)
{
  char *str;

  XmStringGetLtoR(selection->value, XmSTRING_DEFAULT_CHARSET, &str);

  if (sscanf(str, "%lf", G_EPSanswer) != 1)
  {
    printf("can't understand your value!\n");
    return;
  }
}

void cancel(Widget widget, XtPointer client_data,
            XmFileSelectionBoxCallbackStruct * selection)
{
  XtUnmanageChild(widget);
  draw_mesh();
}

void expose(Widget draw, XtPointer client_data,
            XmDrawingAreaCallbackStruct * cbk)
{
  XCopyArea(cbk->event->xexpose.display, pix, cbk->window, gc,
            0, 0, w_width, w_height, 0, 0);
}

void resize(Widget draw, XtPointer client_data,
            XmDrawingAreaCallbackStruct * cbk)
{

  XtVaGetValues(draw, XmNwidth, &w_width, XmNheight, &w_height, NULL);

  /* XFreePixmap(display,pix);
     pix = XCreatePixmap(display, RootWindowOfScreen(screen), 
     w_width, w_height, DefaultDepthOfScreen(screen));
     draw_mesh(); */

  G_Xscrorigin = w_width / 2;
  G_Yscrorigin = w_height / 2;
}

void draw_mesh(void)
{
  Window win = XtWindow(draw);

  XSetForeground(XtDisplay(draw), gc, BlackPixelOfScreen(XtScreen(draw)));
  XFillRectangle(display, pix, gc, 0, 0, w_width, w_height);

  if (mdfopen)
    j_paint_polygons(display, pix, gc, cmap, polygons, Npoly);
  else
    draw_info();

  XCopyArea(display, pix, win, gc, 0, 0, w_width, w_height, 0, 0);
}

void draw_cbk(Widget w, XButtonEvent * event, String * args, int *num_args)
{
  static Position x_b, y_b;     /* begining of the move */
  static Position x_l, y_l;     /* last positions */
  Position xdifpix, ydifpix;
  Window win = XtWindow(draw);

  if (!mdfopen)
    return;

  if (strcmp(args[0], "down") == 0)
  {
    x_b = event->x;
    y_b = event->y;
    XSetForeground(XtDisplay(draw), gc, BlackPixelOfScreen(XtScreen(draw)));
    XFillRectangle(display, win, gc, 0, 0, w_width, w_height);

  }
  else if (strcmp(args[0], "motion") == 0)
  {
    xdifpix = event->x - x_b;
    ydifpix = event->y - y_b;
    switch (wSelection)
    {
      case DS_AZIM_ELEV:
        G_Azim -= (double) xdifpix *j3D_AZIMUTH_INCREMENT / 180.00 * M_PI;
        G_Elev += (double) ydifpix *j3D_ELEVATION_INCREMENT / 180.00 * M_PI;
        j_3Drecalc_rot_matrix();
        break;
      case DS_TCDIST_SCALE:
        G_dTC *= 1.00 + (double) xdifpix *j3D_DISTANCE_MULTIPLIER / 100.00;
        G_scale *= 1.00 - (double) ydifpix *j3D_SCALEFACT_MULTIPLIER / 100.00;
        break;
      case DS_SCRORIGIN:
        G_Xscrorigin += xdifpix;
        G_Yscrorigin += ydifpix;
        break;
      case DS_TARGETCOORDS:
        G_Target[XDIR] -=
          (double) xdifpix / 10.00 * 0.05 * (w_dGlobCoordR - w_dGlobCoordL);
        G_Target[YDIR] +=
          (double) ydifpix / 10.00 * 0.05 * (w_dGlobCoordT - w_dGlobCoordB);
        break;
      case DS_TARGETELEV:
        G_Target[ZDIR] -=
          (double) ydifpix / 10.00 * 0.05 * (w_dGlobCoordU - w_dGlobCoordD);
        break;
    }
    draw_simplefigure();
    x_b = event->x;
    y_b = event->y;
  }
  else if (strcmp(args[0], "up") == 0)
  {
    draw_mesh();
  }
  else if (strcmp(args[0], "key_q") == 0)
  {
    FileExit_call();
  }
  else if (strcmp(args[0], "key_f") == 0)
  {
    FileReload_call();
  }
  else if (strcmp(args[0], "key_c") == 0)
  {
    FileClose_call();
  }
  else if (strcmp(args[0], "key_o") == 0)
  {
    FileOpen_call();
  }
  else if (strcmp(args[0], "key_x") == 0)
  {
    load_removetable();
  }
  else if (strcmp(args[0], "key_p") == 0)
  {
    Plot_call(NULL,0);
  }
  else if (strcmp(args[0], "key_d") == 0)
  {
    Plot_call(NULL,1);
  }
  else if (strcmp(args[0], "key_l") == 0)
  {
    Plot_call(NULL,2);
    if (wNoBorders)
      XtVaSetValues(wg_NB, XmNset, TRUE, NULL);
    else
      XtVaSetValues(wg_NB, XmNset, FALSE, NULL);
  } 
  else if (strcmp(args[0], "key_r") == 0)
  {
    View_call(NULL,0);
    XtVaSetValues(wg_Az, XmNset, TRUE, NULL);
    XtVaSetValues(wg_Dm, XmNset, FALSE, NULL);
    XtVaSetValues(wg_So, XmNset, FALSE, NULL);
    XtVaSetValues(wg_Tp, XmNset, FALSE, NULL);
    XtVaSetValues(wg_Te, XmNset, FALSE, NULL);
  } 
  else if (strcmp(args[0], "key_z") == 0)
  {
    View_call(NULL,1);
    XtVaSetValues(wg_Az, XmNset, FALSE, NULL);
    XtVaSetValues(wg_Dm, XmNset, TRUE, NULL);
    XtVaSetValues(wg_So, XmNset, FALSE, NULL);
    XtVaSetValues(wg_Tp, XmNset, FALSE, NULL);
    XtVaSetValues(wg_Te, XmNset, FALSE, NULL);
  } 
  else if (strcmp(args[0], "key_s") == 0)
  {
    View_call(NULL,2);
    XtVaSetValues(wg_Az, XmNset, FALSE, NULL);
    XtVaSetValues(wg_Dm, XmNset, FALSE, NULL);
    XtVaSetValues(wg_So, XmNset, TRUE, NULL);
    XtVaSetValues(wg_Tp, XmNset, FALSE, NULL);
    XtVaSetValues(wg_Te, XmNset, FALSE, NULL);
  } 
  else if (strcmp(args[0], "key_a") == 0)
  {
    View_call(NULL,3);
    XtVaSetValues(wg_Az, XmNset, FALSE, NULL);
    XtVaSetValues(wg_Dm, XmNset, FALSE, NULL);
    XtVaSetValues(wg_So, XmNset, FALSE, NULL);
    XtVaSetValues(wg_Tp, XmNset, TRUE, NULL);
    XtVaSetValues(wg_Te, XmNset, FALSE, NULL);
  } 
  else if (strcmp(args[0], "key_e") == 0)
  {
    View_call(NULL,4);
    XtVaSetValues(wg_Az, XmNset, FALSE, NULL);
    XtVaSetValues(wg_Dm, XmNset, FALSE, NULL);
    XtVaSetValues(wg_So, XmNset, FALSE, NULL);
    XtVaSetValues(wg_Tp, XmNset, FALSE, NULL);
    XtVaSetValues(wg_Te, XmNset, TRUE, NULL);
  } 
  else if (strcmp(args[0], "key_v") == 0)
  {
    View_call(NULL,5);
  } 
  else if (strcmp(args[0], "key_h") == 0)
  {
    XtVaSetValues(wg_Hl, XmNset, TRUE, NULL);
    Display_call(NULL,7);
  } 
  else if (strcmp(args[0], "key_t") == 0)
  {
    XtVaSetValues(wg_TE, XmNset, TRUE, NULL);
    Display_call(NULL,3);
  } 
  else if (strcmp(args[0], "key_k") == 0)
  {
    printf("\nshortcut keys for eplx\n\n");
    printf("o\topen\n");
    printf("c\tclose\n");
    printf("f\trefresh\n");
    printf("q\tquit\n\n");
    printf("t\ttile elements\n");
    printf("h\thidden lines\n");
    printf("x\tload removal table\n\n");
    printf("r\trotate\n");
    printf("z\tzoom\n");
    printf("s\tscreen origin\n");
    printf("a\tpanning\n");
    printf("e\televation\n");
    printf("v\tnice view\n\n");
    printf("p\tplain mesh\n");
    printf("d\tdecomposition\n");
    printf("l\tlines\n\n");
    printf("k\tshortcut keys\n\n");
  } 
  
  x_l = event->x;
  y_l = event->y;

}

void draw_simplefigure(void)
{
  Window win = XtWindow(draw);
  j_3Dvector qnode1, qnode2, qnode3, qnode4;
  j_3Dvector lnode1, lnode2, lnode3, lnode4;
  j_3Dvector qnode5, qnode6, qnode7, qnode8;
  j_3Dvector lnode5, lnode6, lnode7, lnode8;
  j_3Dvector qTarget, lTarget;
  int T_length;
  int nLength;
  char szBuffer[5];

  XSetForeground(XtDisplay(draw), gc, BlackPixelOfScreen(XtScreen(draw)));
  XFillRectangle(display, win, gc, 0, 0, w_width, w_height);
  fill_pixel = colourpixels[6];
  XSetForeground(display, gc, fill_pixel);

  qnode1[XDIR] = w_dGlobCoordL;
  qnode1[YDIR] = w_dGlobCoordB;
  qnode1[ZDIR] = w_dGlobCoordD;
  qnode2[XDIR] = w_dGlobCoordR;
  qnode2[YDIR] = w_dGlobCoordB;
  qnode2[ZDIR] = w_dGlobCoordD;
  qnode3[XDIR] = w_dGlobCoordR;
  qnode3[YDIR] = w_dGlobCoordT;
  qnode3[ZDIR] = w_dGlobCoordD;
  qnode4[XDIR] = w_dGlobCoordL;
  qnode4[YDIR] = w_dGlobCoordT;
  qnode4[ZDIR] = w_dGlobCoordD;
  qnode5[XDIR] = w_dGlobCoordL;
  qnode5[YDIR] = w_dGlobCoordB;
  qnode5[ZDIR] = w_dGlobCoordU;
  qnode6[XDIR] = w_dGlobCoordR;
  qnode6[YDIR] = w_dGlobCoordB;
  qnode6[ZDIR] = w_dGlobCoordU;
  qnode7[XDIR] = w_dGlobCoordR;
  qnode7[YDIR] = w_dGlobCoordT;
  qnode7[ZDIR] = w_dGlobCoordU;
  qnode8[XDIR] = w_dGlobCoordL;
  qnode8[YDIR] = w_dGlobCoordT;
  qnode8[ZDIR] = w_dGlobCoordU;
  qTarget[XDIR] = G_Target[XDIR];
  qTarget[YDIR] = G_Target[YDIR];
  qTarget[ZDIR] = G_Target[ZDIR];
  j_3Dworld_to_screen(lnode1, qnode1);
  j_3Dworld_to_screen(lnode2, qnode2);
  j_3Dworld_to_screen(lnode3, qnode3);
  j_3Dworld_to_screen(lnode4, qnode4);
  j_3Dworld_to_screen(lnode5, qnode5);
  j_3Dworld_to_screen(lnode6, qnode6);
  j_3Dworld_to_screen(lnode7, qnode7);
  j_3Dworld_to_screen(lnode8, qnode8);
  j_3Dworld_to_screen(lTarget, qTarget);
  j_3Dworld_to_screen(SO, OO);
  j_3Dworld_to_screen(SX, OX);
  j_3Dworld_to_screen(SY, OY);
  j_3Dworld_to_screen(SZ, OZ);

  /* draw the bounding box */
  XDrawLine(display, win, gc,
            (int) lnode1[XDIR], (int) lnode1[YDIR],
            (int) lnode2[XDIR], (int) lnode2[YDIR]);
  XDrawLine(display, win, gc,
            (int) lnode2[XDIR], (int) lnode2[YDIR],
            (int) lnode3[XDIR], (int) lnode3[YDIR]);
  XDrawLine(display, win, gc,
            (int) lnode3[XDIR], (int) lnode3[YDIR],
            (int) lnode4[XDIR], (int) lnode4[YDIR]);
  XDrawLine(display, win, gc,
            (int) lnode4[XDIR], (int) lnode4[YDIR],
            (int) lnode1[XDIR], (int) lnode1[YDIR]);
  XDrawLine(display, win, gc,
            (int) lnode5[XDIR], (int) lnode5[YDIR],
            (int) lnode6[XDIR], (int) lnode6[YDIR]);
  XDrawLine(display, win, gc,
            (int) lnode6[XDIR], (int) lnode6[YDIR],
            (int) lnode7[XDIR], (int) lnode7[YDIR]);
  XDrawLine(display, win, gc,
            (int) lnode7[XDIR], (int) lnode7[YDIR],
            (int) lnode8[XDIR], (int) lnode8[YDIR]);
  XDrawLine(display, win, gc,
            (int) lnode8[XDIR], (int) lnode8[YDIR],
            (int) lnode5[XDIR], (int) lnode5[YDIR]);
  XDrawLine(display, win, gc,
            (int) lnode1[XDIR], (int) lnode1[YDIR],
            (int) lnode5[XDIR], (int) lnode5[YDIR]);
  XDrawLine(display, win, gc,
            (int) lnode2[XDIR], (int) lnode2[YDIR],
            (int) lnode6[XDIR], (int) lnode6[YDIR]);
  XDrawLine(display, win, gc,
            (int) lnode3[XDIR], (int) lnode3[YDIR],
            (int) lnode7[XDIR], (int) lnode7[YDIR]);
  XDrawLine(display, win, gc,
            (int) lnode4[XDIR], (int) lnode4[YDIR],
            (int) lnode8[XDIR], (int) lnode8[YDIR]);

  /* draw target */
  fill_pixel = colourpixels[4];
  XSetForeground(display, gc, fill_pixel);
  T_length = w_width / 50;

  XDrawLine(display, win, gc,
            (int) lTarget[XDIR] - T_length / 2, (int) lTarget[YDIR],
            (int) lTarget[XDIR] + T_length / 2, (int) lTarget[YDIR]);
  XDrawLine(display, win, gc,
            (int) lTarget[XDIR], (int) lTarget[YDIR],
            (int) lTarget[XDIR], (int) lTarget[YDIR] + T_length);

  /* draw axes */
  fill_pixel = colourpixels[2];
  XSetForeground(display, gc, fill_pixel);
  XDrawLine(display, win, gc,
            (int) SO[XDIR], (int) SO[YDIR], (int) SX[XDIR], (int) SX[YDIR]);
  XDrawLine(display, win, gc,
            (int) SO[XDIR], (int) SO[YDIR], (int) SY[XDIR], (int) SY[YDIR]);
  XDrawLine(display, win, gc,
            (int) SO[XDIR], (int) SO[YDIR], (int) SZ[XDIR], (int) SZ[YDIR]);

  /* draw labels on axes */
  nLength = sprintf(szBuffer, "x");
  XDrawString(display, win, gc,
              (int) SX[XDIR], (int) SX[YDIR], szBuffer, nLength);
  nLength = sprintf(szBuffer, "y");
  XDrawString(display, win, gc,
              (int) SY[XDIR], (int) SY[YDIR], szBuffer, nLength);
  nLength = sprintf(szBuffer, "z");
  XDrawString(display, win, gc,
              (int) SZ[XDIR], (int) SZ[YDIR], szBuffer, nLength);
}

void draw_info()
{
  XFontStruct *font_info;
  int font_height;
  short nLength;
  char szBuffer[80];
  Window win = XtWindow(draw);

  j_load_font(display, &font_info);
  font_height = font_info->ascent + font_info->descent;
  XSetForeground(XtDisplay(draw), gc, WhitePixelOfScreen(XtScreen(draw)));

  nLength = sprintf(szBuffer, "SECT Research Group");
  XDrawString(display, pix, gc, 20, 30, szBuffer, nLength);
  nLength = sprintf(szBuffer, "Heriot-Watt University Edinburgh");
  XDrawString(display, pix, gc, 20, 30 + 1 * font_height, szBuffer, nLength);
  nLength = sprintf(szBuffer, "E_PlotX11 X11 viewer for E_library files");
  XDrawString(display, pix, gc, 20, 30 + 3 * font_height, szBuffer, nLength);
  nLength = sprintf(szBuffer, "v%s %s %s", VERSION, __DATE__, __TIME__);
  XDrawString(display, pix, gc, 20, 30 + 4 * font_height, szBuffer, nLength);

  XCopyArea(display, pix, win, gc, 0, 0, w_width, w_height, 0, 0);
}

char *GetCurrentDir()
{
  int allocate_size;
  char *current_dir, *check_dir;
  short home;

  home = 0;
  allocate_size = 1024;
  /* allocate space for the directory name */
  if ((current_dir = e_alloc_c_vect(allocate_size)) == NULL)
    e_GeneralError("no mem");

  check_dir = getcwd(current_dir, allocate_size);
  if (check_dir == NULL)
  {
    e_free(current_dir);
    e_GeneralError("can't get pwd");
  }

  return (current_dir);
}
