#define J_GRAPHICS

#include <stdio.h>
#include <math.h>

#include <e_lib.h>
extern int GetKeyword(char *line, char *word, int *LineNumber,
                      char *FullPathName, FILE *fp);
#include "cursors.h"
#include "jgr.h"

#define EPSILON 1.0e-7

extern MESH *mesh;
extern int *rmv_table;
extern Dimension w_width, w_height;
extern long int fill_pixel;
extern Widget draw;
extern unsigned long colourpixels[8 + MAXCOLS + 9];
static XColor wSDCols[MAXCOLS];
static XColor wSDAltCols[MAXALTCOLS];
static POLYGON *pg;
extern double Hidden_Proj_center[3];
extern int mdfopen;
extern void draw_mesh();

static void j_3Dvec_mult_matrix(j_3Dvector a, j_3Dvector b, j_3Drot_matrix M);
static void j_3Dinverse_rot_matrix(j_3Drot_matrix InvM, j_3Drot_matrix M);
static void j_3Dmatrix_mult(j_3Drot_matrix C, j_3Drot_matrix A,
                            j_3Drot_matrix B);
static void j_3Dfill_rot_matrix(j_3Drot_matrix M, double angle, short axis);
/*static void j_3Dspherical_coords(double *Dist, double *Azim, double *Elev,
                                 j_3Dvector cpoint, j_3Dvector tpoint);*/
/*static void j_3Dcalc_rot_matrix(void);*/
/*static double j_offset(double dx, double dy);*/
/*static int j_comp_tr(const void *a, const void *b);*/



static void j_3Dinverse_rot_matrix(j_3Drot_matrix InvM, j_3Drot_matrix M)
{
  register short i, j;

  for (i = 0; i < 3; i++)
    for (j = 0; j < 3; j++)
      InvM[i][j] = M[j][i];
}

static void j_3Dmatrix_mult(j_3Drot_matrix C, j_3Drot_matrix A,
                            j_3Drot_matrix B)
{
  register double *b = (double *) B;
  register j_3Dvector
    * a = (j_3Dvector *) A, *c = (j_3Dvector *) C, *hi_a = a + 3;

  for (; a < hi_a; a++, c++)
    M_Vec_dot_matrix(*c, *a, b);
}

static void j_3Dfill_rot_matrix(j_3Drot_matrix M, double angle, short axis)
{
  register double s, c;         /* for reasons of speed */

  s = sin(angle);
  c = cos(angle);

  switch (axis)
  {
    case XDIR:
      M[0][0] = 1;
      M[0][1] = 0;
      M[0][2] = 0;
      M[1][0] = 0;
      M[1][1] = c;
      M[1][2] = s;
      M[2][0] = 0;
      M[2][1] = -s;
      M[2][2] = c;
      return;
    case YDIR:
      M[0][0] = c;
      M[0][1] = 0;
      M[0][2] = -s;
      M[1][0] = 0;
      M[1][1] = 1;
      M[1][2] = 0;
      M[2][0] = s;
      M[2][1] = 0;
      M[2][2] = c;
      return;
    case ZDIR:
      M[0][0] = c;
      M[0][1] = s;
      M[0][2] = 0;
      M[1][0] = -s;
      M[1][1] = c;
      M[1][2] = 0;
      M[2][0] = 0;
      M[2][1] = 0;
      M[2][2] = 1;
      return;
  }
}

/*static void j_3Dcalc_rot_matrix(void)
{
  j_3Dspherical_coords(&G_dTC, &G_Azim, &G_Elev, G_Center, G_Target);
  j_3Drecalc_rot_matrix();
}*/

void j_3Drecalc_rot_matrix(void)
{
  double alpha, beta, gamma;    /* The Euler angles */
  j_3Drot_matrix z_alpha, x_beta, z_gamma, xz;

  alpha = G_Twist;
  beta = M_PI / 2.00 - G_Elev;
  gamma = G_Azim;

  j_3Dfill_rot_matrix(z_alpha, alpha, ZDIR);
  j_3Dfill_rot_matrix(x_beta, beta, XDIR);
  j_3Dfill_rot_matrix(z_gamma, gamma, ZDIR);
  j_3Dmatrix_mult(xz, z_alpha, x_beta);
  j_3Dmatrix_mult(G_cpim, xz, z_gamma);
  j_3Dinverse_rot_matrix(G_cpm, G_cpim);
}

/*static void j_3Dspherical_coords(double *Dist, double *Azim, double *Elev,
                                 j_3Dvector cpoint, j_3Dvector tpoint)
{
  j_3Dvector newpoint;

  M_Subtr_vec(newpoint, cpoint, tpoint);
  *Dist = M_Length(newpoint);
  *Azim = j_offset(newpoint[XDIR], newpoint[YDIR]);
  *Elev = asin((double) newpoint[ZDIR] / (*Dist + EPSILON));
}*/

void j_3Dworld_to_screen(j_3Dvector screen, j_3Dvector world)
{
  j_3Dvector world_t;
  double lambda;

  M_Subtr_vec(world_t, world, G_Target);
  j_3Dvec_mult_matrix(screen, world_t, G_cpm);
  lambda = G_scale * G_dTC / (G_dTC - screen[ZDIR]);
  screen[XDIR] = G_Xscrorigin + lambda * screen[XDIR];
  screen[YDIR] = G_Yscrorigin + G_pixel_ratio * lambda * screen[YDIR];
}

void p_3Dworld_to_screen(double *result, double *world)
{
  j_3Dvector world_t;

  M_Subtr_vec(world_t, world, G_Target);
  j_3Dvec_mult_matrix(result, world_t, G_cpm);
}


static void j_3Dvec_mult_matrix(j_3Dvector result, j_3Dvector p,
                                j_3Drot_matrix A)
{
  register double *a = (double *) A, *r = (double *) result, *hi_r = r + 3;

  for (; r < hi_r; r++, a++)
    *r = p[XDIR] * *(a) + p[YDIR] * *(a + 3) + p[ZDIR] * *(a + 6);
}

/*static double j_offset(double dx, double dy)
{
  double angle;

  if (fabs(dx) < EPSILON)
    if (fabs(dy) < EPSILON)   // point on the z-axis
      angle = 0.00;
    else if (dy > 0.0)
      angle = M_PI / 2.0;
    else
      angle = -M_PI / 2.0;
  else if (dx > 0.0)
    angle = atan(dy / dx);
  else
    angle = atan(dy / dx) + M_PI;

  angle += M_PI / 2;

  return (angle);
}*/

/*static int j_comp_tr(const void *a, const void *b)
{
  double **z1, **z2;

  z1 = (double **) a;
  z2 = (double **) b;

  if (*(z1[0]) < *(z2[0]))
    return (-1);
  if (*(z1[0]) > *(z2[0]))
    return (1);
  return (0);
}*/

void j_init_perspective(void)
{
  int i;

  XtVaGetValues(draw, XmNwidth, &w_width, XmNheight, &w_height, NULL);
  G_Xscrorigin = (w_width / 2);
  G_Yscrorigin = (w_height / 2);

  /* calculate extreme nodes for stress window display  */
  w_dGlobCoordB = mesh->MeshPointCoords[0][YDIR];
  w_dGlobCoordT = mesh->MeshPointCoords[0][YDIR];
  w_dGlobCoordL = mesh->MeshPointCoords[0][XDIR];
  w_dGlobCoordR = mesh->MeshPointCoords[0][XDIR];
  w_dGlobCoordU = mesh->MeshPointCoords[0][ZDIR];
  w_dGlobCoordD = mesh->MeshPointCoords[0][ZDIR];
  for (i = 1; i < mesh->NMeshPoints; i++)
  {
    if (w_dGlobCoordT < mesh->MeshPointCoords[i][YDIR])
      w_dGlobCoordT = mesh->MeshPointCoords[i][YDIR];
    if (w_dGlobCoordB > mesh->MeshPointCoords[i][YDIR])
      w_dGlobCoordB = mesh->MeshPointCoords[i][YDIR];
    if (w_dGlobCoordR < mesh->MeshPointCoords[i][XDIR])
      w_dGlobCoordR = mesh->MeshPointCoords[i][XDIR];
    if (w_dGlobCoordL > mesh->MeshPointCoords[i][XDIR])
      w_dGlobCoordL = mesh->MeshPointCoords[i][XDIR];
    if (w_dGlobCoordU < mesh->MeshPointCoords[i][ZDIR])
      w_dGlobCoordU = mesh->MeshPointCoords[i][ZDIR];
    if (w_dGlobCoordD > mesh->MeshPointCoords[i][ZDIR])
      w_dGlobCoordD = mesh->MeshPointCoords[i][ZDIR];
  }

  G_Target[XDIR] = w_dGlobCoordL + (w_dGlobCoordR - w_dGlobCoordL) / 2.00;
  G_Target[YDIR] = w_dGlobCoordB + (w_dGlobCoordT - w_dGlobCoordB) / 2.00;
  G_Target[ZDIR] = w_dGlobCoordD + (w_dGlobCoordU - w_dGlobCoordD) / 2.00;;
  G_Twist = 0.00;
  G_Azim = INIT_AZIMUTH * M_PI / 180.00;
  G_Elev = INIT_ELEVATION * M_PI / 180.00;

  j_3Drecalc_rot_matrix();

  w_maxsize =
    E_MAX((w_dGlobCoordR - w_dGlobCoordL), (w_dGlobCoordT - w_dGlobCoordB));
  w_maxsize = E_MAX(w_maxsize, (w_dGlobCoordU - w_dGlobCoordD));
  G_scale = G_initscale = ((double) w_height) / 2.00 / w_maxsize;
  G_dTC = 2.00 * w_maxsize;
  G_pixel_ratio = -1.00;

  OO[XDIR] = 0.00;
  OO[YDIR] = 0.00;
  OO[ZDIR] = 0.00;
  OX[XDIR] = w_maxsize / 2.00;
  OX[YDIR] = 0.00;
  OX[ZDIR] = 0.00;
  OY[XDIR] = 0.00;
  OY[YDIR] = w_maxsize / 2.00;
  OY[ZDIR] = 0.00;
  OZ[XDIR] = 0.00;
  OZ[YDIR] = 0.00;
  OZ[ZDIR] = w_maxsize / 2.00;

  wZeta = INIT_ZETA;
}

void graphics_info(void)
{
  /* print some info on the current status of the graphics vars */

  XtVaGetValues(draw, XmNwidth, &w_width, XmNheight, &w_height, NULL);
  printf("\n\n");
  printf("# e_plotX11 Graphics Info\n");
  printf("# current file: %s.mdf\n\n", mdfname);
  printf("VERSION          0.4\n\n");
  printf("# SCREEN_WIDTH     %d\n", w_width);
  printf("# SCREEN_HEIGHT    %d\n\n", w_height);
  printf("SCR_ORIGIN_X     %d\n", (int) G_Xscrorigin);
  printf("SCR_ORIGIN_Y     %d\n\n", (int) G_Yscrorigin);
  printf("TARGET_X         %1.5f\n", G_Target[XDIR]);
  printf("TARGET_Y         %1.5f\n", G_Target[YDIR]);
  printf("TARGET_Z         %1.5f\n\n", G_Target[ZDIR]);
  printf("AZIMUT           %1.5f\n", G_Azim);
  printf("ELEVATION        %1.5f\n\n", G_Elev);
  /* printf("TWIST            %1.5f\n",G_Twist); */
  printf("SCALE            %1.5f\n", G_scale);
  printf("DTC              %1.5f\n", G_dTC);
  printf("PLOT_OPT         %s\n",
         (wPlotSelect == PS_SUBDOM) ? "\"SUBDOM\"" :
         (wPlotSelect == PS_PLAINMESH) ? "\"PLAIN\"" : "");
  printf("TILING           %s\n", (wTileElems) ? "\"TRUE\"" : "\"FALSE\"");
  if (wTileElems)
    printf("TILING_PERC      %1.1f\n", w_Element_Size_percent);
  printf("BORDERS          %s\n", (wNoBorders) ? "\"FALSE\"" : "\"TRUE\"");
  printf("HILIGHT_NODES    %s\n", (wHN) ? "\"TRUE\"" : "\"FALSE\"");
  printf("CANVAS_WHITE     \"FALSE\"");
  printf("\n\n");
  fflush(stdout);
}

void save_graphics_info(void)
{
  char *fullfilename;
  FILE *fp;

  fullfilename = e_alloc_c_vect(strlen(mdfname) + 5);
  sprintf(fullfilename, "%s.gra", mdfname);
  if ((fp = fopen(fullfilename, "w")) == NULL)
  {
    printf("\n\ncouldn't save graphics file: %s\n\n", fullfilename);
    return;
  }

  XtVaGetValues(draw, XmNwidth, &w_width, XmNheight, &w_height, NULL);
  fprintf(fp, "# e_plotX11 Graphics Info\n");
  fprintf(fp, "# current file: %s.mdf\n\n", mdfname);
  fprintf(fp, "VERSION          0.4\n\n");
  fprintf(fp, "# SCREEN_WIDTH     %d\n", w_width);
  fprintf(fp, "# SCREEN_HEIGHT    %d\n\n", w_height);
  fprintf(fp, "SCR_ORIGIN_X     %d\n", (int) G_Xscrorigin);
  fprintf(fp, "SCR_ORIGIN_Y     %d\n\n", (int) G_Yscrorigin);
  fprintf(fp, "TARGET_X         %1.5f\n", G_Target[XDIR]);
  fprintf(fp, "TARGET_Y         %1.5f\n", G_Target[YDIR]);
  fprintf(fp, "TARGET_Z         %1.5f\n\n", G_Target[ZDIR]);
  fprintf(fp, "AZIMUT           %1.5f\n", G_Azim);
  fprintf(fp, "ELEVATION        %1.5f\n\n", G_Elev);
  /* fprintf(fp,"TWIST            %1.5f\n",G_Twist); */
  fprintf(fp, "SCALE            %1.5f\n", G_scale);
  fprintf(fp, "DTC              %1.5f\n", G_dTC);
  fprintf(fp, "PLOT_OPT         %s\n",
          (wPlotSelect == PS_SUBDOM) ? "\"SUBDOM\"" :
          (wPlotSelect == PS_PLAINMESH) ? "\"PLAIN\"" : "");
  fprintf(fp, "TILING           %s\n",
          (wTileElems) ? "\"TRUE\"" : "\"FALSE\"");
  if (wTileElems)
    fprintf(fp, "TILING_PERC      %1.1f\n", w_Element_Size_percent);
  fprintf(fp, "BORDERS          %s\n",
          (wNoBorders) ? "\"FALSE\"" : "\"TRUE\"");
  fprintf(fp, "HILIGHT_NODES    %s\n", (wHN) ? "\"TRUE\"" : "\"FALSE\"");
  fprintf(fp, "CANVAS_WHITE     \"FALSE\"");
  fprintf(fp, "\n\n");

  fclose(fp);
}

void load_graphics_info(void)
{
  /* load the graphics settings and apply them */
  char *fullfilename;
  FILE *fp;
  char line[E_LINE_MAX];
  char word[E_LINE_MAX];
  int ln = 0;

  int Xscrorigin;
  int Yscrorigin;
  double ver;
  char sval[E_LINE_MAX];

  fullfilename = e_alloc_c_vect(strlen(mdfname) + 5);
  sprintf(fullfilename, "%s.gra", mdfname);
  if ((fp = fopen(fullfilename, "r")) == NULL)
  {
    printf("\n\ncouldn't load graphics file: %s\n\n", fullfilename);
    return;
  }

  if (!GetKeyword(line, word, &ln, mdfname, fp))
    e_LineError(ln, "can't read the version keyword");
  if (strcmp(word, "VERSION") != 0)
    e_LineError(ln, "version keyword incorrect");
  if (sscanf(line, "%*s %lf", &ver) != 1)
    e_LineError(ln, "version number value incorrect or missing");
  if (ver < 0.3)
    e_GeneralError("eplx requires version>=0.3 graphics info");

  if (!GetKeyword(line, word, &ln, mdfname, fp))
    e_LineError(ln, "can't read x screen origin keyword");
  if (strcmp(word, "SCR_ORIGIN_X") != 0)
    e_LineError(ln, "x screen origin keyword incorrect");
  if (sscanf(line, "%*s %d", &Xscrorigin) != 1)
    e_LineError(ln, "x screen origin value incorrect or missing");
  G_Xscrorigin = Xscrorigin;

  if (!GetKeyword(line, word, &ln, mdfname, fp))
    e_LineError(ln, "can't read y screen origin keyword");
  if (strcmp(word, "SCR_ORIGIN_Y") != 0)
    e_LineError(ln, "y screen origin keyword incorrect");
  if (sscanf(line, "%*s %d", &Yscrorigin) != 1)
    e_LineError(ln, "y screen origin value incorrect or missing");
  G_Yscrorigin = Yscrorigin;

  if (!GetKeyword(line, word, &ln, mdfname, fp))
    e_LineError(ln, "can't read target x keyword");
  if (strcmp(word, "TARGET_X") != 0)
    e_LineError(ln, "target x keyword incorrect");
  if (sscanf(line, "%*s %lf", &(G_Target[XDIR])) != 1)
    e_LineError(ln, "target x value incorrect or missing");

  if (!GetKeyword(line, word, &ln, mdfname, fp))
    e_LineError(ln, "can't read target y keyword");
  if (strcmp(word, "TARGET_Y") != 0)
    e_LineError(ln, "target y keyword incorrect");
  if (sscanf(line, "%*s %lf", &(G_Target[YDIR])) != 1)
    e_LineError(ln, "target y value incorrect or missing");

  if (!GetKeyword(line, word, &ln, mdfname, fp))
    e_LineError(ln, "can't read target z keyword");
  if (strcmp(word, "TARGET_Z") != 0)
    e_LineError(ln, "target z keyword incorrect");
  if (sscanf(line, "%*s %lf", &(G_Target[ZDIR])) != 1)
    e_LineError(ln, "target z value incorrect or missing");

  if (!GetKeyword(line, word, &ln, mdfname, fp))
    e_LineError(ln, "can't read azimut keyword");
  if (strcmp(word, "AZIMUT") != 0)
    e_LineError(ln, "azimut keyword incorrect");
  if (sscanf(line, "%*s %lf", &G_Azim) != 1)
    e_LineError(ln, "azimut value incorrect or missing");

  if (!GetKeyword(line, word, &ln, mdfname, fp))
    e_LineError(ln, "can't read elevation keyword");
  if (strcmp(word, "ELEVATION") != 0)
    e_LineError(ln, "elevation keyword incorrect");
  if (sscanf(line, "%*s %lf", &G_Elev) != 1)
    e_LineError(ln, "elevation value incorrect or missing");

  if (!GetKeyword(line, word, &ln, mdfname, fp))
    e_LineError(ln, "can't read scale keyword");
  if (strcmp(word, "SCALE") != 0)
    e_LineError(ln, "scale keyword incorrect");
  if (sscanf(line, "%*s %lf", &G_scale) != 1)
    e_LineError(ln, "scale value incorrect or missing");

  if (!GetKeyword(line, word, &ln, mdfname, fp))
    e_LineError(ln, "can't read dtc keyword");
  if (strcmp(word, "DTC") != 0)
    e_LineError(ln, "dtc keyword incorrect");
  if (sscanf(line, "%*s %lf", &G_dTC) != 1)
    e_LineError(ln, "dtc value incorrect or missing");

  if (!GetKeyword(line, word, &ln, mdfname, fp))
    e_LineError(ln, "can't read plot option keyword");
  if (strcmp(word, "PLOT_OPT") != 0)
    e_LineError(ln, "plot option keyword incorrect");
  if (sscanf(line, "%*s %s", sval) != 1)
    e_LineError(ln, "plot option value incorrect or missing");
  if ((strcmp(sval, "\"SUBDOM\"") != 0) && (strcmp(sval, "\"PLAIN\"") != 0))
    e_LineError(ln, "plot option value not recognized or not supported");
  if (strcmp(sval, "\"SUBDOM\"") == 0)
  {
    if (!wPlotSelect)
      if (!e_GetDecompData(mesh, mdfname))
        e_GeneralError("graphics option requires loading of decomp data");
    wPlotSelect = PS_SUBDOM;
    XtVaSetValues(wg_Plain, XmNset, FALSE, NULL);
    XtVaSetValues(wg_Subdom, XmNset, TRUE, NULL);
  }
  else if (strcmp(sval, "\"PLAIN\"") == 0)
  {
    wPlotSelect = PS_PLAINMESH;
    XtVaSetValues(wg_Plain, XmNset, TRUE, NULL);
    XtVaSetValues(wg_Subdom, XmNset, FALSE, NULL);
  }

  if (!GetKeyword(line, word, &ln, mdfname, fp))
    e_LineError(ln, "can't read plot tiling keyword");
  if (strcmp(word, "TILING") != 0)
    e_LineError(ln, "tiling keyword incorrect");
  if (sscanf(line, "%*s %s", sval) != 1)
    e_LineError(ln, "tiling value incorrect or missing");
  if ((strcmp(sval, "\"FALSE\"") != 0) && (strcmp(sval, "\"TRUE\"") != 0))
    e_LineError(ln, "tiling value not recognized or not supported");
  if (strcmp(sval, "\"FALSE\"") == 0)
  {
    wTileElems = FALSE;
    XtVaSetValues(wg_TE, XmNset, FALSE, NULL);
    XtVaSetValues(wg_TL, XmNset, FALSE, NULL);
  }
  else
  {
    wTileElems = TRUE;
    XtVaSetValues(wg_TE, XmNset, TRUE, NULL);
    XtVaSetValues(wg_TL, XmNset, TRUE, NULL);
    if (!GetKeyword(line, word, &ln, mdfname, fp))
      e_LineError(ln, "can't read tiling percentage keyword");
    if (strcmp(word, "TILING_PERC") != 0)
      e_LineError(ln, "tiling percentage keyword incorrect");
    if (sscanf(line, "%*s %lf", &w_Element_Size_percent) != 1)
      e_LineError(ln, "tiling percentage value incorrect or missing");
    w_Link_Size_percent = w_Element_Size_percent;
  }

  if (!GetKeyword(line, word, &ln, mdfname, fp))
    e_LineError(ln, "can't read borders keyword");
  if (strcmp(word, "BORDERS") != 0)
    e_LineError(ln, "borders keyword incorrect");
  if (sscanf(line, "%*s %s", sval) != 1)
    e_LineError(ln, "borders value incorrect or missing");
  if ((strcmp(sval, "\"FALSE\"") != 0) && (strcmp(sval, "\"TRUE\"") != 0))
    e_LineError(ln, "borders value not recognized or not supported");
  if (strcmp(sval, "\"FALSE\"") == 0)
  {
    wNoBorders = TRUE;
    XtVaSetValues(wg_NB, XmNset, TRUE, NULL);
  }
  else
  {
    wNoBorders = FALSE;
    XtVaSetValues(wg_NB, XmNset, FALSE, NULL);
  }

  if (!GetKeyword(line, word, &ln, mdfname, fp))
    e_LineError(ln, "can't read hilight nodes keyword");
  if (strcmp(word, "HILIGHT_NODES") != 0)
    e_LineError(ln, "hilight nodes keyword incorrect");
  if (sscanf(line, "%*s %s", sval) != 1)
    e_LineError(ln, "hilight nodes value incorrect or missing");
  if ((strcmp(sval, "\"FALSE\"") != 0) && (strcmp(sval, "\"TRUE\"") != 0))
    e_LineError(ln, "hilight nodes value not recognized or not supported");
  if (strcmp(sval, "\"FALSE\"") == 0)
  {
    wHN = FALSE;
    XtVaSetValues(wg_HN, XmNset, FALSE, NULL);
  }
  else
  {
    wHN = TRUE;
    XtVaSetValues(wg_HN, XmNset, TRUE, NULL);
  }

  /* canvas white property ignored */

  fclose(fp);
  j_3Drecalc_rot_matrix();
  draw_mesh();
}

int j_do_projection(POLYGON * polygons, int NPoly, double *zmax, double *zmin)
{
  /* this function converts the 3D coords in the polygons to
     projected 2D coords */

  int p;                        /* polygon index */
  int v;                        /* vertex index */
  int d;                        /* dimension index */
  double centr[3];              /* center for tiling */
  double zcentr[3];             /* center for z-sorting algo */
  double worldnode[3];
  double screen_centr[3];
  double avsl;                  /* average side length */

  (*zmin) = 1e10;
  (*zmax) = -1e10;

  for (p = 0; p < NPoly; p++)
  {
    for (d = 0; d < 3; d++)
    {
      centr[d] = 0;
      if (polygons[p].elem_type == E_ELEM_TYPE_TETRAH1)
      {
        /* for tetrahedrals tiling is centered around the centroid of the tetrahedral
           not the centroid of the 4 faces individually */
        for (v = 0; v < 4; v++)
          centr[d] +=
            mesh->MeshPointCoords[mesh->
                                  NodesTetrah1[polygons[p].mesh_index][v] -
                                  1][d];
        centr[d] /= 4.0;
      }
      else if (polygons[p].elem_type == E_ELEM_TYPE_BLOCK1)
      {
        for (v = 0; v < 8; v++)
          centr[d] +=
            mesh->MeshPointCoords[mesh->
                                  NodesBlock1[polygons[p].mesh_index][v] -
                                  1][d];
        centr[d] /= 8.0;
      }
      else
      {
        for (v = 0; v < polygons[p].no_vertices; v++)
          centr[d] += polygons[p].world_coords[v][d];
        centr[d] /= (double) polygons[p].no_vertices;
      }
    }
    polygons[p].vis = FALSE;
    for (d = 0; d < 3; d++)
      zcentr[d] = 0;
    for (v = 0; v < polygons[p].no_vertices; v++)
    {
      for (d = 0; d < 3; d++)
      {
        if ((polygons[p].elem_type == E_ELEM_TYPE_LINK1) ||
            (polygons[p].elem_type == E_ELEM_TYPE_LINK2) ||
            (polygons[p].elem_type == E_ELEM_TYPE_LINK3) ||
            (polygons[p].elem_type == E_ELEM_TYPE_LINK4) ||
            (polygons[p].elem_type == E_ELEM_TYPE_LINK5))
          worldnode[d] =
            centr[d] + (polygons[p].world_coords[v][d] -
                        centr[d]) * w_Link_Size_percent / 100.00;
        else
          worldnode[d] =
            centr[d] + (polygons[p].world_coords[v][d] -
                        centr[d]) * w_Element_Size_percent / 100.00;
        zcentr[d] += worldnode[d];
      }
      j_3Dworld_to_screen(polygons[p].screen_coords[v], worldnode);

      if (((short) polygons[p].screen_coords[v][0] > 0) ||
          ((short) polygons[p].screen_coords[v][0] < w_width) ||
          ((short) polygons[p].screen_coords[v][1] > 0) ||
          ((short) polygons[p].screen_coords[v][1] < w_height))
        polygons[p].vis = TRUE;
    }
    for (d = 0; d < 3; d++)
      zcentr[d] /= polygons[p].no_vertices;
    j_3Dworld_to_screen(screen_centr, zcentr);
    polygons[p].z = screen_centr[2];
    avsl = 0.0;
    for (v = 0; v < polygons[p].no_vertices; v++)
      avsl +=
        j_getavsl(polygons[p].world_coords[v],
                  polygons[p].world_coords[(v + 1) %
                                           polygons[p].no_vertices]);
    avsl /= polygons[p].no_vertices;
    polygons[p].avsl = avsl;
    if (polygons[p].z < (*zmin))
      (*zmin) = polygons[p].z;
    if (polygons[p].z > (*zmax))
      (*zmax) = polygons[p].z;
  }
  return 1;
}

void j_sort_polygons(int *order, POLYGON * polygons, int Npoly)
{
    /** pg is the polygons temporarily made global **/
  pg = polygons;

  qsort((void *) order, Npoly, sizeof(int), j_compare_polygons);
  return;
}

int j_compare_polygons(const void *i, const void *j)
{
    /** pg is the polygons temporarily made global **/
  int *ii, *jj;

  ii = (int *) i;
  jj = (int *) j;

  if (pg[*ii].z > pg[*jj].z)
    return (1);
  if (pg[*ii].z < pg[*jj].z)
    return (-1);
  return (0);
}

double j_getavsl(j_3Dvector node1, j_3Dvector node2)
{
  double dx, dy, dz;

  dx = node1[XDIR] - node2[XDIR];
  dy = node1[YDIR] - node2[YDIR];
  dz = node1[ZDIR] - node2[ZDIR];

  return sqrt(pow(dx, 2) + pow(dy, 2) + pow(dz, 2));
}

void j_paint_polygons(Display * dspl, Window wnd, GC gcx, Colormap cmp,
                      POLYGON * polygons, int Npoly)
{
  j_3Dvector tri_cent, scrtri_cent;
  short nLength;
  char szBuffer[80];
  int p, i, e, s;
  int n;
  int line_style = LineSolid;
  unsigned int line_width = 1;
  int cap_style = CapRound;
  int join_style = JoinRound;
  XFontStruct *font_info;
  int font_height;
  XPoint pts[4];
  double zmax, zmin;

  /* for debugging purposes 
  double target_check[3];
  double centre_check[3];
  extern double Hidden_Proj_center[3]; */

  set_watch_cursor(draw);

  if ((wDisplElemNo) || (wDisplPointNo))
  {
    j_load_font(dspl, &font_info);
    font_height = font_info->ascent + font_info->descent;
  }

  j_do_projection(polygons, Npoly, &zmax, &zmin);

  if (wAdvHidden)
  {
    setup_hidden_line2(Npoly, polygons);
    setup_proj_center(G_Azim, G_Elev, G_dTC, G_Target);

    /* j_3Dworld_to_screen(target_check, G_Target);
       j_3Dworld_to_screen(centre_check, Hidden_Proj_center);

       printf("\nTarget:       %lf %lf", target_check[0],
       target_check[1]  );
       printf("\nProj. Centre: %lf %lf", centre_check[0],
       centre_check[1]  ); */

    if (!sort_objects(order, Npoly, polygons))
    {
      e_GeneralWarning("advanced sorting failed");
      for (p = 0; p < Npoly; p++)
          order[p] = p;
    }
  }
  else
  {
    for (p = 0; p < Npoly; p++)
      order[p] = p;
    if (wHidden)
      j_sort_polygons(order, polygons, Npoly);
  }

  for (e = 0; e < Npoly; e++)
  {
    p = order[e];
    if (wRemove)
      if (!rmv_table
          [mesh->FirstNetIndex[polygons[p].elem_type] +
           polygons[p].mesh_index - 1])
        continue;

/*	printf("order[%d]=%d\n",e,p); */
    if (!polygons[p].vis)
      continue;
    if (wClipPlane)
      if (polygons[p].z > 0.0)
        continue;
    if ((polygons[p].elem_type == E_ELEM_TYPE_LINK1) ||
        (polygons[p].elem_type == E_ELEM_TYPE_LINK2) ||
        (polygons[p].elem_type == E_ELEM_TYPE_LINK3) ||
        (polygons[p].elem_type == E_ELEM_TYPE_LINK4) ||
        (polygons[p].elem_type == E_ELEM_TYPE_LINK5))
    {
      if (wPlotSelect == PS_PLAINMESH)
      {
        if (wTileElems)
        {
          if (polygons[p].elem_type == E_ELEM_TYPE_LINK1)
            fill_pixel = colourpixels[2];
          else if (polygons[p].elem_type == E_ELEM_TYPE_LINK2)
            fill_pixel = colourpixels[3];
          else if (polygons[p].elem_type == E_ELEM_TYPE_LINK3)
            fill_pixel = colourpixels[4];
          else if (polygons[p].elem_type == E_ELEM_TYPE_LINK4)
            fill_pixel = colourpixels[5];
          else if (polygons[p].elem_type == E_ELEM_TYPE_LINK5)
            fill_pixel = colourpixels[6];
        }
        else
          fill_pixel = colourpixels[2];
      }
      else if (wPlotSelect == PS_SUBDOM)
      {
        if (!wAltCols)
        {

          if (polygons[p].elem_type == E_ELEM_TYPE_LINK1)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsLink1[polygons[p].mesh_index] -
                             1) % MAXCOLS) + 8];
          else if (polygons[p].elem_type == E_ELEM_TYPE_LINK2)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsLink2[polygons[p].mesh_index] -
                             1) % MAXCOLS) + 8];
          else if (polygons[p].elem_type == E_ELEM_TYPE_LINK3)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsLink3[polygons[p].mesh_index] -
                             1) % MAXCOLS) + 8];
          else if (polygons[p].elem_type == E_ELEM_TYPE_LINK4)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsLink4[polygons[p].mesh_index] -
                             1) % MAXCOLS) + 8];
          else if (polygons[p].elem_type == E_ELEM_TYPE_LINK5)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsLink5[polygons[p].mesh_index] -
                             1) % MAXCOLS) + 8];
        }
        else
        {
          if (polygons[p].elem_type == E_ELEM_TYPE_LINK1)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsLink1[polygons[p].mesh_index] -
                             1) % MAXALTCOLS) + 17 + MAXCOLS];
          else if (polygons[p].elem_type == E_ELEM_TYPE_LINK2)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsLink2[polygons[p].mesh_index] -
                             1) % MAXALTCOLS) + 17 + MAXCOLS];
          else if (polygons[p].elem_type == E_ELEM_TYPE_LINK3)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsLink3[polygons[p].mesh_index] -
                             1) % MAXALTCOLS) + 17 + MAXCOLS];
          else if (polygons[p].elem_type == E_ELEM_TYPE_LINK4)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsLink4[polygons[p].mesh_index] -
                             1) % MAXALTCOLS) + 17 + MAXCOLS];
          else if (polygons[p].elem_type == E_ELEM_TYPE_LINK5)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsLink5[polygons[p].mesh_index] -
                             1) % MAXALTCOLS) + 17 + MAXCOLS];
        }
      }

      XSetForeground(dspl, gcx, fill_pixel);
      XDrawLine(dspl, wnd, gcx,
                (short) (polygons[p].screen_coords[0][0]),
                (short) (polygons[p].screen_coords[0][1]),
                (short) (polygons[p].screen_coords[1][0]),
                (short) (polygons[p].screen_coords[1][1]));

      if (wDisplElemNo)
      {
        fill_pixel = colourpixels[3];
        XSetForeground(dspl, gcx, fill_pixel);
        nLength = sprintf(szBuffer, "%d", polygons[p].mesh_index + 1);
        XDrawString(dspl, wnd, gcx,
                    (short) ((polygons[p].screen_coords[0][0] +
                              polygons[p].screen_coords[1][0]) / 2.0),
                    (short) ((polygons[p].screen_coords[0][1] +
                              polygons[p].screen_coords[1][1]) / 2.0),
                    szBuffer, nLength);
      }
    }
    else if ((polygons[p].elem_type == E_ELEM_TYPE_TRIANG1) ||
             (polygons[p].elem_type == E_ELEM_TYPE_TRIANG2) ||
             (polygons[p].elem_type == E_ELEM_TYPE_TRIANG3) ||
             (polygons[p].elem_type == E_ELEM_TYPE_TRIANG4) ||
             (polygons[p].elem_type == E_ELEM_TYPE_TRIANG5) ||
             (polygons[p].elem_type == E_ELEM_TYPE_TETRAH1))
    {
      for (s = 0; s < 3; s++)
      {
        pts[s].x = (short) (polygons[p].screen_coords[s][0]);
        pts[s].y = (short) (polygons[p].screen_coords[s][1]);
      }
      if (wPlotSelect == PS_PLAINMESH)
      {
        if ((wHidden) || (wAdvHidden))
        {
          fill_pixel = colourpixels[1];
          if (wZC)
          {
            /* fill_pixel = colourpixels[8+MAXCOLS+(int)(9.0*((double)e)/((double)Npoly))]; */
            zmin = zmin - 0.001 * (zmax - zmin);
            zmax = zmax + 0.001 * (zmax - zmin);
            for (s = 0; s < 9; s++)
            {
              if ((polygons[p].z >= zmin + (zmax - zmin) / 9.0 * (double) s)
                  && (polygons[p].z <=
                      zmin + (zmax - zmin) / 9.0 * (double) (s + 1)))
                break;
            }
            if (s == 9)
              e_GeneralError("z colouring problem: %1.4lf (%1.4lf,%1.4lf)",
                             polygons[p].z, zmin, zmax);
            fill_pixel = colourpixels[8 + MAXCOLS + s];
          }
          XSetForeground(dspl, gcx, fill_pixel);
          XFillPolygon(dspl, wnd, gcx, pts, 3, Convex, CoordModeOrigin);
        }
        fill_pixel = colourpixels[2];
        XSetForeground(dspl, gcx, fill_pixel);
        if (!wNoBorders)
        {
          for (s = 0; s < 3; s++)
            XDrawLine(dspl, wnd, gcx,
                      (short) (polygons[p].screen_coords[s][0]),
                      (short) (polygons[p].screen_coords[s][1]),
                      (short) (polygons[p].
                               screen_coords[(s +
                                              1) %
                                             polygons[p].no_vertices][0]),
                      (short) (polygons[p].
                               screen_coords[(s +
                                              1) %
                                             polygons[p].no_vertices][1]));
        }
      }
      else if (wPlotSelect == PS_SUBDOM)
      {
        if (!wAltCols)
        {
          if (polygons[p].elem_type == E_ELEM_TYPE_TRIANG1)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsTriang1[polygons[p].mesh_index] -
                             1) % MAXCOLS) + 8];
          else if (polygons[p].elem_type == E_ELEM_TYPE_TRIANG2)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsTriang2[polygons[p].mesh_index] -
                             1) % MAXCOLS) + 8];
          else if (polygons[p].elem_type == E_ELEM_TYPE_TRIANG3)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsTriang3[polygons[p].mesh_index] -
                             1) % MAXCOLS) + 8];
          else if (polygons[p].elem_type == E_ELEM_TYPE_TRIANG4)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsTriang4[polygons[p].mesh_index] -
                             1) % MAXCOLS) + 8];
          else if (polygons[p].elem_type == E_ELEM_TYPE_TRIANG5)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsTriang5[polygons[p].mesh_index] -
                             1) % MAXCOLS) + 8];
          else if (polygons[p].elem_type == E_ELEM_TYPE_TETRAH1)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsTetrah1[polygons[p].mesh_index] -
                             1) % MAXCOLS) + 8];
        }
        else
        {
          if (polygons[p].elem_type == E_ELEM_TYPE_TRIANG1)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsTriang1[polygons[p].mesh_index] -
                             1) % MAXALTCOLS) + 17 + MAXCOLS];
          else if (polygons[p].elem_type == E_ELEM_TYPE_TRIANG2)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsTriang2[polygons[p].mesh_index] -
                             1) % MAXALTCOLS) + 17 + MAXCOLS];
          else if (polygons[p].elem_type == E_ELEM_TYPE_TRIANG3)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsTriang3[polygons[p].mesh_index] -
                             1) % MAXALTCOLS) + 17 + MAXCOLS];
          else if (polygons[p].elem_type == E_ELEM_TYPE_TRIANG4)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsTriang4[polygons[p].mesh_index] -
                             1) % MAXALTCOLS) + 17 + MAXCOLS];
          else if (polygons[p].elem_type == E_ELEM_TYPE_TRIANG5)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsTriang5[polygons[p].mesh_index] -
                             1) % MAXALTCOLS) + 17 + MAXCOLS];
          else if (polygons[p].elem_type == E_ELEM_TYPE_TETRAH1)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsTetrah1[polygons[p].mesh_index] -
                             1) % MAXALTCOLS) + 17 + MAXCOLS];
        }
        XSetForeground(dspl, gcx, fill_pixel);
        XFillPolygon(dspl, wnd, gcx, pts, 3, Convex, CoordModeOrigin);

        if (!wNoBorders)
        {
          fill_pixel = colourpixels[0];
          XSetForeground(dspl, gcx, fill_pixel);
          for (s = 0; s < 3; s++)
            XDrawLine(dspl, wnd, gcx,
                      (short) (polygons[p].screen_coords[s][0]),
                      (short) (polygons[p].screen_coords[s][1]),
                      (short) (polygons[p].
                               screen_coords[(s +
                                              1) %
                                             polygons[p].no_vertices][0]),
                      (short) (polygons[p].
                               screen_coords[(s +
                                              1) %
                                             polygons[p].no_vertices][1]));
        }
      }

      if (wDisplElemNo)
      {
        fill_pixel = colourpixels[3];
        XSetForeground(dspl, gcx, fill_pixel);
        nLength = sprintf(szBuffer, "%d", polygons[p].mesh_index + 1);
        XDrawString(dspl, wnd, gcx,
                    (short) ((polygons[p].screen_coords[0][0] +
                              polygons[p].screen_coords[1][0] +
                              polygons[p].screen_coords[2][0]) / 3.0),
                    (short) ((polygons[p].screen_coords[0][1] +
                              polygons[p].screen_coords[1][1] +
                              polygons[p].screen_coords[2][1]) / 3.0),
                    szBuffer, nLength);
      }
    }
    else if ((polygons[p].elem_type == E_ELEM_TYPE_QUAD1) ||
             (polygons[p].elem_type == E_ELEM_TYPE_QUAD3) ||
             (polygons[p].elem_type == E_ELEM_TYPE_BLOCK1))
    {
      for (s = 0; s < 4; s++)
      {
        pts[s].x = (short) (polygons[p].screen_coords[s][0]);
        pts[s].y = (short) (polygons[p].screen_coords[s][1]);
      }
      if (wPlotSelect == PS_PLAINMESH)
      {
        if ((wHidden) || (wAdvHidden))
        {
          fill_pixel = colourpixels[1];
          if (wZC)
          {
            /* fill_pixel = colourpixels[8+MAXCOLS+(int)(9.0*((double)e)/((double)Npoly))]; */
            zmin = zmin - 0.001 * (zmax - zmin);
            zmax = zmax + 0.001 * (zmax - zmin);
            for (s = 0; s < 9; s++)
            {
              if ((polygons[p].z >= zmin + (zmax - zmin) / 9.0 * (double) s)
                  && (polygons[p].z <=
                      zmin + (zmax - zmin) / 9.0 * (double) (s + 1)))
                break;
            }
            if (s == 9)
              e_GeneralError("z colouring problem: %1.4lf (%1.4lf,%1.4lf)",
                             polygons[p].z, zmin, zmax);
            fill_pixel = colourpixels[8 + MAXCOLS + s];
          }
          XSetForeground(dspl, gcx, fill_pixel);
          XFillPolygon(dspl, wnd, gcx, pts, 4, Convex, CoordModeOrigin);
        }
        fill_pixel = colourpixels[2];
        XSetForeground(dspl, gcx, fill_pixel);
        if (!wNoBorders)
        {
          for (s = 0; s < 4; s++)
            XDrawLine(dspl, wnd, gcx,
                      (short) (polygons[p].screen_coords[s][0]),
                      (short) (polygons[p].screen_coords[s][1]),
                      (short) (polygons[p].
                               screen_coords[(s +
                                              1) %
                                             polygons[p].no_vertices][0]),
                      (short) (polygons[p].
                               screen_coords[(s +
                                              1) %
                                             polygons[p].no_vertices][1]));
        }
      }
      else if (wPlotSelect == PS_SUBDOM)
      {
        if (!wAltCols)
        {
          if (polygons[p].elem_type == E_ELEM_TYPE_QUAD1)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsQuad1[polygons[p].mesh_index] -
                             1) % MAXCOLS) + 8];
          else if (polygons[p].elem_type == E_ELEM_TYPE_QUAD3)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsQuad3[polygons[p].mesh_index] -
                             1) % MAXCOLS) + 8];
          else if (polygons[p].elem_type == E_ELEM_TYPE_BLOCK1)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsBlock1[polygons[p].mesh_index] -
                             1) % MAXCOLS) + 8];
        }
        else
        {
          if (polygons[p].elem_type == E_ELEM_TYPE_QUAD1)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsQuad1[polygons[p].mesh_index] -
                             1) % MAXALTCOLS) + 17 + MAXCOLS];
          else if (polygons[p].elem_type == E_ELEM_TYPE_QUAD3)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsQuad3[polygons[p].mesh_index] -
                             1) % MAXALTCOLS) + 17 + MAXCOLS];
          else if (polygons[p].elem_type == E_ELEM_TYPE_BLOCK1)
            fill_pixel =
              colourpixels[((mesh->Decomp.
                             SubDomsBlock1[polygons[p].mesh_index] -
                             1) % MAXALTCOLS) + 17 + MAXCOLS];
        }
        XSetForeground(dspl, gcx, fill_pixel);
        XFillPolygon(dspl, wnd, gcx, pts, 4, Complex, CoordModeOrigin);
        if (!wNoBorders)
        {
          fill_pixel = colourpixels[0];
          XSetForeground(dspl, gcx, fill_pixel);
          for (s = 0; s < 4; s++)
            XDrawLine(dspl, wnd, gcx,
                      (short) (polygons[p].screen_coords[s][0]),
                      (short) (polygons[p].screen_coords[s][1]),
                      (short) (polygons[p].
                               screen_coords[(s +
                                              1) %
                                             polygons[p].no_vertices][0]),
                      (short) (polygons[p].
                               screen_coords[(s +
                                              1) %
                                             polygons[p].no_vertices][1]));
        }
      }

      if (wDisplElemNo)
      {
        fill_pixel = colourpixels[3];
        XSetForeground(dspl, gcx, fill_pixel);
        nLength = sprintf(szBuffer, "%d", polygons[p].mesh_index + 1);
        XDrawString(dspl, wnd, gcx,
                    (short) ((polygons[p].screen_coords[0][0] +
                              polygons[p].screen_coords[1][0] +
                              polygons[p].screen_coords[2][0] +
                              polygons[p].screen_coords[3][0]) / 4.0),
                    (short) ((polygons[p].screen_coords[0][1] +
                              polygons[p].screen_coords[1][1] +
                              polygons[p].screen_coords[2][1] +
                              polygons[p].screen_coords[3][1]) / 4.0),
                    szBuffer, nLength);
      }
    }
  }
  if (wDisplPointNo)
  {
    fill_pixel = colourpixels[7];
    XSetForeground(dspl, gcx, fill_pixel);
    for (i = 0; i < mesh->NMeshPoints; i++)
    {
      for (n = 0; n < 3; n++)
        tri_cent[n] = mesh->MeshPointCoords[i][n];
      j_3Dworld_to_screen(scrtri_cent, tri_cent);
      nLength = sprintf(szBuffer, "%d", i + 1);
      XDrawString(dspl, wnd, gcx,
                  (int) scrtri_cent[0], (int) scrtri_cent[1],
                  szBuffer, nLength);
    }
  }
  if (wHN)
  {
    fill_pixel = colourpixels[4];
    XSetForeground(dspl, gcx, fill_pixel);
    XSetLineAttributes(dspl, gcx, 3, line_style, cap_style, join_style);
    XSetFillStyle(dspl, gcx, FillSolid);
    for (i = 0; i < mesh->NMeshPoints; i++)
    {
      for (n = 0; n < 3; n++)
        tri_cent[n] = mesh->MeshPointCoords[i][n];
      j_3Dworld_to_screen(scrtri_cent, tri_cent);
      XDrawArc(dspl, wnd, gcx, (int) (scrtri_cent[0] - 2.5),
               (int) (scrtri_cent[1] - 2.5), 5, 5, 0, 23040);
    }
    XSetLineAttributes(dspl, gcx, line_width, line_style, cap_style,
                       join_style);
  }

  set_normal_cursor(draw);
}

void j_load_font(Display * dspl, XFontStruct ** font_info)
{
  char *fontname = "fixed";

  if ((*font_info = XLoadQueryFont(dspl, fontname)) == NULL)
    e_GeneralError("can't load font!");
}

void j_init_colors(Display * display, Colormap cmp)
{
  XColor xcolour, spare;
  int i;
  
  /* colours from egr */
  int subdomcolors[MAXCOLS][3] = {
    {209, 1, 1},
    {255, 212, 60},
    {42, 33, 234},
    {34, 123, 20},
    {80, 80, 80},
    {188, 15, 171},
    {206, 75, 14},
    {13, 232, 217},
    {164, 232, 30},
    {188, 188, 188},
    {255, 112, 112},
    {0, 8, 86},
    {255, 106, 0},
    {93, 255, 0},
    {255, 0, 140},
    {255, 255, 255}
  };
  
#ifdef OLD_COLOURS
  wSDCols[0].red = 23130;       /* 90/255  */
  wSDCols[0].green = 23130;
  wSDCols[0].blue = 23130;
  wSDCols[1].red = 65535;       /* 255/255 */
  wSDCols[1].green = 21074;     /* 82/255  */
  wSDCols[1].blue = 2056;       /* 8/255   */
  wSDCols[2].red = 2056;        /* 8/255   */
  wSDCols[2].green = 21074;     /* 82/255  */
  wSDCols[2].blue = 65535;      /* 255/255 */
  wSDCols[3].red = 2056;        /* 8/255   */
  wSDCols[3].green = 65535;     /* 255/255 */
  wSDCols[3].blue = 65535;      /* 255/255 */
  wSDCols[4].red = 65535;       /* 255/255 */
  wSDCols[4].green = 44461;     /* 173/255 */
  wSDCols[4].blue = 2056;       /* 8/255   */
  wSDCols[5].red = 65535;       /* 255/255 */
  wSDCols[5].green = 2056;      /* 8/255   */
  wSDCols[5].blue = 44461;      /* 173/255 */
  wSDCols[6].red = 35980;       /* 140/255 */
  wSDCols[6].green = 31611;     /* 123/255 */
  wSDCols[6].blue = 21074;      /* 82/255  */
  wSDCols[7].red = 21074;       /* 82/255  */
  wSDCols[7].green = 65535;     /* 255/255 */
  wSDCols[7].blue = 2056;       /* 8/255   */
#endif
  
  for (i=0;i<MAXCOLS;i++)
  {
    wSDCols[i].red=subdomcolors[i][0]*257;
    wSDCols[i].green=subdomcolors[i][1]*257;
    wSDCols[i].blue=subdomcolors[i][2]*257;
  }
  
  wSDAltCols[0].red = 53713;
  wSDAltCols[0].green = 257;
  wSDAltCols[0].blue = 257;
  wSDAltCols[1].red = 65535;
  wSDAltCols[1].green = 54484;
  wSDAltCols[1].blue = 15420;
  wSDAltCols[2].red = 10794;
  wSDAltCols[2].green = 8481;
  wSDAltCols[2].blue = 60138;
  wSDAltCols[3].red = 8738;
  wSDAltCols[3].green = 31611;
  wSDAltCols[3].blue = 5140;
  wSDAltCols[4].red = 20560;
  wSDAltCols[4].green = 20560;
  wSDAltCols[4].blue = 20560;
  wSDAltCols[5].red = 65535;
  wSDAltCols[5].green = 65535;
  wSDAltCols[5].blue = 65535;

  if (XAllocNamedColor(display, cmp, "White", &xcolour, &spare) == 0)
    e_GeneralError("can't grab colour white");
  colourpixels[0] = xcolour.pixel;
  if (XAllocNamedColor(display, cmp, "Black", &xcolour, &spare) == 0)
    e_GeneralError("can't grab colour black");
  colourpixels[1] = xcolour.pixel;
  if (XAllocNamedColor(display, cmp, "Yellow", &xcolour, &spare) == 0)
    e_GeneralError("can't grab colour yellow");
  colourpixels[2] = xcolour.pixel;
  if (XAllocNamedColor(display, cmp, "Lightblue", &xcolour, &spare) == 0)
    e_GeneralError("can't grab colour lightblue");
  colourpixels[3] = xcolour.pixel;
  if (XAllocNamedColor(display, cmp, "Red", &xcolour, &spare) == 0)
    e_GeneralError("can't grab colour red");
  colourpixels[4] = xcolour.pixel;
  if (XAllocNamedColor(display, cmp, "Blue", &xcolour, &spare) == 0)
    e_GeneralError("can't grab colour blue");
  colourpixels[5] = xcolour.pixel;
  if (XAllocNamedColor(display, cmp, "Green", &xcolour, &spare) == 0)
    e_GeneralError("can't grab colour green");
  colourpixels[6] = xcolour.pixel;
  if (XAllocNamedColor(display, cmp, "Lightgreen", &xcolour, &spare) == 0)
    e_GeneralError("can't grab colour lightgreen");
  colourpixels[7] = xcolour.pixel;
  for (i = 0; i < MAXCOLS; i++)
  {
    if (XAllocColor(display, cmp, &(wSDCols[i])) == 0)
      e_GeneralError("can't grab SD colour %d", i + 1);
    colourpixels[8 + i] = wSDCols[i].pixel;
  }
  if (XAllocNamedColor(display, cmp, "gray10", &xcolour, &spare) == 0)
    e_GeneralError("can't grab colour gray10");
  colourpixels[8 + MAXCOLS] = xcolour.pixel;
  if (XAllocNamedColor(display, cmp, "gray20", &xcolour, &spare) == 0)
    e_GeneralError("can't grab colour gray20");
  colourpixels[8 + MAXCOLS + 1] = xcolour.pixel;
  if (XAllocNamedColor(display, cmp, "gray30", &xcolour, &spare) == 0)
    e_GeneralError("can't grab colour gray30");
  colourpixels[8 + MAXCOLS + 2] = xcolour.pixel;
  if (XAllocNamedColor(display, cmp, "gray40", &xcolour, &spare) == 0)
    e_GeneralError("can't grab colour gray40");
  colourpixels[8 + MAXCOLS + 3] = xcolour.pixel;
  if (XAllocNamedColor(display, cmp, "gray50", &xcolour, &spare) == 0)
    e_GeneralError("can't grab colour gray50");
  colourpixels[8 + MAXCOLS + 4] = xcolour.pixel;
  if (XAllocNamedColor(display, cmp, "gray60", &xcolour, &spare) == 0)
    e_GeneralError("can't grab colour gray60");
  colourpixels[8 + MAXCOLS + 5] = xcolour.pixel;
  if (XAllocNamedColor(display, cmp, "gray70", &xcolour, &spare) == 0)
    e_GeneralError("can't grab colour gray70");
  colourpixels[8 + MAXCOLS + 6] = xcolour.pixel;
  if (XAllocNamedColor(display, cmp, "gray80", &xcolour, &spare) == 0)
    e_GeneralError("can't grab colour gray80");
  colourpixels[8 + MAXCOLS + 7] = xcolour.pixel;
  if (XAllocNamedColor(display, cmp, "gray90", &xcolour, &spare) == 0)
    e_GeneralError("can't grab colour gray90");
  colourpixels[8 + MAXCOLS + 8] = xcolour.pixel;

  for (i = 0; i < MAXALTCOLS; i++)
  {
    if (XAllocColor(display, cmp, &(wSDAltCols[i])) == 0)
      e_GeneralError("can't grab Alternative SD colour %d", i + 1);
    colourpixels[17 + MAXCOLS + i] = wSDAltCols[i].pixel;
  }
}

void EPS_export(POLYGON * polygons, int Npoly)
{
  OPTIONS opt;
  int p;
  double zmin, zmax;

  if (!mdfopen)
    return;

  opt.plot_type = wEPS_plotplan ? 2 : 3;
  opt.hidden_line = wAdvHidden;

  /* opt.azimut=G_Azim;
     opt.elevation=G_Elev;
     opt.distance=G_dTC;
     opt.scale=G_scale; */

  opt.element_number = wDisplElemNo;
  opt.node_number = wDisplPointNo;
  opt.number_size = 0.02;

  if (wPlotSelect == PS_SUBDOM)
    opt.type = EPS_SUBDOMAIN;
  else if (wPlotSelect == PS_PLAINMESH)
    opt.type = EPS_PLAINMESH;

  opt.picture_size_x = wEPS_sizex;
  opt.picture_size_y = wEPS_sizey;
  opt.margin = wEPS_margin;

  opt.border = !wNoBorders;
  opt.bcolor = wEPS_bcolor;
  opt.color = wEPS_color;
  opt.tile = wTileElems;
  opt.percent = w_Element_Size_percent;

  opt.subdom_style = wEPS_subdom_style;
  opt.subdom_b_width = wEPS_subdom_b_width;
  opt.subdom_t_size = wEPS_subdom_t_size;

  set_watch_cursor(draw);

  j_do_projection(polygons, Npoly, &zmin, &zmax);
  order = e_alloc_i_vect(Npoly);
  if (wAdvHidden)
  {
    setup_hidden_line2(Npoly, polygons);
    setup_proj_center(G_Azim, G_Elev, G_dTC, G_Target);

    if (!sort_objects(order, Npoly, polygons))
    {
      e_GeneralWarning("advanced sorting failed");
      for (p = 0; p < Npoly; p++)
          order[p] = p;
    }
  }
  else
  {
    for (p = 0; p < Npoly; p++)
      order[p] = p;
    if (wHidden)
      j_sort_polygons(order, polygons, Npoly);
  }
  if (!e_Write_EPS_File(mesh, mdfname, &opt, Npoly, polygons, order, FALSE))
    e_GeneralWarning("eps export failed !!!");
  set_normal_cursor(draw);
}
