/* FILE chkmdf.c
 * validity checking for e_lib files */

#include "chkmdf.h"


int pslg_check(MESH *mesh)
{
  int t1,t2,s1,s2;
  int res;
  double mu1,mu2;
  int is=TRUE;

  eg_Init();

  if (e_GetElementType(mesh)!=E_ELEM_TYPE_TRIANG1)
  {
    printf("can't run pslg_check on non-triang1 mesh\n");
    return TRUE;
  }
  printf("pslg check may take a while: relax!\n");

  for (t1=0;t1<mesh->NElemsTriang1;t1++)
    for (t2=t1+1;t2<mesh->NElemsTriang1;t2++)
      for (s1=0;s1<3;s1++)
        for (s2=0;s2<3;s2++)
        {
          res=eg_LineLineIntersect2D(
            mesh->MeshPointCoords[mesh->NodesTriang1[t1][s1]-1],
            mesh->MeshPointCoords[mesh->NodesTriang1[t1][(s1+1)%3]-1],
            mesh->MeshPointCoords[mesh->NodesTriang1[t2][s2]-1],
            mesh->MeshPointCoords[mesh->NodesTriang1[t2][(s2+1)%3]-1],
            &mu1,&mu2);
          if (res==EG_INVALID)
          {
            printf("invalid triangle, either t%d or t%d\n",t1+1,t2+1);
            is=FALSE;
          }
          /*else if (res==EG_PARALLEL)
          {
            if (eg_PointLineDist2D(
              mesh->MeshPointCoords[mesh->NodesTriang1[t1][s1]-1],
              mesh->MeshPointCoords[mesh->NodesTriang1[t1][(s1+1)%3]-1],
              mesh->MeshPointCoords[mesh->NodesTriang1[t2][s2]-1],
              &mu1)<eg_GetTolerance(EG_ZERO_DIST_TOL))
            {
              printf("t%d,s%d equals t%d,s%d\n",t1+1,s1+1,t2+1,s2+1);
              is=FALSE;
            } 
          }*/
          else if (res==EG_INTERSECT)
          {
            if ((mu1>0.0001)&&(mu1<0.9999)&&(mu2>0.0001)&&(mu2<0.9999))
            {
              printf("t%d,s%d intersects t%d,s%d (mu1=%e,mu2=%e)\n",
                t1+1,s1+1,t2+1,s2+1,mu1,mu2);
              is=FALSE;
            }   
          }
        }
  return is;
}

void statistics(MESH *mesh)
{
  int et;

  printf("showing mesh statistics: \n");
  printf("  * title of the mdf file: ");
  if (mesh->KeywordSet[E_KW_NUM_TITLE])
    printf("\"%s\"\n", mesh->Title);
  else
    printf("NOT DEFINED\n");
  printf("  * number of mesh-points: ");
  if (mesh->KeywordSet[E_KW_NUM_NMESHPOINTS])
    printf("%d\n", mesh->NMeshPoints);
  else
    printf("NOT DEFINED\n");
  printf("  * number of finite element nodes: ");
  if (mesh->KeywordSet[E_KW_NUM_NNODES])
    printf("%d\n", mesh->NNodes);
  else
    printf("NOT DEFINED\n");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
    printf("  * number of elements LINK1: %d\n", mesh->NElemsLink1);
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
    printf("  * number of elements LINK2: %d\n", mesh->NElemsLink2);
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
    printf("  * number of elements LINK3: %d\n", mesh->NElemsLink3);
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
    printf("  * number of elements LINK4: %d\n", mesh->NElemsLink4);
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
    printf("  * number of elements LINK5: %d\n", mesh->NElemsLink5);
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
    printf("  * number of elements TRIANG1: %d\n", mesh->NElemsTriang1);
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
    printf("  * number of elements TRIANG2: %d\n", mesh->NElemsTriang2);
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
    printf("  * number of elements TRIANG3: %d\n", mesh->NElemsTriang3);
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
    printf("  * number of elements TRIANG4: %d\n", mesh->NElemsTriang4);
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
    printf("  * number of elements TRIANG5: %d\n", mesh->NElemsTriang5);
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
    printf("  * number of elements QUAD1: %d\n", mesh->NElemsQuad1);
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
    printf("  * number of elements QUAD3: %d\n", mesh->NElemsQuad3);
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
    printf("  * number of elements TETRAH1: %d\n", mesh->NElemsTetrah1);
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
    printf("  * number of elements BLOCK1: %d\n", mesh->NElemsBlock1);
  if (e_IsOfOneElementType(mesh))
  {
    printf("  * mesh contains only one element: ");
    et = e_GetElementType(mesh);
    if (et == E_ELEM_TYPE_LINK1)
      printf("LINK1\n");
    else if (et == E_ELEM_TYPE_LINK2)
      printf("LINK2\n");
    else if (et == E_ELEM_TYPE_LINK3)
      printf("LINK3\n");
    else if (et == E_ELEM_TYPE_LINK4)
      printf("LINK4\n");
    else if (et == E_ELEM_TYPE_LINK5)
      printf("LINK5\n");
    else if (et == E_ELEM_TYPE_TRIANG1)
      printf("TRIANG1\n");
    else if (et == E_ELEM_TYPE_TRIANG2)
      printf("TRIANG2\n");
    else if (et == E_ELEM_TYPE_TRIANG3)
      printf("TRIANG3\n");
    else if (et == E_ELEM_TYPE_TRIANG4)
      printf("TRIANG4\n");
    else if (et == E_ELEM_TYPE_TRIANG5)
      printf("TRIANG5\n");
    else if (et == E_ELEM_TYPE_QUAD1)
      printf("QUAD1\n");
    else if (et == E_ELEM_TYPE_QUAD3)
      printf("QUAD3\n");
    else if (et == E_ELEM_TYPE_TETRAH1)
      printf("TETRAH1\n");
    else if (et == E_ELEM_TYPE_BLOCK1)
      printf("BLOCK1\n");
  }
  if (e_LinkElementsOnly(mesh))
    printf("  * mesh contains only link elements\n");
  else
    printf("  * mesh contains non-link elements\n");
  if (e_PlanarElementsOnly(mesh))
    printf("  * mesh contains only planar elements\n");
  else
    printf("  * mesh contains non-planar elements\n");
  printf("  * total number of elements: %d\n", mesh->TotalNElems);
  printf("  * number of boundary condition nodes: ");
  if (mesh->KeywordSet[E_KW_NUM_NBC_NODES])
    printf("%d\n", mesh->NBCNodes);
  else
    printf("NOT DEFINED\n");
  printf("  * number of loaded nodes: ");
  if (mesh->KeywordSet[E_KW_NUM_NLOADED_NODES])
    printf("%d\n", mesh->NLoadedNodes);
  else
    printf("NOT DEFINED\n");

  return;
}

static int nodecompare(const void *e1, const void *e2)
{
  NDC nc1, nc2;

  nc1 = (*((NDC *) e1));
  nc2 = (*((NDC *) e2));

  if (nc1.x < nc2.x)
    return (-1);
  else if (nc1.x > nc2.x)
    return (1);
  else if (nc1.y < nc2.y)
    return (-1);
  else if (nc1.y > nc2.y)
    return (1);
  else if (nc1.z < nc2.z)
    return (-1);
  else if (nc1.z > nc2.z)
    return (1);
  else
    return (0);
}

static int nthlowest(int *el, int nnodes, int prev_low, int max)
{
  int s;
  int low = max + 1;

  for (s = 0; s < nnodes; s++)
    if ((el[s] > prev_low) && (el[s] < low))
      low = el[s];
  return low;
}

static int compne(const void *e1, const void *e2, int nnodes)
{
  int *el1;
  int *el2;
  int s;
  int low1, low2, max1, max2;

  el1 = (int *) e1;
  el2 = (int *) e2;

  max1 = -1;
  for (s = 0; s < nnodes; s++)
    if (el1[s] > max1)
      max1 = el1[s];
  max2 = -1;
  for (s = 0; s < nnodes; s++)
    if (el2[s] > max2)
      max2 = el2[s];

  low1 = -1;
  low2 = -1;
  for (s = 0; s < nnodes; s++)
  {
    low1 = nthlowest(el1, nnodes, low1, max1);
    low2 = nthlowest(el2, nnodes, low2, max2);
    if (low1 < low2)
      return (-1);
    else if (low1 > low2)
      return (1);
  }
  return 0;
}

static int comp2ne(const void *e1, const void *e2)
{
  return compne(e1, e2, 2);
}

static int comp3ne(const void *e1, const void *e2)
{
  return compne(e1, e2, 3);
}

static int comp4ne(const void *e1, const void *e2)
{
  return compne(e1, e2, 4);
}

static int comp8ne(const void *e1, const void *e2)
{
  return compne(e1, e2, 8);
}

int nc_equal(NDC e1, NDC e2)
{
  if (e1.x != e2.x)
    return 0;
  if (e1.y != e2.y)
    return 0;
  if (e1.z != e2.z)
    return 0;
  return 1;
}

int duplicate_node_chk(MESH *mesh)
{
  NDC *nc;
  int i, s, l;
  int dup_fnd;

  printf("checking for duplicate nodes: ");
  dup_fnd = FALSE;

  nc = (NDC *) e_calloc(mesh->NNodes, sizeof(NDC));
  for (i = 0; i < mesh->NNodes; i++)
  {
    nc[i].x = mesh->MeshPointCoords[i][E_XDIR];
    nc[i].y = mesh->MeshPointCoords[i][E_YDIR];
    nc[i].z = mesh->MeshPointCoords[i][E_ZDIR];
    nc[i].n = i + 1;
  }

  qsort((void *) nc, mesh->NNodes, sizeof(NDC), nodecompare);
  l = -1;
  for (i = 1; i < mesh->NNodes; i++)
  {
    if (nc_equal(nc[i], nc[E_MAX(i - 1, l)]))
    {
      dup_fnd = TRUE;
      if (l == -1)
      {
        l = i - 1;
        printf("\n  * following nodes are duplicates: %d %d", nc[l].n,
               nc[i].n);
      }
      else
        printf(" %d", nc[i].n);
    }
    else if (l != -1)
      l = -1;
  }
  if (!dup_fnd)
    printf("NO DUPLICATE NODES");
  printf("\n");

  return dup_fnd;
}

int duplicate_elem_chk(MESH *mesh)
{
  int dup_fnd = FALSE;

  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
    dup_fnd = dup_fnd
      || dupl_elems(mesh->NodesLink1, mesh->NElemsLink1, 2, "LINK1");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
    dup_fnd = dup_fnd
      || dupl_elems(mesh->NodesLink2, mesh->NElemsLink2, 2, "LINK2");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
    dup_fnd = dup_fnd
      || dupl_elems(mesh->NodesLink3, mesh->NElemsLink3, 2, "LINK3");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
    dup_fnd = dup_fnd
      || dupl_elems(mesh->NodesLink4, mesh->NElemsLink4, 2, "LINK4");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
    dup_fnd = dup_fnd
      || dupl_elems(mesh->NodesLink5, mesh->NElemsLink5, 2, "LINK5");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
    dup_fnd = dup_fnd
      || dupl_elems(mesh->NodesTriang1, mesh->NElemsTriang1, 3, "TRIANG1");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
    dup_fnd = dup_fnd
      || dupl_elems(mesh->NodesTriang2, mesh->NElemsTriang2, 3, "TRIANG2");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
    dup_fnd = dup_fnd
      || dupl_elems(mesh->NodesTriang3, mesh->NElemsTriang3, 3, "TRIANG3");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
    dup_fnd = dup_fnd
      || dupl_elems(mesh->NodesTriang4, mesh->NElemsTriang4, 3, "TRIANG4");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
    dup_fnd = dup_fnd
      || dupl_elems(mesh->NodesTriang5, mesh->NElemsTriang5, 3, "TRIANG5");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
    dup_fnd = dup_fnd
      || dupl_elems(mesh->NodesQuad1, mesh->NElemsQuad1, 4, "QUAD1");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
    dup_fnd = dup_fnd
      || dupl_elems(mesh->NodesQuad3, mesh->NElemsQuad3, 4, "QUAD3");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
    dup_fnd = dup_fnd
      || dupl_elems(mesh->NodesTetrah1, mesh->NElemsTetrah1, 4, "TETRAH1");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
    dup_fnd = dup_fnd
      || dupl_elems(mesh->NodesBlock1, mesh->NElemsBlock1, 8, "BLOCK1");
  return dup_fnd;
}

int dupl_elems(int **elems, int nelems, int nnodes, char *type)
{
  int *els;
  int e, s, l;
  int dup_fnd, eq;

  printf("checking for duplicate elements of type %s: ", type);
  dup_fnd = FALSE;

  els = e_alloc_i_vect(nelems * (nnodes + 1));
  for (e = 0; e < nelems; e++)
  {
    for (s = 0; s < nnodes; s++)
      els[e * (nnodes + 1) + s] = elems[e][s];
    els[e * (nnodes + 1) + nnodes] = e + 1;
  }
  if (nnodes == 2)
    qsort(els, nelems, 3 * sizeof(int), comp2ne);
  else if (nnodes == 3)
    qsort(els, nelems, 4 * sizeof(int), comp3ne);
  else if (nnodes == 4)
    qsort(els, nelems, 5 * sizeof(int), comp4ne);
  else if (nnodes == 8)
    qsort(els, nelems, 9 * sizeof(int), comp8ne);

  l = -1;
  for (e = 1; e < nelems; e++)
  {
    if (nnodes == 2)
      eq =
        (comp2ne((void *) &(els[e * 3]), (void *) &(els[E_MAX(e - 1, l) * 3]))
         == 0);
    else if (nnodes == 3)
      eq =
        (comp3ne((void *) &(els[e * 4]), (void *) &(els[E_MAX(e - 1, l) * 4]))
         == 0);
    else if (nnodes == 4)
      eq =
        (comp4ne((void *) &(els[e * 5]), (void *) &(els[E_MAX(e - 1, l) * 5]))
         == 0);
    else if (nnodes == 8)
      eq =
        (comp8ne((void *) &(els[e * 9]), (void *) &(els[E_MAX(e - 1, l) * 9]))
         == 0);
    if (eq)
    {
      dup_fnd = TRUE;
      if (l == -1)
      {
        l = e - 1;
        printf("\n  * following elements are duplicates: %d %d",
               els[l * (nnodes + 1) + nnodes],
               els[e * (nnodes + 1) + nnodes]);
      }
      else
        printf(" %d", els[e * (nnodes + 1) + nnodes]);
    }
    else if (l != -1)
      l = -1;
  }
  if (!dup_fnd)
    printf("NO DUPLICATES FOUND");
  printf("\n");

  e_free(els);
  return dup_fnd;
}

double dCCW(double *A, double *B, double *C)
{
  double d;

  d =
    A[0] * B[1] + C[0] * A[1] + B[0] * C[1] - C[0] * B[1] - B[0] * A[1] -
    A[0] * C[1];
  return d;
}

int elem_valid_chk(MESH *mesh)
{
  int iv = FALSE;
  int cw = FALSE;

  printf("checking for invalid elements: ");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
    iv = iv
      || chk_invalid_elem_def(mesh->NodesLink1, mesh->NElemsLink1, 2,
                              "LINK1");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
    iv = iv
      || chk_invalid_elem_def(mesh->NodesLink2, mesh->NElemsLink2, 2,
                              "LINK2");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
    iv = iv
      || chk_invalid_elem_def(mesh->NodesLink3, mesh->NElemsLink3, 2,
                              "LINK3");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
    iv = iv
      || chk_invalid_elem_def(mesh->NodesLink4, mesh->NElemsLink4, 2,
                              "LINK4");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
    iv = iv
      || chk_invalid_elem_def(mesh->NodesLink5, mesh->NElemsLink5, 2,
                              "LINK5");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
    iv = iv
      || chk_invalid_elem_def(mesh->NodesTriang1, mesh->NElemsTriang1, 3,
                              "TRIANG1");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
    iv = iv
      || chk_invalid_elem_def(mesh->NodesTriang2, mesh->NElemsTriang2, 3,
                              "TRIANG2");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
    iv = iv
      || chk_invalid_elem_def(mesh->NodesTriang3, mesh->NElemsTriang3, 3,
                              "TRIANG3");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
    iv = iv
      || chk_invalid_elem_def(mesh->NodesTriang4, mesh->NElemsTriang4, 3,
                              "TRIANG4");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
    iv = iv
      || chk_invalid_elem_def(mesh->NodesTriang5, mesh->NElemsTriang5, 3,
                              "TRIANG5");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
    iv = iv
      || chk_invalid_elem_def(mesh->NodesQuad1, mesh->NElemsQuad1, 4,
                              "QUAD1");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
    iv = iv
      || chk_invalid_elem_def(mesh->NodesQuad3, mesh->NElemsQuad3, 4,
                              "QUAD3");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
    iv = iv
      || chk_invalid_elem_def(mesh->NodesTetrah1, mesh->NElemsTetrah1, 4,
                              "TETRAH1");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
    iv = iv
      || chk_invalid_elem_def(mesh->NodesBlock1, mesh->NElemsBlock1, 8,
                              "BLOCK1");

  if (iv)
  {
    printf("invalid elements found: skipping orientation checks\n");
    return 1;
  }
  else
    printf("NO INVALID ELEMENTS\n");
  if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1]
      || mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2]
      || mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3]
      || mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4]
      || mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
  {
    printf("checking anti-clock-wise orientation of triangles: ");
    if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
      cw = cw
        || ccw_check(mesh->MeshPointCoords, mesh->NodesTriang1,
                     mesh->NElemsTriang1, "TRIANG1");
    if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
      cw = cw
        || ccw_check(mesh->MeshPointCoords, mesh->NodesTriang2,
                     mesh->NElemsTriang2, "TRIANG2");
    if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
      cw = cw
        || ccw_check(mesh->MeshPointCoords, mesh->NodesTriang3,
                     mesh->NElemsTriang3, "TRIANG3");
    if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
      cw = cw
        || ccw_check(mesh->MeshPointCoords, mesh->NodesTriang4,
                     mesh->NElemsTriang4, "TRIANG4");
    if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
      cw = cw
        || ccw_check(mesh->MeshPointCoords, mesh->NodesTriang5,
                     mesh->NElemsTriang5, "TRIANG5");
    if (!cw)
      printf("ALL ELEMENTS ACW\n");
  }
  return 0;
}

int chk_invalid_elem_def(int **elems, int nelems, int nnodes, char *type)
{
  int e, s, t;
  int valid;
  int first = TRUE;

  for (e = 0; e < nelems; e++)
  {
    valid = TRUE;
    for (s = 0; s < nnodes - 1; s++)
    {
      for (t = s + 1; t < nnodes; t++)
        if (elems[e][s] == elems[e][t])
        {
          valid = FALSE;
          break;
        }
      if (!valid)
        break;
    }
    if (!valid)
    {
      if (first)
      {
        printf("\n  * invalid elements found of type %s: ", type);
        first = FALSE;
      }
      printf("%d ", e + 1);
    }
  }
  if (!first)
  {
    printf("\n");
    return TRUE;
  }
  return FALSE;
}

int ccw_check(double **coords, int **elems, int nelems, char *type)
{
  int e;
  int first = TRUE;

  for (e = 0; e < nelems; e++)
  {
    if (dCCW
        (coords[elems[e][0] - 1], coords[elems[e][1] - 1],
         coords[elems[e][2] - 1]) <= 0)
    {
      if (first)
      {
        printf("\n  * clockwise defined elements of type %s: ", type);
        first = FALSE;
      }
      printf("%d ", e + 1);
    }
  }
  if (!first)
  {
    printf("\n");
    return TRUE;
  }
  return FALSE;
}

int main(int argc, char *argv[])
{
  MESH *mesh;
  int err_fnd = FALSE;

  printf("\nValidity checking of an e_lib mesh definition\n");

  if (argc < 2)
  {
    printf("\nUsage: chkmdf filename\n");
    printf("   filename : name of mesh definition file\n");
    printf("All files should be specified without extension!\n\n");
    e_exit(1);
  }

  printf("\nchecking file: %s.mdf\n", argv[1]);
  if (!e_GetMeshData(&mesh, argv[1]))
  {
    printf("%s.mdf does not exist or is corrupt\n", argv[1]);
    e_GeneralError("execution halted!\n");
  }

  statistics(mesh);
  err_fnd = err_fnd || duplicate_node_chk(mesh);

  if (elem_valid_chk(mesh))
  {
    e_FreeMeshData(mesh);
    printf("correct element validity first before continuing!\n");
    e_GeneralError("execution halted!\n");
  }

  err_fnd = err_fnd || duplicate_elem_chk(mesh);

  err_fnd = err_fnd || pslg_check(mesh);

  if (err_fnd)
    printf("\n***** Warnings or errors were issued! *****\n\n");
  else
    printf("\nMesh validity checked!\n\n");

  return 0;
}
