#include <stdio.h>
#include <stdlib.h>
#include "e_lib.h"

int main(int argc, char *argv[])
{
  MESH *m;
  int	*ch;
  int i,s,n;
  int nn;
  int *tbl;
  double **coords;
  
  if(argc < 3)
  {
    printf("\nRemove unconnected nodes from a mesh.\n");
    printf("\nUsage: cleanup input output\n");
    printf("   input  : name of input mesh definition file\n");
    printf("   output : name of resulting mesh definition file\n");
    printf("All files should be specified without extension!\n\n");
    e_exit(1);
  }  
    
  e_GetMeshData(&m,argv[1]);
  
  if (!e_IsOfOneElementType(m))
      e_GeneralError("cleanup only cleans TRIANG1 meshes");
  if (e_GetElementType(m)!=E_ELEM_TYPE_TRIANG1)
      e_GeneralError("cleanup only cleans TRIANG1 meshes");
  
  ch=e_alloc_i_vect(m->NNodes);
  tbl=e_alloc_i_vect(m->NNodes);
  
  for (i=0;i<m->NNodes;i++)
      ch[i]=FALSE;
        
  for (i=0;i<m->NElemsTriang1;i++)
      for (s=0;s<3;s++)
          ch[m->NodesTriang1[i][s]-1]=TRUE;
    
  /* make translation table */
  n=1;
  for (i=0;i<m->NNodes;i++)
  {
    if (!ch[i])
        tbl[i]=0;
    else
    {
      tbl[i]=n;
      n++;
    }
  }
            
  /* clean up coords matrix */
  nn=n-1;
  coords=e_alloc_cont_d_matr(nn,3);
  for (i=0;i<m->NNodes;i++)
      if (ch[i])
          for (s=0;s<3;s++)
              coords[tbl[i]-1][s]=m->MeshPointCoords[i][s];
  e_free_cont_d_matr(m->NNodes,m->MeshPointCoords);
  m->MeshPointCoords=coords;
  m->NNodes=m->NMeshPoints=nn;
  e_free(m->NodalDOF);
  m->NodalDOF=e_alloc_i_vect(nn);
  for (i=0;i<m->NNodes;i++)
      m->NodalDOF[i]=3;
  m->TotalNodalDOF=3*nn;

  /* translate nodestriang1 */
  for (i=0;i<m->NElemsTriang1;i++)
      for (s=0;s<3;s++)
          m->NodesTriang1[i][s]=tbl[m->NodesTriang1[i][s]-1];
  for (i=0;i<m->NLoadedNodes;i++)
  {
    if (tbl[m->LoadedNodes[i]-1]==0)
        e_GeneralError("a load was applied to an unconnected node");
    m->LoadedNodes[i]=tbl[m->LoadedNodes[i]-1];
  }
  for (i=0;i<m->NBCNodes;i++)
  {
    if (tbl[m->BCNodes[i]-1]==0)
        e_GeneralError("a boundary condition was applied to an unconnected node");
    m->BCNodes[i]=tbl[m->BCNodes[i]-1];
  }
  
  /* clean up gni's */
  e_free(m->GNIs);
  m->GNIs=e_alloc_i_vect(m->NNodes);
  for (i=0;i<m->NNodes;i++)
      m->GNIs[i]=i+1;
  
  e_PutMeshData(m, argv[2]);
  e_FreeMeshData(m);
  
  e_free(ch);
  e_free(tbl);

  return 0;
}
