/* FILE emp2mpr.c
 *
 * to convert element mesh params to nodal mesh params
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "e_lib.h"

int main(int argc, char *argv[])
{
  MESH    *mesh;
  int	   t, s;
  int	  *nn;
  
  if(argc == 3)
  {
    mesh=NULL;
    if (!e_GetMeshData(&mesh, argv[1]))
    {
      e_GeneralError("Mesh file contains faults; execution halted!");
    }
    e_GetMeshParams(mesh, argv[1]);
    if(!mesh->MeshParam.ElemMeshParamsSet)
    {
      e_FreeMeshParams(mesh);
      e_FreeMeshData(mesh);
      e_GeneralError("This program converts FROM element mpr TO nodal mpr!");
    }
    if (!e_IsOfOneElementType(mesh))
      e_GeneralError("emp2mpr only works with TRIANG1 meshes");
    if (e_GetElementType(mesh)!=E_ELEM_TYPE_TRIANG1)
      e_GeneralError("emp2mpr only works with TRIANG1 meshes");
  
    nn=e_alloc_i_vect(mesh->NNodes);
    mesh->MeshParam.NodalMeshParams=e_alloc_d_vect(mesh->NNodes);
    for (t=0;t<mesh->NNodes;t++)
    {
      nn[t]=0;
      mesh->MeshParam.NodalMeshParams[t]=0.0;
    }

    /* sum up all mesh parameters at the node */
    for (t=0;t<mesh->NElemsTriang1;t++)
    {
      for (s=0;s<3;s++)
      {
        mesh->MeshParam.NodalMeshParams[mesh->NodesTriang1[t][s]-1]+=
            mesh->MeshParam.ElemMeshParamsTriang1[t];
        nn[mesh->NodesTriang1[t][s]-1]++;
      }
    }

    /* calculate average mesh parameter */
    for (t=0;t<mesh->NNodes;t++)
    {
      mesh->MeshParam.NodalMeshParams[t]/=nn[t];
    }
    
    e_free(mesh->MeshParam.ElemMeshParamsTriang1);
    mesh->MeshParam.ElemMeshParamsSet=FALSE;
    mesh->MeshParam.NodalMeshParamsSet=TRUE;
    e_PutMeshParams(mesh, argv[2]);
    e_FreeMeshParams(mesh);
    e_FreeMeshData(mesh);
  }
  else
  {
    /* show the usage of the program */
    printf("\nConvert element mesh parameters to nodal mesh prameters\n");
    printf("by averaging all element mesh parameters connected to a node.\n");
    printf("\nUsage: emp2mpr input output\n");
    printf("   input  : name of input mesh file and mesh parameter file\n");
    printf("   output : name of output mesh parameter file\n");
    printf("All files should be specified without extension!\n\n");
    e_exit(1);
  }
  return 0;
}

/* end of file */
