#include <math.h>
#include <e_lib.h>

#include "fem.h"

int main(int argc, char *argv[])
{
    MESH *mesh;
    int mprtype;
    
    if(argc != 4)
    {
	printf("\nadapt v1.1 SECT Research Group, Heriot-Watt University Edinburgh\n");
	printf("usage: adapt input d mprtype\n");
	printf("note all filenames are referred to without extensions\n");
	printf("   input    : name of the mesh definition and stress file\n");
	printf("   d        : permissible error value\n");
	printf("   mprtype  : 1 for element mesh params\n");
	printf("            : 2 for nodal mesh params\n\n");
	exit(0);
    }

    if (!strcmp(argv[3], "1"))
	mprtype=1;
    else if (!strcmp(argv[3], "2"))
	mprtype=2;
    else
	e_GeneralError("unrecognized mesh parameter type\n");
	
    e_GetMeshData(&mesh, argv[1]);
    e_GetMaterials(mesh, argv[1]);
    e_GetStresses(mesh, argv[1]);
  
    J_Calculate_Errors_and_MeshParams(mesh, atof(argv[2]), mprtype);

    e_PutFEErrors(mesh, argv[1]);
    e_PutMeshParams(mesh, argv[1]);
    e_FreeMeshData(mesh);
}

void J_Calculate_Errors_and_MeshParams(MESH *mesh, double perm_error, int mprtype)
{
    int e, a, j;
    int	ElemType;
    int NstressPNode;
    double *Ele_Non_Average_Strain, *Ele_Non_Average_Stress;
    double *Nodal_Average_Strain, *Nodal_Average_Stress;
    double *Ele_Error_Stress_Norm, *Ele_displacement_Energy_Norm;
    double Global_Error_Stress_Norm;
    double Global_displacement_Energy_Norm;
    double *Elem_Size;
    int *Degree;
    int **AdjElems;
    double exi, em;
    
    ElemType=e_GetElementType(mesh);
   
    NstressPNode=3;
    
    Ele_Non_Average_Strain = e_alloc_d_vect(NstressPNode*mesh->TotalNElems);
    Ele_Non_Average_Stress = e_alloc_d_vect(NstressPNode*mesh->TotalNElems);

    Nodal_Average_Strain = e_alloc_d_vect(NstressPNode*mesh->NMeshPoints);
    Nodal_Average_Stress = e_alloc_d_vect(NstressPNode*mesh->NMeshPoints);

    if(ElemType==E_ELEM_TYPE_TRIANG1)
	for (e=0;e<mesh->TotalNElems;e++)
	{
	    a=e*NstressPNode;
	    Ele_Non_Average_Stress[a]=mesh->Stresses.StressesTriang1[e][0][E_SIGMA_X];
	    Ele_Non_Average_Stress[a+1]=mesh->Stresses.StressesTriang1[e][0][E_SIGMA_Y];
	    Ele_Non_Average_Stress[a+2]=mesh->Stresses.StressesTriang1[e][0][E_TAU_XY];
	    for (a=0;a<NstressPNode;a++)
		Ele_Non_Average_Strain[e*NstressPNode+a]=0.0;
	}
    else if(ElemType==E_ELEM_TYPE_QUAD1)
	for (e=0;e<mesh->TotalNElems;e++)
	{
	    a=e*NstressPNode;
	    Ele_Non_Average_Stress[a]=mesh->Stresses.StressesQuad1[e][0][E_SIGMA_X];
	    Ele_Non_Average_Stress[a+1]=mesh->Stresses.StressesQuad1[e][0][E_SIGMA_Y];
	    Ele_Non_Average_Stress[a+2]=mesh->Stresses.StressesQuad1[e][0][E_TAU_XY];
	    for (a=0;a<NstressPNode;a++)
		Ele_Non_Average_Strain[e*NstressPNode+a]=0.0;
	}
    
    if(ElemType==E_ELEM_TYPE_TRIANG1)
	B_get_nodal_stress_Triang1(mesh, Ele_Non_Average_Strain, 
				   Ele_Non_Average_Stress,
				   Nodal_Average_Strain,    
				   Nodal_Average_Stress);
    else if(ElemType==E_ELEM_TYPE_QUAD1)
	e_GeneralError("Quad1 temporarily not supported\n");
        /* B_get_nodal_stress_Quad1(mesh, Force, Ele_Non_Average_Strain,
                                 Ele_Non_Average_Stress,
                                 Nodal_Average_Strain,
                                 Nodal_Average_Stress); */
				 
    Ele_Error_Stress_Norm = e_alloc_d_vect(mesh->TotalNElems);
    Ele_displacement_Energy_Norm = e_alloc_d_vect(mesh->TotalNElems);
    
    if(ElemType==E_ELEM_TYPE_TRIANG1)
	B_Calculate_Norm_e_Triang1(mesh,Ele_Non_Average_Strain, 
				   Ele_Non_Average_Stress,          
				   Nodal_Average_Strain,    
				   Nodal_Average_Stress,
				   Ele_Error_Stress_Norm, 
				   Ele_displacement_Energy_Norm);
    else if(ElemType==E_ELEM_TYPE_QUAD1)
    {
/*	if(Simple_Error)
	    B_Calculate_Norm_e_Quad1(mesh,Force, Nodal_Average_Stress,
				     Ele_Error_Stress_Norm, Ele_displacement_Energy_Norm);
	else  */
	    B_Calculate_Norm_e_QUAD1_average(mesh,Ele_Non_Average_Strain, 
					     Ele_Non_Average_Stress,          
					     Nodal_Average_Strain,    
					     Nodal_Average_Stress,
					     Ele_Error_Stress_Norm, 
					     Ele_displacement_Energy_Norm);
    }
    mesh->SubStructSet[E_SS_NUM_FEERROR]=TRUE;
    
    if(ElemType==E_ELEM_TYPE_TRIANG1)
    {
	mesh->FEError.FEErrorsTriang1=e_alloc_d_vect(mesh->TotalNElems);
	for (e=0;e<mesh->TotalNElems;e++)
	    mesh->FEError.FEErrorsTriang1[e]=
		Ele_Error_Stress_Norm[e]/(Ele_Error_Stress_Norm[e]+Ele_displacement_Energy_Norm[e]);
    }
    else
    {
	mesh->FEError.FEErrorsQuad1=e_alloc_d_vect(mesh->TotalNElems);
	for (e=0;e<mesh->TotalNElems;e++)
	    mesh->FEError.FEErrorsQuad1[e]=
		Ele_Error_Stress_Norm[e]/(Ele_Error_Stress_Norm[e]+Ele_displacement_Energy_Norm[e]);
    }
	    
    Global_Error_Stress_Norm=0.0;
    Global_displacement_Energy_Norm=0.0;
    for(e=0;e<mesh->TotalNElems;e++)
    {
	Global_Error_Stress_Norm+=Ele_Error_Stress_Norm[e];
	Global_displacement_Energy_Norm+=Ele_displacement_Energy_Norm[e];
    }
    if (sqrt(Global_Error_Stress_Norm /(Global_Error_Stress_Norm+Global_displacement_Energy_Norm))<0.0)
	e_GeneralError("Error: Nita is negative in error estimator function ");

    em=perm_error*sqrt((Global_Error_Stress_Norm+Global_displacement_Energy_Norm)/mesh->TotalNElems);
    
    Elem_Size = e_alloc_d_vect(mesh->TotalNElems);
    if(ElemType ==E_ELEM_TYPE_TRIANG1)
	B_Cal_Quad_Elem_Size_Triang1(mesh, Elem_Size);
    else if(ElemType ==E_ELEM_TYPE_QUAD1)
	B_Cal_Quad_Elem_Size_Quad1(mesh, Elem_Size);
    
    for(e=0;e<mesh->TotalNElems;e++)
    {
	exi = 1.1 *sqrt(Ele_Error_Stress_Norm[e])/em;
	if(exi <1.0) exi = 1.0;
	Elem_Size[e] /= exi;
    }

    if (mprtype==1) /* elem mesh params */
    {
	mesh->SubStructSet[E_SS_NUM_MESHPARAM]=TRUE;
	mesh->MeshParam.ElemMeshParamsSet=TRUE;
	mesh->MeshParam.NodalMeshParamsSet=FALSE;
	if(ElemType==E_ELEM_TYPE_TRIANG1)
	{
	    mesh->MeshParam.ElemMeshParamsTriang1=e_alloc_d_vect(mesh->TotalNElems);
	    for(e=0;e<mesh->TotalNElems;e++)
		mesh->MeshParam.ElemMeshParamsTriang1[e]=Elem_Size[e];
	}
	else if(ElemType==E_ELEM_TYPE_QUAD1)
	{
	    mesh->MeshParam.ElemMeshParamsQuad1=e_alloc_d_vect(mesh->TotalNElems);
	    for(e=0;e<mesh->TotalNElems;e++)
		mesh->MeshParam.ElemMeshParamsQuad1[e]=Elem_Size[e];
	}
    }
    else
    {
	mesh->SubStructSet[E_SS_NUM_MESHPARAM]=TRUE;
	mesh->MeshParam.ElemMeshParamsSet=FALSE;
	mesh->MeshParam.NodalMeshParamsSet=TRUE;
	mesh->MeshParam.NodalMeshParams=e_alloc_d_vect(mesh->NMeshPoints);
	
	SetAdjPointsElems(mesh,&Degree,&AdjElems);
	
	for(e=0;e<mesh->NMeshPoints;e++)
	{
	    for(j=0; j<Degree[e]; j++)
               mesh->MeshParam.NodalMeshParams[e]+=Elem_Size[AdjElems[e][j]-1];
	    mesh->MeshParam.NodalMeshParams[e]/=Degree[e];
	}
	
	FreeAdjMeshPointsElems(mesh,Degree,AdjElems);
    }
    
    e_free(Elem_Size);
    e_free(Ele_Non_Average_Strain);
    e_free(Ele_Non_Average_Stress);
    e_free(Ele_Error_Stress_Norm);  
    e_free(Ele_displacement_Energy_Norm);
    e_free(Nodal_Average_Strain);
    e_free(Nodal_Average_Stress); 
}
