/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                             Calculate_displacement()                       */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <math.h>
#include <e_lib.h>
#include "fem.h"


void B_clockwise_check(MESH *mesh,int ElemType)
{
int i;
int **nodes;
int temp;

  nodes = e_GetNodes(mesh);


  for(i=0;i<mesh->TotalNElems;i++)
    {
        if(B_Q_right(mesh->MeshPointCoords[nodes[i][0]-1][0],
                     mesh->MeshPointCoords[nodes[i][0]-1][1],
                     mesh->MeshPointCoords[nodes[i][1]-1][0],
    	             mesh->MeshPointCoords[nodes[i][1]-1][1],
	             mesh->MeshPointCoords[nodes[i][2]-1][0],
	             mesh->MeshPointCoords[nodes[i][2]-1][1]))
           { 

             if(ElemType ==E_ELEM_TYPE_TRIANG1)
              {
                temp =nodes[i][1];
                nodes[i][1] =nodes[i][2];
                nodes[i][2] = temp;
              }
             else if(ElemType ==E_ELEM_TYPE_QUAD1)
              {
                temp =nodes[i][0];
                nodes[i][0] =nodes[i][2];
                nodes[i][2] = temp;
              }
           }
     }
}

/*********************************************************/
int B_Q_right(double x1,double y1,double x2,double y2,double x3,
                     double y3)
{
double D;
D=x3*(y1-y2)+y3*(x2-x1)+(x1*y2-y1*x2);
if(D>1.0e-10)
   return 1;
else
   return 0;
}





/*********************************************************/
void B_Assemble_matrix_TRIANG1(MESH *mesh,double *KV, int *ind, double *Force)
{
int ii,jj;

double YMod, Density, PoissonsRatio, Thickness;
int elem_no;
int i,j;
double quot,Det;

double **Dee, *Guass_Samp, *WT, **Coord, **Jac, **Jac1, **Der, **Deriv, **Bee,
       **DBee, **BTDB, **KM, **BT, *Fun;

int idee, ibee, idbee, ih;
int idof, ibtdb, ibt, ikm;
int ijac, ijac1, it, ider, ideriv;
int icoord, nod, isamp;



 
   nod = icoord = E_NNPE_TRIANG1;
   idee = ibee = idbee = ih = 3;
   idof = ibtdb = ibt = ikm = 2*E_NNPE_TRIANG1;
   ijac = ijac1  = it = ider = ideriv = E_NDIMS-1;
   isamp = 2;

   Dee = e_alloc_cont_d_matr(idee, idee);
   Guass_Samp = e_alloc_d_vect(isamp);
   WT =  e_alloc_d_vect(1);
   Coord = e_alloc_cont_d_matr(icoord, it);
   Jac = e_alloc_cont_d_matr(ijac, ijac);
   Jac1 = e_alloc_cont_d_matr(ijac1, ijac1);
   Der = e_alloc_cont_d_matr(ider, nod);
   Deriv = e_alloc_cont_d_matr(ideriv, nod);
   Bee = e_alloc_cont_d_matr(ibee,   idof);
   DBee = e_alloc_cont_d_matr(idbee, idof);
   BTDB = e_alloc_cont_d_matr(ibtdb, ibtdb);
   KM = e_alloc_cont_d_matr(ikm, ikm);
   BT = e_alloc_cont_d_matr(ibt, ibt);
   Fun = e_alloc_d_vect(nod); 



   B_numint(Guass_Samp,WT); /***********************/

  
  for(elem_no=0; elem_no<mesh->NElemsTriang1; elem_no++)
   {
    B_GetElemMaterial(mesh->MatsTriang1, elem_no, &YMod, 
                      &PoissonsRatio, &Thickness);
    B_Fmdsig(Dee,YMod,PoissonsRatio);
    B_Elem_cord(mesh, elem_no, Coord);
    B_null(KM,ikm,ikm); 

        B_fmtri3(Der,Fun,Guass_Samp);  /***********************/
        B_Matmul(Der,Coord,Jac,it,nod,it);
	B_twoby2(Jac, Jac1,&Det);
        B_Matmul(Jac1,Der,Deriv,it,it,nod);
        B_null(Bee,ih,idof);
	B_form_Bee(Bee, Deriv, nod);
        B_Matmul(Dee,Bee,DBee,ih,ih,idof);
	B_Matran(BT,Bee,ih,idof);
        B_Matmul(BT,DBee,BTDB,idof,ih,idof);
	quot=0.5*Det*WT[0]; /***********************/

        B_Msmult(BTDB,idof,idof,quot);
        B_Matadd(KM,BTDB,idof,idof);
       
    B_Fsparv(mesh, elem_no, KV, KM, ind);
   }

  e_free(Dee);
  e_free(Guass_Samp);
  e_free(WT);
  e_free(Coord);
  e_free(Jac);
  e_free(Jac1);
  e_free(Der);
  e_free(Deriv);
  e_free(Bee);
  e_free(DBee);
  e_free(BTDB);
  e_free(KM);
  e_free(BT);
  e_free(Fun);

 
}



/*********************************************************/

void B_numint(double *s,double *wt)
{

 s[0]=1.0/3.0;
 s[1]=1.0/3.0;
 wt[0]=1.0;  /* quadrature weights */
}


/*********************************************************/

void B_fmtri3(double **der,double *fun,double *samp)
{  

  fun[0]=samp[0];
  fun[1]=samp[1];
  fun[2]=1.0-fun[0]-fun[1];

  der[0][0]=1.0;  der[0][1]=0.0;  der[0][2]=-1.0;
  der[1][0]=0.0;  der[1][1]=1.0;  der[1][2]=-1.0;
}


/*********************************************************/
/* this function forms the elastic plane ***stress****
stress/strain matrix */
void B_Fmdsig(double **D,double e, double v)
{
double c;

c=e/(1-v*v);

D[0][0] = c;
D[1][1] = c;
D[2][2] = 0.5*c*(1-v);
D[0][1] = v*c;
D[1][0] = v*c;
D[0][2] = D[2][0]=D[2][1]=D[1][2]=0.0;
}

/*********************************************************/
/* This function forms the elastic plane ****strain*****
 stress/strain matrix */
void fmdeps(double **dee, double e, double v)
{
double v1,c;

v1 = 1-v;
c = e/((1.0 + v)*(1.0-2.0*v));
dee[0][0] = dee[2][2] = v1*c;
dee[2][2] = 0.5*c*(1-2.0*v);
dee[0][1] = v*c;
dee[1][0] = v*c;
dee[0][2] = dee[2][0] = dee[1][2] = dee[2][1] =0.0;
}


/*********************************************************/

/* Forms the inverse JAC1 of a 2*2 matrix JAC and its
determinant DET */
void B_twoby2(double **Jac, double **Jac1, double *Det)
{
int k,l;

*Det = Jac[0][0]*Jac[1][1]-Jac[0][1]*Jac[1][0];

Jac1[0][0] =  Jac[1][1];
Jac1[0][1] = -Jac[0][1];
Jac1[1][0] = -Jac[1][0];
Jac1[1][1] =  Jac[0][0];

for(k=0;k<2;k++)
 for(l=0;l<2;l++)
   Jac1[k][l]=Jac1[k][l]/(*Det);


}


/*********************************************************/
/* Forms the strain-displacement matrix BEE for an
element with NOD nodes in plane strian. Element
shape function derivatives in DERIV */
void B_form_Bee(double **Bee, double **Deriv,int Nod)
{
int k,l,m;
double x,y;


for(m = 0; m < Nod; m++)
  {
   l=2*m;
   k=2*m+1;

   x=Deriv[0][m];
   Bee[0][l]=x;
   Bee[2][k]=x;

   y=Deriv[1][m];
   Bee[1][k]=y;
   Bee[2][l]=y;
  }

}


/*********************************************************/

void B_Assemble_matrix_QUAD1(MESH *mesh,double *KV, int *ind, double *Force)
{
int ii,jj;
double YMod, Density, PoissonsRatio, Thickness;
int elem_no;
int i,j;
double quot,Det;

double **Dee, **Guass_Samp, **Coord, **Jac, **Jac1, **Der, **Deriv, **Bee,
       **DBee, **BTDB, **KM, **BT, *Fun;

int idee, ibee, idbee, ih;
int idof, ibtdb, ibt, ikm;
int ijac, ijac1, it, ider, ideriv;
int icoord, nod, isamp;



 
   nod = icoord = E_NNPE_QUAD1;
   idee = ibee = idbee = ih = 3;
   idof = ibtdb = ibt = ikm = 2*E_NNPE_QUAD1;
   ijac = ijac1  = it = ider = ideriv = E_NDIMS-1;
   isamp = 2;

   Dee = e_alloc_cont_d_matr(idee, idee);
   Guass_Samp = e_alloc_cont_d_matr(isamp, isamp);
   Coord = e_alloc_cont_d_matr(icoord, it);
   Jac = e_alloc_cont_d_matr(ijac, ijac);
   Jac1 = e_alloc_cont_d_matr(ijac1, ijac1);
   Der = e_alloc_cont_d_matr(ider, nod);
   Deriv = e_alloc_cont_d_matr(ideriv, nod);
   Bee = e_alloc_cont_d_matr(ibee,   idof);
   DBee = e_alloc_cont_d_matr(idbee, idof);
   BTDB = e_alloc_cont_d_matr(ibtdb, ibtdb);
   KM = e_alloc_cont_d_matr(ikm, ikm);
   BT = e_alloc_cont_d_matr(ibt, ibt);
   Fun = e_alloc_d_vect(nod); 

   B_gauss_Quad1(Guass_Samp);

  
  for(elem_no=0; elem_no<mesh->NElemsQuad1; elem_no++)
   {
    B_GetElemMaterial(mesh->MatsQuad1, elem_no, &YMod, 
                      &PoissonsRatio, &Thickness);
    B_Fmdsig(Dee,YMod,PoissonsRatio);
    B_Elem_cord(mesh, elem_no, Coord);
    B_null(KM,ikm,ikm); 

    for(i=0;i<2;i++)
     {
      for(j=0;j<2;j++)
       {
        B_formln(Der,Fun,Guass_Samp,i,j);
        B_Matmul(Der,Coord,Jac,it,nod,it);
	B_twoby2(Jac, Jac1,&Det);
        B_Matmul(Jac1,Der,Deriv,it,it,nod);
        B_null(Bee,ih,idof);
	B_form_Bee(Bee, Deriv, nod);
        B_Matmul(Dee,Bee,DBee,ih,ih,idof);
	B_Matran(BT,Bee,ih,idof);
        B_Matmul(BT,DBee,BTDB,idof,ih,idof);
	quot=Det*Guass_Samp[i][1]*Guass_Samp[j][1];
        B_Msmult(BTDB,idof,idof,quot);
        B_Matadd(KM,BTDB,idof,idof);
       }
     }
    B_Fsparv(mesh, elem_no, KV, KM, ind);
   }

  e_free(Dee);
  e_free(Guass_Samp);
  e_free(Coord);
  e_free(Jac);
  e_free(Jac1);
  e_free(Der);
  e_free(Deriv);
  e_free(Bee);
  e_free(DBee);
  e_free(BTDB);
  e_free(KM);
  e_free(BT);
  e_free(Fun);
}



/*********************************************************/


/*********************************************************/
/* Returns weights and adscissae SAMP of sampling points in
Gauss-Legendre quadrature for NGP integrating points 4 */
void B_gauss_Quad1(double **Guass_samp)
{
Guass_samp[0][0] = 1.0/sqrt(3.0);
Guass_samp[1][0] = -Guass_samp[0][0];
Guass_samp[1][1] = Guass_samp[0][1]=1.0;
}



/*********************************************************/
/* Forms four-node quadrilateral shape functions FUN and
derivatives DER in local coordinates. Gaussian integration
date in SAMP located by counters i,j */
void B_formln(double **Der, double *Fun, double **Guass_samp, int i,
 int j )
{
double eta,xi,etam,etap,xim,xip;


/* to specify the center local coordinates according to i ,j */
if(i==2 && j==2)
 {
  eta=0;
  xi=0;
 }
else
 {
  eta = Guass_samp[i][0];
  xi  = Guass_samp[j][0];
  }

etam=0.25*(1-eta);	etap=0.25*(1+eta);
xim=0.25*(1-xi);	   xip=0.25*(1+xi);

/* All element nodes should be numbered clockwise */

Fun[0]=4*xim*etam;	Fun[1]=4*xim*etap;
Fun[2]=4*xip*etap;	Fun[3]=4*xip*etam;

Der[0][0]=-etam;	   Der[0][1]=-etap;
Der[0][2]=etap;		   Der[0][3]=etam;

Der[1][0]=-xim;		   Der[1][1]=xim;
Der[1][2]=xip;		   Der[1][3]=-xip;
}

/*********************************************************/

void B_Assemble_matrix_TETRAH1(MESH *mesh,double *KV, int *ind, double *Force)
{
int ii,jj;

double YMod, Density, PoissonsRatio, Thickness;
int elem_no;
int i,j;
double quot,Det;

double **Dee, **Guass_Samp, *WT, **Coord, **Jac, **Jac1, **Der, **Deriv, **Bee,
       **DBee, **BTDB, **KM, **BT, *Fun;

int idee, ibee, idbee, ih;
int idof, ibtdb, ibt, ikm;
int ijac, ijac1, it, ider, ideriv;
int icoord, nod, isamp;



 
   nod = icoord = E_NNPE_TETRAH1;
   idee = ibee = idbee = ih = 6;
   idof = ibtdb = ibt = ikm = 3*E_NNPE_TETRAH1; /* x,y,z */
   ijac = ijac1  = it = ider = ideriv = E_NDIMS;
   isamp = 3;

   Dee = e_alloc_cont_d_matr(idee, idee);
   Guass_Samp = e_alloc_cont_d_matr(isamp,isamp);
   WT = e_alloc_d_vect(1);
   Coord = e_alloc_cont_d_matr(icoord, it);
   Jac = e_alloc_cont_d_matr(ijac, ijac);
   Jac1 = e_alloc_cont_d_matr(ijac1, ijac1);
   Der = e_alloc_cont_d_matr(ider, nod);
   Deriv = e_alloc_cont_d_matr(ideriv, nod);
   Bee = e_alloc_cont_d_matr(ibee,   idof);
   DBee = e_alloc_cont_d_matr(idbee, idof);
   BTDB = e_alloc_cont_d_matr(ibtdb, ibtdb);
   KM = e_alloc_cont_d_matr(ikm, ikm);
   BT = e_alloc_cont_d_matr(idof, ibee);
   Fun = e_alloc_d_vect(nod); 


   B_numint_TETRAH1(Guass_Samp,WT); /***********************/

  
  for(elem_no=0; elem_no<mesh->NElemsTriang1; elem_no++)
   {
    B_GetElemMaterial(mesh->MatsTriang1, elem_no, &YMod, 
                      &PoissonsRatio, &Thickness);
    B_formD_TETRAH1(Dee,YMod,PoissonsRatio);
    B_Elem_cord(mesh, elem_no, Coord);
    B_null(KM,ikm,ikm); 

      /*  B_fmtet4(Der,Fun,Guass_Samp); /***********************/
        B_Matmul(Der,Coord,Jac,it,nod,it);
	B_treex3_TETRAH1(Jac,ijac, Jac1,&Det);
        B_Matmul(Jac1,Der,Deriv,it,it,nod);
        B_null(Bee,ih,idof);
	B_formb3_TETRAH1(Bee, Deriv, nod);
        B_Matmul(Dee,Bee,DBee,ih,ih,idof);
	B_Matran(BT,Bee,ih,idof);
        B_Matmul(BT,DBee,BTDB,idof,ih,idof);
	quot = Det*WT[0]/6.0; /***********************/

        B_Msmult(BTDB,idof,idof,quot);
        B_Matadd(KM,BTDB,idof,idof);

   
    
    B_Fsparv(mesh, elem_no, KV, KM, ind);
   }

  e_free(Dee);
  e_free(Guass_Samp);
  e_free(WT);
  e_free(Coord);
  e_free(Jac);
  e_free(Jac1);
  e_free(Der);
  e_free(Deriv);
  e_free(Bee);
  e_free(DBee);
  e_free(BTDB);
  e_free(KM);
  e_free(BT);
  e_free(Fun);

 

}


/*********************************************************/

void B_numint_TETRAH1(double **s,double *wt)
{

 s[0][0]=0.25;
 s[0][1]=0.25;
 s[0][3]=0.25; 
 wt[0]=1.0;
}


/*********************************************************/
void  B_fmtet4(double **Der,double *Fun,double **Guass_Samp)
{

  Fun[0]=Guass_Samp[0][0];
  Fun[1]=Guass_Samp[0][1];
  Fun[2]=Guass_Samp[0][2];
  Fun[3]=1.0-Fun[0]-Fun[1]-Fun[2];

  Der[0][0] = 1.0;
  Der[1][1] = 1.0;
  Der[2][2] = 1.0;
  Der[0][3] = -1.0;
  Der[1][3] = -1.0;
  Der[2][3] = -1.0;
}



/*********************************************************/
void  B_formD_TETRAH1(double **Dee, double e, double v)
{
  int i,j;
  double v1 = v/1.0-v;
  double vv = (1.0-2.0*v)/(1.0-v)*0.5;

  Dee[0][0] = Dee[1][1] = Dee[2][2] = 1.0;
  Dee[0][1] = Dee[1][0] = Dee[0][2] = Dee[2][0] = Dee[1][2] = Dee[2][1] =v1;
  Dee[3][3] = Dee[4][4] = Dee[5][5] = vv;

  for(i=0;i<6;i++)
    for(j=0;j<6;j++)
        Dee[i][j] = Dee[i][j]*e/(2.0*(1.0+v)*vv);
}



/*********************************************************/
void B_treex3_TETRAH1(double **Jac,int ijac, double **Jack1,double *Det)
{

 *Det      = Jac[0][0]*(Jac[1][1]*Jac[2][2]-Jac[2][1]*Jac[1][2])
           - Jac[0][1]*(Jac[1][0]*Jac[2][2]-Jac[2][0]*Jac[1][2])
           + Jac[0][3]*(Jac[1][0]*Jac[2][1]-Jac[2][0]*Jac[1][1]);

B_Matinversion(Jac,ijac,Jack1);
}



/*********************************************************/
void B_formb3_TETRAH1(double **Bee, double **Deriv,int nod)
{
int i,n,k,l;

for(i=0;i<nod;i++)
 { 
   n = 3*i;
   k = n-1;
   l = k-1;

   Bee[0][l] = Bee[3][k] = Bee[5][n] = Deriv[0][i];
   Bee[1][k] = Bee[3][l] = Bee[4][n] = Deriv[1][i];
   Bee[2][n] = Bee[4][k] = Bee[5][l] = Deriv[2][i];
 }
}




/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                             Calculatediagonalelements()                    */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */


/*********************************************************/

void B_form_dia(MESH *mesh,int *ind,int N_displacements)
{
int i,j,k,l,m;
int a,b,c;
int **NodesIndex;
int Temp_E_NNPE_ELEMs;
int Position_last_row;

  Temp_E_NNPE_ELEMs = e_GetNNodesPerElement(mesh);
  NodesIndex = e_GetNodes(mesh);

/* first calculate the rows with even number */
for(i=0;i<mesh->TotalNElems;i++)
{
 for(j=0;j<Temp_E_NNPE_ELEMs;j++)
  {
     a= NodesIndex[i][j];

     for(k=0;k<Temp_E_NNPE_ELEMs; k++)
      {
	     b= NodesIndex[i][k];
	     if((a-b)>ind[N_displacements*a-1])
	       {
	           ind[N_displacements*a-1]=abs(a-b);
	       }
      }
   }
}


/* And then the odd number of row */
for(l=0;l<mesh->NMeshPoints;l++)
 {
  Position_last_row = N_displacements*(l+1)-1;
  ind[Position_last_row] = 
              (ind[Position_last_row]+1)*N_displacements; 
    /* add itself in at the diagonal point */

  for(i=1;i< N_displacements; i++)
           ind[Position_last_row-i]=ind[Position_last_row-i+1]-1;
 }


 for(m=1;m<N_displacements*mesh->NMeshPoints;m++)
                 ind[m]=ind[m]+ind[m-1];
}






/*********************************************************/
/* form the force vector  */
void B_form_loads(MESH *mesh, double *force, int N_displacement_Per_node)
{
int i,  j;
int address;

for(i=0;i<mesh->NLoadedNodes;i++)
  {
	 address=N_displacement_Per_node*(mesh->LoadedNodes[i]-1);
         for(j=0;j<N_displacement_Per_node;j++)
	                 force[address+j] =  mesh->Loads[i][j];
   }

}



/*********************************************************/
void B_GetElemMaterial(MAT **Material,int elem_no, double *e, double *v, double *t)
{
/*   *d = Material[elem_no].Density; */
   *e =Material[elem_no]->YMod;
   *v = Material[elem_no]->PoissonsRatio;
   *t =Material[elem_no]->Thickness;
}







/*********************************************************/
void B_Elem_cord(MESH *mesh, int NElem, double **Coord)
{
int i;
int **NodesIndex;
int Temp_E_NNPE_ELEMs;


  Temp_E_NNPE_ELEMs = e_GetNNodesPerElement(mesh);
  NodesIndex = e_GetNodes(mesh);

for(i=0;i<Temp_E_NNPE_ELEMs;i++)
  {
   Coord[i][0] = mesh->MeshPointCoords[NodesIndex[NElem][i]-1][0];
   Coord[i][1] = mesh->MeshPointCoords[NodesIndex[NElem][i]-1][1];
  }
}








/*********************************************************/
/* Forms sparse global stiffness matrix(vector)BK from constituent
KM element matrices. Steering vector G of length IDOF and number
of terms in each row held in KDIAG */

void B_Fsparv(MESH *mesh, int NElem, double *Kv, double **Km, int *ind)
{
int i,j,l;
int address[10];
int row,col,Diagonal_position,Vector_position_inCurrentRow;
int **NodesIndex;
int Temp_E_NNPE_ELEMs;


  Temp_E_NNPE_ELEMs = e_GetNNodesPerElement(mesh);
  NodesIndex = e_GetNodes(mesh);


for(l=0;l<Temp_E_NNPE_ELEMs;l++)
{
 address[2*l]=2*(NodesIndex[NElem][l]-1);
 address[2*l+1]=address[2*l]+1;
}

for(i = 0;i < 2*Temp_E_NNPE_ELEMs; i++)
 {
   row=address[i];
   for(j=0; j< 2*Temp_E_NNPE_ELEMs; j++)
    {
      col=address[j];
      if(col>row) continue;
      Diagonal_position = ind[address[i]]-1;
      Vector_position_inCurrentRow = row - col;
      Kv[Diagonal_position-Vector_position_inCurrentRow]+=Km[i][j];
    }
  }
 }




/*********************************************************/

void debu_B_Fsparv(MESH *mesh, int NElem, float Kv[12][12], double **Km, int *ind, int NNodesPElem)
{
int i,j,l;
int address[10];
int row,col,point,position;



for(l=0;l<NNodesPElem;l++)
{
 address[2*l]=2*(mesh->NodesQuad1[NElem][l]-1);
 address[2*l+1]=address[2*l]+1;
}

for(i = 0;i < 2*NNodesPElem; i++)
 {
   row=address[i];
   for(j=0; j< 2*NNodesPElem; j++)
    {
      col=address[j];
      Kv[row][col]+=Km[i][j];
    }
  }
 }


/*********************************************************/
/* add the boundary condition to the stiffness matrix */
void B_BoundaryConditions(MESH *mesh, double *kv, double *Force,
int *ind, int Ndisps)
{
int i,j;
int address;

for(i = 0; i < mesh->NBCNodes; i++)
 {
   address = Ndisps*(mesh->BCNodes[i]-1);

   for(j=0;j<Ndisps;j++)
    {

     if(mesh->BCTypes[i][j] == E_BC_NUM_FIXED)
        {
           kv[ind[address+j]-1] = pow(10.0,30.);
      
           if(mesh->BCDispls[i][j]!=0.0)
                Force[address + j] = Force[address + j]+
                        kv[ind[address+j]-1]*mesh->BCDispls[i][j];
         }
 
      else if(mesh->BCTypes[i][j] == E_BC_NUM_SPRING) 
               kv[ind[address+j]-1] += mesh->BCSpringConsts[i][j]; /*???[i][j]????*/

     }
  }

}


/*********************************************************/

/* The Number of row of Matrix A becomes the number of row of Matrix c and the
same as number of column of Matrix B The Nmultiple is the number of column of
Matrix A and the number of row of Matrix B */
 void B_Matmul(double **A, double **B, double **C, int Nrow,int Nmultiple, int Ncolumn)
{
int i,j,k;
double x;

for(i=0;i<Nrow;i++)
 {
  for(j=0;j<Ncolumn;j++)
   {
        x=0.0;
        for(k=0; k<Nmultiple; k++)
          {
           x=x+A[i][k]*B[k][j];
          }
         C[i][j]=x;
    }
  }

}

/*********************************************************/
/* A=B^t where B is M*N. working restrictions IB>m,Ia>n  */
void B_Matran(double **A, double **B, int Nrow, int Ncolumn)
{
int i,j;

for(i=0;i<Nrow;i++)
 {
  for(j=0;j<Ncolumn;j++)
   {
    A[j][i]=B[i][j];
   }
 }
}



/*********************************************************/
void B_Msmult(double **A, int Nrow, int Ncolumn, double scale)
{
int i,j;

for(i=0;i<Nrow;i++)
 {
  for(j=0;j<Ncolumn;j++)
   {
    A[i][j]=scale*A[i][j];
   }
 }

}

/*********************************************************/
void B_Matadd(double **A, double **B, int Nrow, int Ncolumn)
{
int i,j;

 for(i=0;i<Nrow;i++)
 {
  for(j=0;j< Ncolumn;j++)
   {
     A[i][j]=A[i][j]+B[i][j];
   }
 }
}



/*********************************************************/
/* function to read the elements topology and the essential data
for the structural analysis */

void B_null(double **Km, int Nrow, int Ncolumn)
{
int i,j;

 for(i=0;i< Nrow;i++)
    for(j=0;j< Ncolumn;j++)
         Km[i][j]=0;
}



/*********************************************************/
void B_deformationMesh(MESH *mesh,double *Force,double scale, 
                         int N_displacement_Per_node)
{
int i,j;

for(i=0; i< mesh->NMeshPoints; i++)
   for(j=0; j<N_displacement_Per_node; j++)
        mesh->MeshPointCoords[i][j] = mesh->MeshPointCoords[i][j] 
                             + scale*Force[N_displacement_Per_node*i+j];
}




/* ////////////////////////////////////////////////////////////////////////// */
/*  */


int ConGradSolve(int n, double *A, int *index, double *b, double *x, double tolerance)
{
   double alpha,beta;
   double bMagnitude;
   double *p,*r,*w;
   double rho,RhoMinus;
   int i = 0;

   
   if (!AllocateMemory(n,&p,&r,&w))
      return 0;

   memcpy(x,p,n*sizeof(double));        /* set x to zero */
   memcpy(r,b,n*sizeof(double));
   memcpy(p,r,n*sizeof(double));
   rho = DotProduct(n,r,r);
   bMagnitude = sqrt(rho);

   if (!(rho > 0.0))
      return 1;

   PreMultiply(n,A,index,p,w);
   alpha = rho/DotProduct(n,p,w);
   AddVectors(n,alpha,x,p,x);
   AddVectors(n,-alpha,r,w,r);
   RhoMinus = rho;
   rho = DotProduct(n,r,r);

  do {
      beta = rho/RhoMinus;
      AddVectors(n,beta,r,p,p);
      PreMultiply(n,A,index,p,w);
      alpha = rho/DotProduct(n,p,w);
      AddVectors(n,alpha,x,p,x);
      AddVectors(n,-alpha,r,w,r);
      RhoMinus = rho;
      rho = DotProduct(n,r,r);

     printf("%d error = %f\n", ++i,sqrt(rho)/bMagnitude);

   } while (sqrt(rho) > tolerance*bMagnitude );
/*   } while (i<100); */

 
      e_free(p);
      e_free(r);
      e_free(w);

  return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* AllocateMemory: To allocate memory */

/* The return value is 1 if the memory is allocated successfully, else 0. */


int AllocateMemory(int n, double **p, double **r, double **w)
{
   *p = *r = *w = NULL;

   if (   (*p = e_alloc_d_vect(n)) == NULL
       || (*r = e_alloc_d_vect(n)) == NULL
       || (*w = e_alloc_d_vect(n)) == NULL)
   {
      e_free(*p);
      e_free(*r);
      e_free(*w);
      return 0;
   }
    else
       return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* DotProduct: To return the dot product of two vectors (a.b). */


double DotProduct(int n, double *a, double *b)
{
   int i;

   double c;


   c = 0.0;

   for (i = 0; i < n; i++)
      c += a[i]*b[i];

   return c;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* PreMultiply: To determine the product a vector premultiplied by a matrix
   (Ab = c).                                                                  */


void PreMultiply(int n, double *A, int *index,  double *b, double *c)
{
   int i,j;


   for (i = 0; i < n; i++)
   {
      c[i] = 0.0;

      for (j = 0; j < n; j++)
       {
        if(j<=i)  c[i] += Matrix_retrive(n,A,index,i,j)*b[j];
        else      c[i] += Matrix_retrive(n,A,index,j,i)*b[j];
        }
   }
}




double Matrix_retrive(int n, double *A, int *ind, int row, int column)
{
 double A_value;
 int Diagonal_position, Up_Diagonal_position, Length_Current_row,
     Vector_position_inCurrentRow;

 Diagonal_position = ind[row]-1;
 if(row == 0) Up_Diagonal_position = 0;
 else  Up_Diagonal_position = ind[row-1];
 
 Length_Current_row = ind[row]-Up_Diagonal_position;

 Vector_position_inCurrentRow = row - column;

 if(Vector_position_inCurrentRow <0) 
 e_GeneralError("Error: Column is biger than the row in function   Matrix_retrive");
 else if(column < (row +1 -  Length_Current_row)) A_value = 0.0;
 else A_value = A[Diagonal_position - Vector_position_inCurrentRow];

 return A_value;

}


double Matrix_address(int n, double *A, int *ind, int row, int column)
{
 double A_value;
 int Diagonal_position, Up_Diagonal_position, Length_Current_row,
     Vector_position_inCurrentRow;

 Diagonal_position = ind[row]-1;
 if(row == 0) Up_Diagonal_position = 0;
 else  Up_Diagonal_position = ind[row-1];
 
 Length_Current_row = ind[row]-Up_Diagonal_position;

 Vector_position_inCurrentRow = row - column;

 if(Vector_position_inCurrentRow <0) 
 e_GeneralError("Error: Column is biger than the row in function   Matrix_retrive");
 else if(column < (row +1 -  Length_Current_row)) A_value = 0.0;
 else A_value = A[Diagonal_position - Vector_position_inCurrentRow];

 return A_value;

}

/* ////////////////////////////////////////////////////////////////////////// */
/* AddVectors: To add a multiple of the vector "b" to the vector "a" to give
   the vector "c".                                                            */


void AddVectors(int n, double alpha, double *a, double *b, double *c)
{
   int i;


   for (i = 0; i < n; i++)
      c[i] = a[i] + alpha*b[i];
}



void B_bandv(int Order, double *O_kv, double *O_force, int *O_ind)
{
int i,i0,mi,j,j0,mj,mij,ij,jm1,k,ik,kk;
int jk,l,im1,ir,n,jj,ii;

int *ind;
double *force, *kv;

  ind = e_alloc_i_vect(Order+1); 
  force = e_alloc_d_vect(Order+1);
  kv = e_alloc_d_vect(O_ind[Order- 1]+1);


for(i=0;i<Order;i++)
  {
   ind[i+1]   = O_ind[i];
   force[i+1] = O_force[i];
  }

for(i=0;i<O_ind[Order- 1];i++)
  kv[i+1] = O_kv[i];

for(i=1;i<=Order;i++)
 {
     i0=ind[i]-i;
     if(i!=1) 
      { 
          mi=ind[i-1]-i0+1;
          for(j=mi;j<=i;j++)
            {
   	      j0=ind[j]-j;
   	      mj=1;
   	      if(j>=1) mj=ind[j-1]-j0+1;
              mij=mi;
   	      if(mj>mi) mij=mj;
   	      ij=i0+j;
   	      jm1=j-1;

     	      for(k=mij;k<=jm1;k++)
    	        {
     	          ik=i0+k;
     	          kk=ind[k];
     	          jk=j0+k;
     	          kv[ij]=kv[ij]-kv[ik]*kv[kk]*kv[jk];
    	        }
              if(j==i) break;
              jj=ind[j];
              kv[ij]=kv[ij]/kv[jj];

             force[i]=force[i]-kv[ij]*kv[jj]*force[j];

          } 
     }

     ii=i0+i;
     if(kv[ii]==0) e_GeneralError("Matrix singular");
     force[i]=force[i]/kv[ii];
 }



 for(l=2;l<=Order;l++)
   {
     i=Order-l+2;
     i0=ind[i]-i;
     mi=ind[i-1]-i0+1;
     im1=ind[i-1]-i0+1;
     im1=i-1;
    
     for(j=mi;j<=im1;j++)
       {
          ij=i0+j;
          force[j]=force[j]-kv[ij]*force[i];
       }
    }



for(i=0;i<Order;i++)
  O_force[i] = force[i+1];

e_free(kv);
e_free(force);
e_free(ind);

}



