/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                             Calculate_stress_error()                       */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <math.h>
#include <e_lib.h>
#include "fem.h"

/*############################################################*/

void J_Calculate_Stresses(MESH *mesh, double *Force)
{
    int e, j;
    int ElemType;
    int NstressPNode, NDisplPNode, a;
    double *Ele_Non_Average_Strain, *Ele_Non_Average_Stress;
    
    ElemType=e_GetElementType(mesh);
    if(ElemType ==E_ELEM_TYPE_TETRAH1)
    {
	NstressPNode=6;
        NDisplPNode=3;
    }
    else 
    {
        NstressPNode=3;
        NDisplPNode=2;
    }
    Ele_Non_Average_Strain = e_alloc_d_vect(NstressPNode*mesh->TotalNElems);
    Ele_Non_Average_Stress = e_alloc_d_vect(NstressPNode*mesh->TotalNElems);

    B_get_Eleme_Stress_AtCenter(mesh, Force, Ele_Non_Average_Strain, 
                                             Ele_Non_Average_Stress,
			        NstressPNode,NDisplPNode);
    
    if(ElemType ==E_ELEM_TYPE_TRIANG1)
    {
	mesh->NStressPointsTriang1=1;
	mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_TRIANG1]=TRUE;
	e_SetupStresses(mesh);
	for (e=0;e<mesh->TotalNElems;e++)
	{
	    a=NstressPNode*e;
	    mesh->Stresses.StressesTriang1[e][0][E_SIGMA_X]=Ele_Non_Average_Stress[a];
	    mesh->Stresses.StressesTriang1[e][0][E_SIGMA_Y]=Ele_Non_Average_Stress[a+1];
	    mesh->Stresses.StressesTriang1[e][0][E_TAU_XY]=Ele_Non_Average_Stress[a+2];
	}
    }
    else if(ElemType ==E_ELEM_TYPE_QUAD1)
    {
	mesh->NStressPointsQuad1=1;
	mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_QUAD1]=TRUE;
	e_SetupStresses(mesh);
	for (e=0;e<mesh->TotalNElems;e++)
	{
	    a=NstressPNode*e;
	    mesh->Stresses.StressesQuad1[e][0][E_SIGMA_X]=Ele_Non_Average_Stress[a];
	    mesh->Stresses.StressesQuad1[e][0][E_SIGMA_Y]=Ele_Non_Average_Stress[a+1];
	    mesh->Stresses.StressesQuad1[e][0][E_TAU_XY]=Ele_Non_Average_Stress[a+2];
	}
    }
    
    e_free(Ele_Non_Average_Strain);
    e_free(Ele_Non_Average_Stress);
}
/*############################################################*/
/* function to calculate element stress at the center of  iso coordinate
element center */
void B_Calculate_Stress_Error(MESH *mesh, double *Force, char *New_mpr, float Permissible_Error)
{
int ip,i,ii,j,jj,k;
double *Ele_Non_Average_Strain, *Ele_Non_Average_Stress;
double *Nodal_Average_Strain, *Nodal_Average_Stress;
double *Ele_Error_Stress_Norm, *Ele_displacement_Energy_Norm;
double Global_Error_Stress_Norm, Global_displacement_Energy_Norm;
double *Elem_Error, *Nodal_Size_Param;
double *NodalMeshParams;
int NstressPNode,NDisplPNode;
int ElemType;

  ElemType = e_GetElementType(mesh);


 if(ElemType ==E_ELEM_TYPE_TETRAH1)
    {
         NstressPNode = 6;
         NDisplPNode  = 3;
    }
  else 
    {
         NstressPNode = 3;
         NDisplPNode   = 2;
    }


  Ele_Non_Average_Strain = e_alloc_d_vect(NstressPNode*mesh->TotalNElems);
  Ele_Non_Average_Stress = e_alloc_d_vect(NstressPNode*mesh->TotalNElems);

  B_get_Eleme_Stress_AtCenter(mesh, Force, Ele_Non_Average_Strain, 
                                           Ele_Non_Average_Stress,
			      NstressPNode,NDisplPNode);

  Nodal_Average_Strain = e_alloc_d_vect(NstressPNode*mesh->NMeshPoints);
  Nodal_Average_Stress = e_alloc_d_vect(NstressPNode*mesh->NMeshPoints);

 if(ElemType ==E_ELEM_TYPE_TRIANG1)
             B_get_nodal_stress_Triang1(mesh, Ele_Non_Average_Strain, 
                                Ele_Non_Average_Stress,
                                Nodal_Average_Strain,    
                                Nodal_Average_Stress);
  else if(ElemType ==E_ELEM_TYPE_QUAD1)
             B_get_nodal_stress_Quad1(mesh, Force, Ele_Non_Average_Strain,
                                  Ele_Non_Average_Stress,
                                  Nodal_Average_Strain,
                                  Nodal_Average_Stress);


  Ele_Error_Stress_Norm       = e_alloc_d_vect(mesh->TotalNElems);
  Ele_displacement_Energy_Norm = e_alloc_d_vect(mesh->TotalNElems);
 
  

 if(ElemType ==E_ELEM_TYPE_TRIANG1)
      B_Calculate_Norm_e_Triang1(mesh,Ele_Non_Average_Strain, 
			     Ele_Non_Average_Stress,          
			     Nodal_Average_Strain,    
			     Nodal_Average_Stress,
                             Ele_Error_Stress_Norm, 
                            Ele_displacement_Energy_Norm);
 else if(ElemType ==E_ELEM_TYPE_QUAD1)
    {
      if(Simple_Error)
        B_Calculate_Norm_e_Quad1(mesh,Force, Nodal_Average_Stress,
                  Ele_Error_Stress_Norm, Ele_displacement_Energy_Norm);
      else 
         B_Calculate_Norm_e_QUAD1_average(mesh,Ele_Non_Average_Strain, 
			        Ele_Non_Average_Stress,          
			        Nodal_Average_Strain,    
			        Nodal_Average_Stress,
                                Ele_Error_Stress_Norm, 
                                Ele_displacement_Energy_Norm);
      }


  Elem_Error       = e_alloc_d_vect(mesh->TotalNElems);
  Nodal_Size_Param = e_alloc_d_vect(mesh->NMeshPoints);

  B_Error_ElemSize_NodalSize(mesh, Ele_Error_Stress_Norm,
 				      Ele_displacement_Energy_Norm, 
			 	      &Global_Error_Stress_Norm, 
 			              &Global_displacement_Energy_Norm, 
 				      Permissible_Error, 
 			              Elem_Error, 
 			              Nodal_Size_Param);


  NodalMeshParams = e_alloc_d_vect(mesh->NMeshPoints);

  for(i=0;i<mesh->NMeshPoints;i++)
    NodalMeshParams[i] = Nodal_Size_Param[i];

  mesh->MeshParam.NodalMeshParams = NodalMeshParams;

  mesh->MeshParam.NodalMeshParamsSet = TRUE;
  mesh->MeshParam.ElemMeshParamsSet = FALSE;
  mesh->SubStructSet[E_SS_NUM_MESHPARAM] = TRUE;


 e_PutMeshParams(mesh, New_mpr);

 e_free(Ele_Non_Average_Strain); e_free(Ele_Non_Average_Stress);
 e_free(Nodal_Average_Strain); e_free(Ele_Non_Average_Stress);
 e_free(Ele_Error_Stress_Norm);  e_free(Ele_displacement_Energy_Norm);
 e_free(Elem_Error);             e_free(Nodal_Size_Param);

}



/*############################################################*/
void  B_get_Eleme_Stress_AtCenter(MESH *mesh, double *Force,
                                 double *Ele_Non_Average_Strain, 
                                 double  *Ele_Non_Average_Stress,
				 int NstressPNode, int NDisplPNode)
{
int elem_no;

int ih, idee, ibee, nod, ijac, ijac1, ider, ideriv, 
    it, idof, isamp, icoord;
double **Dee, *T_Guass_Samp,**Q_Guass_Samp,**TETRAH1_Guass_Samp, *WT, **Coord, **Jac, **Jac1, **Der, **Deriv, **Bee,*Fun;
double YMod, PoissonsRatio, Thickness, Det;
double T_Bee[3][8];
int i,j,k;
int ElemType,Temp_E_NNPE_ELEMs, Temp_NElems;
int **nodes;
double *temp_node_strain, *temp_node_stress;
MAT **Material;

   temp_node_strain = e_alloc_d_vect(NstressPNode);
   temp_node_stress = e_alloc_d_vect(NstressPNode);

  ElemType = e_GetElementType(mesh);

  if(ElemType ==E_ELEM_TYPE_TRIANG1)
      Material = mesh->MatsTriang1;
  else if(ElemType ==E_ELEM_TYPE_QUAD1)
      Material = mesh->MatsQuad1;
  else if(ElemType ==E_ELEM_TYPE_TETRAH1)
      Material = mesh->MatsTetrah1;

  Temp_E_NNPE_ELEMs = e_GetNNodesPerElement(mesh);
  nodes = e_GetNodes(mesh);


   nod = icoord = Temp_E_NNPE_ELEMs;
   if(ElemType == E_ELEM_TYPE_TETRAH1)
     {
   	 idee = ibee  = ih = 6;
   	 idof = 3*Temp_E_NNPE_ELEMs;
   	 ijac = ijac1  = it = ider = ideriv = E_NDIMS;
   	 isamp = 3;
      }
   else
      {
	   idee = ibee  = ih = 3;
 	   idof = 2*Temp_E_NNPE_ELEMs;
  	   ijac = ijac1  = it = ider = ideriv = E_NDIMS-1;
   	   isamp = 2;
       }
 
  Dee = e_alloc_cont_d_matr(idee, idee);
  T_Guass_Samp = e_alloc_d_vect(isamp);
  Q_Guass_Samp = e_alloc_cont_d_matr(isamp,isamp);
  TETRAH1_Guass_Samp = e_alloc_cont_d_matr(isamp,isamp);
  WT =  e_alloc_d_vect(1);
  Coord = e_alloc_cont_d_matr(icoord, it);
  Jac = e_alloc_cont_d_matr(ijac, ijac);
  Jac1 = e_alloc_cont_d_matr(ijac1, ijac1);
  Der = e_alloc_cont_d_matr(ider, nod);
  Deriv = e_alloc_cont_d_matr(ideriv, nod);
  Bee = e_alloc_cont_d_matr(ibee,   idof);
  Fun = e_alloc_d_vect(nod); 

   if(ElemType ==E_ELEM_TYPE_TRIANG1)
                             B_numint(T_Guass_Samp,WT); 
   else if(ElemType ==E_ELEM_TYPE_QUAD1)
                             B_gauss_Quad1(Q_Guass_Samp);
/*   else if(ElemType ==E_ELEM_TYPE_TETRAH1)
              B_numint_TETRAH1(TETRAH1_Guass_Samp,WT);   ***********************/


  for(elem_no=0; elem_no<mesh->TotalNElems; elem_no++)
   {
    B_GetElemMaterial(Material, elem_no, &YMod, 
                      &PoissonsRatio, &Thickness);
    if(ElemType ==E_ELEM_TYPE_TETRAH1)
              B_formD_TETRAH1(Dee,YMod,PoissonsRatio);
    else      B_Fmdsig(Dee,YMod,PoissonsRatio);
    B_Elem_cord(mesh, elem_no, Coord);

    if(ElemType ==E_ELEM_TYPE_TRIANG1)
                             B_fmtri3(Der,Fun,T_Guass_Samp); 
    else if(ElemType ==E_ELEM_TYPE_QUAD1)
                             B_formln(Der,Fun,Q_Guass_Samp,2,2);
    else if(ElemType ==E_ELEM_TYPE_TETRAH1)
                             B_fmtet4(Der,Fun,TETRAH1_Guass_Samp);

    B_Matmul(Der,Coord,Jac,it,nod,it);
    if(ElemType ==E_ELEM_TYPE_TETRAH1)
	                   B_treex3_TETRAH1(Jac,ijac, Jac1,&Det);
    else                   B_twoby2(Jac, Jac1,&Det);    

    B_Matmul(Jac1,Der,Deriv,it,it,nod);
    B_null(Bee,ih,idof);

    if(ElemType ==E_ELEM_TYPE_TETRAH1)
	      B_formb3_TETRAH1(Bee, Deriv, nod);
    else      B_form_Bee(Bee, Deriv, nod);


    B_form_strain(nodes,Temp_E_NNPE_ELEMs, Bee, Force,
                   temp_node_strain,elem_no,NstressPNode,NDisplPNode);

    B_form_stress(Dee,temp_node_strain, 
                       temp_node_stress, NstressPNode);

    for(k=0;k<NstressPNode;k++)
      {
        Ele_Non_Average_Strain[NstressPNode*elem_no+k]= 
                                  temp_node_strain[k];
        Ele_Non_Average_Stress[NstressPNode*elem_no+k]= 
                                  temp_node_stress[k];
      }

  }


  e_free(Dee); 
  e_free(T_Guass_Samp);
  e_free(Q_Guass_Samp);
  e_free(TETRAH1_Guass_Samp);
  e_free(WT);
  e_free(Coord);
  e_free(Jac);
  e_free(Jac1);
  e_free(Der);
  e_free(Deriv);
  e_free(Bee);
  e_free(Fun);
  e_free(temp_node_strain);
  e_free(temp_node_stress);

}

/*############################################################*/
/* form the strain using the displacement and bee */
void B_form_strain(int **nodes, int Temp_E_NNPE_ELEMs, double **Bee, 
      double *Force, double *temp_node_strain, int NElem, int NstressPNode,
      int NDisplPNode)
{
 int i,j,l;
 int address[12];
 double x;


 for(l=0;l< Temp_E_NNPE_ELEMs;l++)
   {
 	address[NDisplPNode*l]=NDisplPNode*(nodes[NElem][l]-1);
        for(i=1;i< NDisplPNode; i++)
	       address[NDisplPNode*l+i]=address[NDisplPNode*l+i-1]+1;
   }


 for(i = 0; i<NstressPNode; i++)
  {
   x=0.0;
   for(j=0; j<NDisplPNode*Temp_E_NNPE_ELEMs; j++)
    {
     x += Bee[i][j]*Force[address[j]];
    }
   temp_node_strain[i] = x;  /* strain starts from 1 2 3 for x y xy */
  }

}

/*############################################################*/
/* get the stress  */
void B_form_stress(double **Dee, double *temp_node_strain, 
double *temp_node_stress, int NstressPNode)
{
int i,j,jj;
double x;


for(i=0;i<NstressPNode;i++)
 {
  x=0;
  for(j=0;j<NstressPNode;j++)
   {
     x=x+Dee[i][j]*temp_node_strain[j];
   }
  temp_node_stress[i] = x;
 }
}

/*############################################################*/
/* get the nodal stress by their coordinates in the mesh rather than by average 
the non_average_elem_stress. It is fine for constant stress triangle but not 
fine for linear quadrilateral or tethro due to the non_average_elem_stress are 
calculated at the element center and the stresses at the different coordinate are
different in the mesh. It is more resonable to calculate the stress according to
the coordinates with the diferent Bee matrix rather than average the element stresses
at this stage */

void B_get_nodal_stress_Quad1(MESH *mesh, double *Force,
         double  *Ele_Non_Average_Strain,  double *Ele_Non_Average_Stress, 
         double  *Nodal_Average_Strain,    double *Nodal_Average_Stress)

{
int ip,i,ii,j,jj,k;
int elem_no, Node_Number;
 int *Degree;
 int **AdjElems;

double x_strain,y_strain,xy_strain;
double x_stress, y_stress, xy_stress;
int NstressPNode,NDisplPNode;

int ih, idee, ibee, nod, ijac, ijac1, ider, ideriv, it, idof, isamp, icoord;
double **Dee, **Guass_Samp, **Coord, **Jac, **Jac1, **Der, **Deriv, **Bee,*Fun;
double YMod, PoissonsRatio, Thickness, Det;
int Temp_E_NNPE_ELEMs,**nodes;
double *temp_node_strain, *temp_node_stress;


  Temp_E_NNPE_ELEMs = e_GetNNodesPerElement(mesh);
  nodes = e_GetNodes(mesh);

  NstressPNode = 3;
  NDisplPNode = 2;


  temp_node_strain = e_alloc_d_vect(NstressPNode);
  temp_node_stress = e_alloc_d_vect(NstressPNode);

   nod = icoord = E_NNPE_QUAD1;
   idee = ibee  = ih = 3;
   idof = 2*E_NNPE_QUAD1;
   ijac = ijac1  = it = ider = ideriv = E_NDIMS-1;
   isamp = 2;

 
  Dee = e_alloc_cont_d_matr(idee, idee);
  Guass_Samp = e_alloc_cont_d_matr(isamp, isamp);
  Coord = e_alloc_cont_d_matr(icoord, it);
  Jac = e_alloc_cont_d_matr(ijac, ijac);
  Jac1 = e_alloc_cont_d_matr(ijac1, ijac1);
  Der = e_alloc_cont_d_matr(ider, nod);
  Deriv = e_alloc_cont_d_matr(ideriv, nod);
  Bee = e_alloc_cont_d_matr(ibee,   idof);
  Fun = e_alloc_d_vect(nod); 

   B_gauss_Quad1(Guass_Samp);

  for(elem_no=0; elem_no<mesh->NElemsQuad1; elem_no++)
   {
    B_GetElemMaterial(mesh->MatsQuad1, elem_no, &YMod, 
                      &PoissonsRatio, &Thickness);
    B_Fmdsig(Dee,YMod,PoissonsRatio);
    B_Elem_cord(mesh, elem_no, Coord);
    for(i=0;i<2;i++)
     {
       for(j=0;j<2;j++)
         {
          B_form_coner(Der, Fun, i,j);
          B_Matmul(Der,Coord,Jac,it,nod,it);
          B_twoby2(Jac, Jac1,&Det);
          B_Matmul(Jac1,Der,Deriv,it,it,nod);
          B_null(Bee,ih,idof);
          B_form_Bee(Bee, Deriv, nod);

          Node_Number = mesh->NodesQuad1[elem_no][2*i+j]-1;

          B_form_strain(nodes,Temp_E_NNPE_ELEMs, Bee, Force,
                   temp_node_strain,elem_no,NstressPNode,NDisplPNode);

          B_form_stress(Dee,temp_node_strain, 
                       temp_node_stress, NstressPNode);


          for(k=0;k<NstressPNode;k++)
           {
             Nodal_Average_Strain[NstressPNode*Node_Number+k]+= 
                                  temp_node_strain[k];
             Nodal_Average_Stress[NstressPNode*Node_Number+k]+= 
                                  temp_node_stress[k];
           }


         }

      }

  }

  e_free(Dee); 
  e_free(Guass_Samp);
  e_free(Coord);
  e_free(Jac);
  e_free(Jac1);
  e_free(Der);
  e_free(Deriv);
  e_free(Bee);
  e_free(Fun);
  e_free(temp_node_strain);
  e_free(temp_node_stress);

  SetAdjPointsElems(mesh,&Degree,&AdjElems);

  for(i=0;i<mesh->NMeshPoints;i++)
    {
      for(j=0;j<NstressPNode;j++)
       {
          Nodal_Average_Strain[NstressPNode*i+j] /= Degree[i];
          Nodal_Average_Stress[NstressPNode*i+j] /= Degree[i];
       }
   
    }
 
   FreeAdjMeshPointsElems(mesh,Degree,AdjElems);
}


/*############################################################*/
void B_get_nodal_stress_Triang1(MESH *mesh,  double  *Ele_Non_Average_Strain,  double *Ele_Non_Average_Stress, 
 double  *Nodal_Average_Strain,   double *Nodal_Average_Stress)
{
 int i,j,k; 
int *Degree;
 int **AdjElems;
int NstressPNode = 3;



  SetAdjPointsElems(mesh,&Degree,&AdjElems);

  for(i=0;i<mesh->NMeshPoints;i++)
   {
      for(j=0;j<Degree[i];j++)
        {
          for(k=0;k<NstressPNode;k++)
            {
               Nodal_Average_Strain[NstressPNode*i+k]+= 
                       Ele_Non_Average_Strain
                       [NstressPNode*(AdjElems[i][j]-1)+k]; 
               Nodal_Average_Stress[NstressPNode*i+k]+= 
                       Ele_Non_Average_Stress
                       [NstressPNode*(AdjElems[i][j]-1)+k];  
            }
        }

       for(k=0;k<NstressPNode;k++)
        {
          Nodal_Average_Strain[NstressPNode*i+k]/=Degree[i];
          Nodal_Average_Stress[NstressPNode*i+k]/=Degree[i];
        }
   }

   FreeAdjMeshPointsElems(mesh,Degree,AdjElems);
}



/*############################################################*/
/* Forms four-node quadrilateral shape functions FUN
and derivatives DER in local coordinates for each coner node of
relevent element  */
void B_form_coner(double **Der, double *Fun, int i,int j)
{
double eta,xi,etam,etap,xim,xip;

/* still do not know which is right ????????????*/
/*
if(i==0 && j==0)
 {
   eta=1; xi=1;
 }
else if(i==0 && j==1)
  {
   eta= 1; xi =-1;
  }
else if(i==1 && j==0)
  {
    eta=-1; xi=1;
  }
else if(i==1 && j==1)
  {
  eta =-1; xi =-1;
  }
*/


if(i==0 && j==0)
 {
   eta=-1; xi=-1;
 }
else if(i==0 && j==1)
  {
   eta= 1; xi =-1;
  }
else if(i==1 && j==0)
  {
    eta=1; xi=1;
  }
else if(i==1 && j==1)
  {
  eta =-1; xi =1;
  }



etam=0.25*(1-eta);	etap=0.25*(1+eta);
xim=0.25*(1-xi);	   xip=0.25*(1+xi);

/* All element nodes should be numbered clockwise */

Fun[0]=4*xim*etam;	Fun[1]=4*xim*etap;
Fun[2]=4*xip*etap;	Fun[3]=4*xip*etam;

Der[0][0]=-etam;	   Der[0][1]=-etap;
Der[0][2]=etap;		   Der[0][3]=etam;

Der[1][0]=-xim;		   Der[1][1]=xim;
Der[1][2]=xip;		   Der[1][3]=-xip;

}



/*############################################################*/
void B_Calculate_Norm_e_Quad1(MESH *mesh, double *Force, 
                               double *Average_Nodal_Stress, 
                               double *Ele_Error_Stress_Norm, 
                               double *Ele_displacement_Energy_Norm)
{
int elem_no, i,j,k;
double Gauss_Point_strain_x,Gauss_Point_strain_y, Gauss_Point_strain_xy;
double Gauss_Point_stress_x,Gauss_Point_stress_y, Gauss_Point_stress_xy;
double Gauss_Point_Intep_stress_x, Gauss_Point_Intep_stress_y,
       Gauss_Point_Intep_stress_xy;
double Error_stress_x, Error_stress_y, Error_stress_xy;
double Error_Stress_Norm, eDe;
int ih, idee, ibee, nod, ijac, ijac1, ider, ideriv, it, idof, isamp, icoord;
double **Dee, **Guass_Samp, **Coord, **Jac, **Jac1, **Der, **Deriv, **Bee,*Fun;
double YMod, PoissonsRatio, Thickness, Det;
double **Inver_Dee;
int Temp_E_NNPE_ELEMs,**nodes;
int NstressPNode, NDisplPNode;
double *temp_node_strain, *temp_node_stress;

  NstressPNode = 3;
  NDisplPNode  = 2;

  temp_node_strain = e_alloc_d_vect(NstressPNode);
  temp_node_stress = e_alloc_d_vect(NstressPNode);

  Temp_E_NNPE_ELEMs = e_GetNNodesPerElement(mesh);
  nodes = e_GetNodes(mesh);


   nod = icoord = E_NNPE_QUAD1;
   idee = ibee  = ih = 3;
   idof = 2*E_NNPE_QUAD1;
   ijac = ijac1  = it = ider = ideriv = E_NDIMS-1;
   isamp = 2;

 
  Dee = e_alloc_cont_d_matr(idee, idee);
  Guass_Samp = e_alloc_cont_d_matr(isamp, isamp);
  Coord = e_alloc_cont_d_matr(icoord, it);
  Jac = e_alloc_cont_d_matr(ijac, ijac);
  Jac1 = e_alloc_cont_d_matr(ijac1, ijac1);
  Der = e_alloc_cont_d_matr(ider, nod);
  Deriv = e_alloc_cont_d_matr(ideriv, nod);
  Bee = e_alloc_cont_d_matr(ibee,   idof);
  Fun = e_alloc_d_vect(nod); 
  Inver_Dee = e_alloc_cont_d_matr(idee, idee);

  B_gauss_Quad1(Guass_Samp);

  for(elem_no=0; elem_no<mesh->NElemsQuad1; elem_no++)
   {
     B_GetElemMaterial(mesh->MatsQuad1, elem_no, &YMod, 
                       &PoissonsRatio, &Thickness);
     B_Fmdsig(Dee,YMod,PoissonsRatio);
     B_Matinversion(Dee,idee, Inver_Dee);
     B_Elem_cord(mesh, elem_no, Coord);
     Error_Stress_Norm = 0.0;
     for(i=0;i<2;i++)
       {
         for(j=0;j<2;j++)
           {
             B_formln(Der,Fun,Guass_Samp,i,j);
             B_Matmul(Der,Coord,Jac,it,nod,it);
	     B_twoby2(Jac, Jac1,&Det);
             B_Matmul(Jac1,Der,Deriv,it,it,nod);
             B_null(Bee,ih,idof);
	     B_form_Bee(Bee, Deriv, nod);



              B_form_strain(nodes,Temp_E_NNPE_ELEMs, Bee, Force,
                        temp_node_strain,elem_no,NstressPNode,NDisplPNode);

              B_form_stress(Dee,temp_node_strain, 
                            temp_node_stress, NstressPNode);



              B_energy_norm_stress_eDe(Inver_Dee, temp_node_stress[0], 
                                                  temp_node_stress[1], 
                                                  temp_node_stress[2], &eDe); 
 
              Ele_displacement_Energy_Norm[i] += Det*eDe;


             Gauss_Point_stress_x  =  temp_node_stress[0];
             Gauss_Point_stress_y  =  temp_node_stress[1];
             Gauss_Point_stress_xy =  temp_node_stress[2];

         
                   
             Gauss_Point_Intep_stress_x = 
             Gauss_Point_Intep_stress_y = 
             Gauss_Point_Intep_stress_xy =0.0;

             for(k=0;k<4;k++)
               {   
                 Gauss_Point_Intep_stress_x += 
                      Fun[k]* Average_Nodal_Stress
		           [3*(mesh->NodesQuad1[elem_no][k]-1)];
                 Gauss_Point_Intep_stress_y += 
                      Fun[k]*Average_Nodal_Stress
                           [ 3*(mesh->NodesQuad1[elem_no][k]-1)+1];
                 Gauss_Point_Intep_stress_xy += 
                      Fun[k]*Average_Nodal_Stress
                           [ 3*(mesh->NodesQuad1[elem_no][k]-1)+2];
               }
  
            Error_stress_x  = fabs(Gauss_Point_Intep_stress_x  
                                 - Gauss_Point_stress_x);
            Error_stress_y  = fabs(Gauss_Point_Intep_stress_y  
                                 - Gauss_Point_stress_y);
            Error_stress_xy = fabs(Gauss_Point_Intep_stress_xy 
                                 - Gauss_Point_stress_xy);

            B_energy_norm_stress_eDe(Inver_Dee, Error_stress_x, 
                                    Error_stress_y, Error_stress_xy, &eDe);
            Error_Stress_Norm  += Det*eDe;
         }
      }
   Ele_Error_Stress_Norm[elem_no] = Error_Stress_Norm;
  }

  e_free(Dee); 
  e_free(Guass_Samp);
  e_free(Coord);
  e_free(Jac);
  e_free(Jac1);
  e_free(Der);
  e_free(Deriv);
  e_free(Bee);
  e_free(Fun);
  e_free(Inver_Dee);
  e_free(temp_node_strain);
  e_free(temp_node_stress);

}




/*############################################################*/
/*This is the function using normal average technique to get the error of
stresses and calculate the norm */
void B_Calculate_Norm_e_QUAD1_average(MESH *mesh,double  *Ele_Non_Average_Strain,    double  *Ele_Non_Average_Stress,          
			               double  *Nodal_Average_Strain,    
			               double  *Nodal_Average_Stress,
                                       double  *Ele_Error_Stress_Norm, 
                                       double  *Ele_displacement_Energy_Norm)
{

int i,j,k,elem_no; 
double *Ele_Average_Strain, *Ele_Average_Stress;
double Error_stress_x, Error_stress_y, Error_stress_xy;
double area;
double eDe;
double YMod, PoissonsRatio, Thickness, Det;
double **Dee,**Inver_Dee;
int idee = 3;

  Ele_Average_Strain = e_alloc_d_vect(3*mesh->NElemsQuad1);
  Ele_Average_Stress = e_alloc_d_vect(3*mesh->NElemsQuad1);
  Dee = e_alloc_cont_d_matr(idee, idee);
  Inver_Dee = e_alloc_cont_d_matr(idee, idee);


 for(elem_no=0; elem_no<mesh->NElemsQuad1; elem_no++)
   {
     B_GetElemMaterial(mesh->MatsQuad1, elem_no, &YMod, 
                       &PoissonsRatio, &Thickness);
     B_Fmdsig(Dee,YMod,PoissonsRatio);
     B_Matinversion(Dee,idee, Inver_Dee);
 
    for(j=0;j<E_NEPE_QUAD1;j++)
        {
          for(k=0;k<3; k++)
            {
               Ele_Average_Strain[3*elem_no+k]+= 
                    Nodal_Average_Strain[3*(mesh->NodesQuad1[elem_no][j]-1)+k];  
               Ele_Average_Stress[3*elem_no+k]+= 
                    Nodal_Average_Stress[3*(mesh->NodesQuad1[elem_no][j]-1)+k];  
            }
        }
      for(k=0;k<3; k++)
        {
                 Ele_Average_Strain[3*elem_no+k]/=4.0;
                 Ele_Average_Stress[3*elem_no+k]/=4.0;
        }   
  

      Error_stress_x  = 
                 fabs(Ele_Average_Stress[3*elem_no]- Ele_Non_Average_Stress[3*elem_no]);
      Error_stress_y  = 
                 fabs(Ele_Average_Stress[3*elem_no+1]- Ele_Non_Average_Stress[3*elem_no+1]);
      Error_stress_xy  = 
                 fabs(Ele_Average_Stress[3*elem_no+2]- Ele_Non_Average_Stress[3*elem_no+2]);

       area = 1.0;

       B_energy_norm_stress_eDe(Inver_Dee, Ele_Average_Stress[3*elem_no], 
                                           Ele_Average_Stress[3*elem_no+1], 
                                           Ele_Average_Stress[3*elem_no+2], &eDe); 
       Ele_displacement_Energy_Norm[elem_no] =  area*eDe;
       B_energy_norm_stress_eDe(Inver_Dee, Error_stress_x, 
                                    Error_stress_y, Error_stress_xy, &eDe);
        Ele_Error_Stress_Norm[elem_no]  = area*eDe;
   }

e_free(Ele_Average_Strain);
e_free(Ele_Average_Stress);
e_free(Dee);
e_free(Inver_Dee);

}

/*############################################################*/
void B_Calculate_Norm_e_Triang1(MESH *mesh,double  *Ele_Non_Average_Strain, 
			               double  *Ele_Non_Average_Stress,          
			               double  *Nodal_Average_Strain,    
			               double  *Nodal_Average_Stress,
                                       double  *Ele_Error_Stress_Norm, 
                                       double  *Ele_displacement_Energy_Norm)
{

int i,j,k; 
double *Ele_Average_Strain, *Ele_Average_Stress;
double Error_stress_x, Error_stress_y, Error_stress_xy;
double area;
double eDe;
double YMod, PoissonsRatio, Thickness, Det;
double **Dee,**Inver_Dee;
int idee = 3;

  Ele_Average_Strain = e_alloc_d_vect(3*mesh->NElemsTriang1);
  Ele_Average_Stress = e_alloc_d_vect(3*mesh->NElemsTriang1);
  Dee = e_alloc_cont_d_matr(idee, idee);
  Inver_Dee = e_alloc_cont_d_matr(idee, idee);


  for(i=0;i<mesh->NElemsTriang1;i++)
   {
     B_GetElemMaterial(mesh->MatsTriang1, i, &YMod, &PoissonsRatio, &Thickness);
     B_Fmdsig(Dee,YMod,PoissonsRatio);
     B_Matinversion(Dee,idee, Inver_Dee);   
     
     for(j=0;j<E_NEPE_TRIANG1;j++)
        {
          for(k=0;k<3; k++)
            {
               Ele_Average_Strain[3*i+k]+= 
                    Nodal_Average_Strain[3*(mesh->NodesTriang1[i][j]-1)+k];  
               Ele_Average_Stress[3*i+k]+= 
                    Nodal_Average_Stress[3*(mesh->NodesTriang1[i][j]-1)+k];  
            }
        }
      for(k=0;k<3; k++)
        {
                 Ele_Average_Strain[3*i+k]/=3.0;
                 Ele_Average_Stress[3*i+k]/=3.0;
        }   
  

      Error_stress_x  = 
                 fabs(Ele_Average_Stress[3*i]- Ele_Non_Average_Stress[3*i]);
      Error_stress_y  = 
                 fabs(Ele_Average_Stress[3*i+1]- Ele_Non_Average_Stress[3*i+1]);
      Error_stress_xy  = 
                 fabs(Ele_Average_Stress[3*i+2]- Ele_Non_Average_Stress[3*i+2]);

       B_elem_area_Triang1(mesh,i,&area);
       B_energy_norm_stress_eDe(Inver_Dee, Ele_Average_Stress[3*i], 
                                           Ele_Average_Stress[3*i+1], 
                                           Ele_Average_Stress[3*i+2], &eDe); 
       Ele_displacement_Energy_Norm[i] =  area*eDe;
       B_energy_norm_stress_eDe(Inver_Dee, Error_stress_x, 
                                    Error_stress_y, Error_stress_xy, &eDe);
        Ele_Error_Stress_Norm[i]  = area*eDe;
   }

e_free(Ele_Average_Strain);
e_free(Ele_Average_Stress);
e_free(Dee);
e_free(Inver_Dee);

}

/*############################################################*/
/* function calculates the area of the triangular element */
void B_elem_area_Triang1(MESH *mesh,int NElem,double *area)
{
double x1,y1,x2,y2,x3,y3;

   x1 =  mesh->MeshPointCoords[mesh->NodesTriang1[NElem][0]-1][0];
   y1 =  mesh->MeshPointCoords[mesh->NodesTriang1[NElem][0]-1][1];

   x2 =  mesh->MeshPointCoords[mesh->NodesTriang1[NElem][1]-1][0];
   y2 =  mesh->MeshPointCoords[mesh->NodesTriang1[NElem][1]-1][1];

   x3 =  mesh->MeshPointCoords[mesh->NodesTriang1[NElem][2]-1][0];
   y3 =  mesh->MeshPointCoords[mesh->NodesTriang1[NElem][2]-1][1];

 
   *area = fabs(.5*((x2*y3-x3*y2)-(x1*y3-x3*y1)+(x1*y2-x2*y1)));
  }


/*############################################################*/
/* function to calculate the energy norm */
void B_energy_norm_stress_eDe(double **Inver_Dee, double Error_stress_x, 
                    double Error_stress_y, double Error_stress_xy, double *eDe)
{
int j;
int test;
double c[3];

 /* mult row of avg_err and avg_ste with col of dee(inver) */

c[0] = Error_stress_x*Inver_Dee[0][0]+
       Error_stress_y*Inver_Dee[0][1]+
       Error_stress_xy*Inver_Dee[0][2];
c[1] = Error_stress_x*Inver_Dee[1][0]+
       Error_stress_y*Inver_Dee[1][1]+
       Error_stress_xy*Inver_Dee[1][2];
c[2] = Error_stress_x*Inver_Dee[2][0]+
       Error_stress_y*Inver_Dee[2][1]+
       Error_stress_xy*Inver_Dee[2][2];

*eDe = c[0]*Error_stress_x + c[1]*Error_stress_y + c[2]*Error_stress_xy;

}

/*############################################################*/
void B_Matinversion(double **A, int Size_matrix, double **B)
{

int i,j;
double *I;
double **T_AA;
int *temp;

  T_AA = e_alloc_cont_d_matr(Size_matrix, Size_matrix);
  temp = e_alloc_i_vect(Size_matrix);

  for(i=0; i<Size_matrix;i++)
    for(j=0; j<Size_matrix; j++)
       T_AA[i][j] = A[i][j];

  I = e_alloc_d_vect(Size_matrix);

  e_LUFact(Size_matrix, T_AA, temp);

  for(i=0; i<Size_matrix; i++)
    {
      for(j=0; j<Size_matrix; j++)  I[j]=0.0;
      I[i] = 1.0;
 
      e_LUSolve(Size_matrix, T_AA, I,temp);
      for(j=0;j<Size_matrix;j++) B[j][i] = I[j];
    }

  e_free(I);
  e_free(T_AA);
  e_free(temp);
}


/*############################################################*/
void B_Error_ElemSize_NodalSize(MESH *mesh, 
				double *Ele_Error_Stress_Norm,
 				double *Ele_displacement_Energy_Norm, 
			 	double *Global_Error_Stress_Norm, 
 				double *Global_displacement_Energy_Norm, 
 				double Permissib_Error, 
 				double *Elem_Error, 
 				double *Nodal_Size_Param)
{
int i,j;
double Nita,em,exi;
int *Degree;
int **AdjElems;
double *Elem_Size;

int ElemType;

  ElemType = e_GetElementType(mesh);

  Elem_Size = e_alloc_d_vect(mesh->TotalNElems);

 *Global_Error_Stress_Norm = *Global_displacement_Energy_Norm =0.0;

 for(i=0; i<mesh->TotalNElems; i++)
  {
    *Global_Error_Stress_Norm  += Ele_Error_Stress_Norm[i];
    *Global_displacement_Energy_Norm  += Ele_displacement_Energy_Norm[i];
  }

 Nita = sqrt(*Global_Error_Stress_Norm / (*Global_Error_Stress_Norm + 
                                         *Global_displacement_Energy_Norm));

 if(Nita<0.0) e_GeneralError("Error: Nita is negtive in function error estimator ");


 for(i=0; i<mesh->TotalNElems; i++)
   {  
    if((Ele_Error_Stress_Norm[i] + Ele_displacement_Energy_Norm[i])<0.0)
         e_GeneralWarning("Error:Norm sum is negtive----B_Error_ElemSize_NodalSize");
    else if((Ele_Error_Stress_Norm[i] + Ele_displacement_Energy_Norm[i])<1.0e-30)
         Elem_Error[i] = 0.0;
    else Elem_Error[i] = Ele_Error_Stress_Norm[i]/(Ele_Error_Stress_Norm[i] + 
                                             Ele_displacement_Energy_Norm[i]);
   }
 em = Permissib_Error*sqrt((*Global_Error_Stress_Norm +
                             *Global_displacement_Energy_Norm)
                             /mesh->TotalNElems);

 if(ElemType ==E_ELEM_TYPE_TRIANG1)
        B_Cal_Quad_Elem_Size_Triang1(mesh, Elem_Size);
  else if(ElemType ==E_ELEM_TYPE_QUAD1)
              B_Cal_Quad_Elem_Size_Quad1(mesh, Elem_Size);

 
 for(i=0; i<mesh->TotalNElems;i++)
   {
     exi = 1.1 *sqrt(Ele_Error_Stress_Norm[i])/em;

     if(exi <1.0) exi = 1.0;
     Elem_Size[i] /= exi;
   }


 SetAdjPointsElems(mesh,&Degree,&AdjElems);

 for(i=0; i<mesh->NMeshPoints;i++)
   {
     for(j=0; j<Degree[i]; j++)
               Nodal_Size_Param[i]  += Elem_Size[AdjElems[i][j]-1];

     Nodal_Size_Param[i] /=Degree[i];
   }

 FreeAdjMeshPointsElems(mesh,Degree,AdjElems);
 e_free(Elem_Size);
}


/*############################################################*/
void B_Cal_Quad_Elem_Size_Quad1(MESH *mesh, double *Elem_Size)
{
int i;
double l1,l2;

for(i=0;i< mesh->NElemsQuad1;i++)
 {
   l1= sqrt(pow(mesh->MeshPointCoords[mesh->NodesQuad1[i][0]-1][0]-
                mesh->MeshPointCoords[mesh->NodesQuad1[i][2]-1][0],2.0)+
            pow(mesh->MeshPointCoords[mesh->NodesQuad1[i][0]-1][1]-
                mesh->MeshPointCoords[mesh->NodesQuad1[i][2]-1][1],2.0));

   l2= sqrt(pow(mesh->MeshPointCoords[mesh->NodesQuad1[i][1]-1][0]-
                mesh->MeshPointCoords[mesh->NodesQuad1[i][3]-1][0],2.0)+
            pow(mesh->MeshPointCoords[mesh->NodesQuad1[i][1]-1][1]-
                mesh->MeshPointCoords[mesh->NodesQuad1[i][3]-1][1],2.0));
  
   if(l2>l1) Elem_Size[i]=l2; 
   else      Elem_Size[i]=l1;
 } 

}

/*############################################################*/
void B_Cal_Quad_Elem_Size_Triang1(MESH *mesh, double *Elem_Size)
{
int i,j,k;
double l, Max_Length;

for(i=0;i< mesh->NElemsTriang1;i++)
 { 
   Max_Length = 0.0;

   for(j=0;j<3;j++)
    {
      if(j==2) k=0;
      else k=j+1; 
      l= sqrt(pow(mesh->MeshPointCoords[mesh->NodesTriang1[i][j]-1][0]-
                  mesh->MeshPointCoords[mesh->NodesTriang1[i][k]-1][0],2.0)+
              pow(mesh->MeshPointCoords[mesh->NodesTriang1[i][j]-1][1]-
                  mesh->MeshPointCoords[mesh->NodesTriang1[i][k]-1][1],2.0));

     if(l>Max_Length) Max_Length = l;
    }
  Elem_Size[i]=Max_Length; 
  }
}


/*############################################################*/
void B_Average_Elem_stress(MESH *mesh, double *Ele_Average_Stress, 
                    double *Average_Nodal_Stress)
{
 int i,j,k; 

  for(i=0;i<mesh->NElemsQuad1;i++)
   {
      for(j=0;j<4;j++)
        {
          for(k=0;k<3; k++)
            {
               Ele_Average_Stress[3*i+k]+= 
                    Average_Nodal_Stress[3*(mesh->NodesQuad1[i][j]-1)+k];  
            }
        }
      for(k=0;k<3; k++)
                 Ele_Average_Stress[3*i+k]/=4.0;
   }
}

/*############################################################*/
void  B_Average_Nodal_stress(MESH *mesh, double *Ele_Non_Average_Stress)
{

 int i,j,k; 
 int *Degree;
 int **AdjElems;
 double *Ele_Average_Nodal_Stress;

  Ele_Average_Nodal_Stress = e_alloc_d_vect(3*mesh->NMeshPoints);

  SetAdjPointsElems(mesh,&Degree,&AdjElems);

  for(i=0;i<mesh->NMeshPoints;i++)
   {
      for(j=0;j<Degree[i];j++)
        {
          for(k=0;k<3;k++)
            {
               Ele_Average_Nodal_Stress[3*i+k]+= 
                    Ele_Non_Average_Stress[3*(AdjElems[i][j]-1)+k];  
            }
        }
       Ele_Average_Nodal_Stress[3*i+k]/=Degree[i];
   }

  FreeAdjMeshPointsElems(mesh,Degree,AdjElems);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetAdjMeshPoints: To determine which mesh-points are adjacent to which
   (i.e. share an edge).                                                      */


int SetAdjPointsElems(MESH *mesh, int **NAdjMeshPointsElems,
                                         int ***AdjMeshPointsElems)
{
   int i,j,l,k1,k2;
   int ElemType;
   int NNpE;
   int **Vertices;


   ElemType = e_GetElementType(mesh);

   *NAdjMeshPointsElems = e_alloc_i_vect(mesh->NMeshPoints);
   *AdjMeshPointsElems = e_init_frag_i_matr(mesh->NMeshPoints);

   NNpE = e_GetNNodesPerElement(mesh);
   Vertices = e_GetVertices(mesh);

   if (   ElemType == E_ELEM_TYPE_TRIANG1
       || ElemType == E_ELEM_TYPE_TRIANG2
       || ElemType == E_ELEM_TYPE_QUAD1
       || ElemType == E_ELEM_TYPE_TETRAH1  )
   {
      for (i = 0; i < mesh->TotalNElems; i++)
      {
         k2 = i;
         for (j = 0; j < NNpE; j++)
         {
            k1 = Vertices[i][j] - 1;
            UpdateMeshPointsElems(k1, k2, *NAdjMeshPointsElems,
                                          *AdjMeshPointsElems);
         }
      }
   }
   else e_GeneralError("Error: unsupported element type in SetAdjPointsElems");
 

   return 1;
}

/* ////////////////////////////////////////////////////////////////////////// */
/* UpdateMeshPoints: To update the adjacent mesh-point data. */


int UpdateMeshPointsElems(int k1, int k2, int *NAdjMeshPoints,
                                             int **AdjMeshPoints)
{
   
      AdjMeshPoints[k1]
         = e_incr_i_vect(NAdjMeshPoints[k1],AdjMeshPoints[k1],
                                     ADJ_MESH_POINTS_MEM_INCREM);
      AdjMeshPoints[k1][NAdjMeshPoints[k1]++] = k2 + 1;
  
   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* FreeAdjMeshPoints: To free memory allocated for the mesh-point adjacency
   data.                                                                      */


void FreeAdjMeshPointsElems(MESH *mesh, int *NAdjMeshPoints,
                                           int **AdjMeshPoints)
{
   e_free(NAdjMeshPoints);
   e_free_frag_i_matr(mesh->NMeshPoints,AdjMeshPoints);
}
