#include <math.h>
#include <e_lib.h>
#include "fem.h"


int main(int argc, char *argv[])
{
    MESH *mesh;
    int ElemType;
    int N_displacement_Per_node;

    double *KV, *Force; 
    int *ind;
    int i;


    if(argc != 4)
    {
	printf("\nfem v1.1 SECT Research Group, Heriot-Watt University Edinburgh\n");
	printf("usage: fem input output\n");
	printf("note all filenames are referred to without extensions\n");
	printf("   input    : name of the mesh definition and material file\n");
	printf("   output   : name of the output mesh definition file\n");
	printf("   ds	    : displacement scale (default=10)\n\n");   
	printf("you might want to renumber the mesh first!\n");
	printf("and remove unused nodes!\n\n");
	exit(0);
    }

    e_GetMeshData(&mesh, argv[1]);
    e_GetMaterials(mesh, argv[1]);
  
    ElemType = e_GetElementType(mesh);

    B_clockwise_check(mesh, ElemType);

    if(ElemType ==E_ELEM_TYPE_TETRAH1)
	N_displacement_Per_node = 3;
    else N_displacement_Per_node = 2;

    ind = e_alloc_i_vect(N_displacement_Per_node*(mesh->NMeshPoints)); 
    Force = e_alloc_d_vect(N_displacement_Per_node*(mesh->NMeshPoints));
 
    B_form_dia(mesh,ind,N_displacement_Per_node);
    KV = e_alloc_d_vect(ind[N_displacement_Per_node*
			(mesh->NMeshPoints-1)+N_displacement_Per_node - 1]); 

    B_form_loads(mesh, Force,N_displacement_Per_node);

    if(ElemType ==E_ELEM_TYPE_TRIANG1)
	B_Assemble_matrix_TRIANG1(mesh,KV,ind,Force);                   
    else if(ElemType ==E_ELEM_TYPE_QUAD1)
	B_Assemble_matrix_QUAD1(mesh,KV,ind,Force);
    else if(ElemType ==E_ELEM_TYPE_TETRAH1) 
	B_Assemble_matrix_TETRAH1(mesh,KV,ind,Force);
    else e_GeneralError("Error: Unsupported element type");
 
    B_BoundaryConditions(mesh, KV, Force, ind, N_displacement_Per_node); 

    B_bandv(N_displacement_Per_node*mesh->NMeshPoints, KV, Force, ind);

    e_free(ind);
    e_free(KV);
   
    J_Calculate_Stresses(mesh, Force);

    B_deformationMesh(mesh,Force,atof(argv[3]), N_displacement_Per_node); 
    e_PutMeshData(mesh,argv[2]);
    e_PutStresses(mesh,argv[2]);
    e_PutMaterials(mesh,argv[2]);

    e_free(Force);
    e_FreeMeshData(mesh);
}

