/* FILE floor.c
 *
 * to average mesh parameter files
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "e_lib.h"

int main(int argc, char *argv[])
{
  /* variable declarations */
  MESH    *mesh;			/* input mesh structure */
  char     line[E_LINE_MAX];		/* answer */
  int	   i, n, emp;
  double   max, min, ave, newv;

  if(argc == 3)
  {
    mesh=NULL;
    if (!e_GetMeshData(&mesh, argv[1]))
        e_GeneralError("Mesh file contains faults; execution halted!");
    if (!e_GetMeshParams(mesh, argv[1]))
        e_GeneralError("Meshparams corrupted; execution halted!");
    
    emp=mesh->MeshParam.ElemMeshParamsSet;
    if (emp)
        n=mesh->NElemsTriang1;
    else
        n=mesh->NNodes;
    
    max=-1e8;
    min=1e8;
    ave=0.0;
    
    for (i=0;i<n;i++)
    {
      if (emp)
          newv=mesh->MeshParam.ElemMeshParamsTriang1[i];
      else
          newv=mesh->MeshParam.NodalMeshParams[i];
      if (newv>max)
          max=newv;
      if (newv<min)
          min=newv;
      ave+=newv;
    }
    ave/=n;
    
    printf("\nMesh parameter statistics:\n");
    printf("  maximum: %1.8lf\n", max);
    printf("  minimum: %1.8lf\n", min);
    printf("  average: %1.8lf\n", ave);
    printf("\nEnter new maximum: ");
    fgets(line, E_LINE_MAX, stdin);
    if(sscanf(line, "%lf", &max) != 1)
        e_GeneralError("Invalid number!");
    printf("\nEnter new minimum: ");
    fgets(line, E_LINE_MAX, stdin);
    if(sscanf(line, "%lf", &min) != 1)
        e_GeneralError("Invalid number!");
    
    for (i=0;i<n;i++)
    {
      if (emp)
      {
        if (mesh->MeshParam.ElemMeshParamsTriang1[i]<min)
            mesh->MeshParam.ElemMeshParamsTriang1[i]=min;
        if (mesh->MeshParam.ElemMeshParamsTriang1[i]>max)
            mesh->MeshParam.ElemMeshParamsTriang1[i]=max;
      }
      else
      {
        if (mesh->MeshParam.NodalMeshParams[i]<min)
            mesh->MeshParam.NodalMeshParams[i]=min;
        if (mesh->MeshParam.NodalMeshParams[i]>max)
            mesh->MeshParam.NodalMeshParams[i]=max;
      }
    }
    
    e_PutMeshParams(mesh, argv[2]);
    e_FreeMeshData(mesh);
    
  }
  else
  {
    /* show the usage of the program */
    printf("\nLimit the maximum and minimum values of the mesh parameter.\n");
    printf("\nUsage: floor input output\n");
    printf("   input  : name of input mesh file and mesh parameter file\n");
    printf("   output : name of output mesh parameter file\n");
    printf("All files should be specified without extension!\n\n");
    e_exit(1);
  }
  return 0;
}

/* end of file */
