/********************************************************************
 makedom.c --
        makedom creates domain decomposition file (*.dom).
        All elements in the mesh will belong to the same 
        domaint but it is possible to specify domain number
        (if domain number is specified number of subdomains
        will equal domain number and all subdomains will
        be empty except the one with specified number)
 
 ********************************************************************/
    
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include "e_lib.h"

#define VERSION "1.0"

        int 
main (int argc, char **argv)
{
  int c;
  int i;
  extern char *optarg;
  extern int optind;
  int errflag=0;
  int  domNumber = 1; 
  MESH *mesh;

  while ((c = getopt(argc, argv, "hvn:")) != EOF)
  {
    switch (c) 
    {
      case 'h':
      {
        errflag = 1;       
        break;
      }
      case 'v':   
      {
        fprintf (stderr, "%s version %s\n",argv[0], VERSION);
        exit (0); 
      }
      case 'n':
      {
        domNumber = atoi (optarg);
        break;
      }
      case '?':
        errflag = 1;
     }
  }
  if (errflag == 1 || ((optind+1) != argc))
  {
    fprintf(stderr, "\nCreate a domain decomposition file (*.dom).\n");
    fprintf(stderr, "\nUsage: makedom [-v|-h|-n<domain>] mdf_file\n");
    fprintf(stderr, "\t-v - print version information\n");
    fprintf(stderr, "\t-h - print this message\n");
    fprintf(stderr, "\t-n<domain> - set domain number (default 1)\n\n");
    exit(1);
  }

  e_GetMeshData (&mesh, argv[optind]);
  /* allocate memory for subdomain indices */
  mesh->Decomp.SubDoms = e_alloc_i_vect (mesh->TotalNElems); 
  if (mesh->Decomp.SubDoms == NULL)
      e_GeneralError ("Cannot allocate memory... Aborting\n");
  /* fill subdomain information */
  for (i=0; i<mesh->TotalNElems; i++)
  {
    mesh->Decomp.SubDoms[i] = domNumber;
  }
  /* this is a dirty hack,
     it should point to the propert place in the vector */
  mesh->Decomp.SubDomsLink1 = mesh->Decomp.SubDoms;
  mesh->Decomp.SubDomsLink2 = mesh->Decomp.SubDoms;
  mesh->Decomp.SubDomsLink3 = mesh->Decomp.SubDoms;
  mesh->Decomp.SubDomsLink4 = mesh->Decomp.SubDoms;
  mesh->Decomp.SubDomsLink5 = mesh->Decomp.SubDoms;
  mesh->Decomp.SubDomsTriang1 = mesh->Decomp.SubDoms;
  mesh->Decomp.SubDomsTriang2 = mesh->Decomp.SubDoms;
  mesh->Decomp.SubDomsTriang3 = mesh->Decomp.SubDoms;
  mesh->Decomp.SubDomsTriang4 = mesh->Decomp.SubDoms;
  mesh->Decomp.SubDomsTriang5 = mesh->Decomp.SubDoms;
  mesh->Decomp.SubDomsQuad1 = mesh->Decomp.SubDoms;
  mesh->Decomp.SubDomsQuad3 = mesh->Decomp.SubDoms;
  mesh->Decomp.SubDomsTetrah1 = mesh->Decomp.SubDoms;
  mesh->Decomp.SubDomsBlock1 = mesh->Decomp.SubDoms;

  mesh->Decomp.NSubDoms = domNumber;
  mesh->SubStructSet[E_SS_NUM_DECOMP] = TRUE;

  e_PutDecompData(mesh, argv[optind]);
  e_FreeMeshData(mesh);

 return 0;
}



