#include "e_lib.h"

int main(int argc, char *argv[])
{
  MESH *m;
  int	i;
  
  if(argc != 3)
  {
    printf("\nCreates material information and an example material file\n");
    printf("\nUsage: makemat input output\n");
    printf("   input  : name of input mesh file\n");
    printf("   output : name of output mesh and material file\n");
    printf("All files should be specified without extension!\n\n");
    exit(1);
  }	
  
  e_GetMeshData(&m,argv[1]);
  if (!e_IsOfOneElementType(m))
      e_GeneralError("makemat works only for single elem type meshes");
  if (m->TotalNElems!=m->NElemsLink1+m->NElemsTriang3+m->NElemsTriang1)
      e_GeneralError("Only for LINK1, TRIANG1 or TRIANG3 elements are allowed");
  
  m->NMats=1;
  m->KeywordSet[E_KW_NUM_NMATS]=TRUE;
  m->Props=(MAT *)e_calloc(1,sizeof(MAT));
  m->SubStructSet[E_SS_NUM_MAT_PROPS]=TRUE;
  
  if (m->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
  {
    m->Props[0].Name=e_alloc_c_vect(strlen("CLOTH1")+1);
    strcpy(m->Props[0].Name,"CLOTH1");
    m->Props[0].Density=600;
    m->Props[0].KeywordSet[E_MKW_NUM_DENSITY]=TRUE;
    m->Props[0].YMod=1.2e6;
    m->Props[0].KeywordSet[E_MKW_NUM_YMOD]=TRUE;
    m->Props[0].PoissonsRatio=0.3;
    m->Props[0].KeywordSet[E_MKW_NUM_POISSONS_RATIO]=TRUE;
    m->Props[0].Thickness=0.002;
    m->Props[0].KeywordSet[E_MKW_NUM_THICKNESS]=TRUE;
    m->Props[0].KeywordSet[E_MKW_NUM_END]=TRUE;
  }
  if (m->KeywordSet[E_KW_NUM_NELEMS_LINK1])
  {
    m->Props[0].Name=e_alloc_c_vect(strlen("CABLE1")+1);
    strcpy(m->Props[0].Name,"CABLE1");
    m->Props[0].Density=7800;
    m->Props[0].KeywordSet[E_MKW_NUM_DENSITY]=TRUE;
    m->Props[0].YMod=2e09;
    m->Props[0].KeywordSet[E_MKW_NUM_YMOD]=TRUE;
    m->Props[0].iParam1=1;
    m->Props[0].KeywordSet[E_MKW_NUM_IPARAM1]=TRUE;
    m->Props[0].dParam1=0.0003;
    m->Props[0].KeywordSet[E_MKW_NUM_DPARAM1]=TRUE;
    m->Props[0].dParam2=2e14;
    m->Props[0].KeywordSet[E_MKW_NUM_DPARAM2]=TRUE;
    m->Props[0].KeywordSet[E_MKW_NUM_END]=TRUE;
  }
  
  if (m->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
  {
    m->Props[0].Name=e_alloc_c_vect(strlen("STEEL")+1);
    strcpy(m->Props[0].Name,"STEEL");
    m->Props[0].Density=7800;
    m->Props[0].KeywordSet[E_MKW_NUM_DENSITY]=TRUE;
    m->Props[0].YMod=2e12;
    m->Props[0].KeywordSet[E_MKW_NUM_YMOD]=TRUE;
    m->Props[0].PoissonsRatio=0.3;
    m->Props[0].KeywordSet[E_MKW_NUM_POISSONS_RATIO]=TRUE;
    m->Props[0].Thickness=0.01;
    m->Props[0].KeywordSet[E_MKW_NUM_THICKNESS]=TRUE;
    m->Props[0].KeywordSet[E_MKW_NUM_END]=TRUE;
  }
  
  m->Mats=e_alloc_vp_vect(m->TotalNElems);
  if (m->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
  {
    m->MatsTriang3=(MAT **)&(m->Mats[m->NElemsLink1+m->NElemsTriang1]);
    m->KeywordSet[E_KW_NUM_MATS_TRIANG3]=TRUE;
    for (i=0;i<m->NElemsTriang3;i++)
        m->MatsTriang3[i]=&(m->Props[0]);
  }
  if (m->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
  {
    m->MatsTriang1=(MAT **)&(m->Mats[m->NElemsLink1]);
    m->KeywordSet[E_KW_NUM_MATS_TRIANG1]=TRUE;
    for (i=0;i<m->NElemsTriang1;i++)
        m->MatsTriang1[i]=&(m->Props[0]);
  }
  if (m->KeywordSet[E_KW_NUM_NELEMS_LINK1])
  {
    m->MatsLink1=(MAT **)m->Mats;
    m->KeywordSet[E_KW_NUM_MATS_LINK1]=TRUE;
    for (i=0;i<m->NElemsLink1;i++)
        m->MatsLink1[i]=&(m->Props[0]);
  }
  
  e_PutMaterials(m, argv[2]);
  e_PutMeshData(m, argv[2]);
  
  e_FreeMeshData(m);

  return(0);
}
