/********************************************************************
 makempr.c --
        makempr creates mesh parametes file (*.mpr)
        The mesh parameter can be specified either for nodes
        (nodal mesh parameter) or elements or for both.
        Nodal mesh parameter can be calculated automaticaly
        from the edge lengths. Automatic calculation is the default.
        
 ********************************************************************/
    
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include "e_lib.h"
#include "findmpr.h"

#define VERSION "1.1"

        int 
main (int argc, char **argv)
{
  int c;
  int i;
  extern char *optarg;
  extern int optind;
  int errflag=0;
  double nodalMeshParam = 1.0; 
  double elemMeshParam = 1.0;  
  int meshParamType = 'a'; 
  MESH *mesh;

  while ((c = getopt(argc, argv, "hvan:e:")) != EOF)
  {
    switch (c) 
    {
      case 'h':
      {
        errflag = 1;       
        break;
      }
      case 'v':   
      {
        fprintf (stderr, "%s version %s\n",argv[0], VERSION);
        exit (0); 
      }
      case 'n':
      {
        nodalMeshParam = atof (optarg);
        if (meshParamType != 'e' && meshParamType != 'e' + 'n')
        {
          meshParamType = 'n';
        }
        else
        {
          meshParamType = 'n' + 'e';
        }
        break;
      }
      case 'e':
      {
        elemMeshParam = atof (optarg);
        if (meshParamType != 'n' && meshParamType != 'e' + 'n')
        {
          meshParamType = 'e';
        }
        else
        {
          meshParamType = 'n' + 'e';
        }
        break;
      }
      case 'a':
      {
        meshParamType = 'a';
        break;
      }
      case '?':
        errflag = 1;
     }
  }
  if (errflag == 1 || ((optind+1) != argc))
  {
    fprintf(stderr, "\nCreates mesh parameter file (element or nodal)\n");
    fprintf(stderr, "\nUsage: makempr [-v|-h|-n<meshpram>|-e<meshparam>|-a] mdf_file\n");
    fprintf(stderr, "\t-v - print version information\n");
    fprintf(stderr, "\t-h - print this message\n");
    fprintf(stderr,
     "\t-n<meshparam> - set value of nodal mesh parameter \n");
    fprintf(stderr,
     "\t-e<meshparam> - set value of element mesh parameter \n");
    fprintf(stderr,
     "\t-a - automatic calculation of nodal mesh parameter (default)\n");
    
    exit (2);
  }

  e_GetMeshData (&mesh, argv[optind]);

  /* allocate memory for nodal and element mesh parameter and
   * fill the mesh parameters arrays  
   */
  switch (meshParamType)
  {
    case 'a':
    {
      mesh->MeshParam.NodalMeshParams = e_alloc_d_vect (mesh->NMeshPoints); 
      if (mesh->MeshParam.NodalMeshParams == NULL)
         e_GeneralError ("Cannot allocate memory... Aborting\n");

      FindNodalMpr(mesh);

      mesh->MeshParam.NodalMeshParamsSet = TRUE;
      break;
    }
    case 'n':
    {
      mesh->MeshParam.NodalMeshParams = e_alloc_d_vect (mesh->NMeshPoints); 
      if (mesh->MeshParam.NodalMeshParams == NULL)
         e_GeneralError ("Cannot allocate memory... Aborting\n");

      /* fill nodal mesh parameters */
      for (i=0; i<mesh->NMeshPoints; i++)
      {
        mesh->MeshParam.NodalMeshParams[i] = nodalMeshParam;
      }  
      mesh->MeshParam.NodalMeshParamsSet = TRUE;
      break;
    }                
    case ('e'+'n'): /* FALL THROUGH */
    {
      mesh->MeshParam.NodalMeshParams = e_alloc_d_vect (mesh->NMeshPoints); 
      if (mesh->MeshParam.NodalMeshParams == NULL)
         e_GeneralError ("Cannot allocate memory... Aborting\n");
      /* fill nodal mesh parameters */
      for (i=0; i<mesh->NMeshPoints; i++)
      {
        mesh->MeshParam.NodalMeshParams[i] = nodalMeshParam;
      }  
      mesh->MeshParam.NodalMeshParamsSet = TRUE;
    }
    case 'e':
    {
      mesh->MeshParam.ElemMeshParams = e_alloc_d_vect (mesh->TotalNElems); 
      if (mesh->MeshParam.ElemMeshParams == NULL)
         e_GeneralError ("Cannot allocate memory... Aborting\n");
      break; 
      
      /* fill elem mesh parameters */
      for (i=0; i<mesh->TotalNElems; i++)
      {
        mesh->MeshParam.ElemMeshParams[i] = elemMeshParam;
      }  
      /* set element type pointers */
      mesh->MeshParam.ElemMeshParamsLink1 = mesh->MeshParam.ElemMeshParams + 
              mesh->FirstNetIndex[E_ELEM_TYPE_LINK1] - 1; 
      mesh->MeshParam.ElemMeshParamsLink2 = mesh->MeshParam.ElemMeshParams + 
              mesh->FirstNetIndex[E_ELEM_TYPE_LINK2] - 1; 
      mesh->MeshParam.ElemMeshParamsLink3 = mesh->MeshParam.ElemMeshParams + 
              mesh->FirstNetIndex[E_ELEM_TYPE_LINK3] - 1; 
      mesh->MeshParam.ElemMeshParamsLink4 = mesh->MeshParam.ElemMeshParams + 
              mesh->FirstNetIndex[E_ELEM_TYPE_LINK4] - 1; 
      mesh->MeshParam.ElemMeshParamsLink5 = mesh->MeshParam.ElemMeshParams + 
              mesh->FirstNetIndex[E_ELEM_TYPE_LINK5] - 1; 
      mesh->MeshParam.ElemMeshParamsTriang1 = mesh->MeshParam.ElemMeshParams + 
              mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] - 1; 
      mesh->MeshParam.ElemMeshParamsTriang2 = mesh->MeshParam.ElemMeshParams + 
              mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2] - 1; 
      mesh->MeshParam.ElemMeshParamsTriang3 = mesh->MeshParam.ElemMeshParams + 
              mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] - 1; 
      mesh->MeshParam.ElemMeshParamsTriang4 = mesh->MeshParam.ElemMeshParams + 
              mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] - 1; 
      mesh->MeshParam.ElemMeshParamsTriang5 = mesh->MeshParam.ElemMeshParams + 
              mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] - 1; 
      mesh->MeshParam.ElemMeshParamsQuad1 = mesh->MeshParam.ElemMeshParams + 
              mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1] - 1; 
      mesh->MeshParam.ElemMeshParamsQuad3 = mesh->MeshParam.ElemMeshParams + 
              mesh->FirstNetIndex[E_ELEM_TYPE_QUAD3] - 1; 
      mesh->MeshParam.ElemMeshParamsTetrah1 = mesh->MeshParam.ElemMeshParams + 
              mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1] - 1; 
      mesh->MeshParam.ElemMeshParamsBlock1 = mesh->MeshParam.ElemMeshParams + 
              mesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1] - 1; 

      mesh->MeshParam.ElemMeshParamsSet = TRUE;
    }
    default:
    {
      e_GeneralError ("Internal error: Unknown mesh parameter type (%d)\n",
                      meshParamType);
    }
  } 

  mesh->SubStructSet [E_SS_NUM_MESHPARAM] = TRUE;

  e_PutMeshParams (mesh, argv[optind]);
  e_FreeMeshData (mesh);

 return 0;
}
