/********************************************
 Straight forward merging of 2 mdf files.
 All elem types.
 Merging of Decomp values as well.
 But no merging of loads, BCs, Stresses etc.
 ********************************************/

#include <e_lib.h>

int main(int argc,char **argv)
{
  MESH *m1,*m2,*m;
  int mergedecomp=FALSE;
  int	i,s;
  int ne[E_NELEM_TYPES];
  
  if (argc<4)
  {
    printf("\nMerges two mdf files. \n");
    printf("Only merges geometry and domain decomposition,\n");
    printf("but no loads, bc, etc.\n");
    printf("\nUsage: mdfmerge mesh1 mesh2 out [d]\n");
    printf("   mesh1: first mesh definition file\n");
    printf("   mesh2: second mesh definition file\n");
    printf("   out:   output mesh definition file\n");
    printf("   d :    when not zero it also merges decomposition data\n\n");
    exit(1);
  }
  
  if (argc==5)
    if (strcmp(argv[4],"d")==0)
      mergedecomp=TRUE;
  
  e_GetMeshData(&m1,argv[1]);
  e_GetMeshData(&m2,argv[2]);
  if (mergedecomp)
  {
    e_GetDecompData(m1,argv[1]);
    e_GetDecompData(m2,argv[2]);
  }
  
  e_GeneralWarning("mdfmerge only merges element definitions and decomposition\n\
no materials, BCs, loads, stresses, etc. are merged\n");
  
  m=(MESH *)e_calloc(1,sizeof(MESH));
  e_NullifyMesh(m);
  m->Title=e_alloc_c_vect(strlen("merged mesh")+1);
  strcpy(m->Title,"merged mesh");
  m->KeywordSet[E_KW_NUM_TITLE]=TRUE;
  m->NMeshPoints=m1->NMeshPoints+m2->NMeshPoints;
  m->KeywordSet[E_KW_NUM_NMESHPOINTS]=TRUE;
  m->NNodes=m1->NNodes+m2->NNodes;
  m->KeywordSet[E_KW_NUM_NNODES]=TRUE;
  ne[E_ELEM_TYPE_LINK1]=m->NElemsLink1=m1->NElemsLink1+m2->NElemsLink1;
  ne[E_ELEM_TYPE_LINK2]=m->NElemsLink2=m1->NElemsLink2+m2->NElemsLink2;
  ne[E_ELEM_TYPE_LINK3]=m->NElemsLink3=m1->NElemsLink3+m2->NElemsLink3;
  ne[E_ELEM_TYPE_LINK4]=m->NElemsLink4=m1->NElemsLink4+m2->NElemsLink4;
  ne[E_ELEM_TYPE_LINK5]=m->NElemsLink5=m1->NElemsLink5+m2->NElemsLink5;
  ne[E_ELEM_TYPE_TRIANG1]=m->NElemsTriang1=m1->NElemsTriang1+m2->NElemsTriang1;
  ne[E_ELEM_TYPE_TRIANG2]=m->NElemsTriang2=m1->NElemsTriang2+m2->NElemsTriang2;
  ne[E_ELEM_TYPE_TRIANG3]=m->NElemsTriang3=m1->NElemsTriang3+m2->NElemsTriang3;
  ne[E_ELEM_TYPE_TRIANG4]=m->NElemsTriang4=m1->NElemsTriang4+m2->NElemsTriang4;
  ne[E_ELEM_TYPE_TRIANG5]=m->NElemsTriang5=m1->NElemsTriang5+m2->NElemsTriang5;
  ne[E_ELEM_TYPE_QUAD1]=m->NElemsQuad1=m1->NElemsQuad1+m2->NElemsQuad1;
  ne[E_ELEM_TYPE_QUAD3]=m->NElemsQuad3=m1->NElemsQuad3+m2->NElemsQuad3;
  ne[E_ELEM_TYPE_TETRAH1]=m->NElemsTetrah1=m1->NElemsTetrah1+m2->NElemsTetrah1;
  ne[E_ELEM_TYPE_BLOCK1]=m->NElemsBlock1=m1->NElemsBlock1+m2->NElemsBlock1;
  m->TotalNElems=m1->TotalNElems+m2->TotalNElems;
  
#define SET_FNI(Type,TYPE) \
  if (m->NElems##Type##==0) \
    m->FirstNetIndex[E_ELEM_TYPE_##TYPE##]=0; \
  else \
  {	\
    m->FirstNetIndex[E_ELEM_TYPE_##TYPE##]=1; \
    for (i=0;i<E_ELEM_TYPE_##TYPE##;i++) \
      m->FirstNetIndex[E_ELEM_TYPE_##TYPE##]+=ne[i]; \
  }
  
  SET_FNI(Link1,LINK1)
  SET_FNI(Link2,LINK2)
  SET_FNI(Link3,LINK3)
  SET_FNI(Link4,LINK4)
  SET_FNI(Link5,LINK5)
  SET_FNI(Triang1,TRIANG1)
  SET_FNI(Triang2,TRIANG2)
  SET_FNI(Triang3,TRIANG3)
  SET_FNI(Triang4,TRIANG4)
  SET_FNI(Triang5,TRIANG5)
  SET_FNI(Quad1,QUAD1)
  SET_FNI(Quad3,QUAD3)
  SET_FNI(Tetrah1,TETRAH1)

#undef SET_FNI  
      
  m->MeshPointCoords=e_alloc_cont_d_matr(m->NMeshPoints,3);
  for (i=0;i<m1->NMeshPoints;i++)
    for (s=0;s<3;s++)
      m->MeshPointCoords[i][s]=m1->MeshPointCoords[i][s];
  for (i=0;i<m2->NMeshPoints;i++)
    for (s=0;s<3;s++)
      m->MeshPointCoords[m1->NMeshPoints+i][s]=m2->MeshPointCoords[i][s];
  m->KeywordSet[E_KW_NUM_MESHPOINT_COORDS]=TRUE;

#define ADD_ELEM_TYPE(Type,TYPE,n) \
  if (m->NElems##Type##) \
  { \
    m->Nodes##Type##=e_alloc_cont_i_matr(m->NElems##Type##,n); \
    for (i=0;i<m1->NElems##Type##;i++) \
      for (s=0;s<n;s++) \
        m->Nodes##Type##[i][s]=m1->Nodes##Type##[i][s]; \
    for (i=0;i<m2->NElems##Type##;i++) \
      for (s=0;s<n;s++) \
        m->Nodes##Type##[i+m1->NElems##Type##][s]=m2->Nodes##Type##[i][s]+m1->NMeshPoints; \
    m->KeywordSet[E_KW_NUM_NELEMS_##TYPE##]=TRUE; \
    m->KeywordSet[E_KW_NUM_NODES_##TYPE##]=TRUE; \
  } 
   
  ADD_ELEM_TYPE(Link1,LINK1,2)
  ADD_ELEM_TYPE(Link2,LINK2,2)
  ADD_ELEM_TYPE(Link3,LINK3,2)
  ADD_ELEM_TYPE(Link4,LINK4,2)
  ADD_ELEM_TYPE(Link5,LINK5,2)
  ADD_ELEM_TYPE(Triang1,TRIANG1,3)
  ADD_ELEM_TYPE(Triang2,TRIANG2,3)
  ADD_ELEM_TYPE(Triang3,TRIANG3,3)
  ADD_ELEM_TYPE(Triang4,TRIANG4,3)
  ADD_ELEM_TYPE(Triang5,TRIANG5,3)
  ADD_ELEM_TYPE(Quad1,QUAD1,4)
  ADD_ELEM_TYPE(Quad3,QUAD3,4)
  ADD_ELEM_TYPE(Tetrah1,TETRAH1,4)
  ADD_ELEM_TYPE(Block1,BLOCK1,8)
  
#undef ADD_ELEM_TYPE
  
  m->GEIs=e_alloc_i_vect(m->TotalNElems);
  for (i=0;i<m->TotalNElems;i++)
    m->GEIs[i]=i+1;
  m->GNIs=e_alloc_i_vect(m->NNodes);
  for (i=0;i<m->NNodes;i++)
    m->GNIs[i]=i+1;
  
  if (mergedecomp)
  {
    m->Decomp.SubDoms=e_alloc_i_vect(m->TotalNElems);
    m->Decomp.NSubDoms=m1->Decomp.NSubDoms+m2->Decomp.NSubDoms;
    m->SubStructSet[E_SS_NUM_DECOMP]=TRUE;

#define ADD_ELEM_DECOMP(Type,TYPE) \
    if (m->NElems##Type##) \
    { \
      for (i=0;i<m1->NElems##Type##;i++) \
        m->Decomp.SubDoms[m->FirstNetIndex[E_ELEM_TYPE_##TYPE##]-1+i]= \
          m1->Decomp.SubDoms##Type##[i]; \
      for (i=0;i<m2->NElems##Type##;i++) \
        m->Decomp.SubDoms[m->FirstNetIndex[E_ELEM_TYPE_##TYPE##]-1+i+m1->NElems##Type##]= \
          m2->Decomp.SubDoms##Type##[i]+m1->Decomp.NSubDoms; \
      m->Decomp.SubDoms##Type##=&(m->Decomp.SubDoms[m->FirstNetIndex[E_ELEM_TYPE_##TYPE##]-1]); \
    }
    
    ADD_ELEM_DECOMP(Link1,LINK1)
    ADD_ELEM_DECOMP(Link2,LINK2)
    ADD_ELEM_DECOMP(Link3,LINK3)
    ADD_ELEM_DECOMP(Link4,LINK4)
    ADD_ELEM_DECOMP(Link5,LINK5)
    ADD_ELEM_DECOMP(Triang1,TRIANG1)
    ADD_ELEM_DECOMP(Triang2,TRIANG2)
    ADD_ELEM_DECOMP(Triang3,TRIANG3)
    ADD_ELEM_DECOMP(Triang4,TRIANG4)
    ADD_ELEM_DECOMP(Triang5,TRIANG5)
    ADD_ELEM_DECOMP(Quad1,QUAD1)
    ADD_ELEM_DECOMP(Quad3,QUAD3)
    ADD_ELEM_DECOMP(Tetrah1,TETRAH1)
    ADD_ELEM_DECOMP(Block1,BLOCK1)

#undef ADD_ELEM_DECOMP

  }
  
  e_PutMeshData(m,argv[3]);
  if (mergedecomp)
    e_PutDecompData(m,argv[3]);
  
  e_FreeMeshData(m1);
  e_FreeMeshData(m2);
  e_FreeMeshData(m);
  
  return (0);
}
