/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*               Renumber the nodes to reduce the bandwidth                   */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* The origin of the new numbering system is located at each node in turn
so that the number of different renumbering schemes generated is equal
to the numbers of nodes. Using the assumption of the first node should
be start from the smallest relevent node time saved */

#include <e_lib.h>
#include "renumber.h"

/*
/*
/*****************************************************************************/
/*
/**/
int main(int argc, char *argv[])
{
  MESH *mesh;
  int *Degree;
  int **AdjElems;
  int Min_degree, Original_bandwidth,Max_Bandwidth_Elem;
  int Renumber_done;
  int *jnt;
  char outfn[400];

  if(argc < 3)
  {
    printf("\nRenumbers the mesh nodes to reduce the matrix bandwidth.\n");
    printf("\nUsage: renumber input output\n");
    printf("   input  : name of the input mesh definition file\n");
    printf("   output : name of the resulting mesh definition file\n\n");
    printf("All files should be specified without extension!\n\n");
    exit(1);
  }

  
  e_GetMeshData(&mesh, argv[1]);
  strcpy(outfn, argv[1]);
  e_PutMeshData(mesh, outfn);

  if (!e_IsOfOneElementType(mesh))
      e_GeneralError("Error: There is more than one element type in this mesh"); 

  SetAdjMeshPoints(mesh,&Degree,&AdjElems);

  jnt = e_alloc_i_vect(mesh->NMeshPoints+1);
  B_get_orignalmesh_band(mesh,Degree,&Min_degree,&Original_bandwidth,
                          &Max_Bandwidth_Elem);

  Renumber_done = FALSE;

  Renumber_done =
      B_optnum_Nodes(mesh,Degree,AdjElems,Min_degree,
                     jnt, Original_bandwidth);

  if(Renumber_done)
          B_renum_Nodes(mesh, jnt);

  FreeAdjMeshPoints(mesh,Degree,AdjElems);
  e_free(jnt);
  
  strcpy(outfn, argv[2]);
  e_PutMeshData(mesh,outfn);

  return 0;
}


/*
/*
/*****************************************************************************/
/*
/**/
/*  Funtion to find the ooriginal bandwidth and the minume number
of the node relevent */
static void  B_get_orignalmesh_band(MESH *mesh,int *Degree,
                                    int *Min_degree,
			            int *Original_bandwidth,
                                    int *Max_Bandwidth_Elem)
{
  int i,j;
  int Smallest_Node_Elem, Greates_Node_Elem;
  int ElemType,Temp_E_NNPE_ELEMs, Temp_NElems;
  int **nodes;
  
  *Min_degree=100;
  *Original_bandwidth=0;
  *Max_Bandwidth_Elem = 0;
  
  Temp_E_NNPE_ELEMs = e_GetNNodesPerElement(mesh);
  nodes = e_GetNodes(mesh);
  
  for(i=0; i<mesh->NMeshPoints; i++)
      if(*Min_degree > Degree[i])
          *Min_degree = Degree[i];

  for(i=0; i<mesh->TotalNElems;i++)
  {
    Smallest_Node_Elem = mesh->NMeshPoints;
    Greates_Node_Elem = 0;
    
    for(j=0; j < Temp_E_NNPE_ELEMs;j++)
    {
      if(Smallest_Node_Elem > nodes[i][j])
          Smallest_Node_Elem = nodes[i][j];
      
      if(Greates_Node_Elem < nodes[i][j])
          Greates_Node_Elem = nodes[i][j];       
    }
    if( *Original_bandwidth<(Greates_Node_Elem - Smallest_Node_Elem))
    {
      *Original_bandwidth = Greates_Node_Elem - Smallest_Node_Elem;
      *Max_Bandwidth_Elem = i;
    }
  }
}


/*****************************************************************************/
/* main function to perform the new number process. As the new
numbering scheme progresses, two arrays are set up which are
used largely as pointers to eliminate the need to search
through *pe; newjt contains the old node numbers, their locations
in the array being equal to the new joint numbers. For
each new numbering scheme, they are initialized to zero */
static int B_optnum_Nodes(MESH *mesh,int *Degree,int **AdjElems,int Min_degree,
                          int *jnt, int Original_bandwidth)
{
  int *joint, *newjt;
  int i,j,jj,ik,k;
  int Current_TestNode;
  int max,ik4,ndiff,Current_Iteration_minmax,New_search_Node, NInteration=1;
  int success_renumber=0;

  joint = e_alloc_i_vect(mesh->NMeshPoints+1);
  newjt = e_alloc_i_vect(mesh->NMeshPoints+1);

  
  Current_Iteration_minmax = Original_bandwidth;
  
  /* start from 1 rather than using the array from 0. */
  for(ik=1;ik <= mesh->NMeshPoints; ik++)
  {
    
    if(Degree[ik-1]>Min_degree) continue;
    for(j = 0; j <= mesh->NMeshPoints; j++)  joint[j]= newjt[j]= 0;
    max=0;
    New_search_Node = 1;
    newjt[1]=ik;
    joint[ik]=1;
    k=1;

    while(k != mesh->NMeshPoints)
    {
      if(New_search_Node == mesh->NMeshPoints ||
         Degree[newjt[New_search_Node]-1] == 0    )
          e_GeneralError("Error: Error in Nodes renumber function");
      
      for(jj=0; jj<Degree[newjt[New_search_Node]-1]; jj++)
      {
        Current_TestNode = AdjElems[newjt[New_search_Node]-1][jj];
        if(joint[Current_TestNode] == 0)
        {
          k++;
          newjt[k] = Current_TestNode;
          joint[Current_TestNode] = k;
          ndiff=abs(New_search_Node-k);
          if(ndiff >= Current_Iteration_minmax) goto d;
          
          if(ndiff>max) max = ndiff;
        }
      }
      New_search_Node++;
    }
    
    Current_Iteration_minmax = max;
    success_renumber = 1;
    for(j=1;j<=mesh->NMeshPoints;j++)  jnt[j]=joint[j];
    d:   NInteration++;
  }
  
  printf("\nOriginal bandwidth =%d\nNew, reduced bandwidth is %d\n\n",
         Original_bandwidth,Current_Iteration_minmax);
  
  e_free(newjt); e_free(joint);
  return  success_renumber;
}


/*
/*
/*****************************************************************************/
/*
/**/
/* Funtion to renumer the original nodes and element nodes according
to the new numbering scheme */
static void B_renum_Nodes(MESH *mesh,int *jnt)
{
  int i,j;
  int ElemType;
  double **MeshPointCoords;
  int *LoadedNodes;
  int *BCNodes;
  int *NodeTypes;                   /* remeshing node types */
  int *NNurbsCurves;                /* number of NURBS curves per node */
  double **NurbsCurveParams;        /* NURBS curve parameters */
  NURBSCURVE ***NurbsCurves;        /* node NURBS curves */
  
  
  ElemType = e_GetElementType(mesh);
  
  
  if(ElemType ==E_ELEM_TYPE_TRIANG1)
  {
    for(i = 0; i< mesh->NElemsTriang1; i++)
        for(j=0; j< E_NNPE_TRIANG1; j++)
            mesh->NodesTriang1[i][j] = jnt[mesh->NodesTriang1[i][j]];
  }
  else if(ElemType ==E_ELEM_TYPE_QUAD1)
  {
    for(i = 0; i< mesh->NElemsQuad1; i++)
        for(j=0; j< E_NNPE_QUAD1; j++)
            mesh->NodesQuad1[i][j] = jnt[mesh->NodesQuad1[i][j]];
  }
  else if(ElemType ==E_ELEM_TYPE_TETRAH1)
  {
    for(i = 0; i< mesh->NElemsTetrah1; i++)
        for(j=0; j< E_NNPE_TETRAH1; j++)
            mesh->NodesTetrah1[i][j] = jnt[mesh->NodesTetrah1[i][j]];
  }
  else e_GeneralError("Error: Unsupported element type in function renumber");
  
  
  /* Nodes' Coordinates reallocation */
  MeshPointCoords = e_alloc_cont_d_matr(mesh->NMeshPoints, E_NDIMS);
  
  for(i=0; i< mesh->NMeshPoints; i++)
      for(j=0;j<E_NDIMS-1;j++)
          MeshPointCoords[jnt[i+1]-1][j]= mesh->MeshPointCoords[i][j];
  
  e_free_cont_d_matr(mesh->NMeshPoints,mesh->MeshPointCoords);
  mesh->MeshPointCoords = MeshPointCoords;
  
  if (mesh->KeywordSet[E_KW_NUM_REMESH_DATA])
  {
    
    NodeTypes = e_alloc_i_vect(mesh->NMeshPoints);
    NNurbsCurves = e_alloc_i_vect(mesh->NMeshPoints);
    NurbsCurveParams = e_init_frag_d_matr(mesh->NMeshPoints);
    NurbsCurves = (NURBSCURVE ***)e_calloc(mesh->NMeshPoints,
                   sizeof(NURBSCURVE **));

    for(i=0; i< mesh->NMeshPoints; i++)
    {
      NodeTypes[jnt[i+1]-1] = mesh->NodeTypes[i];
      NNurbsCurves[jnt[i+1]-1] = mesh->NNurbsCurves[i];
      NurbsCurveParams[jnt[i+1]-1] = mesh->NurbsCurveParams[i];
      NurbsCurves[jnt[i+1]-1] = mesh->NurbsCurves[i];
    }

    e_free(mesh->NodeTypes);
    e_free(mesh->NNurbsCurves);
    e_free(mesh->NurbsCurveParams);
    e_free(mesh->NurbsCurves);

    mesh->NodeTypes = NodeTypes;
    mesh->NNurbsCurves = NNurbsCurves;
    mesh->NurbsCurveParams = NurbsCurveParams;
    mesh->NurbsCurves = NurbsCurves;
  }

  if (mesh->KeywordSet[E_KW_NUM_NLOADED_NODES])
  {
    LoadedNodes=e_alloc_i_vect(mesh->NLoadedNodes);
    for(i=0;i<mesh->NLoadedNodes;i++)
	LoadedNodes[i]=jnt[mesh->LoadedNodes[i]];
    e_free(mesh->LoadedNodes);
    mesh->LoadedNodes=LoadedNodes;
  }

  if (mesh->KeywordSet[E_KW_NUM_NBC_NODES])
  {
    BCNodes=e_alloc_i_vect(mesh->NBCNodes);
    for(i=0;i<mesh->NBCNodes;i++)
	BCNodes[i]=jnt[mesh->BCNodes[i]];
    e_free(mesh->BCNodes);
    mesh->BCNodes=BCNodes;
  }

  if (mesh->KeywordSet[E_KW_NUM_NMATS])
      e_GeneralWarning("there is no renumbering of materials so far!\n");
    
}

/* ////////////////////////////////////////////////////////////////////////// */
/* SetAdjMeshPoints: To determine which mesh-points are adjacent to which
   (i.e. share an edge).                                                      */
static int SetAdjMeshPoints(MESH *mesh, int **NAdjMeshPoints,
                                         int ***AdjMeshPoints)
{
  int i,j,l,k1,k2;
  int ElemType;
  int NVpE;
  int **Vertices;

  ElemType = e_GetElementType(mesh);
  
  *NAdjMeshPoints = e_alloc_i_vect(mesh->NMeshPoints);
  *AdjMeshPoints = e_init_frag_i_matr(mesh->NMeshPoints);
  
  NVpE = e_GetNVerticesPerElement(mesh);
  Vertices = e_GetVertices(mesh);
  
  if (   ElemType == E_ELEM_TYPE_TRIANG1
      || ElemType == E_ELEM_TYPE_TRIANG2
      || ElemType == E_ELEM_TYPE_QUAD1  )
  {
    for (i = 0; i < mesh->TotalNElems; i++)
    {
      for (j = 0; j < NVpE; j++)
      {
        k1 = Vertices[i][j] - 1;
        k2 = Vertices[i][(j + 1) % NVpE] - 1;
        
        UpdateMeshPoints(k1,k2,*NAdjMeshPoints,*AdjMeshPoints);
      }
    }
  }
  else
  {
    for (i = 0; i < mesh->TotalNElems; i++)
    {
      for (l = 0; l < E_NIPE_TETRAH1; l++)
      {
        for (j = 0; j < E_NVPI_TETRAH1 - 1; j++)
        {
          k1 = mesh->NodesTetrah1[i][g_InterfNodesTetrah[l][j] - 1] - 1;
          k2 = mesh->NodesTetrah1[i][g_InterfNodesTetrah[l][j + 1] - 1] - 1;
          
          UpdateMeshPoints(k1,k2,*NAdjMeshPoints,*AdjMeshPoints);
        }
        
        k1 = mesh->NodesTetrah1[i][g_InterfNodesTetrah[l][E_NVPI_TETRAH1 - 1] - 1] - 1;
        k2 = mesh->NodesTetrah1[i][g_InterfNodesTetrah[l][0] - 1] - 1;
        
        UpdateMeshPoints(k1,k2,*NAdjMeshPoints,*AdjMeshPoints);
      }
    }
  }
  
  return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* UpdateMeshPoints: To update the adjacent mesh-point data. */
static int UpdateMeshPoints(int k1, int k2, int *NAdjMeshPoints,
                            int **AdjMeshPoints)
{
  if (IsANewAdjMeshPoint(NAdjMeshPoints[k1],AdjMeshPoints[k1],k2 + 1))
  {
    AdjMeshPoints[k1]
        = e_incr_i_vect(NAdjMeshPoints[k1],AdjMeshPoints[k1],
                        ADJ_MESHPOINTS_MEM_INCREM);
    AdjMeshPoints[k1][NAdjMeshPoints[k1]++] = k2 + 1;
  }
  
  if (IsANewAdjMeshPoint(NAdjMeshPoints[k2],AdjMeshPoints[k2],k1 + 1))
  {
    AdjMeshPoints[k2]
        = e_incr_i_vect(NAdjMeshPoints[k2],AdjMeshPoints[k2],
                        ADJ_MESHPOINTS_MEM_INCREM);
    AdjMeshPoints[k2][NAdjMeshPoints[k2]++] = k1 + 1;
  }
  
  return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* IsANewAdjMeshPoint: To return whether a mesh-point has a new adjacent
   mesh-point.                                                                */
static int IsANewAdjMeshPoint(int NAdjMeshPoints, int *AdjMeshPoints,
                                                        int MeshPoint)
{
  int j;

  for (j = 0; j < NAdjMeshPoints; j++)
  {
    if (AdjMeshPoints[j] == MeshPoint)
        return 0;
  }
  
  return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* FreeAdjMeshPoints: To free memory allocated for the mesh-point adjacency
   data.                                                                      */
static void FreeAdjMeshPoints(MESH *mesh, int *NAdjMeshPoints,
                                           int **AdjMeshPoints)
{
  e_free(NAdjMeshPoints);
  e_free_frag_i_matr(mesh->NMeshPoints,AdjMeshPoints);
}


