#include "e_lib.h"

#define X_AXIS 1
#define Y_AXIS 2


int main(int argc, char **argv)
{
  MESH *m;
  int    n, open_rev, axis;
  int    i, j, s, in;
  double a;
  double rota;
  
  /* how does it work ? */
  if(argc < 3)
  {
    printf("\nRevolves quad elements around an axis to create blocks.\n");
    printf("The mesh must be in the x-y pane.\n");
    printf("\nUsage: rev mesh n [-o angle] [-x -y]\n");
    printf("   -o angle: Open revolution by angle [degree]\n");
    printf("   -x : Revolution around x axis\n");
    printf("   -y : Revolution around y axis\n\n");
    printf("Defaults are rotation around x axis and rotation by 360 degrees\n");
    printf("to form a closed torus\n\n");
    exit(1);
  }
  
  /* input mesh */
  e_GetMeshData(&m,argv[1]);
  if(!e_IsOfOneElementType(m))
      e_GeneralError("Mesh may only contain QUAD1 elements");
  if(e_GetElementType(m)!=E_ELEM_TYPE_QUAD1)
      e_GeneralError("Mesh may only contain QUAD1 elements");
  
  /* number of divisions */
  if(sscanf(argv[2],"%d",&n)!=1)
      e_GeneralError("Can't read number of angular divisions");

  open_rev = FALSE;
  axis = X_AXIS;
  rota = 2 * M_PI;
  
  for(i = 3; i < argc; i++)
  {
    if(strcmp(argv[i], "-o") == 0)
    {
      open_rev = TRUE;
      i++;
      if(i < argc)
      {
        if(sscanf(argv[i], "%lf", &rota) != 1)
        {
          e_GeneralError("Cannot read rotation angle!");
        }
        if(rota < 0 || rota > 360)
        {
          e_GeneralError("Invalid rotation angle!");
        }
        /* convert to radian */
        rota = rota *  M_PI / 180.0;
      }
      else
      {
        e_GeneralError("Rotation angle is not specified!");
      }
    }
    else if(strcmp(argv[i], "-x") == 0)
    {
      axis = X_AXIS;
    }
    else if(strcmp(argv[i], "-y") == 0)
    {
      axis = Y_AXIS;
    }
    else
    {
      e_GeneralError("Invalid command line option");
    }
  }
  
  if(open_rev)
  {
    if(n<2)
    {
      e_GeneralError("Number of angular divisions must be >1");
    }
    in=n+1;
  }
  else
  {
    if(n<4)
    {
      e_GeneralError("Number of angular divisions must be >3");
    }
    in=n;
  }
  
  /* allocation */
  m->MeshPointCoords=e_ext_cont_d_matr(m->NMeshPoints,3,
                                       m->MeshPointCoords,
                                       (in-1)*m->NMeshPoints);
  m->NMeshPoints *= in;
  m->NNodes       = m->NMeshPoints;
  m->NElemsBlock1 = n*m->NElemsQuad1;
  m->NodesBlock1  = e_alloc_cont_i_matr(m->NElemsBlock1,8);
  m->KeywordSet[E_KW_NUM_NELEMS_QUAD1]  = FALSE;
  m->KeywordSet[E_KW_NUM_NODES_QUAD1]   = FALSE;
  m->KeywordSet[E_KW_NUM_NELEMS_BLOCK1] = TRUE;
  m->KeywordSet[E_KW_NUM_NODES_BLOCK1]  = TRUE;
  
  /* generate extra nodes */
  for(i = 1; i < in; i++)
  {
    a = rota/(double)(n)*(double)i;
    
    for(j = 0; j < m->NMeshPoints/in; j++)
    {
      if(axis == Y_AXIS)
      {
        m->MeshPointCoords[i*m->NMeshPoints/in+j][0]=
            m->MeshPointCoords[j][0]*cos(a);
        m->MeshPointCoords[i*m->NMeshPoints/in+j][1]=
            m->MeshPointCoords[j][1];
        m->MeshPointCoords[i*m->NMeshPoints/in+j][2]=
            m->MeshPointCoords[j][0]*sin(a);
      }
      else if(axis == X_AXIS)
      {
        m->MeshPointCoords[i*m->NMeshPoints/in+j][0]=
            m->MeshPointCoords[j][0];
        m->MeshPointCoords[i*m->NMeshPoints/in+j][1]=
            m->MeshPointCoords[j][1]*cos(a);
        m->MeshPointCoords[i*m->NMeshPoints/in+j][2]=
            m->MeshPointCoords[j][1]*sin(a);
      }
      else
      {
        e_GeneralError("Invalid rotation axis!");
      }
    }
  }
  
  /* generate blocks */
  for(i = 0; i < n; i++)
  {
    for(j = 0; j < m->NElemsQuad1; j++)
    {
      for(s = 0; s < 4; s++)
      {
        m->NodesBlock1[i*m->NElemsQuad1+j][s]=
            i*m->NMeshPoints/in+m->NodesQuad1[j][s];
      }
      
      for(s = 0; s < 4; s++)
      {
        if(open_rev)
        {
          m->NodesBlock1[i*m->NElemsQuad1+j][4+s]=
              ((i+1))*m->NMeshPoints/in+m->NodesQuad1[j][s];
        }
        else
        {
          m->NodesBlock1[i*m->NElemsQuad1+j][4+s]=
              ((i+1)%n)*m->NMeshPoints/in+m->NodesQuad1[j][s];
        }
      }
    }
  }
  
  /* output */
  e_PutMeshData(m, "rev");
  e_FreeMeshData(m);
  
  return (0);
}

