#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <e_lib.h>

int main(int argc, char *argv[])
{
  int     gtr=FALSE;
  MESH    *mesh;
  int     nd;
  int     x,y,z,e,i;
  int     tn, tq, tnn, tqq;
  double  l,l1,l2;
  
  if (argc<3)
  {
    printf("\n");
    printf("\nUsage: circle n [-q|-t]");
    printf("\n   n = subdivs per side");
    printf("\n  -q = generate QUAD1");
    printf("\n  -t = generate TRIANG1\n\n");
    exit(1);
  }
  if (sscanf(argv[1], "%d", &nd)!=1)
  {
    e_GeneralError("Invalid number of divisions");
  }
  if ((strcmp(argv[2], "-q")!=0)&&(strcmp(argv[2], "-t")!=0))
  {
    e_GeneralError("No element type definition");
  }
  
  if (strcmp(argv[2], "-t")==0)
      gtr=TRUE;
  
  if (nd<2)
      e_GeneralError("n must be > 2");
  
  mesh=NULL;
  mesh=(MESH *)e_calloc(1, sizeof(MESH));
  e_NullifyMesh(mesh);
  
  mesh->Title=e_alloc_c_vect(strlen("circle")+1);
  strcpy(mesh->Title, "circle");
  mesh->NMeshPoints=(nd+1)*(nd+1);
  mesh->NNodes=mesh->NMeshPoints;
  mesh->MeshPointCoords=e_alloc_cont_d_matr(mesh->NNodes, 3);
  mesh->NElemsQuad1=nd*nd;
  mesh->NodesQuad1=e_alloc_cont_i_matr(mesh->NElemsQuad1, 4);
  mesh->KeywordSet[E_KW_NUM_TITLE]=TRUE;
  mesh->KeywordSet[E_KW_NUM_NNODES]=TRUE;
  mesh->KeywordSet[E_KW_NUM_NMESHPOINTS]=TRUE;
  mesh->KeywordSet[E_KW_NUM_NODES_QUAD1]=TRUE;
  mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1]=TRUE;
  mesh->KeywordSet[E_KW_NUM_MESHPOINT_COORDS]=TRUE;
  
  for (y=0;y<nd+1;y++)
      for (x=0;x<nd+1;x++)
      {
        /* front nodes */
        mesh->MeshPointCoords[y*(nd+1)+x][0]=-1.0+(2.0/nd)*x;
        mesh->MeshPointCoords[y*(nd+1)+x][1]=-1.0+(2.0/nd)*y;
        mesh->MeshPointCoords[y*(nd+1)+x][2]=1.0;
      }
  for (y=0;y<nd;y++)
      for (x=0;x<nd;x++)
      {
        /* front quads */
        mesh->NodesQuad1[y*nd+x][0]=y*(nd+1)+x;
        mesh->NodesQuad1[y*nd+x][1]=y*(nd+1)+x+1;
        mesh->NodesQuad1[y*nd+x][2]=(y+1)*(nd+1)+x+1;
        mesh->NodesQuad1[y*nd+x][3]=(y+1)*(nd+1)+x;
      }
  for (e=0;e<mesh->NElemsQuad1;e++)
      for (i=0;i<4;i++)
          mesh->NodesQuad1[e][i]++;

  /* spherize */
  for (e=0;e<mesh->NNodes;e++)
  {
    l=0.0;
    for (i=0;i<3;i++)
        l += mesh->MeshPointCoords[e][i] * mesh->MeshPointCoords[e][i];
    l=sqrt(l);
    for (i=0;i<3;i++)
        if (l>1e-8)
            mesh->MeshPointCoords[e][i]/=l;
        else
            mesh->MeshPointCoords[e][i]=0.0;
    /*mesh->MeshPointCoords[e][2]=0.0; */
  }
  
  /* convert to triangs */
  if (gtr)
  {
    mesh->NElemsTriang1=mesh->NElemsQuad1*2;
    mesh->NodesTriang1=e_alloc_cont_i_matr(mesh->NElemsTriang1, 3);
    mesh->KeywordSet[E_KW_NUM_NODES_TRIANG1]=TRUE;
    mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1]=TRUE;
    for (e=0;e<mesh->NElemsQuad1;e++)
    {
      l1=0.0;
      for (i=0;i<3;i++)
          l1+=pow(mesh->MeshPointCoords[mesh->NodesQuad1[e][0]-1][i]-
                  mesh->MeshPointCoords[mesh->NodesQuad1[e][2]-1][i],2);
      l2=0.0;
      for (i=0;i<3;i++)
          l2+=pow(mesh->MeshPointCoords[mesh->NodesQuad1[e][1]-1][i]-
                  mesh->MeshPointCoords[mesh->NodesQuad1[e][3]-1][i],2);
      if (l1<l2)
      {
        mesh->NodesTriang1[2*e][0]=mesh->NodesQuad1[e][0];
        mesh->NodesTriang1[2*e][1]=mesh->NodesQuad1[e][1];
        mesh->NodesTriang1[2*e][2]=mesh->NodesQuad1[e][2];
        mesh->NodesTriang1[2*e+1][0]=mesh->NodesQuad1[e][0];
        mesh->NodesTriang1[2*e+1][1]=mesh->NodesQuad1[e][2];
        mesh->NodesTriang1[2*e+1][2]=mesh->NodesQuad1[e][3];
      }
      else
      {
        mesh->NodesTriang1[2*e][0]=mesh->NodesQuad1[e][0];
        mesh->NodesTriang1[2*e][1]=mesh->NodesQuad1[e][1];
        mesh->NodesTriang1[2*e][2]=mesh->NodesQuad1[e][3];
        mesh->NodesTriang1[2*e+1][0]=mesh->NodesQuad1[e][1];
        mesh->NodesTriang1[2*e+1][1]=mesh->NodesQuad1[e][2];
        mesh->NodesTriang1[2*e+1][2]=mesh->NodesQuad1[e][3];
      }
    }
    mesh->KeywordSet[E_KW_NUM_NODES_QUAD1]=FALSE;
    mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1]=FALSE;
    e_free_cont_i_matr(mesh->NElemsQuad1, mesh->NodesQuad1);
  }
  
  if (gtr)
      e_PutMeshData(mesh, "triangs");
  else
      e_PutMeshData(mesh, "quads");
  e_free(mesh);
  
  return 0;
}

