#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <e_lib.h>

int main(int argc, char *argv[])
{
  int     gtr=FALSE;
  MESH    *mesh;
  int     nd;
  int     x,y,z,e,i;
  int     tn, tq, tnn, tqq;
  double  l,l1,l2;
  
  if (argc<3)
  {
    printf("\nGenerate a surface mesh of a sphere.\n");
    printf("\nUsage: sphere n [-q|-t]\n");
    printf("  n = number of subdivisions per side\n");
    printf(" -q = generate QUAD1 elements\n");
    printf(" -t = generate TRIANG1 elements\n\n");
    exit(1);
  }
  if (sscanf(argv[1], "%d", &nd)!=1)
  {
    e_GeneralError("Could not read number of divisions!");
  }
  if ((strcmp(argv[2], "-q")!=0)&&(strcmp(argv[2], "-t")!=0))
  {
    e_GeneralError("No element type definition!");
  }
  
  if (strcmp(argv[2], "-t")==0)
      gtr=TRUE;
    
  if (nd<2)
      e_GeneralError("n must be > 2");
    
  mesh=NULL;
  mesh=(MESH *)e_calloc(1, sizeof(MESH));
  e_NullifyMesh(mesh);
  
  mesh->Title=e_alloc_c_vect(strlen("sphere")+1);
  strcpy(mesh->Title, "sphere");
  mesh->NMeshPoints=(nd+1)*(nd+1)*2+
      (nd+1)*(nd-1)*2+
      (nd-1)*(nd-1)*2;
  mesh->NNodes=mesh->NMeshPoints;
  mesh->MeshPointCoords=e_alloc_cont_d_matr(mesh->NNodes, 3);
  mesh->NElemsQuad1=nd*nd*6;
  mesh->NodesQuad1=e_alloc_cont_i_matr(mesh->NElemsQuad1, 4);
  mesh->KeywordSet[E_KW_NUM_TITLE]=TRUE;
  mesh->KeywordSet[E_KW_NUM_NNODES]=TRUE;
  mesh->KeywordSet[E_KW_NUM_NMESHPOINTS]=TRUE;
  mesh->KeywordSet[E_KW_NUM_NODES_QUAD1]=TRUE;
  mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1]=TRUE;
  mesh->KeywordSet[E_KW_NUM_MESHPOINT_COORDS]=TRUE;
  
  for (y=0;y<nd+1;y++)
      for (x=0;x<nd+1;x++)
      {
        /* front nodes */
        mesh->MeshPointCoords[y*(nd+1)+x][0]=-1.0+(2.0/nd)*x;
        mesh->MeshPointCoords[y*(nd+1)+x][1]=-1.0+(2.0/nd)*y;
        mesh->MeshPointCoords[y*(nd+1)+x][2]=+1.0;
        /* back nodes */
        mesh->MeshPointCoords[(nd+1+y)*(nd+1)+x][0]=-1.0+(2.0/nd)*x;
        mesh->MeshPointCoords[(nd+1+y)*(nd+1)+x][1]=-1.0+(2.0/nd)*y;
        mesh->MeshPointCoords[(nd+1+y)*(nd+1)+x][2]=-1.0;
      }
  for (y=0;y<nd;y++)
      for (x=0;x<nd;x++)
      {
        /* front quads */
        mesh->NodesQuad1[y*nd+x][0]=y*(nd+1)+x;
        mesh->NodesQuad1[y*nd+x][1]=y*(nd+1)+x+1;
        mesh->NodesQuad1[y*nd+x][2]=(y+1)*(nd+1)+x+1;
        mesh->NodesQuad1[y*nd+x][3]=(y+1)*(nd+1)+x;
        /* back quads */
        mesh->NodesQuad1[(y+nd)*nd+x][0]=(y+nd+1)*(nd+1)+x;
        mesh->NodesQuad1[(y+nd)*nd+x][1]=(y+nd+1+1)*(nd+1)+x;
        mesh->NodesQuad1[(y+nd)*nd+x][2]=(y+nd+1+1)*(nd+1)+x+1;
        mesh->NodesQuad1[(y+nd)*nd+x][3]=(y+nd+1)*(nd+1)+x+1;
      }
  tn=2*(nd+1)*(nd+1);
  tq=2*nd*nd;
  for (y=0;y<nd+1;y++)
      for (z=0;z<nd-1;z++)
      {
        /* right nodes */
        mesh->MeshPointCoords[tn+y*(nd-1)+z][0]=+1.0;
        mesh->MeshPointCoords[tn+y*(nd-1)+z][1]=-1.0+(2.0/nd)*y;
        mesh->MeshPointCoords[tn+y*(nd-1)+z][2]=-1.0+(2.0/nd)*(z+1);
        /* left nodes */
        mesh->MeshPointCoords[tn+(y+nd+1)*(nd-1)+z][0]=-1.0;
        mesh->MeshPointCoords[tn+(y+nd+1)*(nd-1)+z][1]=-1.0+(2.0/nd)*y;
        mesh->MeshPointCoords[tn+(y+nd+1)*(nd-1)+z][2]=-1.0+(2.0/nd)*(z+1);
      }
  for (y=0;y<nd;y++)
      for (z=0;z<nd;z++)
      {
        /* right quads */
        if (z==0)
        {
          mesh->NodesQuad1[tq+y*nd+z][0]=(y+nd+1)*(nd+1)+nd;   
          mesh->NodesQuad1[tq+y*nd+z][1]=(y+nd+1+1)*(nd+1)+nd;
        }
        else
        {
          mesh->NodesQuad1[tq+y*nd+z][0]=tn+y*(nd-1)+z-1;  
          mesh->NodesQuad1[tq+y*nd+z][1]=tn+(y+1)*(nd-1)+z-1;
        }
        if (z==nd-1)
        {
          mesh->NodesQuad1[tq+y*nd+z][2]=(y+1)*(nd+1)+nd;
          mesh->NodesQuad1[tq+y*nd+z][3]=y*(nd+1)+nd;
        }
        else
        {      
          mesh->NodesQuad1[tq+y*nd+z][2]=tn+(y+1)*(nd-1)+z;
          mesh->NodesQuad1[tq+y*nd+z][3]=tn+y*(nd-1)+z;
        }
        /* left quads */
        if (z==0)
        {
          mesh->NodesQuad1[tq+(y+nd)*nd+z][2]=(y+nd+1+1)*(nd+1);
          mesh->NodesQuad1[tq+(y+nd)*nd+z][3]=(y+nd+1)*(nd+1);
        }
        else
        {
          mesh->NodesQuad1[tq+(y+nd)*nd+z][2]=tn+(y+nd+1+1)*(nd-1)+z-1;
          mesh->NodesQuad1[tq+(y+nd)*nd+z][3]=tn+(y+nd+1)*(nd-1)+z-1;
        }
        if (z==nd-1)
        {
          mesh->NodesQuad1[tq+(y+nd)*nd+z][1]=(y+1)*(nd+1);
          mesh->NodesQuad1[tq+(y+nd)*nd+z][0]=y*(nd+1);
        }
        else
        {
          mesh->NodesQuad1[tq+(y+nd)*nd+z][1]=tn+(y+nd+1+1)*(nd-1)+z;
          mesh->NodesQuad1[tq+(y+nd)*nd+z][0]=tn+(y+nd+1)*(nd-1)+z;
        }
      }
  tnn=(nd+1)*(nd+1)*2+(nd+1)*(nd-1)*2;
  tqq=4*nd*nd;
  for (x=0;x<nd-1;x++)
      for (z=0;z<nd-1;z++)
      {
        /* top nodes */
        mesh->MeshPointCoords[tnn+x*(nd-1)+z][0]=-1.0+(2.0/nd)*(x+1);
        mesh->MeshPointCoords[tnn+x*(nd-1)+z][1]=+1.0;
        mesh->MeshPointCoords[tnn+x*(nd-1)+z][2]=-1.0+(2.0/nd)*(z+1);
        /* bottom nodes */
        mesh->MeshPointCoords[tnn+(x+nd-1)*(nd-1)+z][0]=-1.0+(2.0/nd)*(x+1);
        mesh->MeshPointCoords[tnn+(x+nd-1)*(nd-1)+z][1]=-1.0;
        mesh->MeshPointCoords[tnn+(x+nd-1)*(nd-1)+z][2]=-1.0+(2.0/nd)*(z+1);
      }
  for (x=0;x<nd;x++)
      for (z=0;z<nd;z++)
      {
        /* top quads */
        if (x==0)
        {
          if (z==0)
              mesh->NodesQuad1[tqq+x*nd+z][3]=2*(nd+1)*(nd+1)-(nd+1);
          else
              mesh->NodesQuad1[tqq+x*nd+z][3]=tn+2*(nd+1)*(nd-1)-(nd-1)+(z-1);
          if (z==nd-1)
              mesh->NodesQuad1[tqq+x*nd+z][0]=nd*(nd+1);
          else
              mesh->NodesQuad1[tqq+x*nd+z][0]=tn+2*(nd+1)*(nd-1)-(nd-1)+z;
        }
        else
        {
          if (z==0)
              mesh->NodesQuad1[tqq+x*nd+z][3]=2*(nd+1)*(nd+1)-(nd+1)+x;
          else
              mesh->NodesQuad1[tqq+x*nd+z][3]=tnn+(x-1)*(nd-1)+z-1;
          if (z==nd-1)
              mesh->NodesQuad1[tqq+x*nd+z][0]=nd*(nd+1)+x;
          else
              mesh->NodesQuad1[tqq+x*nd+z][0]=tnn+(x-1)*(nd-1)+z;
        }
        if (x==nd-1)    	    	    
        {
          if (z==0)
              mesh->NodesQuad1[tqq+x*nd+z][2]=2*(nd+1)*(nd+1)-1;
          else
              mesh->NodesQuad1[tqq+x*nd+z][2]=tn+(nd+1)*(nd-1)-(nd-1)+(z-1);
          if (z==nd-1)
              mesh->NodesQuad1[tqq+x*nd+z][1]=(nd+1)*(nd+1)-1;
          else
              mesh->NodesQuad1[tqq+x*nd+z][1]=tn+(nd+1)*(nd-1)-(nd-1)+z;
        }
        else
        {	
          if (z==0)
              mesh->NodesQuad1[tqq+x*nd+z][2]=2*(nd+1)*(nd+1)-(nd+1)+x+1;
          else
              mesh->NodesQuad1[tqq+x*nd+z][2]=tnn+x*(nd-1)+z-1;
          if (z==nd-1)
              mesh->NodesQuad1[tqq+x*nd+z][1]=nd*(nd+1)+x+1;
          else
              mesh->NodesQuad1[tqq+x*nd+z][1]=tnn+x*(nd-1)+z;
        }
        /* bottom quads */
        if (x==0)
        {
          if (z==0)
              mesh->NodesQuad1[tqq+(x+nd)*nd+z][2]=(nd+1)*(nd+1);
          else
              mesh->NodesQuad1[tqq+(x+nd)*nd+z][2]=tn+(nd+1)*(nd-1)+(z-1);
          if (z==nd-1)
              mesh->NodesQuad1[tqq+(x+nd)*nd+z][1]=0;
          else
              mesh->NodesQuad1[tqq+(x+nd)*nd+z][1]=tn+(nd+1)*(nd-1)+z;
        }
        else
        {
          if (z==0)
              mesh->NodesQuad1[tqq+(x+nd)*nd+z][2]=(nd+1)*(nd+1)+x;
          else
              mesh->NodesQuad1[tqq+(x+nd)*nd+z][2]=tnn+(x-1+nd-1)*(nd-1)+z-1;
          if (z==nd-1)
              mesh->NodesQuad1[tqq+(x+nd)*nd+z][1]=x;
          else
              mesh->NodesQuad1[tqq+(x+nd)*nd+z][1]=tnn+(x-1+nd-1)*(nd-1)+z;
        }
        if (x==nd-1)    	    	    
        {
          if (z==0)
              mesh->NodesQuad1[tqq+(x+nd)*nd+z][3]=(nd+1)*(nd+1)+nd;
          else
              mesh->NodesQuad1[tqq+(x+nd)*nd+z][3]=tn+(z-1);
          if (z==nd-1)
              mesh->NodesQuad1[tqq+(x+nd)*nd+z][0]=nd;
          else
              mesh->NodesQuad1[tqq+(x+nd)*nd+z][0]=tn+z;
        }
        else
        {	
          if (z==0)
              mesh->NodesQuad1[tqq+(x+nd)*nd+z][3]=(nd+1)*(nd+1)+x+1;
          else
              mesh->NodesQuad1[tqq+(x+nd)*nd+z][3]=tnn+(x+nd-1)*(nd-1)+z-1;
          if (z==nd-1)
              mesh->NodesQuad1[tqq+(x+nd)*nd+z][0]=x+1;
          else
              mesh->NodesQuad1[tqq+(x+nd)*nd+z][0]=tnn+(x+nd-1)*(nd-1)+z;
        }  
      }
  for (e=0;e<mesh->NElemsQuad1;e++)
      for (i=0;i<4;i++)
          mesh->NodesQuad1[e][i]++;
  
  /* spherize */
  for (e=0;e<mesh->NNodes;e++)
  {
    l=0.0;
    for (i=0;i<3;i++)
        l+=pow(mesh->MeshPointCoords[e][i],2);
    l=sqrt(l);
    for (i=0;i<3;i++)
        if (l>1e-8)
            mesh->MeshPointCoords[e][i]/=l;
        else
            mesh->MeshPointCoords[e][i]=0.0;
  }
  
  /* convert to triangs */
  if (gtr)
  {
    mesh->NElemsTriang1=mesh->NElemsQuad1*2;
    mesh->NodesTriang1=e_alloc_cont_i_matr(mesh->NElemsTriang1, 3);
    mesh->KeywordSet[E_KW_NUM_NODES_TRIANG1]=TRUE;
    mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1]=TRUE;
    for (e=0;e<mesh->NElemsQuad1;e++)
    {
      l1=0.0;
      for (i=0;i<3;i++)
          l1+=pow(mesh->MeshPointCoords[mesh->NodesQuad1[e][0]-1][i]-
                  mesh->MeshPointCoords[mesh->NodesQuad1[e][2]-1][i],2);
      l2=0.0;
      for (i=0;i<3;i++)
          l2+=pow(mesh->MeshPointCoords[mesh->NodesQuad1[e][1]-1][i]-
                  mesh->MeshPointCoords[mesh->NodesQuad1[e][3]-1][i],2);
      if (l1<l2)
      {
        mesh->NodesTriang1[2*e][0]=mesh->NodesQuad1[e][0];
        mesh->NodesTriang1[2*e][1]=mesh->NodesQuad1[e][1];
        mesh->NodesTriang1[2*e][2]=mesh->NodesQuad1[e][2];
        mesh->NodesTriang1[2*e+1][0]=mesh->NodesQuad1[e][0];
        mesh->NodesTriang1[2*e+1][1]=mesh->NodesQuad1[e][2];
        mesh->NodesTriang1[2*e+1][2]=mesh->NodesQuad1[e][3];
      }
      else
      {
        mesh->NodesTriang1[2*e][0]=mesh->NodesQuad1[e][0];
        mesh->NodesTriang1[2*e][1]=mesh->NodesQuad1[e][1];
        mesh->NodesTriang1[2*e][2]=mesh->NodesQuad1[e][3];
        mesh->NodesTriang1[2*e+1][0]=mesh->NodesQuad1[e][1];
        mesh->NodesTriang1[2*e+1][1]=mesh->NodesQuad1[e][2];
        mesh->NodesTriang1[2*e+1][2]=mesh->NodesQuad1[e][3];
      }
    }
    mesh->KeywordSet[E_KW_NUM_NODES_QUAD1]=FALSE;
    mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1]=FALSE;
    e_free_cont_i_matr(mesh->NElemsQuad1, mesh->NodesQuad1);
  }
  
  if (gtr)
      e_PutMeshData(mesh, "triangs");
  else
      e_PutMeshData(mesh, "quads");
  
  e_free(mesh);

  return 0;
}

