#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <e_lib.h>

int main(int argc, char *argv[])
{
  MESH    *mesh;
  int     nd;
  int     x,y,z,e,i;
  int     tn, tq, tnn, tqq;
  double  l,l1,l2;
 
  if (argc<2)
  {
    printf("\nGenerate a volume mesh of a sphere.\n");
    printf("\nUsage: vsph n\n");
    printf("  n = number of subdivisions per side\n\n");
    exit(1);
  }
  if (sscanf(argv[1], "%d", &nd)!=1)
  {
    e_GeneralError("Invalid number of divisions");
  }
 
  if (nd<2)
    e_GeneralError("n must be > 2");
 
  mesh=NULL;
  mesh=(MESH *)e_calloc(1, sizeof(MESH));
  e_NullifyMesh(mesh);
 
  mesh->Title=e_alloc_c_vect(strlen("volume sphere")+1);
  strcpy(mesh->Title, "volume sphere");
  mesh->NMeshPoints=(nd+1)*(nd+1)*(nd+1);
  mesh->NNodes=mesh->NMeshPoints;
  mesh->MeshPointCoords=e_alloc_cont_d_matr(mesh->NNodes, 3);
  mesh->NElemsBlock1=nd*nd*nd;
  mesh->NodesBlock1=e_alloc_cont_i_matr(mesh->NElemsBlock1, 8);
  mesh->KeywordSet[E_KW_NUM_TITLE]=TRUE;
  mesh->KeywordSet[E_KW_NUM_NNODES]=TRUE;
  mesh->KeywordSet[E_KW_NUM_NMESHPOINTS]=TRUE;
  mesh->KeywordSet[E_KW_NUM_NODES_BLOCK1]=TRUE;
  mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1]=TRUE;
  mesh->KeywordSet[E_KW_NUM_MESHPOINT_COORDS]=TRUE;
    
  for (z=0;z<nd+1;z++)
    for (y=0;y<nd+1;y++)
      for (x=0;x<nd+1;x++)
      {
    	  mesh->MeshPointCoords[(z*(nd+1)+y)*(nd+1)+x][0]=-1.0+(2.0/nd)*x;
    	  mesh->MeshPointCoords[(z*(nd+1)+y)*(nd+1)+x][1]=-1.0+(2.0/nd)*y;
    	  mesh->MeshPointCoords[(z*(nd+1)+y)*(nd+1)+x][2]=-1.0+(2.0/nd)*z;
    	}
  for (z=0;z<nd;z++)
    for (y=0;y<nd;y++)
      for (x=0;x<nd;x++)
      {
        mesh->NodesBlock1[(z*nd+y)*nd+x][0]=(z*(nd+1)+y)*(nd+1)+x+1;
        mesh->NodesBlock1[(z*nd+y)*nd+x][1]=(z*(nd+1)+y)*(nd+1)+x+2;
        mesh->NodesBlock1[(z*nd+y)*nd+x][2]=(z*(nd+1)+y+1)*(nd+1)+x+2;
        mesh->NodesBlock1[(z*nd+y)*nd+x][3]=(z*(nd+1)+y+1)*(nd+1)+x+1;
        mesh->NodesBlock1[(z*nd+y)*nd+x][4]=((z+1)*(nd+1)+y)*(nd+1)+x+1;
        mesh->NodesBlock1[(z*nd+y)*nd+x][5]=((z+1)*(nd+1)+y)*(nd+1)+x+2;
        mesh->NodesBlock1[(z*nd+y)*nd+x][6]=((z+1)*(nd+1)+y+1)*(nd+1)+x+2;
        mesh->NodesBlock1[(z*nd+y)*nd+x][7]=((z+1)*(nd+1)+y+1)*(nd+1)+x+1;
      }
    
  /* spherize */
  for (e=0;e<mesh->NNodes;e++)
  {
    l=0.0;
    for (i=0;i<3;i++)
    	  l+=pow(mesh->MeshPointCoords[e][i],2);
    l=sqrt(l)/
      E_MAX3(E_ABS(mesh->MeshPointCoords[e][0]),
        		 E_ABS(mesh->MeshPointCoords[e][1]),
             E_ABS(mesh->MeshPointCoords[e][2]));
    for (i=0;i<3;i++)
    	  if (l>1e-8)
    	    mesh->MeshPointCoords[e][i]/=l;
    	  else
    	    mesh->MeshPointCoords[e][i]=0.0;
  } 

  e_PutMeshData(mesh, "vsph");
  e_FreeMeshData(mesh);
  e_free(mesh);

  return 0;
}

