#include <e_lib.h>
#include "nodelem.h"

static void write_hyper_graph(char *filename, int nedge, int nvert,
                              int nlink, int *degree, int **edges);



int main(int argc, char *argv[])
{
  char   line[E_LINE_MAX];
  char  *check;
  char   filename[E_LINE_MAX];
  MESH  *mesh;
  int  **node_elem;
  int   *no_node_elem;

  printf("\n\nProgram to make vertex weighted hypergraphs");
  printf("\nfor cable-membrane structures.");
 
  if(argc > 1)
  {
    strcpy(filename, argv[1]);
  }
  else
  {
    printf("\n\nType in the input filename (.mdf) !");
    printf("\nGive no extension : ");
    check = fgets(line, E_LINE_MAX, stdin);
    if(   (check == NULL)
       || (sscanf(line, "%s", &filename) != 1) )
      e_GeneralError("Error in keyboard input, can not read filename!");
  }

  printf("\nThe projectname is taken as: %s", filename);

  e_GetMeshData(&mesh, filename);

  if(mesh->TotalNElems != mesh->NElemsLink1 +
                          mesh->NElemsLink2 +
                          mesh->NElemsLink3 +
                          mesh->NElemsLink4 +
                          mesh->NElemsLink5 +
                          mesh->NElemsTriang3 +
                          mesh->NElemsTriang4 +
                          mesh->NElemsTriang5   )
  {
    e_GeneralError("Mesh contains non-supported element !");
  }

  node_elem = e_init_frag_i_matr(mesh->NNodes);
  no_node_elem = e_alloc_i_vect(mesh->NNodes);

  e_InitMeshPointElemConnection(mesh, no_node_elem, node_elem);

  write_hyper_graph(filename, mesh->NNodes, mesh->TotalNElems,
                    mesh->NElemsLink1 +
                    mesh->NElemsLink2 +
                    mesh->NElemsLink3 +
                    mesh->NElemsLink4 +
                    mesh->NElemsLink5,
                    no_node_elem, node_elem);

  printf("\n\n");
  
  return(0);
}



static void write_hyper_graph(char *filename, int nedge, int nvert,
                              int nlink, int *degree, int **edges)
{
  char   graph_file[E_LINE_MAX];
  FILE  *fp;
  int    i, j, counter;

  strcpy(graph_file, filename);
  strcat(graph_file, ".graph");

  if ((fp = fopen(graph_file, "w")) == NULL)
  {
    e_GeneralError("Can not open graph file for write !");
  }

  printf("\n\n Graph file: %s", graph_file);

  counter = 0;
  for(i = 0; i < nedge; i++)
  {
    if(degree[i] > 0)
        counter++;
  }
  
  fprintf(fp, "%d %d 10", counter, nvert);

  for(i = 0; i < nedge; i++)
  {
    if(degree[i] > 0)
    {
      fprintf(fp, "\n");
      for(j = 0; j < degree[i]; j++)
      {
        fprintf(fp, "%d ", edges[i][j]);
      }
    }
    else
    {
      printf("\nempty %d", i+1);
    }
    
    
  }

  for(i = 0; i < nvert; i++)
  {
    if(i < nlink )
        fprintf(fp, "\n1");
    else
        fprintf(fp, "\n7");
  }

  fclose(fp);

}


