#include <stdio.h>
#include <math.h>
#include <e_lib.h>
#include "graph.h"

/* 
   it is defined if the vertex weighted graph 
   should be written out
   1 - links
   7 - membrane
*/

 #define WEIGHTED_GRAPH 1 


char        GV_filename[E_LINE_MAX];
MESH       *GV_global_mesh;
int       **GV_node_elem;
int        *GV_no_node_elem;
int       **GV_adjacency;
int        *GV_degree;


static void f_read_in_data(int argc, char *argv[]);
static void write_graph_file(char *filename, int no, 
                             int *degree, int **adjacency);


static void f_read_in_data(int argc, char *argv[])
{
  char   line[E_LINE_MAX];
  char  *check;

  if(argc > 1)
  {
    strcpy(GV_filename, argv[1]);
  }
  else
  {
    printf("\n\nType in the input filename (.mdf) !");
    printf("\nGive no extension : ");
    check = fgets(line, E_LINE_MAX, stdin);
    if(   (check == NULL)
       || (sscanf(line, "%s", &GV_filename) != 1) )
      e_GeneralError("Error in keyboard input, can not read filename!");
  }

  printf("\nThe projectname is taken as: %s", GV_filename);

  e_GetMeshData(&GV_global_mesh, GV_filename);

  if(GV_global_mesh->TotalNElems != GV_global_mesh->NElemsLink1 +
                                    GV_global_mesh->NElemsLink2 +
                                    GV_global_mesh->NElemsLink3 +
                                    GV_global_mesh->NElemsLink4 +
                                    GV_global_mesh->NElemsLink5 +
                                    GV_global_mesh->NElemsTriang3 +
                                    GV_global_mesh->NElemsTriang4 +
                                    GV_global_mesh->NElemsTriang5   )
    e_GeneralError("Mesh contains non-supported element !");
}



static void write_graph_file(char *filename, int no, 
                             int *degree, int **adjacency)
{
  char graph_file[E_LINE_MAX];
  FILE  *fp;
  int i, j, edges;

  strcpy(graph_file, filename);
  strcat(graph_file, ".graph");

  if ((fp = fopen(graph_file, "w")) == NULL)
  {
    e_GeneralError("Can not open graph file for write !");
  }

  printf("\n\n Graph file: %s", graph_file);

#ifndef WEIGHTED_GRAPH
  edges = 0;
  for(i = 0; i < no; i++)
  {
    edges += degree[i];
  }

  fprintf(fp, "%d %d", no, edges/2);

  for(i = 0; i < no; i++)
  {
    fprintf(fp, "\n");
    for(j = 0; j < degree[i]; j++)
    {
      fprintf(fp, "%d ", adjacency[i][j]);
    }
  }
#else
  edges = 0;
  for(i = 0; i < no; i++)
  {
    edges += degree[i];
  }

  fprintf(fp, "%d %d 10", no, edges/2);

  for(i = 0; i < no; i++)
  {
    fprintf(fp, "\n");
    if(i < GV_global_mesh->NElemsLink1 +
           GV_global_mesh->NElemsLink2 +
           GV_global_mesh->NElemsLink3 +
           GV_global_mesh->NElemsLink4 +
           GV_global_mesh->NElemsLink5   )
      fprintf(fp, "1 ");
    else
      fprintf(fp, "7 ");
    for(j = 0; j < degree[i]; j++)
    {
      fprintf(fp, "%d ", adjacency[i][j]);
    }
  }
#endif

  fclose(fp);

}



main(int argc, char *argv[])
{
  int no_row;

#ifdef WEIGHTED_GRAPH
  printf("\n\nWEIGHTED GRAPH GENERATION !!!");
#endif

#ifdef __COMM_GRAPH__
  printf("\nProgram to create the Communication Graph of a mesh.");
  printf("\n  All element is connected with all other elements");
  printf("\n  at a node.");
#endif

#ifdef __DUAL_GRAPH__
  printf("\nProgram to create the Dual Graph of a mesh.");
  printf("\n  Out of plane link elements are connected with");
  printf("\n  all other elements at both end of the link.");
#endif

#ifdef __CIRC_GRAPH__
  printf("\nProgram to create the 'Dual' Graph of a mesh.");
  printf("\n  It tries to determine the Shortest Hamiltonian");
  printf("\n  path around a node. (NP complete)");
#endif

  printf("\n-----------------------------------------------------\n");

  f_read_in_data(argc, argv);

  GV_node_elem = e_init_frag_i_matr(GV_global_mesh->NNodes);
  GV_no_node_elem = e_alloc_i_vect(GV_global_mesh->NNodes);

  e_InitMeshPointElemConnection(GV_global_mesh, GV_no_node_elem, GV_node_elem);

  GV_degree = e_alloc_i_vect(GV_global_mesh->TotalNElems);
  GV_adjacency = e_init_frag_i_matr(GV_global_mesh->TotalNElems);

#ifdef __COMM_GRAPH__
  SetCommAdjacency(GV_global_mesh, GV_degree, GV_adjacency, 
                   GV_no_node_elem, GV_node_elem);
#endif

#ifdef __DUAL_GRAPH__
  SetDualAdjacency(GV_global_mesh, GV_degree, GV_adjacency,
                   GV_no_node_elem, GV_node_elem);
#endif

#ifdef __CIRC_GRAPH__
  SetCircAdjacency(GV_global_mesh, GV_degree, GV_adjacency,
                   GV_no_node_elem, GV_node_elem);
#endif


  no_row    = GV_global_mesh->TotalNElems;
  write_graph_file(GV_filename, no_row, GV_degree, GV_adjacency);

  e_free_frag_i_matr(no_row, GV_adjacency);
  e_free(GV_degree);

  e_free_frag_i_matr(GV_global_mesh->NNodes, GV_node_elem);
  e_free(GV_no_node_elem);


  printf("\n\n");

}



