#include <stdio.h>
#include <math.h>
#include <e_lib.h>
#include "graph.h"


void SetTriangAdjacencyMatrix(MESH *mesh, int *Degree, int **Adjacency,
                              int NElements, int **Vertices, int StartIndex,
                              int *no_node_elem, int **node_elem)
{
  char message[E_LINE_MAX+1];
  int i, i_node, j, k, i_s, ip;
  int ElementIndex, ns, nx, n1, n2, placed, no_link;
  int triangle[MAX_SHARED_TRIANGLES];

  no_link = mesh->NElemsLink1 + mesh->NElemsLink2 + mesh->NElemsLink3 +
            mesh->NElemsLink4 + mesh->NElemsLink5;

  for(i = 0; i < NElements; i++)
  {
    ElementIndex = StartIndex + i;

    /* search for other triangle elements */
    ns = 0;
    for(i_node = 0; i_node < 3; i_node++) /* there are 3 sides to check */
    {
      nx = 2 - i_node;
      n1 = Vertices[i][ns]-1;  /* vertices of the side */
      n2 = Vertices[i][nx]-1;

      placed = 0;
      /* check all possible combination of elements */
      for(j = 0; j < no_node_elem[n1]; j++)
      {
        for(k = 0; k < no_node_elem[n2]; k++)
        {
          /* the element contains the two nodes */
          if((node_elem[n1][j] == node_elem[n2][k])   &&
             (node_elem[n1][j] > no_link)             &&     /* it is a triangle */
             (node_elem[n1][j] != ElementIndex)           )  /* it is not itself */
          {
            /* one side can have only one/two adjacent element */
            if(placed == MAX_SHARED_TRIANGLES)
            {
              sprintf(message, "Triangle %d shares an edge with %d other triangles !",
                      i+1, MAX_SHARED_TRIANGLES+1);
              e_GeneralError(message);
            }
            triangle[placed] = node_elem[n1][j];
            placed++;
          }
        }
      }

      /* store adjacent elements */
      for(ip = 0; ip < placed; ip++)
      {
        /* check whether we already defined this adjacency */
        for(i_s = 0; i_s < Degree[ElementIndex-1]; i_s++)
          if(Adjacency[ElementIndex-1][i_s] == triangle[ip])
            break;

        if(i_s == Degree[ElementIndex-1])
        {
          Adjacency[ElementIndex-1][Degree[ElementIndex-1]++] = triangle[ip];
#ifdef __CIRC_GRAPH__
          GV_handled[ElementIndex-1] = 1;
#endif
        }
      }

      ns = nx;
    }
  }
}


/* This function should be called after that the adjacency matrix has been
   assembled for the Triang elements */

void SetLinkAdjacencyMatrix(MESH *mesh, int *Degree, int **Adjacency,
                           int NElements, int **Vertices, int StartIndex,
                           int *no_node_elem, int **node_elem)
{
  char message[E_LINE_MAX];
  int i, j, k, n1, n2, ElementIndex;
  int triangle[MAX_SHARED_TRIANGLES+1];
  short placed;
  int elem, i_ad, i_s, i_node, no_link, ns, nx;

  no_link = mesh->NElemsLink1 + mesh->NElemsLink2 + mesh->NElemsLink3 +
            mesh->NElemsLink4 + mesh->NElemsLink5;

  for(i = 0; i < NElements; i++)
  {
    n1 = Vertices[i][0] - 1;
    n2 = Vertices[i][1] - 1;

    ElementIndex = StartIndex + i;

    /* search for triangle elements sharing two nodes with the link */
    placed = 0;
    for(j = 0; j < no_node_elem[n1]; j++)
    {
      for(k = 0; k < no_node_elem[n2]; k++)
      {
        if((node_elem[n1][j] == node_elem[n2][k]) &&
           (node_elem[n1][j] > no_link)                )
        {
          /* sanity check */
          if(node_elem[n1][j] <= no_link)
          {
            sprintf(message, "Link %d and %d share the same nodes !", i+1, node_elem[n1][j]);
            e_GeneralError(message);
          }

          if(placed == MAX_SHARED_TRIANGLES+1)
          {
            sprintf(message, "Link %d belongs to too many triangle element !", i+1);
            e_GeneralError(message);
          }

          triangle[placed] = node_elem[n1][j];
          placed++;
        }
      }
    }

/* if there is a triangle sharing two nodes with the link, put the link adjacent to this triangle,
   if there are two such triangles, put the link in between,
   if there are three such triangles, put the link in the middle, cut connection between triangles */

#if 0
    ns = 2;
#endif

    for(i_ad = 0; i_ad < placed; i_ad++)
    {
      elem = triangle[i_ad];

      if(placed == 1)
      {
        /* increase space in triangle adjacency info */
        i_s = Degree[elem-1];
        Degree[elem-1]++;
        Adjacency[elem-1] = e_incr_i_vect(Degree[elem-1]+GRAPH_ALLOC, Adjacency[elem-1], GRAPH_ALLOC);
      }
      else if(placed == 2)
      {
        /* find in the triangle list the adjacent one, which will be replaced */
        for(i_s = 0; i_s < Degree[elem-1]; i_s++)
          if(Adjacency[elem-1][i_s] == triangle[placed-i_ad-1])
            break;
        if(i_s == Degree[elem-1])
        {
          e_GeneralError("Impossible case 2! Could not find proper adjacent triangle!");
        }

#ifdef __CIRC_GRAPH__
        GV_handled[ElementIndex-1] = 1;
#endif

      }
      else if(placed == 3)
      {
#if 0
        if(i_ad == 2)
          continue;

        ns--;
#endif 

        if(i_ad > 0) ns = 0; else ns = 1;
        if(i_ad < 2) nx = 2; else nx = 1;

        /* find the 'first' triangle in adjacency list */
        for(i_s = 0; i_s < Degree[elem-1]; i_s++)
          if(Adjacency[elem-1][i_s] == triangle[ns])
            break;
        if(i_s == Degree[elem-1])
        {
          e_GeneralError("Impossible case 3! Could not find proper adjacent triangle 1 !");
        }

        for( ; i_s < Degree[elem-1]-1; i_s++)
          Adjacency[elem-1][i_s] = Adjacency[elem-1][i_s+1];
        Degree[elem-1]--;

        /* find the 'second' triangle in adjacency list and keep it to replace by link */
        for(i_s = 0; i_s < Degree[elem-1]; i_s++)
          if(Adjacency[elem-1][i_s] == triangle[nx])
            break;
        if(i_s == Degree[elem-1])
        {
          e_GeneralError("Impossible case 3! Could not find proper adjacent triangle 2 !");
        }

#ifdef __CIRC_GRAPH__
        GV_handled[ElementIndex-1] = 1;
#endif

      }

      /* set the adjacent element */
      Adjacency[elem-1][i_s] = ElementIndex;

      /* set the link element */
      Adjacency[ElementIndex-1][Degree[ElementIndex-1]] = elem;
      Degree[ElementIndex-1]++;

      /* initial alloc is GRAPH_ALLOC in dual_graph.c */
      Adjacency[ElementIndex-1] = e_incr_i_vect( Degree[ElementIndex-1]+GRAPH_ALLOC,
                                                 Adjacency[ElementIndex-1], GRAPH_ALLOC);

    }

#ifdef __DUAL_PLANE_GRAPH__
    if(!placed)
    {
      e_GeneralError("Mesh contains links out-of-plane !");
    }
#endif

#ifdef __DUAL_GRAPH__
    if(!placed)
    {
/* link is not adjacent to any triangle, but there might be different triangles
   at both ends, we connect the remaining links with ??? */

      /* go through all elements at node 1 */
      for(i_node = 0; i_node < no_node_elem[n1]; i_node++)
      {
        elem = node_elem[n1][i_node];
        for(i_s = 0; i_s < Degree[elem-1]; i_s++)
          if((Adjacency[elem-1][i_s] == ElementIndex) ||
             (elem == ElementIndex)                      )
            break;

        if((i_s == Degree[elem-1]) && (elem != ElementIndex))
        {
          Adjacency[elem-1][Degree[elem-1]] = ElementIndex;
          Degree[elem-1]++;
          Adjacency[elem-1] = e_incr_i_vect( Degree[elem-1]+GRAPH_ALLOC,
                                             Adjacency[elem-1], GRAPH_ALLOC );
        }

        for(i_s = 0; i_s < Degree[ElementIndex-1]; i_s++)
          if((Adjacency[ElementIndex-1][i_s] == elem) ||
             (elem == ElementIndex)                      )
            break;
        if((i_s == Degree[ElementIndex-1]) && (elem != ElementIndex))
        {
          Adjacency[ElementIndex-1][Degree[ElementIndex-1]] = elem;
          Degree[ElementIndex-1]++;
          Adjacency[ElementIndex-1] = e_incr_i_vect( Degree[ElementIndex-1]+GRAPH_ALLOC,
                                                     Adjacency[ElementIndex-1], GRAPH_ALLOC );
        }
      }

      /* go through all elements at node 2 */
      for(i_node = 0; i_node < no_node_elem[n2]; i_node++)
      {
        elem = node_elem[n2][i_node];
        for(i_s = 0; i_s < Degree[elem-1]; i_s++)
          if((Adjacency[elem-1][i_s] == ElementIndex) ||
             (elem == ElementIndex)                      )
            break;
        if((i_s == Degree[elem-1]) && (elem != ElementIndex))
        {
          Adjacency[elem-1][Degree[elem-1]] = ElementIndex;
          Degree[elem-1]++;
          Adjacency[elem-1] = e_incr_i_vect( Degree[elem-1]+GRAPH_ALLOC,
                                             Adjacency[elem-1], GRAPH_ALLOC );
        }

        for(i_s = 0; i_s < Degree[ElementIndex-1]; i_s++)
          if((Adjacency[ElementIndex-1][i_s] == elem) ||
             (elem == ElementIndex)                      )
            break;
        if((i_s == Degree[ElementIndex-1]) && (elem != ElementIndex))
        {
          Adjacency[ElementIndex-1][Degree[ElementIndex-1]] = elem;
          Degree[ElementIndex-1]++;
          Adjacency[ElementIndex-1] = e_incr_i_vect( Degree[ElementIndex-1]+GRAPH_ALLOC,
                                                     Adjacency[ElementIndex-1], GRAPH_ALLOC );
        }
      }
    }
#endif

  }
}

