/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                    Farhat's Method For Non-Mixed Meshes                    */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include "e_pltfrm.h"

#ifndef __E_T800SAL__
#include <stdio.h>
#endif

#ifdef __E_T800__
#include <chan.h>
#endif

#include <e_lib.h>

#include "frht.h"

/*
#define __DEBUG__
*/

/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                                  farhat()                                  */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* farhat(): To partition a non-mixed mesh using Farhat's method. The return
   value 1 if the partitioning is successful, else 0.                         */


int farhat(MESH *mesh, int NSubdomains, int NStartMinNode, 
           int LinkGrab, int *NElementsPerSubdomain)
{
   int i,k;
   int *FrontElements;
   int NElements;
   int NFrontElements;
   int SubdomainIndex;


   if (!Initialise(mesh, NSubdomains, LinkGrab))
   {
      return 0;
   }

#ifdef __DEBUG__
   e_FreeDecompData(mesh);
   
   mesh->Decomp.NSubDoms = NSubdomains;    
   mesh->Decomp.SubDoms = g_SubdomainIndices;

   mesh->Decomp.SubDomsLink1
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_LINK1] - 1;
   mesh->Decomp.SubDomsLink2
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_LINK2] - 1;
   mesh->Decomp.SubDomsLink3
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_LINK3] - 1;
   mesh->Decomp.SubDomsLink4
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_LINK4] - 1;
   mesh->Decomp.SubDomsLink5
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_LINK5] - 1;
   mesh->Decomp.SubDomsTriang1
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] - 1;
   mesh->Decomp.SubDomsTriang2
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2] - 1;
   mesh->Decomp.SubDomsTriang3
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] - 1;
   mesh->Decomp.SubDomsTriang4
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] - 1;
   mesh->Decomp.SubDomsTriang5
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] - 1;
   mesh->Decomp.SubDomsQuad1
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1] - 1;
   mesh->Decomp.SubDomsTetrah1
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1] - 1;
   mesh->Decomp.SubDomsBlock1
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1] - 1;
   mesh->SubStructSet[E_SS_NUM_DECOMP] = TRUE;

#endif

   for (SubdomainIndex = 1; SubdomainIndex < NSubdomains; SubdomainIndex++)
   {
      k = SubdomainIndex - 1;
      NElements = 0;
      
      while (NElements < NElementsPerSubdomain[k])
      {
         if (!InitialiseFront(NStartMinNode, SubdomainIndex,
                              NElementsPerSubdomain,
                              &NFrontElements,&FrontElements))
         {
            FreeGlobalVariables();
            FreeLocalVariables(NElementsPerSubdomain);
            return 0;
         }
            
         NElements += NFrontElements;
         
         while (NFrontElements > 0)
         {
            if (!AdvanceFront(SubdomainIndex,NElementsPerSubdomain[k],
                              &NElements,&NFrontElements,&FrontElements))
            {
               FreeGlobalVariables();
               FreeLocalVariables(NElementsPerSubdomain);
               return 0;
            }

#ifdef __DEBUG__
{
  static counter = 1;
  char   name[E_LINE_MAX];

  mesh->SubStructSet[E_SS_NUM_DECOMP] = TRUE;

  sprintf(name,"out%d", counter);
  e_PutDecompData(mesh, name);
  counter++;

}
#endif

            if (NElements == NElementsPerSubdomain[k]) break;
         }         
      }
   }

   for (i = 0; i < mesh->TotalNElems; i++)
   {
      if (g_SubdomainIndices[i] == 0)
         g_SubdomainIndices[i] = NSubdomains;
   }

#ifndef __DEBUG__
   e_FreeDecompData(mesh);
   
   mesh->Decomp.NSubDoms = NSubdomains;    
   mesh->Decomp.SubDoms = g_SubdomainIndices;

   mesh->Decomp.SubDomsLink1
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_LINK1] - 1;
   mesh->Decomp.SubDomsLink2
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_LINK2] - 1;
   mesh->Decomp.SubDomsLink3
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_LINK3] - 1;
   mesh->Decomp.SubDomsLink4
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_LINK4] - 1;
   mesh->Decomp.SubDomsLink5
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_LINK5] - 1;
   mesh->Decomp.SubDomsTriang1
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] - 1;
   mesh->Decomp.SubDomsTriang2
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2] - 1;
   mesh->Decomp.SubDomsTriang3
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] - 1;
   mesh->Decomp.SubDomsTriang4
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] - 1;
   mesh->Decomp.SubDomsTriang5
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] - 1;
   mesh->Decomp.SubDomsQuad1
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1] - 1;
   mesh->Decomp.SubDomsTetrah1
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1] - 1;
   mesh->Decomp.SubDomsBlock1
      = g_SubdomainIndices
        + mesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1] - 1;
   mesh->SubStructSet[E_SS_NUM_DECOMP] = TRUE;
#endif

   g_SubdomainIndices = NULL;
   FreeGlobalVariables();
   FreeLocalVariables(NElementsPerSubdomain);

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* Initialise(): To initialise variables. The return value is 1 if the
   operation is successful, else 0.                                           */


static int Initialise(MESH *mesh, int NSubdomains, int LinkGrab)
{
   int i,j,k;
   int remainder;
   int NVpE_Max;
   int n_link, n1, n2, elem;   
   
   if (  mesh->NElemsLink1
       + mesh->NElemsLink2
       + mesh->NElemsLink3
       + mesh->NElemsLink4
       + mesh->NElemsLink5
       + mesh->NElemsTriang1
       + mesh->NElemsTriang2
       + mesh->NElemsTriang3
       + mesh->NElemsTriang4
       + mesh->NElemsTriang5
       + mesh->NElemsQuad1
       + mesh->NElemsTetrah1
       + mesh->NElemsBlock1  != mesh->TotalNElems)
   {
      e_GeneralError("unsupported element type");
      return 0;
   }

   NVpE_Max = 3;

   if(mesh->NElemsTetrah1 > 0 ||
      mesh->NElemsQuad1   > 0    )
      NVpE_Max = 4;
   if (mesh->NElemsBlock1 > 0)
      NVpE_Max = 8;
   if (mesh->NElemsQuad1 > 0)
      NVpE_Max = 9;

   g_NMeshPoints = mesh->NMeshPoints;
   g_TotalNElems = mesh->TotalNElems;

   g_NVpE = NULL;
   g_Vertices = NULL;

   g_NAdjacentElements = NULL;
   g_AdjacentElements = NULL;
   g_Weights = NULL;
   g_SubdomainIndices = NULL;

   if (                   (g_NVpE = e_alloc_i_vect(g_TotalNElems)) == NULL
       || (g_Vertices
                    = e_alloc_frag_i_matr(g_TotalNElems,NVpE_Max)) == NULL
       ||    (g_NAdjacentElements = e_alloc_i_vect(g_NMeshPoints)) == NULL
       || (g_AdjacentElements = e_init_frag_i_matr(g_NMeshPoints)) == NULL
       || (g_SingularElements = e_alloc_cont_i_matr(g_TotalNElems, 18)) == NULL
       ||    (g_NSingularElements = e_alloc_i_vect(g_TotalNElems)) == NULL
       ||              (g_Weights = e_alloc_i_vect(g_NMeshPoints)) == NULL
       ||     (g_SubdomainIndices = e_alloc_i_vect(g_TotalNElems)) == NULL)
   {
      FreeGlobalVariables();
      return 0;
   }

   k = 0;

   for (i = 0; i < mesh->NElemsLink1; i++)
   {
      for (j = 0; j < 2; j++)
         g_Vertices[k][j] = mesh->NodesLink1[i][j];
         
      g_NVpE[k++] = 2;
   }

   for (i = 0; i < mesh->NElemsLink2; i++)
   {
      for (j = 0; j < 2; j++)
         g_Vertices[k][j] = mesh->NodesLink2[i][j];
         
      g_NVpE[k++] = 2;
   }

   for (i = 0; i < mesh->NElemsLink3; i++)
   {
      for (j = 0; j < 2; j++)
         g_Vertices[k][j] = mesh->NodesLink3[i][j];
         
      g_NVpE[k++] = 2;
   }

   for (i = 0; i < mesh->NElemsLink4; i++)
   {
      for (j = 0; j < 2; j++)
         g_Vertices[k][j] = mesh->NodesLink4[i][j];
         
      g_NVpE[k++] = 2;
   }

   for (i = 0; i < mesh->NElemsLink5; i++)
   {
      for (j = 0; j < 2; j++)
         g_Vertices[k][j] = mesh->NodesLink5[i][j];
         
      g_NVpE[k++] = 2;
   }
   
   for (i = 0; i < mesh->NElemsTriang1; i++)
   {
      for (j = 0; j < 3; j++)
         g_Vertices[k][j] = mesh->NodesTriang1[i][j];
         
      g_NVpE[k++] = 3;
   }
   
   for (i = 0; i < mesh->NElemsTriang2; i++)
   {
      for (j = 0; j < 3; j++)
         g_Vertices[k][j] = mesh->NodesTriang2[i][j];
         
      g_NVpE[k++] = 3;
   }

   for (i = 0; i < mesh->NElemsTriang3; i++)
   {
      for (j = 0; j < 3; j++)
         g_Vertices[k][j] = mesh->NodesTriang3[i][j];
         
      g_NVpE[k++] = 3;
   }

   for (i = 0; i < mesh->NElemsTriang4; i++)
   {
      for (j = 0; j < 3; j++)
         g_Vertices[k][j] = mesh->NodesTriang4[i][j];
         
      g_NVpE[k++] = 3;
   }

   for (i = 0; i < mesh->NElemsTriang5; i++)
   {
      for (j = 0; j < 3; j++)
         g_Vertices[k][j] = mesh->NodesTriang5[i][j];
         
      g_NVpE[k++] = 3;
   }
   
   for (i = 0; i < mesh->NElemsQuad1; i++)
   {
      for (j = 0; j < 4; j++)
         g_Vertices[k][j] = mesh->NodesQuad1[i][j];
         
      g_NVpE[k++] = 4;
   }

   for (i = 0; i < mesh->NElemsQuad3; i++)
   {
      for (j = 0; j < 9; j++)
         g_Vertices[k][j] = mesh->NodesQuad3[i][j];
         
      g_NVpE[k++] = 9;
   }
   
   for (i = 0; i < mesh->NElemsTetrah1; i++)
   {
      for (j = 0; j < 4; j++)
         g_Vertices[k][j] = mesh->NodesTetrah1[i][j];
         
      g_NVpE[k++] = 4;
   }
   
   for (i = 0; i < mesh->NElemsBlock1; i++)
   {
      for (j = 0; j < 8; j++)
         g_Vertices[k][j] = mesh->NodesBlock1[i][j];
         
      g_NVpE[k++] = 8;
   }


   for (i = 0; i < g_TotalNElems; i++)
   {
      for (j = 0; j < g_NVpE[i]; j++)
      {
         k = g_Vertices[i][j] - 1;

         g_AdjacentElements[k] = e_incr_i_vect(g_NAdjacentElements[k],
                                               g_AdjacentElements[k],
                                               NADJACENT_MAX          );
         if (g_AdjacentElements[k] == NULL)
         {
            FreeGlobalVariables();
            return 0;
         }
         
         g_AdjacentElements[k][g_NAdjacentElements[k]++] = i + 1;

      }
   }
   
  for (k = 0; k < g_NMeshPoints; k++)
     g_Weights[k] = g_NAdjacentElements[k];
      
/* collect the singular elements (links) for all other elements */
  if(LinkGrab != 0)
  {
    printf("\nLink elements are 'glued' with others.");
    n_link = mesh->NElemsLink1 + mesh->NElemsLink2 +
             mesh->NElemsLink3 + mesh->NElemsLink4 +
             mesh->NElemsLink5;
    for(i = 0; i < n_link; i++)
    {
      n1 = g_Vertices[i][0] - 1;
      n2 = g_Vertices[i][1] - 1;
      for(j = 0; j < g_NAdjacentElements[n1]; j++)
      {
        for(k = 0; k < g_NAdjacentElements[n2]; k++)
        {
          if((g_AdjacentElements[n1][j] == g_AdjacentElements[n2][k]) &&
             (g_AdjacentElements[n1][j] > n_link)                        )
          {
            elem = g_AdjacentElements[n1][j] - 1;
            g_SingularElements[elem][ g_NSingularElements[elem] ] = i+1;
            g_NSingularElements[elem]++;
          }
        }
      }
    }
  }
  else
    printf("\nLink elements are NOT 'glued' with others.");

/* indicates that there is no front at the moment */
   g_FirstFrontHasBeenCreated = FALSE;

   return 1;
}      


/* ////////////////////////////////////////////////////////////////////////// */
/* InitialiseFront(): To initialise a new element front. */


static int InitialiseFront(int NStartMinNode, int SubdomainIndex, 
                           int *NElementsPerSubdomain,
                           int *NFrontElements, int **FrontElements)
{
   int i;
   int MinWeightMeshPoint;


   if (g_FirstFrontHasBeenCreated)
   {
      i = SubdomainIndex - 1;        
      MinWeightMeshPoint
         = GetMinWeightMeshPoint(NStartMinNode, *NFrontElements,
                                 *FrontElements); 

      if (!CreateInitialFront(SubdomainIndex,NElementsPerSubdomain[i],
                              NFrontElements,FrontElements,MinWeightMeshPoint))
         return 0;
   }
   else
   {
      *NFrontElements = g_TotalNElems;

/* 
   Since there is no front, we create an imaginery one consisting of all
   the elements. This will be freed up in "CreateInitialFront" when the real
   front is created.
*/

      if ((*FrontElements = e_alloc_i_vect(g_TotalNElems)) == NULL)
         return 0;

      for (i = 0; i < g_TotalNElems; i++)
         (*FrontElements)[i] = i + 1;
         
      MinWeightMeshPoint
         = GetMinWeightMeshPoint(NStartMinNode, *NFrontElements,
                                 *FrontElements); 
         
      if (!CreateInitialFront(1,NElementsPerSubdomain[0],NFrontElements,
                                        FrontElements,MinWeightMeshPoint))
         return 0;
         
      g_FirstFrontHasBeenCreated = TRUE;
   }

   return 1;
}

static int e_sort_function(const void *a, const void *b)
{
  if( ((int *)a)[0] < ((int *)b)[0] )
    return(-1);
  if( ((int *)a)[0] > ((int *)b)[0] )
    return(1);
  return(0);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetMinWeightMeshPoint(): To return the index of mesh-point on the current
   element front with the lowest weight.                                      */


static int GetMinWeightMeshPoint(int NStartMinNode, int NFrontElements, 
                                 int *FrontElements)
{
   int i,j,k,l;
   int MinWeightMeshPoint;
   int WeightMin;
   int n_min_node;
   int **min_nodes;

  if(g_FirstFrontHasBeenCreated)
  {
    WeightMin = g_TotalNElems + 1;

    /*  First we search the front elements. */
    for (k = 0; k < NFrontElements; k++)
    {
      i = FrontElements[k] - 1;

      for (j = 0; j < g_NVpE[i]; j++)
      {
        l = g_Vertices[i][j] - 1;
         
        if (WeightMin > g_Weights[l] && g_Weights[l] > 0)
        {
          WeightMin = g_Weights[l];
          MinWeightMeshPoint = l + 1;
        }
      }
    }

/*
   If we could not find any element in the front then we search among all the
   elements.
*/
    if (WeightMin > g_TotalNElems)
    {
      e_GeneralWarning("\nStarting point NOT chosen from the front");
      for (i = 0; i < g_TotalNElems; i++)
      {
        if (g_SubdomainIndices[i] != 0)
        {
          for (j = 0; j < g_NVpE[i]; j++)
          {
            l = g_Vertices[i][j] - 1;
               
            if (WeightMin > g_Weights[l] && g_Weights[l] > 0)
            {
              WeightMin = g_Weights[l];
              MinWeightMeshPoint = l + 1;
            }
          }
        }
      }
    }
  }
  else
  {
/* the very first node with minimum weight is given by the user */
    min_nodes = e_alloc_cont_i_matr(g_NMeshPoints, 2);
    for (k = 0; k < NFrontElements; k++)
    {
      i = FrontElements[k] - 1;
      for (j = 0; j < g_NVpE[i]; j++)
      {
        l = g_Vertices[i][j] - 1;
        min_nodes[l][0] = g_Weights[l];
        min_nodes[l][1] = l+1;
        if(g_Weights[l] <= 0)
          e_GeneralError("Invalid weight at node %d", l+1);
        
      }
    }
    for(k = 0; k < g_NMeshPoints; k++)
      if(min_nodes[k][0] == 0)
      {
        min_nodes[k][0] = 10000;
        e_GeneralWarning("zero value: %d", k+1);
      }


    qsort((void *)min_nodes[0], g_NMeshPoints, 2*sizeof(int), e_sort_function);


    n_min_node = min_nodes[0][0];

    for (k = 1; k < g_NMeshPoints; k++)
      if(min_nodes[k][0] != n_min_node)
        break;
    printf("\nNumber of nodes with minimal weight = %d", k);

    if(NStartMinNode > g_NMeshPoints)
    {
      e_GeneralWarning("Starting node does not exist, using first node.");
      NStartMinNode = 1;
    }

    MinWeightMeshPoint = min_nodes[NStartMinNode-1][1];
    WeightMin = min_nodes[NStartMinNode-1][0];

    printf("\nMinWeightMeshPoint = %d weight = %d", MinWeightMeshPoint, WeightMin);
    printf("\n");

    e_free_cont_i_matr(g_NMeshPoints, min_nodes);
  }

  return MinWeightMeshPoint;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* CreateInitialFront(): To create an initial element front. */


static int CreateInitialFront(int SubdomainIndex, int NElementsPerSubdomain,
                                    int *NFrontElements, int **FrontElements,
                                                       int MinWeightMeshPoint)
{
   int i, j, k, ii, jj;
   int *NewFrontElements;
   int NNewFrontElements;
   int n_link;

   NNewFrontElements = 0;
   NewFrontElements = e_alloc_i_vect(NElementsPerSubdomain);

   if (NewFrontElements == NULL)
      return 0;

   k = MinWeightMeshPoint - 1;
   
   for (j = 0; j < g_NAdjacentElements[k]; j++)
   {
      i = g_AdjacentElements[k][j] - 1;
      
      if(g_SubdomainIndices[i] == 0)
      {
         g_SubdomainIndices[i] = SubdomainIndex;
         NewFrontElements[NNewFrontElements++] = i + 1;
         DecreaseVertexWeights(i);
         if (NNewFrontElements == NElementsPerSubdomain) break;

         if(g_NSingularElements[i] > 0)
         {
           for(ii = 0; ii < g_NSingularElements[i]; ii++)
           {
             jj = g_SingularElements[i][ii] - 1;
             if(g_SubdomainIndices[jj] == 0)
             {
               g_SubdomainIndices[jj] = SubdomainIndex;
               NewFrontElements[NNewFrontElements++] = jj + 1;
               DecreaseVertexWeights(jj);
               if(NNewFrontElements == NElementsPerSubdomain) break;
             }
           }
         }
         if (NNewFrontElements == NElementsPerSubdomain) break;

      }
   }
   
   e_free(*FrontElements);

   *NFrontElements = NNewFrontElements;
   *FrontElements = NewFrontElements;
   
   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* AdvanceFront(): To advance the element front by one layer. */


static int AdvanceFront(int SubdomainIndex, int NElementsPerSubdomain,
                                  int *NElements, int *NFrontElements,
                                                   int **FrontElements)
{
   int i,k;
   int *NewFrontElements;
   int NNewFrontElements;

   
   
   NNewFrontElements = 0;
   NewFrontElements = e_alloc_i_vect(NElementsPerSubdomain);
   
   if (NewFrontElements == NULL)
      return 0;

   for (k = 0; k < *NFrontElements; k++)
   {
      i = (*FrontElements)[k] - 1;
      
      PartitionNeighbouringElements(i,SubdomainIndex,NElementsPerSubdomain,
                                               NElements,&NNewFrontElements,
                                                            NewFrontElements);
      if (*NElements == NElementsPerSubdomain) 
         break;
   }

   e_free(*FrontElements);

   *NFrontElements = NNewFrontElements;
   *FrontElements = NewFrontElements;
   
   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* DecreaseVertexWeights(): To decrease the weights of each mesh-point
   attached to an element by 1.                                               */


static void DecreaseVertexWeights(int i)
{                                   
   int j;
   
   for (j = 0; j < g_NVpE[i]; j++)
      g_Weights[g_Vertices[i][j] - 1]--;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* PartitionNeighbouringElements(): To partition the elements adjacent to the
   current element front.                                                     */


static void PartitionNeighbouringElements(int i, int SubdomainIndex,
                                          int NElementsPerSubdomain,
                                          int *NElements,
                                          int *NNewFrontElements,
                                          int *NewFrontElements     )
{
   int j,k,l,m;
   int ii, jj;

   if (*NElements == NElementsPerSubdomain) return;

   for (j = 0; j < g_NVpE[i]; j++)
   {
      k = g_Vertices[i][j] - 1;
      
      for (l = 0; l < g_NAdjacentElements[k]; l++)
      {
         m = g_AdjacentElements[k][l] - 1;

         if (g_SubdomainIndices[m] == 0)
         {
            g_SubdomainIndices[m] = SubdomainIndex;
            NewFrontElements[(*NNewFrontElements)++] = m + 1;
            DecreaseVertexWeights(m);
            if (++(*NElements) == NElementsPerSubdomain) return;

            if(g_NSingularElements[m] > 0)
            {
              for(ii = 0; ii < g_NSingularElements[m]; ii++)
              {
                jj = g_SingularElements[m][ii] - 1;
                if(g_SubdomainIndices[jj] == 0)
                {
                  g_SubdomainIndices[jj] = SubdomainIndex;
                  NewFrontElements[(*NNewFrontElements)++] = jj + 1;
                  DecreaseVertexWeights(jj);
                  if(++(*NElements) == NElementsPerSubdomain) return;
                }
              }
            }
            if ((*NElements) == NElementsPerSubdomain) return;

         }
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* FreeGlobalVariables(): To free global variables. */


static void FreeGlobalVariables()
{
   e_free(g_NVpE);
   e_free_frag_i_matr(g_TotalNElems,g_Vertices);
   e_free(g_NAdjacentElements);
   e_free_frag_i_matr(g_NMeshPoints,g_AdjacentElements);
   e_free(g_Weights);
   e_free(g_SubdomainIndices);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* FreeLocalVariables(): To free local variables. */


static void FreeLocalVariables(int *NElementsPerSubdomain)
{
   e_free(NElementsPerSubdomain);
}
