/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                                 GetFiles()                                 */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include "e_pltfrm.h"

#ifndef __E_T800SAL__
#include <stdio.h>
#endif

#ifdef __E_T800__
#include <chan.h>
#endif

#include <string.h>
#include <e_lib.h>


/* ////////////////////////////////////////////////////////////////////////// */
/* GetFiles: To read input and output filenames from a file. The return value
   is 1 if the filenames are read successfully, else 0.                       */


int GetFiles(char ***InputFile, int NInputFiles, char ***OutputFile,
               int NOutputFiles, char *prompt, int argc, char *argv[])
{
   int i;
   int LineNumber = 0;          /* current line number */

   char *cp;                    /* character pointer */
   char *FilenamesFile;         /* filenames file filename */
   char  line[E_LINE_MAX + 1];  /* input line */
   char *check;
   FILE *fp;                    /* file pointer */


   if (argc > 2)
   {
      e_GeneralError("too many command line arguments");
      return 0;
   }

   if (argc == 1)
   {
      if (prompt == NULL)
         printf("filenames file: ");
      else
         printf("%s ", prompt);

      check = fgets(line, E_LINE_MAX, stdin);
      if(check == NULL)
        e_GeneralError("Can not read file!");


      if ((FilenamesFile = e_alloc_c_vect(strlen(line) + 5)) == NULL)
         return 0;

      sscanf(line,"%s",FilenamesFile);
      strcat(FilenamesFile,".fln");
   }
   else
   {
      if ((FilenamesFile = e_alloc_c_vect(strlen(argv[1]) + 5)) == NULL)
         return 0;

      sscanf(argv[1],"%s",FilenamesFile);
      strcat(FilenamesFile,".fln");
   }

   if ((fp = fopen(FilenamesFile,"r")) == NULL)
   {
      e_FileError(FilenamesFile,"cannot open filenames file");
      e_free(FilenamesFile);
      return 0;
   }

   *InputFile = NULL;
   *OutputFile = NULL;

   if (     (*InputFile = e_init_frag_c_matr(NInputFiles)) == NULL
       || (*OutputFile = e_init_frag_c_matr(NOutputFiles)) == NULL)
   {
      e_free(FilenamesFile);
      e_free(*InputFile);
      e_free(*OutputFile);
      return 0;
   }

   for (i = 0; i < NInputFiles; i++)
   {
      if (!e_GetNextLine(line,&LineNumber,FilenamesFile,fp))
      {
         e_free(FilenamesFile);
         e_free_frag_c_matr(NInputFiles,*InputFile);
         e_free_frag_c_matr(NOutputFiles,*OutputFile);
         return 0;
      }

      if ((cp = strchr(line,'\n')) != NULL)
         *cp = '\0';

      if (((*InputFile)[i] = e_alloc_c_vect(strlen(line) + 1)) == NULL)
      {
         e_free(FilenamesFile);
         e_free_frag_c_matr(NInputFiles,*InputFile);
         e_free_frag_c_matr(NOutputFiles,*OutputFile);
         return 0;
      }

      sscanf(line,"%s",(*InputFile)[i]);
   }

   for (i = 0; i < NOutputFiles; i++)
   {
      if (!e_GetNextLine(line,&LineNumber,FilenamesFile,fp))
      {
         e_free(FilenamesFile);
         e_free_frag_c_matr(NInputFiles,*InputFile);
         e_free_frag_c_matr(NOutputFiles,*OutputFile);
         return 0;
      }

      if ((cp = strchr(line,'\n')) != NULL)
         *cp = '\0';

      if (((*OutputFile)[i] = e_alloc_c_vect(strlen(line) + 1)) == NULL)
      {
         e_free(FilenamesFile);
         e_free_frag_c_matr(NInputFiles,*InputFile);
         e_free_frag_c_matr(NOutputFiles,*OutputFile);
         return 0;
      }

      sscanf(line,"%s",(*OutputFile)[i]);
   }

   e_free(FilenamesFile);
   fclose(fp);

   return 1;
}
