/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*              The Main Function Of The Farhat's Method Program              */
/*                              Non-Mixed Meshes                              */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include "e_pltfrm.h"

#ifndef __E_T800SAL__
#include <stdio.h>
#endif

#ifdef __E_T800__
#include <chan.h>
#endif

#include <time.h>
#include <e_lib.h>

#include "farhat.h"
#include "mainfnct.h"


/* ////////////////////////////////////////////////////////////////////////// */
/* main: To run the Farhat's method program. */


int main(int argc, char *argv[])
{
   int NSubdomains;             /* number of subdomains to be generated */
   int NStartMinNode;           /* which is the starting node */
   int LinkGrab;                /* whether we want the link elements to grab
                                   with others */
   int *ElemsPerSubdomain;      /* No of elements per subdomain */

   time_t StartTime;            /* partitioning start time */
   time_t StopTime;             /* partitioning stop time */

   char *DomFile;               /* output domain decomposition file */
   char *DomStatsFile;          /* output decomposition statistics file */

   MESH *mesh;                  /* MESH structure */


   if (!Input(&mesh, &NSubdomains, &NStartMinNode, 
              &LinkGrab, &ElemsPerSubdomain,
              &DomFile, &DomStatsFile, argc, argv))
      return 1;

   time(&StartTime);

   if (!farhat(mesh, NSubdomains, NStartMinNode, LinkGrab, ElemsPerSubdomain))
   {
      e_FreeMeshData(mesh);
      e_free(DomFile);
      e_free(DomStatsFile);
      return 1;
   }

   time(&StopTime);

   if (   !e_PutDecompData(mesh,DomFile)
       || !Statistics(mesh,StartTime,StopTime,DomStatsFile))
   {
      e_FreeMeshData(mesh);
      e_free(DomFile);
      e_free(DomStatsFile);
      return 1;
   }

   e_FreeMeshData(mesh);
   e_free(DomFile);
   e_free(DomStatsFile);

   return 0;
}
