/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                                Statistics()                                */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include "e_pltfrm.h"

#ifndef __E_T800SAL__
#include <stdio.h>
#endif

#ifdef __E_T800__
#include <chan.h>
#endif

#include <string.h>
#include <time.h>
#include <e_lib.h>


/* ////////////////////////////////////////////////////////////////////////// */
/* Statistics: To output statistics on the partitioning. */

/* The return value is 1 if the operation is successful, else 0. */


int Statistics(MESH *mesh, time_t StartTime, time_t StopTime,
                                           char *DomStatsFile)
{
   int i;
   int FullPathNameLength;
   int *NElements;
   int NInterSubDomNodes;
   int WeightedNInterSubDomNodes;

   double AbsError,SumAbsError;
   double RelError,SumRelError;
   double TargetNElements;

   char extension[4] = "sts";
   char *FullPathName;

   FILE *fp;


   FullPathNameLength = strlen(DomStatsFile) + strlen("sts") + 1;

   FullPathName = NULL;
   NElements = NULL;

   if (   !e_GetNInterSubDomNodes(mesh,&NInterSubDomNodes,
                                  &WeightedNInterSubDomNodes)
       || (FullPathName = e_alloc_c_vect(FullPathNameLength + 1)) == NULL
       || (NElements = e_alloc_i_vect(mesh->Decomp.NSubDoms)) == NULL    )
   {
      e_free(FullPathName);
      e_free(NElements);
      return 0;
   }

   strcpy(FullPathName,DomStatsFile);
   strcat(FullPathName,".");
   strcat(FullPathName,extension);

   if ((fp = fopen(FullPathName,"w")) == NULL)
   {
      e_FileError(FullPathName,"cannot open file");
      e_free(FullPathName);
      e_free(NElements);
      return 0;
   }

   printf("\n                       partitioning time: ");
   printf("%ds\n",StopTime - StartTime);
   printf("         number of inter-subdomain nodes: ");
   printf("%d\n",NInterSubDomNodes);
   printf("weighted number of inter-subdomain nodes: ");
   printf("%d\n",WeightedNInterSubDomNodes);
   printf("\nSubdomain  NElements  Abs Error  Rel Error\n");

   fprintf(fp,"                       partitioning time: ");
   fprintf(fp,"%ds\n",StopTime - StartTime);
   fprintf(fp,"         number of inter-subdomain nodes: ");
   fprintf(fp,"%d\n",NInterSubDomNodes);
   fprintf(fp,"weighted number of inter-subdomain nodes: ");
   fprintf(fp,"%d\n",WeightedNInterSubDomNodes);
   fprintf(fp,"\nSubdomain  NElements  Abs Error  Rel Error\n");

   TargetNElements = (double)(mesh->TotalNElems)/mesh->Decomp.NSubDoms;

   for (i = 0; i < mesh->TotalNElems; i++)
      NElements[mesh->Decomp.SubDoms[i] - 1]++;

   SumAbsError = 0.0;
   SumRelError = 0.0;

   for (i = 0; i < mesh->Decomp.NSubDoms; i++)
   {
      AbsError = NElements[i] - TargetNElements;
      RelError = 100.0*(NElements[i]/TargetNElements - 1.0);
      SumAbsError += E_ABS(AbsError);
      SumRelError += E_ABS(RelError);
      printf("   %2d        %4d      %5.1f     %5.1f%%\n",
                   i + 1,NElements[i],AbsError,RelError);
      fprintf(fp,"   %2d        %4d      %5.1f     %5.1f%%\n",
                       i + 1,NElements[i],AbsError,RelError);
   }

   printf("\n Abs Sum     %4d      %5.1f     %5.1f%%\n",
            mesh->TotalNElems,SumAbsError,SumRelError);
   fprintf(fp,"\n Abs Sum     %4d      %5.1f     %5.1f%%\n",
                mesh->TotalNElems,SumAbsError,SumRelError);

   e_free(FullPathName);
   e_free(NElements);

   return 1;
}
