#include <stdio.h>
#include <e_lib.h>
#include "graph.h"

static int adj_sort_function(const void *a, const void *b)
{
  if( *((int *) a) > *((int *) b) )
    return(-1);
  if( *((int *) a) < *((int *) b) )
    return(1);
  return(0);
}


void SetCommAdjacency(MESH *mesh, int *degree, int **adjacency,
                      int *no_node_elem, int **node_elem)
{
  int i, j_elem, i_elem, i_deg;
  int elem, other_elem;

  for(i = 0; i < mesh->NNodes; i++)
  {

    /* at the node go through all elements */
    for(j_elem = 0; j_elem < no_node_elem[i]; j_elem++)
    {
      /* select the element */
      elem = node_elem[i][j_elem];
      for(i_elem = j_elem+1; i_elem < no_node_elem[i]; i_elem++)
      {
        /* select the other element */
        other_elem = node_elem[i][i_elem];
        /* put the newly selected element in the adjacency, 
           if it is not there */
        for(i_deg = 0; i_deg < degree[elem-1]; i_deg++)
        {
          if(adjacency[elem-1][i_deg] == other_elem)
            break;
        }
        if(i_deg == degree[elem-1])
        {
          adjacency[elem-1] = e_incr_i_vect(i_deg, adjacency[elem-1], 1);
          adjacency[elem-1][i_deg] = other_elem;
          degree[elem-1]++;
        }
        /* put the first selected element in the other's adjacency list,
           if it is not there */
        for(i_deg = 0; i_deg < degree[other_elem-1]; i_deg++)
        {
          if(adjacency[other_elem-1][i_deg] == elem)
            break;
        }
        if(i_deg == degree[other_elem-1])
        {
          adjacency[other_elem-1] = 
            e_incr_i_vect(i_deg, adjacency[other_elem-1], 1);
          adjacency[other_elem-1][i_deg] = elem;
          degree[other_elem-1]++;
        }
      }
    }
  }


  for(i = 0; i < mesh->TotalNElems; i++)
  {
    qsort(adjacency[i], degree[i], sizeof(int), adj_sort_function);
  }


}
