/* incremental update + total allocation */
/* with sorting it gives exactly the same as the original algorithm */

/* It is not efficient since there are more memory allocations it 
is necessary. The InterSubdomainElements and relevant
structures can be approximated by an upper bound, like the max. number of
elements in the bigger subdomain. */



/* It does not try to swap the dummy elements in one round, but
according the agressivity constant. This enhances the subdomains,
because there is a time when the required dummies are swaped
and the algorithm only enhances the current boundary. */





/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                               KernighanLin()                               */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <e_lib.h>

#include "kerlin.h"

#define NCOSTS_MAX  3           /* number of maximum costs stored */
#define HIGH_COST   2           /* high external cost for dummy elements */

/* Specifies the number to divide */

#define DUMMY_AGRESSIVITY 2

/* This switch automatically controls the swap of dummy elements.
   If this is defined, then in every turn only the number of interface
   elements divided by two, dummy elements will be selected for swap. */

#define __AUTOMATIC_AGRESSIVITY__ 


/* This switch determines, how many times the origram should shuffle
   the boundary elements when there was an unsuccessfull interchange.
   By shuffling, the selection order is changed increasing the possibility
   of a successfull interchange. Only in case of __AUTOMATIC_AGRESSIVITY__ */

#define MAX_ATTEMPT 10



typedef struct
{
  int elem;
  int cost;
} KL_SORTING;




/* ////////////////////////////////////////////////////////////////////////// */
/* KernighanLin: To perform a Kernighan and Lin local modification to an
   inter-subdomain boundary of two subdomains. The dummy elements should move 
   over from subdomain2 to subdomain 1 */

/* The return value is 1 if the operations is successful, else 0. */


int KernighanLin(int NElems, int NDummyElems, int subdomain1, int subdomain2,
                 int *Degree, int **Adjacency, int *Subdomains)
{
   int   k,i1,i2,i, attempt = 0;
   int   kPartialSumMax;

   int  *BoundMask;
   int   NInterSubdomainElements;
   int  *InterSubdomainElements, *WorkingBoundElements;
   int  *NetExternalCosts, *WorkingCosts;
   int   NExchangePairs;          /* number of element exchange pairs */
   int **ExchangeElements;	/* element exchange pairs */
   int  *ExchangeGains;          /* element exchange losses */
   int   PartialSum;              /* partial sum of positive exchange losses */
   int   PartialSumMax;           /* current maximum "PartialSum" */
   int   dummy_exchange;
   int   processed_dummys;
   int   alloc_number;
   int  *new_subdom, *new_degree;
   int **new_adjacency;

   if(NDummyElems > 0)
   {
     new_degree = e_alloc_i_vect(NElems + NDummyElems);
     memcpy((void *)new_degree, (void *)Degree, NElems*sizeof(int) );
     for(i = 0; i < NDummyElems; i++)
       new_degree[i+NElems] = 0;

     new_adjacency = (int **)e_alloc_vp_vect(NElems + NDummyElems);
     memcpy((void **)new_adjacency, (void **)Adjacency, NElems*sizeof(int *) );
     for(i = 0; i < NDummyElems; i++)
       new_adjacency[i+NElems] = NULL;

     new_subdom = e_alloc_i_vect(NElems + NDummyElems);
     memcpy((void *)new_subdom, (void *)Subdomains, NElems*sizeof(int) );
     for(i = 0; i < NDummyElems; i++)
       new_subdom[i+NElems] = subdomain2;
   }
   else
   {
     new_degree = Degree;
     new_adjacency = Adjacency;
     new_subdom = Subdomains;
   }

   alloc_number = NElems + NDummyElems;
   InterSubdomainElements = e_alloc_i_vect(alloc_number);
   NetExternalCosts       = e_alloc_i_vect(alloc_number);
   WorkingBoundElements   = e_alloc_i_vect(alloc_number);
   WorkingCosts           = e_alloc_i_vect(alloc_number);
   BoundMask              = e_alloc_i_vect(alloc_number);

   alloc_number /= 2;
   ExchangeElements  = e_alloc_cont_i_matr(alloc_number,2);
   ExchangeGains     = e_alloc_i_vect(alloc_number);

   if(!SetInitialCost(NElems, new_degree, new_adjacency,
                      subdomain1, subdomain2, new_subdom,
                      &NInterSubdomainElements,
                      InterSubdomainElements,
                      NetExternalCosts,
                      BoundMask                          ))
   {
     return 0;
   }


   dummy_exchange = 0;
   processed_dummys = 0;

   srand(5457);

   do
   {
/*
      printf("\n%d interface elements between %d and %d", 
             NInterSubdomainElements, subdomain1, subdomain2);
*/

      /* create the working sets */
      memcpy((void *)WorkingBoundElements, 
             (void *)InterSubdomainElements,
             NInterSubdomainElements * sizeof(int) );
      memcpy((void *)WorkingCosts, 
             (void *)NetExternalCosts,
             NInterSubdomainElements * sizeof(int) );

      if (!SetExchangePairs(NElems, processed_dummys, 
                            new_degree, new_adjacency, 
                            subdomain1, subdomain2, new_subdom,
                            NInterSubdomainElements,
                            WorkingBoundElements, WorkingCosts,
                            &NExchangePairs, ExchangeElements, 
                            ExchangeGains                       ))
      {
         return 0;
      }

      PartialSum = 0;
      PartialSumMax = -INT_MAX;

      for (k = 0; k < NExchangePairs; k++)
      {
         PartialSum += ExchangeGains[k];

         if (PartialSumMax < PartialSum)
         {
            kPartialSumMax = k;
            PartialSumMax = PartialSum;
         }
      }

      if (PartialSumMax > 0)
      {
/*  printf("\n-----"); */
         for (k = 0; k <= kPartialSumMax; k++)
         {
/*  printf("\n%d - %d", ExchangeElements[k][0], ExchangeElements[k][1]); */
            i1 = ExchangeElements[k][0] - 1;
            if(i1 >= NElems)
            {
              dummy_exchange++;
              e_GeneralWarning("Dummy is moving back !");
            }
            i2 = ExchangeElements[k][1] - 1;
            if(i2 >= NElems)
              dummy_exchange++;
            new_subdom[i1] = subdomain2;
            new_subdom[i2] = subdomain1;
         }
         UpdateBoundary(NElems, new_degree, new_adjacency,
                        subdomain1, subdomain2, new_subdom,
                        &NInterSubdomainElements, 
                        InterSubdomainElements,
                        NetExternalCosts,
                        BoundMask                          );
         attempt = 0;
      }
      else
      {
#ifdef __AUTOMATIC_AGRESSIVITY__
        int increment;

        if(attempt < MAX_ATTEMPT)
        {
          BoundaryShuffle(NInterSubdomainElements,
                          InterSubdomainElements,
                          NetExternalCosts        );
          attempt++;
        }
        else
        {
          attempt = 0;
          increment = NInterSubdomainElements / DUMMY_AGRESSIVITY;
          if(increment == 0)
            increment = 1;
          if(NDummyElems > increment)
          {
            if(((processed_dummys + increment) > NDummyElems) &&
               (processed_dummys != NDummyElems)                 )
              processed_dummys = NDummyElems;
            else
              processed_dummys += increment;
          }
          else
            processed_dummys += NDummyElems;
#else
          if(NDummyElems > DUMMY_AGRESSIVITY)
            processed_dummys += NDummyElems / DUMMY_AGRESSIVITY;
          else
            processed_dummys += NDummyElems;
#endif
#ifdef __AUTOMATIC_AGRESSIVITY__
        }
#endif
      }
   }
   while ((PartialSumMax > 0) || 
          ((processed_dummys <= NDummyElems) && NDummyElems));


   if(NDummyElems > 0)
   {
/*
     printf("\n%d dummy element moved from subdom %d to subdom %d", 
            dummy_exchange, subdomain2, subdomain1);
*/
     fprintf(stderr, "*");
     e_free(new_degree);
     e_free(new_adjacency);
     memcpy((void *)Subdomains, (void *)new_subdom, NElems*sizeof(int) );
     e_free(new_subdom);
   }

   return 1;
}



/* ////////////////////////////////////////////////////////////////////////// */
/* SetInitialCost: To determine the initial cost for boundary elements        */

/* The return value is 1 if the operations is successful, else 0. */

static int SetInitialCost(int  NElems, 
                           int *Degree, int **Adjacency,
                           int  subdom1, int subdom2,
                           int *Subdomain,
                           int *NInterSubdomainElements,
                           int *InterSubdomainElements,
                           int *NetExternalCosts,
                           int *BoundMask                )
{
   int i, j, k, s, n;
   int NInterSubdomainInterfaces;

   n = 0;
   for (i = 0; i < NElems; i++)
   {
      NInterSubdomainInterfaces = 0;

      if(Subdomain[i] == subdom1)
        s = subdom2;
      else if(Subdomain[i] == subdom2)
        s = subdom1;
      else
        s = -1;

      for (j = 0; j < Degree[i]; j++)
      {
         k = Adjacency[i][j] - 1;

         if (Subdomain[k] == s)
            NInterSubdomainInterfaces++;
      }

      if (NInterSubdomainInterfaces > 0)
      {
         InterSubdomainElements[n] = i + 1;
         /* !!!!! Degree should express the number of interfaces !!!!! */
         NetExternalCosts[n] = 2*NInterSubdomainInterfaces - Degree[i];
         n++;
         BoundMask[i] = n;
      }
   }

/* If there is no boundary between the two subdomain */
   if(n == 0)
   {
     e_GeneralWarning("No boundary between the two subdomain !");
     return 0;
   }

/*
   if(n == 0)
   {
     for (i = 0; i < NElems; i++)
       if(Subdomain[i] == subdom1)
       {         
         InterSubdomainElements[n] = i + 1;
         NetExternalCosts[n] = - Degree[i];
         n++;
         BoundMask[i] = n;
         break;
       }
   }
*/

   (*NInterSubdomainElements) = n;
   return 1;
}



static int bound_sort_function(const void *a, const void *b)
{
  if( ((KL_SORTING *)a)[0].elem < ((KL_SORTING *)b)[0].elem )
    return(-1);
  if( ((KL_SORTING *)a)[0].elem > ((KL_SORTING *)b)[0].elem )
    return(1);
  return(0);
}



/* ////////////////////////////////////////////////////////////////////////// */
/* UpdateBoundary: To determine the new boundary elements 
   and recalculate costs */

static void UpdateBoundary(int NElems, int *Degree, int **Adjacency, 
                           int subdom1, int subdom2, int *Subdom,
                           int *NInterSubdomainElements, 
                           int *InterSubdomainElements,
                           int *NetExternalCosts,
                           int *BoundMask               )
{
  int i, j, jj, no_bound_elem, other_subdom, elem_other_subdom;
  int NBoundInterfaces, NInterfacesSurr;
  int elem, surr_elem, bound_elem;
  int start;

  no_bound_elem = *NInterSubdomainElements;
  for (i = 0; i < *NInterSubdomainElements; i++)
  {
    bound_elem = InterSubdomainElements[i]-1;
    if(Subdom[bound_elem] == subdom1)
      other_subdom = subdom2;
    else if(Subdom[bound_elem] == subdom2)
      other_subdom = subdom1;
    else
      other_subdom = -1;

    /* ordinary element */
    if(bound_elem < NElems)
    {
      NBoundInterfaces = 0;
      for(j = 0; j < Degree[bound_elem]; j ++)
      {
        elem = Adjacency[bound_elem][j] - 1;
        if(Subdom[elem] == other_subdom)
          NBoundInterfaces++;

        /* check surrounding elements whether they will 
        end up on the boundary, if it is already there, 
        then either we will process it later or it is a new
        boundary and no processing is necessary second time */
        if(BoundMask[elem] == 0)
        {
          NInterfacesSurr = 0;
          if(Subdom[elem] == subdom1)
            elem_other_subdom = subdom2;
          else if(Subdom[elem] == subdom2)
            elem_other_subdom = subdom1;
          else
            elem_other_subdom = -1;
          for(jj = 0; jj < Degree[elem]; jj++)
          {
            surr_elem = Adjacency[elem][jj] - 1;
            if(Subdom[surr_elem] == elem_other_subdom)
              NInterfacesSurr++;
          }
          if(NInterfacesSurr > 0)
          {
            InterSubdomainElements[no_bound_elem] = elem + 1;
            NetExternalCosts[no_bound_elem] = 2*NInterfacesSurr - Degree[elem];
            no_bound_elem++;
            BoundMask[elem] = no_bound_elem;
          }
        }
      }
      if(NBoundInterfaces > 0)
      {
        NetExternalCosts[i] = 2*NBoundInterfaces - Degree[bound_elem];
      }
      else
      {
        BoundMask[bound_elem] = 0;
        InterSubdomainElements[i] = 0;
      }
    }
    /* dummy element */
    else
    {
      /* if the dummy element in the correct subdomain, then simply
      remove it from the boundary element list */
      if(Subdom[bound_elem] == subdom1)
      {
        BoundMask[bound_elem] = 0;
        InterSubdomainElements[i] = 0;
      }
    }
  }

  /* compress the vectors, taking out the zero elements 
  (deleted boundary elements), update other structures as well */
  start = -1;
  *NInterSubdomainElements = no_bound_elem;
  for(i = 0; i < (*NInterSubdomainElements); i++)
  {
    if((InterSubdomainElements[i] == 0) && (start == -1))
    {
      start = i;
      continue;
    }
    if((InterSubdomainElements[i] != 0) && (start != -1))
    {
      memcpy((void *)(&(InterSubdomainElements[start])),
             (void *)(&(InterSubdomainElements[i])),
             ((*NInterSubdomainElements) - i) * sizeof(int) );
      memcpy((void *)(&(NetExternalCosts[start])),
             (void *)(&(NetExternalCosts[i])),
             ((*NInterSubdomainElements) - i) * sizeof(int) );
      /* adjust BoundMask index by the movement */
      for(jj = 0; jj < (*NInterSubdomainElements) - i; jj++)
      {
        bound_elem = InterSubdomainElements[start+jj] - 1;
        if(BoundMask[bound_elem] != 0)
          BoundMask[bound_elem] = start + jj + 1;
      }
      (*NInterSubdomainElements) -= (i - start);
      /* it is enough to reset it to start, since we know, that
      a non-zero value will be there, and "i" will be increased
      immediately to examine the next element */
      i = start;
      start = -1;
    }
  }
  /* there were elements with zero contents at the back of the list */
  if(start != -1)
  {
    *NInterSubdomainElements = start;
  }

/*
{
  KL_SORTING *new_bound;

  new_bound = (KL_SORTING *)e_calloc(*NInterSubdomainElements, 
                                     sizeof(KL_SORTING));

  for(i = 0; i < *NInterSubdomainElements; i++)
  {
    new_bound[i].elem = InterSubdomainElements[i];
    new_bound[i].cost = NetExternalCosts[i];
  }

  qsort(new_bound, *NInterSubdomainElements, 
        sizeof(KL_SORTING), bound_sort_function);

  for(i = 0; i < *NInterSubdomainElements; i++)
  {
    InterSubdomainElements[i] = new_bound[i].elem;
    NetExternalCosts[i] = new_bound[i].cost;
  }

  e_free(new_bound);

} */

}



/* ////////////////////////////////////////////////////////////////////////// */
/* SetExchangePairs: To determine element exchange pairs. */

/* The return value is 1 if the operations is successful, else 0. */

static int SetExchangePairs(int    NElems, int NDummyElems,
                            int   *Degree, int **Adjacency, 
                            int    subdom1, int subdom2,
                            int   *Subdomain,
                            int    NInterSubdomainElements,
                            int   *InterSubdomainElements,
                            int   *NetExternalCosts,
                            int   *NExchangePairs, int **ExchangeElements,
                            int   *ExchangeGains   )
{
   int i,j,k,l,s;
   int index;

   /* count dummy elements in the wrong subdomain */
   for(i = 0; i < NDummyElems; i++)
   {
     if(Subdomain[NElems+i] != subdom1)
     {
       NetExternalCosts[NInterSubdomainElements] = HIGH_COST;
       InterSubdomainElements[NInterSubdomainElements] = i + NElems + 1;
       NInterSubdomainElements++;
     }
   }

/*
   printf("\n-----------");
   for(i = 0; i < NInterSubdomainElements; i++)
   {
      printf("\n%d. bound elem: %d cost %d", i+1, 
             InterSubdomainElements[i], NetExternalCosts[i]);
   }
*/

   (*NExchangePairs) = NInterSubdomainElements / 2;

   for (l = 0; l < *NExchangePairs; l++)
   {

      if (!SetNextExchangePair(Degree,Adjacency,
                               Subdomain, subdom1,
                               NInterSubdomainElements, 
                               InterSubdomainElements,
                               NetExternalCosts,
                               l, ExchangeElements,
                               ExchangeGains           ))
      {
         *NExchangePairs = l;
         break;
      }

      if (!ReduceSet(ExchangeElements[l], &NInterSubdomainElements,
                     InterSubdomainElements, NetExternalCosts))
      {
         return 0;
      }


   }

   return 1;
}



/* ////////////////////////////////////////////////////////////////////////// */
/* SetNextExchangePair: To determine a (potential) exchange pair and the
   reduction in external cost as a result of the exchange.                    */

/* The return value is 1 is an exchange pair is found, else 0. */


static int SetNextExchangePair(int  *Degree, int **Adjacency,
                               int  *Subdomain, int   subdom1,
                               int   NInterSubdomainElements,
                               int  *InterSubdomainElements,
                               int  *NetExternalCosts,
                               int   ExchangePairsIndex,
                               int **ExchangeElements,
                               int  *ExchangeGains             )
{
   int i,k,s;
   int iCostMax,kCostMax;
   int Cost,CostMax;
   int DeltaCost[2];
   int ElementIndex1,ElementIndex2;
   int MaxNetExternalCost[2][NCOSTS_MAX];
   int MaxNetExternalCostIndex[2][NCOSTS_MAX];

   for (k = 0; k < NCOSTS_MAX; k++)
   {
      MaxNetExternalCostIndex[0][k] = -1;
      MaxNetExternalCostIndex[1][k] = -1;
      MaxNetExternalCost[0][k] = -INT_MAX;
      MaxNetExternalCost[1][k] = -INT_MAX;
   }

   /* for the first time we do not have to do adjustment in the costs */
   if(ExchangePairsIndex == 0)
   {
     /* search for the max gain */
     for (i = 0; i < NInterSubdomainElements; i++)
     {
       if(Subdomain[InterSubdomainElements[i] - 1] == subdom1) 
         s = 0; 
       else 
         s = 1;

       if (MaxNetExternalCost[s][0] < NetExternalCosts[i])
       {
         for (k = NCOSTS_MAX - 1; k >= 1; k--)
         {
           MaxNetExternalCostIndex[s][k]
             = MaxNetExternalCostIndex[s][k - 1];
           MaxNetExternalCost[s][k]
             = MaxNetExternalCost[s][k - 1];
         }

         MaxNetExternalCostIndex[s][0] = i;
         MaxNetExternalCost[s][0] = NetExternalCosts[i];
       }
     }
   }
   else
   {
     for (i = 0; i < NInterSubdomainElements; i++)
     {
        if (ElementsAreAdjacent(Degree, Adjacency,
                                ExchangeElements[ExchangePairsIndex-1][0],
                                InterSubdomainElements[i]))
           DeltaCost[0] = 2;
        else
           DeltaCost[0] = 0;

        if (ElementsAreAdjacent(Degree, Adjacency,
                                ExchangeElements[ExchangePairsIndex-1][1],
                                InterSubdomainElements[i]))
           DeltaCost[1] = -2;
        else
           DeltaCost[1] = 0;

        if (Subdomain[InterSubdomainElements[i] - 1] == subdom1)
        {
           NetExternalCosts[i] += DeltaCost[0] + DeltaCost[1];
           s = 0;
        }
        else
        {
           NetExternalCosts[i] -= DeltaCost[0] + DeltaCost[1];
           s = 1;
        }

        if (MaxNetExternalCost[s][0] < NetExternalCosts[i])
        {
          for (k = NCOSTS_MAX - 1; k >= 1; k--)
          {
            MaxNetExternalCostIndex[s][k]
              = MaxNetExternalCostIndex[s][k - 1];
            MaxNetExternalCost[s][k]
              = MaxNetExternalCost[s][k - 1];
          }

          MaxNetExternalCostIndex[s][0] = i;
          MaxNetExternalCost[s][0] = NetExternalCosts[i];
        }
     }
   }

   CostMax = -INT_MAX;
   iCostMax = -1;
   kCostMax = -1;

   for (i = 0; i < NCOSTS_MAX; i++)
   {
      if (MaxNetExternalCostIndex[0][i] >= 0)
      {
         ElementIndex1
            = InterSubdomainElements[MaxNetExternalCostIndex[0][i]];

         for (k = 0; k < NCOSTS_MAX; k++)
         {
            if (MaxNetExternalCostIndex[1][k] >= 0)
            {
               ElementIndex2
                  = InterSubdomainElements[MaxNetExternalCostIndex[1][k]];
               Cost = MaxNetExternalCost[0][i] + MaxNetExternalCost[1][k];

               if (ElementsAreAdjacent(Degree,Adjacency,
                                       ElementIndex1,ElementIndex2))
               {
                  Cost -= 2;
               }

               if (CostMax < Cost)
               {
                  iCostMax = i;
                  kCostMax = k;
                  CostMax = Cost;
               }
            }
         }
      }
   }

   if (   iCostMax >= 0
       && kCostMax >= 0
       && MaxNetExternalCostIndex[0][iCostMax] >= 0
       && MaxNetExternalCostIndex[1][kCostMax] >= 0)
   {
      ExchangeElements[ExchangePairsIndex][0]
         = InterSubdomainElements[MaxNetExternalCostIndex[0][iCostMax]];

      ExchangeElements[ExchangePairsIndex][1]
         = InterSubdomainElements[MaxNetExternalCostIndex[1][kCostMax]];

      ExchangeGains[ExchangePairsIndex] = CostMax;

      return 1;
   }
   else
      return 0;
}




/* ////////////////////////////////////////////////////////////////////////// */
/* ReduceSet: To extract the two current exchange elements from the set
   of inter-subdomain elements.                                               */

/* The return value is 1 if the operations is successful, else 0. */


static int ReduceSet(int *ExchangeElements, int *NInterSubdomainElements,
                     int *InterSubdomainElements, int *NetExternalCosts  )
{
   int i,k;

   k = 0;

   for (i = 0; i < *NInterSubdomainElements; i++)
   {
      if (   (InterSubdomainElements[i] != ExchangeElements[0])
          && (InterSubdomainElements[i] != ExchangeElements[1]))
      {
         InterSubdomainElements[k] = InterSubdomainElements[i];
         NetExternalCosts[k] = NetExternalCosts[i];
         k++;
      }
   }
   *NInterSubdomainElements -= 2;

   return 1;
}





/* ////////////////////////////////////////////////////////////////////////// */
/* ElementsAreAdjacent: To return whether two elements are adjacent. */

static int ElementsAreAdjacent(int *Degree, int **Adjacency,
                               int ElementIndex1, int ElementIndex2)
{
   int j,k;


   k = ElementIndex1 - 1;

   for (j = 0; j < Degree[k]; j++)
   {
      if (Adjacency[k][j] == ElementIndex2)
         return 1;
   }

   return 0;
}

/* ////////////////////////////////////////////////////////////////////////// */

double p_rand()
{
  return (double) rand() / (double) RAND_MAX;
}

int p_rnd(int low, int high)
{
  if (low >= high)
    return low;
  else
  {
    return (int)(p_rand() * ((double)(high) - (double)(low))) + low;
  }
}


void BoundaryShuffle(int  NInterSubdomainElements,
                     int *InterSubdomainElements,
                     int *NetExternalCosts         )
{
  int i, z, temp;

  for(i = NInterSubdomainElements - 1; i > 0; i--)
  {
    z = p_rnd(0, i);

    temp = InterSubdomainElements[i];
    InterSubdomainElements[i] = InterSubdomainElements[z];
    InterSubdomainElements[z] = temp;

    temp = NetExternalCosts[i];
    NetExternalCosts[i] = NetExternalCosts[z];
    NetExternalCosts[z] = temp;
  }

}


