/* if two partitions looses connection during the shift of the boundaries
it recalculates the subdom-graph and the diffusion */

/* the diffusion algorithm has been changed to use double values
and at the end it rounds it to int. Now it works, but need some clean up
in the code !! */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "e_lib.h"
#include "graph.h"

/* whether the diffusion should be carried out in some order */

#define ORDER_DIFFUSION 1


/* diffusion on the whole front or one subdomain finished after each
other */

#define TOTAL_FRONT 1


/* when it works on the whole front of a subdomain, how many element 
should be moved per round */
#define ELEM_TO_MOVE	 10
/* connection between subdomains with less this many nodes will not
be considered as connected */
#define NO_SHARED_NODES  5

int       **GV_node_elem;
int        *GV_no_node_elem;
int       **GV_adjacency;
int        *GV_degree;


int Create_Subdomain_Graph(MESH *mesh, 
                           int *SubdomGraphDegree, int **SubdomGraph);
void Diffuse_Elements(MESH *mesh, int *ElemsPerSubdomain,
                      int *SubdomGraphDegree, int **SubdomGraph,
                      int **ElemsToMove);
void partition(MESH *mesh, int *Degree, int **Adjacency, int *Subdomains);
void diffusion_order(int NSubdom, int *SubdomGraphDegree,
                     int **SubdomGraph, int **ElemsToMove,
                     int *order);


main(int argc, char *argv[])
{
  char        GV_filename[E_LINE_MAX];
  char        line[E_LINE_MAX];
  char       *check;
  MESH       *GV_mesh;

  strcpy(GV_filename, "");
  if(argc > 1)
  {
    strcpy(GV_filename, argv[1]);
  }

  if (!strlen(GV_filename))
  {
    printf("\n\nType in the input filename (.mdf) !");
    printf("\nGive no extension : ");
    check = fgets(line, E_LINE_MAX, stdin);
    if(   (check == NULL)
       || (sscanf(line, "%s", &GV_filename) != 1))
      e_GeneralError("Can not read filename!");
  }

  e_GetMeshData(&GV_mesh, GV_filename);
  e_GetDecompData(GV_mesh, GV_filename);

  if(GV_mesh->TotalNElems != GV_mesh->NElemsLink1 +
                             GV_mesh->NElemsLink2 +
                             GV_mesh->NElemsLink3 +
                             GV_mesh->NElemsLink4 +
                             GV_mesh->NElemsLink5 +
                             GV_mesh->NElemsTriang3 +
                             GV_mesh->NElemsTriang4 +
                             GV_mesh->NElemsTriang5   )
    e_GeneralError("Mesh contains non-supported element !");


  GV_node_elem = e_init_frag_i_matr(GV_mesh->NNodes);
  GV_no_node_elem = e_alloc_i_vect(GV_mesh->NNodes);

  e_InitMeshPointElemConnection(GV_mesh, GV_no_node_elem, GV_node_elem);

  GV_degree = e_alloc_i_vect(GV_mesh->TotalNElems);
  GV_adjacency = e_init_frag_i_matr(GV_mesh->TotalNElems);

#ifdef __CIRC_GRAPH__
  SetCircAdjacency(GV_mesh, GV_degree, GV_adjacency,
                   GV_no_node_elem, GV_node_elem);
#endif

#ifdef __DUAL_GRAPH__
  SetDualAdjacency(GV_mesh, GV_degree, GV_adjacency,
                   GV_no_node_elem, GV_node_elem);
#endif

#ifdef __COMM_GRAPH__
  SetCommAdjacency(GV_mesh, GV_degree, GV_adjacency, 
                   GV_no_node_elem, GV_node_elem);
#endif


  partition(GV_mesh, GV_degree, GV_adjacency, 
                   GV_mesh->Decomp.SubDoms);

  strcat(GV_filename, "out");
  e_PutDecompData(GV_mesh, GV_filename);


}

/* ////////////////////////////////////////////////////////////////////////// */
/* Create_Subdomain_Graph: To create the connectivity graph of the subdomains.

/* The return value is 1 if the operations is successful, else 0. */

int Create_Subdomain_Graph(MESH *mesh, 
                           int *SubdomGraphDegree, int **SubdomGraph)
{
  int    NInterSubDomNodes;
  int   *InterSubDomNodes;
  int   *NNodeSubDoms;
  int  **NodeSubDoms;

  int    i,j,k,jj,s1;
  int  **no_subdom_graph_ref;

  e_GetInterSubDomNodes(mesh, &NInterSubDomNodes, 
                        &InterSubDomNodes, &NNodeSubDoms, 
                        &NodeSubDoms);

  no_subdom_graph_ref = e_alloc_cont_i_matr(mesh->Decomp.NSubDoms,
                                            mesh->Decomp.NSubDoms-1);

  /* determine which subdomain is adjacent to which other,
     this information comes from the shared meshpoints,
     every subdomain stores the connection to the subdomains
     with higher ID number only, therefore connections are 
     not replicated */
  for(i = 0; i < NInterSubDomNodes; i++)
  {
    for(j = 0; j < NNodeSubDoms[i]; j++)
    {
      s1 = NodeSubDoms[i][j] - 1;
      for(jj = j+1; jj < NNodeSubDoms[i]; jj++)
      {
        for(k = 0; k < SubdomGraphDegree[s1]; k++)
        {
          if(SubdomGraph[s1][k] == NodeSubDoms[i][jj])
          {
            no_subdom_graph_ref[s1][k]++;
            break;
          }
        }
        if (k == SubdomGraphDegree[s1])
        {
          if(NodeSubDoms[i][j] < NodeSubDoms[i][jj])
          {
            SubdomGraph[s1][k] = NodeSubDoms[i][jj];
            SubdomGraphDegree[s1]++;
            no_subdom_graph_ref[s1][k]++;
          }
        }
      }
    }
  }

  e_FreeInterSubDomNodes(NInterSubDomNodes, InterSubDomNodes,
                         NNodeSubDoms, NodeSubDoms);

  /* if a subdomain shares only a small number meshpoint with another one
     that is not good enough, remove the connection between the 
     two subdomains */
  for(i = 0; i < mesh->Decomp.NSubDoms; i++)
  {
    for(j = 0; j < SubdomGraphDegree[i]; j++)
    {
      if(no_subdom_graph_ref[i][j] < NO_SHARED_NODES)
      {
        for(k = j+1; k < SubdomGraphDegree[i]; k++)
        {
          SubdomGraph[i][k-1]         = SubdomGraph[i][k];
          no_subdom_graph_ref[i][k-1] = no_subdom_graph_ref[i][k];
        }
        SubdomGraphDegree[i]--;
        j--;
      }
    }

/*
    if((SubdomGraphDegree[i] == 0) && (mesh->Decomp.NSubDoms > 2))
    {
      fprintf(stderr, "Subdomain %d became disconnected !", i+1);
      e_free_cont_i_matr(mesh->Decomp.NSubDoms, no_subdom_graph_ref);
      return 0;
    }
*/

  }

  e_free_cont_i_matr(mesh->Decomp.NSubDoms, no_subdom_graph_ref);

  return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* Diffuse_Elements: To calculate how many elements should be moved between
   the subdomains. */

void Diffuse_Elements(MESH *mesh, int *ElemsPerSubdomain,
                      int *SubdomGraphDegree, int **SubdomGraph,
                      int **ElemsToMove)
{
  int    i, j;
  double ddiff, frac_diff;
  int    idiff, s, NotBalanced;
  int   *MaxDegree;
  int    TotalMaxDegree;
  int    Deviation;

  double    AvgElemsPerSubdom;
  double  **round_move;
  double   *ElemsInSubdom;
  double  **ElemsMove;

  ElemsMove = e_alloc_cont_d_matr(mesh->Decomp.NSubDoms,
                                  mesh->Decomp.NSubDoms-1);
  ElemsInSubdom = e_alloc_d_vect(mesh->Decomp.NSubDoms);
  for(i = 0; i < mesh->Decomp.NSubDoms; i++)
    ElemsInSubdom[i] = (double)ElemsPerSubdomain[i];

  round_move = e_alloc_cont_d_matr(mesh->Decomp.NSubDoms,
                                   mesh->Decomp.NSubDoms-1);

  MaxDegree = e_alloc_i_vect(mesh->Decomp.NSubDoms);
  TotalMaxDegree = -1;

  /* determine the max. degree of the vertices of the subdomain graph
     and the number of elements of the balanced state */
  AvgElemsPerSubdom = 0;
  for(i = 0; i < mesh->Decomp.NSubDoms; i++)
  {
    for(j = 0; j < SubdomGraphDegree[i]; j++)
    {
      s = SubdomGraph[i][j] - 1;
      MaxDegree[i]++;
      MaxDegree[s]++;
    }
    AvgElemsPerSubdom += ElemsInSubdom[i];
  }
  AvgElemsPerSubdom /= (double)mesh->Decomp.NSubDoms;

  for(i = 0; i < mesh->Decomp.NSubDoms; i++)
  {
    if(TotalMaxDegree < MaxDegree[i])
      TotalMaxDegree = MaxDegree[i];
  }
  if((TotalMaxDegree <= 0) && (mesh->Decomp.NSubDoms > 2))
    e_GeneralError("Max. degree is not valid !");

  /* calculate the number of elements to move, 
     positive means to send, negative means to receive */
  do
  {
    NotBalanced = 0;
    for(i = 0; i < mesh->Decomp.NSubDoms; i++)
    {
      for(j = 0; j < SubdomGraphDegree[i]; j++)
      {
        s = SubdomGraph[i][j] - 1;
        ddiff = (ElemsInSubdom[i] - ElemsInSubdom[s]) / 
                (double)(MaxDegree[i] + 1);
/*
                (double)(TotalMaxDegree + 1);
*/
        ElemsMove[i][j] += ddiff;
        round_move[i][j]  += ddiff;
      }
    }
    for(i = 0; i < mesh->Decomp.NSubDoms; i++)
    {
      for(j = 0; j < SubdomGraphDegree[i]; j++)
      {
        s = SubdomGraph[i][j] - 1;
        ElemsInSubdom[i] -= round_move[i][j];
        ElemsInSubdom[s] += round_move[i][j];
        round_move[i][j] = 0;
      }
      if( E_ABS(ElemsInSubdom[i] - AvgElemsPerSubdom) > 1 )
        NotBalanced = 1;
/*      printf("\n%d subdom %d elem", i+1, ElemsPerSubdomain[i]); */
    }
/*    printf("\n\n"); */
  } while(NotBalanced);

  for(i = 0; i < mesh->Decomp.NSubDoms; i++)
    for(j = 0; j < mesh->Decomp.NSubDoms-1; j++)
      ElemsToMove[i][j] = (int)ElemsMove[i][j];

  e_free_cont_d_matr(mesh->Decomp.NSubDoms, round_move);
  e_free(MaxDegree);
}


void partition(MESH *mesh, int *Degree, int **Adjacency, int *Subdomains)
{
  int  *SubdomGraphDegree;
  int **SubdomGraph;
  int **ElemsToMove;
  int  *ElemsPerSubdomain;
  int  *ReqElemsPerSubdomain;
  int  *RealElemsPerSubdomain;
  char  line[E_LINE_MAX];
  char *check;
  int   i, j, s1, s2, s, EqualElems;
  short recalc;
  int  *order, *stat;
  int   index;

  ReqElemsPerSubdomain = e_alloc_i_vect(mesh->Decomp.NSubDoms);
  for(i = 0; i < mesh->Decomp.NSubDoms; i++)
  {
    printf("\nNumber of elements in %d subdomain : ", i+1);
    check = fgets(line, E_LINE_MAX, stdin);
    if(   (check == NULL)
       || (sscanf(line, "%d", &(ReqElemsPerSubdomain[i])) != 1))
      e_GeneralError("Can not read the number of elements!");
  }

  do
  {
    recalc = 0;
    RealElemsPerSubdomain = e_alloc_i_vect(mesh->Decomp.NSubDoms);
    for(i = 0; i < mesh->TotalNElems; i++)
      RealElemsPerSubdomain[mesh->Decomp.SubDoms[i]-1]++;

    ElemsPerSubdomain = e_alloc_i_vect(mesh->Decomp.NSubDoms);
    EqualElems = mesh->TotalNElems / mesh->Decomp.NSubDoms;
    for(i = 0; i < mesh->Decomp.NSubDoms; i++)
    {
      ElemsPerSubdomain[i] = EqualElems + (ReqElemsPerSubdomain[i] - 
                                           RealElemsPerSubdomain[i]  );
    }

    SubdomGraphDegree = e_alloc_i_vect(mesh->Decomp.NSubDoms);
    SubdomGraph       = e_alloc_cont_i_matr(mesh->Decomp.NSubDoms, 
                                            mesh->Decomp.NSubDoms-1);
    if(!Create_Subdomain_Graph(mesh, SubdomGraphDegree, SubdomGraph))
      exit(-1);

    fprintf(stderr, "\nSubdomain graph is created !");

    ElemsToMove = e_alloc_cont_i_matr(mesh->Decomp.NSubDoms,
                                      mesh->Decomp.NSubDoms-1);
    Diffuse_Elements(mesh, ElemsPerSubdomain, SubdomGraphDegree,
                     SubdomGraph, ElemsToMove);

    fprintf(stderr, "\nDiffusion calculation is finished !");

    for(i = 0; i < mesh->Decomp.NSubDoms; i++)
    {
      for(j = 0; j < SubdomGraphDegree[i]; j++)
      {
        printf("\n%d subdom will move/receive %d elements to/from %d subdom", 
               i+1, ElemsToMove[i][j], SubdomGraph[i][j]);
      }
    }
    printf("\n\n");

#ifdef ORDER_DIFFUSION
    order = e_alloc_i_vect(mesh->Decomp.NSubDoms);
    diffusion_order(mesh->Decomp.NSubDoms, SubdomGraphDegree,
                    SubdomGraph, ElemsToMove,
                    order);

    fprintf(stderr, "\nDiffusion order has been determined !\n");

    fflush(stdout);
#endif

    stat = e_alloc_i_vect(mesh->Decomp.NSubDoms);

    for(i = 0; i < mesh->Decomp.NSubDoms; i++)
    {
      int expand;
      char save[E_LINE_MAX];
      
#ifdef ORDER_DIFFUSION
      index = order[i];
#else
      index = i;
#endif
      stat[index]++;
      /* handle all boundary of the subdomain together, until all
      required movement has been finished */
      do
      {
        expand = 0;
        /* the whole front of the subdomain is handled, bit-by-bit */
        for(j = 0; j < SubdomGraphDegree[index]; j++)
        {
          /* if there is something to move and that connection has not been
          finished yet */
          if(ElemsToMove[index][j] != 0 && !stat[SubdomGraph[index][j]-1])
          {
            int move;

            expand = 1;

            s1 = index+1;
            s2 = SubdomGraph[index][j];
            if(ElemsToMove[index][j] > 0)
            {
              s  = s1;
              s1 = s2;
              s2 = s;
            }
            
            /* determine the number of elements to move on that boundary */
#ifdef TOTAL_FRONT
            if(ELEM_TO_MOVE > E_ABS(ElemsToMove[index][j]))
              move = E_ABS(ElemsToMove[index][j]);
            else
              move = ELEM_TO_MOVE;
#else
            move = E_ABS(ElemsToMove[index][j]);
#endif
            
            /* do Kernighan-Lin */
            if(!KernighanLin(mesh->TotalNElems, move, 
                             s1, s2, Degree, Adjacency, Subdomains))
            {
              recalc = 1;
              break;
            }
            
            /* reflect the real number of elements which should be moved */
            ElemsToMove[index][j] += move * (ElemsToMove[index][j] > 0 ? -1 : 1);
          }
        }
        if(recalc)
          break;
      }
      while(expand);
/*
      sprintf(save, "test%d", i+1);
      e_PutDecompData(mesh, save);
*/

      if(recalc)
        break;
    }

    e_free(stat);
    e_free(RealElemsPerSubdomain);
    e_free(ElemsPerSubdomain);
    e_free(SubdomGraphDegree);
    e_free_cont_i_matr(mesh->Decomp.NSubDoms, SubdomGraph);
    e_free_cont_i_matr(mesh->Decomp.NSubDoms, ElemsToMove);
  } while(recalc);

  e_free_cont_i_matr(mesh->Decomp.NSubDoms, ElemsToMove);
  e_free(SubdomGraphDegree);
  e_free_cont_i_matr(mesh->Decomp.NSubDoms, SubdomGraph);

}

/* determines the order of the diffusion */
void diffusion_order(int NSubdom, int *SubdomGraphDegree,
                     int **SubdomGraph, int **ElemsToMove,
                     int *order)
{
  int *stat;
  int i, j;
  int gr, pos, prev_pos;

  /* make the graph complete */
  for(i = 0; i < NSubdom; i++)
  {
    for(j = 0; j < SubdomGraphDegree[i]; j++)
    {
      gr = SubdomGraph[i][j]-1;
      if(gr > i)
      {
        SubdomGraph[gr][SubdomGraphDegree[gr]] = i+1;
        ElemsToMove[gr][SubdomGraphDegree[gr]] = ElemsToMove[i][j] * -1;
        SubdomGraphDegree[gr]++;
      }
    }
  }

  /* print out the subdomain graph */
  printf("\nSubdomain graph");
  for(i = 0; i < NSubdom; i++)
  {
    printf("\n");
    for(j = 0; j < SubdomGraphDegree[i]; j++)
    {
      printf("%d ", SubdomGraph[i][j]);
    }
  }

  stat = e_alloc_i_vect(NSubdom);
  pos = 0;

  while(pos < NSubdom)
  {
    prev_pos = pos;
    for(i = 0; i < NSubdom; i++)
    {
      if(stat[i])
        continue;

      for(j = 0; j < SubdomGraphDegree[i]; j++)
      {
        gr = SubdomGraph[i][j]-1;
        if(stat[gr])
          continue;
        if(ElemsToMove[i][j] > 0)
          break;
      }
      if(j == SubdomGraphDegree[i])
      {
        order[pos] = i;
        pos++;
        stat[i]++;
      }
    }
    if(prev_pos == pos)
      e_GeneralError("No sorting possible ...");
  }

  /* print out the order os diffusion */
  printf("\n\nOrder of diffusion:");
  for(i = 0; i < NSubdom; i++)
    printf("\n%d", order[i]);


  e_free(stat);
}





