#include "ssgm.h"

static MESH *SubMesh;

/* !!!!! selects a node and an angle !!!!! */
double linear_objective(long nvars, DESIGNPARAMS vars)
{
  long      selected, no_interfaces;
  double    angle, selected_x, selected_y, other_x, other_y;
  int       doms[2];
  double    fitness;
  long      i, j, left, right, subdom;

  selected = (long) vars[0].parameter;
  angle = vars[1].parameter;

  selected_x = SubMesh->MeshPointCoords[selected][0];
  selected_y = SubMesh->MeshPointCoords[selected][1];

  other_x = selected_x + 10.00 * cos(angle);
  other_y = selected_y + 10.00 * sin(angle);

/* put the strips to the left or the right of the separator, 
   except the selected */
  doms[0] = 0;
  doms[1] = 0;
  for(i = 0; i < G_master->Decomp.NSubDoms; i++)
  {
    if(G_proc_subdom == G_act_subdom[i])
    {
      if(dirnode(selected_x, selected_y, other_x, other_y,
                 G_center[i][0], G_center[i][1]))
      {
        doms[1] += G_no_elems_in_strip[i];
        G_proc_strip[i] = 2;
      }
      else
      {
        doms[0] += G_no_elems_in_strip[i];
        G_proc_strip[i] = 1;
      }
    }
    else
      G_proc_strip[i] = 0;
  }

  no_interfaces = 0;
  for(i = 0; i < G_master->Decomp.NSubDoms; i++)
  {
    if(G_proc_subdom == G_act_subdom[i])
    {
      for(j = 0; j < SubdomGraphDegree[i]; j++)
      {
        subdom = SubdomGraph[i][j]-1;
        if(   (G_act_subdom[subdom] == G_proc_subdom)
           && (G_proc_strip[subdom] != G_proc_strip[i]) )
          no_interfaces += SubdomGraphRef[i][j];
      }
    }
  }


  left = doms[0];
  right = doms[1];

  fitness = (left + right) -
            (1.00 - GA_j_gamma) * E_ABS(left - right) -
            GA_j_gamma * no_interfaces;

  if(GA_printing == 3)
  {
    printf("\nleft = %d right = %d no. of interfaces = %d >>> fitness = %f",
           left, right, no_interfaces, fitness);
    printf("\nselected strip = %d --- angle = %f", selected, angle);

  }
  return fitness;
}




int dirnode(double x1, double y1, double x2, double y2,
            double x3, double y3)
{
  double    a11, a12, a21, a22, D, c1, c2, s;

/* Function dirnode returns 1 if the  node(x3,y3) is on the
   left of the seg(1,2) else 0 */

  a11 = x2 - x1;
  a12 = y2 - y1;
  a21 = y2 - y1;
  a22 = -1.00 * (x2 - x1);

  D = a11 * a22 - a12 * a21;

  c1 = x3 - x1;
  c2 = y3 - y1;

  s = (a11 * c2 - a21 * c1) / D;

  if(s < 0.00)
    return(1);

  return(0);
}                                /*  end of function dirnode()  */








void do_linear_decomposition()
{
  long      i, j, iter, subdom_iter, i_st;
  long      no_something;
  int       no_subdom, pr;
  int      *stat_subdom;
  DESIGNPARAMS design_var;
  int     **strips, *doms;
  int       subdom, elem;
  double    angle, selected_x, selected_y;

  printf("\nLinear cutting decomposition ...\n");

  G_proc_strip = e_alloc_i_vect(G_master->Decomp.NSubDoms);
  G_act_subdom = e_alloc_i_vect(G_master->Decomp.NSubDoms);
  for(i = 0; i < G_master->Decomp.NSubDoms; i++)
    G_act_subdom[i] = 1;

/* quick and dirty way to get the same mesh with different subdomain info */

  e_GetMeshData(&G_fem1, G_projectname);
  e_SetupDecompData(G_fem1);

/*
  Create_Subdom_Struct_with_Zero(G_fem1);
  G_fem1->Decomp.NSubDoms = 1;
  for(i = 0; i < G_fem1->TotalNElems; i++)
    G_fem1->Decomp.SubDoms[i] = 1;
*/

  for(iter = 0; iter < G_total_recursions; iter++)
  {
    for(subdom_iter = 0; subdom_iter < pow(2, iter); subdom_iter++)
    {
      G_proc_subdom = subdom_iter + 1;

      printf("\n\nRecursion number = %d >>> Subdomain number = %d",
             iter + 1, subdom_iter + 1);
      printf("\n---------------------------------------------");

      e_CreateSubMesh(G_fem1, &SubMesh, "Working mesh", subdom_iter+1);
      no_something = SubMesh->NMeshPoints;

      design_var = (DESIGNPARAMS)e_calloc(2, sizeof(PARAM));

/* number of elements */
      design_var[0].minparam = 0.00;
      design_var[0].maxparam = no_something - 1;
      pr = 1;
      while(no_something > pow(2, pr))
        pr++;
      design_var[0].lparam = pr;

/* angle */
      design_var[1].minparam = -M_PI;
      design_var[1].maxparam = M_PI;
      design_var[1].lparam = 9;        /* 360 degree - min. 1 degree precision */

      srand(GA_seed);
      genetic_algorithm(GA_popsize, GA_pcross, GA_pmutation, 2, design_var,
                        GA_maxgen, GA_linear, GA_printing);

      linear_objective(2, design_var);

/* save the output of the step */
      (G_fem1->Decomp.NSubDoms)++;

      for(i_st = 0; i_st < G_master->Decomp.NSubDoms; i_st++)
        if((G_act_subdom[i_st] == G_proc_subdom) &&
           (G_proc_strip[i_st] == 2) )
          G_act_subdom[i_st] = G_fem1->Decomp.NSubDoms;

      for(j = 0; j < SubMesh->NElemsTriang3; j++)
      {
        elem = SubMesh->GEIsTriang3[j]-1;
        G_fem1->Decomp.SubDoms[elem] = 
          G_act_subdom[G_master->Decomp.SubDoms[elem]-1];
      }
      for(j = 0; j < SubMesh->NElemsTriang4; j++)
      {
        elem = SubMesh->GEIsTriang4[j]-1;
        G_fem1->Decomp.SubDoms[elem] = 
          G_act_subdom[G_master->Decomp.SubDoms[elem]-1];
      }
      for(j = 0; j < SubMesh->NElemsTriang5; j++)
      {
        elem = SubMesh->GEIsTriang5[j]-1;
        G_fem1->Decomp.SubDoms[elem] = 
          G_act_subdom[G_master->Decomp.SubDoms[elem]-1];
      }

      e_FreeMeshData(SubMesh);
    }
  }

  G_no_elems_in_subdom = e_alloc_i_vect(G_total_no_subdomains);
  for(i = 0; i < G_master->Decomp.NSubDoms; i++)
  {
    G_no_elems_in_subdom[G_act_subdom[i]-1] += G_no_elems_in_strip[i];
  }


  e_free(design_var);

}
