#include "ssgm.h"

#define YES 1
#define NO  0

int           **G_total_link_info;
static int      G_method;        /* determines the method to use */

static int      G_interactive_keyboard_input;
static char     G_kbdinputfilename[E_LINE_MAX];
static FILE    *G_keyboard_input_fp;
static int      G_kbdinputfile_linecount;

/* //////////////////////////////////////////////////////////////////////// */
static void     j_analise_main_args(int argc, char *argv[]);
static void     j_read_projectname(void);
static int      j_read_GA_method(void);
static int      j_read_no_subdomains(void);
static void     j_read_GA_initdata(void);

static void     Strip_Information_for_Type(int no, int **nodes, double **coords, 
                                           int *subdoms, double *StripArea);
static void     Strip_Information_Creation();
static void     Assign_Link(int no, int *Degree, int **Adjacency,
                            int *SubDoms, int StartIndex);
static void     Assign_Triang(int no, int *subdom, int StartIndex);
static void     Assign_Elements_Subdom(MESH * mesh,
                                       int *degree, int **adjacency);
/* ------------------------------------------------------------------------- */
/*                              Input functions                              */
/* ------------------------------------------------------------------------- */

static void j_analise_main_args(int argc, char *argv[])
{
  if(argc > 1)
  {
    if(argc != 3 || strcmp("-i", argv[1]) != 0)
    {
      printf("\nUsage:  esgm [-i keyboard_input_filename]");
      printf("\n...going for interactive keyboard input....\n\n");
      G_interactive_keyboard_input = YES;
      return;
    }
    if((G_keyboard_input_fp = fopen(argv[2], "r")) == NULL)
    {
      printf("\nCan not open file >%s<\n", argv[2]);
      printf("\nUsage:  esgm [-i keyboard_input_filename]");
      printf("\n...going for interactive keyboard input....\n\n");
      G_interactive_keyboard_input = YES;
      return;
    }
    G_interactive_keyboard_input = NO;
    G_kbdinputfile_linecount = 0;
    strcpy(G_kbdinputfilename, argv[2]);
  }
  else
  {
    printf("\n...going for interactive keyboard input....\n\n");
    G_interactive_keyboard_input = YES;
  }
}                                /* end of function j_analise_main_args()  */

static void j_read_projectname(void)
{
  char      line[E_LINE_MAX];
  char     *check;

  if(G_interactive_keyboard_input == NO)
  {
    e_GetLine(line, E_LINE_MAX, &G_kbdinputfile_linecount, G_keyboard_input_fp);
    if(sscanf(line, "%s", G_projectname) != 1)
      e_FileLineError(G_kbdinputfile_linecount,
                      G_kbdinputfilename,
                      "Error in keyboard file, can not read projectname!");

  }
  else
    /* read from keyboard  */
  {
    printf("\n\nType in the projectname which is a Mesh Definition File (.mdf) !");
    printf("\nGive no extension : ");
    check = fgets(line, E_LINE_MAX, stdin);
    if(   (check == NULL)
       || (sscanf(line, "%s", G_projectname) != 1) )
      e_GeneralError("Error in keyboard input, can not read projectname!");
  }
  return;
}                                /* end of function j_read_projectname()  */

static int j_read_GA_method(void)
{
  char      line[E_LINE_MAX];
  char     *check;
  int       inp;

  if(G_interactive_keyboard_input == NO)
  {
    e_GetLine(line, E_LINE_MAX, &G_kbdinputfile_linecount, G_keyboard_input_fp);
    if(sscanf(line, "%d", &inp) != 1)
      e_FileLineError(G_kbdinputfile_linecount, G_kbdinputfilename,
               "Error in keyboard file, can not read optimisation method!");

    if(inp != 1 && inp != 2)
    {
      sprintf(line,
              "Error in keyboard file, can not have %d optimisation method!", 
              inp);
      e_FileLineError(G_kbdinputfile_linecount, G_kbdinputfilename, line);
    }
  }
  else
    /* read from keyboard  */
  {
    printf("\n\nSpecify which optimisation method to use!");
    printf("\n(Give `%d' - for linear cutting", 1);
    printf("\n      `%d' - for multi-variable cutting", 2);
    printf("\nYour choice is [%d|%d] : ", 1, 2);
    check = fgets(line, E_LINE_MAX, stdin);
    if(   (check == NULL)
       || (sscanf(line, "%d", &inp) != 1) )
      e_GeneralError("Error in keyboard input, can not read opt. method!");

    if(inp != 1 && inp != 2)
    {
      sprintf(line,
              "Error in keyboard input, can not have %d optimisation method!", 
              inp);
      e_GeneralError(line);
    }
  }

  return(inp);
}

static int j_read_no_subdomains(void)
{
  char      line[E_LINE_MAX];
  char     *check;
  int       subdom;

  if(G_interactive_keyboard_input == NO)
  {
    e_GetLine(line, E_LINE_MAX, &G_kbdinputfile_linecount, G_keyboard_input_fp);
    if(sscanf(line, "%d", &subdom) != 1)
      e_FileLineError(G_kbdinputfile_linecount,
                      G_kbdinputfilename,
                    "Error in keyboard file, can not read no. subdomains!");
    if(G_method == 1)
    {
      G_total_recursions = 1;
      while(subdom > pow(2, G_total_recursions))
        G_total_recursions++;
      if(subdom != pow(2, G_total_recursions))
      {
        sprintf(line,
          "Error in keyboard file, can not have %d no. subdomains!", 
          subdom);
        e_FileLineError(G_kbdinputfile_linecount, G_kbdinputfilename, line);
      }
    }
  }
  else
    /* read from keyboard  */
  {
    printf("\n\nNumber of subdomains : ");
    check = fgets(line, E_LINE_MAX, stdin);
    if(   (check == NULL)
       || (sscanf(line, "%d", &subdom) != 1) )
      e_GeneralError("Error in keyboard input, can not read no. subdomains!");

    if(G_method == 1)
    {
      G_total_recursions = 1;
      while(subdom > pow(2, G_total_recursions))
        G_total_recursions++;
      if(subdom != pow(2, G_total_recursions))
      {
        sprintf(line,
          "Error in keyboard file, can not have %d no. subdomains!", 
          subdom);
        e_FileLineError(G_kbdinputfile_linecount, G_kbdinputfilename, line);
      }
    }
  }
  return(subdom);
}                                /* end of function j_read_no_subdomains()  */

static void j_read_GA_initdata(void)
{
  FILE     *fp;
  char      line[E_LINE_MAX];
  int       linecount;
  char      GA_init_datafile[E_LINE_MAX];

  strcpy(GA_init_datafile, "ssgm.dat");

  if((fp = fopen(GA_init_datafile, "r")) == NULL)
    e_FileError(GA_init_datafile, "Can not open GA init data file!");

  linecount = 0;
  e_GetLine(line, E_LINE_MAX, &linecount, fp);
  if(sscanf(line, "%d", &(GA_popsize)) != 1)
    e_FileLineError(linecount, GA_init_datafile,
                    "Could not convert this line into integer!");

  e_GetLine(line, E_LINE_MAX, &linecount, fp);
  if(sscanf(line, "%d", &(GA_maxgen)) != 1)
    e_FileLineError(linecount, GA_init_datafile,
                    "Could not convert this line into integer!");

  e_GetLine(line, E_LINE_MAX, &linecount, fp);
  if(sscanf(line, "%lf", &(GA_pcross)) != 1)
    e_FileLineError(linecount, GA_init_datafile,
                    "Could not convert this line into double!");

  e_GetLine(line, E_LINE_MAX, &linecount, fp);
  if(sscanf(line, "%lf", &(GA_pmutation)) != 1)
    e_FileLineError(linecount, GA_init_datafile,
                    "Could not convert this line into double!");

  e_GetLine(line, E_LINE_MAX, &linecount, fp);
  if(sscanf(line, "%lf", &(GA_j_gamma)) != 1)
    e_FileLineError(linecount, GA_init_datafile,
                    "Could not convert this line into double!");
  if(GA_j_gamma < 0.00 || GA_j_gamma > 1.00)
  {
    sprintf(line, "Can not have %f for gamma value in fitness evaluation!",
            GA_j_gamma);
    e_FileLineError(linecount, GA_init_datafile, line);
  }

  e_GetLine(line, E_LINE_MAX, &linecount, fp);
  if(sscanf(line, "%d", &(GA_seed)) != 1)
    e_FileLineError(linecount, GA_init_datafile,
                    "Could not convert this line into integer!");

  e_GetLine(line, E_LINE_MAX, &linecount, fp);
  if(sscanf(line, "%d", &(GA_linear)) != 1)
    e_FileLineError(linecount, GA_init_datafile,
                    "Could not convert this line into integer!");

  e_GetLine(line, E_LINE_MAX, &linecount, fp);
  if(sscanf(line, "%d", &(GA_printing)) != 1)
    e_FileLineError(linecount, GA_init_datafile,
                    "Could not convert this line into integer!");

  fclose(fp);
}                                /* end of function j_read_GA_initdata()  */


/* ------------------------------------------------------------------------- */
/*                  Preparation of strip information                         */
/* ------------------------------------------------------------------------- */

static void Strip_Information_for_Type(int no, int **nodes, double **coords, 
                                       int *subdoms, double *StripArea)
{
  double    mid_x, mid_y, x21, y21, z21, x31, y31, z31, area;
  double    norm[3];
  int       n1, n2, n3, i;


  for(i = 0; i < no; i++)
  {
    G_no_elems_in_strip[subdoms[i] - 1]++;

    n1 = nodes[i][0] - 1;
    n2 = nodes[i][1] - 1;
    n3 = nodes[i][2] - 1;

    mid_x = (coords[n1][0] + coords[n2][0] + coords[n3][0]) / 3.0;
    mid_y = (coords[n1][1] + coords[n2][1] + coords[n3][1]) / 3.0;

    x21 = coords[n2][0] - coords[n1][0];
    y21 = coords[n2][1] - coords[n1][1];
    z21 = coords[n2][2] - coords[n1][2];

    x31 = coords[n3][0] - coords[n1][0];
    y31 = coords[n3][1] - coords[n1][1];
    z31 = coords[n3][2] - coords[n1][2];

    norm[0] = (y21 * z31) - (y31 * z21);
    norm[1] = (z21 * x31) - (z31 * x21);
    norm[2] = (x21 * y31) - (x31 * y21);

    area = sqrt(E_SQUARE(norm[0]) + E_SQUARE(norm[1]) + E_SQUARE(norm[2]));
    area = area / 2;

    G_center[subdoms[i] - 1][0] += area * mid_x;
    G_center[subdoms[i] - 1][1] += area * mid_y;

    StripArea[subdoms[i] - 1] += area;
  }
}

static void Strip_Information_Creation()
{
  double   *StripArea;
  int       i;

  StripArea = e_alloc_d_vect(G_master->Decomp.NSubDoms);
  G_center = e_alloc_cont_d_matr(G_master->Decomp.NSubDoms, 2);
  G_no_elems_in_strip = e_alloc_i_vect(G_master->Decomp.NSubDoms);


  Strip_Information_for_Type(G_master->NElemsTriang3, G_master->NodesTriang3, 
                             G_master->MeshPointCoords, G_master->Decomp.SubDomsTriang3,
                             StripArea);
  Strip_Information_for_Type(G_master->NElemsTriang4, G_master->NodesTriang4, 
                             G_master->MeshPointCoords, G_master->Decomp.SubDomsTriang4,
                             StripArea);
  Strip_Information_for_Type(G_master->NElemsTriang5, G_master->NodesTriang5, 
                             G_master->MeshPointCoords, G_master->Decomp.SubDomsTriang5,
                             StripArea);

/* create the center of each strip */
  for(i = 0; i < G_master->Decomp.NSubDoms; i++)
  {
    G_center[i][0] = G_center[i][0] / StripArea[i];
    G_center[i][1] = G_center[i][1] / StripArea[i];
  }

  e_free(StripArea);
}

/* ------------------------------------------------------------------------- */
/*                              Link assignment                              */
/* ------------------------------------------------------------------------- */

static void Assign_Link(int no, int *Degree, int **Adjacency, 
                        int *SubDoms, int StartIndex)
{
  int i, ElemIndex;


  for(i = 0; i < no; i++)
  {
    ElemIndex = StartIndex + i;
    switch(Degree[ElemIndex-1])
    {
      case 0:
        e_GeneralError("Unconnected Link element !");
        break;
      case 1:
        SubDoms[ElemIndex-1] = SubDoms[Adjacency[ElemIndex-1][0]-1];
        break;
      case 2:
        if(SubDoms[Adjacency[ElemIndex-1][0]-1] == 
           SubDoms[Adjacency[ElemIndex-1][1]-1]    )
        {
          SubDoms[ElemIndex-1] = SubDoms[Adjacency[ElemIndex-1][0]-1];
          G_no_elems_in_subdom[SubDoms[Adjacency[ElemIndex-1][0]-1]-1]++;
        }
        else
        {
          if(G_no_elems_in_subdom[SubDoms[Adjacency[ElemIndex-1][0]-1]-1] < 
             G_no_elems_in_subdom[SubDoms[Adjacency[ElemIndex-1][1]-1]-1]   )
          {
            SubDoms[ElemIndex-1] = SubDoms[Adjacency[ElemIndex-1][0]-1];
            G_no_elems_in_subdom[SubDoms[Adjacency[ElemIndex-1][0]-1]-1]++;
          }
          else
          {
            SubDoms[ElemIndex-1] = SubDoms[Adjacency[ElemIndex-1][1]-1];
            G_no_elems_in_subdom[SubDoms[Adjacency[ElemIndex-1][1]-1]-1]++;
          }
        }
        break;
      default:
        e_GeneralError("Link connects to more than two element !");
    }
  }
}


static void Assign_Triang(int no, int *subdom, int StartIndex)
{
  int i, ElemIndex;

  for(i = 0; i < no; i++)
  {
    ElemIndex = StartIndex + i;
    subdom[i] = G_act_subdom[G_master->Decomp.SubDoms[ElemIndex-1]-1];
  }
}


static void Assign_Elements_Subdom(MESH *mesh,
                                   int *degree, int **adjacency)
{

  Assign_Triang(mesh->NElemsTriang3, mesh->Decomp.SubDomsTriang3,
                mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3]);
  Assign_Triang(mesh->NElemsTriang4, mesh->Decomp.SubDomsTriang4,
                mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4]);
  Assign_Triang(mesh->NElemsTriang5, mesh->Decomp.SubDomsTriang5,
                mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5]);

  Assign_Link(mesh->NElemsLink1, degree, adjacency,
              mesh->Decomp.SubDoms,
              mesh->FirstNetIndex[E_ELEM_TYPE_LINK1]);
  Assign_Link(mesh->NElemsLink2, degree, adjacency,
              mesh->Decomp.SubDoms,
              mesh->FirstNetIndex[E_ELEM_TYPE_LINK2]);
  Assign_Link(mesh->NElemsLink3, degree, adjacency,
              mesh->Decomp.SubDoms,
              mesh->FirstNetIndex[E_ELEM_TYPE_LINK3]);
  Assign_Link(mesh->NElemsLink4, degree, adjacency,
              mesh->Decomp.SubDoms,
              mesh->FirstNetIndex[E_ELEM_TYPE_LINK4]);
  Assign_Link(mesh->NElemsLink5, degree, adjacency,
              mesh->Decomp.SubDoms,
              mesh->FirstNetIndex[E_ELEM_TYPE_LINK5]);

}


/* ------------------------------------------------------------------------- */
/*                            Objective function                             */
/* ------------------------------------------------------------------------- */

double objective_function(long nvars, DESIGNPARAMS vars)
{
  if(G_method == 1)
  {
    /* linear GA cut */
    return(linear_objective(nvars, vars));
  }
  else
  {
    /* multi GA cut */
    return(multi_objective(nvars, vars));
  }
}


/* ------------------------------------------------------------------------- */
/*                               Main function                               */
/* ------------------------------------------------------------------------- */

main(int argc, char *argv[])
{
  int *stat_subdom;
  int i;

  /* reading in the necessary information */
  j_analise_main_args(argc, argv);
  j_read_projectname();
  printf("\nThe projectname is taken as: %s", G_projectname);
  printf("\n    This is the name of the input .MDF file.");

  e_GetMeshData(&G_master, G_projectname);
  if((G_master->NElemsLink1   +
      G_master->NElemsLink2   +
      G_master->NElemsLink3   +
      G_master->NElemsLink4   +
      G_master->NElemsLink5   +
      G_master->NElemsTriang3 +
      G_master->NElemsTriang4 +
      G_master->NElemsTriang5   ) != G_master->TotalNElems)
    e_GeneralError("Mesh contains unsupported element(s) !");

  printf("\n\nThe project has been loaded in succesfully with the title:\n<%s>",
         G_master->Title);
  printf("\n    total No. elements: %d", G_master->TotalNElems);
  printf("\n    total No. co-ordinate points: %d", G_master->NMeshPoints);
  printf("\n    total No. nodes: %d", G_master->NNodes);

  G_method              = j_read_GA_method();
  G_total_no_subdomains = j_read_no_subdomains();
  j_read_GA_initdata();
  printf("\n\n    The required number of subdomains is taken as: %d",
         G_total_no_subdomains);


/* Initialise the dual graph, 
   if there is an out-of-plane link, QUIT */
  NoNodeElem = e_alloc_i_vect(G_master->NNodes);
  NodeElem   = e_init_frag_i_matr(G_master->NNodes);
  Degree     = e_alloc_i_vect(G_master->TotalNElems);
  Adjacency  = e_init_frag_i_matr(G_master->TotalNElems);

  e_InitMeshPointElemConnection(G_master, NoNodeElem, NodeElem);
  SetDualAdjacency(G_master, Degree, Adjacency,
                   NoNodeElem, NodeElem);

  /* creating the strip information */
  Create_Strips(G_master, Degree, Adjacency);
  printf("\n\n    Number of strips: %d", G_master->Decomp.NSubDoms);

  /* creating the subdomain graph */
  SubdomGraphDegree = e_alloc_i_vect(G_master->Decomp.NSubDoms);
  SubdomGraph       = e_alloc_cont_i_matr(G_master->Decomp.NSubDoms, 
                                          G_master->Decomp.NSubDoms-1);
  SubdomGraphRef    = e_alloc_cont_i_matr(G_master->Decomp.NSubDoms, 
                                          G_master->Decomp.NSubDoms-1);
  Create_Subdomain_Graph(G_master, SubdomGraphDegree, SubdomGraph,
                         SubdomGraphRef);

  Strip_Information_Creation();

  if(G_method == 1)
    do_linear_decomposition();
  else if(G_method == 2)
    do_multi_decomposition();
  else
    e_GeneralError("Unknown method !");


  Assign_Elements_Subdom(G_fem1, Degree, Adjacency);

  strcat(G_projectname, "out");
  e_PutDecompData(G_fem1, G_projectname);

  stat_subdom = e_alloc_i_vect(G_fem1->Decomp.NSubDoms);

  for(i = 0; i < G_fem1->TotalNElems; i++)
    stat_subdom[ G_fem1->Decomp.SubDoms[i]-1 ]++;

  printf("\n\n*******************************************");
  printf("\n*****           Results               *****");
  printf("\n*******************************************");
  for(i = 0; i < G_fem1->Decomp.NSubDoms; i++)
    printf("\n %d elements in %d. subdom", stat_subdom[i], i+1);

  e_free(stat_subdom);

  printf("\n\n");

}
