#include "ssgm.h"

static double **GV_subdom_centre;
static int      GA_no_Pareto, GA_all_time_Pareto_length;
static int    **GA_Pareto;
static double  *GA_obj_func;



double multi_objective(long nvars, DESIGNPARAMS vars)
{
  long      no_interfaces;
  long      i, j, subdom;
  long      diff, ideal, total;
  double    fitness, ekm;

/* these are only for Pareto optimum */
  static  int very_first_time = 1, counter = 0;
  static  double n_this, n_new;
  double  min_prox, par_result, prox;
  short   pareto_smaller, pareto_equal, pareto_bigger;

  no_interfaces = 0;
  ideal = 0;
  for(i = 0; i < G_total_no_subdomains; i++)
    G_no_elems_in_subdom[i] = 0;

/* fill G_act_subdom with the appropriate number */
  for(i = 0; i < G_master->Decomp.NSubDoms; i++)
  {
    subdom = (int) vars[i].parameter;
    G_act_subdom[i] = subdom + 1;
    G_no_elems_in_subdom[subdom] += G_no_elems_in_strip[i];
    ideal += G_no_elems_in_strip[i];
  }

/* determine the communication approximately */
  for(i = 0; i < G_master->Decomp.NSubDoms; i++)
  {
    for(j = 0; j < SubdomGraphDegree[i]; j++)
    {
      subdom = SubdomGraph[i][j]-1;
      if(G_act_subdom[subdom] != G_act_subdom[i])
        no_interfaces += SubdomGraphRef[i][j];
    }
  }

  total = ideal;
  diff = 0;
  ideal /= G_total_no_subdomains;
  for(i = 0; i < G_total_no_subdomains; i++)
    diff += E_ABS(ideal - G_no_elems_in_subdom[i]);


/* aspect ratio */
  if(GA_aspect)
  {
    for(i = 0; i < G_total_no_subdomains; i++)
    {
      GV_subdom_centre[i][0] = 0.00;
      GV_subdom_centre[i][1] = 0.00;
    }

    for(i = 0; i < G_master->Decomp.NSubDoms; i++)
    {
      GV_subdom_centre[G_act_subdom[i]-1][0] += G_center[i][0];
      GV_subdom_centre[G_act_subdom[i]-1][0] += G_center[i][0];
    }

    for(i = 0; i < G_total_no_subdomains; i++)
    {
      GV_subdom_centre[i][0] /= G_no_elems_in_subdom[i];
      GV_subdom_centre[i][1] /= G_no_elems_in_subdom[i];
    }

    ekm = 0.00;

    for(i = 0; i < G_master->Decomp.NSubDoms; i++)
    {
      ekm += E_SQUARE(GV_subdom_centre[G_act_subdom[i]-1][0] -
                   G_center[i][0]);
      ekm += E_SQUARE(GV_subdom_centre[G_act_subdom[i]-1][1] -
                   G_center[i][1]);
    }
  }

  if(GA_Pareto_method)
  {
    GA_obj_func[0] = (double)no_interfaces;
    GA_obj_func[1] = diff;
    GA_obj_func[2] = ekm;

    if(counter == GA_popsize)
    {
      n_this = n_new;
      counter = 0;
    }

    if(very_first_time)
    {
      /* only for the very first time 
         contains initialization as well
      */
      n_this = total; /* total number of elements */
      n_new = n_this;

      min_prox = 0.0;

      GA_no_Pareto = 0;
      i = GA_no_Pareto;
      GA_all_time_Pareto_length = GA_no_Pareto;
      GA_Pareto = NULL;

      very_first_time = 0;
    }
    else
    {
      /* calculate proximity values */
      min_prox = 1e50;
      for(i = 0; i < GA_no_Pareto; i++)
      {
        prox = 0.00;
        for(j = 0; j < G_total_no_subdomains+1; j++)
          prox += E_SQUARE(((double)GA_Pareto[i][j] - (double)GA_obj_func[j]) / 
                            (double)GA_Pareto[i][j] );
        prox = sqrt(prox);
        if(prox < min_prox)
          min_prox = prox;
      }

      /* check whether it is Pareto optimum and dominance */
      for(i = 0; i < GA_no_Pareto; i++)
      {
        pareto_smaller = 0;
        pareto_equal   = 0;
        pareto_bigger  = 0;
        for(j = 0; j < 3; j++)
        {
          if(GA_Pareto[i][j] < GA_obj_func[j])
            pareto_smaller = 1;
          if(GA_Pareto[i][j] == GA_obj_func[j])
            pareto_equal = 1;
          if(GA_Pareto[i][j] > GA_obj_func[j])
            pareto_bigger = 1;
        }
        if((pareto_smaller  && !pareto_equal && !pareto_bigger) ||
           (pareto_smaller  &&  pareto_equal && !pareto_bigger) ||
           (!pareto_smaller &&  pareto_equal && !pareto_bigger)    )
          break;

        if((!pareto_smaller && !pareto_equal && pareto_bigger) ||
           (!pareto_smaller &&  pareto_equal && pareto_bigger)    )
          GA_Pareto[i][G_total_no_subdomains+1] = 1;
      }
    } /* else very_first_time */


    /* new Pareto optimum */
    if(i == GA_no_Pareto) 
    {
      /* remove dominated Pareto optimums */
      for(i = 0; i < GA_no_Pareto; i++)
      {
        if(GA_Pareto[i][G_total_no_subdomains+1])
        {
          e_free(GA_Pareto[i]);
          for(j = i+1; j < GA_no_Pareto; j++)
            GA_Pareto[j-1] = GA_Pareto[j];
          GA_no_Pareto--;
          i--;
        }
      }

      /* add new Pareto optimum */
      if(GA_all_time_Pareto_length == GA_no_Pareto)
      {
        GA_Pareto = e_incr_frag_i_matr(GA_all_time_Pareto_length, GA_Pareto, 1);
        GA_all_time_Pareto_length++;
      }

      GA_Pareto[GA_no_Pareto] = e_alloc_i_vect(G_total_no_subdomains+2);
      for(j = 0; j < G_total_no_subdomains+1; j++)
        GA_Pareto[GA_no_Pareto][j] = GA_obj_func[j];
      GA_no_Pareto++;

      par_result = n_this + min_prox;
      if(par_result > n_new)
        n_new = par_result;
    }
    else
    {
      par_result = n_this - min_prox;
    }

    counter++;
    return par_result;
  }
/* not using Pareto optimum */
  else
  {
    fitness = G_master->TotalNElems * 3 - (1.00 - GA_j_gamma) * diff -
              GA_j_gamma * (no_interfaces);

    fitness -= 45000-ekm;
  }

  return(fitness);
}


void do_multi_decomposition()
{
  DESIGNPARAMS  design_var;
  int           pr, i, j, k;
  double        sx, sy, sz;
  int          *stat_subdom;

  printf("\nMulti variable decomposition ...\n");

  GA_aspect = 1;
  GA_Pareto_method = 0;
  if(GA_Pareto_method)
  {
    GA_obj_func = e_alloc_d_vect(3);
  }

  G_act_subdom = e_alloc_i_vect(G_master->Decomp.NSubDoms);
  G_no_elems_in_subdom = e_alloc_i_vect(G_total_no_subdomains);
  GV_subdom_centre = e_alloc_cont_d_matr(G_total_no_subdomains, 2);

  /* quick and dirty way to get the same mesh with different subdomain info */
  e_GetMeshData(&G_fem1, G_projectname);
  e_SetupDecompData(G_fem1);
/*
  Create_Subdom_Struct_with_Zero(G_fem1);
*/
  G_fem1->Decomp.NSubDoms = G_total_no_subdomains;

  /* initialise the design variables */
  design_var = (DESIGNPARAMS) e_calloc(G_master->Decomp.NSubDoms, sizeof(PARAM));
  pr = 1;
  while(G_total_no_subdomains > pow(2, pr))
    pr++;
  for(i = 0; i < G_master->Decomp.NSubDoms; i++)
  {
    design_var[i].minparam = 0.00;
    design_var[i].maxparam = G_total_no_subdomains-1;
    design_var[i].lparam = pr*2;
  }

  /* start GA */
  srand(GA_seed);
  genetic_algorithm(GA_popsize, GA_pcross, GA_pmutation,
                    G_master->Decomp.NSubDoms, design_var, GA_maxgen, 
                    GA_linear, GA_printing);

  multi_objective(G_master->Decomp.NSubDoms, design_var);

  printf("\nPareto optimums : %d", GA_no_Pareto);


  e_free(design_var);
}

