#include "e_lib.h"

#include "strip_create.h"

#define ALLOC_SIZE  16


static void AddAllTriangsToStrip(MESH *mesh, int *degree, int **adjacency,
                                 int NElems, int *Subdoms, int ElemType);
static int AddAdjTriangToStrip(MESH *mesh, int elem_index, 
                               int **strip_triangs, int *len_tri_list, 
                               int strip_index);


static int AddAdjTriangToStrip(MESH *mesh, int elem_index, 
                               int **strip_triangs, int *len_tri_list, 
                               int strip_index)
{
  int boundary = -1;

  /* determine element type */
  if(mesh->FirstNetIndex[E_ELEM_TYPE_LINK1] &&
     mesh->FirstNetIndex[E_ELEM_TYPE_LINK1] <= (elem_index))
  {
    boundary = 1;
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_LINK2] &&
     mesh->FirstNetIndex[E_ELEM_TYPE_LINK2] <= (elem_index))
  {
    boundary = 1;
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_LINK3] &&
     mesh->FirstNetIndex[E_ELEM_TYPE_LINK3] <= (elem_index))
  {
    boundary = 1;
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_LINK4] &&
     mesh->FirstNetIndex[E_ELEM_TYPE_LINK4] <= (elem_index))
  {
    boundary = 1;
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_LINK5] &&
     mesh->FirstNetIndex[E_ELEM_TYPE_LINK5] <= (elem_index))
  {
    boundary = 1;
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] &&
     mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] <= (elem_index))
  {
    boundary = 0;
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] &&
     mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] <= (elem_index))
  {
    boundary = 0;
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] &&
     mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] <= (elem_index))
  {
    boundary = 0;
  }

  if((boundary == 0) && (mesh->Decomp.SubDoms[elem_index-1] == 0))
  {
    mesh->Decomp.SubDoms[elem_index-1] = strip_index;
    (*strip_triangs) = e_incr_i_vect(*len_tri_list, *strip_triangs, ALLOC_SIZE);
    (*strip_triangs)[(*len_tri_list)] = elem_index;
    (*len_tri_list)++;
  }

  return(boundary);
}



static void AddAllTriangsToStrip(MESH *mesh, int *degree, int **adjacency,
                                 int NElems, int *Subdoms, int ElemType)
{
  int   i, i_deg, act_tri, bndtype;
  int   ptr_tri_list, len_tri_list;
  int  *strip_triangs;

  for(i = 0; i < NElems; i++)
  {
    /* elements not yet decided */
    if(Subdoms[i] == 0)
    {
      /* new strip starts */
      mesh->Decomp.NSubDoms++;
      /* first element in the strip is explicitly defined here */
      ptr_tri_list = 0;
      len_tri_list = 1;
      strip_triangs = e_alloc_i_vect(ALLOC_SIZE);
      strip_triangs[ptr_tri_list] = mesh->FirstNetIndex[ElemType] + i;
      Subdoms[i] = mesh->Decomp.NSubDoms;
      
      /* go, until all elements in the list has been processed */
      while(ptr_tri_list != len_tri_list)
      {
        act_tri = strip_triangs[ptr_tri_list++];

        for(i_deg = 0; i_deg < degree[act_tri-1]; i_deg++)
        {
          bndtype = AddAdjTriangToStrip(mesh, adjacency[act_tri-1][i_deg], 
                                        &strip_triangs, &len_tri_list, 
                                        mesh->Decomp.NSubDoms);
        }
      }
      /* free up memory */
      e_free(strip_triangs);
      strip_triangs = NULL;
    }
  }

}



void Create_Strips(MESH *mesh, int *degree, int **adjacency)
{
  int  i;

  if(mesh->NElemsTriang3 +
     mesh->NElemsTriang4 +
     mesh->NElemsTriang5 +
     mesh->NElemsLink1 +
     mesh->NElemsLink2 +
     mesh->NElemsLink3 +
     mesh->NElemsLink4 +
     mesh->NElemsLink5   != mesh->TotalNElems)
  {
    e_GeneralError("Create_Strips: Unsupported element type !");
  }

  /* setup valid data structure */
  e_SetupDecompData(mesh);
  /* zero it out */
  for(i = 0; i < mesh->TotalNElems; i++)
  {
    mesh->Decomp.SubDoms[i] = 0;
  }
  mesh->Decomp.NSubDoms = 0;

  /* process valid triangle element types */
  AddAllTriangsToStrip(mesh, degree, adjacency,
                       mesh->NElemsTriang3,
                       mesh->Decomp.SubDomsTriang3,
                       E_ELEM_TYPE_TRIANG3);
                       
  AddAllTriangsToStrip(mesh, degree, adjacency,
                       mesh->NElemsTriang4,
                       mesh->Decomp.SubDomsTriang4,
                       E_ELEM_TYPE_TRIANG4);
                       
  AddAllTriangsToStrip(mesh, degree, adjacency,
                       mesh->NElemsTriang5,
                       mesh->Decomp.SubDomsTriang5,
                       E_ELEM_TYPE_TRIANG5);

}




