/* Convert a "graph.ptn" file into a "dom" file */


#include <stdio.h>
#include <math.h>
#include <e_lib.h>


char        GV_filename[E_LINE_MAX];
MESH       *GV_global_mesh;


void f_read_in_data(int argc, char *argv[]);
void create_subdom_struct(MESH *mesh);


void f_read_in_data(int argc, char *argv[])
{
  char   line[E_LINE_MAX];
  char  *check;

  if(argc > 1)
  {
    strcpy(GV_filename, argv[1]);
  }
  else
  {
    printf("\n\nType in the input filename (.mdf) !");
    printf("\nGive no extension : ");
    check = fgets(line, E_LINE_MAX, stdin);
    if(   (check == NULL)
       || (sscanf(line, "%s", &GV_filename) != 1))
      e_GeneralError("Error in keyboard input, can not read filename!");
  }

  printf("\nThe projectname is taken as: %s", GV_filename);

  e_GetMeshData(&GV_global_mesh, GV_filename);

  if(GV_global_mesh->TotalNElems != GV_global_mesh->NElemsLink1 +
                                    GV_global_mesh->NElemsLink2 +
                                    GV_global_mesh->NElemsLink3 +
                                    GV_global_mesh->NElemsLink4 +
                                    GV_global_mesh->NElemsLink5 +
                                    GV_global_mesh->NElemsTriang3 +
                                    GV_global_mesh->NElemsTriang4 +
                                    GV_global_mesh->NElemsTriang5   )
    e_GeneralError("Mesh contains non-supported element !");
}



main(int argc, char *argv[])
{
  char part_file[E_LINE_MAX];
  FILE  *fp;
  char line[E_LINE_MAX];
  int i, linecount, subdom, max_subdom = 0;
  int  *strip_subdom;
  int  *NoNodeElem;
  int  **NodeElem;
  int   *Degree;
  int  **Adjacency;

  printf("\n\nConvert a \"str.graph.ptn\" file into a \"dom\" file\n");

  f_read_in_data(argc, argv);

  strcpy(part_file, GV_filename);
  strcat(part_file, ".str.graph.ptn");
  printf("\nPartition file: %s", part_file);

  if ((fp = fopen(part_file, "r")) == NULL)
  {
    e_GeneralError("Can not open partition file !");
  }

/* Initialise the dual graph, 
   if there is an out-of-plane link, QUIT */
  NoNodeElem = e_alloc_i_vect(GV_global_mesh->NNodes);
  NodeElem = e_init_frag_i_matr(GV_global_mesh->NNodes);
  Degree = e_alloc_i_vect(GV_global_mesh->TotalNElems);
  Adjacency = e_init_frag_i_matr(GV_global_mesh->TotalNElems);

  e_InitMeshPointElemConnection(GV_global_mesh, NoNodeElem, NodeElem);
  SetDualAdjacency(GV_global_mesh, Degree, Adjacency,
                   NoNodeElem, NodeElem);

  /* creating the strip information */
  Create_Strips(GV_global_mesh, Degree, Adjacency);
  printf("\n\n    Number of strips: %d", GV_global_mesh->Decomp.NSubDoms);

  strip_subdom = e_alloc_i_vect(GV_global_mesh->Decomp.NSubDoms);

  for(i = 0; i < GV_global_mesh->Decomp.NSubDoms; i++)
  {
    e_GetNextLine(line, &linecount, part_file, fp);
    if(sscanf(line, "%d", &(strip_subdom[i])) != 1)
      e_FileLineError(linecount, part_file, "Cannot interpret the line !");
    if(max_subdom < strip_subdom[i])
      max_subdom = strip_subdom[i];
  }

  for(i = 0; i < GV_global_mesh->TotalNElems; i++)
  {
    subdom = strip_subdom[GV_global_mesh->Decomp.SubDoms[i]-1]+1;
    GV_global_mesh->Decomp.SubDoms[i] = subdom;
  }

  GV_global_mesh->Decomp.NSubDoms = max_subdom+1;

  fclose(fp);

  e_PutDecompData(GV_global_mesh, GV_filename);

  printf("\n\n");

}
