#include <stdio.h>
#include <math.h>
#include <e_lib.h>

char            GV_filename[E_LINE_MAX];
MESH           *GV_global_mesh, *GV_new_mesh;


int            *NoNodeElem;
int           **NodeElem;
int            *Degree;
int           **Adjacency;
double        **G_center;

char *GetStrArg(char *pLine, char *format, void *argument);
static void f_read_in_data(int argc, char *argv[]);
static void create_nodes(MESH *mesh);
static void create_links(MESH *mesh);
static void create_mesh();
static void Strip_Information_for_Type(int no, int **nodes, double **coords, 
                                       int *subdoms, double *StripArea);
static void Strip_Information_Creation();


static void f_read_in_data(int argc, char *argv[])
{
  char   line[E_LINE_MAX];
  char  *check;


  if(argc > 1)
  {
    strcpy(GV_filename, argv[1]);
  }
  else
  {
    printf("\n\nType in the input filename (.mdf) !");
    printf("\nGive no extension : ");
    check = fgets(line, E_LINE_MAX, stdin);
    if(   (check == NULL)
       || (sscanf(line, "%s", &GV_filename) != 1) )
      e_GeneralError("Error in keyboard input, can not read filename!");
  }

  printf("\nThe projectname is taken as: %s", GV_filename);

  e_GetMeshData(&GV_global_mesh, GV_filename);

  if(GV_global_mesh->TotalNElems != GV_global_mesh->NElemsLink1 +
                                    GV_global_mesh->NElemsLink2 +
                                    GV_global_mesh->NElemsLink3 +
                                    GV_global_mesh->NElemsLink4 +
                                    GV_global_mesh->NElemsLink5 +
                                    GV_global_mesh->NElemsTriang3 +
                                    GV_global_mesh->NElemsTriang4 +
                                    GV_global_mesh->NElemsTriang5   )
    e_GeneralError("Mesh contains non-supported element !");
}


static void create_nodes(MESH *mesh)
{
  int   i, i_cent, i_dim;
  int **nodes;
  int   no_nodes_in_elem, elem_index;

  mesh->MeshPointCoords = e_alloc_cont_d_matr(mesh->NMeshPoints, 3);

  for(i = 0; i < GV_global_mesh->Decomp.NSubDoms; i++)
  {
    mesh->MeshPointCoords[i][0] = G_center[i][0];
    mesh->MeshPointCoords[i][1] = G_center[i][1];
    mesh->MeshPointCoords[i][2] = 0.00;
  }

  mesh->KeywordSet[E_KW_NUM_MESHPOINT_COORDS] = TRUE;
}


static void create_links(MESH *mesh)
{

  char  graph_file[E_LINE_MAX];
  char  line[E_LINE_MAX];
  char  *pLine;
  FILE *fp;
  int   linecount, elem, elem_index, no_row, no_edge, i, dummy;

  strcpy(graph_file, GV_filename);
  strcat(graph_file, ".str.graph");
  printf("\n\nGraph file: %s", graph_file);

  if ((fp = fopen(graph_file, "r")) == NULL)
  {
    e_GeneralError("Can not open graph file for reading !");
  }

  e_GetNextLine(line, &linecount, graph_file, fp);
  if(sscanf(line, "%d %d %d", &no_row, &no_edge, &dummy) != 3)
    e_FileLineError(linecount, graph_file, 
        "Could not read parameters !");

  mesh->NElemsLink1 = no_edge;
  mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1] = TRUE;

  mesh->NodesLink1 = e_alloc_cont_i_matr(no_edge, 2);

  elem_index = 0;
  for(i = 0; i < no_row; i++)
  {
    e_GetNextLine(line, &linecount, graph_file, fp);
    pLine = GetStrArg(line,"%d",&dummy);
    while(1)
    {
      pLine = GetStrArg(pLine,"%d",&elem);
      if(pLine == NULL)
        break;
      else
      {
        if(elem > i+1)
        {
          mesh->NodesLink1[elem_index][0] = i+1;
          mesh->NodesLink1[elem_index][1] = elem;
          elem_index++;
        }
      }
#ifdef __WEIGHTED_VERTEX_EDGE__
      pLine = GetStrArg(pLine,"%d",&dummy);
      if(pLine == NULL)
        break;
#endif
    }
  }

  fclose(fp);

  mesh->KeywordSet[E_KW_NUM_NODES_LINK1] = TRUE;


  if(elem_index != no_edge)
    e_GeneralError("Edge and element number are different !");
}


static void create_mesh()
{
  int i;

  GV_new_mesh = (MESH *)e_calloc(1, sizeof(MESH));
  for(i = 0; i < E_NKEYWORDS; i++)
    GV_new_mesh->KeywordSet[i] = FALSE;

  for(i = 0; i < E_NSUBSTRUCTS; i++)
    GV_new_mesh->SubStructSet[i] = FALSE;

  GV_new_mesh->Title = e_alloc_c_vect(6);
  strncpy(GV_new_mesh->Title, "graph", 5);
  GV_new_mesh->Title[5] = '\0';
  GV_new_mesh->KeywordSet[E_KW_NUM_TITLE] = TRUE;

  GV_new_mesh->NMeshPoints = GV_global_mesh->Decomp.NSubDoms;
  GV_new_mesh->NNodes      = GV_global_mesh->Decomp.NSubDoms;
  GV_new_mesh->KeywordSet[E_KW_NUM_NMESHPOINTS] = TRUE;
  GV_new_mesh->KeywordSet[E_KW_NUM_NNODES] = TRUE;

  create_nodes(GV_new_mesh);

  create_links(GV_new_mesh);
}


static void Strip_Information_for_Type(int no, int **nodes, double **coords, 
                                       int *subdoms, double *StripArea)
{
  double    mid_x, mid_y, x21, y21, z21, x31, y31, z31, area;
  double    norm[3];
  int       n1, n2, n3, i;


  for(i = 0; i < no; i++)
  {
    n1 = nodes[i][0] - 1;
    n2 = nodes[i][1] - 1;
    n3 = nodes[i][2] - 1;

    mid_x = (coords[n1][0] + coords[n2][0] + coords[n3][0]) / 3.0;
    mid_y = (coords[n1][1] + coords[n2][1] + coords[n3][1]) / 3.0;

    x21 = coords[n2][0] - coords[n1][0];
    y21 = coords[n2][1] - coords[n1][1];
    z21 = coords[n2][2] - coords[n1][2];

    x31 = coords[n3][0] - coords[n1][0];
    y31 = coords[n3][1] - coords[n1][1];
    z31 = coords[n3][2] - coords[n1][2];

    norm[0] = (y21 * z31) - (y31 * z21);
    norm[1] = (z21 * x31) - (z31 * x21);
    norm[2] = (x21 * y31) - (x31 * y21);

    area = sqrt(E_SQUARE(norm[0]) + E_SQUARE(norm[1]) + E_SQUARE(norm[2]));
    area = area / 2;

    G_center[subdoms[i] - 1][0] += area * mid_x;
    G_center[subdoms[i] - 1][1] += area * mid_y;

    StripArea[subdoms[i] - 1] += area;
  }
}

static void Strip_Information_Creation()
{
  double   *StripArea;
  int       i;

  StripArea = e_alloc_d_vect(GV_global_mesh->Decomp.NSubDoms);
  G_center = e_alloc_cont_d_matr(GV_global_mesh->Decomp.NSubDoms, 2);

  Strip_Information_for_Type(GV_global_mesh->NElemsTriang3, GV_global_mesh->NodesTriang3, 
                             GV_global_mesh->MeshPointCoords, GV_global_mesh->Decomp.SubDomsTriang3,
                             StripArea);
  Strip_Information_for_Type(GV_global_mesh->NElemsTriang4, GV_global_mesh->NodesTriang4, 
                             GV_global_mesh->MeshPointCoords, GV_global_mesh->Decomp.SubDomsTriang4,
                             StripArea);
  Strip_Information_for_Type(GV_global_mesh->NElemsTriang5, GV_global_mesh->NodesTriang5, 
                             GV_global_mesh->MeshPointCoords, GV_global_mesh->Decomp.SubDomsTriang5,
                             StripArea);

/* create the center of each strip */
  for(i = 0; i < GV_global_mesh->Decomp.NSubDoms; i++)
  {
    G_center[i][0] = G_center[i][0] / StripArea[i];
    G_center[i][1] = G_center[i][1] / StripArea[i];
  }

  e_free(StripArea);
}


main(int argc, char *argv[])
{
  char  new_mesh[E_LINE_MAX];

  printf("\nProgram to create a mesh file representing graph connectivity");

  f_read_in_data(argc, argv);

/* Initialise the dual graph, 
   if there is an out-of-plane link, QUIT */
  NoNodeElem = e_alloc_i_vect(GV_global_mesh->NNodes);
  NodeElem = e_init_frag_i_matr(GV_global_mesh->NNodes);
  Degree = e_alloc_i_vect(GV_global_mesh->TotalNElems);
  Adjacency = e_init_frag_i_matr(GV_global_mesh->TotalNElems);

  e_InitMeshPointElemConnection(GV_global_mesh, NoNodeElem, NodeElem);
  SetDualAdjacency(GV_global_mesh, Degree, Adjacency,
                   NoNodeElem, NodeElem);

  /* creating the strip information */
  Create_Strips(GV_global_mesh, Degree, Adjacency);
  printf("\n\n    Number of strips: %d", GV_global_mesh->Decomp.NSubDoms);

  Strip_Information_Creation();
  create_mesh();

  strcpy(new_mesh, GV_filename);
  strcat(new_mesh, "_gr");

  printf("\nMesh file representing graph connectivity: %s", new_mesh);

  e_PutMeshData(GV_new_mesh, new_mesh);

  printf("\n\n");

}



