#include <stdio.h>
#include "e_lib.h"

#include "subdom_graph.h"

/* /////////////////////////////////////////////////////////////////////////// */
/* Create_Subdomain_Graph: To create the connectivity graph of the subdomains. */

/* The return value is 1 if the operations is successful, else 0. */

void Create_Subdomain_Graph(MESH *mesh, 
                            int *SubdomGraphDegree, int **SubdomGraph,
                            int **SubdomGraphRef)
{
  int    NInterSubDomNodes;
  int   *InterSubDomNodes;
  int   *NNodeSubDoms;
  int  **NodeSubDoms;

  int    i,j,k,jj,s1;
  int  **no_subdom_graph_ref;

  e_GetInterSubDomNodes(mesh, &NInterSubDomNodes, 
                        &InterSubDomNodes, &NNodeSubDoms, 
                        &NodeSubDoms);

  /* determine which subdomain is adjacent to which other,
     this information comes from the shared meshpoints,
     every subdomain stores the connection to the subdomains
     with higher ID number only, therefore connections are 
     not replicated */
  for(i = 0; i < NInterSubDomNodes; i++)
  {
    for(j = 0; j < NNodeSubDoms[i]; j++)
    {
      s1 = NodeSubDoms[i][j] - 1;
      for(jj = j+1; jj < NNodeSubDoms[i]; jj++)
      {
        for(k = 0; k < SubdomGraphDegree[s1]; k++)
        {
          if(SubdomGraph[s1][k] == NodeSubDoms[i][jj])
          {
            SubdomGraphRef[s1][k]++;
            break;
          }
        }
        if (k == SubdomGraphDegree[s1])
        {
          if(NodeSubDoms[i][j] < NodeSubDoms[i][jj])
          {
            SubdomGraph[s1][k] = NodeSubDoms[i][jj];
            SubdomGraphDegree[s1]++;
            SubdomGraphRef[s1][k]++;
          }
        }
      }
    }
  }

  e_FreeInterSubDomNodes(NInterSubDomNodes, InterSubDomNodes,
                         NNodeSubDoms, NodeSubDoms);
}
