#include "e_lib.h"


char        GV_filename[E_LINE_MAX];
MESH       *GV_global_mesh;


void f_read_in_data(int argc, char *argv[]);
void create_centre(int no_elem, int no_node, int **nodes, 
                   double **coords, FILE *fp);


void f_read_in_data(int argc, char *argv[])
{
  char   line[E_LINE_MAX];
  char  *check;

  if(argc > 1)
  {
    strcpy(GV_filename, argv[1]);
  }
  else
  {
    printf("\n\nType in the input filename (.mdf) !");
    printf("\nGive no extension : ");
    check = fgets(line, E_LINE_MAX, stdin);
    if(   (check == NULL)
       || (sscanf(line, "%s", &GV_filename) != 1))
      e_GeneralError("Error in keyboard input, can not read filename!");
  }

  printf("\nThe projectname is taken as: %s", GV_filename);

  e_GetMeshData(&GV_global_mesh, GV_filename);

  if(GV_global_mesh->TotalNElems != GV_global_mesh->NElemsLink1 +
                                    GV_global_mesh->NElemsLink2 +
                                    GV_global_mesh->NElemsLink3 +
                                    GV_global_mesh->NElemsLink4 +
                                    GV_global_mesh->NElemsLink5 +
                                    GV_global_mesh->NElemsTriang3 +
                                    GV_global_mesh->NElemsTriang4 +
                                    GV_global_mesh->NElemsTriang5   )
    e_GeneralError("Mesh contains non-supported element !");
}


void create_centre(int no_elem, int no_node, int **nodes, 
                   double **coords, FILE *fp)
{
  int i, j, n;
  double xc, yc, zc;

  for(i = 0; i < no_elem; i++)
  {
    xc = 0.0;
    yc = 0.0;
    zc = 0.0;
    for(j = 0; j < no_node; j ++)
    {
      n = nodes[i][j]-1;
      xc += coords[n][0];
      yc += coords[n][1];
      zc += coords[n][2];
    }
    xc /= no_node;
    yc /= no_node;
    zc /= no_node;
    fprintf(fp, "%lf %lf %lf\n", xc, yc, zc);

  }
}



main(int argc, char *argv[])
{
  char coord_file[E_LINE_MAX];
  FILE  *fp;
  char line[E_LINE_MAX];

  f_read_in_data(argc, argv);

  strcpy(coord_file, GV_filename);
  strcat(coord_file, ".cor");

  if ((fp = fopen(coord_file, "w")) == NULL)
  {
    e_GeneralError("Can not open coordinate file !");
  }

  create_centre(GV_global_mesh->NElemsLink1, 2, 
                GV_global_mesh->NodesLink1, 
                GV_global_mesh->MeshPointCoords, fp);
  create_centre(GV_global_mesh->NElemsLink2, 2, 
                GV_global_mesh->NodesLink2, 
                GV_global_mesh->MeshPointCoords, fp);
  create_centre(GV_global_mesh->NElemsLink3, 2, 
                GV_global_mesh->NodesLink3, 
                GV_global_mesh->MeshPointCoords, fp);
  create_centre(GV_global_mesh->NElemsLink4, 2, 
                GV_global_mesh->NodesLink4, 
                GV_global_mesh->MeshPointCoords, fp);
  create_centre(GV_global_mesh->NElemsLink5, 2, 
                GV_global_mesh->NodesLink5, 
                GV_global_mesh->MeshPointCoords, fp);

  create_centre(GV_global_mesh->NElemsTriang3, 3, 
                GV_global_mesh->NodesTriang3, 
                GV_global_mesh->MeshPointCoords, fp);
  create_centre(GV_global_mesh->NElemsTriang4, 3, 
                GV_global_mesh->NodesTriang4, 
                GV_global_mesh->MeshPointCoords, fp);
  create_centre(GV_global_mesh->NElemsTriang5, 3, 
                GV_global_mesh->NodesTriang5, 
                GV_global_mesh->MeshPointCoords, fp);

  fprintf(fp, "\n");
  fclose(fp);

  printf("\n\n");

}
