/* Convert a "graph.ptn" file into a "dom" file */


#include <stdio.h>
#include <math.h>
#include <e_lib.h>


char        GV_filename[E_LINE_MAX];
MESH       *GV_global_mesh;


void f_read_in_data(int argc, char *argv[]);
void create_subdom_struct(MESH *mesh);


void f_read_in_data(int argc, char *argv[])
{
  char   line[E_LINE_MAX];
  char  *check;

  if(argc > 1)
  {
    strcpy(GV_filename, argv[1]);
  }
  else
  {
    printf("\n\nType in the input filename (.mdf) !");
    printf("\nGive no extension : ");
    check = fgets(line, E_LINE_MAX, stdin);
    if(   (check == NULL)
       || (sscanf(line, "%s", &GV_filename) != 1))
      e_GeneralError("Error in keyboard input, can not read filename!");
  }

  printf("\nThe projectname is taken as: %s", GV_filename);

  e_GetMeshData(&GV_global_mesh, GV_filename);

  if(GV_global_mesh->TotalNElems != GV_global_mesh->NElemsLink1 +
                                    GV_global_mesh->NElemsLink2 +
                                    GV_global_mesh->NElemsLink3 +
                                    GV_global_mesh->NElemsLink4 +
                                    GV_global_mesh->NElemsLink5 +
                                    GV_global_mesh->NElemsTriang3 +
                                    GV_global_mesh->NElemsTriang4 +
                                    GV_global_mesh->NElemsTriang5   )
    e_GeneralError("Mesh contains non-supported element !");
}


void create_subdom_struct(MESH *mesh)
{
  int i;

  mesh->Decomp.SubDoms = e_alloc_i_vect(mesh->TotalNElems);

  for(i = 0; i < mesh->TotalNElems; i++)
    mesh->Decomp.SubDoms[i] = 1;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
    mesh->Decomp.SubDomsLink1 = mesh->Decomp.SubDoms +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_LINK1] - 1;
  else
    mesh->Decomp.SubDomsLink1 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
    mesh->Decomp.SubDomsLink2 = mesh->Decomp.SubDoms +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_LINK2] - 1;
  else
    mesh->Decomp.SubDomsLink2 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
    mesh->Decomp.SubDomsLink3 = mesh->Decomp.SubDoms +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_LINK3] - 1;
  else
    mesh->Decomp.SubDomsLink3 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
    mesh->Decomp.SubDomsLink4 = mesh->Decomp.SubDoms +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_LINK4] - 1;
  else
    mesh->Decomp.SubDomsLink4 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
    mesh->Decomp.SubDomsLink5 = mesh->Decomp.SubDoms +
                                  mesh->FirstNetIndex[E_ELEM_TYPE_LINK5] - 1;
  else
    mesh->Decomp.SubDomsLink5 = NULL;

  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
    mesh->Decomp.SubDomsTriang3 = mesh->Decomp.SubDoms +
                                    mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] - 1;
  else
    mesh->Decomp.SubDomsTriang3 = NULL;
  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
    mesh->Decomp.SubDomsTriang4 = mesh->Decomp.SubDoms +
                                    mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] - 1;
  else
    mesh->Decomp.SubDomsTriang4 = NULL;
  if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
    mesh->Decomp.SubDomsTriang5 = mesh->Decomp.SubDoms +
                                    mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] - 1;
  else
    mesh->Decomp.SubDomsTriang5 = NULL;

  mesh->Decomp.NSubDoms = 1;
  mesh->SubStructSet[E_SS_NUM_DECOMP] = TRUE;
}




main(int argc, char *argv[])
{
  char part_file[E_LINE_MAX];
  FILE  *fp;
  char line[E_LINE_MAX];
  int i, linecount, subdom, no_subdom = 0;

  printf("\n\nConvert a \"graph.ptn\" file into a \"dom\" file\n");

  f_read_in_data(argc, argv);

  create_subdom_struct(GV_global_mesh);

  strcpy(part_file, GV_filename);
  strcat(part_file, ".graph.ptn");

  if ((fp = fopen(part_file, "r")) == NULL)
  {
    e_GeneralError("Can not open partition file !");
  }

  printf("\nPartition file: %s", part_file);

  for(i = 0; i < GV_global_mesh->TotalNElems; i++)
  {
    e_GetNextLine(line, &linecount, part_file, fp);
    if(sscanf(line, "%d", &subdom) != 1)
      e_FileLineError(linecount, part_file, "Cannot interpret the line !");
    GV_global_mesh->Decomp.SubDoms[i] = subdom+1;
    if(no_subdom < subdom+1)
      no_subdom = subdom+1;
  }

  GV_global_mesh->Decomp.NSubDoms = no_subdom;

  fclose(fp);

  e_PutDecompData(GV_global_mesh, GV_filename);

  printf("\n\n");

}
