#include "e_lib.h"


char        GV_filename[E_LINE_MAX];
MESH       *GV_global_mesh;


void f_read_in_data(int argc, char *argv[]);


void f_read_in_data(int argc, char *argv[])
{
  char   line[E_LINE_MAX];
  char  *check;

  if(argc > 1)
  {
    strcpy(GV_filename, argv[1]);
  }
  else
  {
    printf("\n\nType in the input filename (.mdf) !");
    printf("\nGive no extension : ");
    check = fgets(line, E_LINE_MAX, stdin);
    if(   (check == NULL)
       || (sscanf(line, "%s", &GV_filename) != 1))
      e_GeneralError("Error in keyboard input, can not read filename!");
  }

  printf("\nThe projectname is taken as: %s", GV_filename);

  e_GetMeshData(&GV_global_mesh, GV_filename);
  e_GetDecompData(GV_global_mesh, GV_filename);

  if(GV_global_mesh->TotalNElems != GV_global_mesh->NElemsLink1 +
                                    GV_global_mesh->NElemsLink2 +
                                    GV_global_mesh->NElemsLink3 +
                                    GV_global_mesh->NElemsLink4 +
                                    GV_global_mesh->NElemsLink5 +
                                    GV_global_mesh->NElemsTriang3 +
                                    GV_global_mesh->NElemsTriang4 +
                                    GV_global_mesh->NElemsTriang5   )
    e_GeneralError("Mesh contains non-supported element !");
}


void create_centre(int no_elem, int no_node, int **nodes, 
                   double **coords, int *Elemsubdom,
                   double **subdom_centre, int *subdom_elem)
{
  int i, j, n, subdom;
  double xc, yc, zc;

  for(i = 0; i < no_elem; i++)
  {
    xc = 0.0;
    yc = 0.0;
    zc = 0.0;
    for(j = 0; j < no_node; j ++)
    {
      n = nodes[i][j]-1;
      xc += coords[n][0];
      yc += coords[n][1];
      zc += coords[n][2];
    }
    xc /= no_node;
    yc /= no_node;
    zc /= no_node;

    subdom = Elemsubdom[i]-1;

    subdom_centre[subdom][0] += xc;
    subdom_centre[subdom][1] += yc;
    subdom_centre[subdom][2] += zc;

    subdom_elem[subdom]++;
  }
}



main(int argc, char *argv[])
{
  char coord_file[E_LINE_MAX];
  FILE  *fp;
  char line[E_LINE_MAX];
  int  i;
  double  **Subdom_centre;
  int      *Subdom_elem;

  f_read_in_data(argc, argv);

  Subdom_centre = e_alloc_cont_d_matr(GV_global_mesh->Decomp.NSubDoms, 3);
  Subdom_elem = e_alloc_i_vect(GV_global_mesh->Decomp.NSubDoms);

  create_centre(GV_global_mesh->NElemsLink1, 2, 
                GV_global_mesh->NodesLink1, 
                GV_global_mesh->MeshPointCoords, 
                GV_global_mesh->Decomp.SubDomsLink1,
                Subdom_centre, Subdom_elem);
  create_centre(GV_global_mesh->NElemsLink2, 2, 
                GV_global_mesh->NodesLink2, 
                GV_global_mesh->MeshPointCoords, 
                GV_global_mesh->Decomp.SubDomsLink2,
                Subdom_centre, Subdom_elem);
  create_centre(GV_global_mesh->NElemsLink3, 2, 
                GV_global_mesh->NodesLink3, 
                GV_global_mesh->MeshPointCoords, 
                GV_global_mesh->Decomp.SubDomsLink3,
                Subdom_centre, Subdom_elem);
  create_centre(GV_global_mesh->NElemsLink4, 2, 
                GV_global_mesh->NodesLink4, 
                GV_global_mesh->MeshPointCoords, 
                GV_global_mesh->Decomp.SubDomsLink4,
                Subdom_centre, Subdom_elem);
  create_centre(GV_global_mesh->NElemsLink5, 2, 
                GV_global_mesh->NodesLink5, 
                GV_global_mesh->MeshPointCoords, 
                GV_global_mesh->Decomp.SubDomsLink5,
                Subdom_centre, Subdom_elem);

  create_centre(GV_global_mesh->NElemsTriang3, 3, 
                GV_global_mesh->NodesTriang3, 
                GV_global_mesh->MeshPointCoords, 
                GV_global_mesh->Decomp.SubDomsTriang3,
                Subdom_centre, Subdom_elem);
  create_centre(GV_global_mesh->NElemsTriang4, 3, 
                GV_global_mesh->NodesTriang4, 
                GV_global_mesh->MeshPointCoords, 
                GV_global_mesh->Decomp.SubDomsTriang4,
                Subdom_centre, Subdom_elem);
  create_centre(GV_global_mesh->NElemsTriang5, 3, 
                GV_global_mesh->NodesTriang5, 
                GV_global_mesh->MeshPointCoords, 
                GV_global_mesh->Decomp.SubDomsTriang5,
                Subdom_centre, Subdom_elem);

  strcpy(coord_file, GV_filename);
  strcat(coord_file, ".part.cor");

  if ((fp = fopen(coord_file, "w")) == NULL)
  {
    e_GeneralError("Can not open coordinate file !");
  }

  for(i = 0; i < GV_global_mesh->Decomp.NSubDoms; i++)
  {
    Subdom_centre[i][0] /= Subdom_elem[i];
    Subdom_centre[i][1] /= Subdom_elem[i];
    Subdom_centre[i][2] /= Subdom_elem[i];
    fprintf(fp, "%lf %lf %lf\n", Subdom_centre[i][0], 
            Subdom_centre[i][1], Subdom_centre[i][2]);
  }


  fprintf(fp, "\n");
  fclose(fp);

  printf("\n\n");

}
