/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                       Memory Deallocation Functions                        */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include "e_pltfrm.h"
#include <stdio.h>
#include <e_lib.h>


/* ////////////////////////////////////////////////////////////////////////// */
/* e_FreeMeshData: To deallocate all memory of the MESH structure. */


void e_FreeMeshData(MESH *mesh)
{
   int k;


   if (mesh == NULL) return;

   e_free(mesh->Title);

   if (mesh->KeywordSet[E_KW_NUM_MESHPOINT_COORDS])
      e_free_cont_d_matr(mesh->NMeshPoints,mesh->MeshPointCoords);

   if (mesh->KeywordSet[E_KW_NUM_NODES_LINK1])
      e_free_cont_i_matr(mesh->NElemsLink1,mesh->NodesLink1);

   if (mesh->KeywordSet[E_KW_NUM_NODES_LINK2])
      e_free_cont_i_matr(mesh->NElemsLink2,mesh->NodesLink2);

   if (mesh->KeywordSet[E_KW_NUM_NODES_LINK3])
      e_free_cont_i_matr(mesh->NElemsLink3,mesh->NodesLink3);

   if (mesh->KeywordSet[E_KW_NUM_NODES_LINK4])
      e_free_cont_i_matr(mesh->NElemsLink4,mesh->NodesLink4);

   if (mesh->KeywordSet[E_KW_NUM_NODES_LINK5])
      e_free_cont_i_matr(mesh->NElemsLink5,mesh->NodesLink5);

   if (mesh->KeywordSet[E_KW_NUM_NODES_TRIANG1])
      e_free_cont_i_matr(mesh->NElemsTriang1,mesh->NodesTriang1);

   if (mesh->KeywordSet[E_KW_NUM_NODES_TRIANG2])
      e_free_cont_i_matr(mesh->NElemsTriang2,mesh->NodesTriang2);

   if (mesh->KeywordSet[E_KW_NUM_NODES_TRIANG3])
      e_free_cont_i_matr(mesh->NElemsTriang3,mesh->NodesTriang3);

   if (mesh->KeywordSet[E_KW_NUM_NODES_TRIANG4])
      e_free_cont_i_matr(mesh->NElemsTriang4,mesh->NodesTriang4);

   if (mesh->KeywordSet[E_KW_NUM_NODES_TRIANG5])
      e_free_cont_i_matr(mesh->NElemsTriang5,mesh->NodesTriang5);

   if (mesh->KeywordSet[E_KW_NUM_NODES_QUAD1])
      e_free_cont_i_matr(mesh->NElemsQuad1,mesh->NodesQuad1);

   if (mesh->KeywordSet[E_KW_NUM_NODES_QUAD3])
      e_free_cont_i_matr(mesh->NElemsQuad3,mesh->NodesQuad3);

   if (mesh->KeywordSet[E_KW_NUM_NODES_TETRAH1])
      e_free_cont_i_matr(mesh->NElemsTetrah1,mesh->NodesTetrah1);

   if (mesh->KeywordSet[E_KW_NUM_NODES_BLOCK1])
      e_free_cont_i_matr(mesh->NElemsBlock1,mesh->NodesBlock1);

   e_free(mesh->NodalDOF);

   if (mesh->KeywordSet[E_KW_NUM_BCS])
   {
      e_free(mesh->BCNodes);
      e_free_frag_i_matr(mesh->NBCNodes,mesh->BCTypes);
      e_free_frag_d_matr(mesh->NBCNodes,mesh->BCDispls);
      e_free_frag_d_matr(mesh->NBCNodes,mesh->BCSpringConsts);
   }

   if (mesh->KeywordSet[E_KW_NUM_LOADS])
   {
      e_free(mesh->LoadedNodes);
      e_free_frag_d_matr(mesh->NLoadedNodes,mesh->Loads);
   }

   if (mesh->SubStructSet[E_SS_NUM_GEOM])
      e_FreeGeomModel(mesh);

   if (mesh->KeywordSet[E_KW_NUM_REMESH_DATA])
   {
      for (k = 0; k < mesh->Geom.NNurbsCurves; k++)
         e_free(mesh->Geom.NurbsCurves[k].Name);

      e_free(mesh->Geom.NurbsCurves);
      e_free(mesh->NodeTypes);
      e_free(mesh->NNurbsCurves);
      e_free_frag_d_matr(mesh->NMeshPoints,mesh->NurbsCurveParams);

      for (k = 0; k < mesh->NMeshPoints; k++)
         e_free(mesh->NurbsCurves[k]);

      e_free(mesh->NurbsCurves);
   }

   e_free(mesh->GNIs);
   e_free(mesh->GEIs);

   /* !!!!! it relies on material information !!!!! */
   e_FreeStresses(mesh);

   if (mesh->SubStructSet[E_SS_NUM_MAT_PROPS])
   {
      for (k = 0; k < mesh->NCompMatsType1; k++)
      {
         e_free(mesh->CompPropsType1[k].LayerMats);
         e_free(mesh->CompPropsType1[k].Thicknesses);
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NMATS])
   {
      for (k = 0; k < mesh->NMats; k++)
         e_free(mesh->Props[k].Name);

      e_free(mesh->Mats);
      e_free(mesh->Props);
   }

   if (mesh->KeywordSet[E_KW_NUM_NCOMP_MATS_TYPE1])
   {
      for (k = 0; k < mesh->NCompMatsType1; k++)
         e_free(mesh->CompPropsType1[k].Name);

      e_free(mesh->CompPropsType1);
   }

   e_FreeDecompData(mesh);
   e_FreeMeshParams(mesh);
   e_FreeFEErrors(mesh);

   e_free(mesh);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_FreeGeomModel: To deallocate (.gmf) geometric model data. */


void e_FreeGeomModel(MESH *mesh)
{
   int k;


   if (mesh == NULL || !mesh->SubStructSet[E_SS_NUM_GEOM])
      return;

   e_free_cont_d_matr(mesh->Geom.NEndPoints,mesh->Geom.EndPointCoords);

   for (k = 0; k < mesh->Geom.NNurbsCurves; k++)
   {
      e_free_cont_d_matr(mesh->Geom.NurbsCurves->Degree + 1,
                         mesh->Geom.NurbsCurves[k].CntlPoints);

      if (mesh->Geom.NurbsCurves[k].FixityType != E_NC_NUM_NO_FIXITY)
      {
         e_free(mesh->Geom.NurbsCurves[k].BCTypes);
         e_free(mesh->Geom.NurbsCurves[k].BCDispls);
         e_free(mesh->Geom.NurbsCurves[k].BCSpringConsts);
      }

      if (mesh->Geom.NurbsCurves[k].LoadTypeSet[E_NC_NUM_POINT_LOAD])
         e_free(mesh->Geom.NurbsCurves[k].PointLoad);

      if (mesh->Geom.NurbsCurves[k].LoadTypeSet[E_NC_NUM_LIN_LOAD_GC])
      {
         e_free(mesh->Geom.NurbsCurves[k].LinLoadGCBeg);
         e_free(mesh->Geom.NurbsCurves[k].LinLoadGCEnd);
      }

      if (mesh->Geom.NurbsCurves[k].LoadTypeSet[E_NC_NUM_LIN_LOAD_CC])
      {
         e_free(mesh->Geom.NurbsCurves[k].LinLoadCCBeg);
         e_free(mesh->Geom.NurbsCurves[k].LinLoadCCEnd);
      }
   }

   mesh->SubStructSet[E_SS_NUM_GEOM] = FALSE;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_FreeMatProps: To deallocate (.mat) material properties data. */


void e_FreeMatProps(MESH *mesh)
{
   int k;


   if (mesh == NULL || !mesh->SubStructSet[E_SS_NUM_MAT_PROPS])
      return;

   for (k = 0; k < mesh->NCompMatsType1; k++)
   {
      e_free(mesh->CompPropsType1[k].LayerMats);
      e_free(mesh->CompPropsType1[k].Thicknesses);
   }

   for (k = mesh->NMdfMats; k < mesh->NMats; k++)
   {
      e_free(mesh->Props[k].Name);
      mesh->Props[k].Name = NULL;
   }

   mesh->SubStructSet[E_SS_NUM_MAT_PROPS] = FALSE;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_FreeDecompData: To deallocate all memory of the DECOMP structure. */


void e_FreeDecompData(MESH *mesh)
{
   if (mesh == NULL || !mesh->SubStructSet[E_SS_NUM_DECOMP])
      return;

   e_free(mesh->Decomp.SubDoms);

   mesh->SubStructSet[E_SS_NUM_DECOMP] = FALSE;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_FreeNodalMeshParams: To free memory allocated for the nodal mesh
   parameters.                                                                */


void e_FreeNodalMeshParams(MESH *mesh)
{
   if (mesh == NULL || !mesh->SubStructSet[E_SS_NUM_MESHPARAM])
      return;

   if (mesh->MeshParam.NodalMeshParamsSet)
      e_free(mesh->MeshParam.NodalMeshParams);

   mesh->MeshParam.NodalMeshParamsSet = FALSE;

   if (!mesh->MeshParam.ElemMeshParamsSet)
      mesh->SubStructSet[E_SS_NUM_MESHPARAM] = FALSE;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_FreeElemMeshParams: To free memory allocated for the element mesh
   parameters.                                                                */


void e_FreeElemMeshParams(MESH *mesh)
{
   if (mesh == NULL || !mesh->SubStructSet[E_SS_NUM_MESHPARAM])
      return;

   if (mesh->MeshParam.ElemMeshParamsSet)
      e_free(mesh->MeshParam.ElemMeshParams);

   mesh->MeshParam.ElemMeshParamsSet = FALSE;

   if (!mesh->MeshParam.NodalMeshParamsSet)
      mesh->SubStructSet[E_SS_NUM_MESHPARAM] = FALSE;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_FreeMeshParams: To deallocate all memory of the MESHPARAMS structure. */


void e_FreeMeshParams(MESH *mesh)
{
   e_FreeNodalMeshParams(mesh);
   e_FreeElemMeshParams(mesh);
}

/* ////////////////////////////////////////////////////////////////////////// */
/* e_FreeStresses: To deallocate all memory of the STRESS structure. */


void e_FreeStresses(MESH *mesh)
{
  int i, maxNLayerTriang2 = 0, maxNLayerQuad3 = 0;

   if (mesh == NULL || !mesh->SubStructSet[E_SS_NUM_STRESS])
      return;

   if (   (   (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2]) 
           || (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])   )
       && (!mesh->SubStructSet[E_SS_NUM_MAT_PROPS])         )
   {
      e_GeneralError("No material information is available for freeing (Triang2/Quad3).");
   }

   for(i = 0; i < mesh->NElemsTriang2; i++)
     if(maxNLayerTriang2 < mesh->CompMatsTriang2[i]->NLayers)
       maxNLayerTriang2 = mesh->CompMatsTriang2[i]->NLayers;

   for(i = 0; i < mesh->NElemsQuad3; i++)
     if(maxNLayerQuad3 < mesh->CompMatsQuad3[i]->NLayers)
       maxNLayerQuad3 = mesh->CompMatsQuad3[i]->NLayers;

   if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
     e_free_cont_d_3D_matr(mesh->NElemsLink1, 
                           mesh->NStressPointsLink1, 
                           mesh->Stresses.StressesLink1);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
     e_free_cont_d_3D_matr(mesh->NElemsLink2, 
                           mesh->NStressPointsLink2, 
                           mesh->Stresses.StressesLink2);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
     e_free_cont_d_3D_matr(mesh->NElemsLink3, 
                           mesh->NStressPointsLink3, 
                           mesh->Stresses.StressesLink3);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
     e_free_cont_d_3D_matr(mesh->NElemsLink4, 
                           mesh->NStressPointsLink4, 
                           mesh->Stresses.StressesLink4);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
     e_free_cont_d_3D_matr(mesh->NElemsLink5, 
                           mesh->NStressPointsLink5, 
                           mesh->Stresses.StressesLink5);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
     e_free_cont_d_3D_matr(mesh->NElemsTriang1, 
                           mesh->NStressPointsTriang1, 
                           mesh->Stresses.StressesTriang1);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
     e_free_cont_d_4D_matr(mesh->NElemsTriang2, 
                           maxNLayerTriang2, 
                           mesh->NStressPointsTriang2, 
                           mesh->Stresses.StressesTriang2);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
     e_free_cont_d_3D_matr(mesh->NElemsTriang3, 
                           mesh->NStressPointsTriang3, 
                           mesh->Stresses.StressesTriang3);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
     e_free_cont_d_3D_matr(mesh->NElemsTriang4, 
                           mesh->NStressPointsTriang4, 
                           mesh->Stresses.StressesTriang4);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
     e_free_cont_d_3D_matr(mesh->NElemsTriang5, 
                           mesh->NStressPointsTriang5, 
                           mesh->Stresses.StressesTriang5);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
     e_free_cont_d_3D_matr(mesh->NElemsQuad1, 
                           mesh->NStressPointsQuad1, 
                           mesh->Stresses.StressesQuad1);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
     e_free_cont_d_4D_matr(mesh->NElemsQuad3, 
                           maxNLayerQuad3, 
                           mesh->NStressPointsQuad3, 
                           mesh->Stresses.StressesQuad3);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
     e_free_cont_d_3D_matr(mesh->NElemsTetrah1, 
                           mesh->NStressPointsTetrah1, 
                           mesh->Stresses.StressesTetrah1);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
     e_free_cont_d_3D_matr(mesh->NElemsBlock1, 
                           mesh->NStressPointsBlock1, 
                           mesh->Stresses.StressesBlock1);

   mesh->SubStructSet[E_SS_NUM_STRESS] = FALSE;

}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_FreeFEErrors: To deallocate all memory of the FEERROR structure. */


void e_FreeFEErrors(MESH *mesh)
{
   if (mesh == NULL || !mesh->SubStructSet[E_SS_NUM_FEERROR])
      return;

   e_free(mesh->FEError.FEErrors);

   mesh->SubStructSet[E_SS_NUM_FEERROR] = FALSE;
}
