/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                     e_GetMaterials(), e_PutMaterials()                     */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include "e_pltfrm.h"
#include <stdio.h>
#include <string.h>
#include <e_lib.h>

#include "comsrc.h"
#include "matprops.h"


/* ////////////////////////////////////////////////////////////////////////// */
/* e_GetMaterials: To read in material properties. */
/* The return value is 1 if no input errors occur, else 0. */


int e_GetMaterials(MESH *mesh, char *InputFile)
{
   if (!mesh->KeywordSet[E_KW_NUM_NMATS])
   {
      e_GeneralError("e_GetMaterials: material types not declared");
      return 0;
   }
   else if ((FullPathName = OpenInputFile(InputFile,E_MAT,&fp)) == NULL)
   {
      return 0;
   }

   if (   !Initialise(mesh)
       || !GetCompMaterialsType1()
       || !GetUsedMaterials()     )
   {
      DeallocateAllMemory();
      fclose(fp);
      return 0;
   }

   SetMeshData(mesh);

   e_free(FullPathName);
   fclose(fp);
   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* initialise: To initialise variables. */

/* The return value is 1 if the variables are initialised successfully,
   else 0.                                                                    */


static int Initialise(MESH *mesh)
{
   int i,k;


   for (i = 0; i < mesh->NCompMatsType1; i++)
   {
      mesh->CompPropsType1[i].LayerMats = NULL;
      mesh->CompPropsType1[i].Thicknesses = NULL;
   }

   for (i = mesh->NMdfMats; i < mesh->NMats; i++)
      mesh->Props[i].Name = NULL;

   MaterialDefined = NULL;
   CompMatDefinedType1 = NULL;
   KeywordSet = NULL;

   if (                (MaterialDefined = e_alloc_c_vect(mesh->NMats)) == NULL
       || (CompMatDefinedType1 = e_alloc_c_vect(mesh->NCompMatsType1)) == NULL
       || (KeywordSet
              = e_alloc_frag_i_matr(mesh->NMats,E_NMAT_PROP_KEYWORDS)) == NULL)
   {
      return 0;
   }

   strcpy(Keyword[E_MKW_NUM_MAT_TYPE],E_MKW_MAT_TYPE);
   strcpy(Keyword[E_MKW_NUM_DENSITY],E_MKW_DENSITY);
   strcpy(Keyword[E_MKW_NUM_YMOD],E_MKW_YMOD);
   strcpy(Keyword[E_MKW_NUM_YMOD_X],E_MKW_YMOD_X);
   strcpy(Keyword[E_MKW_NUM_YMOD_Y],E_MKW_YMOD_Y);
   strcpy(Keyword[E_MKW_NUM_YMOD_Z],E_MKW_YMOD_Z);
   strcpy(Keyword[E_MKW_NUM_POISSONS_RATIO],E_MKW_POISSONS_RATIO);
   strcpy(Keyword[E_MKW_NUM_THICKNESS],E_MKW_THICKNESS);

   strcpy(Keyword[E_MKW_NUM_IPARAM1],E_MKW_IPARAM1);
   strcpy(Keyword[E_MKW_NUM_IPARAM2],E_MKW_IPARAM2);
   strcpy(Keyword[E_MKW_NUM_IPARAM3],E_MKW_IPARAM3);
   strcpy(Keyword[E_MKW_NUM_IPARAM4],E_MKW_IPARAM4);
   strcpy(Keyword[E_MKW_NUM_IPARAM5],E_MKW_IPARAM5);
   strcpy(Keyword[E_MKW_NUM_IPARAM6],E_MKW_IPARAM6);

   strcpy(Keyword[E_MKW_NUM_DPARAM1],E_MKW_DPARAM1);
   strcpy(Keyword[E_MKW_NUM_DPARAM2],E_MKW_DPARAM2);
   strcpy(Keyword[E_MKW_NUM_DPARAM3],E_MKW_DPARAM3);
   strcpy(Keyword[E_MKW_NUM_DPARAM4],E_MKW_DPARAM4);
   strcpy(Keyword[E_MKW_NUM_DPARAM5],E_MKW_DPARAM5);
   strcpy(Keyword[E_MKW_NUM_DPARAM6],E_MKW_DPARAM6);
   strcpy(Keyword[E_MKW_NUM_DPARAM7],E_MKW_DPARAM7);
   strcpy(Keyword[E_MKW_NUM_DPARAM8],E_MKW_DPARAM8);
   strcpy(Keyword[E_MKW_NUM_DPARAM9],E_MKW_DPARAM9);
   strcpy(Keyword[E_MKW_NUM_DPARAM10],E_MKW_DPARAM10);
   strcpy(Keyword[E_MKW_NUM_DPARAM11],E_MKW_DPARAM11);
   strcpy(Keyword[E_MKW_NUM_DPARAM12],E_MKW_DPARAM12);
   strcpy(Keyword[E_MKW_NUM_DPARAM13],E_MKW_DPARAM13);
   strcpy(Keyword[E_MKW_NUM_DPARAM14],E_MKW_DPARAM14);
   strcpy(Keyword[E_MKW_NUM_DPARAM15],E_MKW_DPARAM15);
   strcpy(Keyword[E_MKW_NUM_DPARAM16],E_MKW_DPARAM16);
   strcpy(Keyword[E_MKW_NUM_DPARAM17],E_MKW_DPARAM17);
   strcpy(Keyword[E_MKW_NUM_DPARAM18],E_MKW_DPARAM18);
   strcpy(Keyword[E_MKW_NUM_DPARAM19],E_MKW_DPARAM19);
   strcpy(Keyword[E_MKW_NUM_DPARAM20],E_MKW_DPARAM20);

   strcpy(Keyword[E_MKW_NUM_END],E_MKW_END);

   NMats = mesh->NMats;
   NCompMatsType1 = mesh->NCompMatsType1;
   Props = mesh->Props;
   CompPropsType1 = mesh->CompPropsType1;
   CurrentNMats = mesh->NMdfMats;
   NMdfMats = mesh->NMdfMats;
   CurrentNCompMatsType1 = 0;
   LineNumber = 0;

   for (i = 0; i < mesh->NCompMatsType1; i++)
      CompMatDefinedType1[i] = FALSE;

   for (i = 0; i < mesh->NMats; i++)
   {
      MaterialDefined[i] = FALSE;

      for (k = 0; k < E_NMAT_PROP_KEYWORDS; k++)
         KeywordSet[i][k] = FALSE;
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetCompMaterialsType1: To read in the TYPE1 composite material data. */

/* The return value is 1 if the data are read successfully or the end of file
   is encountered, else 0.                                                    */


static int GetCompMaterialsType1()
{
   int k;
   int CompPropsIndex;
   int ReturnValue;

   char *CompMaterialName;
   char *pFirstQuote;
   char *pLastQuote;


   if ((CompMaterialName = e_alloc_c_vect(E_LINE_MAX + 1)) == NULL)
      return 0;

   while ((ReturnValue
              = GetKeyword(line,word,&LineNumber,FullPathName,fp)) == 1)
   {
      if (strcmp(E_MKW_COMP_MAT_TYPE1,word) == 0)
      {
         if (         sscanf(line,"%*s %s",CompMaterialName) != 1
             || (pFirstQuote = strchr(CompMaterialName,'"')) == NULL
             || (pLastQuote = strrchr(CompMaterialName,'"')) == pFirstQuote
             ||              (int)(pLastQuote - pFirstQuote) == 1          )
         {
            e_FileLineError(LineNumber,FullPathName,
                            "cannot read composite material name");
            e_free(CompMaterialName);
            return 0;
         }
         else
         {
            CompPropsIndex = GetCompPropsType1Index(CompMaterialName);

            if (CompPropsIndex > 0)
            {
               k = CompPropsIndex - 1;

               if (CompMatDefinedType1[k])
               {
                  e_FileLineError(LineNumber,FullPathName,
                                  "composite material multiply defined");
                  e_free(CompMaterialName);
                  return 0;
               }
               else
                  CompMatDefinedType1[k] = TRUE;

               CurrentNCompMatsType1++;

               if (!GetCompMatPropsType1(&CompPropsType1[k]))
               {
                  e_free(CompMaterialName);
                  return 0;
               }
            }
         }
      }
   }

   e_free(CompMaterialName);

   if (ReturnValue == EOF)
   {
      if (CurrentNCompMatsType1 != NCompMatsType1)
      {
         e_GeneralError("incorrect number of TYPE1 composite materials");
         return 0;
      }
      else if (CurrentNMats != NMats)
      {
         e_GeneralError("incorrect number of materials");
         return 0;
      }
      else
         return 1;
   }
   else
      return 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetUsedMaterials: To read in the material properties. */

/* The return value is 1 if the data are read successfully or the end of file
   is encountered, else 0.                                                    */


static int GetUsedMaterials()
{
   int k;
   int NDefinedMaterials;
   int PropsIndex;
   int ReturnValue;

   char *pFirstQuote;
   char *pLastQuote;
   char *UsedMaterialName;


   rewind(fp);
   LineNumber = 0;

   if ((UsedMaterialName = e_alloc_c_vect(E_LINE_MAX + 1)) == NULL)
      return 0;

   NDefinedMaterials = 0;

   while ((ReturnValue
              = GetKeyword(line,word,&LineNumber,FullPathName,fp)) == 1)
   {
      if (strcmp(E_MKW_MAT,word) == 0)
      {
         if (         sscanf(line,"%*s %s",UsedMaterialName) != 1
             || (pFirstQuote = strchr(UsedMaterialName,'"')) == NULL
             || (pLastQuote = strrchr(UsedMaterialName,'"')) == pFirstQuote
             ||              (int)(pLastQuote - pFirstQuote) == 1          )
         {
            e_FileLineError(LineNumber,FullPathName,
                            "cannot read material name");
            e_free(UsedMaterialName);
            return 0;
         }
         else
         {
            PropsIndex = GetPropsIndex(UsedMaterialName);

            if (PropsIndex > 0)
            {
               CurrentPropsIndex = PropsIndex;
               k = PropsIndex - 1;

               if (MaterialDefined[k])
               {
                  e_FileLineError(LineNumber,FullPathName,
                                  "material multiply defined");
                  e_free(UsedMaterialName);
                  return 0;
               }
               else
               {
                  NDefinedMaterials++;
                  MaterialDefined[k] = TRUE;
               }

               if (!GetMaterialProperties(&Props[k]))
               {
                  e_free(UsedMaterialName);
                  return 0;
               }
            }
         }
      }
   }

   e_free(UsedMaterialName);

   if (ReturnValue == EOF)
   {
      if (NDefinedMaterials != NMats)
      {
         e_GeneralError("incorrect number of materials");
         return 0;
      }
      else 
         return 1;
   }
   else
      return 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetCompPropsType1Index: To return the properties index corresponding to a
   TYPE1 composite material name if the composite material name has already
   been read, else 0.                                                         */


static int GetCompPropsType1Index(char *CompMaterialName)
{
   int  i;
   char newname[E_LINE_MAX + 1];
   
   /* assemble quote-less string */
   strcpy(newname, CompMaterialName+1);
   newname[strlen(CompMaterialName) - 2] = '\0';

   for (i = 0; i < NCompMatsType1; i++)
   {
      if (strcmp(CompPropsType1[i].Name, newname) == 0)
        return i + 1;
   }

   return 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetPropsIndex: To return the properties index corresponding to a material
   name if the material name has already been read, else 0.                   */


static int GetPropsIndex(char *string)
{
   int i;
   char newname[E_LINE_MAX + 1];
   
   /* assemble quote-less string */
   strcpy(newname, string+1);
   newname[strlen(string) - 2] = '\0';

   for (i = 0; i < CurrentNMats; i++)
   {
      if (strcmp(Props[i].Name, newname) == 0)
        return i + 1;
   }

   return 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetCompMatPropsType1: To read in one set of TYPE1 composite material
   properties.                                                                */

/* The return value is 1 if the data are read successfully or the end of file
   is encountered, else 0.                                                    */


static int GetCompMatPropsType1(COMPMAT1 *CompPropsType1)
{
   if (   GetKeyword(line,word,&LineNumber,FullPathName,fp) != 1
       || !WordMatchesKeyword(E_MKW_NLAYERS,word)
       || !GetNLayers(CompPropsType1)                           )
   {
      return 0;
   }

   if (   (CompPropsType1->LayerMats
              = (MAT **)e_calloc(CompPropsType1->NLayers,
                                            sizeof(MAT *))) == NULL
       || (CompPropsType1->Thicknesses
                 = e_alloc_d_vect(CompPropsType1->NLayers)) == NULL)
   {
      return 0;
   }

   if (   GetKeyword(line,word,&LineNumber,FullPathName,fp) != 1
       || !WordMatchesKeyword(E_MKW_LAYER_MATS,word)
       || !GetLayerMaterials(CompPropsType1)                    )
   {
      return 0;
   }

   if (   GetKeyword(line,word,&LineNumber,FullPathName,fp) != 1
       || !WordMatchesKeyword(E_MKW_LAYER_THICKNESSES,word)
       || !GetLayerThicknesses(CompPropsType1)                  )
   {
      return 0;
   }

   if (   GetKeyword(line,word,&LineNumber,FullPathName,fp) != 1
       || !WordMatchesKeyword(E_MKW_END,word)                     )
   {
      return 0;
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* Get... input functions: To read in data corresponding to a keyword. */

/* The return value is 1 if the data are read successfully, else 0. */


static int GetNLayers(COMPMAT1 *CompPropsType1)
{
   if (sscanf(line,"%*s %d",&CompPropsType1->NLayers) == 1)
   {
      if (CompPropsType1->NLayers < 1)
      {
         e_FileLineError(LineNumber,FullPathName,
                         "ill-defined number of material layers");
         return 0;
      }
   }
   else
   {
      e_FileLineError(LineNumber,FullPathName,
                      "cannot read number of material layers");
      return 0;
   }

   return 1;
}


/* ////////////////////////////////// */


static int GetLayerMaterials(COMPMAT1 *CompPropsType1)
{
   int i;
   int LayerIndex;
   int NameLength;
   int PropsIndex;

   char *pFirstQuote;
   char *pLastQuote;
   char *string;


   if ((string = e_alloc_c_vect(E_LINE_MAX + 1)) == NULL)
      return 0;

   for (i = 0; i < CompPropsType1->NLayers; i++)
   {
      if (e_GetNextLine(line,&LineNumber,FullPathName,fp) == 0)
      {
         e_free(string);
         return 0;
      }
      else if (    sscanf(line,"%d %s",&LayerIndex,string) != 2
                ||      (pFirstQuote = strchr(string,'"')) == NULL
                ||      (pLastQuote = strrchr(string,'"')) == pFirstQuote
                ||         (int)(pLastQuote - pFirstQuote) == 1          )
      {
         e_FileLineError(LineNumber,FullPathName,
                         "cannot read layer material");
         e_free(string);
         return 0;
      }
      else if (LayerIndex != i + 1)
      {
         e_FileLineError(LineNumber,FullPathName,
                         "layer index out of sequence");
         e_free(string);
         return 0;
      }

      PropsIndex = GetPropsIndex(string);

      if (PropsIndex > 0)
      {
         CompPropsType1->LayerMats[i] = &Props[PropsIndex - 1];
      }
      else
      {
         if (CurrentNMats == NMats)
         {
            e_GeneralError("incorrect number of materials");
            e_free(string);
            return 0;
         }

         NameLength = (int)(pLastQuote - pFirstQuote) - 1;
         Props[CurrentNMats].Name = e_alloc_c_vect(NameLength + 1);

         if (Props[CurrentNMats].Name == NULL)
         {
            e_free(string);
            return 0;
         }

         CompPropsType1->LayerMats[i] = &Props[CurrentNMats];
         strncpy(Props[CurrentNMats].Name,pFirstQuote + 1,NameLength);
         CurrentNMats++;
      }
   }

   e_free(string);
   return 1;
}


/* ////////////////////////////////// */


static int GetLayerThicknesses(COMPMAT1 *CompPropsType1)
{
   int i;
   int ThicknessIndex;


   for (i = 0; i < CompPropsType1->NLayers; i++)
   {
      if (e_GetNextLine(line,&LineNumber,FullPathName,fp) == 0)
         return 0;

      if (sscanf(line,"%d %lf ",&ThicknessIndex,
                 &CompPropsType1->Thicknesses[i]) == 2)
      {
         if (ThicknessIndex != i + 1)
         {
            e_FileLineError(LineNumber,FullPathName,
                            "layer thickness index out of sequence");
            return 0;
         }
      }
      else
      {
         e_FileLineError(LineNumber,FullPathName,
                         "cannot read layer thickness");
         return 0;
      }
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetMaterialProperties: To set the properties of a material. */

/* The return value is 1 if the data are set successfully or the end of file
   is encountered, else 0.                                                    */


static int GetMaterialProperties(MAT *Props)
{
   int i;
   int ReturnValue;


   i = 0;

   if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
   {
      e_FileError(FullPathName,"unexpected end of file");
      return 0;
   }

   while (TRUE)
   {
      if ((ReturnValue = WordMatchesPropKeyword(i,Keyword[i],word)) > 0)
      {
         if (GetKeywordData(i++,Props) == 1)
         {
            if (i == E_NMAT_PROP_KEYWORDS) 
            {
               return 1;
            }
            else if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
            {
               e_FileError(FullPathName,"unexpected end of file");
               return 0;
            }
         }
         else
            return 0;
      }
      else if (ReturnValue < 0)
      {
         i++;
      }
      else
         return 0;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* WordMatchesKeyword: To return whether a word is a specified keyword. */

/* The return value is 1 if the word matches the keyword, else 0. */


static int WordMatchesKeyword(char *Keyword, char *word)
{
   if (strcmp(Keyword,word) == 0)
   {
      return 1;
   }
   else
   {
      if (IsAKeyword(word))
         e_FileLineError(LineNumber,FullPathName,"keyword out of sequence");
      else
         e_FileLineError(LineNumber,FullPathName,"undefined keyword");

      return 0;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* WordMatchesPropKeyword: To return whether a word is a specified material
   property keyword.                                                          */

/* The return value is 1 if the word matches the keyword, -1 if it matches
   a "later" material keyword, else 0.                                        */


static int WordMatchesPropKeyword(int i, char *Keyword, char *word)
{
   int PropertyKeywordIndex;


   if (strcmp(Keyword,word) == 0)
   {
      return 1;
   }
   else
   {
      PropertyKeywordIndex = GetPropsKeywordIndex(word);

      if (PropertyKeywordIndex > i + 1)
      {
         return -1;
      }
      else if (   (    PropertyKeywordIndex > 0
                    && PropertyKeywordIndex < i + 1)
               || IsAKeyword(word)                  )
      {
         e_FileLineError(LineNumber,FullPathName,"keyword out of sequence");
         return 0;
      }
      else
      {
         e_FileLineError(LineNumber,FullPathName,"undefined keyword");
         return 0;
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* IsAKeyword: To return whether a word is a material keyword. */

/* The return value is 1 if "word" is a material keyword, else 0. */


static int IsAKeyword(char *word)
{
   int i;


   if (                 strcmp(E_MKW_MAT,word) == 0
       ||    strcmp(E_MKW_COMP_MAT_TYPE1,word) == 0
       ||           strcmp(E_MKW_NLAYERS,word) == 0
       ||        strcmp(E_MKW_LAYER_MATS,word) == 0
       || strcmp(E_MKW_LAYER_THICKNESSES,word) == 0
       ||               strcmp(E_MKW_END,word) == 0)
   {
      return TRUE;
   }

   for (i = 0; i < E_NMAT_PROP_KEYWORDS; i++)
   {
      if (strcmp(Keyword[i],word) == 0)
         return TRUE;
   }

   return FALSE;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetPropsKeywordIndex: To return the index of a material properties
   keyword, else 0.                                                           */


static int GetPropsKeywordIndex(char *word)
{
   int i;


   for (i = 0; i < E_NMAT_PROP_KEYWORDS; i++)
   {
      if (strcmp(Keyword[i],word) == 0)
         return i + 1;
   }

   return 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetKeywordData: To read in keyword data. */

/* The return value is 1 if the data are read successfully, else 0. */


static int GetKeywordData(int i, MAT *Props)
{
   switch (i)
   {
      case E_MKW_NUM_MAT_TYPE:
      {
         return GetMatType(&Props->MatType);
      }
      case E_MKW_NUM_DENSITY:
      {
         return GetDensity(&Props->Density);
      }
      case E_MKW_NUM_YMOD:
      {
         return GetYMod(&Props->YMod);
      }
      case E_MKW_NUM_YMOD_X:
      {
         return GetYMod_x(&Props->YMod_x);
      }
      case E_MKW_NUM_YMOD_Y:
      {
         return GetYMod_y(&Props->YMod_y);
      }
      case E_MKW_NUM_YMOD_Z:
      {
         return GetYMod_z(&Props->YMod_z);
      }
      case E_MKW_NUM_POISSONS_RATIO:
      {
         return GetPoissonsRatio(&Props->PoissonsRatio);
      }
      case E_MKW_NUM_THICKNESS:
      {
         return GetThickness(&Props->Thickness);
      }
      case E_MKW_NUM_IPARAM1:
      {
         return GetiParam(&Props->iParam1,E_MKW_NUM_IPARAM1,E_MKW_IPARAM1);
      }
      case E_MKW_NUM_IPARAM2:
      {
         return GetiParam(&Props->iParam2,E_MKW_NUM_IPARAM2,E_MKW_IPARAM2);
      }
      case E_MKW_NUM_IPARAM3:
      {
         return GetiParam(&Props->iParam3,E_MKW_NUM_IPARAM3,E_MKW_IPARAM3);
      }
      case E_MKW_NUM_IPARAM4:
      {
         return GetiParam(&Props->iParam4,E_MKW_NUM_IPARAM4,E_MKW_IPARAM4);
      }
      case E_MKW_NUM_IPARAM5:
      {
         return GetiParam(&Props->iParam5,E_MKW_NUM_IPARAM5,E_MKW_IPARAM5);
      }
      case E_MKW_NUM_IPARAM6:
      {
         return GetiParam(&Props->iParam6,E_MKW_NUM_IPARAM6,E_MKW_IPARAM6);
      }
      case E_MKW_NUM_DPARAM1:
      {
         return GetdParam(&Props->dParam1,E_MKW_NUM_DPARAM1,E_MKW_DPARAM1);
      }
      case E_MKW_NUM_DPARAM2:
      {
         return GetdParam(&Props->dParam2,E_MKW_NUM_DPARAM2,E_MKW_DPARAM2);
      }
      case E_MKW_NUM_DPARAM3:
      {
         return GetdParam(&Props->dParam3,E_MKW_NUM_DPARAM3,E_MKW_DPARAM3);
      }
      case E_MKW_NUM_DPARAM4:
      {
         return GetdParam(&Props->dParam4,E_MKW_NUM_DPARAM4,E_MKW_DPARAM4);
      }
      case E_MKW_NUM_DPARAM5:
      {
         return GetdParam(&Props->dParam5,E_MKW_NUM_DPARAM5,E_MKW_DPARAM5);
      }
      case E_MKW_NUM_DPARAM6:
      {
         return GetdParam(&Props->dParam6,E_MKW_NUM_DPARAM6,E_MKW_DPARAM6);
      }
      case E_MKW_NUM_DPARAM7:
      {
         return GetdParam(&Props->dParam7,E_MKW_NUM_DPARAM7,E_MKW_DPARAM7);
      }
      case E_MKW_NUM_DPARAM8:
      {
         return GetdParam(&Props->dParam8,E_MKW_NUM_DPARAM8,E_MKW_DPARAM8);
      }
      case E_MKW_NUM_DPARAM9:
      {
         return GetdParam(&Props->dParam9,E_MKW_NUM_DPARAM9,E_MKW_DPARAM9);
      }
      case E_MKW_NUM_DPARAM10:
      {
         return GetdParam(&Props->dParam10,E_MKW_NUM_DPARAM10,E_MKW_DPARAM10);
      }
      case E_MKW_NUM_DPARAM11:
      {
         return GetdParam(&Props->dParam11,E_MKW_NUM_DPARAM11,E_MKW_DPARAM11);
      }
      case E_MKW_NUM_DPARAM12:
      {
         return GetdParam(&Props->dParam12,E_MKW_NUM_DPARAM12,E_MKW_DPARAM12);
      }
      case E_MKW_NUM_DPARAM13:
      {
         return GetdParam(&Props->dParam13,E_MKW_NUM_DPARAM13,E_MKW_DPARAM13);
      }
      case E_MKW_NUM_DPARAM14:
      {
         return GetdParam(&Props->dParam14,E_MKW_NUM_DPARAM14,E_MKW_DPARAM14);
      }
      case E_MKW_NUM_DPARAM15:
      {
         return GetdParam(&Props->dParam15,E_MKW_NUM_DPARAM15,E_MKW_DPARAM15);
      }
      case E_MKW_NUM_DPARAM16:
      {
         return GetdParam(&Props->dParam16,E_MKW_NUM_DPARAM16,E_MKW_DPARAM16);
      }
      case E_MKW_NUM_DPARAM17:
      {
         return GetdParam(&Props->dParam17,E_MKW_NUM_DPARAM17,E_MKW_DPARAM17);
      }
      case E_MKW_NUM_DPARAM18:
      {
         return GetdParam(&Props->dParam18,E_MKW_NUM_DPARAM18,E_MKW_DPARAM18);
      }
      case E_MKW_NUM_DPARAM19:
      {
         return GetdParam(&Props->dParam19,E_MKW_NUM_DPARAM19,E_MKW_DPARAM19);
      }
      case E_MKW_NUM_DPARAM20:
      {
         return GetdParam(&Props->dParam20,E_MKW_NUM_DPARAM20,E_MKW_DPARAM20);
      }
      case E_MKW_NUM_END:
      {
         return 1;
      }
      default:
      {
         return 0;
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* Get... input functions: To read in data corresponding to a keyword. */

/* The return value is 1 if the data are read successfully, else 0. */


static int GetMatType(int *MatType)
{
   if (sscanf(line,"%*s %d",MatType) != 1)
   {
      e_FileLineError(LineNumber,FullPathName,
                      "cannot read material model type");
      return 0;
   }

   KeywordSet[CurrentPropsIndex - 1][E_MKW_NUM_MAT_TYPE] = TRUE;
   return 1;
}


/* ////////////////////////////////// */


static int GetDensity(double *Density)
{
   if (sscanf(line,"%*s %lf",Density) == 1)
   {
      if (!(*Density > 0.0))
      {
         e_FileLineError(LineNumber,FullPathName,"ill-defined density");
         return 0;
      }
   }
   else
   {
      e_FileLineError(LineNumber,FullPathName,"cannot read density");
      return 0;
   }

   KeywordSet[CurrentPropsIndex - 1][E_MKW_NUM_DENSITY] = TRUE;
   return 1;
}


/* ////////////////////////////////// */


static int GetYMod(double *YMod)
{
   if (sscanf(line,"%*s %lf",YMod) == 1)
   {
      if (!(*YMod > 0.0))
      {
         e_FileLineError(LineNumber,FullPathName,
                         "ill-defined Young's modulus");
         return 0;
      }
   }
   else
   {
      e_FileLineError(LineNumber,FullPathName,
                      "cannot read Young's modulus");
      return 0;
   }

   KeywordSet[CurrentPropsIndex - 1][E_MKW_NUM_YMOD] = TRUE;
   return 1;
}


/* ////////////////////////////////// */


static int GetYMod_x(double *YMod_x)
{
   if (KeywordSet[CurrentPropsIndex - 1][E_MKW_NUM_YMOD])
   {
      e_FileLineError(LineNumber,FullPathName,
                      "Young's modulus multiply defined");
      return 0;
   }
   else if (sscanf(line,"%*s %lf",YMod_x) == 1)
   {
      if (!(*YMod_x > 0.0))
      {
         e_FileLineError(LineNumber,FullPathName,
                         "ill-defined Young's modulus");
         return 0;
      }
   }
   else
   {
      e_FileLineError(LineNumber,FullPathName,
                      "cannot read Young's modulus");
      return 0;
   }

   KeywordSet[CurrentPropsIndex - 1][E_MKW_NUM_YMOD_X] = TRUE;
   return 1;
}


/* ////////////////////////////////// */


static int GetYMod_y(double *YMod_y)
{
   if (KeywordSet[CurrentPropsIndex - 1][E_MKW_NUM_YMOD])
   {
      e_FileLineError(LineNumber,FullPathName,
                      "Young's modulus multiply defined");
      return 0;
   }
   else if (sscanf(line,"%*s %lf",YMod_y) == 1)
   {
      if (!(*YMod_y > 0.0))
      {
         e_FileLineError(LineNumber,FullPathName,
                         "ill-defined Young's modulus");
         return 0;
      }
   }
   else
   {
      e_FileLineError(LineNumber,FullPathName,
                      "cannot read Young's modulus");
      return 0;
   }

   KeywordSet[CurrentPropsIndex - 1][E_MKW_NUM_YMOD_Y] = TRUE;
   return 1;
}


/* ////////////////////////////////// */


static int GetYMod_z(double *YMod_z)
{
   if (KeywordSet[CurrentPropsIndex - 1][E_MKW_NUM_YMOD])
   {
      e_FileLineError(LineNumber,FullPathName,
                      "Young's modulus multiply defined");
      return 0;
   }
   else if (sscanf(line,"%*s %lf",YMod_z) == 1)
   {
      if (!(*YMod_z > 0.0))
      {
         e_FileLineError(LineNumber,FullPathName,
                         "ill-defined Young's modulus");
         return 0;
      }
   }
   else
   {
      e_FileLineError(LineNumber,FullPathName,
                      "cannot read Young's modulus");
      return 0;
   }

   KeywordSet[CurrentPropsIndex - 1][E_MKW_NUM_YMOD_Z] = TRUE;
   return 1;
}


/* ////////////////////////////////// */


static int GetPoissonsRatio(double *PoissonsRatio)
{
   if (sscanf(line,"%*s %lf",PoissonsRatio) == 1)
   {
      if (!(*PoissonsRatio > 0.0))
      {
         e_FileLineError(LineNumber,FullPathName,
                         "ill-defined Poisson's ratio");
         return 0;
      }
   }
   else
   {
      e_FileLineError(LineNumber,FullPathName,"cannot read Poisson's ratio");
      return 0;
   }

   KeywordSet[CurrentPropsIndex - 1][E_MKW_NUM_POISSONS_RATIO] = TRUE;
   return 1;
}


/* ////////////////////////////////// */


static int GetThickness(double *thickness)
{
   if (sscanf(line,"%*s %lf",thickness) == 1)
   {
      if (!(*thickness > 0.0))
      {
         e_FileLineError(LineNumber,FullPathName,"ill-defined thickness");
         return 0;
      }
   }
   else
   {
      e_FileLineError(LineNumber,FullPathName,"cannot read thickness");
      return 0;
   }

   KeywordSet[CurrentPropsIndex - 1][E_MKW_NUM_THICKNESS] = TRUE;
   return 1;
}


/* ////////////////////////////////// */


static int GetiParam(int *iParam, int iParamNumber, char *iParamName)
{
   char ErrorMessage[E_DOF_MAX + 12 + 1];


   if (sscanf(line,"%*s %d",iParam) != 1)
   {
      strcpy(ErrorMessage,"cannot read ");
      strcpy(&ErrorMessage[12],iParamName);
      e_FileLineError(LineNumber,FullPathName,ErrorMessage);
      return 0;
   }

   KeywordSet[CurrentPropsIndex - 1][iParamNumber] = TRUE;
   return 1;
}



/* ////////////////////////////////// */



static int GetdParam(double *dParam, int dParamNumber, char *dParamName)
{
   char ErrorMessage[E_DOF_MAX + 12 + 1];


   if (sscanf(line,"%*s %lf",dParam) != 1)
   {
      strcpy(ErrorMessage,"cannot read ");
      strcpy(&ErrorMessage[12],dParamName);
      e_FileLineError(LineNumber,FullPathName,ErrorMessage);
      return 0;
   }

   KeywordSet[CurrentPropsIndex - 1][dParamNumber] = TRUE;
   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetMeshData: To set the MESH structure data. */


static void SetMeshData(MESH *mesh)
{
   int k,j;


   for (k = 0; k < NMats; k++)
   {
      for (j = 0; j < E_NMAT_PROP_KEYWORDS; j++)
         mesh->Props[k].KeywordSet[j] = KeywordSet[k][j];
   }

   mesh->SubStructSet[E_SS_NUM_MAT_PROPS] = TRUE;

   e_free(MaterialDefined);
   e_free(CompMatDefinedType1);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* DeallocateAllMemory: To deallocate all memory. */


static void DeallocateAllMemory()
{
   int k;


   for (k = 0; k < NCompMatsType1; k++)
   {
      e_free(CompPropsType1[k].LayerMats);
      e_free(CompPropsType1[k].Thicknesses);
   }

   for (k = NMdfMats; k < NMats; k++)
   {
      e_free(Props[k].Name);
      Props[k].Name = NULL;
   }

   e_free(MaterialDefined);
   e_free(CompMatDefinedType1);
   e_free_frag_i_matr(NMats,KeywordSet);
   e_free(FullPathName);
}


/* ////////////////////////////////////////////////////////////////////////// */










/* ////////////////////////////////////////////////////////////////////////// */
/* e_PutMaterials: To output material properties. */
/* The return value is 1 if the data are written successfully, else 0. */


int e_PutMaterials(MESH *mesh, char *OutputFile)
{
   int i,j,k;

   char FirstMaterials = 1;     /* whether first material properties */

   FILE *fp;                    /* output file pointer */


   if (!mesh->SubStructSet[E_SS_NUM_MAT_PROPS])
   {
      e_GeneralError("e_PutMaterials: material properties not set");
      return 0;
   }
   else if (OpenOutputFile(OutputFile,E_MAT,&fp) == NULL)
   {
      return 0;
   }

   for (k = 0; k < mesh->NCompMatsType1; k++)
   {
      if (FirstMaterials)
         FirstMaterials = 0;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"%s  \"%s\"\n",E_MKW_COMP_MAT_TYPE1,
                         mesh->CompPropsType1[k].Name);
      fprintf(fp,"%s  %d\n",E_MKW_NLAYERS,
              mesh->CompPropsType1[k].NLayers);
      fprintf(fp,"%s\n",E_MKW_LAYER_MATS);

      for (j = 0; j < mesh->CompPropsType1[k].NLayers; j++)
      {
         i = (int)(mesh->CompPropsType1[k].LayerMats[j] - mesh->Props);
         fprintf(fp,"%3d   \"%s\"\n",j + 1,mesh->Props[i].Name);
      }

      fprintf(fp,"%s\n",E_MKW_LAYER_THICKNESSES);

      for (j = 0; j < mesh->CompPropsType1[k].NLayers; j++)
         fprintf(fp,"%3d   %20.8f\n",j + 1,
                 mesh->CompPropsType1[k].Thicknesses[j]);

      fprintf(fp,"%s\n",E_MKW_END);
   }

   for (k = 0; k < mesh->NMats; k++)
   {
      if (FirstMaterials)
         FirstMaterials = 0;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"%s  \"%s\"\n",E_MKW_MAT,mesh->Props[k].Name);

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_MAT_TYPE])
         fprintf(fp,"%-9s  %d\n",E_MKW_MAT_TYPE,mesh->Props[k].MatType);

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DENSITY])
         fprintf(fp,"%-9s  %f\n",E_MKW_DENSITY,mesh->Props[k].Density);

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_YMOD])
      {
         fprintf(fp,"%-9s  ",E_MKW_YMOD);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].YMod) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_YMOD_X])
      {
         fprintf(fp,"%-9s  ",E_MKW_YMOD_X);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].YMod_x) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_YMOD_Y])
      {
         fprintf(fp,"%-9s  ",E_MKW_YMOD_Y);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].YMod_y) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_YMOD_Z])
      {
         fprintf(fp,"%-9s  ",E_MKW_YMOD_Z);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].YMod_z) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_POISSONS_RATIO])
         fprintf(fp,"%-9s  %1.8f\n",E_MKW_POISSONS_RATIO,
                                    mesh->Props[k].PoissonsRatio);

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_THICKNESS])
         fprintf(fp,"%-9s  %1.8f\n",E_MKW_THICKNESS,mesh->Props[k].Thickness);

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_IPARAM1])
         fprintf(fp,"%-9s  %d\n",E_MKW_IPARAM1,mesh->Props[k].iParam1);

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_IPARAM2])
         fprintf(fp,"%-9s  %d\n",E_MKW_IPARAM2,mesh->Props[k].iParam2);

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_IPARAM3])
         fprintf(fp,"%-9s  %d\n",E_MKW_IPARAM3,mesh->Props[k].iParam3);

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_IPARAM4])
         fprintf(fp,"%-9s  %d\n",E_MKW_IPARAM4,mesh->Props[k].iParam4);

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_IPARAM5])
         fprintf(fp,"%-9s  %d\n",E_MKW_IPARAM5,mesh->Props[k].iParam5);

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_IPARAM6])
         fprintf(fp,"%-9s  %d\n",E_MKW_IPARAM5,mesh->Props[k].iParam6);

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM1])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM1);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam1) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM2])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM2);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam2) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM3])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM3);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam3) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM4])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM4);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam4) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM5])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM5);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam5) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM6])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM6);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam6) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM7])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM7);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam7) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM8])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM8);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam8) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM9])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM9);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam9) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM10])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM10);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam10) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM11])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM11);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam11) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM12])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM12);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam12) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM13])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM13);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam13) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM14])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM14);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam14) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM15])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM15);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam15) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM16])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM16);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam16) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM17])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM17);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam17) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM18])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM18);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam18) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM19])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM19);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam19) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      if (mesh->Props[k].KeywordSet[E_MKW_NUM_DPARAM20])
      {
         fprintf(fp,"%-9s  ",E_MKW_DPARAM20);

         if (PrintEng(fp,"%14.6e",mesh->Props[k].dParam20) < 0)
            return 0;
         else
            fprintf(fp,"\n");
      }

      fprintf(fp,"%s\n",E_MKW_END);
   }

   fclose(fp);
   return 1;
}
