/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                    e_GetMeshParams(), e_PutMeshParams()                    */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include "e_pltfrm.h"
#include <stdio.h>
#include <string.h>
#include <e_lib.h>

#include "comsrc.h"
#include "meshpara.h"


/* ////////////////////////////////////////////////////////////////////////// */
/* e_GetMeshParams: To read in mesh parameters. */
/* The return value is 1 if no input errors occur, else 0. */


int e_GetMeshParams(MESH *mesh, char *InputFile)
{
   if ((FullPathName = OpenInputFile(InputFile,E_MPR,&fp)) == NULL)
      return 0;

   Initialise();

   if (GetData(mesh))
   {
      SetMeshData(mesh);
      e_free(FullPathName);
      fclose(fp);
      return 1;
   }
   else
   {
      DeallocateAllMemory();
      fclose(fp);
      return 0;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* Initialise: To Initialise the variables. */


static void Initialise()
{
   LineNumber = 0;
   NodalMeshParamsSet = FALSE;
   ElemMeshParamsSet = FALSE;
   NodalMeshParams = NULL;
   ElemMeshParams = NULL;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetData: To read in the input data. */

/* The return value is 1 if the data are read successfully or the end of file
   is encountered, else 0.                                                    */


static int GetData(MESH *mesh)
{
   int sum = 0;

   char ReadNextLine = FALSE;
   char word[E_LINE_MAX + 1];      /* potential keyword */


   if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
   {
      e_FileError(FullPathName,"unexpected end of file");
      return 0;
   }
   else if (strcmp(E_KW_NODAL_MESHPARAMS,word) == 0)
   {
      if (   (NodalMeshParams = e_alloc_d_vect(mesh->NMeshPoints)) == NULL
          || !GetNodalMeshParams(mesh->NMeshPoints)                       )
      {
         return 0;
      }
      else if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
      {
         NodalMeshParamsSet = TRUE;
         return 1;
      }

      NodalMeshParamsSet = TRUE;
   }

   if ((ElemMeshParams = e_alloc_d_vect(mesh->TotalNElems)) == NULL)
      return 0;

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
   {
      ElemMeshParamsLink1 = ElemMeshParams + sum;
      sum += mesh->NElemsLink1;

      if (   ReadNextLine
          && GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         return 0;
      }
      else if (   !WordMatchesKeyword(E_KW_MESHPARAMS_LINK1,word)
               || !GetElemMeshParams(mesh->NElemsLink1,ElemMeshParamsLink1))
      {
         return 0;
      }

      ReadNextLine = TRUE;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
   {
      ElemMeshParamsLink2 = ElemMeshParams + sum;
      sum += mesh->NElemsLink2;

      if (   ReadNextLine
          && GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         return 0;
      }
      else if (   !WordMatchesKeyword(E_KW_MESHPARAMS_LINK2,word)
               || !GetElemMeshParams(mesh->NElemsLink2,ElemMeshParamsLink2))
      {
         return 0;
      }

      ReadNextLine = TRUE;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
   {
      ElemMeshParamsLink3 = ElemMeshParams + sum;
      sum += mesh->NElemsLink3;

      if (   ReadNextLine
          && GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         return 0;
      }
      else if (   !WordMatchesKeyword(E_KW_MESHPARAMS_LINK3,word)
               || !GetElemMeshParams(mesh->NElemsLink3,ElemMeshParamsLink3))
      {
         return 0;
      }

      ReadNextLine = TRUE;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
   {
      ElemMeshParamsLink4 = ElemMeshParams + sum;
      sum += mesh->NElemsLink4;

      if (   ReadNextLine
          && GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         return 0;
      }
      else if (   !WordMatchesKeyword(E_KW_MESHPARAMS_LINK4,word)
               || !GetElemMeshParams(mesh->NElemsLink4,ElemMeshParamsLink4))
      {
         return 0;
      }

      ReadNextLine = TRUE;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
   {
      ElemMeshParamsLink5 = ElemMeshParams + sum;
      sum += mesh->NElemsLink5;

      if (   ReadNextLine
          && GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         return 0;
      }
      else if (   !WordMatchesKeyword(E_KW_MESHPARAMS_LINK5,word)
               || !GetElemMeshParams(mesh->NElemsLink5,ElemMeshParamsLink5))
      {
         return 0;
      }

      ReadNextLine = TRUE;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
   {
      ElemMeshParamsTriang1 = ElemMeshParams + sum;
      sum += mesh->NElemsTriang1;

      if (   ReadNextLine
          && GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         return 0;
      }
      else if (   !WordMatchesKeyword(E_KW_MESHPARAMS_TRIANG1,word)
               || !GetElemMeshParams(mesh->NElemsTriang1,ElemMeshParamsTriang1))
      {
         return 0;
      }

      ReadNextLine = TRUE;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
   {
      ElemMeshParamsTriang2 = ElemMeshParams + sum;
      sum += mesh->NElemsTriang2;

      if (   ReadNextLine
          && GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         return 0;
      }
      else if (   !WordMatchesKeyword(E_KW_MESHPARAMS_TRIANG2,word)
               || !GetElemMeshParams(mesh->NElemsTriang2,ElemMeshParamsTriang2))
      {
         return 0;
      }

      ReadNextLine = TRUE;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
   {
      ElemMeshParamsTriang3 = ElemMeshParams + sum;
      sum += mesh->NElemsTriang3;

      if (   ReadNextLine
          && GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         return 0;
      }
      else if (   !WordMatchesKeyword(E_KW_MESHPARAMS_TRIANG3,word)
               || !GetElemMeshParams(mesh->NElemsTriang3,ElemMeshParamsTriang3))
      {
         return 0;
      }

      ReadNextLine = TRUE;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
   {
      ElemMeshParamsTriang4 = ElemMeshParams + sum;
      sum += mesh->NElemsTriang4;

      if (   ReadNextLine
          && GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         return 0;
      }
      else if (   !WordMatchesKeyword(E_KW_MESHPARAMS_TRIANG4,word)
               || !GetElemMeshParams(mesh->NElemsTriang4,ElemMeshParamsTriang4))
      {
         return 0;
      }

      ReadNextLine = TRUE;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
   {
      ElemMeshParamsTriang5 = ElemMeshParams + sum;
      sum += mesh->NElemsTriang5;

      if (   ReadNextLine
          && GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         return 0;
      }
      else if (   !WordMatchesKeyword(E_KW_MESHPARAMS_TRIANG5,word)
               || !GetElemMeshParams(mesh->NElemsTriang5,ElemMeshParamsTriang5))
      {
         return 0;
      }

      ReadNextLine = TRUE;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
   {
      ElemMeshParamsQuad1 = ElemMeshParams + sum;
      sum += mesh->NElemsQuad1;

      if (   ReadNextLine
          && GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         return 0;
      }
      else if (   !WordMatchesKeyword(E_KW_MESHPARAMS_QUAD1,word)
               || !GetElemMeshParams(mesh->NElemsQuad1,ElemMeshParamsQuad1))
      {
         return 0;
      }

      ReadNextLine = TRUE;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
   {
      ElemMeshParamsQuad3 = ElemMeshParams + sum;
      sum += mesh->NElemsQuad3;

      if (   ReadNextLine
          && GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         return 0;
      }
      else if (   !WordMatchesKeyword(E_KW_MESHPARAMS_QUAD3,word)
               || !GetElemMeshParams(mesh->NElemsQuad3,ElemMeshParamsQuad3))
      {
         return 0;
      }

      ReadNextLine = TRUE;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
   {
      ElemMeshParamsTetrah1 = ElemMeshParams + sum;
      sum += mesh->NElemsTetrah1;

      if (   ReadNextLine
          && GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         return 0;
      }
      else if (   !WordMatchesKeyword(E_KW_MESHPARAMS_TETRAH1,word)
               || !GetElemMeshParams(mesh->NElemsTetrah1,ElemMeshParamsTetrah1))
      {
         return 0;
      }

      ReadNextLine = TRUE;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
   {
      ElemMeshParamsBlock1 = ElemMeshParams + sum;
      sum += mesh->NElemsBlock1;

      if (   ReadNextLine
          && GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         return 0;
      }
      else if (   !WordMatchesKeyword(E_KW_MESHPARAMS_BLOCK1,word)
               || !GetElemMeshParams(mesh->NElemsBlock1,ElemMeshParamsBlock1))
      {
         return 0;
      }

      ReadNextLine = TRUE;
   }

   if (GetKeyword(line,word,&LineNumber,FullPathName,fp) != EOF)
   {
      e_FileLineError(LineNumber,FullPathName,"undefined keyword");
      return 0;
   }

   ElemMeshParamsSet = TRUE;

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* WordMatchesKeyword: To return whether a word is a specified keyword. */

/* The return value is 1 if the word matches the keyword, else 0. */


static int WordMatchesKeyword(char *Keyword, char *word)
{
   if (strcmp(Keyword,word) != 0)
   {
      if (IsAKeyword(word))
         e_FileLineError(LineNumber,FullPathName,"keyword out of sequence");
      else
         e_FileLineError(LineNumber,FullPathName,"undefined keyword");

      return 0;
   }
   else
      return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* Get... input functions: To read in data corresponding to a keyword. */

/* The return value is 1 if the data are read successfully, else 0. */


static int GetNodalMeshParams(int NMeshPoints)
{
   int k;
   int ParamIndex;


   for (k = 0; k < NMeshPoints; k++)
   {
      if (e_GetNextLine(line,&LineNumber,FullPathName,fp) == 0)
         return 0;

      if (sscanf(line,"%d %lf",&ParamIndex,&NodalMeshParams[k]) == 2)
      {
         if (ParamIndex != k + 1)
         {
            e_FileLineError(LineNumber,FullPathName,
                            "nodal mesh parameter index out of sequence");
            return 0;
         }

         if (!(NodalMeshParams[k] > 0))
         {
            e_FileLineError(LineNumber,FullPathName,
                            "ill-defined nodal mesh parameter");
            return 0;
         }
      }
      else
      {
         e_FileLineError(LineNumber,FullPathName,
                         "cannot read nodal mesh parameter");
         return 0;
      }
   }

   return 1;
}


/* ////////////////////////////////// */


static int GetElemMeshParams(int NElements, double *ElemMeshParams)
{
   int i;
   int ParamIndex;


   for (i = 0; i < NElements; i++)
   {
      if (e_GetNextLine(line,&LineNumber,FullPathName,fp) == 0)
         return 0;

      if (sscanf(line,"%d %lf",&ParamIndex,&ElemMeshParams[i]) == 2)
      {
         if (ParamIndex != i + 1)
         {
            e_FileLineError(LineNumber,FullPathName,
                            "element mesh parameter index out of sequence");
            return 0;
         }

         if (!(ElemMeshParams[i] > 0))
         {
            e_FileLineError(LineNumber,FullPathName,
                            "ill-defined element mesh parameter");
            return 0;
         }
      }
      else
      {
         e_FileLineError(LineNumber,FullPathName,
                         "cannot read element mesh parameter");
         return 0;
      }
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* IsAKeyword: To return whether a word is a keyword. */

/* The return value is 1 if "word" is a keyword, else 0. */


static int IsAKeyword(char *word)
{
   if (     strcmp(E_KW_NODAL_MESHPARAMS,word) == 0
       ||   strcmp(E_KW_MESHPARAMS_LINK1,word) == 0
       ||   strcmp(E_KW_MESHPARAMS_LINK2,word) == 0
       ||   strcmp(E_KW_MESHPARAMS_LINK3,word) == 0
       ||   strcmp(E_KW_MESHPARAMS_LINK4,word) == 0
       ||   strcmp(E_KW_MESHPARAMS_LINK5,word) == 0
       || strcmp(E_KW_MESHPARAMS_TRIANG1,word) == 0
       || strcmp(E_KW_MESHPARAMS_TRIANG2,word) == 0
       || strcmp(E_KW_MESHPARAMS_TRIANG3,word) == 0
       || strcmp(E_KW_MESHPARAMS_TRIANG4,word) == 0
       || strcmp(E_KW_MESHPARAMS_TRIANG5,word) == 0
       ||   strcmp(E_KW_MESHPARAMS_QUAD1,word) == 0
       ||   strcmp(E_KW_MESHPARAMS_QUAD3,word) == 0
       || strcmp(E_KW_MESHPARAMS_TETRAH1,word) == 0
       ||  strcmp(E_KW_MESHPARAMS_BLOCK1,word) == 0)
   {
      return TRUE;
   }
   else
   {
      return FALSE;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetMeshData: To set the MESH structure data. */


static void SetMeshData(MESH *mesh)
{
   mesh->MeshParam.NodalMeshParams = NodalMeshParams;
   mesh->MeshParam.ElemMeshParams = ElemMeshParams;
   mesh->MeshParam.ElemMeshParamsLink1 = ElemMeshParamsLink1;
   mesh->MeshParam.ElemMeshParamsLink2 = ElemMeshParamsLink2;
   mesh->MeshParam.ElemMeshParamsLink3 = ElemMeshParamsLink3;
   mesh->MeshParam.ElemMeshParamsLink4 = ElemMeshParamsLink4;
   mesh->MeshParam.ElemMeshParamsLink5 = ElemMeshParamsLink5;
   mesh->MeshParam.ElemMeshParamsTriang1 = ElemMeshParamsTriang1;
   mesh->MeshParam.ElemMeshParamsTriang2 = ElemMeshParamsTriang2;
   mesh->MeshParam.ElemMeshParamsTriang3 = ElemMeshParamsTriang3;
   mesh->MeshParam.ElemMeshParamsTriang4 = ElemMeshParamsTriang4;
   mesh->MeshParam.ElemMeshParamsTriang5 = ElemMeshParamsTriang5;
   mesh->MeshParam.ElemMeshParamsQuad1 = ElemMeshParamsQuad1;
   mesh->MeshParam.ElemMeshParamsQuad3 = ElemMeshParamsQuad3;
   mesh->MeshParam.ElemMeshParamsTetrah1 = ElemMeshParamsTetrah1;
   mesh->MeshParam.ElemMeshParamsBlock1 = ElemMeshParamsBlock1;

   mesh->MeshParam.NodalMeshParamsSet = NodalMeshParamsSet;
   mesh->MeshParam.ElemMeshParamsSet = ElemMeshParamsSet;

   mesh->SubStructSet[E_SS_NUM_MESHPARAM] = TRUE;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* DeallocateAllMemory: To deallocate all memory. */


static void DeallocateAllMemory()
{
   e_free(NodalMeshParams);
   e_free(ElemMeshParams);
   e_free(FullPathName);
}


/* ////////////////////////////////////////////////////////////////////////// */










/* ////////////////////////////////////////////////////////////////////////// */
/* e_PutMeshParams: To output mesh parameters. */
/* The return value is 1 if the data are written successfully, else 0. */


int e_PutMeshParams(MESH *mesh, char *OutputFile)
{
   int i,k;

   char FirstParams = TRUE;     /* whether first set of mesh parameters */

   FILE *fp;                    /* output file pointer */


   if (!mesh->SubStructSet[E_SS_NUM_MESHPARAM])
   {
      e_GeneralError("e_PutMeshParams: mesh parameters not set");
      return 0;
   }
   else if (OpenOutputFile(OutputFile,E_MPR,&fp) == NULL)
   {
      return 0;
   }

   if (mesh->MeshParam.NodalMeshParamsSet)
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# Nodal mesh parameters\n");
      fprintf(fp,"# node index\n");
      fprintf(fp,"#             mesh parameter\n");

      fprintf(fp,"\n%s\n",E_KW_NODAL_MESHPARAMS);

      for (k = 0; k < mesh->NMeshPoints; k++)
         fprintf(fp,"%7d   %20.8f\n", k + 1,
                 mesh->MeshParam.NodalMeshParams[k]);
   }

   if (   mesh->MeshParam.ElemMeshParamsSet
       && mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# LINK1 element mesh parameters\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#             mesh parameter\n");

      fprintf(fp,"\n%s\n",E_KW_MESHPARAMS_LINK1);

      for (i = 0; i < mesh->NElemsLink1; i++)
         fprintf(fp,"%7d   %20.8f\n", i + 1,
                 mesh->MeshParam.ElemMeshParamsLink1[i]);
   }

   if (   mesh->MeshParam.ElemMeshParamsSet
       && mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# LINK2 element mesh parameters\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#             mesh parameter\n");

      fprintf(fp,"\n%s\n",E_KW_MESHPARAMS_LINK2);

      for (i = 0; i < mesh->NElemsLink2; i++)
         fprintf(fp,"%7d   %20.8f\n", i + 1,
                 mesh->MeshParam.ElemMeshParamsLink2[i]);
   }

   if (   mesh->MeshParam.ElemMeshParamsSet
       && mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# LINK3 element mesh parameters\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#             mesh parameter\n");

      fprintf(fp,"\n%s\n",E_KW_MESHPARAMS_LINK3);

      for (i = 0; i < mesh->NElemsLink3; i++)
         fprintf(fp,"%7d   %20.8f\n", i + 1,
                 mesh->MeshParam.ElemMeshParamsLink3[i]);
   }

   if (   mesh->MeshParam.ElemMeshParamsSet
       && mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# LINK4 element mesh parameters\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#             mesh parameter\n");

      fprintf(fp,"\n%s\n",E_KW_MESHPARAMS_LINK4);

      for (i = 0; i < mesh->NElemsLink4; i++)
         fprintf(fp,"%7d   %20.8f\n", i + 1,
                 mesh->MeshParam.ElemMeshParamsLink4[i]);
   }

   if (   mesh->MeshParam.ElemMeshParamsSet
       && mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# LINK5 element mesh parameters\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#             mesh parameter\n");

      fprintf(fp,"\n%s\n",E_KW_MESHPARAMS_LINK5);

      for (i = 0; i < mesh->NElemsLink5; i++)
         fprintf(fp,"%7d   %20.8f\n", i + 1,
                 mesh->MeshParam.ElemMeshParamsLink5[i]);
   }

   if (   mesh->MeshParam.ElemMeshParamsSet
       && mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# TRIANG1 element mesh parameters\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#             mesh parameter\n");

      fprintf(fp,"\n%s\n",E_KW_MESHPARAMS_TRIANG1);

      for (i = 0; i < mesh->NElemsTriang1; i++)
         fprintf(fp,"%7d   %20.8f\n", i + 1,
                 mesh->MeshParam.ElemMeshParamsTriang1[i]);
   }

   if (   mesh->MeshParam.ElemMeshParamsSet
       && mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# TRIANG2 element mesh parameters\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#             mesh parameter\n");

      fprintf(fp,"\n%s\n",E_KW_MESHPARAMS_TRIANG2);

      for (i = 0; i < mesh->NElemsTriang2; i++)
         fprintf(fp,"%7d   %20.8f\n", i + 1,
                 mesh->MeshParam.ElemMeshParamsTriang2[i]);
   }

   if (   mesh->MeshParam.ElemMeshParamsSet
       && mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# TRIANG3 element mesh parameters\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#             mesh parameter\n");

      fprintf(fp,"\n%s\n",E_KW_MESHPARAMS_TRIANG3);

      for (i = 0; i < mesh->NElemsTriang3; i++)
         fprintf(fp,"%7d   %20.8f\n", i + 1,
                 mesh->MeshParam.ElemMeshParamsTriang3[i]);
   }

   if (   mesh->MeshParam.ElemMeshParamsSet
       && mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# TRIANG4 element mesh parameters\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#             mesh parameter\n");

      fprintf(fp,"\n%s\n",E_KW_MESHPARAMS_TRIANG4);

      for (i = 0; i < mesh->NElemsTriang4; i++)
         fprintf(fp,"%7d   %20.8f\n", i + 1,
                 mesh->MeshParam.ElemMeshParamsTriang4[i]);
   }

   if (   mesh->MeshParam.ElemMeshParamsSet
       && mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# TRIANG5 element mesh parameters\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#             mesh parameter\n");

      fprintf(fp,"\n%s\n",E_KW_MESHPARAMS_TRIANG5);

      for (i = 0; i < mesh->NElemsTriang5; i++)
         fprintf(fp,"%7d   %20.8f\n", i + 1,
                 mesh->MeshParam.ElemMeshParamsTriang5[i]);
   }

   if (   mesh->MeshParam.ElemMeshParamsSet
       && mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# QUAD1 element mesh parameters\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#             mesh parameter\n");

      fprintf(fp,"\n%s\n",E_KW_MESHPARAMS_QUAD1);

      for (i = 0; i < mesh->NElemsQuad1; i++)
         fprintf(fp,"%7d   %20.8f\n", i + 1,
                 mesh->MeshParam.ElemMeshParamsQuad1[i]);
   }

   if (   mesh->MeshParam.ElemMeshParamsSet
       && mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# QUAD3 element mesh parameters\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#             mesh parameter\n");

      fprintf(fp,"\n%s\n",E_KW_MESHPARAMS_QUAD3);

      for (i = 0; i < mesh->NElemsQuad3; i++)
         fprintf(fp,"%7d   %20.8f\n", i + 1,
                 mesh->MeshParam.ElemMeshParamsQuad3[i]);
   }

   if (   mesh->MeshParam.ElemMeshParamsSet
       && mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# TETRAH1 element mesh parameters\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#             mesh parameter\n");

      fprintf(fp,"\n%s\n",E_KW_MESHPARAMS_TETRAH1);

      for (i = 0; i < mesh->NElemsTetrah1; i++)
         fprintf(fp,"%7d   %20.8f\n", i + 1,
                 mesh->MeshParam.ElemMeshParamsTetrah1[i]);
   }

   if (   mesh->MeshParam.ElemMeshParamsSet
       && mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# BLOCK1 element mesh parameters\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#               mesh parameter\n");

      fprintf(fp,"\n%s\n",E_KW_MESHPARAMS_BLOCK1);

      for (i = 0; i < mesh->NElemsBlock1; i++)
         fprintf(fp,"%7d   %20.8f\n", i + 1,
                 mesh->MeshParam.ElemMeshParamsBlock1[i]);
   }

   fclose(fp);
   return 1;
}
