/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                       e_GetStresses(), e_PutStresses()                     */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include "e_pltfrm.h"
#include <stdio.h>
#include <string.h>
#include <e_lib.h>

#include "comsrc.h"
#include "stresses.h"


/* ////////////////////////////////////////////////////////////////////////// */
/* e_SetupStresses */
/* The return value is 1 if no errors occur, else 0. */

int e_SetupStresses(MESH *mesh)
{
   if (Initialise(mesh))
   {
      SetMeshData(mesh);
      return 1;
   }
   else
   {
      DeallocateAllMemory(mesh);
      return 0;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_GetStresses: To read in element stress data. */
/* The return value is 1 if no input errors occur, else 0. */


int e_GetStresses(MESH *mesh, char *InputFile)
{
   if ((FullPathName = OpenInputFile(InputFile,E_STE,&fp)) == NULL)
     return 0;

   if (   Initialise(mesh)
       && GetData(mesh))
   {
      SetMeshData(mesh);
      e_free(FullPathName);
      fclose(fp);
      return 1;
   }
   else
   {
      DeallocateAllMemory(mesh);
      fclose(fp);
      return 0;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* initialise: To initialise variables. */


static int Initialise(MESH *mesh)
{
   int i;

   LineNumber = 0;
   StressesLink1 = NULL;
   StressesLink2 = NULL;
   StressesLink3 = NULL;
   StressesLink4 = NULL;
   StressesLink5 = NULL;
   StressesTriang1 = NULL;
   StressesTriang2 = NULL;
   StressesTriang3 = NULL;
   StressesTriang4 = NULL;
   StressesTriang5 = NULL;
   StressesQuad1 = NULL;
   StressesQuad3 = NULL;
   StressesTetrah1 = NULL;
   StressesBlock1 = NULL;
   maxNLayerTriang2 = 0;
   maxNLayerQuad3   = 0;

   if (   (   (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2]) 
           || (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])   )
       && (!mesh->SubStructSet[E_SS_NUM_MAT_PROPS])         )
   {
      e_PlatformMessage("No material information is available for stresses (Triang2/Quad3).");
      return 0;
   }

   for(i = 0; i < mesh->NElemsTriang2; i++)
     if(maxNLayerTriang2 < mesh->CompMatsTriang2[i]->NLayers)
       maxNLayerTriang2 = mesh->CompMatsTriang2[i]->NLayers;

   for(i = 0; i < mesh->NElemsQuad3; i++)
     if(maxNLayerQuad3 < mesh->CompMatsQuad3[i]->NLayers)
       maxNLayerQuad3 = mesh->CompMatsQuad3[i]->NLayers;

   if(   (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
      && (maxNLayerTriang2 <= 0)                     )
     return 0;

   if(   (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
      && (maxNLayerQuad3 <= 0)                     )
     return 0;


   if(   (   (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
          && (!mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_LINK1]) )
      || (   (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
          && (!mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_LINK2]) )
      || (   (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
          && (!mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_LINK3]) )
      || (   (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
          && (!mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_LINK4]) )
      || (   (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
          && (!mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_LINK5]) )
      || (   (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
          && (!mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_TRIANG1]) )
      || (   (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
          && (!mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_TRIANG2]) )
      || (   (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
          && (!mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_TRIANG3]) )
      || (   (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
          && (!mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_TRIANG4]) )
      || (   (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
          && (!mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_TRIANG5]) )
      || (   (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
          && (!mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_QUAD1]) )
      || (   (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
          && (!mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_QUAD3]) )
      || (   (mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
          && (!mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_TETRAH1]) )
      || (   (mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
          && (!mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_BLOCK1]) )  )
   {
      e_PlatformMessage("Number of stress points is not specified for stresses.");
      return 0;
   }

   if(   (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
      && ((StressesLink1 = e_alloc_cont_d_3D_matr(mesh->NElemsLink1, 
                                                  mesh->NStressPointsLink1, 
                                                  E_NUM_STRESS_COMPONENTS)) == NULL))
     return 0;
   if(   (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
      && ((StressesLink2 = e_alloc_cont_d_3D_matr(mesh->NElemsLink2,
                                                  mesh->NStressPointsLink2,
                                                  E_NUM_STRESS_COMPONENTS)) == NULL))
     return 0;
   if(   (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
      && ((StressesLink3 = e_alloc_cont_d_3D_matr(mesh->NElemsLink3,
                                                  mesh->NStressPointsLink3,
                                                  E_NUM_STRESS_COMPONENTS)) == NULL))
     return 0;
   if(   (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
      && ((StressesLink4 = e_alloc_cont_d_3D_matr(mesh->NElemsLink4,
                                                  mesh->NStressPointsLink4,
                                                  E_NUM_STRESS_COMPONENTS)) == NULL))
     return 0;
   if(   (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
      && ((StressesLink5 = e_alloc_cont_d_3D_matr(mesh->NElemsLink5, 
                                                  mesh->NStressPointsLink5, 
                                                  E_NUM_STRESS_COMPONENTS)) == NULL))
     return 0;
   if(   (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
      && ((StressesTriang1 = e_alloc_cont_d_3D_matr(mesh->NElemsTriang1, 
                                                    mesh->NStressPointsTriang1, 
                                                    E_NUM_STRESS_COMPONENTS)) == NULL))
     return 0;
   if(   (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
      && ((StressesTriang2 = e_alloc_cont_d_4D_matr(mesh->NElemsTriang2, 
                                                    maxNLayerTriang2,
                                                    mesh->NStressPointsTriang2, 
                                                    E_NUM_STRESS_COMPONENTS)) == NULL))
     return 0;
   if(   (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
      && ((StressesTriang3 = e_alloc_cont_d_3D_matr(mesh->NElemsTriang3, 
                                                    mesh->NStressPointsTriang3, 
                                                    E_NUM_STRESS_COMPONENTS)) == NULL))
     return 0;
   if(   (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
      && ((StressesTriang4 = e_alloc_cont_d_3D_matr(mesh->NElemsTriang4, 
                                                    mesh->NStressPointsTriang4, 
                                                    E_NUM_STRESS_COMPONENTS)) == NULL))
     return 0;
   if(   (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
      && ((StressesTriang5 = e_alloc_cont_d_3D_matr(mesh->NElemsTriang5, 
                                                    mesh->NStressPointsTriang5, 
                                                    E_NUM_STRESS_COMPONENTS)) == NULL))
     return 0;
   if(   (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
      && ((StressesQuad1 = e_alloc_cont_d_3D_matr(mesh->NElemsQuad1, 
                                                  mesh->NStressPointsQuad1, 
                                                  E_NUM_STRESS_COMPONENTS)) == NULL))
     return 0;
   if(   (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
      && ((StressesQuad3 = e_alloc_cont_d_4D_matr(mesh->NElemsQuad3, 
                                                  maxNLayerQuad3,
                                                  mesh->NStressPointsQuad3, 
                                                  E_NUM_STRESS_COMPONENTS)) == NULL))
     return 0;
   if(   (mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
      && ((StressesTetrah1 = e_alloc_cont_d_3D_matr(mesh->NElemsTetrah1, 
                                                    mesh->NStressPointsTetrah1, 
                                                    E_NUM_STRESS_COMPONENTS)) == NULL))
     return 0;
   if(   (mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
      && ((StressesBlock1 = e_alloc_cont_d_3D_matr(mesh->NElemsBlock1, 
                                                   mesh->NStressPointsBlock1, 
                                                   E_NUM_STRESS_COMPONENTS)) == NULL))
     return 0;

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetData: To read in the input data. */

/* The return value is 1 if the data are read successfully or the end of file
   is encountered, else 0.                                                    */


static int GetData(MESH *mesh)
{
   char word[E_LINE_MAX + 1];      /* potential keyword */

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
   {
     if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
     {
       e_FileError(FullPathName,"unexpected end of file");
       return 0;
     }
     else if (   !WordMatchesKeyword(E_KW_STRESSES_LINK1,word)
              || !GetStresses(mesh->NElemsLink1, 
                              mesh->NStressPointsLink1, StressesLink1))
     {
       return 0;
     }
   }
   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
   {
     if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
     {
       e_FileError(FullPathName,"unexpected end of file");
       return 0;
     }
     else if (   !WordMatchesKeyword(E_KW_STRESSES_LINK2,word)
              || !GetStresses(mesh->NElemsLink2, 
                              mesh->NStressPointsLink2, StressesLink2))
     {
       return 0;
     }
   }
   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
   {
     if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
     {
       e_FileError(FullPathName,"unexpected end of file");
       return 0;
     }
     else if (   !WordMatchesKeyword(E_KW_STRESSES_LINK3,word)
              || !GetStresses(mesh->NElemsLink3,
                              mesh->NStressPointsLink3, StressesLink3))
     {
       return 0;
     }
   }
   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
   {
     if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
     {
       e_FileError(FullPathName,"unexpected end of file");
       return 0;
     }
     else if (   !WordMatchesKeyword(E_KW_STRESSES_LINK4,word)
              || !GetStresses(mesh->NElemsLink4,
                              mesh->NStressPointsLink4, StressesLink4))
     {
       return 0;
     }
   }
   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
   {
     if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
     {
       e_FileError(FullPathName,"unexpected end of file");
       return 0;
     }
     else if (   !WordMatchesKeyword(E_KW_STRESSES_LINK5,word)
              || !GetStresses(mesh->NElemsLink5,
                              mesh->NStressPointsLink5, StressesLink5))
     {
       return 0;
     }
   }
   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
   {
     if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
     {
       e_FileError(FullPathName,"unexpected end of file");
       return 0;
     }
     else if (   !WordMatchesKeyword(E_KW_STRESSES_TRIANG1,word)
              || !GetStresses(mesh->NElemsTriang1,
                              mesh->NStressPointsTriang1, StressesTriang1))
     {
       return 0;
     }
   }
   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
   {
     if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
     {
       e_FileError(FullPathName,"unexpected end of file");
       return 0;
     }
     else if (   !WordMatchesKeyword(E_KW_STRESSES_TRIANG2,word)
              || !GetStressesPerLayer(mesh->NElemsTriang2,
                                      mesh->CompMatsTriang2,
                                      mesh->NStressPointsTriang2,
                                      StressesTriang2))
     {
       return 0;
     }
   }
   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
   {
     if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
     {
       e_FileError(FullPathName,"unexpected end of file");
       return 0;
     }
     else if (   !WordMatchesKeyword(E_KW_STRESSES_TRIANG3,word)
              || !GetStresses(mesh->NElemsTriang3,
                              mesh->NStressPointsTriang3, StressesTriang3))
     {
       return 0;
     }
   }
   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
   {
     if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
     {
       e_FileError(FullPathName,"unexpected end of file");
       return 0;
     }
     else if (   !WordMatchesKeyword(E_KW_STRESSES_TRIANG4,word)
              || !GetStresses(mesh->NElemsTriang4,
                              mesh->NStressPointsTriang4, StressesTriang4))
     {
       return 0;
     }
   }
   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
   {
     if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
     {
       e_FileError(FullPathName,"unexpected end of file");
       return 0;
     }
     else if (   !WordMatchesKeyword(E_KW_STRESSES_TRIANG5,word)
              || !GetStresses(mesh->NElemsTriang5,
                              mesh->NStressPointsTriang5, StressesTriang5))
     {
       return 0;
     }
   }
   if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
   {
     if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
     {
       e_FileError(FullPathName,"unexpected end of file");
       return 0;
     }
     else if (   !WordMatchesKeyword(E_KW_STRESSES_QUAD1,word)
              || !GetStresses(mesh->NElemsQuad1,
                              mesh->NStressPointsQuad1, StressesQuad1))
     {
       return 0;
     }
   }
   if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
   {
     if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
     {
       e_FileError(FullPathName,"unexpected end of file");
       return 0;
     }
     else if (   !WordMatchesKeyword(E_KW_STRESSES_QUAD3,word)
              || !GetStressesPerLayer(mesh->NElemsQuad3,
                                      mesh->CompMatsQuad3,
                                      mesh->NStressPointsQuad3,
                                      StressesQuad3))
     {
       return 0;
     }
   }
   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
   {
     if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
     {
       e_FileError(FullPathName,"unexpected end of file");
       return 0;
     }
     else if (   !WordMatchesKeyword(E_KW_STRESSES_TETRAH1,word)
              || !GetStresses(mesh->NElemsTetrah1,
                              mesh->NStressPointsTetrah1, StressesTetrah1))
     {
       return 0;
     }
   }
   if (mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
   {
     if (GetKeyword(line,word,&LineNumber,FullPathName,fp) == EOF)
     {
       e_FileError(FullPathName,"unexpected end of file");
       return 0;
     }
     else if (   !WordMatchesKeyword(E_KW_STRESSES_BLOCK1,word)
              || !GetStresses(mesh->NElemsBlock1,
                              mesh->NStressPointsBlock1, StressesBlock1))
     {
       return 0;
     }
   }

   if (GetKeyword(line,word,&LineNumber,FullPathName,fp) != EOF)
   {
      e_FileLineError(LineNumber,FullPathName,"undefined keyword");
      return 0;
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* WordMatchesKeyword: To return whether a word is a specified keyword. */

/* The return value is 1 if the word matches the keyword, else 0. */


static int WordMatchesKeyword(char *Keyword, char *word)
{
   if (strcmp(Keyword,word) != 0)
   {
      if (IsAKeyword(word))
         e_FileLineError(LineNumber,FullPathName,"keyword out of sequence");
      else
         e_FileLineError(LineNumber,FullPathName,"undefined keyword");

      return 0;
   }
   else
      return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* Get... input functions: To read in data corresponding to a keyword. */

/* The return value is 1 if the data are read successfully, else 0. */

static int GetStresses(int NElements, int NStressPoints, double ***ElemStresses)
{
   int i, j;
   int StressIndex, Point;


   for (i = 0; i < NElements; i++)
     for(j = 0; j < NStressPoints; j++)
     {
       if (e_GetNextLine(line,&LineNumber,FullPathName,fp) == 0)
         return 0;

       if (sscanf(line,"%d %d %lf %lf %lf %lf %lf %lf",
             &StressIndex, &Point,
             &(ElemStresses[i][j][E_SIGMA_X]),
             &(ElemStresses[i][j][E_SIGMA_Y]),
             &(ElemStresses[i][j][E_SIGMA_Z]),
             &(ElemStresses[i][j][E_TAU_XY]),
             &(ElemStresses[i][j][E_TAU_YZ]),
             &(ElemStresses[i][j][E_TAU_ZX])  ) == 8)
       {
          if (StressIndex != i + 1)
          {
             e_FileLineError(LineNumber,FullPathName,
                             "stress index out of sequence");
             return 0;
          }
          if (Point != j + 1)
          {
             e_FileLineError(LineNumber,FullPathName,
                             "Gauss point index out of sequence");
             return 0;
          }
       }
       else
       {
          e_FileLineError(LineNumber,FullPathName,"cannot read elem stresses");
          return 0;
       }
     }

   return 1;
}

/* /////////////////////////////////////////// */

static int GetStressesPerLayer(int NElements,  COMPMAT1 **Material, 
                               int NStressPoints, double ****ElemStresses)
{
   int i, j, k;
   int StressIndex, Layer, Point;


   for (i = 0; i < NElements; i++)
     for(j = 0; j < Material[i]->NLayers; j++)
       for(k = 0; k < NStressPoints; k++)
       {
         if (e_GetNextLine(line,&LineNumber,FullPathName,fp) == 0)
           return 0;

         if (sscanf(line,"%d %d %d %lf %lf %lf %lf %lf %lf",
               &StressIndex, &Layer, &Point,
               &(ElemStresses[i][j][k][E_SIGMA_X]),
               &(ElemStresses[i][j][k][E_SIGMA_Y]),
               &(ElemStresses[i][j][k][E_SIGMA_Z]),
               &(ElemStresses[i][j][k][E_TAU_XY]),
               &(ElemStresses[i][j][k][E_TAU_YZ]),
               &(ElemStresses[i][j][k][E_TAU_ZX])  ) == 9)
         {
           if (StressIndex != i + 1)
           {
             e_FileLineError(LineNumber,FullPathName,
                             "elem index out of sequence");
             return 0;
           }
           if (Layer != j + 1)
           {
             e_FileLineError(LineNumber,FullPathName,
                             "layer index out of sequence");
             return 0;
           }
           if (Point != k + 1)
           {
             e_FileLineError(LineNumber,FullPathName,
                             "Gauss point index out of sequence");
             return 0;
           }
         }
         else
         {
           e_FileLineError(LineNumber,FullPathName,"cannot read stresses");
           return 0;
         }
       }

   return 1;
}



/* ////////////////////////////////////////////////////////////////////////// */
/* IsAKeyword: To return whether a word is a keyword. */

/* The return value is 1 if "word" is a keyword, else 0. */


static int IsAKeyword(char *word)
{
   if (     strcmp(E_KW_STRESSES_LINK1,word) == 0
       ||   strcmp(E_KW_STRESSES_LINK2,word) == 0
       ||   strcmp(E_KW_STRESSES_LINK3,word) == 0
       ||   strcmp(E_KW_STRESSES_LINK4,word) == 0
       ||   strcmp(E_KW_STRESSES_LINK5,word) == 0
       || strcmp(E_KW_STRESSES_TRIANG1,word) == 0
       || strcmp(E_KW_STRESSES_TRIANG2,word) == 0
       || strcmp(E_KW_STRESSES_TRIANG3,word) == 0
       || strcmp(E_KW_STRESSES_TRIANG4,word) == 0
       || strcmp(E_KW_STRESSES_TRIANG5,word) == 0
       ||   strcmp(E_KW_STRESSES_QUAD1,word) == 0
       ||   strcmp(E_KW_STRESSES_QUAD3,word) == 0
       || strcmp(E_KW_STRESSES_TETRAH1,word) == 0
       ||  strcmp(E_KW_STRESSES_BLOCK1,word) == 0)
      return TRUE;
   else
      return FALSE;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetMeshData: To set the MESH structure data. */


static void SetMeshData(MESH *mesh)
{
   mesh->Stresses.StressesLink1   = StressesLink1;
   mesh->Stresses.StressesLink2   = StressesLink2;
   mesh->Stresses.StressesLink3   = StressesLink3;
   mesh->Stresses.StressesLink4   = StressesLink4;
   mesh->Stresses.StressesLink5   = StressesLink5;
   mesh->Stresses.StressesTriang1 = StressesTriang1;
   mesh->Stresses.StressesTriang2 = StressesTriang2;
   mesh->Stresses.StressesTriang3 = StressesTriang3;
   mesh->Stresses.StressesTriang4 = StressesTriang4;
   mesh->Stresses.StressesTriang5 = StressesTriang5;
   mesh->Stresses.StressesQuad1   = StressesQuad1;
   mesh->Stresses.StressesQuad3   = StressesQuad3;
   mesh->Stresses.StressesTetrah1 = StressesTetrah1;
   mesh->Stresses.StressesBlock1  = StressesBlock1;

   mesh->SubStructSet[E_SS_NUM_STRESS] = TRUE;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* DeallocateAllMemory: To deallocate all memory. */


static void DeallocateAllMemory(MESH *mesh)
{
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
     e_free_cont_d_3D_matr(mesh->NElemsLink1, 
                           mesh->NStressPointsLink1, StressesLink1);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
     e_free_cont_d_3D_matr(mesh->NElemsLink2, 
                           mesh->NStressPointsLink2, StressesLink2);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
     e_free_cont_d_3D_matr(mesh->NElemsLink3, 
                           mesh->NStressPointsLink3, StressesLink3);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
     e_free_cont_d_3D_matr(mesh->NElemsLink4, 
                           mesh->NStressPointsLink4, StressesLink4);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
     e_free_cont_d_3D_matr(mesh->NElemsLink5, 
                           mesh->NStressPointsLink5, StressesLink5);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
     e_free_cont_d_3D_matr(mesh->NElemsTriang1, 
                           mesh->NStressPointsTriang1, StressesTriang1);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
     e_free_cont_d_4D_matr(mesh->NElemsTriang2, 
                           mesh->NStressPointsTriang2, 
                           maxNLayerTriang2, StressesTriang2);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
     e_free_cont_d_3D_matr(mesh->NElemsTriang3, 
                           mesh->NStressPointsTriang3, StressesTriang3);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
     e_free_cont_d_3D_matr(mesh->NElemsTriang4, 
                           mesh->NStressPointsTriang4, StressesTriang4);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
     e_free_cont_d_3D_matr(mesh->NElemsTriang5, 
                           mesh->NStressPointsTriang5, StressesTriang5);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
     e_free_cont_d_3D_matr(mesh->NElemsQuad1, 
                           mesh->NStressPointsQuad1, StressesQuad1);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
     e_free_cont_d_4D_matr(mesh->NElemsQuad3, 
                           mesh->NStressPointsQuad3, 
                           maxNLayerQuad3, StressesQuad3);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
     e_free_cont_d_3D_matr(mesh->NElemsTetrah1, 
                           mesh->NStressPointsTetrah1, StressesTetrah1);
   if(mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
     e_free_cont_d_3D_matr(mesh->NElemsBlock1, 
                           mesh->NStressPointsBlock1, StressesBlock1);

   e_free(FullPathName);
}


/* ////////////////////////////////////////////////////////////////////////// */












/* ////////////////////////////////////////////////////////////////////////// */
/* e_PutStresses: To output stresses. */
/* The return value is 1 if the data are written successfully, else 0. */


int e_PutStresses(MESH *mesh, char *OutputFile)
{
   int k, j ,i;

   char FirstParams = TRUE;        /* whether first mesh element type */

   FILE *fp;                       /* output file pointer */


   if (!mesh->SubStructSet[E_SS_NUM_STRESS])
   {
      e_GeneralError("e_PutStresses: stresses are not set");
      return 0;
   }
   else if (OpenOutputFile(OutputFile,E_STE,&fp) == NULL)
   {
      return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# Link1 stresses\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#   Stress point index\n");
      fprintf(fp,"#             Sigma x        Sigma y        Sigma z         Tau xy         Tau yz         Tau zx   \n");
      fprintf(fp,"#__________________________________________________________________________________________________\n");

      fprintf(fp,"\n%s\n",E_KW_STRESSES_LINK1);

      for (k = 0; k < mesh->NElemsLink1; k++)
      {
        for(i = 0; i < mesh->NStressPointsLink1; i++)
        {
          fprintf(fp,"%5d %2d ", k+1, i+1);
          if (   PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink1[k][i][E_SIGMA_X]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink1[k][i][E_SIGMA_Y]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink1[k][i][E_SIGMA_Z]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink1[k][i][E_TAU_XY]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink1[k][i][E_TAU_YZ]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink1[k][i][E_TAU_ZX]) < 0)
          {
            fclose(fp);
            return 0;
          }
          fprintf(fp,"\n");
        }
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# Link2 stresses\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#   Stress point index\n");
      fprintf(fp,"#             Sigma x        Sigma y        Sigma z         Tau xy         Tau yz         Tau zx   \n");
      fprintf(fp,"#__________________________________________________________________________________________________\n");

      fprintf(fp,"\n%s\n",E_KW_STRESSES_LINK2);

      for (k = 0; k < mesh->NElemsLink2; k++)
      {
        for(i = 0; i < mesh->NStressPointsLink2; i++)
        {
          fprintf(fp,"%5d %2d ", k+1, i+1);
          if (   PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink2[k][i][E_SIGMA_X]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink2[k][i][E_SIGMA_Y]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink2[k][i][E_SIGMA_Z]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink2[k][i][E_TAU_XY]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink2[k][i][E_TAU_YZ]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink2[k][i][E_TAU_ZX]) < 0)
          {
            fclose(fp);
            return 0;
          }
          fprintf(fp,"\n");
        }
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# Link3 stresses\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#   Stress point index\n");
      fprintf(fp,"#             Sigma x        Sigma y        Sigma z         Tau xy         Tau yz         Tau zx   \n");
      fprintf(fp,"#__________________________________________________________________________________________________\n");

      fprintf(fp,"\n%s\n",E_KW_STRESSES_LINK3);

      for (k = 0; k < mesh->NElemsLink3; k++)
      {
        for(i = 0; i < mesh->NStressPointsLink3; i++)
        {
          fprintf(fp,"%5d %2d ", k+1, i+1);
          if (   PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink3[k][i][E_SIGMA_X]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink3[k][i][E_SIGMA_Y]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink3[k][i][E_SIGMA_Z]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink3[k][i][E_TAU_XY]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink3[k][i][E_TAU_YZ]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink3[k][i][E_TAU_ZX]) < 0)
          {
            fclose(fp);
            return 0;
          }
          fprintf(fp,"\n");
        }
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# Link4 stresses\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#   Stress point index\n");
      fprintf(fp,"#             Sigma x        Sigma y        Sigma z         Tau xy         Tau yz         Tau zx   \n");
      fprintf(fp,"#__________________________________________________________________________________________________\n");

      fprintf(fp,"\n%s\n",E_KW_STRESSES_LINK4);

      for (k = 0; k < mesh->NElemsLink4; k++)
      {
        for(i = 0; i < mesh->NStressPointsLink4; i++)
        {
          fprintf(fp,"%5d %2d ", k+1, i+1);
          if (   PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink4[k][i][E_SIGMA_X]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink4[k][i][E_SIGMA_Y]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink4[k][i][E_SIGMA_Z]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink4[k][i][E_TAU_XY]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink4[k][i][E_TAU_YZ]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink4[k][i][E_TAU_ZX]) < 0)
          {
            fclose(fp);
            return 0;
          }
          fprintf(fp,"\n");
        }
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# Link5 stresses\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#   Stress point index\n");
      fprintf(fp,"#             Sigma x        Sigma y        Sigma z         Tau xy         Tau yz         Tau zx   \n");
      fprintf(fp,"#__________________________________________________________________________________________________\n");

      fprintf(fp,"\n%s\n",E_KW_STRESSES_LINK5);

      for (k = 0; k < mesh->NElemsLink5; k++)
      {
        for(i = 0; i < mesh->NStressPointsLink5; i++)
        {
          fprintf(fp,"%5d %2d ", k+1, i+1);
          if (   PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink5[k][i][E_SIGMA_X]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink5[k][i][E_SIGMA_Y]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink5[k][i][E_SIGMA_Z]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink5[k][i][E_TAU_XY]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink5[k][i][E_TAU_YZ]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesLink5[k][i][E_TAU_ZX]) < 0)
          {
            fclose(fp);
            return 0;
          }
          fprintf(fp,"\n");
        }
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# Triang1 stresses\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#   Stress point index\n");
      fprintf(fp,"#             Sigma x        Sigma y        Sigma z         Tau xy         Tau yz         Tau zx   \n");
      fprintf(fp,"#__________________________________________________________________________________________________\n");

      fprintf(fp,"\n%s\n",E_KW_STRESSES_TRIANG1);

      for (k = 0; k < mesh->NElemsTriang1; k++)
      {
        for(i = 0; i < mesh->NStressPointsTriang1; i++)
        {
          fprintf(fp,"%5d %2d ", k+1, i+1);
          if (   PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang1[k][i][E_SIGMA_X]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang1[k][i][E_SIGMA_Y]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang1[k][i][E_SIGMA_Z]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang1[k][i][E_TAU_XY]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang1[k][i][E_TAU_YZ]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang1[k][i][E_TAU_ZX]) < 0)
          {
            fclose(fp);
            return 0;
          }
          fprintf(fp,"\n");
        }
      }
   }


   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# Triang2 stresses\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#   Layer index\n");
      fprintf(fp,"#     Gauss point index\n");
      fprintf(fp,"#                Sigma x        Sigma y        Sigma z         Tau xy         Tau yz         Tau zx   \n");
      fprintf(fp,"#_____________________________________________________________________________________________________\n");

      fprintf(fp,"\n%s\n",E_KW_STRESSES_TRIANG2);

      for (k = 0; k < mesh->NElemsTriang2; k++)
        for(j = 0; j < mesh->CompMatsTriang2[k]->NLayers; j++)
          for(i = 0; i < mesh->NStressPointsTriang2; i++)
          {
            fprintf(fp,"%5d %2d %2d ", k+1, j+1, i+1);
            if (   PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang2[k][j][i][E_SIGMA_X]) < 0
                || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang2[k][j][i][E_SIGMA_Y]) < 0
                || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang2[k][j][i][E_SIGMA_Z]) < 0
                || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang2[k][j][i][E_TAU_XY]) < 0
                || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang2[k][j][i][E_TAU_YZ]) < 0
                || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang2[k][j][i][E_TAU_ZX]) < 0)
            {
              fclose(fp);
              return 0;
            }
            fprintf(fp,"\n");
          }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# Triang3 stresses\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#   Stress point index\n");
      fprintf(fp,"#             Sigma x        Sigma y        Sigma z         Tau xy         Tau yz         Tau zx   \n");
      fprintf(fp,"#__________________________________________________________________________________________________\n");

      fprintf(fp,"\n%s\n",E_KW_STRESSES_TRIANG3);

      for (k = 0; k < mesh->NElemsTriang3; k++)
      {
        for(i = 0; i < mesh->NStressPointsTriang3; i++)
        {
          fprintf(fp,"%5d %2d ", k+1, i+1);
          if (   PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang3[k][i][E_SIGMA_X]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang3[k][i][E_SIGMA_Y]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang3[k][i][E_SIGMA_Z]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang3[k][i][E_TAU_XY]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang3[k][i][E_TAU_YZ]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang3[k][i][E_TAU_ZX]) < 0)
          {
            fclose(fp);
            return 0;
          }
          fprintf(fp,"\n");
        }
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# Triang4 stresses\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#   Stress point index\n");
      fprintf(fp,"#             Sigma x        Sigma y        Sigma z         Tau xy         Tau yz         Tau zx   \n");
      fprintf(fp,"#__________________________________________________________________________________________________\n");

      fprintf(fp,"\n%s\n",E_KW_STRESSES_TRIANG4);

      for (k = 0; k < mesh->NElemsTriang4; k++)
      {
        for(i = 0; i < mesh->NStressPointsTriang4; i++)
        {
          fprintf(fp,"%5d %2d ", k+1, i+1);
          if (   PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang4[k][i][E_SIGMA_X]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang4[k][i][E_SIGMA_Y]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang4[k][i][E_SIGMA_Z]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang4[k][i][E_TAU_XY]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang4[k][i][E_TAU_YZ]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang4[k][i][E_TAU_ZX]) < 0)
          {
            fclose(fp);
            return 0;
          }
          fprintf(fp,"\n");
        }
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# Triang5 stresses\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#   Stress point index\n");
      fprintf(fp,"#             Sigma x        Sigma y        Sigma z         Tau xy         Tau yz         Tau zx   \n");
      fprintf(fp,"#__________________________________________________________________________________________________\n");

      fprintf(fp,"\n%s\n",E_KW_STRESSES_TRIANG5);

      for (k = 0; k < mesh->NElemsTriang5; k++)
      {
        for(i = 0; i < mesh->NStressPointsTriang5; i++)
        {
          fprintf(fp,"%5d %2d ", k+1, i+1);
          if (   PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang5[k][i][E_SIGMA_X]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang5[k][i][E_SIGMA_Y]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang5[k][i][E_SIGMA_Z]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang5[k][i][E_TAU_XY]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang5[k][i][E_TAU_YZ]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTriang5[k][i][E_TAU_ZX]) < 0)
          {
            fclose(fp);
            return 0;
          }
          fprintf(fp,"\n");
        }
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# Quad1 stresses\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#   Stress point index\n");
      fprintf(fp,"#             Sigma x        Sigma y        Sigma z         Tau xy         Tau yz         Tau zx   \n");
      fprintf(fp,"#__________________________________________________________________________________________________\n");

      fprintf(fp,"\n%s\n",E_KW_STRESSES_QUAD1);

      for (k = 0; k < mesh->NElemsQuad1; k++)
      {
        for(i = 0; i < mesh->NStressPointsQuad1; i++)
        {
          fprintf(fp,"%5d %2d ", k+1, i+1);
          if (   PrintEng(fp,"%15e.5",mesh->Stresses.StressesQuad1[k][i][E_SIGMA_X]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesQuad1[k][i][E_SIGMA_Y]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesQuad1[k][i][E_SIGMA_Z]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesQuad1[k][i][E_TAU_XY]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesQuad1[k][i][E_TAU_YZ]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesQuad1[k][i][E_TAU_ZX]) < 0)
          {
            fclose(fp);
            return 0;
          }
          fprintf(fp,"\n");
        }
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# Quad3 stresses\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#   Layer index\n");
      fprintf(fp,"#     Gauss point index\n");
      fprintf(fp,"#                Sigma x        Sigma y        Sigma z         Tau xy         Tau yz         Tau zx   \n");
      fprintf(fp,"#_____________________________________________________________________________________________________\n");

      fprintf(fp,"\n%s\n",E_KW_STRESSES_QUAD3);

      for (k = 0; k < mesh->NElemsQuad3; k++)
        for(j = 0; j < mesh->CompMatsQuad3[k]->NLayers; j++)
          for(i = 0; i < mesh->NStressPointsQuad3; i++)
          {
            fprintf(fp,"%5d %2d %2d ", k+1, j+1, i+1);
            if (   PrintEng(fp,"%15e.5",mesh->Stresses.StressesQuad3[k][j][i][E_SIGMA_X]) < 0
                || PrintEng(fp,"%15e.5",mesh->Stresses.StressesQuad3[k][j][i][E_SIGMA_Y]) < 0
                || PrintEng(fp,"%15e.5",mesh->Stresses.StressesQuad3[k][j][i][E_SIGMA_Z]) < 0
                || PrintEng(fp,"%15e.5",mesh->Stresses.StressesQuad3[k][j][i][E_TAU_XY]) < 0
                || PrintEng(fp,"%15e.5",mesh->Stresses.StressesQuad3[k][j][i][E_TAU_YZ]) < 0
                || PrintEng(fp,"%15e.5",mesh->Stresses.StressesQuad3[k][j][i][E_TAU_ZX]) < 0)
            {
              fclose(fp);
              return 0;
            }
            fprintf(fp,"\n");
          }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# Tetrah1 stresses\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#   Stress point index\n");
      fprintf(fp,"#             Sigma x        Sigma y        Sigma z         Tau xy         Tau yz         Tau zx   \n");
      fprintf(fp,"#__________________________________________________________________________________________________\n");

      fprintf(fp,"\n%s\n",E_KW_STRESSES_TETRAH1);

      for (k = 0; k < mesh->NElemsTetrah1; k++)
      {
        for(i = 0; i < mesh->NStressPointsTetrah1; i++)
        {
          fprintf(fp,"%5d %2d ", k+1, i+1);
          if (   PrintEng(fp,"%15e.5",mesh->Stresses.StressesTetrah1[k][i][E_SIGMA_X]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTetrah1[k][i][E_SIGMA_Y]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTetrah1[k][i][E_SIGMA_Z]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTetrah1[k][i][E_TAU_XY]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTetrah1[k][i][E_TAU_YZ]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesTetrah1[k][i][E_TAU_ZX]) < 0)
          {
            fclose(fp);
            return 0;
          }
          fprintf(fp,"\n");
        }
      }
   }

   if (mesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
   {
      if (FirstParams)
         FirstParams = FALSE;
      else
         fprintf(fp,"\n\n");

      fprintf(fp,"# Block1 stresses\n");
      fprintf(fp,"# element index\n");
      fprintf(fp,"#   Stress point index\n");
      fprintf(fp,"#             Sigma x        Sigma y        Sigma z         Tau xy         Tau yz         Tau zx   \n");
      fprintf(fp,"#__________________________________________________________________________________________________\n");

      fprintf(fp,"\n%s\n",E_KW_STRESSES_BLOCK1);

      for (k = 0; k < mesh->NElemsBlock1; k++)
      {
        for(i = 0; i < mesh->NStressPointsBlock1; i++)
        {
          fprintf(fp,"%5d %2d ", k+1, i+1);
          if (   PrintEng(fp,"%15e.5",mesh->Stresses.StressesBlock1[k][i][E_SIGMA_X]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesBlock1[k][i][E_SIGMA_Y]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesBlock1[k][i][E_SIGMA_Z]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesBlock1[k][i][E_TAU_XY]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesBlock1[k][i][E_TAU_YZ]) < 0
              || PrintEng(fp,"%15e.5",mesh->Stresses.StressesBlock1[k][i][E_TAU_ZX]) < 0)
          {
            fclose(fp);
            return 0;
          }
          fprintf(fp,"\n");
        }
      }
   }

   fclose(fp);
   return 1;
}
