/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                             e_CreateSubMesh()                              */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include "e_pltfrm.h"
#include <string.h>
#include <stdio.h>
#include <e_lib.h>

#include "submesh.h"


/* ////////////////////////////////////////////////////////////////////////// */
/* e_CreateSubMesh: To create a submesh. */
/* The return value is 1 if no input errors occur, else 0. */


int e_CreateSubMesh(MESH *mesh, MESH **SubMesh, char *SubMeshTitle,
                                                 int SubdomainIndex)
{
   if (  mesh->NElemsLink1
       + mesh->NElemsLink2
       + mesh->NElemsLink3
       + mesh->NElemsLink4
       + mesh->NElemsLink5
       + mesh->NElemsTriang1
       + mesh->NElemsTriang2
       + mesh->NElemsTriang3
       + mesh->NElemsTriang4
       + mesh->NElemsTriang5
       + mesh->NElemsQuad1
       + mesh->NElemsQuad3
       + mesh->NElemsTetrah1
       + mesh->NElemsBlock1  != mesh->TotalNElems)
   {
      e_GeneralError("e_CreateSubMesh: unsupported element type");
      return 0;
   }
   else if (!mesh->SubStructSet[E_SS_NUM_DECOMP])
   {
      e_GeneralError("e_CreateSubMesh: decomposition data not set");
      return 0;
   }
   else if (!Initialise(mesh,SubMesh,SubMeshTitle,SubdomainIndex))
   {
      return 0;
   }
   else if (!CreateMesh(mesh,*SubMesh,SubMeshTitle,SubdomainIndex))
   {
      DeallocateAllMemory(*SubMesh);
      return 0;
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* Initialise: To initialise variables. */

/* The return value is 1 if the initialisation is successfully, else 0. */


static int Initialise(MESH *mesh, MESH **SubMesh, char* SubMeshTitle,
                                                   int SubdomainIndex)
{
   if (!AllocateMemory(mesh,SubMesh,SubMeshTitle,SubdomainIndex))
      return 0;

   ZeroSubMesh(*SubMesh);

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* AllocateMemory: To allocate sub-mesh memory. */

/* The return value is 1 if the allocation is successfully, else 0. */


static int AllocateMemory(MESH *mesh, MESH **SubMesh, char* SubMeshTitle,
                                                       int SubdomainIndex)
{
   int k;


   if ((*SubMesh = (MESH *)e_calloc(1,sizeof(MESH))) == NULL)
      return 0;

   e_NullifyMesh(*SubMesh);

   if (!CountSubMesh(mesh,*SubMesh,SubdomainIndex))
      return 0;

   if (   ((*SubMesh)->Title
               = e_alloc_c_vect(strlen(SubMeshTitle) + 1)) == NULL

       || ((*SubMesh)->MeshPointCoords
               = e_alloc_cont_d_matr((*SubMesh)->NMeshPoints,E_NDIMS)) == NULL

       || (    (*SubMesh)->NElemsLink1 > 0
            && ((*SubMesh)->NodesLink1
                    = e_alloc_cont_i_matr
                         ((*SubMesh)->NElemsLink1,E_NNPE_LINK1)) == NULL)

       || (    (*SubMesh)->NElemsLink2 > 0
            && ((*SubMesh)->NodesLink2
                    = e_alloc_cont_i_matr
                         ((*SubMesh)->NElemsLink2,E_NNPE_LINK2)) == NULL)

       || (    (*SubMesh)->NElemsLink3 > 0
            && ((*SubMesh)->NodesLink3
                    = e_alloc_cont_i_matr
                         ((*SubMesh)->NElemsLink3,E_NNPE_LINK3)) == NULL)

       || (    (*SubMesh)->NElemsLink4 > 0
            && ((*SubMesh)->NodesLink4
                    = e_alloc_cont_i_matr
                         ((*SubMesh)->NElemsLink4,E_NNPE_LINK4)) == NULL)

       || (    (*SubMesh)->NElemsLink5 > 0
            && ((*SubMesh)->NodesLink5
                    = e_alloc_cont_i_matr
                         ((*SubMesh)->NElemsLink5,E_NNPE_LINK5)) == NULL)

       || (    (*SubMesh)->NElemsTriang1 > 0
            && ((*SubMesh)->NodesTriang1
                    = e_alloc_cont_i_matr
                         ((*SubMesh)->NElemsTriang1,E_NNPE_TRIANG1)) == NULL)

       || (    (*SubMesh)->NElemsTriang2 > 0
            && ((*SubMesh)->NodesTriang2
                    = e_alloc_cont_i_matr
                         ((*SubMesh)->NElemsTriang2,E_NNPE_TRIANG2)) == NULL)

       || (    (*SubMesh)->NElemsTriang3 > 0
            && ((*SubMesh)->NodesTriang3
                    = e_alloc_cont_i_matr
                         ((*SubMesh)->NElemsTriang3,E_NNPE_TRIANG3)) == NULL)

       || (    (*SubMesh)->NElemsTriang4 > 0
            && ((*SubMesh)->NodesTriang4
                    = e_alloc_cont_i_matr
                         ((*SubMesh)->NElemsTriang4,E_NNPE_TRIANG4)) == NULL)

       || (    (*SubMesh)->NElemsTriang5 > 0
            && ((*SubMesh)->NodesTriang5
                    = e_alloc_cont_i_matr
                         ((*SubMesh)->NElemsTriang5,E_NNPE_TRIANG5)) == NULL)

       || (    (*SubMesh)->NElemsQuad1 > 0
            && ((*SubMesh)->NodesQuad1
                    = e_alloc_cont_i_matr
                         ((*SubMesh)->NElemsQuad1,E_NNPE_QUAD1)) == NULL)

       || (    (*SubMesh)->NElemsQuad3 > 0
            && ((*SubMesh)->NodesQuad3
                    = e_alloc_cont_i_matr
                         ((*SubMesh)->NElemsQuad3,E_NNPE_QUAD3)) == NULL)

       || (    (*SubMesh)->NElemsTetrah1 > 0
            && ((*SubMesh)->NodesTetrah1
                    = e_alloc_cont_i_matr
                         ((*SubMesh)->NElemsTetrah1,E_NNPE_TETRAH1)) == NULL)

       || (    (*SubMesh)->NElemsBlock1 > 0
            && ((*SubMesh)->NodesBlock1
                    = e_alloc_cont_i_matr
                         ((*SubMesh)->NElemsBlock1,E_NNPE_BLOCK1)) == NULL)

       || (((*SubMesh)->NodalDOF
                = e_alloc_i_vect((*SubMesh)->NNodes)) == NULL)

       || (    (*SubMesh)->NBCNodes > 0
            && (   ((*SubMesh)->BCNodes
                        = e_alloc_i_vect((*SubMesh)->NBCNodes)) == NULL
                || ((*SubMesh)->BCTypes
                        = e_alloc_frag_i_matr
                              ((*SubMesh)->NBCNodes,E_DOF_MAX)) == NULL
                || ((*SubMesh)->BCDispls
                        = e_alloc_frag_d_matr
                              ((*SubMesh)->NBCNodes,E_DOF_MAX)) == NULL
                || ((*SubMesh)->BCSpringConsts
                        = e_alloc_frag_d_matr
                              ((*SubMesh)->NBCNodes,E_DOF_MAX)) == NULL))

       || (    (*SubMesh)->NLoadedNodes > 0
            && (   ((*SubMesh)->LoadedNodes
                        = e_alloc_i_vect((*SubMesh)->NLoadedNodes)) == NULL
                || ((*SubMesh)->Loads
                        = e_alloc_frag_d_matr
                              ((*SubMesh)->NLoadedNodes,E_DOF_MAX)) == NULL))

       || (   (*SubMesh)->NMats > 0
           && (   ((*SubMesh)->Mats
                       = (void **)
                         e_calloc((*SubMesh)->TotalNElems,
                                                 sizeof(void *))) == NULL
               || ((*SubMesh)->Props
                       = (MAT *)
                         e_calloc((*SubMesh)->NMats,sizeof(MAT))) == NULL))   )
   {
      DeallocateAllMemory(*SubMesh);
      return 0;
   }

   for (k = 0; k < (*SubMesh)->NMats; k++)
      (*SubMesh)->Props[k].Name = NULL;

   if (   (*SubMesh)->NCompMatsType1 > 0
       && (((*SubMesh)->CompPropsType1
                = (COMPMAT1 *)e_calloc((*SubMesh)->NCompMatsType1,
                                          sizeof(COMPMAT1))) == NULL))
   {
      DeallocateAllMemory(*SubMesh);
      return 0;
   }

   for (k = 0; k < (*SubMesh)->NCompMatsType1; k++)
   {
      (*SubMesh)->CompPropsType1[k].Name = NULL;
      (*SubMesh)->CompPropsType1[k].LayerMats = NULL;
      (*SubMesh)->CompPropsType1[k].Thicknesses = NULL;
   }

   if (mesh->KeywordSet[E_KW_NUM_REMESH_DATA])
   {
      if (   ((*SubMesh)->NodeTypes
                      = e_alloc_i_vect((*SubMesh)->NMeshPoints)) == NULL
          || ((*SubMesh)->NNurbsCurves
                      = e_alloc_i_vect((*SubMesh)->NMeshPoints)) == NULL
          || ((*SubMesh)->NurbsCurveParams
                  = e_init_frag_d_matr((*SubMesh)->NMeshPoints)) == NULL
          || ((*SubMesh)->NurbsCurves
                  = (NURBSCURVE ***)
                    e_calloc((*SubMesh)->NMeshPoints,
                                         sizeof(NURBSCURVE **))) == NULL)
      {
         DeallocateAllMemory(*SubMesh);
         return 0;
      }

      for (k = 0; k < (*SubMesh)->NMeshPoints; k++)
         (*SubMesh)->NurbsCurves[k] = NULL;

      if (   (*SubMesh)->Geom.NNurbsCurves > 0
          && ((*SubMesh)->Geom.NurbsCurves
                  = (NURBSCURVE *)
                     e_calloc((*SubMesh)->Geom.NNurbsCurves,
                                          sizeof(NURBSCURVE))) == NULL)
      {
         DeallocateAllMemory(*SubMesh);
         return 0;
      }

      for (k = 0; k < (*SubMesh)->Geom.NNurbsCurves; k++)
      {
         (*SubMesh)->Geom.NurbsCurves[k].Name = NULL;
         (*SubMesh)->Geom.NurbsCurves[k].CntlPoints = NULL;
         (*SubMesh)->Geom.NurbsCurves[k].BCTypes = NULL;
         (*SubMesh)->Geom.NurbsCurves[k].BCDispls = NULL;
         (*SubMesh)->Geom.NurbsCurves[k].BCSpringConsts = NULL;
         (*SubMesh)->Geom.NurbsCurves[k].PointLoad = NULL;
         (*SubMesh)->Geom.NurbsCurves[k].LinLoadGCBeg = NULL;
         (*SubMesh)->Geom.NurbsCurves[k].LinLoadGCEnd = NULL;
         (*SubMesh)->Geom.NurbsCurves[k].LinLoadCCBeg = NULL;
         (*SubMesh)->Geom.NurbsCurves[k].LinLoadCCEnd = NULL;
      }
   }

   if (   ((*SubMesh)->GNIs = e_alloc_i_vect((*SubMesh)->NNodes)) == NULL

       || ((*SubMesh)->GEIs = e_alloc_i_vect((*SubMesh)->TotalNElems)) == NULL

       || (   (*SubMesh)->Geom.NEndPoints > 0
           && ((*SubMesh)->Geom.EndPointCoords
                   = e_alloc_cont_d_matr
                     ((*SubMesh)->Geom.NEndPoints,E_NDIMS)) == NULL)

       || (   mesh->SubStructSet[E_SS_NUM_MESHPARAM]
           && mesh->MeshParam.NodalMeshParamsSet
           && ((*SubMesh)->MeshParam.NodalMeshParams
                   = e_alloc_d_vect((*SubMesh)->NMeshPoints)) == NULL)

       || (   mesh->SubStructSet[E_SS_NUM_MESHPARAM]
           && mesh->MeshParam.ElemMeshParamsSet
           && ((*SubMesh)->MeshParam.ElemMeshParams
                   = e_alloc_d_vect((*SubMesh)->TotalNElems)) == NULL)

       || (   mesh->SubStructSet[E_SS_NUM_FEERROR]
           && ((*SubMesh)->FEError.FEErrors
                   = e_alloc_d_vect((*SubMesh)->TotalNElems)) == NULL)       )
   {
      DeallocateAllMemory(*SubMesh);
      return 0;
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* CountSubMesh: To determine the sizes of the sub-mesh arrays. */

/* The return value is 1 if the sizes are calculated successfully, else 0. */


static int CountSubMesh(MESH *mesh, MESH *SubMesh, int SubdomainIndex)
{
   int k;
   int *BCNumbers;
   int *CompPropsIndicesType1;
   int *LoadedNodeNumbers;
   int *PropsIndices;

   char *EndPointStored;
   char *NodeStored;
   char *NurbsCurveStored;


   NodeStored = NULL;
   EndPointStored = NULL;
   NurbsCurveStored = NULL;
   BCNumbers = NULL;
   LoadedNodeNumbers = NULL;
   PropsIndices = NULL;
   CompPropsIndicesType1 = NULL;

   if ((NodeStored = e_alloc_c_vect(mesh->NNodes)) == NULL)
      return 0;

   ZeroSubMesh(SubMesh);

   CountNodesAndElements(mesh,SubMesh,
                         E_NNPE_LINK1,E_NVPE_LINK1,
                         mesh->NElemsLink1,
                         mesh->NodesLink1,
                         mesh->FirstNetIndex[E_ELEM_TYPE_LINK1],
                         &SubMesh->NElemsLink1,
                         NodeStored,SubdomainIndex              );
   CountNodesAndElements(mesh,SubMesh,
                         E_NNPE_LINK2,E_NVPE_LINK2,
                         mesh->NElemsLink2,
                         mesh->NodesLink2,
                         mesh->FirstNetIndex[E_ELEM_TYPE_LINK2],
                         &SubMesh->NElemsLink2,
                         NodeStored,SubdomainIndex              );
   CountNodesAndElements(mesh,SubMesh,
                         E_NNPE_LINK3,E_NVPE_LINK3,
                         mesh->NElemsLink3,
                         mesh->NodesLink3,
                         mesh->FirstNetIndex[E_ELEM_TYPE_LINK3],
                         &SubMesh->NElemsLink3,
                         NodeStored,SubdomainIndex              );
   CountNodesAndElements(mesh,SubMesh,
                         E_NNPE_LINK4,E_NVPE_LINK4,
                         mesh->NElemsLink4,
                         mesh->NodesLink4,
                         mesh->FirstNetIndex[E_ELEM_TYPE_LINK4],
                         &SubMesh->NElemsLink4,
                         NodeStored,SubdomainIndex              );
   CountNodesAndElements(mesh,SubMesh,
                         E_NNPE_LINK5,E_NVPE_LINK5,
                         mesh->NElemsLink5,
                         mesh->NodesLink5,
                         mesh->FirstNetIndex[E_ELEM_TYPE_LINK5],
                         &SubMesh->NElemsLink5,
                         NodeStored,SubdomainIndex              );
   CountNodesAndElements(mesh,SubMesh,
                         E_NNPE_TRIANG1,E_NVPE_TRIANG1,
                         mesh->NElemsTriang1,
                         mesh->NodesTriang1,
                         mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1],
                         &SubMesh->NElemsTriang1,
                         NodeStored,SubdomainIndex                );
   CountNodesAndElements(mesh,SubMesh,
                         E_NNPE_TRIANG2,E_NVPE_TRIANG2,
                         mesh->NElemsTriang2,
                         mesh->NodesTriang2,
                         mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2],
                         &SubMesh->NElemsTriang2,
                         NodeStored,SubdomainIndex                );
   CountNodesAndElements(mesh,SubMesh,
                         E_NNPE_TRIANG3,E_NVPE_TRIANG3,
                         mesh->NElemsTriang3,
                         mesh->NodesTriang3,
                         mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3],
                         &SubMesh->NElemsTriang3,
                         NodeStored,SubdomainIndex                );
   CountNodesAndElements(mesh,SubMesh,
                         E_NNPE_TRIANG4,E_NVPE_TRIANG4,
                         mesh->NElemsTriang4,
                         mesh->NodesTriang4,
                         mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4],
                         &SubMesh->NElemsTriang4,
                         NodeStored,SubdomainIndex                );
   CountNodesAndElements(mesh,SubMesh,
                         E_NNPE_TRIANG5,E_NVPE_TRIANG5,
                         mesh->NElemsTriang5,
                         mesh->NodesTriang5,
                         mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5],
                         &SubMesh->NElemsTriang5,
                         NodeStored,SubdomainIndex                );
   CountNodesAndElements(mesh,SubMesh,
                         E_NNPE_QUAD1,E_NVPE_QUAD1,
                         mesh->NElemsQuad1,
                         mesh->NodesQuad1,
                         mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1],
                         &SubMesh->NElemsQuad1,
                         NodeStored,SubdomainIndex              );
   CountNodesAndElements(mesh,SubMesh,
                         E_NNPE_QUAD3,E_NVPE_QUAD3,
                         mesh->NElemsQuad3,
                         mesh->NodesQuad3,
                         mesh->FirstNetIndex[E_ELEM_TYPE_QUAD3],
                         &SubMesh->NElemsQuad3,
                         NodeStored,SubdomainIndex              );
   CountNodesAndElements(mesh,SubMesh,
                         E_NNPE_TETRAH1,E_NVPE_TETRAH1,
                         mesh->NElemsTetrah1,
                         mesh->NodesTetrah1,
                         mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1],
                         &SubMesh->NElemsTetrah1,
                         NodeStored,SubdomainIndex               );
   CountNodesAndElements(mesh,SubMesh,
                         E_NNPE_BLOCK1,E_NVPE_BLOCK1,
                         mesh->NElemsBlock1,
                         mesh->NodesBlock1,
                         mesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1],
                         &SubMesh->NElemsBlock1,
                         NodeStored,SubdomainIndex               );

   if (SubMesh->TotalNElems == 0)
   {
      e_GeneralError("e_CreateSubMesh: sub-mesh contains zero elements");
      e_free(NodeStored);
      return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_BCS])
   {
      if ((BCNumbers = e_alloc_i_vect(mesh->NNodes)) == NULL)
      {
         e_free(NodeStored);
         return 0;
      }

      for (k = 0; k < mesh->NNodes; k++)
         NodeStored[k] = FALSE;

      for (k = 0; k < mesh->NBCNodes; k++)
         BCNumbers[mesh->BCNodes[k] - 1] = k + 1;

      CountBCNodes(mesh,SubMesh,E_NNPE_LINK1,
                   mesh->NElemsLink1,
                   mesh->NodesLink1,
                   mesh->FirstNetIndex[E_ELEM_TYPE_LINK1],
                   BCNumbers,NodeStored,SubdomainIndex    );
      CountBCNodes(mesh,SubMesh,E_NNPE_LINK2,
                   mesh->NElemsLink2,
                   mesh->NodesLink2,
                   mesh->FirstNetIndex[E_ELEM_TYPE_LINK2],
                   BCNumbers,NodeStored,SubdomainIndex    );
      CountBCNodes(mesh,SubMesh,E_NNPE_LINK3,
                   mesh->NElemsLink3,
                   mesh->NodesLink3,
                   mesh->FirstNetIndex[E_ELEM_TYPE_LINK3],
                   BCNumbers,NodeStored,SubdomainIndex    );
      CountBCNodes(mesh,SubMesh,E_NNPE_LINK4,
                   mesh->NElemsLink4,
                   mesh->NodesLink4,
                   mesh->FirstNetIndex[E_ELEM_TYPE_LINK4],
                   BCNumbers,NodeStored,SubdomainIndex    );
      CountBCNodes(mesh,SubMesh,E_NNPE_LINK5,
                   mesh->NElemsLink5,
                   mesh->NodesLink5,
                   mesh->FirstNetIndex[E_ELEM_TYPE_LINK5],
                   BCNumbers,NodeStored,SubdomainIndex    );
      CountBCNodes(mesh,SubMesh,E_NNPE_TRIANG1,
                   mesh->NElemsTriang1,
                   mesh->NodesTriang1,
                   mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1],
                   BCNumbers,NodeStored,SubdomainIndex      );
      CountBCNodes(mesh,SubMesh,E_NNPE_TRIANG2,
                   mesh->NElemsTriang2,
                   mesh->NodesTriang2,
                   mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2],
                   BCNumbers,NodeStored,SubdomainIndex      );
      CountBCNodes(mesh,SubMesh,E_NNPE_TRIANG3,
                   mesh->NElemsTriang3,
                   mesh->NodesTriang3,
                   mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3],
                   BCNumbers,NodeStored,SubdomainIndex      );
      CountBCNodes(mesh,SubMesh,E_NNPE_TRIANG4,
                   mesh->NElemsTriang4,
                   mesh->NodesTriang4,
                   mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4],
                   BCNumbers,NodeStored,SubdomainIndex      );
      CountBCNodes(mesh,SubMesh,E_NNPE_TRIANG5,
                   mesh->NElemsTriang5,
                   mesh->NodesTriang5,
                   mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5],
                   BCNumbers,NodeStored,SubdomainIndex      );
      CountBCNodes(mesh,SubMesh,E_NNPE_QUAD1,
                   mesh->NElemsQuad1,
                   mesh->NodesQuad1,
                   mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1],
                   BCNumbers,NodeStored,SubdomainIndex    );
      CountBCNodes(mesh,SubMesh,E_NNPE_QUAD3,
                   mesh->NElemsQuad3,
                   mesh->NodesQuad3,
                   mesh->FirstNetIndex[E_ELEM_TYPE_QUAD3],
                   BCNumbers,NodeStored,SubdomainIndex    );
      CountBCNodes(mesh,SubMesh,E_NNPE_TETRAH1,
                   mesh->NElemsTetrah1,
                   mesh->NodesTetrah1,
                   mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1],
                   BCNumbers,NodeStored,SubdomainIndex      );
      CountBCNodes(mesh,SubMesh,E_NNPE_BLOCK1,
                   mesh->NElemsBlock1,
                   mesh->NodesBlock1,
                   mesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1],
                   BCNumbers,NodeStored,SubdomainIndex     );

      e_free(BCNumbers);
   }

   if (mesh->KeywordSet[E_KW_NUM_LOADS])
   {
      if ((LoadedNodeNumbers = e_alloc_i_vect(mesh->NNodes)) == NULL)
      {
         e_free(NodeStored);
         return 0;
      }

      for (k = 0; k < mesh->NNodes; k++)
         NodeStored[k] = FALSE;

      for (k = 0; k < mesh->NLoadedNodes; k++)
         LoadedNodeNumbers[mesh->LoadedNodes[k] - 1] = k + 1;

      CountLoadedNodes(mesh,SubMesh,E_NNPE_LINK1,
                       mesh->NElemsLink1,
                       mesh->NodesLink1,
                       mesh->FirstNetIndex[E_ELEM_TYPE_LINK1],
                       LoadedNodeNumbers,NodeStored,SubdomainIndex);
      CountLoadedNodes(mesh,SubMesh,E_NNPE_LINK2,
                       mesh->NElemsLink2,
                       mesh->NodesLink2,
                       mesh->FirstNetIndex[E_ELEM_TYPE_LINK2],
                       LoadedNodeNumbers,NodeStored,SubdomainIndex);
      CountLoadedNodes(mesh,SubMesh,E_NNPE_LINK3,
                       mesh->NElemsLink3,
                       mesh->NodesLink3,
                       mesh->FirstNetIndex[E_ELEM_TYPE_LINK3],
                       LoadedNodeNumbers,NodeStored,SubdomainIndex);
      CountLoadedNodes(mesh,SubMesh,E_NNPE_LINK4,
                       mesh->NElemsLink4,
                       mesh->NodesLink4,
                       mesh->FirstNetIndex[E_ELEM_TYPE_LINK4],
                       LoadedNodeNumbers,NodeStored,SubdomainIndex);
      CountLoadedNodes(mesh,SubMesh,E_NNPE_LINK5,
                       mesh->NElemsLink5,
                       mesh->NodesLink5,
                       mesh->FirstNetIndex[E_ELEM_TYPE_LINK5],
                       LoadedNodeNumbers,NodeStored,SubdomainIndex);
      CountLoadedNodes(mesh,SubMesh,E_NNPE_TRIANG1,
                       mesh->NElemsTriang1,
                       mesh->NodesTriang1,
                       mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1],
                       LoadedNodeNumbers,NodeStored,SubdomainIndex);
      CountLoadedNodes(mesh,SubMesh,E_NNPE_TRIANG2,
                       mesh->NElemsTriang2,
                       mesh->NodesTriang2,
                       mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2],
                       LoadedNodeNumbers,NodeStored,SubdomainIndex);
      CountLoadedNodes(mesh,SubMesh,E_NNPE_TRIANG3,
                       mesh->NElemsTriang3,
                       mesh->NodesTriang3,
                       mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3],
                       LoadedNodeNumbers,NodeStored,SubdomainIndex);
      CountLoadedNodes(mesh,SubMesh,E_NNPE_TRIANG4,
                       mesh->NElemsTriang4,
                       mesh->NodesTriang4,
                       mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4],
                       LoadedNodeNumbers,NodeStored,SubdomainIndex);
      CountLoadedNodes(mesh,SubMesh,E_NNPE_TRIANG5,
                       mesh->NElemsTriang5,
                       mesh->NodesTriang5,
                       mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5],
                       LoadedNodeNumbers,NodeStored,SubdomainIndex);
      CountLoadedNodes(mesh,SubMesh,E_NNPE_QUAD1,
                       mesh->NElemsQuad1,
                       mesh->NodesQuad1,
                       mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1],
                       LoadedNodeNumbers,NodeStored,SubdomainIndex);
      CountLoadedNodes(mesh,SubMesh,E_NNPE_QUAD3,
                       mesh->NElemsQuad3,
                       mesh->NodesQuad3,
                       mesh->FirstNetIndex[E_ELEM_TYPE_QUAD3],
                       LoadedNodeNumbers,NodeStored,SubdomainIndex);
      CountLoadedNodes(mesh,SubMesh,E_NNPE_TETRAH1,
                       mesh->NElemsTetrah1,
                       mesh->NodesTetrah1,
                       mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1],
                       LoadedNodeNumbers,NodeStored,SubdomainIndex);
      CountLoadedNodes(mesh,SubMesh,E_NNPE_BLOCK1,
                       mesh->NElemsBlock1,
                       mesh->NodesBlock1,
                       mesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1],
                       LoadedNodeNumbers,NodeStored,SubdomainIndex);

      e_free(LoadedNodeNumbers);
   }

   if (   (PropsIndices = e_alloc_i_vect(mesh->NMats)) == NULL
       || (CompPropsIndicesType1
               = e_alloc_i_vect(mesh->NCompMatsType1)) == NULL)
   {
      e_free(NodeStored);
      e_free(PropsIndices);
      e_free(CompPropsIndicesType1);
      return 0;
   }

   if (mesh->KeywordSet[E_KW_NUM_MATS_LINK1])
      CountMaterials(mesh,SubMesh,
                     mesh->NElemsLink1,
                     mesh->FirstNetIndex[E_ELEM_TYPE_LINK1],
                     PropsIndices,SubdomainIndex            );

   if (mesh->KeywordSet[E_KW_NUM_MATS_LINK2])
      CountMaterials(mesh,SubMesh,
                     mesh->NElemsLink2,
                     mesh->FirstNetIndex[E_ELEM_TYPE_LINK2],
                     PropsIndices,SubdomainIndex            );

   if (mesh->KeywordSet[E_KW_NUM_MATS_LINK3])
      CountMaterials(mesh,SubMesh,
                     mesh->NElemsLink3,
                     mesh->FirstNetIndex[E_ELEM_TYPE_LINK3],
                     PropsIndices,SubdomainIndex            );

   if (mesh->KeywordSet[E_KW_NUM_MATS_LINK4])
      CountMaterials(mesh,SubMesh,
                     mesh->NElemsLink4,
                     mesh->FirstNetIndex[E_ELEM_TYPE_LINK4],
                     PropsIndices,SubdomainIndex            );

   if (mesh->KeywordSet[E_KW_NUM_MATS_LINK5])
      CountMaterials(mesh,SubMesh,
                     mesh->NElemsLink5,
                     mesh->FirstNetIndex[E_ELEM_TYPE_LINK5],
                     PropsIndices,SubdomainIndex            );

   if (mesh->KeywordSet[E_KW_NUM_MATS_TRIANG1])
      CountMaterials(mesh,SubMesh,
                     mesh->NElemsTriang1,
                     mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1],
                     PropsIndices,SubdomainIndex              );

   if (mesh->KeywordSet[E_KW_NUM_COMP_MATS_TRIANG2])
      CountMatsType1(mesh,SubMesh,
                     mesh->NElemsTriang2,
                     mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2],
                     PropsIndices,CompPropsIndicesType1,SubdomainIndex);

   if (mesh->KeywordSet[E_KW_NUM_MATS_TRIANG3])
      CountMaterials(mesh,SubMesh,
                     mesh->NElemsTriang3,
                     mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3],
                     PropsIndices,SubdomainIndex              );

   if (mesh->KeywordSet[E_KW_NUM_MATS_TRIANG4])
      CountMaterials(mesh,SubMesh,
                     mesh->NElemsTriang4,
                     mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4],
                     PropsIndices,SubdomainIndex              );

   if (mesh->KeywordSet[E_KW_NUM_MATS_TRIANG5])
      CountMaterials(mesh,SubMesh,
                     mesh->NElemsTriang5,
                     mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5],
                     PropsIndices,SubdomainIndex              );

   if (mesh->KeywordSet[E_KW_NUM_MATS_QUAD1])
      CountMaterials(mesh,SubMesh,
                     mesh->NElemsQuad1,
                     mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1],
                     PropsIndices,SubdomainIndex            );

   if (mesh->KeywordSet[E_KW_NUM_COMP_MATS_QUAD3])
      CountMatsType1(mesh,SubMesh,
                     mesh->NElemsQuad3,
                     mesh->FirstNetIndex[E_ELEM_TYPE_QUAD3],
                     PropsIndices,CompPropsIndicesType1,SubdomainIndex);

   if (mesh->KeywordSet[E_KW_NUM_MATS_TETRAH1])
      CountMaterials(mesh,SubMesh,
                     mesh->NElemsTetrah1,
                     mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1],
                     PropsIndices,SubdomainIndex              );

   if (mesh->KeywordSet[E_KW_NUM_MATS_BLOCK1])
      CountMaterials(mesh,SubMesh,
                     mesh->NElemsBlock1,
                     mesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1],
                     PropsIndices,SubdomainIndex             );

   if (mesh->KeywordSet[E_KW_NUM_REMESH_DATA])
   {
      if (mesh->SubStructSet[E_SS_NUM_GEOM])
      {
         if (   (EndPointStored
                      = e_alloc_c_vect(mesh->Geom.NEndPoints)) == NULL
             || (NurbsCurveStored
                    = e_alloc_c_vect(mesh->Geom.NNurbsCurves)) == NULL)
         {
            e_free(NodeStored);
            e_free(PropsIndices);
            e_free(CompPropsIndicesType1);
            e_free(EndPointStored);
            e_free(NurbsCurveStored);
            return 0;
         }
      }
      else
      {
         if ((NurbsCurveStored
                 = e_alloc_c_vect(mesh->Geom.NNurbsCurves)) == NULL)
         {
            e_free(NodeStored);
            e_free(PropsIndices);
            e_free(CompPropsIndicesType1);
            return 0;
         }
      }

      CountNurbsCurves(mesh,SubMesh,E_NVPE_LINK1,mesh->NElemsLink1,
                       mesh->NodesLink1,
                       mesh->FirstNetIndex[E_ELEM_TYPE_LINK1],
                       EndPointStored,NurbsCurveStored,SubdomainIndex);

      CountNurbsCurves(mesh,SubMesh,E_NVPE_LINK2,mesh->NElemsLink2,
                       mesh->NodesLink2,
                       mesh->FirstNetIndex[E_ELEM_TYPE_LINK2],
                       EndPointStored,NurbsCurveStored,SubdomainIndex);

      CountNurbsCurves(mesh,SubMesh,E_NVPE_LINK3,mesh->NElemsLink3,
                       mesh->NodesLink3,
                       mesh->FirstNetIndex[E_ELEM_TYPE_LINK3],
                       EndPointStored,NurbsCurveStored,SubdomainIndex);

      CountNurbsCurves(mesh,SubMesh,E_NVPE_LINK4,mesh->NElemsLink4,
                       mesh->NodesLink4,
                       mesh->FirstNetIndex[E_ELEM_TYPE_LINK4],
                       EndPointStored,NurbsCurveStored,SubdomainIndex);

      CountNurbsCurves(mesh,SubMesh,E_NVPE_LINK5,mesh->NElemsLink5,
                       mesh->NodesLink5,
                       mesh->FirstNetIndex[E_ELEM_TYPE_LINK5],
                       EndPointStored,NurbsCurveStored,SubdomainIndex);

      CountNurbsCurves(mesh,SubMesh,E_NVPE_TRIANG1,mesh->NElemsTriang1,
                       mesh->NodesTriang1,
                       mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1],
                       EndPointStored,NurbsCurveStored,SubdomainIndex  );

      CountNurbsCurves(mesh,SubMesh,E_NVPE_TRIANG2,mesh->NElemsTriang2,
                       mesh->NodesTriang2,
                       mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2],
                       EndPointStored,NurbsCurveStored,SubdomainIndex  );

      CountNurbsCurves(mesh,SubMesh,E_NVPE_TRIANG3,mesh->NElemsTriang3,
                       mesh->NodesTriang3,
                       mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3],
                       EndPointStored,NurbsCurveStored,SubdomainIndex  );

      CountNurbsCurves(mesh,SubMesh,E_NVPE_TRIANG4,mesh->NElemsTriang4,
                       mesh->NodesTriang4,
                       mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4],
                       EndPointStored,NurbsCurveStored,SubdomainIndex  );

      CountNurbsCurves(mesh,SubMesh,E_NVPE_TRIANG5,mesh->NElemsTriang5,
                       mesh->NodesTriang5,
                       mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5],
                       EndPointStored,NurbsCurveStored,SubdomainIndex  );

      CountNurbsCurves(mesh,SubMesh,E_NVPE_QUAD1,mesh->NElemsQuad1,
                       mesh->NodesQuad1,
                       mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1],
                       EndPointStored,NurbsCurveStored,SubdomainIndex);

      CountNurbsCurves(mesh,SubMesh,E_NVPE_QUAD3,mesh->NElemsQuad3,
                       mesh->NodesQuad3,
                       mesh->FirstNetIndex[E_ELEM_TYPE_QUAD3],
                       EndPointStored,NurbsCurveStored,SubdomainIndex);

      CountNurbsCurves(mesh,SubMesh,E_NVPE_TETRAH1,mesh->NElemsTetrah1,
                       mesh->NodesTetrah1,
                       mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1],
                       EndPointStored,NurbsCurveStored,SubdomainIndex  );

      CountNurbsCurves(mesh,SubMesh,E_NVPE_BLOCK1,mesh->NElemsBlock1,
                       mesh->NodesBlock1,
                       mesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1],
                       EndPointStored,NurbsCurveStored,SubdomainIndex);

      e_free(EndPointStored);
      e_free(NurbsCurveStored);
   }

   e_free(NodeStored);
   e_free(PropsIndices);
   e_free(CompPropsIndicesType1);

   g_NMeshPoints = SubMesh->NMeshPoints;
   g_NNodes = SubMesh->NNodes;
   g_NElemsLink1 = SubMesh->NElemsLink1;
   g_NElemsLink2 = SubMesh->NElemsLink2;
   g_NElemsLink3 = SubMesh->NElemsLink3;
   g_NElemsLink4 = SubMesh->NElemsLink4;
   g_NElemsLink5 = SubMesh->NElemsLink5;
   g_NElemsTriang1 = SubMesh->NElemsTriang1;
   g_NElemsTriang2 = SubMesh->NElemsTriang2;
   g_NElemsTriang3 = SubMesh->NElemsTriang3;
   g_NElemsTriang4 = SubMesh->NElemsTriang4;
   g_NElemsTriang5 = SubMesh->NElemsTriang5;
   g_NElemsQuad1 = SubMesh->NElemsQuad1;
   g_NElemsQuad3 = SubMesh->NElemsQuad3;
   g_NElemsTetrah1 = SubMesh->NElemsTetrah1;
   g_NElemsBlock1 = SubMesh->NElemsBlock1;
   g_TotalNElems = SubMesh->TotalNElems;
   g_NBCNodes = SubMesh->NBCNodes;
   g_NLoadedNodes = SubMesh->NLoadedNodes;
   g_NMats = SubMesh->NMats;
   g_NMdfMats = SubMesh->NMdfMats;
   g_NCompMatsType1 = SubMesh->NCompMatsType1;
   g_TotalNodalDOF = SubMesh->TotalNodalDOF;

   g_NEndPoints = SubMesh->Geom.NEndPoints;
   g_NNurbsCurves = SubMesh->Geom.NNurbsCurves;

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* CountNodesAndElements: To count the number of sub-mesh nodes, mesh-points
   and elements for a particular element type.                                */


static void CountNodesAndElements
               (MESH *mesh, MESH *SubMesh, int NNodesPerElement,
                int NVerticesPerElement, int NMeshElements,
                int **MeshNodes, int FirstNetIndex,
                int *NSubMeshElements, char *NodeStored,
                int SubdomainIndex                              )
{
   int i,j,k;
   int MeshElementNetIndex;


   MeshElementNetIndex = FirstNetIndex - 1;

   for (i = 0; i < NMeshElements; i++)
   {
      if (mesh->Decomp.SubDoms[MeshElementNetIndex++] == SubdomainIndex)
      {
         (*NSubMeshElements)++;
         SubMesh->TotalNElems++;

         for (j = 0; j < NVerticesPerElement; j++)
         {
            k = MeshNodes[i][j] - 1;

            if (!NodeStored[k])
            {
               SubMesh->NNodes++;
               SubMesh->NMeshPoints++;
               NodeStored[k] = TRUE;
            }
         }

         for (j = NVerticesPerElement ; j < NNodesPerElement; j++)
         {
            k = MeshNodes[i][j] - 1;

            if (!NodeStored[k])
            {
               SubMesh->NNodes++;
               NodeStored[k] = TRUE;
            }
         }
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* CountBCNodes: To count the number of sub-mesh boundary condition nodes for
   a particular element type.                                                 */


static void CountBCNodes(MESH *mesh, MESH *SubMesh, int NNodesPerElement,
                         int NMeshElements, int **MeshNodes,
                         int FirstNetIndex, int *BCNumbers,
                         char *NodeStored, int SubdomainIndex            )
{
   int i,j,k;
   int MeshElementNetIndex;


   MeshElementNetIndex = FirstNetIndex - 1;

   for (i = 0; i < NMeshElements; i++)
   {
      if (mesh->Decomp.SubDoms[MeshElementNetIndex++] == SubdomainIndex)
      {
         for (j = 0; j < NNodesPerElement; j++)
         {
            k = MeshNodes[i][j] - 1;

            if (!NodeStored[k] && BCNumbers[k] > 0)
            {
               SubMesh->NBCNodes++;
               NodeStored[k] = TRUE;
            }
         }
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* CountLoadedNodes: To count the number of sub-mesh loaded nodes for a
   particular element type.                                                   */


static void CountLoadedNodes(MESH *mesh, MESH *SubMesh, int NNodesPerElement,
                             int NMeshElements, int **MeshNodes,
                             int FirstNetIndex, int *LoadedNodeNumbers,
                             char *NodeStored, int SubdomainIndex            )
{
   int i,j,k;
   int MeshElementNetIndex;


   MeshElementNetIndex = FirstNetIndex - 1;

   for (i = 0; i < NMeshElements; i++)
   {
      if (mesh->Decomp.SubDoms[MeshElementNetIndex++] == SubdomainIndex)
      {
         for (j = 0; j < NNodesPerElement; j++)
         {
            k = MeshNodes[i][j] - 1;

            if (!NodeStored[k] && LoadedNodeNumbers[k] > 0)
            {
                  SubMesh->NLoadedNodes++;
                  NodeStored[k] = TRUE;
            }
         }
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* CountMaterials: To count the number of sub-mesh materials for a particular
   element type.                                                              */


static void CountMaterials(MESH *mesh, MESH *SubMesh, int NMeshElements,
                           int FirstNetIndex, int *PropsIndices,
                           int SubdomainIndex                           )
{
   int i,r;
   int MeshElementNetIndex;


   MeshElementNetIndex = FirstNetIndex - 1;

   for (i = 0; i < NMeshElements; i++)
   {
      if (mesh->Decomp.SubDoms[MeshElementNetIndex] == SubdomainIndex)
      {
         r = (int)((MAT *)mesh->Mats[MeshElementNetIndex] - mesh->Props);

         if (GetPropsIndex(r,SubMesh->NMats,PropsIndices) == 0)
            PropsIndices[SubMesh->NMats++] = r;
      }

      MeshElementNetIndex++;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* CountMatsType1: To count the number of sub-mesh TYPE1 composite materials
   and that composite material's materials for a particular element type.     */


static void CountMatsType1(MESH *mesh, MESH *SubMesh, int NMeshElements,
                           int FirstNetIndex, int *PropsIndices,
                           int *CompPropsIndicesType1, int SubdomainIndex)
{
   int i,l,r,s;
   int MeshElementNetIndex;


   MeshElementNetIndex = FirstNetIndex - 1;

   for (i = 0; i < NMeshElements; i++)
   {
      if (mesh->Decomp.SubDoms[MeshElementNetIndex] == SubdomainIndex)
      {
         r = (int)((COMPMAT1 *)mesh->Mats[MeshElementNetIndex]
                                        - mesh->CompPropsType1);

         if (GetPropsIndex(r,SubMesh->NCompMatsType1,
                                CompPropsIndicesType1) == 0)
         {
            CompPropsIndicesType1[SubMesh->NCompMatsType1++] = r;

            for (l = 0; l < mesh->CompPropsType1[r].NLayers; l++)
            {
               s = (int)(mesh->CompPropsType1[r].LayerMats[l] - mesh->Props);

               if (GetPropsIndex(s,SubMesh->NMats,PropsIndices) == 0)
                  PropsIndices[SubMesh->NMats++] = s;
            }
         }
      }

      MeshElementNetIndex++;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* CountNurbsCurves: To count the number of NURBS curves and curve
   end-points.                                                                */


static void CountNurbsCurves(MESH *mesh, MESH *SubMesh,
                             int NVerticesPerElement, int NMeshElements,
                             int **MeshNodes, int FirstNetIndex,
                             char *EndPointStored, char *NurbsCurveStored,
                             int SubdomainIndex                           )
{
   int i,j,k,l,m,r;
   int MeshElementNetIndex;


   MeshElementNetIndex = FirstNetIndex - 1;

   if (mesh->SubStructSet[E_SS_NUM_GEOM])
   {
      for (i = 0; i < NMeshElements; i++)
      {
         if (mesh->Decomp.SubDoms[MeshElementNetIndex++] == SubdomainIndex)
         {
            for (j = 0; j < NVerticesPerElement; j++)
            {
               k = MeshNodes[i][j] - 1;

               if (mesh->NodeTypes[k] == E_RM_NUM_EXTERIOR)
               {
                  for (l = 0; l < mesh->NNurbsCurves[k]; l++)
                  {
                     r = (int)(mesh->NurbsCurves[k][l]
                               - mesh->Geom.NurbsCurves);

                     if (!NurbsCurveStored[r])
                     {
                        NurbsCurveStored[r] = TRUE;
                        SubMesh->Geom.NNurbsCurves++;

                        m = mesh->Geom.NurbsCurves[r].EndPoint1 - 1;

                        if (!EndPointStored[m])
                        {
                           EndPointStored[m] = TRUE;
                           SubMesh->Geom.NEndPoints++;
                        }

                        m = mesh->Geom.NurbsCurves[r].EndPoint2 - 1;

                        if (!EndPointStored[m])
                        {
                           EndPointStored[m] = TRUE;
                           SubMesh->Geom.NEndPoints++;
                        }
                     }
                  }
               }
            }
         }
      }
   }
   else
   {
      for (i = 0; i < NMeshElements; i++)
      {
         if (mesh->Decomp.SubDoms[MeshElementNetIndex++] == SubdomainIndex)
         {
            for (j = 0; j < NVerticesPerElement; j++)
            {
               k = MeshNodes[i][j] - 1;

               if (mesh->NodeTypes[k] == E_RM_NUM_EXTERIOR)
               {
                  for (l = 0; l < mesh->NNurbsCurves[k]; l++)
                  {
                     r = (int)(mesh->NurbsCurves[k][l]
                               - mesh->Geom.NurbsCurves);

                     if (!NurbsCurveStored[r])
                     {
                        NurbsCurveStored[r] = TRUE;
                        SubMesh->Geom.NNurbsCurves++;
                     }
                  }
               }
            }
         }
      }
   }
}



/* ////////////////////////////////////////////////////////////////////////// */
/* ZeroSubMesh: To set sub-mesh variables to zero. */


static void ZeroSubMesh(MESH *SubMesh)
{
   SubMesh->NMeshPoints = 0;
   SubMesh->NNodes = 0;
   SubMesh->NElemsLink1 = 0;
   SubMesh->NElemsLink2 = 0;
   SubMesh->NElemsLink3 = 0;
   SubMesh->NElemsLink4 = 0;
   SubMesh->NElemsLink5 = 0;
   SubMesh->NElemsTriang1 = 0;
   SubMesh->NElemsTriang2 = 0;
   SubMesh->NElemsTriang3 = 0;
   SubMesh->NElemsTriang4 = 0;
   SubMesh->NElemsTriang5 = 0;
   SubMesh->NElemsQuad1 = 0;
   SubMesh->NElemsQuad3 = 0;
   SubMesh->NElemsTetrah1 = 0;
   SubMesh->NElemsBlock1 = 0;
   SubMesh->TotalNElems = 0;
   SubMesh->NBCNodes = 0;
   SubMesh->NLoadedNodes = 0;
   SubMesh->NMats = 0;
   SubMesh->NMdfMats = 0;
   SubMesh->NCompMatsType1 = 0;
   SubMesh->TotalNodalDOF = 0;

   SubMesh->Geom.NEndPoints = 0;
   SubMesh->Geom.NNurbsCurves = 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* CreateMesh: To create the sub-mesh structure. */

/* The return value is 1 if the mesh is created successfully, else 0. */


static int CreateMesh(MESH *mesh, MESH *SubMesh, char *SubMeshTitle,
                                                  int SubdomainIndex)
{
   int k;
   int *BCNumbers;
   int *CompPropsIndicesType1;
   int *EndPoints;
   int *LoadedNodeNumbers;
   int *MeshToSubMeshPropsIndices;
   int *NewNodeIndex;
   int *NurbsCurveIndices;
   int *PropsIndices;

   char *NodeStored;


   NodeStored = NULL;
   NewNodeIndex = NULL;
   PropsIndices = NULL;
   CompPropsIndicesType1 = NULL;
   MeshToSubMeshPropsIndices = NULL;

   if (     (NodeStored = e_alloc_c_vect(mesh->NNodes)) == NULL
       || (NewNodeIndex = e_alloc_i_vect(mesh->NNodes)) == NULL)
   {
      e_free(NodeStored);
      e_free(NewNodeIndex);
      return 0;
   }

   SetMeshPointNodesAndElements(mesh,SubMesh,
                                E_NVPE_LINK1,
                                mesh->NElemsLink1,
                                mesh->NodesLink1,
                                mesh->FirstNetIndex[E_ELEM_TYPE_LINK1],
                                &SubMesh->NElemsLink1,
                                SubMesh->NodesLink1,
                                NewNodeIndex,NodeStored,SubdomainIndex );
   SetMeshPointNodesAndElements(mesh,SubMesh,
                                E_NVPE_LINK2,
                                mesh->NElemsLink2,
                                mesh->NodesLink2,
                                mesh->FirstNetIndex[E_ELEM_TYPE_LINK2],
                                &SubMesh->NElemsLink2,
                                SubMesh->NodesLink2,
                                NewNodeIndex,NodeStored,SubdomainIndex );
   SetMeshPointNodesAndElements(mesh,SubMesh,
                                E_NVPE_LINK3,
                                mesh->NElemsLink3,
                                mesh->NodesLink3,
                                mesh->FirstNetIndex[E_ELEM_TYPE_LINK3],
                                &SubMesh->NElemsLink3,
                                SubMesh->NodesLink3,
                                NewNodeIndex,NodeStored,SubdomainIndex );
   SetMeshPointNodesAndElements(mesh,SubMesh,
                                E_NVPE_LINK4,
                                mesh->NElemsLink4,
                                mesh->NodesLink4,
                                mesh->FirstNetIndex[E_ELEM_TYPE_LINK4],
                                &SubMesh->NElemsLink4,
                                SubMesh->NodesLink4,
                                NewNodeIndex,NodeStored,SubdomainIndex );
   SetMeshPointNodesAndElements(mesh,SubMesh,
                                E_NVPE_LINK5,
                                mesh->NElemsLink5,
                                mesh->NodesLink5,
                                mesh->FirstNetIndex[E_ELEM_TYPE_LINK5],
                                &SubMesh->NElemsLink5,
                                SubMesh->NodesLink5,
                                NewNodeIndex,NodeStored,SubdomainIndex );
   SetMeshPointNodesAndElements(mesh,SubMesh,
                                E_NVPE_TRIANG1,
                                mesh->NElemsTriang1,
                                mesh->NodesTriang1,
                                mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1],
                                &SubMesh->NElemsTriang1,
                                SubMesh->NodesTriang1,
                                NewNodeIndex,NodeStored,SubdomainIndex   );
   SetMeshPointNodesAndElements(mesh,SubMesh,
                                E_NVPE_TRIANG2,
                                mesh->NElemsTriang2,
                                mesh->NodesTriang2,
                                mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2],
                                &SubMesh->NElemsTriang2,
                                SubMesh->NodesTriang2,
                                NewNodeIndex,NodeStored,SubdomainIndex   );
   SetMeshPointNodesAndElements(mesh,SubMesh,
                                E_NVPE_TRIANG3,
                                mesh->NElemsTriang3,
                                mesh->NodesTriang3,
                                mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3],
                                &SubMesh->NElemsTriang3,
                                SubMesh->NodesTriang3,
                                NewNodeIndex,NodeStored,SubdomainIndex   );
   SetMeshPointNodesAndElements(mesh,SubMesh,
                                E_NVPE_TRIANG4,
                                mesh->NElemsTriang4,
                                mesh->NodesTriang4,
                                mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4],
                                &SubMesh->NElemsTriang4,
                                SubMesh->NodesTriang4,
                                NewNodeIndex,NodeStored,SubdomainIndex   );
   SetMeshPointNodesAndElements(mesh,SubMesh,
                                E_NVPE_TRIANG5,
                                mesh->NElemsTriang5,
                                mesh->NodesTriang5,
                                mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5],
                                &SubMesh->NElemsTriang5,
                                SubMesh->NodesTriang5,
                                NewNodeIndex,NodeStored,SubdomainIndex   );
   SetMeshPointNodesAndElements(mesh,SubMesh,
                                E_NVPE_QUAD1,
                                mesh->NElemsQuad1,
                                mesh->NodesQuad1,
                                mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1],
                                &SubMesh->NElemsQuad1,
                                SubMesh->NodesQuad1,
                                NewNodeIndex,NodeStored,SubdomainIndex);
   SetMeshPointNodesAndElements(mesh,SubMesh,
                                E_NVPE_QUAD3,
                                mesh->NElemsQuad3,
                                mesh->NodesQuad3,
                                mesh->FirstNetIndex[E_ELEM_TYPE_QUAD3],
                                &SubMesh->NElemsQuad3,
                                SubMesh->NodesQuad3,
                                NewNodeIndex,NodeStored,SubdomainIndex);
   SetMeshPointNodesAndElements(mesh,SubMesh,
                                E_NVPE_TETRAH1,
                                mesh->NElemsTetrah1,
                                mesh->NodesTetrah1,
                                mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1],
                                &SubMesh->NElemsTetrah1,
                                SubMesh->NodesTetrah1,
                                NewNodeIndex,NodeStored,SubdomainIndex   );
   SetMeshPointNodesAndElements(mesh,SubMesh,
                                E_NVPE_BLOCK1,
                                mesh->NElemsBlock1,
                                mesh->NodesBlock1,
                                mesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1],
                                &SubMesh->NElemsBlock1,
                                SubMesh->NodesBlock1,
                                NewNodeIndex,NodeStored,SubdomainIndex  );

   SetNonMeshPointNodes(mesh,SubMesh,
                        E_NNPE_TRIANG2,E_NVPE_TRIANG2,
                        mesh->NElemsTriang2,
                        mesh->NodesTriang2,
                        mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2],
                        SubMesh->NodesTriang2,
                        NewNodeIndex,NodeStored,SubdomainIndex   );

   SetNonMeshPointNodes(mesh,SubMesh,
                        E_NNPE_QUAD3,E_NVPE_QUAD3,
                        mesh->NElemsQuad3,
                        mesh->NodesQuad3,
                        mesh->FirstNetIndex[E_ELEM_TYPE_QUAD3],
                        SubMesh->NodesQuad3,
                        NewNodeIndex,NodeStored,SubdomainIndex   );



   /* "FirstNetIndex" and mesh parameters are set
      here so that "NewNodeIndex" can be freed    */

   SetFirstNetIndices(SubMesh);

   if (mesh->SubStructSet[E_SS_NUM_MESHPARAM])
      SetMeshParams(mesh,SubMesh,NewNodeIndex,SubdomainIndex);

   e_free(NewNodeIndex);

   if (mesh->KeywordSet[E_KW_NUM_BCS])
   {
      if ((BCNumbers = e_alloc_i_vect(mesh->NNodes)) == NULL)
      {
         e_free(NodeStored);
         return 0;
      }

      for (k = 0; k < mesh->NNodes; k++)
         NodeStored[k] = FALSE;

      for (k = 0; k < mesh->NBCNodes; k++)
         BCNumbers[mesh->BCNodes[k] - 1] = k + 1;

      SubMesh->NNodes = 0;

      SetMeshPointBCs(mesh,SubMesh,
                      E_NVPE_LINK1,
                      mesh->NElemsLink1,
                      mesh->NodesLink1,
                      mesh->FirstNetIndex[E_ELEM_TYPE_LINK1],
                      BCNumbers,NodeStored,SubdomainIndex    );
      SetMeshPointBCs(mesh,SubMesh,
                      E_NVPE_LINK2,
                      mesh->NElemsLink2,
                      mesh->NodesLink2,
                      mesh->FirstNetIndex[E_ELEM_TYPE_LINK2],
                      BCNumbers,NodeStored,SubdomainIndex    );
      SetMeshPointBCs(mesh,SubMesh,
                      E_NVPE_LINK3,
                      mesh->NElemsLink3,
                      mesh->NodesLink3,
                      mesh->FirstNetIndex[E_ELEM_TYPE_LINK3],
                      BCNumbers,NodeStored,SubdomainIndex    );
      SetMeshPointBCs(mesh,SubMesh,
                      E_NVPE_LINK4,
                      mesh->NElemsLink4,
                      mesh->NodesLink4,
                      mesh->FirstNetIndex[E_ELEM_TYPE_LINK4],
                      BCNumbers,NodeStored,SubdomainIndex    );
      SetMeshPointBCs(mesh,SubMesh,
                      E_NVPE_LINK5,
                      mesh->NElemsLink5,
                      mesh->NodesLink5,
                      mesh->FirstNetIndex[E_ELEM_TYPE_LINK5],
                      BCNumbers,NodeStored,SubdomainIndex    );
      SetMeshPointBCs(mesh,SubMesh,
                      E_NVPE_TRIANG1,
                      mesh->NElemsTriang1,
                      mesh->NodesTriang1,
                      mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1],
                      BCNumbers,NodeStored,SubdomainIndex      );
      SetMeshPointBCs(mesh,SubMesh,
                      E_NVPE_TRIANG2,
                      mesh->NElemsTriang2,
                      mesh->NodesTriang2,
                      mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2],
                      BCNumbers,NodeStored,SubdomainIndex      );
      SetMeshPointBCs(mesh,SubMesh,
                      E_NVPE_TRIANG3,
                      mesh->NElemsTriang3,
                      mesh->NodesTriang3,
                      mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3],
                      BCNumbers,NodeStored,SubdomainIndex      );
      SetMeshPointBCs(mesh,SubMesh,
                      E_NVPE_TRIANG4,
                      mesh->NElemsTriang4,
                      mesh->NodesTriang4,
                      mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4],
                      BCNumbers,NodeStored,SubdomainIndex      );
      SetMeshPointBCs(mesh,SubMesh,
                      E_NVPE_TRIANG5,
                      mesh->NElemsTriang5,
                      mesh->NodesTriang5,
                      mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5],
                      BCNumbers,NodeStored,SubdomainIndex      );
      SetMeshPointBCs(mesh,SubMesh,
                      E_NVPE_QUAD1,
                      mesh->NElemsQuad1,
                      mesh->NodesQuad1,
                      mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1],
                      BCNumbers,NodeStored,SubdomainIndex    );
      SetMeshPointBCs(mesh,SubMesh,
                      E_NVPE_QUAD3,
                      mesh->NElemsQuad3,
                      mesh->NodesQuad3,
                      mesh->FirstNetIndex[E_ELEM_TYPE_QUAD3],
                      BCNumbers,NodeStored,SubdomainIndex    );
      SetMeshPointBCs(mesh,SubMesh,
                      E_NVPE_TETRAH1,
                      mesh->NElemsTetrah1,
                      mesh->NodesTetrah1,
                      mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1],
                      BCNumbers,NodeStored,SubdomainIndex      );
      SetMeshPointBCs(mesh,SubMesh,
                      E_NVPE_BLOCK1,
                      mesh->NElemsBlock1,
                      mesh->NodesBlock1,
                      mesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1],
                      BCNumbers,NodeStored,SubdomainIndex     );
      SetNonMeshPointBCs(mesh,SubMesh,
                         E_NNPE_TRIANG2,E_NVPE_TRIANG2,
                         mesh->NElemsTriang2,
                         mesh->NodesTriang2,
                         mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2],
                         BCNumbers,NodeStored,SubdomainIndex      );
      SetNonMeshPointBCs(mesh,SubMesh,
                         E_NNPE_QUAD3,E_NVPE_QUAD3,
                         mesh->NElemsQuad3,
                         mesh->NodesQuad3,
                         mesh->FirstNetIndex[E_ELEM_TYPE_QUAD3],
                         BCNumbers,NodeStored,SubdomainIndex      );

      e_free(BCNumbers);
   }

   if (mesh->KeywordSet[E_KW_NUM_LOADS])
   {
      if ((LoadedNodeNumbers = e_alloc_i_vect(mesh->NNodes)) == NULL)
      {
         e_free(NodeStored);
         return 0;
      }

      for (k = 0; k < mesh->NNodes; k++)
         NodeStored[k] = FALSE;

      for (k = 0; k < mesh->NLoadedNodes; k++)
         LoadedNodeNumbers[mesh->LoadedNodes[k] - 1] = k + 1;

      SubMesh->NNodes = 0;

      SetMeshPointLoads(mesh,SubMesh,
                        E_NVPE_LINK1,
                        mesh->NElemsLink1,
                        mesh->NodesLink1,
                        mesh->FirstNetIndex[E_ELEM_TYPE_LINK1],
                        LoadedNodeNumbers,NodeStored,SubdomainIndex);
      SetMeshPointLoads(mesh,SubMesh,
                        E_NVPE_LINK2,
                        mesh->NElemsLink2,
                        mesh->NodesLink2,
                        mesh->FirstNetIndex[E_ELEM_TYPE_LINK2],
                        LoadedNodeNumbers,NodeStored,SubdomainIndex);
      SetMeshPointLoads(mesh,SubMesh,
                        E_NVPE_LINK3,
                        mesh->NElemsLink3,
                        mesh->NodesLink3,
                        mesh->FirstNetIndex[E_ELEM_TYPE_LINK3],
                        LoadedNodeNumbers,NodeStored,SubdomainIndex);
      SetMeshPointLoads(mesh,SubMesh,
                        E_NVPE_LINK4,
                        mesh->NElemsLink4,
                        mesh->NodesLink4,
                        mesh->FirstNetIndex[E_ELEM_TYPE_LINK4],
                        LoadedNodeNumbers,NodeStored,SubdomainIndex);
      SetMeshPointLoads(mesh,SubMesh,
                        E_NVPE_LINK5,
                        mesh->NElemsLink5,
                        mesh->NodesLink5,
                        mesh->FirstNetIndex[E_ELEM_TYPE_LINK5],
                        LoadedNodeNumbers,NodeStored,SubdomainIndex);
      SetMeshPointLoads(mesh,SubMesh,
                        E_NVPE_TRIANG1,
                        mesh->NElemsTriang1,
                        mesh->NodesTriang1,
                        mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1],
                        LoadedNodeNumbers,NodeStored,SubdomainIndex);
      SetMeshPointLoads(mesh,SubMesh,
                        E_NVPE_TRIANG2,
                        mesh->NElemsTriang2,
                        mesh->NodesTriang2,
                        mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2],
                        LoadedNodeNumbers,NodeStored,SubdomainIndex);
      SetMeshPointLoads(mesh,SubMesh,
                        E_NVPE_TRIANG3,
                        mesh->NElemsTriang3,
                        mesh->NodesTriang3,
                        mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3],
                        LoadedNodeNumbers,NodeStored,SubdomainIndex);
      SetMeshPointLoads(mesh,SubMesh,
                        E_NVPE_TRIANG4,
                        mesh->NElemsTriang4,
                        mesh->NodesTriang4,
                        mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4],
                        LoadedNodeNumbers,NodeStored,SubdomainIndex);
      SetMeshPointLoads(mesh,SubMesh,
                        E_NVPE_TRIANG5,
                        mesh->NElemsTriang5,
                        mesh->NodesTriang5,
                        mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5],
                        LoadedNodeNumbers,NodeStored,SubdomainIndex);
      SetMeshPointLoads(mesh,SubMesh,
                        E_NVPE_QUAD1,
                        mesh->NElemsQuad1,
                        mesh->NodesQuad1,
                        mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1],
                        LoadedNodeNumbers,NodeStored,SubdomainIndex);
      SetMeshPointLoads(mesh,SubMesh,
                        E_NVPE_QUAD3,
                        mesh->NElemsQuad3,
                        mesh->NodesQuad3,
                        mesh->FirstNetIndex[E_ELEM_TYPE_QUAD3],
                        LoadedNodeNumbers,NodeStored,SubdomainIndex);
      SetMeshPointLoads(mesh,SubMesh,
                        E_NVPE_TETRAH1,
                        mesh->NElemsTetrah1,
                        mesh->NodesTetrah1,
                        mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1],
                        LoadedNodeNumbers,NodeStored,SubdomainIndex);
      SetMeshPointLoads(mesh,SubMesh,
                        E_NVPE_BLOCK1,
                        mesh->NElemsBlock1,
                        mesh->NodesBlock1,
                        mesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1],
                        LoadedNodeNumbers,NodeStored,SubdomainIndex);
      SetNonMeshPointLoads(mesh,SubMesh,
                           E_NNPE_TRIANG2,E_NVPE_TRIANG2,
                           mesh->NElemsTriang2,
                           mesh->NodesTriang2,
                           mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2],
                           LoadedNodeNumbers,NodeStored,SubdomainIndex);
      SetNonMeshPointLoads(mesh,SubMesh,
                           E_NNPE_QUAD3,E_NVPE_QUAD3,
                           mesh->NElemsQuad3,
                           mesh->NodesQuad3,
                           mesh->FirstNetIndex[E_ELEM_TYPE_QUAD3],
                           LoadedNodeNumbers,NodeStored,SubdomainIndex);

      e_free(LoadedNodeNumbers);
   }

   e_free(NodeStored);

   if (                (PropsIndices = e_alloc_i_vect(mesh->NMats)) == NULL
       || (CompPropsIndicesType1
                            = e_alloc_i_vect(mesh->NCompMatsType1)) == NULL
       || (MeshToSubMeshPropsIndices = e_alloc_i_vect(mesh->NMats)) == NULL)
   {
      e_free(PropsIndices);
      e_free(CompPropsIndicesType1);
      e_free(MeshToSubMeshPropsIndices);
      return 0;
   }

   if (   mesh->KeywordSet[E_KW_NUM_MATS_LINK1]
       || mesh->KeywordSet[E_KW_NUM_MATS_LINK1]
       || mesh->KeywordSet[E_KW_NUM_MATS_LINK3]
       || mesh->KeywordSet[E_KW_NUM_MATS_LINK4]
       || mesh->KeywordSet[E_KW_NUM_MATS_LINK5]
       || mesh->KeywordSet[E_KW_NUM_MATS_TRIANG1]
       || mesh->KeywordSet[E_KW_NUM_COMP_MATS_TRIANG2]
       || mesh->KeywordSet[E_KW_NUM_MATS_TRIANG3]
       || mesh->KeywordSet[E_KW_NUM_MATS_TRIANG4]
       || mesh->KeywordSet[E_KW_NUM_MATS_TRIANG5]
       || mesh->KeywordSet[E_KW_NUM_MATS_QUAD1]
       || mesh->KeywordSet[E_KW_NUM_COMP_MATS_QUAD3]
       || mesh->KeywordSet[E_KW_NUM_MATS_TETRAH1]
       || mesh->KeywordSet[E_KW_NUM_MATS_BLOCK1]      )
   {
      SubMesh->TotalNElems = 0;
      SubMesh->NMats = 0;
      SubMesh->NCompMatsType1 = 0;
   }

   if (   (   mesh->KeywordSet[E_KW_NUM_MATS_LINK1]
           && !SetMaterials(mesh,SubMesh,mesh->NElemsLink1,
                            mesh->FirstNetIndex[E_ELEM_TYPE_LINK1],
                            PropsIndices,MeshToSubMeshPropsIndices,
                            SubdomainIndex                         ))

       || (   mesh->KeywordSet[E_KW_NUM_MATS_LINK2]
           && !SetMaterials(mesh,SubMesh,mesh->NElemsLink2,
                            mesh->FirstNetIndex[E_ELEM_TYPE_LINK2],
                            PropsIndices,MeshToSubMeshPropsIndices,
                            SubdomainIndex                         ))

       || (   mesh->KeywordSet[E_KW_NUM_MATS_LINK3]
           && !SetMaterials(mesh,SubMesh,mesh->NElemsLink3,
                            mesh->FirstNetIndex[E_ELEM_TYPE_LINK3],
                            PropsIndices,MeshToSubMeshPropsIndices,
                            SubdomainIndex                         ))

       || (   mesh->KeywordSet[E_KW_NUM_MATS_LINK4]
           && !SetMaterials(mesh,SubMesh,mesh->NElemsLink4,
                            mesh->FirstNetIndex[E_ELEM_TYPE_LINK4],
                            PropsIndices,MeshToSubMeshPropsIndices,
                            SubdomainIndex                         ))

       || (   mesh->KeywordSet[E_KW_NUM_MATS_LINK5]
           && !SetMaterials(mesh,SubMesh,mesh->NElemsLink5,
                            mesh->FirstNetIndex[E_ELEM_TYPE_LINK5],
                            PropsIndices,MeshToSubMeshPropsIndices,
                            SubdomainIndex                         ))

       || (   mesh->KeywordSet[E_KW_NUM_MATS_TRIANG1]
           && !SetMaterials(mesh,SubMesh,mesh->NElemsTriang1,
                            mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1],
                            PropsIndices,MeshToSubMeshPropsIndices,
                            SubdomainIndex                           ))

       || (   mesh->KeywordSet[E_KW_NUM_COMP_MATS_TRIANG2]
           && !SetMatsType1(mesh,SubMesh,mesh->NElemsTriang2,
                            mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2],
                            PropsIndices,CompPropsIndicesType1,
                            MeshToSubMeshPropsIndices,SubdomainIndex ))

       || (   mesh->KeywordSet[E_KW_NUM_MATS_TRIANG3]
           && !SetMaterials(mesh,SubMesh,mesh->NElemsTriang3,
                            mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3],
                            PropsIndices,MeshToSubMeshPropsIndices,
                            SubdomainIndex                           ))

       || (   mesh->KeywordSet[E_KW_NUM_MATS_TRIANG4]
           && !SetMaterials(mesh,SubMesh,mesh->NElemsTriang4,
                            mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4],
                            PropsIndices,MeshToSubMeshPropsIndices,
                            SubdomainIndex                           ))

       || (   mesh->KeywordSet[E_KW_NUM_MATS_TRIANG5]
           && !SetMaterials(mesh,SubMesh,mesh->NElemsTriang5,
                            mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5],
                            PropsIndices,MeshToSubMeshPropsIndices,
                            SubdomainIndex                           ))

       || (   mesh->KeywordSet[E_KW_NUM_MATS_QUAD1]
           && !SetMaterials(mesh,SubMesh,mesh->NElemsQuad1,
                            mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1],
                            PropsIndices,MeshToSubMeshPropsIndices,
                            SubdomainIndex                         ))

       || (   mesh->KeywordSet[E_KW_NUM_COMP_MATS_QUAD3]
           && !SetMatsType1(mesh,SubMesh,mesh->NElemsQuad3,
                            mesh->FirstNetIndex[E_ELEM_TYPE_QUAD3],
                            PropsIndices,CompPropsIndicesType1,
                            MeshToSubMeshPropsIndices,SubdomainIndex ))

       || (   mesh->KeywordSet[E_KW_NUM_MATS_TETRAH1]
           && !SetMaterials(mesh,SubMesh,mesh->NElemsTetrah1,
                            mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1],
                            PropsIndices,MeshToSubMeshPropsIndices,
                            SubdomainIndex                           ))

       || (   mesh->KeywordSet[E_KW_NUM_MATS_BLOCK1]
           && !SetMaterials(mesh,SubMesh,mesh->NElemsBlock1,
                            mesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1],
                            PropsIndices,MeshToSubMeshPropsIndices,
                            SubdomainIndex                          ))

       || (   mesh->SubStructSet[E_SS_NUM_MAT_PROPS]
           && !SetMaterialProperties(mesh,SubMesh,PropsIndices,
                                    CompPropsIndicesType1,
                                    MeshToSubMeshPropsIndices ))       )
   {
      e_free(PropsIndices);
      e_free(CompPropsIndicesType1);
      e_free(MeshToSubMeshPropsIndices);
      return 0;
   }
   else
   {
      e_free(PropsIndices);
      e_free(CompPropsIndicesType1);
      e_free(MeshToSubMeshPropsIndices);
   }

   /* The following will also set GEOM data if appropriate */

   if (mesh->KeywordSet[E_KW_NUM_REMESH_DATA])
   {
      NodeStored = NULL;
      EndPoints = NULL;
      NurbsCurveIndices = NULL;

      if (   (NodeStored = e_alloc_c_vect(mesh->NMeshPoints)) == NULL

          || (   mesh->SubStructSet[E_SS_NUM_GEOM]
              && (EndPoints
                     = e_alloc_i_vect(mesh->Geom.NEndPoints)) == NULL)

          || (NurbsCurveIndices
                   = e_alloc_i_vect(mesh->Geom.NNurbsCurves)) == NULL)
      {
         e_free(NodeStored);
         e_free(EndPoints);
         e_free(NurbsCurveIndices);
         return 0;
      }

      SubMesh->NMeshPoints = 0;
      SubMesh->Geom.NNurbsCurves = 0;
      SubMesh->Geom.NEndPoints = 0;

      if (   !SetNurbsCurves(mesh,SubMesh,E_NVPE_LINK1,
                             mesh->NElemsLink1,mesh->NodesLink1,
                             mesh->FirstNetIndex[E_ELEM_TYPE_LINK1],
                             EndPoints,NurbsCurveIndices,NodeStored,
                             SubdomainIndex                         )
          || !SetNurbsCurves(mesh,SubMesh,E_NVPE_LINK2,
                             mesh->NElemsLink2,mesh->NodesLink2,
                             mesh->FirstNetIndex[E_ELEM_TYPE_LINK2],
                             EndPoints,NurbsCurveIndices,NodeStored,
                             SubdomainIndex                         )
          || !SetNurbsCurves(mesh,SubMesh,E_NVPE_LINK3,
                             mesh->NElemsLink3,mesh->NodesLink3,
                             mesh->FirstNetIndex[E_ELEM_TYPE_LINK3],
                             EndPoints,NurbsCurveIndices,NodeStored,
                             SubdomainIndex                         )
          || !SetNurbsCurves(mesh,SubMesh,E_NVPE_LINK4,
                             mesh->NElemsLink4,mesh->NodesLink4,
                             mesh->FirstNetIndex[E_ELEM_TYPE_LINK4],
                             EndPoints,NurbsCurveIndices,NodeStored,
                             SubdomainIndex                         )
          || !SetNurbsCurves(mesh,SubMesh,E_NVPE_LINK5,
                             mesh->NElemsLink5,mesh->NodesLink5,
                             mesh->FirstNetIndex[E_ELEM_TYPE_LINK5],
                             EndPoints,NurbsCurveIndices,NodeStored,
                             SubdomainIndex                         )
          || !SetNurbsCurves(mesh,SubMesh,E_NVPE_TRIANG1,
                             mesh->NElemsTriang1,mesh->NodesTriang1,
                             mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1],
                             EndPoints,NurbsCurveIndices,NodeStored,
                             SubdomainIndex                           )
          || !SetNurbsCurves(mesh,SubMesh,E_NVPE_TRIANG2,
                             mesh->NElemsTriang2,mesh->NodesTriang2,
                             mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2],
                             EndPoints,NurbsCurveIndices,NodeStored,
                             SubdomainIndex                           )
          || !SetNurbsCurves(mesh,SubMesh,E_NVPE_TRIANG3,
                             mesh->NElemsTriang3,mesh->NodesTriang3,
                             mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3],
                             EndPoints,NurbsCurveIndices,NodeStored,
                             SubdomainIndex                           )
          || !SetNurbsCurves(mesh,SubMesh,E_NVPE_TRIANG4,
                             mesh->NElemsTriang4,mesh->NodesTriang4,
                             mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4],
                             EndPoints,NurbsCurveIndices,NodeStored,
                             SubdomainIndex                           )
          || !SetNurbsCurves(mesh,SubMesh,E_NVPE_TRIANG5,
                             mesh->NElemsTriang5,mesh->NodesTriang5,
                             mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5],
                             EndPoints,NurbsCurveIndices,NodeStored,
                             SubdomainIndex                           )
          || !SetNurbsCurves(mesh,SubMesh,E_NVPE_QUAD1,
                             mesh->NElemsQuad1,mesh->NodesQuad1,
                             mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1],
                             EndPoints,NurbsCurveIndices,NodeStored,
                             SubdomainIndex                         )
          || !SetNurbsCurves(mesh,SubMesh,E_NVPE_QUAD3,
                             mesh->NElemsQuad3,mesh->NodesQuad3,
                             mesh->FirstNetIndex[E_ELEM_TYPE_QUAD3],
                             EndPoints,NurbsCurveIndices,NodeStored,
                             SubdomainIndex                         )
          || !SetNurbsCurves(mesh,SubMesh,E_NVPE_TETRAH1,
                             mesh->NElemsTetrah1,mesh->NodesTetrah1,
                             mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1],
                             EndPoints,NurbsCurveIndices,NodeStored,
                             SubdomainIndex                           )
          || !SetNurbsCurves(mesh,SubMesh,E_NVPE_BLOCK1,
                             mesh->NElemsBlock1,mesh->NodesBlock1,
                             mesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1],
                             EndPoints,NurbsCurveIndices,NodeStored,
                             SubdomainIndex                          ) )
      {
         e_free(NodeStored);
         e_free(EndPoints);
         e_free(NurbsCurveIndices);
         return 0;
      }
      else
      {
         e_free(NodeStored);
         e_free(EndPoints);
         e_free(NurbsCurveIndices);
      }
   }


   SetNodalDOF(SubMesh);
   SetElementMaterials(SubMesh);
   SetGlobalElementIndices(SubMesh);

   /* Mesh parameters are above */

   if (mesh->SubStructSet[E_SS_NUM_FEERROR])
      SetFEErrors(mesh,SubMesh,SubdomainIndex);

   /* this should be executed before the stresses are set */
   if (!SetRemainingData(mesh,SubMesh,SubMeshTitle))
     return 0;

   if (mesh->SubStructSet[E_SS_NUM_STRESS])
   {
     if(!e_SetupStresses(SubMesh))
       return 0;
     SetStresses(mesh,SubMesh, SubdomainIndex);
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetMeshPointNodesAndElements: To set the sub-mesh mesh-point node
   variables and element variables for a particular element type.             */


static void SetMeshPointNodesAndElements
               (MESH *mesh, MESH *SubMesh, int NVerticesPerElement,
                int NMeshElements, int **MeshNodes, int FirstNetIndex,
                int *NSubMeshElements, int **SubMeshNodes, int *NewNodeIndex,
                char *NodeStored, int SubdomainIndex                         )
{
   int i,j,k,l,m,n;
   int MeshElementNetIndex;


   MeshElementNetIndex = FirstNetIndex - 1;

   for (i = 0; i < NMeshElements; i++)
   {
      if (mesh->Decomp.SubDoms[MeshElementNetIndex] == SubdomainIndex)
      {
         n = (*NSubMeshElements)++;
         SubMesh->GEIs[SubMesh->TotalNElems++]
            = mesh->GEIs[MeshElementNetIndex];

         for (j = 0; j < NVerticesPerElement; j++)
         {
            k = MeshNodes[i][j] - 1;

            if (!NodeStored[k])
            {
               l = SubMesh->NNodes++;
               SubMesh->NMeshPoints++;

               for (m = 0; m < E_NDIMS; m++)
                  SubMesh->MeshPointCoords[l][m]
                     = mesh->MeshPointCoords[k][m];

               SubMesh->GNIs[l] = mesh->GNIs[k];
               NewNodeIndex[k] = SubMesh->NNodes;
               NodeStored[k] = TRUE;
            }

            SubMeshNodes[n][j] = NewNodeIndex[k];
         }
      }

      MeshElementNetIndex++;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetNonMeshPointNodes: To set the sub-mesh non-mesh-point node variables
   for a particular element type.                                             */


static void SetNonMeshPointNodes
               (MESH *mesh, MESH *SubMesh, int NNodesPerElement,
                int NVerticesPerElement, int NMeshElements, int **MeshNodes,
                int FirstNetIndex, int **SubMeshNodes, int *NewNodeIndex,
                char *NodeStored, int SubdomainIndex                        )
{
   int i,j,k,l,n;
   int MeshElementNetIndex;
   int NSubMeshElements;


   MeshElementNetIndex = FirstNetIndex - 1;
   NSubMeshElements = 0;

   for (i = 0; i < NMeshElements; i++)
   {
      if (mesh->Decomp.SubDoms[MeshElementNetIndex++] == SubdomainIndex)
      {
         n = NSubMeshElements++;

         for (j = NVerticesPerElement ; j < NNodesPerElement; j++)
         {
            k = MeshNodes[i][j] - 1;

            if (!NodeStored[k])
            {
               l = SubMesh->NNodes++;
               SubMesh->GNIs[l] = mesh->GNIs[k];
               NewNodeIndex[k] = SubMesh->NNodes;
               NodeStored[k] = TRUE;
            }

            SubMeshNodes[n][j] = NewNodeIndex[k];
         }
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetFirstNetIndices: To set the sub-mesh first net indices vector.   */


static void SetFirstNetIndices(MESH *SubMesh)
{
   int i;
   int sum;


   for (i = 0; i < E_NELEM_TYPES; i++)
      SubMesh->FirstNetIndex[i] = 0;

   sum = 0;

   if (SubMesh->NElemsLink1 > 0)
   {
      SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK1] = sum + 1;
      sum += SubMesh->NElemsLink1;
   }

   if (SubMesh->NElemsLink2 > 0)
   {
      SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK2] = sum + 1;
      sum += SubMesh->NElemsLink2;
   }

   if (SubMesh->NElemsLink3 > 0)
   {
      SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK3] = sum + 1;
      sum += SubMesh->NElemsLink3;
   }

   if (SubMesh->NElemsLink4 > 0)
   {
      SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK4] = sum + 1;
      sum += SubMesh->NElemsLink4;
   }

   if (SubMesh->NElemsLink5 > 0)
   {
      SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK5] = sum + 1;
      sum += SubMesh->NElemsLink5;
   }

   if (SubMesh->NElemsTriang1 > 0)
   {
      SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] = sum + 1;
      sum += SubMesh->NElemsTriang1;
   }

   if (SubMesh->NElemsTriang2 > 0)
   {
      SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2] = sum + 1;
      sum += SubMesh->NElemsTriang2;
   }

   if (SubMesh->NElemsTriang3 > 0)
   {
      SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] = sum + 1;
      sum += SubMesh->NElemsTriang3;
   }

   if (SubMesh->NElemsTriang4 > 0)
   {
      SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] = sum + 1;
      sum += SubMesh->NElemsTriang4;
   }

   if (SubMesh->NElemsTriang5 > 0)
   {
      SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] = sum + 1;
      sum += SubMesh->NElemsTriang5;
   }

   if (SubMesh->NElemsQuad1 > 0)
   {
      SubMesh->FirstNetIndex[E_ELEM_TYPE_QUAD1] = sum + 1;
      sum += SubMesh->NElemsQuad1;
   }

   if (SubMesh->NElemsQuad3 > 0)
   {
      SubMesh->FirstNetIndex[E_ELEM_TYPE_QUAD3] = sum + 1;
      sum += SubMesh->NElemsQuad3;
   }

   if (SubMesh->NElemsTetrah1 > 0)
   {
      SubMesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1] = sum + 1;
      sum += SubMesh->NElemsTetrah1;
   }

   if (SubMesh->NElemsBlock1 > 0)
   {
      SubMesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1] = sum + 1;
      sum += SubMesh->NElemsBlock1;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetNodalDOF: To set the sub-mesh nodal degrees of freedom.   */


static void SetNodalDOF(MESH *SubMesh)
{
   int i,j,k;


   for (i = 0; i < SubMesh->NElemsLink1; i++)
   {
      for (j = 0; j < E_NNPE_LINK1; j++)
      {
         k = SubMesh->NodesLink1[i][j] - 1;

         if (SubMesh->NodalDOF[k] == 0)
         {
            SubMesh->TotalNodalDOF += E_VERTEX_NODE_DOF_LINK1;
            SubMesh->NodalDOF[k] = E_VERTEX_NODE_DOF_LINK1;
         }
      }
   }

   for (i = 0; i < SubMesh->NElemsLink2; i++)
   {
      for (j = 0; j < E_NNPE_LINK2; j++)
      {
         k = SubMesh->NodesLink2[i][j] - 1;

         if (SubMesh->NodalDOF[k] == 0)
         {
            SubMesh->TotalNodalDOF += E_VERTEX_NODE_DOF_LINK2;
            SubMesh->NodalDOF[k] = E_VERTEX_NODE_DOF_LINK2;
         }
      }
   }

   for (i = 0; i < SubMesh->NElemsLink3; i++)
   {
      for (j = 0; j < E_NNPE_LINK3; j++)
      {
         k = SubMesh->NodesLink3[i][j] - 1;

         if (SubMesh->NodalDOF[k] == 0)
         {
            SubMesh->TotalNodalDOF += E_VERTEX_NODE_DOF_LINK3;
            SubMesh->NodalDOF[k] = E_VERTEX_NODE_DOF_LINK3;
         }
      }
   }

   for (i = 0; i < SubMesh->NElemsLink4; i++)
   {
      for (j = 0; j < E_NNPE_LINK4; j++)
      {
         k = SubMesh->NodesLink4[i][j] - 1;

         if (SubMesh->NodalDOF[k] == 0)
         {
            SubMesh->TotalNodalDOF += E_VERTEX_NODE_DOF_LINK4;
            SubMesh->NodalDOF[k] = E_VERTEX_NODE_DOF_LINK4;
         }
      }
   }

   for (i = 0; i < SubMesh->NElemsLink5; i++)
   {
      for (j = 0; j < E_NNPE_LINK5; j++)
      {
         k = SubMesh->NodesLink5[i][j] - 1;

         if (SubMesh->NodalDOF[k] == 0)
         {
            SubMesh->TotalNodalDOF += E_VERTEX_NODE_DOF_LINK5;
            SubMesh->NodalDOF[k] = E_VERTEX_NODE_DOF_LINK5;
         }
      }
   }

   for (i = 0; i < SubMesh->NElemsTriang1; i++)
   {
      for (j = 0; j < E_NNPE_TRIANG1; j++)
      {
         k = SubMesh->NodesTriang1[i][j] - 1;

         if (SubMesh->NodalDOF[k] == 0)
         {
            SubMesh->TotalNodalDOF += E_VERTEX_NODE_DOF_TRIANG1;
            SubMesh->NodalDOF[k] = E_VERTEX_NODE_DOF_TRIANG1;
         }
      }
   }

   for (i = 0; i < SubMesh->NElemsTriang2; i++)
   {
      for (j = 0; j < E_NVPE_TRIANG2; j++)
      {
         k = SubMesh->NodesTriang2[i][j] - 1;

         if (SubMesh->NodalDOF[k] == 0)
         {
            SubMesh->TotalNodalDOF += E_VERTEX_NODE_DOF_TRIANG2;
            SubMesh->NodalDOF[k] = E_VERTEX_NODE_DOF_TRIANG2;
         }
      }

      for ( ; j < E_NNPE_TRIANG2; j++)
      {
         k = SubMesh->NodesTriang2[i][j] - 1;

         if (SubMesh->NodalDOF[k] == 0)
         {
            SubMesh->TotalNodalDOF += E_MIDSIDE_NODE_DOF_TRIANG2;
            SubMesh->NodalDOF[k] = E_MIDSIDE_NODE_DOF_TRIANG2;
         }
      }
   }

   for (i = 0; i < SubMesh->NElemsTriang3; i++)
   {
      for (j = 0; j < E_NNPE_TRIANG3; j++)
      {
         k = SubMesh->NodesTriang3[i][j] - 1;

         if (SubMesh->NodalDOF[k] == 0)
         {
            SubMesh->TotalNodalDOF += E_VERTEX_NODE_DOF_TRIANG3;
            SubMesh->NodalDOF[k] = E_VERTEX_NODE_DOF_TRIANG3;
         }
      }
   }

   for (i = 0; i < SubMesh->NElemsTriang4; i++)
   {
      for (j = 0; j < E_NNPE_TRIANG4; j++)
      {
         k = SubMesh->NodesTriang4[i][j] - 1;

         if (SubMesh->NodalDOF[k] == 0)
         {
            SubMesh->TotalNodalDOF += E_VERTEX_NODE_DOF_TRIANG4;
            SubMesh->NodalDOF[k] = E_VERTEX_NODE_DOF_TRIANG4;
         }
      }
   }

   for (i = 0; i < SubMesh->NElemsTriang5; i++)
   {
      for (j = 0; j < E_NNPE_TRIANG5; j++)
      {
         k = SubMesh->NodesTriang5[i][j] - 1;

         if (SubMesh->NodalDOF[k] == 0)
         {
            SubMesh->TotalNodalDOF += E_VERTEX_NODE_DOF_TRIANG5;
            SubMesh->NodalDOF[k] = E_VERTEX_NODE_DOF_TRIANG5;
         }
      }
   }

   for (i = 0; i < SubMesh->NElemsQuad1; i++)
   {
      for (j = 0; j < E_NNPE_QUAD1; j++)
      {
         k = SubMesh->NodesQuad1[i][j] - 1;

         if (SubMesh->NodalDOF[k] == 0)
         {
            SubMesh->TotalNodalDOF += E_VERTEX_NODE_DOF_QUAD1;
            SubMesh->NodalDOF[k] = E_VERTEX_NODE_DOF_QUAD1;
         }
      }
   }

   for (i = 0; i < SubMesh->NElemsQuad3; i++)
   {
      for (j = 0; j < E_NVPE_QUAD3; j++)
      {
         k = SubMesh->NodesQuad3[i][j] - 1;

         if (SubMesh->NodalDOF[k] == 0)
         {
            SubMesh->TotalNodalDOF += E_VERTEX_NODE_DOF_QUAD3;
            SubMesh->NodalDOF[k] = E_VERTEX_NODE_DOF_QUAD3;
         }
      }

      for ( ; j < E_NNPE_QUAD3; j++)
      {
         k = SubMesh->NodesQuad3[i][j] - 1;

         if (SubMesh->NodalDOF[k] == 0)
         {
            SubMesh->TotalNodalDOF += E_MIDSIDE_NODE_DOF_QUAD3;
            SubMesh->NodalDOF[k] = E_MIDSIDE_NODE_DOF_QUAD3;
         }
      }
   }


   for (i = 0; i < SubMesh->NElemsTetrah1; i++)
   {
      for (j = 0; j < E_NNPE_TETRAH1; j++)
      {
         k = SubMesh->NodesTetrah1[i][j] - 1;

         if (SubMesh->NodalDOF[k] == 0)
         {
            SubMesh->TotalNodalDOF += E_VERTEX_NODE_DOF_TETRAH1;
            SubMesh->NodalDOF[k] = E_VERTEX_NODE_DOF_TETRAH1;
         }
      }
   }

   for (i = 0; i < SubMesh->NElemsBlock1; i++)
   {
      for (j = 0; j < E_NNPE_BLOCK1; j++)
      {
         k = SubMesh->NodesBlock1[i][j] - 1;

         if (SubMesh->NodalDOF[k] == 0)
         {
            SubMesh->TotalNodalDOF += E_VERTEX_NODE_DOF_BLOCK1;
            SubMesh->NodalDOF[k] = E_VERTEX_NODE_DOF_BLOCK1;
         }
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetMeshPointBCs: To set the sub-mesh mesh-point boundary conditions for a
   particular element type.                                                   */


static void SetMeshPointBCs
               (MESH *mesh, MESH *SubMesh, int NVerticesPerElement,
                int NMeshElements, int **MeshNodes, int FirstNetIndex,
                int *BCNumbers, char *NodeStored, int SubdomainIndex  )
{
   int i,j,k,l,m,n;
   int MeshElementNetIndex;


   MeshElementNetIndex = FirstNetIndex - 1;

   for (i = 0; i < NMeshElements; i++)
   {
      if (mesh->Decomp.SubDoms[MeshElementNetIndex++] == SubdomainIndex)
      {
         for (j = 0; j < NVerticesPerElement; j++)
         {
            k = MeshNodes[i][j] - 1;

            if (!NodeStored[k])
            {
               (SubMesh->NNodes)++;

               if (BCNumbers[k] > 0)
               {
                  l = SubMesh->NBCNodes++;
                  m = BCNumbers[k] - 1;
                  SubMesh->BCNodes[l] = SubMesh->NNodes;

                  for (n = 0; n < E_DOF_MAX; n++)
                  {
                     SubMesh->BCTypes[l][n] = mesh->BCTypes[m][n];
                     SubMesh->BCDispls[l][n] = mesh->BCDispls[m][n];
                     SubMesh->BCSpringConsts[l][n]
                         = mesh->BCSpringConsts[m][n];
                  }
               }

               NodeStored[k] = TRUE;
            }
         }
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetNonMeshPointBCs: To set the sub-mesh non-mesh-point boundary conditions
   for a particular element type.                                             */


static void SetNonMeshPointBCs
               (MESH *mesh, MESH *SubMesh, int NNodesPerElement,
                int NVerticesPerElement, int NMeshElements, int **MeshNodes,
                int FirstNetIndex, int *BCNumbers, char *NodeStored,
                int SubdomainIndex                                          )
{
   int i,j,k,l,m,n;
   int MeshElementNetIndex;


   MeshElementNetIndex = FirstNetIndex - 1;

   for (i = 0; i < NMeshElements; i++)
   {
      if (mesh->Decomp.SubDoms[MeshElementNetIndex++] == SubdomainIndex)
      {
         for (j = NVerticesPerElement; j < NNodesPerElement; j++)
         {
            k = MeshNodes[i][j] - 1;

            if (!NodeStored[k])
            {
               (SubMesh->NNodes)++;

               if (BCNumbers[k] > 0)
               {
                  l = SubMesh->NBCNodes++;
                  m = BCNumbers[k] - 1;
                  SubMesh->BCNodes[l] = SubMesh->NNodes;

                  for (n = 0; n < E_DOF_MAX; n++)
                  {
                     SubMesh->BCTypes[l][n] = mesh->BCTypes[m][n];
                     SubMesh->BCDispls[l][n] = mesh->BCDispls[m][n];
                     SubMesh->BCSpringConsts[l][n]
                         = mesh->BCSpringConsts[m][n];
                  }
               }

               NodeStored[k] = TRUE;
            }
         }
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetMeshPointLoads: To set the sub-mesh mesh-point loads for a particular
   element type.                                                              */


static void SetMeshPointLoads
               (MESH *mesh, MESH *SubMesh, int NVerticesPerElement,
                int NMeshElements, int **MeshNodes, int FirstNetIndex,
                int *LoadedNodeNumbers, char *NodeStored,
                int SubdomainIndex                                    )
{
   int i,j,k,l,m,n;
   int MeshElementNetIndex;


   MeshElementNetIndex = FirstNetIndex - 1;

   for (i = 0; i < NMeshElements; i++)
   {
      if (mesh->Decomp.SubDoms[MeshElementNetIndex++] == SubdomainIndex)
      {
         for (j = 0; j < NVerticesPerElement; j++)
         {
            k = MeshNodes[i][j] - 1;

            if (!NodeStored[k])
            {
               (SubMesh->NNodes)++;

               if (LoadedNodeNumbers[k] > 0)
               {
                  l = SubMesh->NLoadedNodes++;
                  m = LoadedNodeNumbers[k] - 1;
                  SubMesh->LoadedNodes[l] = SubMesh->NNodes;

                  for (n = 0; n < E_DOF_MAX; n++)
                     SubMesh->Loads[l][n] = mesh->Loads[m][n];
               }

               NodeStored[k] = TRUE;
            }
         }
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetNonMeshPointLoads: To set the sub-mesh non-mesh-point loads for a
   particular element type.                                                   */


static void SetNonMeshPointLoads
               (MESH *mesh, MESH *SubMesh, int NNodesPerElement,
                int NVerticesPerElement, int NMeshElements, int **MeshNodes,
                int FirstNetIndex, int *LoadedNodeNumbers, char *NodeStored,
                int SubdomainIndex                                          )
{
   int i,j,k,l,m,n;
   int MeshElementNetIndex;


   MeshElementNetIndex = FirstNetIndex - 1;

   for (i = 0; i < NMeshElements; i++)
   {
      if (mesh->Decomp.SubDoms[MeshElementNetIndex++] == SubdomainIndex)
      {
         for (j = NVerticesPerElement; j < NNodesPerElement; j++)
         {
            k = MeshNodes[i][j] - 1;

            if (!NodeStored[k])
            {
               (SubMesh->NNodes)++;

               if (LoadedNodeNumbers[k] > 0)
               {
                  l = SubMesh->NLoadedNodes++;
                  m = LoadedNodeNumbers[k] - 1;
                  SubMesh->LoadedNodes[l] = SubMesh->NNodes;

                  for (n = 0; n < E_DOF_MAX; n++)
                     SubMesh->Loads[l][n] = mesh->Loads[m][n];
               }

               NodeStored[k] = TRUE;
            }
         }
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetMaterials: To set the sub-mesh material data for a particular element
   type.                                                                      */

/* The return value is 1 if the data are set successfully, else 0. */


static int SetMaterials(MESH *mesh, MESH *SubMesh, int NMeshElements,
                        int FirstNetIndex, int *PropsIndices,
                        int *MeshToSubMeshPropsIndices, int SubdomainIndex)
{
   int i,k,l,r;
   int MeshElementNetIndex;
   int PropsIndex;


   MeshElementNetIndex = FirstNetIndex - 1;

   for (i = 0; i < NMeshElements; i++)
   {
      if (mesh->Decomp.SubDoms[MeshElementNetIndex] == SubdomainIndex)
      {
         k = (SubMesh->TotalNElems)++;
         r = (int)((MAT *)mesh->Mats[MeshElementNetIndex] - mesh->Props);
         PropsIndex = GetPropsIndex(r,SubMesh->NMats,PropsIndices);

         if (PropsIndex > 0)
         {
            SubMesh->Mats[k] = &SubMesh->Props[PropsIndex - 1];
         }
         else
         {
            l = SubMesh->NMats++;
            PropsIndices[l] = r;
            MeshToSubMeshPropsIndices[r] = l;
            SubMesh->Mats[k] = &SubMesh->Props[l];
            SubMesh->NMdfMats++;

            if ((SubMesh->Props[l].Name
                    = e_alloc_c_vect(strlen(mesh->Props[r].Name) + 1)) == NULL)
            {
               return 0;
            }

            strcpy(SubMesh->Props[l].Name,mesh->Props[r].Name);
         }
      }

      MeshElementNetIndex++;
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetMatsType1: To set the sub-mesh TYPE1 composite material data for a
   particular element type.                                                   */

/* The return value is 1 if the data are set successfully, else 0. */


static int SetMatsType1(MESH *mesh, MESH *SubMesh, int NMeshElements,
                        int FirstNetIndex, int *PropsIndices,
                        int *CompPropsIndicesType1,
                        int *MeshToSubMeshPropsIndices, int SubdomainIndex)

{
   int i,k,l,m,r,s;
   int CompPropsIndex;
   int MeshElementNetIndex;
   int PropsIndex;


   MeshElementNetIndex = FirstNetIndex - 1;

   for (i = 0; i < NMeshElements; i++)
   {
      if (mesh->Decomp.SubDoms[MeshElementNetIndex] == SubdomainIndex)
      {
         k = (SubMesh->TotalNElems)++;
         r = (int)((COMPMAT1 *)mesh->Mats[MeshElementNetIndex]
                                        - mesh->CompPropsType1);
         CompPropsIndex = GetPropsIndex(r,SubMesh->NCompMatsType1,
                                             CompPropsIndicesType1);
         if (CompPropsIndex > 0)
         {
            SubMesh->Mats[k] = &SubMesh->CompPropsType1[CompPropsIndex - 1];
         }
         else
         {
            l = SubMesh->NCompMatsType1++;
            CompPropsIndicesType1[l] = r;
            SubMesh->Mats[k] = &SubMesh->CompPropsType1[l];

            if ((SubMesh->CompPropsType1[l].Name
                    = e_alloc_c_vect
                      (strlen(mesh->CompPropsType1[r].Name) + 1)) == NULL)
            {
               return 0;
            }

            strcpy(SubMesh->CompPropsType1[l].Name,
                      mesh->CompPropsType1[r].Name );

            for (m = 0; m < mesh->CompPropsType1[r].NLayers; m++)
            {
               s = (int)(mesh->CompPropsType1[r].LayerMats[m] - mesh->Props);
               PropsIndex = GetPropsIndex(s,SubMesh->NMats,PropsIndices);

               if (PropsIndex == 0)
               {
                  l = SubMesh->NMats++;
                  PropsIndices[l] = s;
                  MeshToSubMeshPropsIndices[s] = l;

                  if ((SubMesh->Props[l].Name
                          = e_alloc_c_vect
                            (strlen(mesh->Props[s].Name) + 1)) == NULL)
                  {
                     return 0;
                  }

                  strcpy(SubMesh->Props[l].Name,mesh->Props[s].Name);
               }
            }
         }
      }

      MeshElementNetIndex++;
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetMaterialProperties: To set the sub-mesh material properties. */

/* The return value is 1 if the properties are set successfully, else 0. */


static int SetMaterialProperties(MESH *mesh, MESH *SubMesh,
                                 int *PropsIndices,
                                 int *CompPropsIndicesType1,
                                 int *MeshToSubMeshPropsIndices)
{
   int j,k,l,r,s;

   char *pTemp;


   for (k = 0; k < SubMesh->NMats; k++)
   {
      r = PropsIndices[k];
      pTemp = SubMesh->Props[k].Name;
      memcpy(&SubMesh->Props[k],&mesh->Props[r],sizeof(MAT));
      SubMesh->Props[k].Name = pTemp;
   }

   for (k = 0; k < SubMesh->NCompMatsType1; k++)
   {
      r = CompPropsIndicesType1[k];

      if (   (SubMesh->CompPropsType1[k].LayerMats
                = (MAT **)
                       e_calloc(mesh->CompPropsType1[r].NLayers,
                                                   sizeof(MAT *))) == NULL
          || (SubMesh->CompPropsType1[k].Thicknesses
                = e_alloc_d_vect(mesh->CompPropsType1[r].NLayers)) == NULL)
      {
         return 0;
      }

      SubMesh->CompPropsType1[k].NLayers = mesh->CompPropsType1[r].NLayers;

      for (j = 0; j < SubMesh->CompPropsType1[k].NLayers; j++)
      {
         s = (int)(mesh->CompPropsType1[r].LayerMats[j] - mesh->Props);
         l = MeshToSubMeshPropsIndices[s];

         SubMesh->CompPropsType1[k].LayerMats[j] = &SubMesh->Props[l];
         SubMesh->CompPropsType1[k].Thicknesses[j]
            = mesh->CompPropsType1[r].Thicknesses[j];
      }
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetNurbsCurves: To set the remeshing and NURBS curve data. */

/* The return value is 1 if the mesh is created successfully, else 0. */


static int SetNurbsCurves(MESH *mesh, MESH *SubMesh, int NVerticesPerElement,
                          int NMeshElements, int **MeshNodes,
                          int FirstNetIndex, int *EndPoints,
                          int *NurbsCurveIndices, char *NodeStored,
                          int SubdomainIndex                                 )
{
   int i,j,k,l,m,p,q,r,s,t,u;
   int Degree;
   int MeshElementNetIndex;
   int NameLength;


   MeshElementNetIndex = FirstNetIndex - 1;

   if (mesh->SubStructSet[E_SS_NUM_GEOM])
   {
      for (i = 0; i < NMeshElements; i++)
      {
         if (mesh->Decomp.SubDoms[MeshElementNetIndex++] != SubdomainIndex)
            continue;

         for (j = 0; j < NVerticesPerElement; j++)
         {
            k = MeshNodes[i][j] - 1;

            if (NodeStored[k]) continue;

            NodeStored[k] = TRUE;
            m = SubMesh->NMeshPoints++;
            SubMesh->NodeTypes[m] = mesh->NodeTypes[k];
            SubMesh->NNurbsCurves[m] = mesh->NNurbsCurves[k];

            if (SubMesh->NodeTypes[m] != E_RM_NUM_EXTERIOR) continue;

            if (   (SubMesh->NurbsCurveParams[m]
                       = e_alloc_d_vect(SubMesh->NNurbsCurves[m])) == NULL
                || (SubMesh->NurbsCurves[m]
                       = (NURBSCURVE **)
                          e_calloc(SubMesh->NNurbsCurves[m],
                                            sizeof(NURBSCURVE *))) == NULL)
            {
               return 0;
            }

            for (l = 0; l < SubMesh->NNurbsCurves[m]; l++)
            {
               r = (int)(mesh->NurbsCurves[k][l] - mesh->Geom.NurbsCurves);

               if (NurbsCurveIndices[r] > 0)
               {
                  s = NurbsCurveIndices[r] - 1;
                  SubMesh->NurbsCurveParams[m][l]
                     = mesh->NurbsCurveParams[k][l];
                  SubMesh->NurbsCurves[m][l] = &SubMesh->Geom.NurbsCurves[s];
                  continue;
               }

               s = SubMesh->Geom.NNurbsCurves++;
               NurbsCurveIndices[r] = SubMesh->Geom.NNurbsCurves;
               SubMesh->NurbsCurveParams[m][l] = mesh->NurbsCurveParams[k][l];
               SubMesh->NurbsCurves[m][l] = &SubMesh->Geom.NurbsCurves[s];

               Degree = SubMesh->Geom.NurbsCurves[s].Degree
                      = mesh->Geom.NurbsCurves[r].Degree;
               NameLength = strlen(mesh->Geom.NurbsCurves[r].Name);

               if (   (SubMesh->Geom.NurbsCurves[s].Name
                                 = e_alloc_c_vect(NameLength + 1)) == NULL
                   || (SubMesh->Geom.NurbsCurves[s].CntlPoints
                          = e_alloc_cont_d_matr
                                         (Degree + 1,E_NDIMS + 1)) == NULL)
               {
                  return 0;
               }

               strcpy(SubMesh->Geom.NurbsCurves[s].Name,
                         mesh->Geom.NurbsCurves[r].Name );

               for (t = 0; t < Degree + 1; t++)
               {
                  SubMesh->Geom.NurbsCurves[s].CntlPoints[t][0]
                     = mesh->Geom.NurbsCurves[r].CntlPoints[t][0];
                  SubMesh->Geom.NurbsCurves[s].CntlPoints[t][1]
                     = mesh->Geom.NurbsCurves[r].CntlPoints[t][1];
                  SubMesh->Geom.NurbsCurves[s].CntlPoints[t][2]
                     = mesh->Geom.NurbsCurves[r].CntlPoints[t][2];
                  SubMesh->Geom.NurbsCurves[s].CntlPoints[t][3]
                     = mesh->Geom.NurbsCurves[r].CntlPoints[t][3];
               }

               SubMesh->Geom.NurbsCurves[s].FixityType
                  = mesh->Geom.NurbsCurves[r].FixityType;

               if (   SubMesh->Geom.NurbsCurves[s].FixityType
                                                  == E_NC_NUM_FIXITY_GC
                   || SubMesh->Geom.NurbsCurves[s].FixityType
                                                  == E_NC_NUM_FIXITY_CC )
               {
                  if (   (SubMesh->Geom.NurbsCurves[s].BCTypes
                                          = e_alloc_i_vect(E_DOF_MAX)) == NULL
                      || (SubMesh->Geom.NurbsCurves[s].BCDispls
                                          = e_alloc_d_vect(E_DOF_MAX)) == NULL
                      || (SubMesh->Geom.NurbsCurves[s].BCSpringConsts
                                          = e_alloc_d_vect(E_DOF_MAX)) == NULL)
                  {
                     return 0;
                  }

                  for (u = 0; u < E_DOF_MAX; u++)
                  {
                     SubMesh->Geom.NurbsCurves[s].BCTypes[u]
                        = mesh->Geom.NurbsCurves[r].BCTypes[u];
                     SubMesh->Geom.NurbsCurves[s].BCDispls[u]
                        = mesh->Geom.NurbsCurves[r].BCDispls[u];
                     SubMesh->Geom.NurbsCurves[s].BCSpringConsts[u]
                        = mesh->Geom.NurbsCurves[r].BCSpringConsts[u];
                  }
               }

               for (u = 0; u < E_NC_NLOAD_TYPES; u++)
                  SubMesh->Geom.NurbsCurves[s].LoadTypeSet[u]
                     = mesh->Geom.NurbsCurves[r].LoadTypeSet[u];

               if (SubMesh->Geom.NurbsCurves[s].LoadTypeSet
                                                [E_NC_NUM_POINT_LOAD])
               {
                  if ((SubMesh->Geom.NurbsCurves[s].PointLoad
                                  = e_alloc_d_vect(E_DOF_MAX)) == NULL)
                  {
                     return 0;
                  }

                  for (u = 0; u < E_DOF_MAX; u++)
                     SubMesh->Geom.NurbsCurves[s].PointLoad[u]
                        = mesh->Geom.NurbsCurves[r].PointLoad[u];

                  SubMesh->Geom.NurbsCurves[s].PointLoadParam
                     = mesh->Geom.NurbsCurves[r].PointLoadParam;
               }

               if (SubMesh->Geom.NurbsCurves[s].LoadTypeSet
                                                [E_NC_NUM_LIN_LOAD_GC])
               {
                  if (   (SubMesh->Geom.NurbsCurves[s].LinLoadGCBeg
                                           = e_alloc_d_vect(E_DOF_MAX)) == NULL
                      || (SubMesh->Geom.NurbsCurves[s].LinLoadGCEnd
                                           = e_alloc_d_vect(E_DOF_MAX)) == NULL)
                  {
                     return 0;
                  }

                  for (u = 0; u < E_DOF_MAX; u++)
                  {
                     SubMesh->Geom.NurbsCurves[s].LinLoadGCBeg[u]
                        = mesh->Geom.NurbsCurves[r].LinLoadGCBeg[u];
                     SubMesh->Geom.NurbsCurves[s].LinLoadGCEnd[u]
                        = mesh->Geom.NurbsCurves[r].LinLoadGCEnd[u];
                  }
               }

               if (SubMesh->Geom.NurbsCurves[s].LoadTypeSet
                                                [E_NC_NUM_LIN_LOAD_CC])
               {
                  if (   (SubMesh->Geom.NurbsCurves[s].LinLoadCCBeg
                                           = e_alloc_d_vect(E_DOF_MAX)) == NULL
                      || (SubMesh->Geom.NurbsCurves[s].LinLoadCCEnd
                                           = e_alloc_d_vect(E_DOF_MAX)) == NULL)
                  {
                     return 0;
                  }

                  for (u = 0; u < E_DOF_MAX; u++)
                  {
                     SubMesh->Geom.NurbsCurves[s].LinLoadCCBeg[u]
                        = mesh->Geom.NurbsCurves[r].LinLoadCCBeg[u];
                     SubMesh->Geom.NurbsCurves[s].LinLoadCCEnd[u]
                        = mesh->Geom.NurbsCurves[r].LinLoadCCEnd[u];
                  }
               }

               p = mesh->Geom.NurbsCurves[r].EndPoint1 - 1;

               if (EndPoints[p] == 0)
               {
                  q = SubMesh->Geom.NEndPoints++;
                  EndPoints[p] = SubMesh->Geom.NEndPoints;
                  SubMesh->Geom.EndPointCoords[q][0]
                     = mesh->Geom.EndPointCoords[p][0];
                  SubMesh->Geom.EndPointCoords[q][1]
                     = mesh->Geom.EndPointCoords[p][1];
                  SubMesh->Geom.EndPointCoords[q][2]
                     = mesh->Geom.EndPointCoords[p][2];
               }

               SubMesh->Geom.NurbsCurves[s].EndPoint1 = EndPoints[p];
               p = mesh->Geom.NurbsCurves[r].EndPoint2 - 1;

               if (EndPoints[p] == 0)
               {
                  q = SubMesh->Geom.NEndPoints++;
                  EndPoints[p] = SubMesh->Geom.NEndPoints;
                  SubMesh->Geom.EndPointCoords[q][0]
                     = mesh->Geom.EndPointCoords[p][0];
                  SubMesh->Geom.EndPointCoords[q][1]
                     = mesh->Geom.EndPointCoords[p][1];
                  SubMesh->Geom.EndPointCoords[q][2]
                     = mesh->Geom.EndPointCoords[p][2];
               }

               SubMesh->Geom.NurbsCurves[s].EndPoint2 = EndPoints[p];
            }
         }
      }
   }
   else
   {
      for (i = 0; i < NMeshElements; i++)
      {
         if (mesh->Decomp.SubDoms[MeshElementNetIndex++] != SubdomainIndex)
            continue;

         for (j = 0; j < NVerticesPerElement; j++)
         {
            k = MeshNodes[i][j] - 1;

            if (NodeStored[k]) continue;

            NodeStored[k] = TRUE;
            m = SubMesh->NMeshPoints++;
            SubMesh->NodeTypes[m] = mesh->NodeTypes[k];
            SubMesh->NNurbsCurves[m] = mesh->NNurbsCurves[k];

            if (SubMesh->NodeTypes[m] != E_RM_NUM_EXTERIOR) continue;

            if (   (SubMesh->NurbsCurveParams[m]
                       = e_alloc_d_vect(SubMesh->NNurbsCurves[m])) == NULL
                || (SubMesh->NurbsCurves[m]
                       = (NURBSCURVE **)
                          e_calloc(SubMesh->NNurbsCurves[m],
                                            sizeof(NURBSCURVE *))) == NULL)
            {
               return 0;
            }

            for (l = 0; l < SubMesh->NNurbsCurves[m]; l++)
            {
               r = (int)(mesh->NurbsCurves[k][l] - mesh->Geom.NurbsCurves);

               if (NurbsCurveIndices[r] > 0)
               {
                  s = NurbsCurveIndices[r] - 1;
                  SubMesh->NurbsCurveParams[m][l]
                     = mesh->NurbsCurveParams[k][l];
                  SubMesh->NurbsCurves[m][l] = &SubMesh->Geom.NurbsCurves[s];
                  continue;
               }

               s = SubMesh->Geom.NNurbsCurves++;
               NurbsCurveIndices[r] = SubMesh->Geom.NNurbsCurves;
               SubMesh->NurbsCurveParams[m][l] = mesh->NurbsCurveParams[k][l];
               SubMesh->NurbsCurves[m][l] = &SubMesh->Geom.NurbsCurves[s];

               NameLength = strlen(mesh->Geom.NurbsCurves[r].Name);

               if (   (SubMesh->Geom.NurbsCurves[s].Name
                                 = e_alloc_c_vect(NameLength + 1)) == NULL)
               {
                  return 0;
               }

               strcpy(SubMesh->Geom.NurbsCurves[s].Name,
                         mesh->Geom.NurbsCurves[r].Name );
            }
         }
      }
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetElementMaterials: To set the sub-mesh element material pointers.        */


static void SetElementMaterials(MESH *SubMesh)
{
   /* only if there are materials defined */
   if(SubMesh->NMats > 0) 
   {
     SubMesh->MatsLink1
        = (MAT **)
          (SubMesh->Mats + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK1] - 1);
     SubMesh->MatsLink2
        = (MAT **)
          (SubMesh->Mats + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK2] - 1);
     SubMesh->MatsLink3
        = (MAT **)
          (SubMesh->Mats + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK3] - 1);
     SubMesh->MatsLink4
        = (MAT **)
          (SubMesh->Mats + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK4] - 1);   
     SubMesh->MatsLink5
        = (MAT **)
          (SubMesh->Mats + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK5] - 1);
     SubMesh->MatsTriang1
        = (MAT **)
          (SubMesh->Mats + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] - 1);
     SubMesh->CompMatsTriang2
        = (COMPMAT1 **)
          (SubMesh->Mats + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2] - 1);
     SubMesh->MatsTriang3
        = (MAT **)
          (SubMesh->Mats + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] - 1);
     SubMesh->MatsTriang4
        = (MAT **)
          (SubMesh->Mats + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] - 1);   
     SubMesh->MatsTriang5
        = (MAT **)
          (SubMesh->Mats + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] - 1);
     SubMesh->MatsQuad1
        = (MAT **)
          (SubMesh->Mats + SubMesh->FirstNetIndex[E_ELEM_TYPE_QUAD1] - 1);
     SubMesh->CompMatsQuad3
        = (COMPMAT1 **)
          (SubMesh->Mats + SubMesh->FirstNetIndex[E_ELEM_TYPE_QUAD3] - 1);
     SubMesh->MatsTetrah1
        = (MAT **)
          (SubMesh->Mats + SubMesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1] - 1);
     SubMesh->MatsBlock1
        = (MAT **)
          (SubMesh->Mats + SubMesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1] - 1);
  }
}



/* ////////////////////////////////////////////////////////////////////////// */
/* SetGlobalElementIndices: To set the sub-mesh global element indices.       */


static void SetGlobalElementIndices(MESH *SubMesh)
{
   SubMesh->GEIsLink1
      = SubMesh->GEIs + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK1] - 1;
   SubMesh->GEIsLink2
      = SubMesh->GEIs + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK2] - 1;
   SubMesh->GEIsLink3
      = SubMesh->GEIs + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK3] - 1;
   SubMesh->GEIsLink4
      = SubMesh->GEIs + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK4] - 1;
   SubMesh->GEIsLink5
      = SubMesh->GEIs + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK5] - 1;
   SubMesh->GEIsTriang1
      = SubMesh->GEIs + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] - 1;
   SubMesh->GEIsTriang2
      = SubMesh->GEIs + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2] - 1;
   SubMesh->GEIsTriang3
      = SubMesh->GEIs + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] - 1;
   SubMesh->GEIsTriang4
      = SubMesh->GEIs + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] - 1;
   SubMesh->GEIsTriang5
      = SubMesh->GEIs + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] - 1;
   SubMesh->GEIsQuad1
      = SubMesh->GEIs + SubMesh->FirstNetIndex[E_ELEM_TYPE_QUAD1] - 1;
   SubMesh->GEIsQuad3
      = SubMesh->GEIs + SubMesh->FirstNetIndex[E_ELEM_TYPE_QUAD3] - 1;
   SubMesh->GEIsTetrah1
      = SubMesh->GEIs + SubMesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1] - 1;
   SubMesh->GEIsBlock1
      = SubMesh->GEIs + SubMesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1] - 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetMeshParams: To set the sub-mesh mesh parameters. */


static void SetMeshParams(MESH *mesh, MESH *SubMesh, int *NewNodeIndex,
                                                     int SubdomainIndex)
{
   int i,k,l;


   SubMesh->MeshParam.NodalMeshParamsSet
      = mesh->MeshParam.NodalMeshParamsSet;
   SubMesh->MeshParam.ElemMeshParamsSet
      = mesh->MeshParam.ElemMeshParamsSet;

   if (SubMesh->MeshParam.NodalMeshParamsSet)
   {
      for (k = 0; k < mesh->NMeshPoints; k++)
      {
         if (NewNodeIndex[k] > 0)
         {
            l = NewNodeIndex[k] - 1;
            SubMesh->MeshParam.NodalMeshParams[l] =
               mesh->MeshParam.NodalMeshParams[k];
         }
      }
   }

   if (SubMesh->MeshParam.ElemMeshParamsSet)
   {
      k = 0;

      for (i = 0; i < mesh->TotalNElems; i++)
      {
         if (mesh->Decomp.SubDoms[i] == SubdomainIndex)
            SubMesh->MeshParam.ElemMeshParams[k++]
               = mesh->MeshParam.ElemMeshParams[i];
      }

      SubMesh->MeshParam.ElemMeshParamsLink1
         = SubMesh->MeshParam.ElemMeshParams
           + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK1] - 1;
      SubMesh->MeshParam.ElemMeshParamsLink2
         = SubMesh->MeshParam.ElemMeshParams
           + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK2] - 1;
      SubMesh->MeshParam.ElemMeshParamsLink3
         = SubMesh->MeshParam.ElemMeshParams
           + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK3] - 1;
      SubMesh->MeshParam.ElemMeshParamsLink4
         = SubMesh->MeshParam.ElemMeshParams
           + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK4] - 1;
      SubMesh->MeshParam.ElemMeshParamsLink5
         = SubMesh->MeshParam.ElemMeshParams
           + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK5] - 1;
      SubMesh->MeshParam.ElemMeshParamsTriang1
         = SubMesh->MeshParam.ElemMeshParams
           + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] - 1;
      SubMesh->MeshParam.ElemMeshParamsTriang2
         = SubMesh->MeshParam.ElemMeshParams
           + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2] - 1;
      SubMesh->MeshParam.ElemMeshParamsTriang3
         = SubMesh->MeshParam.ElemMeshParams
           + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] - 1;
      SubMesh->MeshParam.ElemMeshParamsTriang4
         = SubMesh->MeshParam.ElemMeshParams
           + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] - 1;
      SubMesh->MeshParam.ElemMeshParamsTriang5
         = SubMesh->MeshParam.ElemMeshParams
           + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] - 1;
      SubMesh->MeshParam.ElemMeshParamsQuad1
         = SubMesh->MeshParam.ElemMeshParams
           + SubMesh->FirstNetIndex[E_ELEM_TYPE_QUAD1] - 1;
      SubMesh->MeshParam.ElemMeshParamsQuad3
         = SubMesh->MeshParam.ElemMeshParams
           + SubMesh->FirstNetIndex[E_ELEM_TYPE_QUAD3] - 1;
      SubMesh->MeshParam.ElemMeshParamsTetrah1
         = SubMesh->MeshParam.ElemMeshParams
           + SubMesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1] - 1;
      SubMesh->MeshParam.ElemMeshParamsBlock1
         = SubMesh->MeshParam.ElemMeshParams
           + SubMesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1] - 1;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetStresses: To set the sub-mesh stresses. */

static void SetStresses(MESH *mesh, MESH *Submesh, int SubdomainIndex)
{
  int i, j, k, l, n;

  if (Submesh->KeywordSet[E_KW_NUM_NELEMS_LINK1])
  {
    n = 0;
    for (i = 0; i < mesh->NElemsLink1; i++)
      if (mesh->Decomp.SubDomsLink1[i] == SubdomainIndex)
      {
        for(j = 0; j < Submesh->NStressPointsLink1; j++)
          for(k = 0; k < E_NUM_STRESS_COMPONENTS; k++)
            Submesh->Stresses.StressesLink1[n][j][k] =
              mesh->Stresses.StressesLink1[i][j][k];
        n++;
      }
    if(n != Submesh->NElemsLink1)
      e_GeneralError("Internal structure damaged !");
  }

  if (Submesh->KeywordSet[E_KW_NUM_NELEMS_LINK2])
  {
    n = 0;
    for (i = 0; i < mesh->NElemsLink2; i++)
      if (mesh->Decomp.SubDomsLink2[i] == SubdomainIndex)
      {
        for(j = 0; j < Submesh->NStressPointsLink2; j++)
          for(k = 0; k < E_NUM_STRESS_COMPONENTS; k++)
            Submesh->Stresses.StressesLink2[n][j][k] =
              mesh->Stresses.StressesLink2[i][j][k];
        n++;
      }
    if(n != Submesh->NElemsLink2)
      e_GeneralError("Internal structure damaged !");
  }

  if (Submesh->KeywordSet[E_KW_NUM_NELEMS_LINK3])
  {
    n = 0;
    for (i = 0; i < mesh->NElemsLink3; i++)
      if (mesh->Decomp.SubDomsLink3[i] == SubdomainIndex)
      {
        for(j = 0; j < Submesh->NStressPointsLink3; j++)
          for(k = 0; k < E_NUM_STRESS_COMPONENTS; k++)
            Submesh->Stresses.StressesLink3[n][j][k] =
              mesh->Stresses.StressesLink3[i][j][k];
        n++;
      }
    if(n != Submesh->NElemsLink3)
      e_GeneralError("Internal structure damaged !");
  }

  if (Submesh->KeywordSet[E_KW_NUM_NELEMS_LINK4])
  {
    n = 0;
    for (i = 0; i < mesh->NElemsLink4; i++)
      if (mesh->Decomp.SubDomsLink4[i] == SubdomainIndex)
      {
        for(j = 0; j < Submesh->NStressPointsLink4; j++)
          for(k = 0; k < E_NUM_STRESS_COMPONENTS; k++)
            Submesh->Stresses.StressesLink4[n][j][k] =
              mesh->Stresses.StressesLink4[i][j][k];
        n++;
      }
    if(n != Submesh->NElemsLink4)
      e_GeneralError("Internal structure damaged !");
  }

  if (Submesh->KeywordSet[E_KW_NUM_NELEMS_LINK5])
  {
    n = 0;
    for (i = 0; i < mesh->NElemsLink5; i++)
      if (mesh->Decomp.SubDomsLink5[i] == SubdomainIndex)
      {
        for(j = 0; j < Submesh->NStressPointsLink5; j++)
          for(k = 0; k < E_NUM_STRESS_COMPONENTS; k++)
            Submesh->Stresses.StressesLink5[n][j][k] =
              mesh->Stresses.StressesLink5[i][j][k];
        n++;
      }
    if(n != Submesh->NElemsLink5)
      e_GeneralError("Internal structure damaged !");
  }

  if (Submesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1])
  {
    n = 0;
    for (i = 0; i < mesh->NElemsTriang1; i++)
      if (mesh->Decomp.SubDomsTriang1[i] == SubdomainIndex)
      {
        for(j = 0; j < Submesh->NStressPointsTriang1; j++)
          for(k = 0; k < E_NUM_STRESS_COMPONENTS; k++)
            Submesh->Stresses.StressesTriang1[n][j][k] =
              mesh->Stresses.StressesTriang1[i][j][k];
        n++;
      }
    if(n != Submesh->NElemsTriang1)
      e_GeneralError("Internal structure damaged !");
  }

  if (Submesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2])
  {
    n = 0;
    for (i = 0; i < mesh->NElemsTriang2; i++)
      if (mesh->Decomp.SubDomsTriang2[i] == SubdomainIndex)
      {
        for(j = 0; j < mesh->CompMatsTriang2[i]->NLayers; j++)
          for(k = 0; k < Submesh->NStressPointsTriang2; k++)
            for(l = 0; l < E_NUM_STRESS_COMPONENTS; l++)
              Submesh->Stresses.StressesTriang2[n][j][k][l] =
                mesh->Stresses.StressesTriang2[i][j][k][l];
        n++;
      }
    if(n != Submesh->NElemsTriang2)
      e_GeneralError("Internal structure damaged !");
  }

  if (Submesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3])
  {
    n = 0;
    for (i = 0; i < mesh->NElemsTriang3; i++)
      if (mesh->Decomp.SubDomsTriang3[i] == SubdomainIndex)
      {
        for(j = 0; j < Submesh->NStressPointsTriang3; j++)
          for(k = 0; k < E_NUM_STRESS_COMPONENTS; k++)
            Submesh->Stresses.StressesTriang3[n][j][k] =
              mesh->Stresses.StressesTriang3[i][j][k];
        n++;
      }
    if(n != Submesh->NElemsTriang3)
      e_GeneralError("Internal structure damaged !");
  }

  if (Submesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4])
  {
    n = 0;
    for (i = 0; i < mesh->NElemsTriang4; i++)
      if (mesh->Decomp.SubDomsTriang4[i] == SubdomainIndex)
      {
        for(j = 0; j < Submesh->NStressPointsTriang4; j++)
          for(k = 0; k < E_NUM_STRESS_COMPONENTS; k++)
            Submesh->Stresses.StressesTriang4[n][j][k] =
              mesh->Stresses.StressesTriang4[i][j][k];
        n++;
      }
    if(n != Submesh->NElemsTriang4)
      e_GeneralError("Internal structure damaged !");
  }

  if (Submesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5])
  {
    n = 0;
    for (i = 0; i < mesh->NElemsTriang5; i++)
      if (mesh->Decomp.SubDomsTriang5[i] == SubdomainIndex)
      {
        for(j = 0; j < Submesh->NStressPointsTriang5; j++)
          for(k = 0; k < E_NUM_STRESS_COMPONENTS; k++)
            Submesh->Stresses.StressesTriang5[n][j][k] =
              mesh->Stresses.StressesTriang5[i][j][k];
        n++;
      }
    if(n != Submesh->NElemsTriang5)
      e_GeneralError("Internal structure damaged !");
  }

  if (Submesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1])
  {
    n = 0;
    for (i = 0; i < mesh->NElemsQuad1; i++)
      if (mesh->Decomp.SubDomsQuad1[i] == SubdomainIndex)
      {
        for(j = 0; j < Submesh->NStressPointsQuad1; j++)
          for(k = 0; k < E_NUM_STRESS_COMPONENTS; k++)
            Submesh->Stresses.StressesQuad1[n][j][k] =
              mesh->Stresses.StressesQuad1[i][j][k];
        n++;
      }
    if(n != Submesh->NElemsQuad1)
      e_GeneralError("Internal structure damaged !");
  }

  if (Submesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3])
  {
    n = 0;
    for (i = 0; i < mesh->NElemsQuad3; i++)
      if (mesh->Decomp.SubDomsQuad3[i] == SubdomainIndex)
      {
        for(j = 0; j < mesh->CompMatsQuad3[i]->NLayers; j++)
          for(k = 0; k < Submesh->NStressPointsQuad3; k++)
            for(l = 0; l < E_NUM_STRESS_COMPONENTS; l++)
              Submesh->Stresses.StressesQuad3[n][j][k][l] =
                mesh->Stresses.StressesQuad3[i][j][k][l];
        n++;
      }
    if(n != Submesh->NElemsQuad3)
      e_GeneralError("Internal structure damaged !");
  }

  if (Submesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1])
  {
    n = 0;
    for (i = 0; i < mesh->NElemsTetrah1; i++)
      if (mesh->Decomp.SubDomsTetrah1[i] == SubdomainIndex)
      {
        for(j = 0; j < Submesh->NStressPointsTetrah1; j++)
          for(k = 0; k < E_NUM_STRESS_COMPONENTS; k++)
            Submesh->Stresses.StressesTetrah1[n][j][k] =
              mesh->Stresses.StressesTetrah1[i][j][k];
        n++;
      }
    if(n != Submesh->NElemsTetrah1)
      e_GeneralError("Internal structure damaged !");
  }

  if (Submesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1])
  {
    n = 0;
    for (i = 0; i < mesh->NElemsBlock1; i++)
      if (mesh->Decomp.SubDomsBlock1[i] == SubdomainIndex)
      {
        for(j = 0; j < Submesh->NStressPointsBlock1; j++)
          for(k = 0; k < E_NUM_STRESS_COMPONENTS; k++)
            Submesh->Stresses.StressesBlock1[n][j][k] =
              mesh->Stresses.StressesBlock1[i][j][k];
        n++;
      }
    if(n != Submesh->NElemsBlock1)
      e_GeneralError("Internal structure damaged !");
  }

}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetFEErrors: To set the sub-mesh finite element errors.                    */


static void SetFEErrors(MESH *mesh, MESH *SubMesh, int SubdomainIndex)
{
   int i,k;


   k = 0;

   for (i = 0; i < mesh->TotalNElems; i++)
   {
      if (mesh->Decomp.SubDoms[i] == SubdomainIndex)
      {
         SubMesh->FEError.FEErrors[k++]
           = mesh->FEError.FEErrors[i];
      }
   }

   SubMesh->FEError.FEErrorsLink1
      = SubMesh->FEError.FEErrors
        + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK1] - 1;
   SubMesh->FEError.FEErrorsLink2
      = SubMesh->FEError.FEErrors
        + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK2] - 1;
   SubMesh->FEError.FEErrorsLink3
      = SubMesh->FEError.FEErrors
        + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK3] - 1;
   SubMesh->FEError.FEErrorsLink4
      = SubMesh->FEError.FEErrors
        + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK4] - 1;
   SubMesh->FEError.FEErrorsLink5
      = SubMesh->FEError.FEErrors
        + SubMesh->FirstNetIndex[E_ELEM_TYPE_LINK5] - 1;
   SubMesh->FEError.FEErrorsTriang1
      = SubMesh->FEError.FEErrors
        + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] - 1;
   SubMesh->FEError.FEErrorsTriang2
      = SubMesh->FEError.FEErrors
        + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG2] - 1;
   SubMesh->FEError.FEErrorsTriang3
      = SubMesh->FEError.FEErrors
        + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] - 1;
   SubMesh->FEError.FEErrorsTriang4
      = SubMesh->FEError.FEErrors
        + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] - 1;
   SubMesh->FEError.FEErrorsTriang5
      = SubMesh->FEError.FEErrors
        + SubMesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] - 1;
   SubMesh->FEError.FEErrorsQuad1
      = SubMesh->FEError.FEErrors
        + SubMesh->FirstNetIndex[E_ELEM_TYPE_QUAD1] - 1;
   SubMesh->FEError.FEErrorsQuad3
      = SubMesh->FEError.FEErrors
        + SubMesh->FirstNetIndex[E_ELEM_TYPE_QUAD3] - 1;
   SubMesh->FEError.FEErrorsTetrah1
      = SubMesh->FEError.FEErrors
        + SubMesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1] - 1;
   SubMesh->FEError.FEErrorsBlock1
      = SubMesh->FEError.FEErrors
        + SubMesh->FirstNetIndex[E_ELEM_TYPE_BLOCK1] - 1;

}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetRemainingData: To set all remaining sub-mesh data. */

/* The return value is 1 if the data are set successfully, else 0. */


static int SetRemainingData(MESH *mesh, MESH *SubMesh, char *SubMeshTitle)
{
   int k;


   strcpy(SubMesh->Title,SubMeshTitle);

   SubMesh->TimeStep = mesh->TimeStep;
   SubMesh->NTimeSteps = mesh->NTimeSteps;
   SubMesh->NInternalTimeSteps = mesh->NInternalTimeSteps;
   SubMesh->NExternalTimeSteps = mesh->NExternalTimeSteps;
   SubMesh->DampingFactor = mesh->DampingFactor;
   SubMesh->Beta = mesh->Beta;
   SubMesh->Gamma = mesh->Gamma;

   SubMesh->NStressPointsLink1   = mesh->NStressPointsLink1;
   SubMesh->NStressPointsLink2   = mesh->NStressPointsLink2;
   SubMesh->NStressPointsLink3   = mesh->NStressPointsLink3;
   SubMesh->NStressPointsLink4   = mesh->NStressPointsLink4;
   SubMesh->NStressPointsLink5   = mesh->NStressPointsLink5;
   SubMesh->NStressPointsTriang1 = mesh->NStressPointsTriang1;
   SubMesh->NStressPointsTriang2 = mesh->NStressPointsTriang2;
   SubMesh->NStressPointsTriang3 = mesh->NStressPointsTriang3;
   SubMesh->NStressPointsTriang4 = mesh->NStressPointsTriang4;
   SubMesh->NStressPointsTriang5 = mesh->NStressPointsTriang5;
   SubMesh->NStressPointsQuad1   = mesh->NStressPointsQuad1;
   SubMesh->NStressPointsQuad3   = mesh->NStressPointsQuad3;
   SubMesh->NStressPointsTetrah1 = mesh->NStressPointsTetrah1;
   SubMesh->NStressPointsBlock1  = mesh->NStressPointsBlock1;

   for (k = 0; k < E_NKEYWORDS; k++)
      SubMesh->KeywordSet[k] = mesh->KeywordSet[k];

   if (SubMesh->NElemsLink1 == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NELEMS_LINK1] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NODES_LINK1] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_MATS_LINK1] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_LINK1] = FALSE;
   }

   if (SubMesh->NElemsLink2 == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NELEMS_LINK2] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NODES_LINK2] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_MATS_LINK2] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_LINK2] = FALSE;
   }

   if (SubMesh->NElemsLink3 == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NELEMS_LINK3] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NODES_LINK3] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_MATS_LINK3] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_LINK3] = FALSE;
   }

   if (SubMesh->NElemsLink4 == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NELEMS_LINK4] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NODES_LINK4] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_MATS_LINK4] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_LINK4] = FALSE;
   }

   if (SubMesh->NElemsLink5 == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NELEMS_LINK5] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NODES_LINK5] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_MATS_LINK5] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_LINK5] = FALSE;
   }

   if (SubMesh->NElemsTriang1 == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NODES_TRIANG1] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_MATS_TRIANG1] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_TRIANG1] = FALSE;
   }

   if (SubMesh->NElemsTriang2 == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG2] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NODES_TRIANG2] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_COMP_MATS_TRIANG2] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_TRIANG2] = FALSE;
   }

   if (SubMesh->NElemsTriang3 == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NODES_TRIANG3] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_MATS_TRIANG3] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_TRIANG3] = FALSE;
   }

   if (SubMesh->NElemsTriang4 == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NODES_TRIANG4] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_MATS_TRIANG4] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_TRIANG4] = FALSE;
   }

   if (SubMesh->NElemsTriang5 == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NODES_TRIANG5] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_MATS_TRIANG5] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_TRIANG5] = FALSE;
   }

   if (SubMesh->NElemsQuad1 == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NODES_QUAD1] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_MATS_QUAD1] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_QUAD1] = FALSE;
   }

   if (SubMesh->NElemsQuad3 == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NELEMS_QUAD3] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NODES_QUAD3] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_COMP_MATS_QUAD3] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_QUAD3] = FALSE;
   }

   if (SubMesh->NElemsTetrah1 == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NODES_TETRAH1] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_MATS_TETRAH1] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_TETRAH1] = FALSE;
   }

   if (SubMesh->NElemsBlock1 == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NELEMS_BLOCK1] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NODES_BLOCK1] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_MATS_BLOCK1] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_BLOCK1] = FALSE;
   }

   if (SubMesh->NBCNodes == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NBC_NODES] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_BCS] = FALSE;
   }

   if (SubMesh->NLoadedNodes == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NLOADED_NODES] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_LOADS] = FALSE;
   }

   if (SubMesh->NMats == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NMATS] = FALSE;
      SubMesh->KeywordSet[E_KW_NUM_NCOMP_MATS_TYPE1] = FALSE;
   }

   if (SubMesh->NCompMatsType1 == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NCOMP_MATS_TYPE1] = FALSE;
   }

   for (k = 0; k < E_NSUBSTRUCTS; k++)
      SubMesh->SubStructSet[k] = mesh->SubStructSet[k];

   if (SubMesh->Geom.NNurbsCurves == 0)
   {
      SubMesh->KeywordSet[E_KW_NUM_NNURBS_CURVES] = FALSE;
      SubMesh->SubStructSet[E_SS_NUM_GEOM] = FALSE;
   }

   SubMesh->SubStructSet[E_SS_NUM_DECOMP] = FALSE;

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetPropsIndex: To return the properties index corresponding to a
   material if the material has not already been assigned to the sub-mesh,
   else 0.                                                                    */


static int GetPropsIndex(int r, int NMaterials, int *PropsIndices)
{
   int i;


   for (i = 0; i < NMaterials; i++)
   {
      if (r == PropsIndices[i])
        return i + 1;
   }

   return 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* DeallocateAllMemory: To all deallocate memory. */
/* Does not deallocate the stresses because they are handled specially */

static void DeallocateAllMemory(MESH *SubMesh)
{
   int k;

   SubMesh->NMeshPoints = g_NMeshPoints;
   SubMesh->NNodes = g_NNodes;
   SubMesh->NElemsLink1 = g_NElemsLink1;
   SubMesh->NElemsLink2 = g_NElemsLink2;
   SubMesh->NElemsLink3 = g_NElemsLink3;
   SubMesh->NElemsLink4 = g_NElemsLink4;
   SubMesh->NElemsLink5 = g_NElemsLink5;
   SubMesh->NElemsTriang1 = g_NElemsTriang1;
   SubMesh->NElemsTriang2 = g_NElemsTriang2;
   SubMesh->NElemsTriang3 = g_NElemsTriang3;
   SubMesh->NElemsTriang4 = g_NElemsTriang4;
   SubMesh->NElemsTriang5 = g_NElemsTriang5;
   SubMesh->NElemsQuad1 = g_NElemsQuad1;
   SubMesh->NElemsQuad3 = g_NElemsQuad3;
   SubMesh->NElemsTetrah1 = g_NElemsTetrah1;
   SubMesh->NElemsBlock1 = g_NElemsBlock1;
   SubMesh->TotalNElems = g_TotalNElems;
   SubMesh->NBCNodes = g_NBCNodes;
   SubMesh->NLoadedNodes = g_NLoadedNodes;
   SubMesh->NMats = g_NMats;
   SubMesh->NMdfMats = g_NMdfMats;
   SubMesh->NCompMatsType1 = g_NCompMatsType1;
   SubMesh->TotalNodalDOF = g_TotalNodalDOF;

   SubMesh->Geom.NEndPoints = g_NEndPoints;
   SubMesh->Geom.NNurbsCurves = g_NNurbsCurves;

   e_free(SubMesh->Title);

   e_free_cont_d_matr(SubMesh->NMeshPoints,SubMesh->MeshPointCoords);
   e_free_cont_i_matr(SubMesh->NElemsLink1,SubMesh->NodesLink1);
   e_free_cont_i_matr(SubMesh->NElemsLink2,SubMesh->NodesLink2);
   e_free_cont_i_matr(SubMesh->NElemsLink3,SubMesh->NodesLink3);
   e_free_cont_i_matr(SubMesh->NElemsLink4,SubMesh->NodesLink4);
   e_free_cont_i_matr(SubMesh->NElemsLink5,SubMesh->NodesLink5);
   e_free_cont_i_matr(SubMesh->NElemsTriang1,SubMesh->NodesTriang1);
   e_free_cont_i_matr(SubMesh->NElemsTriang2,SubMesh->NodesTriang2);
   e_free_cont_i_matr(SubMesh->NElemsTriang3,SubMesh->NodesTriang3);
   e_free_cont_i_matr(SubMesh->NElemsTriang4,SubMesh->NodesTriang4);
   e_free_cont_i_matr(SubMesh->NElemsTriang5,SubMesh->NodesTriang5);
   e_free_cont_i_matr(SubMesh->NElemsQuad1,SubMesh->NodesQuad1);
   e_free_cont_i_matr(SubMesh->NElemsQuad3,SubMesh->NodesQuad3);
   e_free_cont_i_matr(SubMesh->NElemsTetrah1,SubMesh->NodesTetrah1);
   e_free_cont_i_matr(SubMesh->NElemsBlock1,SubMesh->NodesBlock1);

   e_free(SubMesh->NodalDOF);
   e_free(SubMesh->BCNodes);
   e_free_frag_i_matr(SubMesh->NBCNodes,SubMesh->BCTypes);
   e_free_frag_d_matr(SubMesh->NBCNodes,SubMesh->BCDispls);
   e_free_frag_d_matr(SubMesh->NBCNodes,SubMesh->BCSpringConsts);
   e_free(SubMesh->LoadedNodes);
   e_free_frag_d_matr(SubMesh->NLoadedNodes,SubMesh->Loads);
   e_free(SubMesh->Mats);

   if (SubMesh->Props != NULL)
   {
      for (k = 0; k < SubMesh->NMats; k++)
         e_free(SubMesh->Props[k].Name);
   }

   if (SubMesh->CompPropsType1 != NULL)
   {
      for (k = 0; k < SubMesh->NCompMatsType1; k++)
      {
         e_free(SubMesh->CompPropsType1[k].Name);
         e_free(SubMesh->CompPropsType1[k].LayerMats);
         e_free(SubMesh->CompPropsType1[k].Thicknesses);
      }
   }

   e_free(SubMesh->Props);
   e_free(SubMesh->CompPropsType1);
   e_free(SubMesh->NodeTypes);
   e_free(SubMesh->NNurbsCurves);
   e_free_frag_d_matr(SubMesh->NMeshPoints,SubMesh->NurbsCurveParams);

   if (SubMesh->NurbsCurves != NULL)
   {
      for (k = 0; k < SubMesh->NMeshPoints; k++)
         e_free(SubMesh->NurbsCurves[k]);
   }

   e_free(SubMesh->NurbsCurves);
   e_free(SubMesh->GNIs);
   e_free(SubMesh->GEIs);

   if (SubMesh->Geom.NurbsCurves != NULL)
   {
      for (k = 0; k < SubMesh->Geom.NNurbsCurves; k++)
      {
         e_free(SubMesh->Geom.NurbsCurves[k].Name);
         e_free_cont_d_matr(SubMesh->Geom.NurbsCurves[k].Degree + 1,
                            SubMesh->Geom.NurbsCurves[k].CntlPoints );
         e_free(SubMesh->Geom.NurbsCurves[k].BCTypes);
         e_free(SubMesh->Geom.NurbsCurves[k].BCDispls);
         e_free(SubMesh->Geom.NurbsCurves[k].BCSpringConsts);
         e_free(SubMesh->Geom.NurbsCurves[k].PointLoad);
         e_free(SubMesh->Geom.NurbsCurves[k].LinLoadGCBeg);
         e_free(SubMesh->Geom.NurbsCurves[k].LinLoadGCEnd);
         e_free(SubMesh->Geom.NurbsCurves[k].LinLoadCCBeg);
         e_free(SubMesh->Geom.NurbsCurves[k].LinLoadCCEnd);
      }
   }

   e_free(SubMesh->Geom.NurbsCurves);
   e_free_cont_d_matr(SubMesh->Geom.NEndPoints,SubMesh->Geom.EndPointCoords);

   e_free(SubMesh->MeshParam.NodalMeshParams);
   e_free(SubMesh->MeshParam.ElemMeshParams);
   e_free(SubMesh->FEError.FEErrors);


}
