/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                             e_CreateSubMesh()                              */
/*                                Header File                                 */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */


static int AllocateMemory(MESH *mesh, MESH **SubMesh, char* SubMeshTitle,
                                                       int SubdomainIndex);
static int CountSubMesh(MESH *mesh, MESH *SubMesh, int SubdomainIndex);
static int CreateMesh(MESH *mesh, MESH *SubMesh, char *SubMeshTitle,
                                                  int SubdomainIndex);
static int GetPropsIndex(int r, int NMaterials, int *PropsIndices);
static int Initialise(MESH *mesh, MESH **SubMesh, char* SubMeshTitle,
                                                   int SubdomainIndex);


static void CountBCNodes(MESH *mesh, MESH *SubMesh, int NNodesPerElement,
                         int NMeshElements, int **MeshNodes,
                         int FirstNetIndex, int *BCNumbers,
                         char *NodeStored, int SubdomainIndex            );
static void CountLoadedNodes(MESH *mesh, MESH *SubMesh, int NNodesPerElement,
                             int NMeshElements, int **MeshNodes,
                             int FirstNetIndex, int *LoadedNodeNumbers,
                             char *NodeStored, int SubdomainIndex            );
static void CountMaterials(MESH *mesh, MESH *SubMesh, int NMeshElements,
                           int FirstNetIndex, int *PropsIndices,
                           int SubdomainIndex                           );
static void CountMatsType1(MESH *mesh, MESH *SubMesh, int NMeshElements,
                           int FirstNetIndex, int *PropsIndices,
                           int *CompPropsIndicesType1, int SubdomainIndex);
static void CountNodesAndElements
               (MESH *mesh, MESH *SubMesh, int NNodesPerElement,
                int NVerticesPerElement, int NMeshElements,
                int **MeshNodes, int FirstNetIndex,
                int *NSubMeshElements, char *NodeStored,
                int SubdomainIndex                              );
static void CountNurbsCurves(MESH *mesh, MESH *SubMesh,
                             int NVerticesPerElement, int NMeshElements,
                             int **MeshNodes, int FirstNetIndex,
                             char *EndPointStored, char *NurbsCurveStored,
                             int SubdomainIndex                           );
static void DeallocateAllMemory(MESH *SubMesh);
static void SetElementMaterials(MESH *SubMesh);
static void SetMeshParams(MESH *mesh, MESH *SubMesh, int *NewNodeIndex,
                                                     int SubdomainIndex);
static void SetFEErrors(MESH *mesh, MESH *SubMesh, int SubdomainIndex);
static void SetFirstNetIndices(MESH *SubMesh);
static void SetGlobalElementIndices(MESH *SubMesh);
static void SetMeshPointBCs
               (MESH *mesh, MESH *SubMesh, int NVerticesPerElement,
                int NMeshElements, int **MeshNodes, int FirstNetIndex,
                int *BCNumbers, char *NodeStored, int SubdomainIndex  );
static void SetMeshPointLoads
               (MESH *mesh, MESH *SubMesh, int NVerticesPerElement,
                int NMeshElements, int **MeshNodes, int FirstNetIndex,
                int *LoadedNodeNumbers, char *NodeStored,
                int SubdomainIndex                                    );
static int SetMaterialProperties(MESH *mesh, MESH *SubMesh,
                                 int *PropsIndices,
                                 int *CompPropsIndicesType1,
                                 int *MeshToSubMeshPropsIndices);
static int SetMaterials(MESH *mesh, MESH *SubMesh, int NMeshElements,
                        int FirstNetIndex, int *PropsIndices,
                        int *MeshToSubMeshPropsIndices, int SubdomainIndex);
static int SetMatsType1(MESH *mesh, MESH *SubMesh, int NMeshElements,
                        int FirstNetIndex, int *PropsIndices,
                        int *CompPropsIndicesType1,
                        int *MeshToSubMeshPropsIndices, int SubdomainIndex);
static int SetNurbsCurves(MESH *mesh, MESH *SubMesh, int NVerticesPerElement,
                          int NMeshElements, int **MeshNodes,
                          int FirstNetIndex, int *EndPoints,
                          int *NurbsCurveIndices, char *NodeStored,
                          int SubdomainIndex                                 );
static int SetRemainingData(MESH *mesh, MESH *SubMesh, char *SubMeshTitle);

static void SetMeshPointNodesAndElements
               (MESH *mesh, MESH *SubMesh, int NVerticesPerElement,
                int NMeshElements, int **MeshNodes, int FirstNetIndex,
                int *NSubMeshElements, int **SubMeshNodes, int *NewNodeIndex,
                char *NodeStored, int SubdomainIndex                         );
static void SetNodalDOF(MESH *SubMesh);
static void SetNonMeshPointBCs
               (MESH *mesh, MESH *SubMesh, int NNodesPerElement,
                int NVerticesPerElement, int NMeshElements, int **MeshNodes,
                int FirstNetIndex, int *BCNumbers, char *NodeStored,
                int SubdomainIndex                                          );
static void SetNonMeshPointLoads
               (MESH *mesh, MESH *SubMesh, int NNodesPerElement,
                int NVerticesPerElement, int NMeshElements, int **MeshNodes,
                int FirstNetIndex, int *LoadedNodeNumbers, char *NodeStored,
                int SubdomainIndex                                          );
static void SetNonMeshPointNodes
               (MESH *mesh, MESH *SubMesh, int NNodesPerElement,
                int NVerticesPerElement, int NMeshElements, int **MeshNodes,
                int FirstNetIndex, int **SubMeshNodes, int *NewNodeIndex,
                char *NodeStored, int SubdomainIndex                        );
static void SetStresses(MESH *mesh, MESH *Submesh, int SubdomainIndex);
static void ZeroSubMesh(MESH *SubMesh);


/* ////////////////////////////////////////////////////////////////////////// */

/* NOTE: Certain integer members of the MESH structure are temporarily set to
         zero after memory has been allocated. To ensure that all memory is
         freed in the event of failure, the following global variables have
         been defined. Their values are set in the function "CountSubMesh"
         and used in the function "DeallocateAllMemory".                      */


static int g_NMeshPoints;
static int g_NNodes;
static int g_NElemsLink1;
static int g_NElemsLink2;
static int g_NElemsLink3;
static int g_NElemsLink4;
static int g_NElemsLink5;
static int g_NElemsTriang1;
static int g_NElemsTriang2;
static int g_NElemsTriang3;
static int g_NElemsTriang4;
static int g_NElemsTriang5;
static int g_NElemsQuad1;
static int g_NElemsQuad3;
static int g_NElemsTetrah1;
static int g_NElemsBlock1;
static int g_TotalNElems;
static int g_NBCNodes;
static int g_NLoadedNodes;
static int g_NMats;
static int g_NMdfMats;
static int g_NCompMatsType1;
static int g_TotalNodalDOF;

static int g_NEndPoints;
static int g_NNurbsCurves;
