#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "e_lib.h"


#define NO_COLOR         8
#define DEFAULT_SCALE    2.0
#define DEFAULT_PERCENT  80.0
#define DEFAULT_MARGIN   0.0

/* subdomain plotting style */
#define SUBDOM_COLOR     1
#define SUBDOM_BOUNDARY  2
#define SUBDOM_NUMBER    4

/* possibility for the plot type */

#define EPS_PLAINMESH   1
#define EPS_SUBDOMAIN   2
#define EPS_ERROR       3
#define EPS_STRESS      4


/* user parameters */
typedef struct 
{
  FILE   *fp;              /* file to write into */
  char    filename[E_LINE_MAX]; /* filename */

  int     plot_type;       /* 2 = 2D, 3 = 3D */
  int     hidden_line;
  double  azimuth;
  double  elevation;
  double  distance;
  double  scale;           /* scaling factor */
  double  proj_center[3];

  int     type;            /* plot type, plain, subdom, error, stress */
  double  picture_size_x;
  double  picture_size_y;
  double  margin;          /* margin around the picture */
  int     border;          /* not for plain mesh, show element boundary, [1/0] */
  int     bcolor;          /* if show element boundary, black or white color [1/0] */
  int     color;           /* color or grey scale picture [1/0]*/
  int     tile;            /* whether tile elements or not [1/0] */
  double  percent;         /* for tiling the percentage [0-100] */

  int     subdom_style;    /* whether to plot subdom boundary and/or color it and/or number it */
  double  subdom_b_width;  /* width of the line to draw the boundary around */
  double  subdom_t_size;   /* size of the subdom number */

  double  color_correction; /* color correction factor for errors and stresses [0.0 - 1.0]*/

  int     user_value;       /* user specified value for error or stresses [1/0] */

  double  max_error;        /* max error and above to plot as red */

  int     stress_type;  /* what kind of stress to plot, 1 - 6*/
  int     stress_point; /* which stress point to plot, 1 - n */
  int     stress_layer; /* 1 - n */
  double  MaxStrIntensity; /* percentage correction to plot 0-100 */
  double  MaxTotalStress;  /* max stresses to scale others with */
}
OPTIONS;



typedef struct
{
  double normal[3];
  double cnst;
  short  info;

  int    priority;
}
PLANE;


typedef struct
{
  int         index;
  int         no_vertices;
  double    **world_coords;
  double    **screen_coords;
  int         elem_type;     /* elib values */
  int         mesh_index;    /* 0 -> n-1 */

  double      avsl;          /* for Jelle */
  double      z;
  int         vis;

  short       obscured;
  double      box_min[3];
  double      box_max[3];
  double      center[3];
  double      max_rad;

  short       is_in_front;
  short       is_in_back;

  PLANE      *sep_plane;
  PLANE       local_plane;

}
POLYGON;


/* ----------------------------------------------------------- */

void setup_polygons(MESH *mesh, POLYGON **polygons, int *NPoly);
void free_polygons(int NPoly, POLYGON *polygons);

void setup_hidden_line1(int NPoly, POLYGON *pool);
void setup_hidden_line2(int NPoly, POLYGON *pool);
void setup_proj_center(double azim, double elev, double dist, double *target);
int  sort_objects(int *order, int n, POLYGON *poly, int plane);


void p_3Dworld_to_screen(double *result, double *world);

/* ----------------------------------------------------------- */

int e_Write_EPS_File(MESH *mesh, char *filename, OPTIONS *options, 
                     int NPoly, POLYGON *Polygons, int *Order, int error);

/* ----------------------------------------------------------- */

int draw_polygon(MESH *mesh, POLYGON polygon, OPTIONS options, int error);
int Subdom_Boundary(MESH *mesh, OPTIONS *options);
void world_to_screen(double *screen, double *world);

