/* File: 3D.C  */



#include<stdlib.h>
#include<stdio.h>
#include<math.h>


#include"j_3d.h"


#define EXIT_FAIL 1
#define EPSILON 1.0e-7




void j_3Dinverse_rot_matrix(InvM, M)
j_3Drot_matrix InvM, M;
{
register short i, j;

for(i = 0; i < 3; i++)
    for(j = 0; j < 3; j++)
        InvM[i][j] = M[j][i];
}



void j_3Dmatrix_mult(C, A, B)
j_3Drot_matrix C, A, B;
{
register double *b = (double *) B;
register j_3Dvector *a = (j_3Dvector *) A,
                *c = (j_3Dvector *) C,
                *hi_a = a + 3;

for(; a < hi_a; a++, c++)
    M_Vec_dot_matrix(*c, *a, b);
}





void j_3Dfill_rot_matrix(M, angle, axis)
j_3Drot_matrix M;
double angle;
short axis;
{
register double s, c;  /* for reasons of speed */

s = sin(angle); c = cos(angle);

switch(axis)
    {
    case XDIR: M[0][0] = 1; M[0][1] =  0; M[0][2] = 0;
               M[1][0] = 0; M[1][1] =  c; M[1][2] = s;
               M[2][0] = 0; M[2][1] = -s; M[2][2] = c; return;

    case YDIR: M[0][0] = c; M[0][1] = 0; M[0][2] = -s;
               M[1][0] = 0; M[1][1] = 1; M[1][2] =  0;
               M[2][0] = s; M[2][1] = 0; M[2][2] =  c; return;

    case ZDIR: M[0][0] =  c; M[0][1] = s; M[0][2] = 0;
               M[1][0] = -s; M[1][1] = c; M[1][2] = 0;
               M[2][0] =  0; M[2][1] = 0; M[2][2] = 1; return;
    }
}






void j_3Dcalc_rot_matrix(void)

{
j_3Dspherical_coords(&G_dTC,&G_Azim,&G_Elev,G_Center,G_Target);
j_3Drecalc_rot_matrix();
}



void j_3Drecalc_rot_matrix(void)

{
double alpha, beta, gamma; /* The Euler angles */
j_3Drot_matrix z_alpha, x_beta, z_gamma, xz;

alpha = G_Twist;
beta  = M_PI / 2.00 - G_Elev;
gamma = G_Azim;

j_3Dfill_rot_matrix(z_alpha, alpha, ZDIR);
j_3Dfill_rot_matrix(x_beta,  beta,  XDIR);
j_3Dfill_rot_matrix(z_gamma, gamma, ZDIR);
j_3Dmatrix_mult(xz, z_alpha, x_beta);
j_3Dmatrix_mult(G_cpim, xz, z_gamma);
j_3Dinverse_rot_matrix(G_cpm,G_cpim);
}






void j_3Dspherical_coords(Dist,Azim,Elev,cpoint,tpoint)
double *Dist;
double *Azim, *Elev;
j_3Dvector cpoint,tpoint;
{
j_3Dvector newpoint;

M_Subtr_vec(newpoint,cpoint,tpoint);
*Dist = M_Length(newpoint);
*Azim = j_offset(newpoint[XDIR], newpoint[YDIR]);
*Elev = asin((double) newpoint[ZDIR] / (*Dist + EPSILON));
}




void j_3Dworld_to_screen(screen,world)

j_3Dvector screen,world;
{
j_3Dvector world_t;
double lambda;


M_Subtr_vec(world_t,world,G_Target);
j_3Dvec_mult_matrix(screen,world_t,G_cpm);
lambda = G_scale * G_dTC / (G_dTC - screen[ZDIR]);
screen[XDIR] = G_Xscrorigin + lambda * screen[XDIR];
screen[YDIR] = G_Yscrorigin + G_pixel_ratio * lambda * screen[YDIR];
}

void p_3Dworld_to_screen(double *result,double *world)
{
	j_3Dvector world_t;
	
    M_Subtr_vec(world_t,world,G_Target);
	j_3Dvec_mult_matrix(result,world_t,G_cpm);
}


void j_3Dvec_mult_matrix(result, p, A)

j_3Dvector result, p;
j_3Drot_matrix A;
{
register double *a = (double *)A, *r = (double *) result, *hi_r = r + 3;

for(; r < hi_r; r++, a++)
   *r = p[XDIR] * *(a) + p[YDIR] * *(a+ 3) + p[ZDIR] * *(a+ 6);
}



double  j_offset(double dx, double dy)

{
double  angle;

if(fabs(dx) < EPSILON)
    if(fabs(dy) < EPSILON)
         /* point on the z-axis  */
        angle = 0.00;
/*        j_error_exit("Co-ordinates are the same in offset()",EXIT_FAIL);  */
    else
        if(dy > 0.0)
            angle = M_PI / 2.0;
        else
            angle = -M_PI / 2.0;
	else
    if(dx > 0.0)
        angle = atan(dy/dx);
    else
        angle = atan(dy/dx) + M_PI;

 	angle += M_PI/2;
return(angle);
}     /* end of function offset()  */





/* End Of File:  3D.C    */