#include "eps.h"


static void setup_2Delements(int *n, POLYGON *pool, 
                             int n_elem, int no_vertices, int elem_type,
                             int **nodes, double **coords)
{
  int i, j, k;
  int index;

  index = *n;

  for(i = 0; i < n_elem; i++)
  {
    pool[index].index = index;
    pool[index].no_vertices = no_vertices;
    pool[index].world_coords  = e_alloc_cont_d_matr(no_vertices, 3);
    pool[index].screen_coords = e_alloc_cont_d_matr(no_vertices, 3);
    pool[index].elem_type = elem_type;
    pool[index].mesh_index = i;

    pool[index].sep_plane = NULL;

    pool[index].center[0] = 0.0;
    pool[index].center[1] = 0.0;
    pool[index].center[2] = 0.0;
    for(j = 0; j < no_vertices; j++)
    {
      for(k = 0; k < 3; k++)
      {
        pool[index].world_coords[j][k] = coords[nodes[i][j]-1][k];
        pool[index].center[k] += pool[index].world_coords[j][k];
      }
    }
    pool[index].center[0] /= pool[index].no_vertices;
    pool[index].center[1] /= pool[index].no_vertices;
    pool[index].center[2] /= pool[index].no_vertices;

    index++;
  }

  *n = index;
}


static void setup_tetrah_elements(int *n, POLYGON *pool, 
                                  int n_elem, int elem_type,
                                  int **nodes, double **coords)
{
  int i, j, k, ii;
  int index;
  int tris[4][3] = { 0,1,2,  1,2,3,  0,1,3,  2,0,3 };

  index = *n;

  for(i = 0; i < n_elem; i++)
  {
    for(ii = 0; ii < 4; ii++)
    {
      pool[index].index = index;
      pool[index].no_vertices = 3;
      pool[index].world_coords = e_alloc_cont_d_matr(3, 3);
      pool[index].screen_coords = e_alloc_cont_d_matr(3, 3);
      pool[index].elem_type = elem_type;
      pool[index].mesh_index = i;

      pool[index].sep_plane = NULL;

      pool[index].center[0] = 0.0;
      pool[index].center[1] = 0.0;
      pool[index].center[2] = 0.0;
      for(j = 0; j < 3; j++)
      {
        for(k = 0; k < 3; k++)
        {
          pool[index].world_coords[j][k] = coords[nodes[i][tris[ii][j]]-1][k];
          pool[index].center[k] += pool[index].world_coords[j][k];
        }
      }
      pool[index].center[0] /= pool[index].no_vertices;
      pool[index].center[1] /= pool[index].no_vertices;
      pool[index].center[2] /= pool[index].no_vertices;

      index++;
    }
  }

  *n = index;
}


static void setup_block_elements(int *n, POLYGON *pool, 
                                 int n_elem, int elem_type,
                                 int **nodes, double **coords)
{
  int i, j, k, ii;
  int index;
  int quads[6][4] = { 0,1,2,3,  0,1,5,4,  1,2,6,5,
                      2,3,7,6,  3,0,4,7,  4,5,6,7  };

  index = *n;

  for(i = 0; i < n_elem; i++)
  {
    for(ii = 0; ii < 6; ii++)
    {
      pool[index].index = index;
      pool[index].no_vertices = 4;
      pool[index].world_coords = e_alloc_cont_d_matr(4, 3);
      pool[index].screen_coords = e_alloc_cont_d_matr(4, 3);
      pool[index].elem_type = elem_type;
      pool[index].mesh_index = i;

      pool[index].sep_plane = NULL;

      pool[index].center[0] = 0.0;
      pool[index].center[1] = 0.0;
      pool[index].center[2] = 0.0;
      for(j = 0; j < 4; j++)
      {
        for(k = 0; k < 3; k++)
        {
          pool[index].world_coords[j][k] = coords[nodes[i][quads[ii][j]]-1][k];
          pool[index].center[k] += pool[index].world_coords[j][k];
        }
      }
      pool[index].center[0] /= pool[index].no_vertices;
      pool[index].center[1] /= pool[index].no_vertices;
      pool[index].center[2] /= pool[index].no_vertices;

      index++;
    }
  }

  *n = index;
}


void setup_polygons(MESH *mesh, POLYGON **polygons, int *NPoly)
{
  int n;

  n = mesh->NElemsLink1   + mesh->NElemsLink2 +
      mesh->NElemsLink3   + mesh->NElemsLink4 +
      mesh->NElemsLink5   + mesh->NElemsTriang1 +
      mesh->NElemsTriang2 + mesh->NElemsTriang3 +
      mesh->NElemsTriang4 + mesh->NElemsTriang5 +
      mesh->NElemsQuad1   + mesh->NElemsQuad3 +
      mesh->NElemsTetrah1 * 4 +
      mesh->NElemsBlock1  * 6;

  *polygons = e_calloc(n, sizeof(POLYGON));

  *NPoly = 0;

  /* 2D elements */

  setup_2Delements(NPoly, *polygons, mesh->NElemsTriang1, E_NVPE_TRIANG1, E_ELEM_TYPE_TRIANG1,
                   mesh->NodesTriang1, mesh->MeshPointCoords);

  setup_2Delements(NPoly, *polygons, mesh->NElemsTriang2, E_NVPE_TRIANG2, E_ELEM_TYPE_TRIANG2,
                   mesh->NodesTriang2, mesh->MeshPointCoords);

  setup_2Delements(NPoly, *polygons, mesh->NElemsTriang3, E_NVPE_TRIANG3, E_ELEM_TYPE_TRIANG3,
                   mesh->NodesTriang3, mesh->MeshPointCoords);

  setup_2Delements(NPoly, *polygons, mesh->NElemsTriang4, E_NVPE_TRIANG4, E_ELEM_TYPE_TRIANG4,
                   mesh->NodesTriang4, mesh->MeshPointCoords);

  setup_2Delements(NPoly, *polygons, mesh->NElemsTriang5, E_NVPE_TRIANG5, E_ELEM_TYPE_TRIANG5,
                   mesh->NodesTriang5, mesh->MeshPointCoords);


  setup_2Delements(NPoly, *polygons, mesh->NElemsQuad1, E_NVPE_QUAD1, E_ELEM_TYPE_QUAD1,
                   mesh->NodesQuad1, mesh->MeshPointCoords);

  setup_2Delements(NPoly, *polygons, mesh->NElemsQuad3, E_NVPE_QUAD3, E_ELEM_TYPE_QUAD3,
                   mesh->NodesQuad3, mesh->MeshPointCoords);

  /* 3D elements are divided up into faces */

  setup_tetrah_elements(NPoly, *polygons, mesh->NElemsTetrah1, E_ELEM_TYPE_TETRAH1,
                        mesh->NodesTetrah1, mesh->MeshPointCoords);

  setup_block_elements(NPoly, *polygons, mesh->NElemsBlock1, E_ELEM_TYPE_BLOCK1,
                       mesh->NodesBlock1, mesh->MeshPointCoords);

  /* 1D elements are last */

  setup_2Delements(NPoly, *polygons, mesh->NElemsLink1, E_NVPE_LINK1, E_ELEM_TYPE_LINK1,
                   mesh->NodesLink1, mesh->MeshPointCoords);

  setup_2Delements(NPoly, *polygons, mesh->NElemsLink2, E_NVPE_LINK2, E_ELEM_TYPE_LINK2,
                   mesh->NodesLink2, mesh->MeshPointCoords);

  setup_2Delements(NPoly, *polygons, mesh->NElemsLink3, E_NVPE_LINK3, E_ELEM_TYPE_LINK3,
                   mesh->NodesLink3, mesh->MeshPointCoords);

  setup_2Delements(NPoly, *polygons, mesh->NElemsLink4, E_NVPE_LINK4, E_ELEM_TYPE_LINK4,
                   mesh->NodesLink4, mesh->MeshPointCoords);

  setup_2Delements(NPoly, *polygons, mesh->NElemsLink5, E_NVPE_LINK5, E_ELEM_TYPE_LINK5,
                   mesh->NodesLink5, mesh->MeshPointCoords);


}



void free_polygons(int NPoly, POLYGON *polygons)
{
  int i;

  for(i = 0; i < NPoly; i++)
  {
    e_free_cont_d_matr(polygons[i].no_vertices, polygons[i].world_coords);
    e_free_cont_d_matr(polygons[i].no_vertices, polygons[i].screen_coords);

    if(polygons[i].sep_plane != NULL)
      e_free(polygons[i].sep_plane);
  }

  e_free(polygons);

}
